@@ -10,7 +10,7 @@
 
 ;;; Code:
 
-(defvar agda2-version "2.3.2.2"
+(defvar agda2-version "2.4.0.2"
   "The version of the Agda mode.
 Note that the same version of the Agda executable must be used.")
 
@@ -54,6 +54,9 @@ properties to add to the result."
 (defmacro agda2-let (varbind funcbind &rest body)
   "Expands to (let* VARBIND (cl-labels FUNCBIND BODY...)).
 Or possibly (let* VARBIND (labels FUNCBIND BODY...))."
+  (declare (debug ((&rest [&or symbolp (symbolp form)])
+                   (&rest (cl-defun))
+                   body)))
   `(let* ,varbind (cl-labels ,funcbind ,@body)))
 (put 'agda2-let 'lisp-indent-function 2)
 
@@ -103,6 +106,7 @@ argument, and does not need to be listed here."
   "MAlonzo"
   "The backend which is used to compile Agda programs."
   :type '(choice (const "MAlonzo")
+                 (const "MAlonzoNoMain")
                  (const "Epic")
                  (const "JS"))
   :group 'agda2)
@@ -238,7 +242,9 @@ constituents.")
     (agda2-make-case                         "\C-c\C-c"           (local)        "Case")
     (agda2-goal-type                         "\C-c\C-t"           (local)        "Goal type")
     (agda2-show-context                      "\C-c\C-e"           (local)        "Context (environment)")
+    (agda2-helper-function-type              "\C-c\C-h"           (local)        "Helper function type")
     (agda2-infer-type-maybe-toplevel         "\C-c\C-d"           (local global) "Infer (deduce) type")
+    (agda2-why-in-scope-maybe-toplevel       "\C-c\C-w"           (local global) "Explain why a particular name is in scope")
     (agda2-goal-and-context                  ,(kbd "C-c C-,")     (local)        "Goal type and context")
     (agda2-goal-and-context-and-inferred     ,(kbd "C-c C-.")     (local)        "Goal type, context and inferred type")
     (agda2-module-contents-maybe-toplevel    ,(kbd "C-c C-o")     (local global) "Module contents")
@@ -511,7 +517,7 @@ exist, then an attempt is made to restart the process."
     (error "Another command is currently in progress
 \(if a command has been aborted you may want to restart Agda)")))
 
-(defun agda2-go (responses-expected highlight &rest args)
+(defun agda2-go (responses-expected highlight do-abort &rest args)
   "Executes commands in the Agda2 interpreter.
 Sends the list of strings ARGS to the Agda2 interpreter, waits for
 output and executes the responses, if any. If no responses are
@@ -519,7 +525,7 @@ received, and RESPONSES-EXPECTED is non-nil, then an error is
 raised. If HIGHLIGHT is non-nil, then the buffer's syntax
 highlighting may be updated."
 
-  (agda2-abort-if-in-progress)
+  (if do-abort (agda2-abort-if-in-progress))
 
   (setq agda2-in-progress             t
         agda2-highlight-in-progress   highlight
@@ -603,12 +609,14 @@ reloaded from `agda2-highlighting-file', unless
               (when cmd
                 (unless (and (symbolp (car cmd))
                              (let ((case-fold-search nil))
-                             (string-match "^agda2-highlight-"
-                                           (symbol-name (car cmd)))))
+                               (string-match "^agda2-highlight-"
+                                             (symbol-name (car cmd)))))
                     ;; Do not echo highlighting commands.
                   (with-current-buffer agda2-process-buffer
-                    (insert line)
-                    (insert "\n"))
+                    (save-excursion
+                      (goto-char (point-max))
+                      (insert line)
+                      (insert "\n")))
                   (incf agda2-responses))
                 (if (equal 'last (car-safe (car cmd)))
                     (push (cons (cdr (car cmd)) (cdr cmd))
@@ -709,7 +717,7 @@ An error is raised if no responses are received."
                   (or ask (string-match "\\`\\s *\\'" txt)))
              (setq txt (read-string (concat want ": ") nil nil txt t)))
             (t (setq input-from-goal t)))
-      (apply 'agda2-go t nil cmd
+      (apply 'agda2-go t t t cmd
              (format "%d" g)
              (if input-from-goal (agda2-goal-Range o) "noRange")
              (agda2-string-quote txt) args))))
@@ -731,7 +739,7 @@ An error is raised if no responses are received."
 (defun agda2-load ()
   "Load current buffer."
   (interactive)
-  (agda2-go t t "Cmd_load"
+  (agda2-go t t t "Cmd_load"
             (agda2-string-quote (buffer-file-name))
             (agda2-list-quote agda2-include-dirs)
             ))
@@ -763,7 +771,7 @@ resulting time (represented as a string)."
 
 The variable `agda2-backend' determines which backend is used."
   (interactive)
-  (agda2-go t t "Cmd_compile"
+  (agda2-go t t t "Cmd_compile"
             agda2-backend
             (agda2-string-quote (buffer-file-name))
             (agda2-list-quote agda2-include-dirs)
@@ -775,7 +783,11 @@ The variable `agda2-backend' determines which backend is used."
 
 (defun agda2-give-action (old-g paren)
   "Update the goal OLD-G with the expression in it."
-  (agda2-update old-g paren))
+  (let
+     ;; Don't run modification hooks: we don't want this to
+      ;; trigger agda2-abort-highlighting.
+      ((inhibit-modification-hooks t))
+  (agda2-update old-g paren)))
 
 (defun agda2-refine (pmlambda)
   "Refine the goal at point.
@@ -830,7 +842,7 @@ Assumes that <clause> = {!<variables>!} is on one line."
 	 (bracketCount 0)
 	 cl)
     (re-search-backward "{!")
-      (while (and (not (equal (preceding-char) ?;)) (>= bracketCount 0))
+      (while (and (not (equal (preceding-char) ?\;)) (>= bracketCount 0))
 	(backward-char)
 	(if (equal (preceding-char) ?}) (incf bracketCount))
 	(if (equal (preceding-char) ?{) (decf bracketCount)))
@@ -888,6 +900,16 @@ If APPEND is nil, then any previous text is removed before TEXT
 is inserted, and point is placed before this text."
   (interactive)
   (with-current-buffer (agda2-info-buffer)
+    ;; In some cases the jump-to-position-mentioned-in-text
+    ;; functionality (see compilation-error-regexp-alist above) didn't
+    ;; work: Emacs jumped to the wrong position. However, it seems to
+    ;; work if compilation-forget-errors is used. This problem may be
+    ;; related to Emacs bug #9679
+    ;; (http://debbugs.gnu.org/cgi/bugreport.cgi?bug=9679). The idea
+    ;; to use compilation-forget-errors comes from a comment due to
+    ;; Oleksandr Manzyuk
+    ;; (https://github.com/haskell/haskell-mode/issues/67).
+    (compilation-forget-errors)
     (unless append (erase-buffer))
     (save-excursion
       (goto-char (point-max))
@@ -898,22 +920,41 @@ is inserted, and point is placed before this text."
       (let (;; If there is only one window, then the info window
             ;; should be created above or below the code window, not
             ;; to the left or right.
-            (split-width-threshold nil))
-        (pop-to-buffer (current-buffer) nil 'norecord)
-        (fit-window-to-buffer
-         nil (truncate
+            (split-width-threshold nil)
+            (buf (current-buffer))
+           )
+        ;; Andreas, 2014-02-23, issue 1061
+        ;; If the buffer is already displayed in some window,
+        ;; do not display it again.  Also, do not raise its frame.
+        ;; This allows undisturbed working on something else while
+        ;; Agda is type-checking.
+        ;; The solution is to query for the window displaying the buffer
+        ;; via get-buffer-window.  Only if it return nil, we pop-to-buffer.
+        ;; Credits go to Iqbal Ansari who anwered my question on
+        ;; http://stackoverflow.com/questions/21955162/emacs-how-to-display-a-buffer-without-switching-window-and-without-raising-fram
+        ;; 2014-03-01 DISABLED FIX
+        ;; (unless (get-buffer-window buf t)
+          (pop-to-buffer buf nil 'norecord)
+          (fit-window-to-buffer nil
+            (truncate
               (* (frame-height) agda2-information-window-max-height)))
+        ;; )
         (if append
             (goto-char (point-max))
           (goto-char (point-min)))))))
 
+(defun agda2-info-action-and-copy (name text &optional append)
+  "Same as agda2-info-action but also puts TEXT in the kill ring."
+  (kill-new text)
+  (agda2-info-action name text append))
+
 (defun agda2-show-goals()
   "Show all goals." (interactive)
-  (agda2-go t t "Cmd_metas"))
+  (agda2-go t t t "Cmd_metas"))
 
 (defun agda2-show-constraints()
   "Show constraints." (interactive)
-  (agda2-go t t "Cmd_constraints"))
+  (agda2-go t t t "Cmd_constraints"))
 
 (defun agda2-remove-annotations ()
   "Removes buffer annotations (overlays and text properties)."
@@ -960,12 +1001,25 @@ normalise types or not when running CMD (through
 `agda2-goal-cmd'; WANT is used as `agda2-goal-cmd's WANT
 argument)."
   (let ((eval (make-symbol "eval")))
-  `(defun ,name (&optional not-normalise)
+  `(defun ,name (&optional prefix)
      ,(concat comment ".
 
-With a prefix argument the result is not explicitly normalised.")
+The form of the result depends on the prefix argument:
+
+* If the prefix argument is `nil' (i.e., if no prefix argument is
+  given), then the result is simplified.
+
+* If the prefix argument is `(4)' (for instance if C-u is typed
+  exactly once right before the command is invoked), then the
+  result is neither explicitly normalised nor simplified.
+
+* If any other prefix argument is used (for instance if C-u is
+  typed twice right before the command is invoked), then the
+  result is normalised.")
      (interactive "P")
-     (let ((,eval (if not-normalise "Instantiated" "Normalised")))
+     (let ((,eval (cond ((equal prefix nil) "Simplified")
+                        ((equal prefix '(4)) "Instantiated")
+                        ("Normalised"))))
        (agda2-goal-cmd (concat ,cmd " " ,eval)
                        ,want)))))
 
@@ -974,16 +1028,29 @@ With a prefix argument the result is not explicitly normalised.")
 COMMENT is used to build the function's comments. The function
 NAME takes a prefix argument which tells whether it should
 normalise types or not when running CMD (through
-`agda2-go' t nil; the string PROMPT is used as the goal command
+`agda2-go' t nil t; the string PROMPT is used as the goal command
 prompt)."
   (let ((eval (make-symbol "eval")))
-    `(defun ,name (not-normalise expr)
+    `(defun ,name (prefix expr)
        ,(concat comment ".
 
-With a prefix argument the result is not explicitly normalised.")
+The form of the result depends on the prefix argument:
+
+* If the prefix argument is `nil' (i.e., if no prefix argument is
+  given), then the result is simplified.
+
+* If the prefix argument is `(4)' (for instance if C-u is typed
+  exactly once right before the command is invoked), then the
+  result is neither explicitly normalised nor simplified.
+
+* If any other prefix argument is used (for instance if C-u is
+  typed twice right before the command is invoked), then the
+  result is normalised.")
        (interactive ,(concat "P\nM" prompt ": "))
-       (let ((,eval (if not-normalise "Instantiated" "Normalised")))
-         (agda2-go t nil
+       (let ((,eval (cond ((equal prefix nil) "Simplified")
+                          ((equal prefix '(4)) "Instantiated")
+                          ("Normalised"))))
+         (agda2-go t nil t
                    (concat ,cmd " " ,eval " "
                            (agda2-string-quote expr)))))))
 
@@ -1016,6 +1083,25 @@ top-level scope."
                           'agda2-infer-type
                         'agda2-infer-type-toplevel)))
 
+(defun agda2-why-in-scope ()
+  "Explain why something is in scope in a goal."
+  (interactive)
+  (agda2-goal-cmd "Cmd_why_in_scope" "Name"))
+
+(defun agda2-why-in-scope-toplevel (name)
+  "Explain why something is in scope at the top level."
+  (interactive "MName: ")
+  (agda2-go t nil t
+            "Cmd_why_in_scope_toplevel"
+            (agda2-string-quote name)))
+
+(defun agda2-why-in-scope-maybe-toplevel ()
+  "Explains why a given name is in scope."
+  (interactive)
+  (call-interactively (if (agda2-goal-at (point))
+                          'agda2-why-in-scope
+                          'agda2-why-in-scope-toplevel)))
+
 (agda2-maybe-normalised
  agda2-goal-and-context
  "Shows the type of the goal at point and the currect context"
@@ -1034,6 +1120,12 @@ top-level scope."
  "Cmd_context"
  nil)
 
+(agda2-maybe-normalised
+ agda2-helper-function-type
+  "Compute the type of a hypothetical helper function."
+  "Cmd_helper_function"
+  "Expression")
+
 (defun agda2-module-contents ()
   "Shows all the top-level names in the given module.
 Along with their types."
@@ -1044,7 +1136,7 @@ Along with their types."
   "Shows all the top-level names in the given module.
 Along with their types."
   (interactive "MModule name: ")
-  (agda2-go t nil
+  (agda2-go t nil t
             "Cmd_show_module_contents_toplevel"
             (agda2-string-quote module)))
 
@@ -1062,7 +1154,7 @@ a goal, the top-level scope."
 (defun agda2-solveAll ()
   "Solves all goals that are already instantiated internally."
   (interactive)
-  (agda2-go t t "Cmd_solveAll"))
+  (agda2-go t t t "Cmd_solveAll"))
 
 (defun agda2-solveAll-action (iss)
   (while iss
@@ -1095,7 +1187,7 @@ With a prefix argument \"abstract\" is ignored during the computation."
   (let ((cmd (concat "Cmd_compute_toplevel"
                      (if arg " True" " False")
                      " ")))
-    (agda2-go t nil (concat cmd (agda2-string-quote expr)))))
+    (agda2-go t nil t (concat cmd (agda2-string-quote expr)))))
 
 (defun agda2-compute-normalised-maybe-toplevel ()
   "Computes the normal form of the given expression,
@@ -1113,7 +1205,7 @@ With a prefix argument \"abstract\" is ignored during the computation."
 (defun agda2-highlight-reload nil
   "Loads precomputed syntax highlighting info for the current buffer.
 If there is any to load."
-  (agda2-go nil t
+  (agda2-go nil t t
             "Cmd_load_highlighting_info"
             (agda2-string-quote (buffer-file-name))))
 
@@ -1137,12 +1229,21 @@ ways."
        ;; trigger agda2-abort-highlighting.
        (inhibit-modification-hooks t))
       ((delims() (re-search-forward "[?]\\|[{][-!]\\|[-!][}]\\|--\\|\\\\begin{code}\\|\\\\end{code}" nil t))
-       (is-lone-questionmark ()
+       ;; is-proper checks whether string s (e.g. "?" or "--") is proper
+       ;; i.e., is not part of an identifier.
+       ;; comment-starter is true if s starts a comment (e.g. "--")
+       (is-proper (s comment-starter)
           (save-excursion
             (save-match-data
-                (backward-char 3)
-                (looking-at
-                 "\\(.[{(]\\|\\(.\\|\n\\)\\s \\)[?]\\(\\s \\|[)};]\\|$\\)"))))
+              (backward-char (length s))
+              (unless (bolp) (backward-char 1))  ;; bolp = pointer at beginning of line
+              ;; Andreas, 2014-05-17 Issue 1132
+              ;; A questionmark can also follow immediately after a .
+              ;; for instance to be a place holder for a dot pattern.
+              (looking-at (concat "\\([.{}();]\\|^\\|\\s \\)"  ;; \\s = whitespace
+                                  (regexp-quote s)
+                                  (unless comment-starter
+                                    "\\([{}();]\\|$\\|\\s \\)"))))))
        (make(p)  (agda2-make-goal p (point) (pop goals)))
        (inside-comment() (and stk (null     (car stk))))
        (inside-goal()    (and stk (integerp (car stk))))
@@ -1163,7 +1264,8 @@ ways."
           (cond
            ((c "\\begin{code}") (when (outside-code)               (pop stk)))
            ((c "\\end{code}")   (when (not stk)                    (push 'outside stk)))
-           ((c "--")            (when (not stk)                    (end-of-line)))
+           ((c "--")            (when (and (not stk)
+                                           (is-proper "--" t))     (end-of-line)))
            ((c "{-")            (when (and (inside-code)
                                            (not (inside-goal)))    (push nil           stk)))
            ((c "-}")            (when (inside-comment)             (pop stk)))
@@ -1173,7 +1275,7 @@ ways."
                                   (setq top (pop stk))
                                   (unless stk (make top))))
            ((c "?")             (progn
-                                  (when (and (not stk) (is-lone-questionmark))
+                                  (when (and (not stk) (is-proper "?" nil))
                                     (delete-char -1)
                                     (insert "{!!}")
                                     (make (- (point) 4)))))))))))
@@ -1232,7 +1334,14 @@ text properties."
              (goto-char (+ p 2)) (insert "(")))
       (multiple-value-bind (p q) (agda2-range-of-goal old-g)
         (delete-region (- q 2) q)
-        (delete-region p (+ p 2))))))
+        (delete-region p (+ p 2)))
+        ;; Update highlighting
+        (if (and (not (equal new-txt 'paren)) (not (equal new-txt 'no-paren)))
+            (apply 'agda2-go t t nil "Cmd_highlight"
+              (format "%d" old-g)
+              (agda2-mkRange p (- q 2))
+              (agda2-string-quote new-txt) nil))
+    )))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;; Misc
@@ -1248,9 +1357,14 @@ text properties."
 
 (defun agda2-goal-Range (o)
   "The Haskell Range of goal overlay O."
+  (agda2-mkRange (+ (overlay-start o) 2)
+                 (- (overlay-end   o) 2)))
+
+(defun agda2-mkRange (p q)
+  "The Haskell Range corresponding to two points."
   (format "(Range [Interval %s %s])"
-          (agda2-mkPos (+ (overlay-start o) 2))
-          (agda2-mkPos (- (overlay-end   o) 2))))
+          (agda2-mkPos p)
+          (agda2-mkPos q)))
 
 (defun agda2-mkPos (&optional p)
   "The Haskell Position corresponding to P or `point'."
@@ -1462,10 +1576,10 @@ invoked."
 With prefix argument, turn on display of implicit arguments if
 the argument is a positive number, otherwise turn it off."
   (interactive "P")
-  (cond ((eq arg nil)       (agda2-go t t "ToggleImplicitArgs"))
+  (cond ((eq arg nil)       (agda2-go t t t "ToggleImplicitArgs"))
         ((and (numberp arg)
-              (> arg 0))    (agda2-go t t "ShowImplicitArgs" "True"))
-        (t                  (agda2-go t t "ShowImplicitArgs" "False"))))
+              (> arg 0))    (agda2-go t t t "ShowImplicitArgs" "True"))
+        (t                  (agda2-go t t t "ShowImplicitArgs" "False"))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;;
