@@ -2,38 +2,32 @@
 
 module Agda.Auto.Auto (auto) where
 
-import Agda.Utils.Impossible
-#include "../undefined.h"
-
 import Control.Monad.Error
 import Control.Monad.State
-import System.IO.Unsafe (unsafePerformIO)
 import Data.List
-import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.IORef
 import qualified System.Timeout
-import Data.Maybe (catMaybes)
+import Data.Maybe (catMaybes, isNothing)
+import Data.Tuple (swap)
+import Data.Functor
 
 import Agda.Utils.Permutation (permute, takeP)
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.MetaVars
 import Agda.TypeChecking.Monad.Context
 import Agda.TypeChecking.Monad.Signature
-import Agda.TypeChecking.Monad.State (getScope)
 import Agda.TypeChecking.Substitute
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Abstract.Pretty (prettyA)
 import qualified Text.PrettyPrint as PP
-import qualified Agda.Syntax.Concrete as C
 import Agda.Syntax.Position
 import qualified Agda.Syntax.Internal as I
-import Agda.Syntax.Common
 import Agda.Syntax.Translation.InternalToAbstract
-import Agda.Syntax.Translation.AbstractToConcrete (abstractToConcrete, abstractToConcrete_, makeEnv, runAbsToCon, toConcrete)
+import Agda.Syntax.Translation.AbstractToConcrete (abstractToConcreteEnv, abstractToConcrete_, makeEnv, runAbsToCon, toConcrete)
 import Agda.Interaction.BasicOps hiding (refine)
-import Agda.Interaction.MakeCase (findClause)
 import Agda.TypeChecking.Reduce (normalise)
+import Agda.Syntax.Common
 import qualified Agda.Syntax.Scope.Base
 import Agda.Syntax.Scope.Monad (withCurrentModule)
 import qualified Agda.Syntax.Abstract.Name as AN
@@ -47,10 +41,12 @@ import Agda.Auto.Syntax
 import Agda.Auto.SearchControl
 import Agda.Auto.Typecheck
 
-
 import Agda.Auto.CaseSplit
 
+#include "../undefined.h"
+import Agda.Utils.Impossible
 
+insertAbsurdPattern :: String -> String
 insertAbsurdPattern [] = []
 insertAbsurdPattern s@(_:_) | take (length abspatvarname) s == abspatvarname = "()" ++ drop (length abspatvarname) s
 insertAbsurdPattern (c:s) = c : insertAbsurdPattern s
@@ -63,6 +59,25 @@ getName _ = Nothing
 
 dispmsg msg = return (Left [], Just msg)
 
+-- | Entry point for Auto tactic (Agsy).
+--
+--     @auto ii rng s = return (res, mmsg)@
+--
+--   If @mmsg = Just msg@, the message @msg@ produced by Agsy should
+--   be displayed to the user.
+--
+--   The result @res@ of the Auto tactic can be one of the following three:
+--
+--   1. @Left [(ii,s)]@
+--      A list of solutions @s@ for interaction ids @ii@.
+--      In particular, @Left []@ means Agsy found no solution.
+--
+--   2. @Right (Left cs)@
+--      A list of clauses (the user allowed case-split).
+--
+--   3. @Right (Right s)@
+--      A refinement for the interaction id @ii@ in which Auto was invoked.
+
 auto :: InteractionId -> Range -> String -> TCM (Either [(InteractionId, String)] (Either [String] String), Maybe String)
 auto ii rng argstr = liftTCM $ do
          let (hints, timeout, pick, mode, hintmode) = parseargs argstr
@@ -222,28 +237,27 @@ auto ii rng argstr = liftTCM $ do
                          (Map.elems tccons)
                         ) eqcons
                  res <- exsearch initprop recinfo defdfv
-                 iis <- getInteractionPoints
-                 riis <- mapM (\ii -> lookupInteractionId ii >>= \mi -> return (mi, ii)) iis
+                 riis <- map swap <$> getInteractionIdsAndMetas
+                 let timeoutString | isNothing res = " after timeout (" ++ show timeout ++ "s)"
+                                   | otherwise     = ""
                  if listmode then do
                    rsols <- liftM reverse $ liftIO $ readIORef sols
                    if null rsols then do
                      nsol' <- liftIO $ readIORef nsol
-                     dispmsg $ insuffsols (pick + 10 - nsol')
+                     dispmsg $ insuffsols (pick + 10 - nsol') ++ timeoutString
                     else do
                      aexprss <- mapM getsols rsols
                      cexprss <- mapM (mapM (\(mi, e) -> lookupMeta mi >>= \mv -> withMetaInfo (getMetaInfo mv) $ abstractToConcrete_ e >>= \e' -> return (mi, e'))) aexprss
                      let disp [(_, cexpr)] = show cexpr
                          disp cexprs = concat (map (\(mi, cexpr) -> case lookup mi riis of {Nothing -> show mi; Just ii -> show ii} ++ " := " ++ show cexpr ++ " ") cexprs)
                      ticks <- liftIO $ readIORef ticks
-                     dispmsg $ "Listing solution(s) " ++ show pick ++ "-" ++ show (pick + length rsols - 1) ++
-
-
+                     dispmsg $ "Listing solution(s) " ++ show pick ++ "-" ++ show (pick + length rsols - 1) ++ timeoutString ++
                                "\n" ++ unlines (map (\(x, y) -> show y ++ "  " ++ disp x) $ zip cexprss [pick..])
                   else
                    case res of
                     Nothing -> do
                      nsol' <- liftIO $ readIORef nsol
-                     dispmsg $ insuffsols (pick + 1 - nsol') ++ " at time out (" ++ show timeout ++ "s)"
+                     dispmsg $ insuffsols (pick + 1 - nsol') ++ timeoutString
                     Just depthreached -> do
                      ticks <- liftIO $ readIORef ticks
                      rsols <- liftIO $ readIORef sols
@@ -256,13 +270,16 @@ auto ii rng argstr = liftTCM $ do
                        giveress <-
                         mapM (\(mi, expr) ->
                          case lookup mi riis of
-                          Nothing -> giveExpr mi expr >>= \_ -> return Nothing
-                          Just ii' -> do (ae, []) <- give ii' Nothing expr
+                          Nothing ->
+                           catchError
+                            (giveExpr mi expr >>= \_ -> return (Nothing, Nothing))
+                            (\_ -> return (Nothing, Just ("Failed to give expr for side solution of " ++ show mi)))
+                          Just ii' -> do ae <- give ii' Nothing expr
                                          mv <- lookupMeta mi
                                          let scope = getMetaScope mv
-                                             ce = abstractToConcrete (makeEnv scope) ae
+                                         ce <- abstractToConcreteEnv (makeEnv scope) ae
                                          let cmnt = if ii' == ii then agsyinfo ticks else ""
-                                         return $ Just (ii', show ce ++ cmnt)
+                                         return (Just (ii', show ce ++ cmnt), Nothing)
                          ) exprs
                        let msg = if length exprs == 1 then
                                   Nothing
@@ -271,7 +288,11 @@ auto ii rng argstr = liftTCM $ do
                                           concatMap (\(mi', _) ->
                                            if mi' == mi then "" else (" " ++ case lookup mi' riis of {Nothing -> show mi'; Just ii -> show ii})
                                           ) exprs
-                       return (Left $ catMaybes giveress, msg)
+                       let msgs = catMaybes $ msg : map snd giveress
+                           msg' = case msgs of
+                                   [] -> Nothing
+                                   _ -> Just $ unlines msgs
+                       return (Left $ catMaybes $ map fst giveress, msg')
 
             MCaseSplit -> do
              case thisdefinfo of
@@ -281,8 +302,7 @@ auto ii rng argstr = liftTCM $ do
 
 
                  (ids, pats) <- constructPats cmap mi clause
-                 let pids = concat $ map (\(_, x) -> " " ++ case x of {Id s -> s; NoId -> "noid"}) ids
-                     ctx = map (\((hid, id), t) -> HI hid (id, t)) (zip ids mylocalVars)
+                 let ctx = map (\((hid, id), t) -> HI hid (id, t)) (zip ids mylocalVars)
                  ticks <- liftIO $ newIORef 0
                  let [rectyp'] = mymrectyp
                  defdfv <- getdfv mi def
@@ -300,13 +320,12 @@ auto ii rng argstr = liftTCM $ do
                    case cls' of
                     Left{} -> dispmsg "No solution found"
                     Right cls' -> do
-                     cls'' <- mapM (\(I.Clause _ tel perm ps body) ->
+                     cls'' <- forM cls' $ \ (I.Clause _ tel perm ps body t) -> do
                        withCurrentModule (AN.qnameModule def) $ do
                         -- Normalise the dot patterns
                         ps <- addCtxTel tel $ normalise ps
                         body <- etaContractBody body
-                        liftM modifyAbstractClause $ inContext [] $ reify $ AN.QNamed def $ I.Clause noRange tel perm ps body
-                      ) cls'
+                        liftM modifyAbstractClause $ inContext [] $ reify $ AN.QNamed def $ I.Clause noRange tel perm ps body t
                      pcs <- withInteractionId ii $ mapM prettyA cls''
                      ticks <- liftIO $ readIORef ticks
 
