@@ -1,8 +1,8 @@
-{-# LANGUAGE Rank2Types, ScopedTypeVariables, CPP #-}
-module Agda.Auto.CaseSplit where
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE Rank2Types #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 
-import Agda.Utils.Impossible
-#include "../undefined.h"
+module Agda.Auto.CaseSplit where
 
 import Data.IORef
 import Data.List (findIndex, union)
@@ -11,10 +11,13 @@ import qualified Data.IntMap as IntMap
 import Agda.Auto.NarrowingSearch
 import Agda.Auto.Syntax
 
-
 import Agda.Auto.SearchControl
 import Agda.Auto.Typecheck
 
+#include "../undefined.h"
+import Agda.Utils.Impossible
+
+abspatvarname :: String
 abspatvarname = "\0absurdPattern"
 
 costCaseSplitVeryHigh = 10000
@@ -135,7 +138,7 @@ caseSplitSearch' branchsearch depthinterval depth recdef ctx tt pats = do
            case findperm (map snd (drophid ctx)) of
             Just perm ->
              let HI scrhid(_, scrt) = ctx !! scrut
-                 ctx1 = take scrut ctx ++ (HI scrhid (Id abspatvarname, scrt)) : drop (scrut + 1) ctx
+                 ctx1 = take scrut ctx ++ (HI scrhid (stringToMyId abspatvarname, scrt)) : drop (scrut + 1) ctx
                  (ctx', _, pats') = applyperm perm ctx1 tt ({-map (replacep scrut 1 CSAbsurd __IMPOSSIBLE__) -}pats)
              in [(ctx', pats', Nothing)]
             Nothing -> __IMPOSSIBLE__ -- no permutation found
@@ -291,6 +294,12 @@ unifyexp e1 e2 = r e1 e2 (\unif -> Just unif) []
    (Lam hid1 (Abs _ b1), Lam hid2 (Abs _ b2)) | hid1 == hid2 -> r b1 b2 cont unif
    (Pi _ hid1 _ it1 (Abs _ ot1), Pi _ hid2 _ it2 (Abs _ ot2)) | hid1 == hid2 -> r it1 it2 (r ot1 ot2 cont) unif
    (Sort _, Sort _) -> cont unif -- a bit sloppy
+   (App _ _ (Var v) (NotM ALNil), App _ _ (Var u) (NotM ALNil))
+     | v == u -> cont unif
+   (App _ _ (Var v) (NotM ALNil), _)
+     | elem v (freevars e2) -> Nothing -- Occurs check
+   (_, App _ _ (Var v) (NotM ALNil))
+     | elem v (freevars e1) -> Nothing -- Occurs check
    (App _ _ (Var v) (NotM ALNil), _) ->
     case lookup v unif of
      Nothing -> cont ((v, e2) : unif)
