@@ -2,9 +2,6 @@
 
 module Agda.Auto.Convert where
 
-import Agda.Utils.Impossible
-#include "../undefined.h"
-
 import Control.Applicative hiding (getConst, Const(..))
 import Data.IORef
 import Data.Map (Map)
@@ -13,25 +10,24 @@ import Control.Monad.State
 import Control.Monad.Error
 
 import qualified Agda.Syntax.Internal as I
-import qualified Agda.Syntax.Literal as I
 import qualified Agda.Syntax.Common as C
 import qualified Agda.Syntax.Abstract.Name as AN
 import qualified Agda.Syntax.Abstract as A
 import qualified Agda.Syntax.Position as SP
 import qualified Agda.TypeChecking.Monad.Base as MB
 import Agda.TypeChecking.Monad.Signature (getConstInfo, getDefFreeVars, getImportedSignature)
-import Agda.Utils.Permutation (Permutation(Perm), idP, permute, takeP)
-import Agda.Interaction.BasicOps (rewrite, Rewrite(..))
+import Agda.Utils.Permutation (Permutation(Perm), permute, takeP, compactP)
 import Agda.TypeChecking.Level (reallyUnLevelView)
 import Agda.TypeChecking.Monad.Base (mvJudgement, mvPermutation, getMetaInfo, ctxEntry, envContext, clEnv)
 import Agda.TypeChecking.Monad.MetaVars (lookupMeta, withMetaInfo)
 import Agda.TypeChecking.Monad.Context (getContextArgs)
 import Agda.TypeChecking.Monad.Constraints (getAllConstraints)
-import Agda.TypeChecking.Substitute (piApply, raise)
+import Agda.TypeChecking.Substitute (piApply, applySubst)
+import Agda.TypeChecking.Telescope (renamingR)
 import qualified Agda.TypeChecking.Substitute as I (absBody)
 import Agda.TypeChecking.Reduce (Normalise, normalise, instantiate)
 import Agda.TypeChecking.EtaContract (etaContract)
-import Agda.TypeChecking.Primitive (constructorForm)
+import Agda.TypeChecking.Monad.Builtin (constructorForm)
 import Agda.TypeChecking.Free (freeIn)
 import qualified Agda.Utils.HashMap as HMap
 
@@ -40,6 +36,8 @@ import Agda.Auto.Syntax
 
 import Agda.Auto.CaseSplit hiding (lift)
 
+#include "../undefined.h"
+import Agda.Utils.Impossible
 
 norm :: Normalise t => t -> MB.TCM t
 norm x = normalise x
@@ -88,21 +86,23 @@ tomy imi icns typs = do
            clauses' <- tomyClauses clauses
            let narg = case clauses of
                         [] -> 0
-                        I.Clause {I.clausePats = xs} : _ -> length xs
+                        I.Clause {I.namedClausePats = xs} : _ -> length xs
            return (Def narg clauses' Nothing Nothing, [])
      (cont, projfcns2) <- case defn of
       MB.Axiom {} -> return (Postulate, [])
       MB.Function {MB.funClauses = clauses} -> clausesToDef clauses
-      MB.Primitive {MB.primClauses = Just clauses} -> clausesToDef clauses
-      MB.Primitive {} -> throwError $ strMsg "Auto: Primitive functions are not supported"
+      -- MB.Primitive {MB.primClauses = []} -> throwError $ strMsg "Auto: Primitive functions are not supported" -- Andreas, 2013-06-17 breaks interaction/AutoMisc
+      MB.Primitive {MB.primClauses = clauses} -> clausesToDef clauses
       MB.Datatype {MB.dataCons = cons} -> do
        cons2 <- mapM (\con -> getConst True con TMAll) cons
        return (Datatype cons2 [], [])
       MB.Record {MB.recFields = fields, MB.recTel = tel} -> do -- the value of recPars seems unreliable or don't know what it signifies
-       let pars n (I.El _ (I.Pi it typ)) = C.Arg (C.domHiding it) (C.domRelevance it) (I.var n) : pars (n - 1) (I.unAbs typ)
+       let pars n (I.El _ (I.Pi it typ)) = C.Arg (C.domInfo it) (I.var n) :
+                                           pars (n - 1) (I.unAbs typ)
            pars n (I.El s (I.Shared p))  = pars n (I.El s (I.derefPtr p))
            pars _ (I.El _ _) = []
-           contyp npar I.EmptyTel = I.El (I.mkType 0 {- arbitrary -}) (I.Def cn (pars (npar - 1) typ))
+           contyp npar I.EmptyTel = I.El (I.mkType 0 {- arbitrary -}) $
+                                      I.Def cn $ map I.Apply $ pars (npar - 1) typ
            contyp npar (I.ExtendTel it (I.Abs v tel)) = I.El (I.mkType 0 {- arbitrary -}) (I.Pi it (I.Abs v (contyp (npar + 1) tel)))
            contyp npar (I.ExtendTel it I.NoAbs{})     = __IMPOSSIBLE__
        contyp' <- tomyType $ contyp 0 tel
@@ -136,7 +136,7 @@ tomy imi icns typs = do
                       lift $ withMetaInfo (getMetaInfo mv) $ do
                        args <- getContextArgs
                        --sol <- norm (I.MetaV mi args)
-                       sol <- instantiate (I.MetaV mi args)
+                       sol <- instantiate $ I.MetaV mi $ map I.Apply $ permute (takeP (length args) $ mvPermutation mv) args
                        return $ Just sol
                      _ -> return Nothing
        case msol of
@@ -193,7 +193,8 @@ getConst :: Bool -> AN.QName -> TMode -> TOM (ConstRef O)
 getConst iscon name mode = do
  def <- lift $ getConstInfo name
  case MB.theDef def of
-  MB.Record {MB.recCon = conname} -> do
+  MB.Record {MB.recConHead = con} -> do
+   let conname = I.conName con
    cmap <- fst `liftM` gets sConsts
    case Map.lookup name cmap of
     Just (mode', c) ->
@@ -263,6 +264,9 @@ getEqs = do
  eqs' <- r eqs
  return $ concat eqs'
 
+copatternsNotImplemented :: MB.TCM a
+copatternsNotImplemented = MB.typeError $ MB.NotImplemented $
+  "The Agda synthesizer (Agsy) does not support copatterns yet"
 
 tomyClauses [] = return []
 tomyClauses (cl:cls) = do
@@ -272,7 +276,8 @@ tomyClauses (cl:cls) = do
   Just cl' -> cl' : cls'
   Nothing -> cls'
 
-tomyClause cl@(I.Clause {I.clausePerm = Perm n ps, I.clausePats = pats, I.clauseBody = body}) = do
+tomyClause cl@(I.Clause {I.clausePerm = Perm n ps, I.clauseBody = body}) = do
+ let pats = I.clausePats cl
  pats' <- mapM tomyPat pats
  body' <- tomyBody body
  return $ case body' of
@@ -280,11 +285,13 @@ tomyClause cl@(I.Clause {I.clausePerm = Perm n ps, I.clausePats = pats, I.clause
            Nothing -> Nothing
 
 tomyPat p = case C.unArg p of
+ I.ProjP _ -> lift $ copatternsNotImplemented
  I.VarP n -> return $ PatVar (show n)
  I.DotP _ -> return $ PatVar "_" -- because Agda includes these when referring to variables in the body
- I.ConP n _ pats -> do
+ I.ConP con _ pats -> do
+  let n = I.conName con
   c <- getConst True n TMAll
-  pats' <- mapM tomyPat pats
+  pats' <- mapM (tomyPat . fmap C.namedThing) pats
   def <- lift $ getConstInfo n
   cc <- lift $ liftIO $ readIORef c
   let Just npar = fst $ cdorigin cc
@@ -346,57 +353,64 @@ tomyType :: I.Type -> TOM (MExp O)
 tomyType (I.El _ t) = tomyExp t -- sort info is thrown away
 
 tomyExp :: I.Term -> TOM (MExp O)
-tomyExp (I.Var v as) = do
- as' <- tomyExps as
- return $ NotM $ App Nothing (NotM OKVal) (Var v) as'
-tomyExp (I.Lam hid b) = do
- b' <- tomyExp (I.absBody b)
- return $ NotM $ Lam (cnvh hid) (Abs (Id $ I.absName b) b')
-tomyExp t@(I.Lit{}) = do
- t <- lift $ constructorForm t
- case t of
-  I.Lit{} -> throwError $ strMsg "Auto: Literals in terms are not supported"
-  _ -> tomyExp t
-tomyExp (I.Level l) = tomyExp =<< lift (reallyUnLevelView l)
-tomyExp (I.Def name as) = do
- c <- getConst False name TMAll
- as' <- tomyExps as
- return $ NotM $ App Nothing (NotM OKVal) (Const c) as'
-tomyExp (I.Con name as) = do
- c <- getConst True name TMAll
- as' <- tomyExps as
- def <- lift $ getConstInfo name
- cc <- lift $ liftIO $ readIORef c
- let Just npar = fst $ cdorigin cc
- return $ NotM $ App Nothing (NotM OKVal) (Const c) (foldl (\x _ -> NotM $ ALConPar x) as' [1..npar])
-tomyExp (I.Pi (C.Dom hid _ x) b) = do
- let y    = I.absBody b
-     name = I.absName b
- x' <- tomyType x
- y' <- tomyType y
- return $ NotM $ Pi Nothing (cnvh hid) (Agda.TypeChecking.Free.freeIn 0 y) x' (Abs (Id name) y')
-tomyExp (I.Sort (I.Type (I.Max [I.ClosedLevel l]))) = return $ NotM $ Sort $ Set $ fromIntegral l
-tomyExp (I.Sort _) = return $ NotM $ Sort UnknownSort
-tomyExp t@I.MetaV{} = do
- t <- lift $ instantiate t
- case t of
-  I.MetaV mid _ -> do
-   mcurmeta <- gets sCurMeta
-   case mcurmeta of
-    Nothing -> return ()
-    Just curmeta ->
-     modify (\s -> s {sMetas = (Map.adjust (\(m, x, deps) -> (m, x, mid : deps)) curmeta (fst $ sMetas s), snd $ sMetas s)})
-   m <- getMeta mid
-   return $ Meta m
-  _ -> tomyExp t
-tomyExp (I.DontCare _) = return $ NotM $ dontCare
-tomyExp (I.Shared p) = tomyExp $ I.derefPtr p
+tomyExp v0 =
+  case I.unSpine v0 of
+    I.Var v es -> do
+      let Just as = I.allApplyElims es
+      as' <- tomyExps as
+      return $ NotM $ App Nothing (NotM OKVal) (Var v) as'
+    I.Lam info b -> do
+      b' <- tomyExp (I.absBody b)
+      return $ NotM $ Lam (cnvh info) (Abs (Id $ I.absName b) b')
+    t@I.Lit{} -> do
+      t <- lift $ constructorForm t
+      case t of
+        I.Lit{} -> throwError $ strMsg "Auto: Literals in terms are not supported"
+        _       -> tomyExp t
+    I.Level l -> tomyExp =<< lift (reallyUnLevelView l)
+    I.Def name es -> do
+      let Just as = I.allApplyElims es
+      c   <- getConst False name TMAll
+      as' <- tomyExps as
+      return $ NotM $ App Nothing (NotM OKVal) (Const c) as'
+    I.Con con as -> do
+      let name = I.conName con
+      c   <- getConst True name TMAll
+      as' <- tomyExps as
+      def <- lift $ getConstInfo name
+      cc  <- lift $ liftIO $ readIORef c
+      let Just npar = fst $ cdorigin cc
+      return $ NotM $ App Nothing (NotM OKVal) (Const c) (foldl (\x _ -> NotM $ ALConPar x) as' [1..npar])
+    I.Pi (C.Dom info x) b -> do
+      let y    = I.absBody b
+          name = I.absName b
+      x' <- tomyType x
+      y' <- tomyType y
+      return $ NotM $ Pi Nothing (cnvh info) (Agda.TypeChecking.Free.freeIn 0 y) x' (Abs (Id name) y')
+    I.Sort (I.Type (I.Max [I.ClosedLevel l])) -> return $ NotM $ Sort $ Set $ fromIntegral l
+    I.Sort _ -> return $ NotM $ Sort UnknownSort
+    t@I.MetaV{} -> do
+      t <- lift $ instantiate t
+      case t of
+        I.MetaV mid _ -> do
+          mcurmeta <- gets sCurMeta
+          case mcurmeta of
+            Nothing -> return ()
+            Just curmeta ->
+              modify $ \ s -> s { sMetas = ( Map.adjust (\(m, x, deps) -> (m, x, mid : deps)) curmeta (fst $ sMetas s)
+                                           , snd $ sMetas s
+                                           ) }
+          m <- getMeta mid
+          return $ Meta m
+        _ -> tomyExp t
+    I.DontCare _ -> return $ NotM $ dontCare
+    I.Shared p -> tomyExp $ I.derefPtr p
 
 tomyExps [] = return $ NotM ALNil
-tomyExps (C.Arg hid _ a : as) = do
+tomyExps (C.Arg info a : as) = do
  a' <- tomyExp a
  as' <- tomyExps as
- return $ NotM $ ALCons (cnvh hid) a' as'
+ return $ NotM $ ALCons (cnvh info) a' as'
 
 tomyIneq MB.CmpEq = False
 tomyIneq MB.CmpLeq = True
@@ -407,11 +421,11 @@ fmType :: I.MetaId -> I.Type -> Bool
 fmType m (I.El _ t) = fmExp m t
 
 fmExp :: I.MetaId -> I.Term -> Bool
-fmExp m (I.Var _ as) = fmExps m as
+fmExp m (I.Var _ as) = fmExps m $ I.argsFromElims as
 fmExp m (I.Lam _ b) = fmExp m (I.unAbs b)
 fmExp m (I.Lit _) = False
 fmExp m (I.Level (I.Max as)) = any (fmLevel m) as
-fmExp m (I.Def _ as) = fmExps m as
+fmExp m (I.Def _ as) = fmExps m $ I.argsFromElims as
 fmExp m (I.Con _ as) = fmExps m as
 fmExp m (I.Pi x y)  = fmType m (C.unDom x) || fmType m (I.unAbs y)
 fmExp m (I.Sort _) = False
@@ -432,12 +446,16 @@ fmLevel m (I.Plus _ l) = case l of
 
 -- ---------------------------------------------
 
-cnvh C.NotHidden = NotHidden
-cnvh C.Instance = Instance
-cnvh C.Hidden = Hidden
-icnvh NotHidden = C.NotHidden
-icnvh Instance = C.Instance
-icnvh Hidden = C.Hidden
+cnvh info = case C.getHiding info of
+    C.NotHidden -> NotHidden
+    C.Instance  -> Instance
+    C.Hidden    -> Hidden
+icnvh h = (C.setHiding h' C.defaultArgInfo)
+    where
+    h' = case h of
+        NotHidden -> C.NotHidden
+        Instance  -> C.Instance
+        Hidden    -> C.Hidden
 
 -- ---------------------------------------------
 
@@ -461,7 +479,15 @@ frommyExp (NotM e) =
   App _ _ (Const c) as -> do
    cdef <- lift $ readIORef c
    let (iscon, name) = cdorigin cdef
-       (ndrop, h) = case iscon of {Just n -> (n, I.Con); Nothing -> (0, I.Def)}
+{-
+   case iscon of
+      Just n -> do
+        v <- getConTerm name -- We are not in TCM
+        frommyExps n as v
+-}
+       (ndrop, h) = case iscon of
+                      Just n -> (n, \ q -> I.Con (I.ConHead q [])) -- TODO: restore fields
+                      Nothing -> (0, \ f vs -> I.Def f $ map I.Apply vs)
    frommyExps ndrop as (h name [])
   Lam hid (Abs mid t) -> do
    t' <- frommyExp t
@@ -469,7 +495,7 @@ frommyExp (NotM e) =
   Pi _ hid _ x (Abs mid y) -> do
    x' <- frommyType x
    y' <- frommyType y
-   return $ I.Pi (C.Dom (icnvh hid) C.Relevant x') (I.Abs (case mid of {NoId -> "x"; Id id -> id}) y')
+   return $ I.Pi (C.Dom (icnvh hid) x') (I.Abs (case mid of {NoId -> "x"; Id id -> id}) y')
    -- maybe have case for Pi where possdep is False which produces Fun (and has to unweaken y), return $ I.Fun (C.Arg (icnvh hid) x') y'
   Sort (Set l) ->
    return $ I.Sort (I.mkType (fromIntegral l))
@@ -492,8 +518,9 @@ frommyExps ndrop (NotM as) trm =
   ALCons _ _ xs | ndrop > 0 -> frommyExps (ndrop - 1) xs trm
   ALCons hid x xs -> do
    x' <- frommyExp x
-   frommyExps ndrop xs (addend (C.Arg (icnvh hid) C.Relevant x') trm)
+   frommyExps ndrop xs (addend (C.Arg (icnvh hid) x') trm)
 
+  -- Andreas, 2013-10-19 TODO: restore postfix projections
   ALProj eas idx hid xs -> do
    idx <- lift $ expandbind idx
    c <- case idx of
@@ -502,14 +529,14 @@ frommyExps ndrop (NotM as) trm =
    cdef <- lift $ readIORef c
    let name = snd $ cdorigin cdef
    trm2 <- frommyExps 0 eas (I.Def name [])
-   frommyExps 0 xs (addend (C.Arg (icnvh hid) C.Relevant trm) trm2)
+   frommyExps 0 xs (addend (C.Arg (icnvh hid) trm) trm2)
 
   ALConPar xs | ndrop > 0 -> frommyExps (ndrop - 1) xs trm
   ALConPar _ -> __IMPOSSIBLE__
  where
-  addend x (I.Var h xs) = I.Var h (xs ++ [x])
+  addend x (I.Var h xs) = I.Var h (xs ++ [I.Apply x])
   addend x (I.Con h xs) = I.Con h (xs ++ [x])
-  addend x (I.Def h xs) = I.Def h (xs ++ [x])
+  addend x (I.Def h xs) = I.Def h (xs ++ [I.Apply x])
   addend x (I.Shared p) = addend x (I.derefPtr p)
   addend _ _ = __IMPOSSIBLE__
 
@@ -520,8 +547,10 @@ abslamvarname = "\0absurdlambda"
 modifyAbstractExpr :: A.Expr -> A.Expr
 modifyAbstractExpr = f
  where
-  f (A.App i e1 (C.Arg h r (C.Named n e2))) = A.App i (f e1) (C.Arg h r (C.Named n (f e2)))
-  f (A.Lam i (A.DomainFree h rel n) _) | show n == abslamvarname = A.AbsurdLam i h
+  f (A.App i e1 (C.Arg info (C.Named n e2))) =
+        A.App i (f e1) (C.Arg info (C.Named n (f e2)))
+  f (A.Lam i (A.DomainFree info n) _) | show n == abslamvarname =
+        A.AbsurdLam i $ C.argInfoHiding info
   f (A.Lam i b e) = A.Lam i b (f e)
   f (A.Rec i xs) = A.Rec i (map (\(n, e) -> (n, f e)) xs)
   f (A.RecUpdate i e xs) = A.RecUpdate i (f e) (map (\(n, e) -> (n, f e)) xs)
@@ -543,10 +572,11 @@ constructPats cmap mainm clause = do
       (ns'', p') <- cnvp ns' p
       return (ns'', p' : ps')
      cnvp ns p =
-      let hid = cnvh $ C.argHiding p
-      in case C.unArg p of
+      let hid = cnvh $ C.argInfo p
+      in case C.namedArg p of
        I.VarP n -> return ((hid, Id n) : ns, HI hid (CSPatVar $ length ns))
-       I.ConP c _ ps -> do
+       I.ConP con _ ps -> do
+        let c = I.conName con
         (c2, _) <- runStateT (getConst True c TMAll) (S {sConsts = (cmap, []), sMetas = initMapS, sEqs = initMapS, sCurMeta = Nothing, sMainMeta = mainm})
         (ns', ps') <- cnvps ns ps
         cc <- liftIO $ readIORef c2
@@ -555,8 +585,9 @@ constructPats cmap mainm clause = do
        I.DotP t -> do
         (t2, _) <- runStateT (tomyExp t) (S {sConsts = (cmap, []), sMetas = initMapS, sEqs = initMapS, sCurMeta = Nothing, sMainMeta = mainm})
         return (ns, HI hid (CSPatExp t2))
+       I.ProjP{} -> copatternsNotImplemented
        _ -> __IMPOSSIBLE__
- (names, pats) <- cnvps [] (I.clausePats clause)
+ (names, pats) <- cnvps [] (I.namedClausePats clause)
  return (reverse names, pats)
 
 
@@ -567,7 +598,7 @@ frommyClause (ids, pats, mrhs) = do
       let Id id = mid
       tel <- ctel ctx
       t' <- frommyType t
-      return $ I.ExtendTel (C.Dom (icnvh hid) C.Relevant t') (I.Abs id tel)
+      return $ I.ExtendTel (C.Dom (icnvh hid) t') (I.Abs id tel)
  tel <- ctel $ reverse ids
  let getperms 0 [] perm nv = return (perm, nv)
      getperms n [] _ _ = __IMPOSSIBLE__
@@ -608,23 +639,32 @@ frommyClause (ids, pats, mrhs) = do
         cdef <- lift $ readIORef c
         let (Just ndrop, name) = cdorigin cdef
         ps' <- cnvps ndrop ps
-        return (I.ConP name Nothing ps')
+        let con = I.ConHead name [] -- TODO: restore record fields!
+        return (I.ConP con Nothing ps')
        CSPatExp e -> do
-        e' <- frommyExp {- renm e -} e  -- rename disabled to match (incorrect?) Agda reification of clauses
+        e' <- frommyExp e {- renm e -} -- renaming before adding to clause below
         return (I.DotP e')
        CSAbsurd -> __IMPOSSIBLE__ -- CSAbsurd not used
        _ -> __IMPOSSIBLE__
-      return $ C.Arg (icnvh hid) C.Relevant p'
+      return $ C.Arg (icnvh hid) $ C.unnamed p'   -- TODO: recover names
  ps <- cnvps 0 pats
  body <- case mrhs of
           Nothing -> return $ I.NoBody
           Just e -> do
-           e' <- frommyExp {- renm e -} e  -- rename disabled to match (incorrect?) Agda reification of clauses
+           e' <- frommyExp e {- renm e -} -- renaming before adding to clause below
            let r 0 = I.Body e'
                r n = I.Bind $ I.Abs "h" $ r (n - 1)
-               e'' = r ({-length ids + -}nv)
+               e'' = r nv
            return e''
- return $ I.Clause SP.noRange tel (Perm (nv{- + length ids-}) perm) ps body
+ let cperm =  Perm nv perm
+ return $ I.Clause
+   { I.clauseRange = SP.noRange
+   , I.clauseTel   = tel
+   , I.clausePerm  = cperm
+   , I.namedClausePats = applySubst (renamingR $ compactP cperm) ps
+   , I.clauseBody  = applySubst (renamingR cperm) <$> body
+   , I.clauseType  = Nothing -- TODO: compute clause type
+   }
 
 contains_constructor :: [CSPat O] -> Bool
 contains_constructor = any f
@@ -669,7 +709,7 @@ freeIn = f
 
 
 negtype :: ConstRef o -> MExp o -> MExp o
-negtype ee = f 0
+negtype ee = f (0 :: Int)
  where
   mr x = let NotM x' = x in x'
   f n e = case mr e of
@@ -698,11 +738,11 @@ findClauseDeep m = do
                   I.Body e -> f e
     findMeta e =
      case I.ignoreSharing e of
-      I.Var _ as -> findMetas as
+      I.Var _ es -> findMetas $ I.argsFromElims es
       I.Lam _ b -> findMeta (I.absBody b)
       I.Lit{} -> False
       I.Level (I.Max as) -> any (fmLevel m) as
-      I.Def _ as -> findMetas as
+      I.Def _ es -> findMetas $ I.argsFromElims es
       I.Con _ as -> findMetas as
       I.Pi it ot -> findMetat (C.unDom it) || findMetat (I.unAbs ot)
       I.Sort{} -> False
@@ -736,23 +776,28 @@ matchType cdfv tctx ctyp ttyp = trmodps cdfv ctyp
     ft nl n c (I.El _ e1) (I.El _ e2) = f nl n c e1 e2
     f nl n c e1 e2 = case I.ignoreSharing e1 of
      I.Var v1 as1 | v1 < nl -> case e2 of
-      I.Var v2 as2 | v1 == v2 -> fs nl (n + 1) c as1 as2
+      I.Var v2 as2 | v1 == v2 -> fes nl (n + 1) c as1 as2
       _ -> Nothing
      I.Var v1 _ | v1 < nl + na -> c n -- unify vars with no args?
      I.Var v1 as1 -> case e2 of
-      I.Var v2 as2 | cdfv + na + nl - v1 == tctx + nl - v2 -> fs nl (n + 1) c as1 as2
+      I.Var v2 as2 | cdfv + na + nl - v1 == tctx + nl - v2 -> fes nl (n + 1) c as1 as2
       _ -> Nothing
      _ -> case (I.ignoreSharing e1, I.ignoreSharing e2) of
       (I.MetaV{}, _) -> c n
       (_, I.MetaV{}) -> c n
       (I.Lam hid1 b1, I.Lam hid2 b2) | hid1 == hid2 -> f (nl + 1) n c (I.absBody b1) (I.absBody b2)
       (I.Lit lit1, I.Lit lit2) | lit1 == lit2 -> c (n + 1)
-      (I.Def n1 as1, I.Def n2 as2) | n1 == n2 -> fs nl (n + 1) c as1 as2
+      (I.Def n1 as1, I.Def n2 as2) | n1 == n2 -> fes nl (n + 1) c as1 as2
       (I.Con n1 as1, I.Con n2 as2) | n1 == n2 -> fs nl (n + 1) c as1 as2
-      (I.Pi (C.Dom hid1 rel1 it1) ot1, I.Pi (C.Dom hid2 rel2 it2) ot2) | hid1 == hid2 -> ft nl n (\n -> ft (nl + 1) n c (I.absBody ot1) (I.absBody ot2)) it1 it2
+      (I.Pi (C.Dom info1 it1) ot1, I.Pi (C.Dom info2 it2) ot2) | C.argInfoHiding info1 == C.argInfoHiding info2 -> ft nl n (\n -> ft (nl + 1) n c (I.absBody ot1) (I.absBody ot2)) it1 it2
       (I.Sort{}, I.Sort{}) -> c n -- sloppy
       _ -> Nothing
     fs nl n c es1 es2 = case (es1, es2) of
      ([], []) -> c n
-     (C.Arg hid1 rel1 e1 : es1, C.Arg hid2 rel2 e2 : es2) | hid1 == hid2 -> f nl n (\n -> fs nl n c es1 es2) e1 e2
+     (C.Arg info1 e1 : es1, C.Arg info2 e2 : es2) | C.argInfoHiding info1 == C.argInfoHiding info2 -> f nl n (\n -> fs nl n c es1 es2) e1 e2
+     _ -> Nothing
+    fes nl n c es1 es2 = case (es1, es2) of
+     ([], []) -> c n
+     (I.Proj f : es1, I.Proj f' : es2) | f == f' -> fes nl n c es1 es2
+     (I.Apply (C.Arg info1 e1) : es1, I.Apply (C.Arg info2 e2) : es2) | C.argInfoHiding info1 == C.argInfoHiding info2 -> f nl n (\n -> fes nl n c es1 es2) e1 e2
      _ -> Nothing
