@@ -1,14 +1,16 @@
-{-# LANGUAGE ExistentialQuantification, CPP #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE ExistentialQuantification #-}
 
 module Agda.Auto.Syntax where
 
-import Agda.Utils.Impossible
-#include "../undefined.h"
-
 import Data.IORef
 
 import Agda.Auto.NarrowingSearch
 
+#include "../undefined.h"
+import Agda.Utils.Impossible
+
+-- | Unique identifiers for variable occurrences in unification.
 type UId o = Metavar (Exp o) (RefInfo o)
 
 data HintMode = HMNormal
@@ -20,26 +22,40 @@ data EqReasoningConsts o = EqReasoningConsts {eqrcId, eqrcBegin, eqrcStep, eqrcE
 data EqReasoningState = EqRSNone | EqRSChain | EqRSPrf1 | EqRSPrf2 | EqRSPrf3
  deriving (Eq, Show)
 
+-- | The concrete instance of the 'blk' parameter in 'Metavar'.
+--   I.e., the information passed to the search control.
 
-data RefInfo o = RIEnv {rieHints :: [(ConstRef o, HintMode)], rieDefFreeVars :: Nat -- Nat - deffreevars (to make cost of using module parameters correspond to that of hints)
+data RefInfo o
+  = RIEnv
+    { rieHints :: [(ConstRef o, HintMode)]
+    , rieDefFreeVars :: Nat
+      -- ^ Nat - deffreevars
+      --   (to make cost of using module parameters correspond to that of hints).
+    , rieEqReasoningConsts :: Maybe (EqReasoningConsts o)
+    }
+  | RIMainInfo
+    { riMainCxtLength :: Nat
+      -- ^ Size of typing context in which meta was created.
+    , riMainType      :: HNExp o
+      -- ^ Head normal form of type of meta.
+    , riMainIota      :: Bool
+       -- ^ True if iota steps performed when normalising target type
+       --   (used to put cost when traversing a definition
+       --    by construction instantiation).
+    }
+  | RIUnifInfo [CAction o] (HNExp o) -- meta environment, opp hne
+  | RICopyInfo (ICExp o)
+  | RIIotaStep Bool -- True - semiflex
+  | RIInferredTypeUnknown
+  | RINotConstructor
+  | RIUsedVars [UId o] [Elr o]
+  | RIPickSubsvar
 
-                        , rieEqReasoningConsts :: Maybe (EqReasoningConsts o)
-
-                       }
-               | RIMainInfo Nat (HNExp o) Bool -- true if iota steps performed when normalising target type (used to put cost when traversing a definition by construction instantiation)
-               | forall a . RIUnifInfo [CAction o] (HNExp o) -- meta environment, opp hne
-               | RICopyInfo (ICExp o)
-               | RIIotaStep Bool -- True - semiflex
-               | RIInferredTypeUnknown
-               | RINotConstructor
-               | RIUsedVars [UId o] [Elr o]
-               | RIPickSubsvar
-
-               | RIEqRState EqReasoningState
+  | RIEqRState EqReasoningState
 
 
-               | RICheckElim Bool -- isdep
-               | RICheckProjIndex [ConstRef o] -- noof proj functions
+  | RICheckElim Bool -- isdep
+  | RICheckProjIndex [ConstRef o] -- noof proj functions
 
 
 type MyPB o = PB (RefInfo o)
@@ -47,7 +63,7 @@ type MyMB a o = MB a (RefInfo o)
 
 type Nat = Int
 
-
+-- | 'Hiding' in Agda.
 data FMode = Hidden
            | Instance
            | NotHidden
@@ -57,59 +73,106 @@ data FMode = Hidden
 data MId = Id String
          | NoId
 
+stringToMyId :: String -> MId
+stringToMyId = Id
+
+-- | Abstraction with maybe a name.
+--
+--   Different from Agda, where there is also info
+--   whether function is constant.
 data Abs a = Abs MId a
 
-data ConstDef o = ConstDef {cdname :: String, cdorigin :: o, cdtype :: MExp o, cdcont :: DeclCont o
+-- | Constant signatures.
 
-                   , cddeffreevars :: Nat
+data ConstDef o = ConstDef
+  { cdname        :: String
+    -- ^ For debug printing.
+  , cdorigin      :: o
+    -- ^ Reference to the Agda constant.
+  , cdtype        :: MExp o
+    -- ^ Type of constant.
+  , cdcont        :: DeclCont o
+    -- ^ Constant definition.
+  , cddeffreevars :: Nat
+    -- ^ Free vars of the module where the constant is defined..
+  } -- contains no metas
 
-                  } -- contains no metas
+-- | Constant definitions.
 
-data DeclCont o = Def Nat [Clause o] (Maybe Nat) (Maybe Nat) -- maybe an index to elimand argument, maybe index to elim arg if semiflex
-                | Datatype [ConstRef o] -- constructors
+data DeclCont o
+  = Def Nat [Clause o] (Maybe Nat) (Maybe Nat) -- maybe an index to elimand argument, maybe index to elim arg if semiflex
+  | Datatype [ConstRef o] -- constructors
+             [ConstRef o] -- projection functions (in case it is a record)
 
-                           [ConstRef o] -- projection functions
-
-                | Constructor Nat -- number of omitted args
-                | Postulate
+  | Constructor Nat -- number of omitted args
+  | Postulate
 
 type Clause o = ([Pat o], MExp o)
 
-data Pat o = PatConApp (ConstRef o) [Pat o]
-           | PatVar String
-           | PatExp
+data Pat o
+  = PatConApp (ConstRef o) [Pat o]
+  | PatVar String
+  | PatExp
+    -- ^ Dot pattern.
+{- TODO: projection patterns.
+  | PatProj (ConstRef o)
+    -- ^ Projection pattern.
+-}
 
 type ConstRef o = IORef (ConstDef o)
 
-data Elr o = Var Nat
-           | Const (ConstRef o)
+-- | Head of application (elimination).
+data Elr o
+  = Var Nat
+  | Const (ConstRef o)
 
-data Sort = Set Nat
+data Sort
+  = Set Nat
+  | UnknownSort
+  | Type
 
-          | UnknownSort
-
-          | Type
-
-data Exp o = App (Maybe (UId o)) (OKHandle (RefInfo o)) (Elr o) (MArgList o)
-           | Lam FMode (Abs (MExp o))
-           | Pi (Maybe (UId o)) FMode Bool (MExp o) (Abs (MExp o)) -- true if possibly dependent (var not known to not occur)
-           | Sort Sort
-
-           | AbsurdLambda FMode
+-- | Agsy's internal syntax.
+data Exp o
+  = App
+    { appUId   :: Maybe (UId o)
+      -- ^ Unique identifier of the head.
+    , appOK    :: OKHandle (RefInfo o)
+      -- ^ This application has been type-checked.
+    , appHead  :: Elr o
+      -- ^ Head.
+    , appElims :: MArgList o
+      -- ^ Arguments.
+    }
+  | Lam FMode (Abs (MExp o))
+    -- ^ Lambda with hiding information.
+  | Pi (Maybe (UId o)) FMode Bool (MExp o) (Abs (MExp o))
+    -- ^ @True@ if possibly dependent (var not known to not occur).
+    --   @False@ if non-dependent.
+  | Sort Sort
+  | AbsurdLambda FMode
+    -- ^ Absurd lambda with hiding information.
 
 
 dontCare = Sort UnknownSort
 
 
+-- | "Maybe expression":  Expression or reference to meta variable.
 type MExp o = MM (Exp o) (RefInfo o)
 
-data ArgList o = ALNil
-               | ALCons FMode (MExp o) (MArgList o)
+data ArgList o
+  = ALNil
+    -- ^ No more eliminations.
+  | ALCons FMode (MExp o) (MArgList o)
+    -- ^ Application and tail.
 
-               | ALProj (MArgList o) (MM (ConstRef o) (RefInfo o)) FMode (MArgList o) -- proj pre args, projfcn idx, tail
+  | ALProj (MArgList o) (MM (ConstRef o) (RefInfo o)) FMode (MArgList o)
+    -- ^ proj pre args, projfcn idx, tail
 
-
-               | ALConPar (MArgList o) -- inserted to cover glitch of polymorphic constructor applications coming from Agda
+  | ALConPar (MArgList o)
+    -- ^ Constructor parameter (missing in Agda).
+    --   Agsy has monomorphic constructors.
+    --   Inserted to cover glitch of polymorphic constructor
+    --   applications coming from Agda
 
 
 type MArgList o = MM (ArgList o) (RefInfo o)
@@ -119,25 +182,37 @@ data HNExp o = HNApp [Maybe (UId o)] (Elr o) (ICArgList o)
              | HNPi [Maybe (UId o)] FMode Bool (ICExp o) (Abs (ICExp o))
              | HNSort Sort
 
+-- | Head-normal form of 'ICArgList'.  First entry is exposed.
+--
+--   Q: Why are there no projection eliminations?
 data HNArgList o = HNALNil
                  | HNALCons FMode (ICExp o) (ICArgList o)
-
                  | HNALConPar (ICArgList o)
 
-
-type ICExp o = Clos (MExp o) o
-
-type CExp o = TrBr (ICExp o) o
-
+-- | Lazy concatenation of argument lists under explicit substitutions.
 data ICArgList o = CALNil
                  | CALConcat (Clos (MArgList o) o) (ICArgList o)
 
+-- | An expression @a@ in an explicit substitution @[CAction a]@.
+type ICExp o  = Clos (MExp o) o
 data Clos a o = Clos [CAction o] a
+
+type CExp o   = TrBr (ICExp o) o
 data TrBr a o = TrBr [MExp o] a
 
-data CAction o = Sub (ICExp o)
-               | Skip
-               | Weak Nat
+-- | Entry of an explicit substitution.
+--
+--   An explicit substitution is a list of @CAction@s.
+--   This is isomorphic to the usual presentation where
+--   @Skip@ and @Weak@ would be constructors of exp. substs.
+
+data CAction o
+  = Sub (ICExp o)
+    -- ^ Instantation of variable.
+  | Skip
+    -- ^ For going under a binder, often called "Lift".
+  | Weak Nat
+    -- ^ Shifting substitution (going to a larger context).
 
 type Ctx o = [(MId, CExp o)]
 
@@ -314,12 +389,14 @@ weakelr 0 elr = elr
 weakelr n (Var v) = Var (v + n)
 weakelr _ elr@(Const _) = elr
 
+-- | Substituting for a variable.
 doclos :: [CAction o] -> Nat -> Either Nat (ICExp o)
 doclos = f 0
  where
-  f ns [] i = Left (ns + i)
+  -- ns is the number of weakenings
+  f ns []            i = Left (ns + i)
   f ns (Weak n : xs) i = f (ns + n) xs i
-  f ns (Sub s : _) 0 = Right (weaki ns s)
-  f ns (Skip : _) 0 = Left ns
-  f ns (Skip : xs) i = f (ns + 1) xs (i - 1)
-  f ns (Sub _ : xs) i = f ns xs (i - 1)
+  f ns (Sub s  : _ ) 0 = Right (weaki ns s)
+  f ns (Skip   : _ ) 0 = Left ns
+  f ns (Skip   : xs) i = f (ns + 1) xs (i - 1)
+  f ns (Sub _  : xs) i = f ns xs (i - 1)
