@@ -1,11 +1,10 @@
-{-# LANGUAGE ExistentialQuantification, FlexibleContexts,
-             ScopedTypeVariables, CPP #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE ExistentialQuantification #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 
 module Agda.Auto.Typecheck where
 
-import Agda.Utils.Impossible
-#include "../undefined.h"
-
 import Data.IORef
 import Control.Monad (liftM)
 
@@ -13,9 +12,13 @@ import Agda.Auto.NarrowingSearch
 import Agda.Auto.Syntax
 import Agda.Auto.SearchControl
 
+#include "../undefined.h"
+import Agda.Utils.Impossible
 
 -- ---------------------------------
 
+-- | Typechecker drives the solution of metas.
+
 tcExp :: Bool -> Ctx o -> CExp o -> MExp o -> EE (MyPB o)
 tcExp isdep ctx typ@(TrBr typtrs ityp@(Clos _ itypexp)) trm =
   mbpcase prioTypeUnknown Nothing (hnn_checkstep ityp) $ \(hntyp, iotastepdone) ->
@@ -597,10 +600,7 @@ comp' ineq lhs@(TrBr trs1 e1) rhs@(TrBr trs2 e2) = comp ineq e1 e2
            Left _ -> return True
            Right e -> boringExp e
           ALCons{} -> return False
-
-          ALProj{} -> __IMPOSSIBLE__
-
-
+          ALProj{} -> return False
           ALConPar{} -> return False
 
        _ -> return False
