@@ -1,9 +1,9 @@
 -- | Remove forced arguments from constructors.
 {-# LANGUAGE CPP #-}
+
 module Agda.Compiler.Epic.ForceConstrs where
 
 import Control.Applicative
-import Control.Monad.Trans
 
 import Agda.Compiler.Epic.AuxAST
 import Agda.Compiler.Epic.CompileState
@@ -11,8 +11,7 @@ import Agda.Compiler.Epic.Interface
 
 import qualified Agda.Syntax.Common   as S
 import qualified Agda.Syntax.Internal as T
-import Agda.TypeChecking.Monad (TCM, reportSDoc)
-import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Monad (TCM)
 
 #include "../../undefined.h"
 import Agda.Utils.Impossible
@@ -23,8 +22,8 @@ makeForcedArgs (T.El _ term) = case term of
     T.Pi  arg ab  -> isRel arg : makeForcedArgs (T.unAbs ab)
     _ -> []
   where
-    isRel :: S.Dom T.Type -> Forced
-    isRel arg = case S.domRelevance arg of
+    isRel :: T.Dom T.Type -> Forced
+    isRel arg = case S.getRelevance arg of
       S.Relevant   -> NotForced
       S.Irrelevant -> Forced
       S.UnusedArg  -> Forced
