@@ -1,34 +1,32 @@
-{-# LANGUAGE CPP, ScopedTypeVariables #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE ScopedTypeVariables #-}
+
 module Agda.Compiler.Epic.Forcing where
 
 import Control.Applicative
-import Control.Arrow (first, second)
 import Control.Monad
 import Control.Monad.State
-import Control.Monad.Trans
 
 import Data.Char
 import Data.List hiding (sort)
-import qualified Data.Map as M
 import Data.Maybe
 
 import Agda.Syntax.Common
-import qualified Agda.Syntax.Internal as SI
+import qualified Agda.Syntax.Internal as I
 import Agda.Syntax.Literal
 import Agda.Syntax.Position(noRange)
 import Agda.Syntax.Internal(Tele(..), Telescope, Term, Abs(..), unAbs, absName, Type, Args, QName, unEl)
 import Agda.TypeChecking.Monad
+import Agda.TypeChecking.Rules.LHS.Problem (FlexibleVars, defaultFlexibleVar)
 import Agda.TypeChecking.Rules.LHS.Unify
-import Agda.TypeChecking.Rules.LHS.Instantiate
 import Agda.TypeChecking.Substitute
-  (raiseFrom, raise, applySubst, apply, wkS, raiseS, dropS, (++#), TelV(..))
+  (applySubst, apply, wkS, raiseS, dropS, (++#), TelV(..))
 import qualified Agda.TypeChecking.Substitute as S
 import Agda.TypeChecking.Pretty as P
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Telescope
 import Agda.Utils.List
 import Agda.Utils.Monad
-import Agda.Utils.Permutation
 import Agda.Utils.Size
 import qualified Agda.Utils.HashMap as HM
 
@@ -41,7 +39,6 @@ import qualified Agda.Compiler.Epic.FromAgda as FA
 #include "../../undefined.h"
 import Agda.Utils.Impossible
 
-
 -- | Returns how many parameters a datatype has
 dataParameters :: QName -> Compile TCM Nat
 dataParameters = lift . dataParametersTCM
@@ -102,8 +99,8 @@ insertTele x 0 ins term (ExtendTel t to) = do
       , text "term:" <+> prettyTCM term
       , text "to:"   <+> prettyTCM (unAbs to)
       ]
-    (st, arg) <- case SI.unEl . unDom $ t' of
-            SI.Def st arg -> return (st, arg)
+    (st, arg) <- case I.unEl . unDom $ t' of
+            I.Def st es -> return (st, fromMaybe __IMPOSSIBLE__ $ I.allApplyElims es)
             s          -> do
               report 10 $ vcat
                 [ text "ERROR!!!"
@@ -140,9 +137,10 @@ insertTele er n ins term (ExtendTel x xs) = do
     (xs', typ) <- insertTele er (n - 1) ins term (unAbs xs)
     return (ExtendTel x $ Abs (absName xs) xs' , typ)
 
-mkCon c n = SI.Con c [ defaultArg $ SI.Var (fromIntegral i) [] | i <- [n - 1, n - 2 .. 0] ]
+-- TODO: restore fields in ConHead
+mkCon c n = I.Con (I.ConHead c []) [ defaultArg $ I.Var (fromIntegral i) [] | i <- [n - 1, n - 2 .. 0] ]
 
-unifyI :: Telescope -> [Nat] -> Type -> Args -> Args -> Compile TCM [Maybe Term]
+unifyI :: Telescope -> FlexibleVars -> Type -> Args -> Args -> Compile TCM [Maybe Term]
 unifyI tele flex typ a1 a2 = lift $ addCtxTel tele $ unifyIndices_ flex typ a1 a2
 
 takeTele 0 _ = EmptyTel
@@ -188,7 +186,7 @@ forcedExpr vars tele expr = case expr of
         let n = fromMaybe __IMPOSSIBLE__ $ elemIndex x vars
         (Case v <$>) . forM brs $ \ br -> case br of
             BrInt i e -> do
-              (tele'', _) <-  insertTele __IMPOSSIBLE__ n Nothing (SI.Lit (LitChar noRange (chr i))) tele
+              (tele'', _) <-  insertTele __IMPOSSIBLE__ n Nothing (I.Lit (LitChar noRange (chr i))) tele
               BrInt i <$> forcedExpr (replaceAt n vars []) tele'' e
 
             Default e -> Default <$> rec e
@@ -205,7 +203,9 @@ forcedExpr vars tele expr = case expr of
                   else do
                     -- unify the telescope type with the return type of the constructor
                     unif <- case (unEl ntyp, unEl ctyp) of
-                        (SI.Def st a1, SI.Def st' a2) | st == st' -> do
+                        (I.Def st es1, I.Def st' es2) | st == st' -> do
+                            let a1 = fromMaybe __IMPOSSIBLE__ $ I.allApplyElims es1
+                            let a2 = fromMaybe __IMPOSSIBLE__ $ I.allApplyElims es2
                             typPars <- dataParameters st
                             setType <- getType st
                             report 10 $ vcat
@@ -213,14 +213,14 @@ forcedExpr vars tele expr = case expr of
                               , text "ctyp:" <+> prettyTCM ctyp
                               ]
                             unifyI (takeTele (n + length as) tele'')
-                                   [0 .. n + length as]
+                                   (map defaultFlexibleVar [0 .. n + length as])
                                    (setType `apply` take typPars a1)
                                    (drop typPars a1)
                                    (drop typPars a2)
                         _ -> __IMPOSSIBLE__
                     let
                         lower = wkS (-1) . dropS 1
-                        subT 0 tel = let ss = [fromMaybe (SI.Var n []) t
+                        subT 0 tel = let ss = [fromMaybe (I.Var n []) t
                                                 | (n , t) <- zip [0..] unif] ++#
                                               raiseS (length unif)
                                       in (applySubst ss tel, lower ss)
@@ -246,13 +246,13 @@ forcedExpr vars tele expr = case expr of
     rec = forcedExpr vars tele
 
 -- | replace the forcedVar with pattern matching from the outside.
-replaceForced :: ([Var],[Var]) -> Telescope -> [Var] -> [Maybe SI.Term] -> Expr -> Compile TCM Expr
+replaceForced :: ([Var],[Var]) -> Telescope -> [Var] -> [Maybe I.Term] -> Expr -> Compile TCM Expr
 replaceForced (vars,_) tele [] _ e = forcedExpr vars tele e
 replaceForced (vars,uvars) tele (fvar : fvars) unif e = do
     let n = fromMaybe __IMPOSSIBLE__ $ elemIndex fvar uvars
     mpos <- findPosition n unif
     case mpos of
-        Nothing -> case unif !! n of
+        Nothing -> case fromMaybe __IMPOSSIBLE__ $ unif !!! n of
             Nothing | fvar `notElem` fv e ->
               replaceForced (vars, uvars) tele fvars unif e
             Nothing -> do
@@ -270,7 +270,7 @@ replaceForced (vars,uvars) tele (fvar : fvars) unif e = do
                 subst fvar v <$> replaceForced (vars, uvars)
                                                tele fvars unif (Let v te e)
         Just (pos , term) -> do
-            (build, v) <- buildTerm (uvars !! pos) n term
+            (build, v) <- buildTerm (fromMaybe __IMPOSSIBLE__ $ uvars !!! pos) n term
             build . subst fvar v <$> replaceForced (vars, uvars) tele fvars unif
                                      e
   where
@@ -279,21 +279,22 @@ replaceForced (vars,uvars) tele (fvar : fvars) unif e = do
 -- | Given a term containg the forced var, dig out the variable by inserting
 -- the proper case-expressions.
 buildTerm :: Var -> Nat -> Term -> Compile TCM (Expr -> Expr, Var)
-buildTerm var idx (SI.Var i _) | idx == i = return (id, var)
-buildTerm var idx (SI.Con c args) = do
+buildTerm var idx (I.Var i _) | idx == i = return (id, var)
+buildTerm var idx (I.Con con args) = do
+    let c = I.conName con
     vs <- replicateM (length args) newName
     (pos , arg) <- fromMaybe __IMPOSSIBLE__ <$> findPosition idx (map (Just . unArg) args)
-    (fun' , v) <- buildTerm (vs !! pos) idx arg
+    (fun2 , v) <- buildTerm (fromMaybe __IMPOSSIBLE__ $ vs !!! pos) idx arg
     tag <- getConstrTag c
-    let fun e = casee (Var var) [Branch tag c vs e]
-    return (fun . fun' , v)
+    let fun1 e = casee (Var var) [Branch tag c vs e]
+    return (fun1 . fun2 , v)
 buildTerm _ _ _ = __IMPOSSIBLE__
 
 
 -- | Find the location where a certain Variable index is by searching the constructors
 --   aswell. i.e find a term that can be transformed into a pattern that contains the
 --   same value the index. This fails if no such term is present.
-findPosition :: Nat -> [Maybe SI.Term] -> Compile TCM (Maybe (Nat, SI.Term))
+findPosition :: Nat -> [Maybe I.Term] -> Compile TCM (Maybe (Nat, I.Term))
 findPosition var ts = (listToMaybe . catMaybes <$>) . forM (zip [0..] ts) $ \ (n, mt) -> do
     ifM (maybe (return False) pred mt)
         (return (Just (n, fromMaybe __IMPOSSIBLE__ mt)))
@@ -301,8 +302,8 @@ findPosition var ts = (listToMaybe . catMaybes <$>) . forM (zip [0..] ts) $ \ (n
   where
     pred :: Term -> Compile TCM Bool
     pred t = case t of
-      SI.Var i _ | var == i -> return True
-      SI.Con c args         -> do
-          forc <- getForcedArgs c
+      I.Var i _ | var == i -> return True
+      I.Con c args         -> do
+          forc <- getForcedArgs $ I.conName c
           or <$> mapM (pred . unArg) (notForced forc args)
       _                  -> return False
