@@ -7,11 +7,9 @@ import Control.Applicative
 import Control.Monad
 import Control.Monad.State
 import Data.Char
-import Data.Map(Map)
 import qualified Data.Map as M
 import Data.Maybe
 
-import Agda.Interaction.Options
 import Agda.Syntax.Common
 import Agda.Syntax.Internal hiding (Term(..))
 import qualified Agda.Syntax.Internal as T
@@ -21,6 +19,7 @@ import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Level (reallyUnLevelView)
 import qualified Agda.TypeChecking.Substitute as S
 import Agda.TypeChecking.Pretty
+import Agda.Utils.List
 
 import Agda.Compiler.Epic.AuxAST
 import Agda.Compiler.Epic.CompileState
@@ -46,8 +45,10 @@ translateDefn msharp (n, defini) =
         vars <- replicateM (dataPars d + dataIxs d) newName
         return . return $ Fun True n' (Just n) ("datatype: " ++ show n) vars UNIT
     f@(Function{}) -> do
-        let projArgs = maybe 0 (pred . snd) (funProjection f)
-        ccs  <- reverseCCBody projArgs <$> normaliseStatic (funCompiled f)
+        let projArgs = projectionArgs f
+            cc       = fromMaybe __IMPOSSIBLE__ $ funCompiled f
+        -- let projArgs = maybe 0 (pred . projIndex) (funProjection f)
+        ccs  <- reverseCCBody projArgs <$> normaliseStatic cc
         let len   = (+ projArgs) . length . clausePats . head .  funClauses $ f
             toEta = arity (defType defini) - len
         -- forcing <- lift $ gets (optForcing . stPersistentOptions)
@@ -182,11 +183,13 @@ compileClauses name nargs c = do
         CC.Case n nc -> case length env <= n of
            True -> __IMPOSSIBLE__
            False -> case CC.catchAllBranch nc of
-            Nothing -> Case (Var (env !! n)) <$> compileCase env omniDefault n nc
+            Nothing -> Case (Var (fromMaybe __IMPOSSIBLE__ $ env !!! n)) <$>
+                         compileCase env omniDefault n nc
             Just de -> do
                 def <- compileClauses' env omniDefault de
                 bindExpr (lazy def) $ \ var ->
-                  Case (Var (env !! n)) <$> compileCase env (Just var) n nc
+                  Case (Var (fromMaybe __IMPOSSIBLE__ $ env !!! n)) <$>
+                    compileCase env (Just var) n nc
         CC.Done _ t -> substTerm ({- reverse -} env) t
         CC.Fail     -> return IMPOSSIBLE
 
@@ -218,10 +221,13 @@ compileClauses name nargs c = do
 --   from patternmatching. Agda terms are in de Bruijn so we just check the new
 --   names in the position.
 substTerm :: [Var] -> T.Term -> Compile TCM Expr
-substTerm env term = case term of
-    T.Var ind args -> case length env <= ind of
+substTerm env term = case T.unSpine term of
+    T.Var ind es -> do
+      let args = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
+      case length env <= ind of
         True  -> __IMPOSSIBLE__
-        False -> apps (env !! ind) <$> mapM (substTerm env . unArg) args
+        False -> apps (fromMaybe __IMPOSSIBLE__ $ env !!! ind) <$>
+                   mapM (substTerm env . unArg) args
     T.Lam _ (Abs _ te) -> do
        name <- newName
        Lam name <$> substTerm (name : env) te
@@ -230,19 +236,23 @@ substTerm env term = case term of
        Lam name <$> substTerm env te
     T.Lit l -> Lit <$> substLit l
     T.Level l -> substTerm env =<< lift (reallyUnLevelView l)
-    T.Def q args -> do
+    T.Def q es -> do
+      let args = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
       let name = unqname q
       del <- getDelayed q
       def <- theDef <$> lift (getConstInfo q)
-      let nr = case def of
-                Function{funProjection = Just (_ , x)} -> pred x
+      let nr = projectionArgs def
+{- MOVED to Signature.hs
+             case def of
+                Function{funProjection = Just p} -> pred $ projIndex p
                 _ -> 0
+ -}
       f <- apps name . (replicate nr UNIT ++) <$> mapM (substTerm env . unArg) args
       return $ case del of
         True  -> Lazy f
         False -> f
-    T.Con q args -> do
-        let con = unqname q
+    T.Con c args -> do
+        let con = unqname $ conName c
         apps con <$> mapM (substTerm env . unArg) args
     T.Shared p -> substTerm env $ derefPtr p
     T.Pi _ _ -> return UNIT
