@@ -1,10 +1,15 @@
-{-# LANGUAGE CPP, TypeOperators, PatternGuards #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE TypeOperators #-}
+{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE OverlappingInstances #-}
+
 module Agda.Compiler.Epic.Injection where
 
 import Control.Monad.State
+import Control.Monad.Reader
 
-import Data.Function
-import Data.Ix
 import Data.List
 import Data.Map(Map)
 import qualified Data.Map as M
@@ -13,21 +18,21 @@ import Data.Set(Set)
 import qualified Data.Set as S
 
 import Agda.Syntax.Common
-import Agda.Syntax.Internal
+import Agda.Syntax.Internal as I
+import Agda.Syntax.Internal.Pattern (FunArity(..))
 import Agda.Syntax.Literal
 import Agda.TypeChecking.CompiledClause
-import Agda.TypeChecking.Monad hiding ((!!!))
+import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
-import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Pretty hiding (empty)
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
-import Agda.TypeChecking.Telescope
+import Agda.Utils.List
 import Agda.Utils.Monad
-import Agda.Utils.Size
+import Agda.Utils.Null
 import qualified Agda.Utils.HashMap as HM
 
 import Agda.Compiler.Epic.CompileState
-import qualified Agda.Compiler.Epic.FromAgda as FA
 import Agda.Compiler.Epic.Interface as Interface
 
 #include "../../undefined.h"
@@ -47,8 +52,8 @@ findInjection defs = do
         Just inj@(InjectiveFun nvar arity) -> case theDef def of
             f@(Function{})   -> do
                 modifyEI $ \s -> s { injectiveFuns = M.insert q inj (injectiveFuns s) }
-                let ns = replicate arity (Arg NotHidden Relevant "")
-                return $ (,) q $ def { theDef = f { funCompiled = Done ns $
+                let ns = replicate arity (defaultArg empty)
+                return $ (,) q $ def { theDef = f { funCompiled = Just $ Done ns $
                                                       var $ arity - nvar - 1 } }
             _                -> __IMPOSSIBLE__
 
@@ -68,7 +73,7 @@ replaceFunCC name cc = do
   where
     replaceDef :: Definition -> Definition
     replaceDef def = case theDef def of
-        f@(Function{}) -> def {theDef = f { funCompiled = cc } }
+        f@(Function{}) -> def {theDef = f { funCompiled = Just $ cc } }
         x                -> __IMPOSSIBLE__
 
 -- | If the pairs of constructor names have the same tags, the function is
@@ -83,7 +88,8 @@ isInjective :: QName    -- ^ Name of the function being tested
                                   ))
 isInjective nam []  = return Nothing
 isInjective nam cls@(cl : _) = do
-    let total = genericLength . clausePats $ cl
+    lift $ reportSLn "epic.injection" 20 $ "checking isInjective " ++ show nam
+    let total = funArity cls
     (listToMaybe . catMaybes <$>) . forM [0 .. total - 1] $ \i -> do
         cli <- forM cls $ \ cl -> isInjectiveHere nam i  cl
         let cli' = catMaybes cli
@@ -91,18 +97,6 @@ isInjective nam cls@(cl : _) = do
              then Just ((nam, InjectiveFun i total), concat cli')
              else Nothing
 
-remAbs :: ClauseBody -> Term
-remAbs b = case b of
-    Body t     -> t
-    Bind ab    -> remAbs $ absBody ab
-    NoBody     -> __IMPOSSIBLE__
-
-isNoBody :: ClauseBody -> Bool
-isNoBody b = case b of
-    Body t     -> False
-    Bind ab    -> isNoBody $ absBody ab
-    NoBody     -> True
-
 patternToTerm :: Nat -> Pattern -> Term
 patternToTerm n p = case p of
     VarP v          -> var n
@@ -111,43 +105,56 @@ patternToTerm n p = case p of
                              $ snd
                              $ foldr (\ arg (n, ts) -> (n + nrBinds arg, patternToTerm n arg : ts))
                                      (n , [])
-                             $ map unArg args
+                             $ map namedArg args
     LitP l          -> Lit l
+    ProjP d         -> Def d [] -- Andreas, 2012-10-31 that might not be enought to get a term from list of patterns (TODO)
 
 nrBinds :: Num i => Pattern -> i
 nrBinds p = case p of
     VarP v          -> 1
     DotP t          -> 0
-    ConP c typ args -> sum $ map (nrBinds . unArg) args
+    ConP c typ args -> sum $ map (nrBinds . namedArg) args
     LitP l          -> 0
+    ProjP{}         -> 0
 
-substForDot :: [Arg Pattern] -> Substitution
+substForDot :: [I.NamedArg Pattern] -> Substitution
 substForDot = makeSubst 0 0 . reverse . calcDots
   where
     makeSubst i accum [] = raiseS (i + accum)
     makeSubst i accum (True  : ps) = makeSubst i (accum +1) ps
     makeSubst i accum (False : ps) = var (i + accum) :# makeSubst (i+1) accum ps
 
-    calcDots = concatMap calcDots' . map unArg
+    calcDots = concatMap calcDots' . map namedArg
     calcDots' p = case p of
         VarP v          -> [False]
         DotP t          -> [True]
         ConP c typ args -> calcDots args
         LitP l          -> [False]
+        ProjP{}         -> [False]
 
 isInjectiveHere :: QName  -- ^ Name of the function being tested
                 -> Int    -- ^ The current argument
                 -> Clause
                 -> Compile TCM InjConstraints
-isInjectiveHere nam idx Clause {clauseBody = body} | isNoBody body = return emptyC
 isInjectiveHere nam idx clause = do
-    let t    = patternToTerm idxR $ unArg $ clausePats clause !! idx
-        t'   = applySubst (substForDot $ clausePats clause) t
+ lift $ reportSDoc "epic.injection" 40 $ sep
+   [ text "isInjectiveHere"
+   , prettyTCM nam
+   , text ("argumentNo=" ++ show idx)
+   -- , prettyTCM (clausePats clause)
+   ]
+ case getBody clause of
+  Nothing -> return emptyC
+  Just body -> do
+    let t    = patternToTerm idxR $ unArg $ fromMaybe __IMPOSSIBLE__ $
+                 clausePats clause !!! idx
+        t'   = applySubst (substForDot $ namedClausePats clause) t
         idxR = sum . map (nrBinds . unArg) . genericDrop (idx + 1) $ clausePats clause
-        body = remAbs $ clauseBody clause
     body' <- lift $ reduce body
+    lift $ reportSLn "epic.injection" 40 "reduced body"
     injFs <- gets (injectiveFuns . importedModules)
-    res <- (t' <: body') (M.insert nam (InjectiveFun idx
+    lift $ reportSLn "epic.injection" 40 "calculated injFs"
+    res <- (t' <: body') `runReaderT` (M.insert nam (InjectiveFun idx
                                                      (length (clausePats clause))) injFs)
     lift $ reportSDoc "epic.injection" 20 $ vcat
       [ text "isInjective:" <+> text (show nam)
@@ -162,6 +169,7 @@ isInjectiveHere nam idx clause = do
       ]
     return res
 
+-- | Turn NATURAL literal n into suc^n zero.
 litToCon :: Literal -> TCM Term
 litToCon l = case l of
     LitInt   r n | n > 0     -> do
@@ -172,9 +180,9 @@ litToCon l = case l of
 --    LitLevel _ n -> -- Does not really matter
     lit          -> return $ Lit lit
 
-litCon :: Literal -> Bool
-litCon (LitInt _ _) = True
-litCon _          = False
+litInt :: Literal -> Bool
+litInt LitInt{} = True
+litInt _        = False
 
 insertAt :: (Nat,Term) -> Term -> Term
 insertAt (index, ins) =
@@ -182,6 +190,8 @@ insertAt (index, ins) =
 
 solve :: [QName] -> [((QName, InjectiveFun), [(QName,QName)])] -> Compile TCM [(QName, InjectiveFun)]
 solve newNames xs = do
+    lift $ reportSDoc "epic.injection" 30 $
+      sep $ text "Epic.Injection.solve" : map prettyTCM newNames
     -- Only primitive lists should be in the current module at this point,
     -- but we still want them
     conGraph <- M.union <$> gets (constrTags . curModule) <*> gets (constrTags . importedModules)
@@ -210,7 +220,7 @@ solve newNames xs = do
             _         -> __IMPOSSIBLE__
         case M.toList eqs of
             (c, Same n) : _ -> do
-                let grp = eqGroups tags !!! n
+                let grp = eqGroups tags !!!! n
                 tag <- assignConstrTag' c (S.toList grp)
                 updateTags . fromMaybe __IMPOSSIBLE__ =<< setTag n tag tags { constrGroup = eqs }
             _              -> return ()
@@ -236,38 +246,89 @@ unionConstraints (Just c : cs) = do
     return (c ++ cs')
 
 -- | Are two terms injectible?
+--   Tries to find a mapping between constructors that equates the terms.
+--
 --   Precondition: t1 is normalised, t2 is in WHNF
 -- When reducing t2, it may become a literal, which makes this not work in some cases...
-(<:) :: Term -> Term -> (QName :-> InjectiveFun) -> Compile TCM InjConstraints
-(Lit l        <:  t1)          injs | litCon l = do
-    l' <- lift $ litToCon l
-    (l' <: t1) injs
-(t1 <: Lit l)                  injs | litCon l = do
-    l' <- lift $ litToCon l
-    (t1 <: l') injs
-(t1           <: Def n2 args2) injs | Just (InjectiveFun argn arit) <- M.lookup n2 injs =
-    if genericLength args2 /= arit
-        then return Nothing
-        else do
-            arg <- lift $ reduce $ unArg $ args2 !! argn
-            (t1 <: arg) injs
--- (Var n1 []    <: Var n2 [])    nam idx = return $ if n1 == n2 then emptyC else Nothing
-(Var n1 args1 <: Var n2 args2) injs | n1 == n2 && length args1 == length args2 = do
-    args1' <- map unArg <$> mapM (lift . reduce) args1
-    args2' <- map unArg <$> mapM (lift . reduce) args2
-    unionConstraints <$> zipWithM (\a b -> (a <: b) injs) args1' args2'
-(Def q1 args1 <: Def q2 args2) injs | q1 == q2 && length args1 == length args2 = do
-    args1' <- map unArg <$> mapM (lift . reduce) args1
-    args2' <- map unArg <$> mapM (lift . reduce) args2
-    unionConstraints <$> zipWithM (\a b -> (a <: b) injs) args1' args2'
-(Con c1 args1 <: Con c2 args2) injs = do
-    args1' <- map unArg <$> flip notForced args1 <$> getForcedArgs c1
-    args2' <- map unArg <$> (mapM (lift . reduce) =<< flip notForced args2 <$> getForcedArgs c2)
-    if length args1' == length args2'
-        then addConstraint c1 c2 <$> unionConstraints <$> zipWithM (\a b -> (a <: b) injs) args1' args2'
-        else return Nothing
-(_            <: _) _ = return Nothing
+class Injectible a where
+  (<:) :: a -> a -> ReaderT (QName :-> InjectiveFun) (Compile TCM) InjConstraints
 
+instance Injectible a => Injectible (I.Arg a) where
+  a1 <: a2 = unArg a1 <: unArg a2
+
+instance Injectible a => Injectible [a] where
+  l1 <: l2
+    | length l1 == length l2 = unionConstraints <$> zipWithM (<:) l1 l2
+    | otherwise              = return Nothing
+
+instance Injectible a => Injectible (Elim' a) where
+  e1 <: e2 =
+    case (e1, e2) of
+      (Proj f1 , Proj f2 ) | f1 == f2 -> return $ Just []
+      (Apply a1, Apply a2)            -> a1 <: a2
+      _                               -> return Nothing
+
+instance Injectible Term where
+  t1 <: t2 = do
+    injs <- ask
+    -- Andreas, 2013-10-18: ignoring the precondition (NF, WHNF) since I am not maintaining it
+    -- in recursive calls.
+    -- The original code did not follow this invariant in the Var-Var and Def-Def case,
+    -- thus, I am not trusting it.  Also the call site does not seem to ensure it.
+    -- It could be restored by only reducing the right argument in the Arg-instance.
+
+    -- (t1, t2) <- lift . lift . reduce $ (t1, t2)  -- NOTE: reduce *introduces* Lit! Loops!
+    case (t1, t2) of
+      (Lit l, Lit l') | l == l' -> return $ Just []
+      (Lit l, _) | litInt l -> do
+        l' <- lift . lift $ litToCon l
+        l' <: t2
+      (_,  Lit l) | litInt l -> do
+        l' <- lift . lift $ litToCon l
+        t1 <: l'
+      (_, Def n2 es2) | Just (InjectiveFun argn arit) <- M.lookup n2 injs -> do
+        if genericLength es2 /= arit
+          then return Nothing
+          else do
+            case es2 !!! argn of
+              Nothing        -> __IMPOSSIBLE__
+              Just (Proj{})  -> __IMPOSSIBLE__
+              Just (Apply a) -> t1 <: unArg a
+      (Var i1 es1, Var i2 es2) | i1 == i2 -> es1 <: es2
+      (Def q1 es1, Def q2 es2) | q1 == q2 -> es1 <: es2
+      (Con con1 args1, Con con2 args2) -> do
+        let c1 = conName con1
+            c2 = conName con2
+        args1' <- flip notForced args1 <$> do lift . getForcedArgs $ c1
+        args2' <- flip notForced args2 <$> do lift . getForcedArgs $ c2
+        addConstraint c1 c2 <$> do
+          args1' <: args2'
+      _ -> return Nothing
+{-
+      (_, Def n2 args2) | Just (InjectiveFun argn arit) <- M.lookup n2 injs -> do
+        if genericLength args2 /= arit
+          then return Nothing
+          else do
+              arg <- lift . lift . reduce $ unArg $ args2 !! argn
+              t1 <: arg
+      (Var n1 args1, Var n2 args2) | n1 == n2 && length args1 == length args2 -> do
+        args1' <- map unArg <$> mapM (lift . lift . reduce) args1
+        args2' <- map unArg <$> mapM (lift . lift . reduce) args2
+        unionConstraints <$> zipWithM (\a b -> (a <: b)) args1' args2'
+      (Def q1 args1, Def q2 args2) | q1 == q2 && length args1 == length args2 -> do
+        args1' <- map unArg <$> mapM (lift . lift . reduce) args1
+        args2' <- map unArg <$> mapM (lift . lift . reduce) args2
+        unionConstraints <$> zipWithM (\a b -> (a <: b)) args1' args2'
+      (Con con1 args1, Con con2 args2) -> do
+        let c1 = conName con1
+            c2 = conName con2
+        args1' <- map unArg <$> flip notForced args1 <$> getForcedArgs c1
+        args2' <- map unArg <$> (mapM (lift . lift . reduce) =<< flip notForced args2 <$> getForcedArgs c2)
+        if length args1' == length args2'
+            then addConstraint c1 c2 <$> unionConstraints <$> zipWithM (\a b -> (a <: b)) args1' args2'
+            else return Nothing
+      _ -> return Nothing
+-}
 data TagEq
     = Same Int
     | IsTag Tag
@@ -286,8 +347,8 @@ initialTags setTags newNames = Tags
 
 unify :: QName -> QName -> Tags -> Compile TCM (Maybe Tags)
 unify c1 c2 ts = do
-    let g1 = constrGroup ts !!! c1
-        g2 = constrGroup ts !!! c2
+    let g1 = constrGroup ts !!!! c1
+        g2 = constrGroup ts !!!! c2
     case (g1, g2) of
         (Same n1, Same n2)   | n1 == n2 -> return $ Just ts
         (IsTag t1, IsTag t2) | t1 == t2 -> return $ Just ts
@@ -298,18 +359,20 @@ unify c1 c2 ts = do
 
 setTag :: Int -> Tag -> Tags -> Compile TCM (Maybe Tags)
 setTag gid tag ts = return $ Just $ ts
-    { constrGroup = foldr (\c -> M.insert c (IsTag tag)) (constrGroup ts) (S.toList $ eqGroups ts !!! gid)}
+    { constrGroup = foldr (\c -> M.insert c (IsTag tag)) (constrGroup ts) (S.toList $ eqGroups ts !!!! gid)}
 
 mergeGroups :: Int -> Int -> Tags -> Compile TCM (Maybe Tags)
 mergeGroups n1 n2 ts = do
-    let g1s = eqGroups ts !!! n1
-        g2s = eqGroups ts !!! n2
+    let g1s = eqGroups ts !!!! n1
+        g2s = eqGroups ts !!!! n2
         gs  = S.union g1s g2s
-    ifM (not <$> andM [unifiable e1 e2 | e1 <- S.toList g1s, e2 <- S.toList g2s])
+        g1l = S.toList g1s
+        g2l = S.toList g2s
+    ifNotM (andM $ zipWith unifiable g1l g2l)
         (return Nothing) $
         return $ Just $ ts
             { eqGroups    = M.delete n2 $ M.insert n1 gs (eqGroups ts)
-            , constrGroup = M.fromList [(e2, Same n1) | e2 <- S.toList g2s] `M.union` constrGroup ts
+            , constrGroup = M.fromList [ (e2, Same n1) | e2 <- g2l ] `M.union` constrGroup ts
             }
 
 unifiable :: QName -> QName -> Compile TCM Bool
@@ -318,7 +381,7 @@ unifiable c1 c2 = do
     d2 <- getConData c2
     return $ d1 /= d2
 
-(!!!) :: Ord k => k :-> v -> k -> v
-m !!!  k = case M.lookup k m of
+(!!!!) :: Ord k => k :-> v -> k -> v
+m !!!!  k = case M.lookup k m of
     Nothing -> __IMPOSSIBLE__
     Just x  -> x
