@@ -1,17 +1,13 @@
 {-# LANGUAGE CPP #-}
+
 -- | Smash functions which return something that can be inferred
 --   (something of a type with only one element)
 
 module Agda.Compiler.Epic.Smashing where
 
-import Control.Arrow((&&&))
-import Control.Monad
 import Control.Monad.State
-import Control.Monad.Trans
 
 import Data.List
-import qualified Data.Map as M
-import Data.Map (Map)
 import Data.Maybe
 
 import qualified Data.Set as S
@@ -24,7 +20,6 @@ import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Reduce
-import Agda.TypeChecking.Rules.LHS.Unify
 
 import Agda.Compiler.Epic.AuxAST as AA
 import Agda.Compiler.Epic.CompileState
@@ -75,7 +70,7 @@ xs +++ ys = unflattenTel names $ map (raise (size ys)) (flattenTel xs) ++ flatte
   where names = teleNames xs ++ teleNames ys
 
 -- | Can a datatype be inferred? If so, return the only possible value.
-inferable :: Set QName -> QName -> [Arg Term] ->  Compile TCM (Maybe Expr)
+inferable :: Set QName -> QName -> [SI.Arg Term] ->  Compile TCM (Maybe Expr)
 inferable visited dat args | dat `S.member` visited = return Nothing
 inferable visited dat args = do
   lift $ reportSLn "epic.smashing" 10 $ "  inferring:" ++ (show dat)
@@ -88,7 +83,7 @@ inferable visited dat args = do
             _   -> return Nothing
       r@Record{}   -> inferableArgs (recCon r) (recPars r)
       f@Function{} -> do
-        term <- lift $ normalise $ Def dat args
+        term <- lift $ normalise $ Def dat $ map SI.Apply args
         inferableTerm visited' term
       d -> do
         lift $ reportSLn "epic.smashing" 10 $ "  failed (inferable): " ++ (show d)
@@ -108,8 +103,12 @@ inferable visited dat args = do
         (AA.Con tag c <$>) <$> sequence <$> forM (notForced forc $ flattenTel tel) (inferableTerm visited' . unEl . unDom)
     visited' = S.insert dat visited
 
-inferableTerm visited t = case t of
-    Def q as     -> inferable visited q as
+inferableTerm visited t = do
+  case t of
+    Def q es    ->
+      case allApplyElims es of
+        Just vs -> inferable visited q vs
+        Nothing -> return Nothing
     Pi _   b    -> (AA.Lam "_" <$>) <$> inferableTerm visited (unEl $ unAbs b)
     Sort {}     -> return . return $ AA.UNIT
     t           -> do
