@@ -2,22 +2,23 @@
 --   there.
 
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE TypeSynonymInstances #-}
 
 module Agda.Compiler.Epic.Static where
 
 import Control.Applicative
 import Control.Monad
 import Control.Monad.State
-import Control.Monad.Trans
 
 import qualified Data.Map as M
+import Data.Traversable (traverse)
 
 import Agda.Syntax.Common
-import Agda.Syntax.Internal
+import Agda.Syntax.Internal as I
 
 import Agda.TypeChecking.CompiledClause
 import Agda.TypeChecking.Monad.Base
-import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Monad.Sharing
 import Agda.TypeChecking.Reduce
@@ -48,7 +49,7 @@ evaluateCC ccs = case ccs of
             , litBranches    = M.fromList lbrs
             , catchAllBranch = cab
             }
-    Done n t   -> Done n <$> evaluateTerm t
+    Done n t   -> Done n <$> evaluate t
     Fail       -> return Fail
 
 etaExpand :: Term -> Compile TCM Term
@@ -59,42 +60,58 @@ etaExpand def@(Def n ts) = do
         toEta :: Num a => a
         toEta = fromIntegral $ len - length ts
         term  = raise toEta def `apply` [ defaultArg $ Var i [] | i <- [toEta - 1, toEta - 2 .. 0]]
-    return $ foldr (\ v t -> Lam NotHidden (Abs v t)) term $ replicate toEta "staticVar"
+    return $ foldr (\ v t -> Lam defaultArgInfo (Abs v t)) term $ replicate toEta "staticVar"
 etaExpand x = return x
 
-evaluateTerm :: Term -> Compile TCM Term
-evaluateTerm term = case term of
-    Var x as     -> Var x <$>  evaluateTerms as
-    Lam h ab     -> do
+class Evaluate a where
+  evaluate :: a -> Compile TCM a
+
+instance Evaluate a => Evaluate [a] where
+  evaluate = traverse evaluate
+
+instance Evaluate a => Evaluate (I.Arg a) where
+  evaluate = traverse evaluate
+
+instance Evaluate a => Evaluate (Abs a) where
+  evaluate = traverse evaluate  -- TODO: maintain context
+
+instance Evaluate a => Evaluate (Elim' a) where
+  evaluate = traverse evaluate
+
+instance Evaluate Term where
+  evaluate term = case term of
+    Var x es     -> Var x <$> evaluate es
+    Lam h b      -> Lam h <$> evaluate b
+{-
       ab' <- evaluateTerm (unAbs ab)
       return $ Lam h $ Abs (absName ab) ab'
+-}
     Lit l        -> return $ Lit l
-    Def n ts -> do
-        ifM (not <$> isStatic n)
-            (Def n <$> evaluateTerms ts) $ do
-                feta <- return term -- etaExpand term
-                f <- lift $ normalise feta
-                lift $ reportSDoc "epic.static" 10 $ vcat
-                  [ text "STATIC pragma fired"
-                  , nest 2 $ vcat
-                    [ text "before :" <+> prettyTCM term
-                    , text "after  :" <+> prettyTCM f
-                    ]
-                  ]
-                return f
-    Con c args   -> Con c <$> evaluateTerms args
-    Pi  arg abs  -> return $ Pi  arg abs
-    Sort s       -> return $ Sort s
-    MetaV i args -> return $ MetaV i args
-    Level l      -> return $ Level l
-    DontCare i   -> return $ DontCare i
-    Shared{}     -> updateSharedTermT evaluateTerm term
-  where
+    Def n es     -> ifNotM (isStatic n) {- then -} (Def n <$> evaluate es) $ {- else -} do
+        feta <- return term -- etaExpand term
+        f <- lift $ normalise feta
+        lift $ reportSDoc "epic.static" 10 $ vcat
+          [ text "STATIC pragma fired"
+          , nest 2 $ vcat
+            [ text "before :" <+> prettyTCM term
+            , text "after  :" <+> prettyTCM f
+            ]
+          ]
+        return f
+    Con c args   -> Con c <$> evaluate args
+    Pi  arg abs  -> return term
+    Sort s       -> return term
+    MetaV i args -> return term
+    Level l      -> return term
+    DontCare i   -> return term
+    Shared{}     -> updateSharedTermT evaluate term
+    where
+{-
     evaluateTerms :: Args -> Compile TCM Args
     evaluateTerms as = forM as $ \x -> do
       y <- evaluateTerm (unArg x)
       return x { unArg = y }
-
+-}
     isStatic :: QName -> Compile TCM Bool
     isStatic q = do
       defs <- lift (gets (sigDefinitions . stImports))
