@@ -1,53 +1,58 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE NamedFieldPuns #-}
 
 module Agda.Compiler.JS.Compiler where
 
 import Prelude hiding ( null, writeFile )
 import Control.Monad.Reader ( liftIO )
-import Control.Monad.State ( get, put )
-import Data.List ( intercalate, map, filter, isPrefixOf, concat, genericDrop, genericLength, partition )
-import Data.Set ( Set, empty, null, insert, difference, delete )
-import Data.Map ( Map, fold, singleton, fromList, toList, toAscList, insertWith, elems )
+import Data.List ( intercalate, genericLength, partition )
+import Data.Maybe ( isJust )
+import Data.Set ( Set, null, insert, difference, delete )
+import Data.Map ( fromList, elems )
 import qualified Data.Set as Set
 import qualified Data.Map as Map
 import System.Directory ( createDirectoryIfMissing )
-import System.FilePath ( pathSeparator, splitFileName, (</>) )
+import System.FilePath ( splitFileName, (</>) )
 
 import Agda.Interaction.FindFile ( findFile, findInterfaceFile )
 import Agda.Interaction.Imports ( isNewerThan )
 import Agda.Interaction.Options ( optCompileDir )
-import Agda.Syntax.Common ( Nat, Arg, unArg )
+import Agda.Syntax.Common ( Nat, unArg, namedArg )
 import Agda.Syntax.Concrete.Name ( projectRoot )
 import Agda.Syntax.Abstract.Name
-  ( ModuleName(MName), QName(QName),
+  ( ModuleName(MName), QName,
     mnameToList, qnameName, qnameModule, isInModule, nameId )
 import Agda.Syntax.Internal
   ( Name, Args, Type,
-    Clause(Clause), Pattern(VarP,DotP,LitP,ConP), Abs(Abs),
-    ClauseBody(Body,NoBody,Bind),
+    Clause, Pattern(VarP,DotP,LitP,ConP,ProjP),
+    ClauseBodyF(Body,NoBody,Bind),ClauseBody,
     Term(Var,Lam,Lit,Level,Def,Con,Pi,Sort,MetaV,DontCare,Shared),
+    unSpine, allApplyElims,
+    conName,
     derefPtr,
     toTopLevelModuleName, clausePats, clauseBody, arity, unEl, unAbs )
 import Agda.TypeChecking.Substitute ( absBody )
 import Agda.Syntax.Literal ( Literal(LitInt,LitFloat,LitString,LitChar,LitQName) )
 import Agda.TypeChecking.Level ( reallyUnLevelView )
 import Agda.TypeChecking.Monad
-  ( TCM, Definition(Defn), Definitions, Interface,
+  ( TCM, Definition(Defn), Interface,
     JSCode, Defn(Record,Datatype,Constructor,Primitive,Function,Axiom),
-    iModuleName, iImportedModules, theDef, getConstInfo, typeOfConst,
+    Projection(Projection), projProper, projFromType, projIndex,
+    iModuleName, iImportedModules, theDef, getConstInfo,
     ignoreAbstractMode, miInterface, getVisitedModules,
-    defName, defType, funClauses, funProjection,
-    dataPars, dataIxs, dataClause, dataCons,
-    conPars, conData, conSrcCon,
-    recClause, recCon, recFields, recPars, recNamedCon,
-    primClauses, defJSDef )
+    defName, defType, funClauses, funProjection, projectionArgs,
+    dataPars, dataCons,
+    conPars, conData,
+    recConHead, recFields, recNamedCon,
+    localTCState,
+    typeError, TypeError(NotImplemented),
+    defJSDef )
 import Agda.TypeChecking.Monad.Options ( setCommandLineOptions, commandLineOptions, reportSLn )
 import Agda.TypeChecking.Reduce ( instantiateFull, normalise )
 import Agda.Utils.FileName ( filePath )
 import Agda.Utils.Function ( iterate' )
-import Agda.Utils.Monad ( (<$>), (<*>), localState, ifM )
+import Agda.Utils.Monad ( (<$>), (<*>), ifM )
 import Agda.Utils.IO.UTF8 ( writeFile )
-import Agda.Utils.Impossible ( Impossible(Impossible), throwImpossible )
 import qualified Agda.Utils.HashMap as HMap
 import Agda.Compiler.MAlonzo.Misc ( curDefs, curIF, curMName, setInterface )
 import Agda.Compiler.MAlonzo.Primitives ( repl )
@@ -57,11 +62,12 @@ import Agda.Compiler.JS.Syntax
     LocalId(LocalId), GlobalId(GlobalId), MemberId(MemberId), Export(Export), Module(Module),
     modName, expName, uses )
 import Agda.Compiler.JS.Substitution
-  ( curriedLambda, curriedApply, fix, emp, object, subst, apply )
+  ( curriedLambda, curriedApply, emp, subst, apply )
 import Agda.Compiler.JS.Case ( Tag(Tag), Case(Case), Patt(VarPatt,Tagged), lambda )
 import Agda.Compiler.JS.Pretty ( pretty )
 
 #include "../../undefined.h"
+import Agda.Utils.Impossible ( Impossible(Impossible), throwImpossible )
 
 --------------------------------------------------
 -- Entry point into the compiler
@@ -70,7 +76,7 @@ import Agda.Compiler.JS.Pretty ( pretty )
 compilerMain :: Interface -> TCM ()
 compilerMain mainI =
   -- Preserve the state (the compiler modifies the state).
-  localState $ do
+  localTCState $ do
 
     -- Compute the output directory.
     opts <- commandLineOptions
@@ -200,7 +206,7 @@ insertAfter us e (f:fs) | otherwise = f : insertAfter (delete (expName f) us) e
 curModule :: TCM Module
 curModule = do
   m <- (jsMod <$> curMName)
-  is <- map jsMod <$> (iImportedModules <$> curIF)
+  is <- map jsMod <$> (map fst . iImportedModules <$> curIF)
   es <- mapM definition =<< (HMap.toList <$> curDefs)
   return (Module m (reorder es))
 
@@ -235,7 +241,16 @@ defn q ls t Nothing (Function { funProjection = proj, funClauses = cls }) = do
     Just e ->
       return (curriedLambda (arity t) e)
     Nothing -> case proj of
-      Just (p,i) -> do
+      Just Projection{ projProper, projFromType = p, projIndex = i } -> do
+        -- Andreas, 2013-05-20: whether a projection is proper is now stored.
+        if isJust projProper then
+          -- For projections from records we use a field lookup
+            return (curriedLambda (numPars cls)
+              (Lookup (Local (LocalId 0)) (last ls)))
+         else
+            -- For anything else we generate code, after adding (i-1) dummy lambdas
+            return (dummyLambda (i-1) (lambda cs))
+{- OLD way of finding out whether a projection is proper (ie. from record)
         d <- getConstInfo p
         case theDef d of
           -- For projections from records we use a field lookup
@@ -245,6 +260,7 @@ defn q ls t Nothing (Function { funProjection = proj, funClauses = cls }) = do
           _ ->
             -- For anything else we generate code, after adding (i-1) dummy lambdas
             return (dummyLambda (i-1) (lambda cs))
+-}
       Nothing ->
         return (lambda cs)
 defn q ls t (Just e) (Primitive {}) =
@@ -297,18 +313,21 @@ mapping :: [Pattern] -> (Nat,Nat,[Exp])
 mapping = foldr mapping' (0,0,[])
 
 mapping' :: Pattern -> (Nat,Nat,[Exp]) -> (Nat,Nat,[Exp])
+mapping' (ProjP _)     (av,bv,es) =
+  __IMPOSSIBLE__
 mapping' (VarP _)      (av,bv,es) = (av+1, bv+1, Local (LocalId bv) : es)
 mapping' (DotP _)      (av,bv,es) = (av+1, bv+1, Local (LocalId bv) : es)
 mapping' (ConP _ _ ps) (av,bv,es) = (av',bv'+1,es') where
-  (av',bv',es') = foldr mapping' (av,bv,es) (map unArg ps)
+  (av',bv',es') = foldr mapping' (av,bv,es) (map namedArg ps)
 mapping' (LitP _)      (av,bv,es) = (av, bv+1, es)
 
 -- Not doing literal patterns yet
 
 pattern :: Pattern -> TCM Patt
-pattern (ConP q _ ps) = do
-  l <- tag q
-  ps <- mapM (pattern . unArg) ps
+pattern (ProjP _)     = typeError $ NotImplemented $ "Compilation of copatterns"
+pattern (ConP c _ ps) = do
+  l <- tag $ conName c
+  ps <- mapM (pattern . namedArg) ps
   return (Tagged l ps)
 pattern _             = return VarPatt
 
@@ -342,53 +361,58 @@ body (Bind b) = body (unAbs b)
 body (NoBody) = return Undefined
 
 term :: Term -> TCM Exp
-term (Var   i as)         = do
-  e <- return (Local (LocalId i))
-  es <- args Nothing as
-  return (curriedApply e es)
-term (Lam   _ at)         = Lambda 1 <$> term (absBody at)
-term (Lit   l)            = return (literal l)
-term (Level l)            = term =<< reallyUnLevelView l
-term (Shared p)           = term $ derefPtr p
-term (Def q as) = do
-  d <- getConstInfo q
-  case theDef d of
-    -- Datatypes and records are erased
-    Datatype {} -> return (String "*")
-    Record {} -> return (String "*")
-    _ -> case defJSDef d of
-      -- Inline functions with an FFI definition
-      Just e -> do
-        es <- args (defProjection d) as
-        return (curriedApply e es)
-      Nothing -> do
-        t <- normalise (defType d)
-        s <- isSingleton t
-        case s of
-          -- Inline and eta-expand singleton types
-          Just e ->
-            return (curriedLambda (arity t) e)
-          -- Everything else we leave non-inline
-          Nothing -> do
-            e <- qname q
-            es <- args (defProjection d) as
+term v = do
+  case unSpine v of
+    (Var   i es)         -> do
+      let Just as = allApplyElims es
+      e <- return (Local (LocalId i))
+      es <- args 0 as
+      return (curriedApply e es)
+    (Lam   _ at)         -> Lambda 1 <$> term (absBody at)
+    (Lit   l)            -> return (literal l)
+    (Level l)            -> term =<< reallyUnLevelView l
+    (Shared p)           -> term $ derefPtr p
+    (Def q es)           -> do
+      let Just as = allApplyElims es
+      d <- getConstInfo q
+      case theDef d of
+        -- Datatypes and records are erased
+        Datatype {} -> return (String "*")
+        Record {} -> return (String "*")
+        _ -> case defJSDef d of
+          -- Inline functions with an FFI definition
+          Just e -> do
+            es <- args (projectionArgs $ theDef d) as
             return (curriedApply e es)
-term (Con q as) = do
-  d <- getConstInfo q
-  case defJSDef d of
-    -- Inline functions with an FFI definition
-    Just e -> do
-      es <- args Nothing as
-      return (curriedApply e es)
-    -- Everything else we leave non-inline
-    Nothing -> do
-      e <- qname q
-      es <- args Nothing as
-      return (curriedApply e es)
-term (Pi    _ _)          = return (String "*")
-term (Sort  _)            = return (String "*")
-term (MetaV _ _)          = return (Undefined)
-term (DontCare _)         = return (Undefined)
+          Nothing -> do
+            t <- normalise (defType d)
+            s <- isSingleton t
+            case s of
+              -- Inline and eta-expand singleton types
+              Just e ->
+                return (curriedLambda (arity t) e)
+              -- Everything else we leave non-inline
+              Nothing -> do
+                e <- qname q
+                es <- args (projectionArgs $ theDef d) as
+                return (curriedApply e es)
+    (Con con as)         -> do
+      let q = conName con
+      d <- getConstInfo q
+      case defJSDef d of
+        -- Inline functions with an FFI definition
+        Just e -> do
+          es <- args 0 as
+          return (curriedApply e es)
+        -- Everything else we leave non-inline
+        Nothing -> do
+          e <- qname q
+          es <- args 0 as
+          return (curriedApply e es)
+    (Pi    _ _)          -> return (String "*")
+    (Sort  _)            -> return (String "*")
+    (MetaV _ _)          -> return (Undefined)
+    (DontCare _)         -> return (Undefined)
 
 -- Check to see if a type is a singleton, and if so, return its only
 -- member.  Singleton types are of the form T1 -> ... -> Tn -> T where
@@ -410,21 +434,29 @@ isSingleton t = case unEl t of
         case (arity (defType c) == np) of
           True -> Just <$> qname p
           False -> return (Nothing)
-      Record { recCon = p, recFields = [] } ->
-        Just <$> qname p
+      Record { recConHead = con, recFields = [] } ->
+        Just <$> qname (conName con)
       _ -> return (Nothing)
   _              -> return (Nothing)
 
+{- OBSOLETE, see Signature.hs
 defProjection :: Definition -> Maybe (QName, Int)
 defProjection Defn { theDef = Function { funProjection = p } } = p
 defProjection _                                                = Nothing
+-}
 
-args :: Maybe(QName, Int) -> Args -> TCM [Exp]
+{-
+args :: Maybe Projection -> Args -> TCM [Exp]
 args Nothing as =
   mapM (term . unArg) as
-args (Just (q,i)) as = do
+args (Just p) as = do
   es <- mapM (term . unArg) as
-  return (replicate (i-1) Undefined ++ es)
+  return (replicate (projIndex p - 1) Undefined ++ es)
+-}
+
+args :: Int -> Args -> TCM [Exp]
+args n as = (replicate n Undefined ++) <$>
+  mapM (term . unArg) as
 
 qname :: QName -> TCM Exp
 qname q = do
