@@ -1,19 +1,17 @@
-{-# LANGUAGE CPP, PatternGuards #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE PatternGuards #-}
 
 module Agda.Compiler.MAlonzo.Compiler where
 
 import Control.Applicative
-import Control.Monad ((<=<))
 import Control.Monad.Reader
 import Control.Monad.State
-import Data.Char
 import Data.List as L
 import Data.Map as M
 import Data.Set as S
 import qualified Language.Haskell.Exts.Extension as HS
 import qualified Language.Haskell.Exts.Parser as HS
 import qualified Language.Haskell.Exts.Syntax as HS
-import System.Cmd
 import System.Directory (createDirectoryIfMissing)
 import System.FilePath hiding (normalise)
 
@@ -26,13 +24,11 @@ import Agda.Interaction.Imports
 import Agda.Interaction.Options
 import Agda.Syntax.Common
 import qualified Agda.Syntax.Concrete.Name as CN
-import Agda.Syntax.Internal
+import Agda.Syntax.Internal as I
 import Agda.Syntax.Literal
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
-import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Reduce
--- import Agda.TypeChecking.Rules.Builtin.Coinduction
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
@@ -40,15 +36,17 @@ import Agda.TypeChecking.Level (reallyUnLevelView)
 import Agda.Utils.FileName
 import Agda.Utils.Monad
 import qualified Agda.Utils.IO.UTF8 as UTF8
-import Agda.Utils.Impossible
 import qualified Agda.Utils.HashMap as HMap
 
 #include "../../undefined.h"
+import Agda.Utils.Impossible
 
-compilerMain :: Interface -> TCM ()
-compilerMain mainI =
+compilerMain :: Bool -> Interface -> TCM ()
+compilerMain modIsMain mainI =
   -- Preserve the state (the compiler modifies the state).
-  localState $ do
+  -- Andreas, 2014-03-23 But we might want to collect Benchmark info,
+  -- so use localTCState.
+  localTCState $ do
 
     -- Compute the output directory.
     opts <- commandLineOptions
@@ -65,7 +63,7 @@ compilerMain mainI =
     ignoreAbstractMode $ do
       mapM_ (compile . miInterface) =<< (M.elems <$> getVisitedModules)
       writeModule rteModule
-      callGHC mainI
+      callGHC modIsMain mainI
 
 compile :: Interface -> TCM ()
 compile i = do
@@ -96,7 +94,7 @@ imports = (++) <$> hsImps <*> imps where
              ((++) <$> importsForPrim <*> (L.map mazMod <$> mnames))
   decl m = HS.ImportDecl dummy m True False Nothing Nothing Nothing
   mnames = (++) <$> (S.elems <$> gets stImportedModules)
-                <*> (iImportedModules <$> curIF)
+                <*> (L.map fst . iImportedModules <$> curIF)
   uniq   = L.map head . group . L.sort
 
 --------------------------------------------------
@@ -125,12 +123,19 @@ definitions defs = do
 definition :: Maybe CoinductionKit -> Definition -> TCM [HS.Decl]
 -- ignore irrelevant definitions
 {- Andreas, 2012-10-02: Invariant no longer holds
-definition kit (Defn Forced     _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
-definition kit (Defn UnusedArg  _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
-definition kit (Defn NonStrict  _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
+definition kit (Defn Forced    _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
+definition kit (Defn UnusedArg _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
+definition kit (Defn NonStrict _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
 -}
-definition kit (Defn Irrelevant _ _  _ _ _ _ _ _) = return []
-definition kit (Defn _          q ty _ _ _ _ compiled d) = do
+definition kit (Defn info      q _ _ _ _ _ _ _) | isIrrelevant info = do
+  reportSDoc "malonzo.definition" 10 $
+    text "Not compiling" <+> prettyTCM q <> text "."
+  return []
+definition kit (Defn _         q ty _ _ _ _ compiled d) = do
+  reportSDoc "malonzo.definition" 10 $ vcat
+    [ text "Compiling" <+> prettyTCM q <> text ":"
+    , nest 2 $ text (show d)
+    ]
   checkTypeOfMain q ty $ do
   (infodecl q :) <$> case d of
 
@@ -177,10 +182,10 @@ definition kit (Defn _          q ty _ _ _ _ compiled d) = do
                                (HS.BDecls [])]
         ]
 
-    Axiom{}                                -> return $ fb axiomErr
-    Function{ funClauses =        cls } -> function cls
-    Primitive{ primClauses = Just cls } -> function cls
-    Primitive{ primClauses = Nothing, primName = s } -> fb <$> primBody s
+    Axiom{} -> return $ fb axiomErr
+    Primitive{ primClauses = [], primName = s } -> fb <$> primBody s
+    Primitive{ primClauses = cls } -> function cls Nothing
+    Function{ funClauses =   cls } -> function cls (exportHaskell compiled)
     Datatype{ dataPars = np, dataIxs = ni, dataClause = cl, dataCons = cs }
       | Just (HsType ty) <- compiledHaskell compiled -> do
       ccs <- concat <$> mapM checkConstructorType cs
@@ -190,7 +195,8 @@ definition kit (Defn _          q ty _ _ _ _ compiled d) = do
       (ars, cds) <- unzip <$> mapM condecl cs
       return $ tvaldecl q (dataInduction d) (maximum (np:ars) - np) (np + ni) cds cl
     Constructor{} -> return []
-    Record{ recClause = cl, recCon = c, recFields = flds } -> do
+    Record{ recClause = cl, recConHead = con, recFields = flds } -> do
+      let c = conName con
       let noFields = genericLength flds
       let ar = arity ty
       cd <- snd <$> condecl c
@@ -199,7 +205,10 @@ definition kit (Defn _          q ty _ _ _ _ compiled d) = do
 --         Just c  -> snd <$> condecl c
       return $ tvaldecl q Inductive noFields ar [cd] cl
   where
-  function cls = mkwhere <$> mapM (clause q) (tag 0 cls)
+  function cls (Just (HsExport t name)) =
+    (HS.TypeSig dummy [HS.Ident name] (fakeType t) :) <$>
+    mkwhere <$> mapM (clause q (Just name)) (tag 0 cls)
+  function cls Nothing = mkwhere <$> mapM (clause q Nothing) (tag 0 cls)
   tag _ []       = []
   tag i [cl]     = (i, True , cl): []
   tag i (cl:cls) = (i, False, cl): tag (i + 1) cls
@@ -248,8 +257,8 @@ conArityAndPars q = do
       n = genericLength (telToList tel)
   return (n - np, np)
 
-clause :: QName -> (Nat, Bool, Clause) -> TCM HS.Decl
-clause q (i, isLast, Clause{ clausePats = ps, clauseBody = b }) =
+clause :: QName -> Maybe String -> (Nat, Bool, Clause) -> TCM HS.Decl
+clause q maybeName (i, isLast, Clause{ namedClausePats = ps, clauseBody = b }) =
   HS.FunBind . (: cont) <$> main where
   main = match <$> argpatts ps (bvars b (0::Nat)) <*> clausebody b
   cont | isLast && any isCon ps = [match (L.map HS.PVar cvs) failrhs]
@@ -259,25 +268,26 @@ clause q (i, isLast, Clause{ clausePats = ps, clauseBody = b }) =
   crhs = hsCast$ L.foldl HS.App (hsVarUQ $ dsubname q (i + 1)) (L.map hsVarUQ cvs)
   failrhs = rtmIncompleteMatch q  -- Andreas, 2011-11-16 call to RTE instead of inlined error
 --  failrhs = rtmError $ "incomplete pattern matching: " ++ show q
-  match hps rhs = HS.Match dummy (dsubname q i) hps Nothing
+  match hps rhs = HS.Match dummy (maybe (dsubname q i) HS.Ident maybeName) hps Nothing
                            (HS.UnGuardedRhs rhs) (HS.BDecls [])
   bvars (Body _)           _ = []
   bvars (Bind (Abs _ b'))  n = HS.PVar (ihname "v" n) : bvars b' (n + 1)
   bvars (Bind (NoAbs _ b)) n = HS.PWildCard : bvars b n
   bvars NoBody             _ = repeat HS.PWildCard -- ?
 
-  isCon (Arg _ _ ConP{}) = True
-  isCon _                = False
+  isCon (Arg _ (Named _ ConP{})) = True
+  isCon _                        = False
 
 -- argpatts aps xs = hps
 -- xs is alist of haskell *variables* in form of patterns (because of wildcard)
-argpatts :: [Arg Pattern] -> [HS.Pat] -> TCM [HS.Pat]
+argpatts :: [I.NamedArg Pattern] -> [HS.Pat] -> TCM [HS.Pat]
 argpatts ps0 bvs = evalStateT (mapM pat' ps0) bvs
   where
+  pat   (ProjP _  ) = lift $ typeError $ NotImplemented $ "Compilation of copatterns"
   pat   (VarP _   ) = do v <- gets head; modify tail; return v
   pat   (DotP _   ) = pat (VarP dummy) -- WHY NOT: return HS.PWildCard -- SEE ABOVE
   pat   (LitP l   ) = return $ HS.PLit $ hslit l
-  pat p@(ConP q _ ps) = do
+  pat p@(ConP c _ ps) = do
     -- Note that irr is applied once for every subpattern, so in the
     -- worst case it is quadratic in the size of the pattern. I
     -- suspect that this will not be a problem in practice, though.
@@ -286,27 +296,34 @@ argpatts ps0 bvs = evalStateT (mapM pat' ps0) bvs
                 then HS.PParen . HS.PIrrPat
                 else id
     (tilde . HS.PParen) <$>
-      (HS.PApp <$> lift (conhqn q) <*> mapM pat' ps)
+      (HS.PApp <$> lift (conhqn $ conName c) <*> mapM pat' ps)
+
+  {- Andreas, 2013-02-15 this triggers Issue 794,
+     because it fails to count the variables bound in p,
+     thus, the following variables bound by patterns do
+     not correspond to the according rhs-variables.
 
   -- Andreas, 2010-09-29
   -- do not match against irrelevant stuff
-  pat' (Arg _ Irrelevant _) = return $ HS.PWildCard
-  pat' (Arg _ _          p) = pat p
+  pat' a | isIrrelevant a = return $ HS.PWildCard
+-}
+  pat' a = pat $ namedArg a
 
   tildesEnabled = False
 
   -- | Is the pattern irrefutable?
   irr :: Pattern -> TCM Bool
+  irr (ProjP {})  = __IMPOSSIBLE__
   irr (VarP {})   = return True
   irr (DotP {})   = return True
   irr (LitP {})   = return False
-  irr (ConP q _ ps) =
-    (&&) <$> singleConstructorType q
+  irr (ConP c _ ps) =
+    (&&) <$> singleConstructorType (conName c)
          <*> (andM $ L.map irr' ps)
 
   -- | Irrelevant patterns are naturally irrefutable.
-  irr' (Arg _ Irrelevant _) = return $ True
-  irr' (Arg _ _          p) = irr p
+  irr' a | isIrrelevant a = return $ True
+  irr' a = irr $ namedArg a
 
 clausebody :: ClauseBody -> TCM HS.Exp
 clausebody b0 = runReaderT (go b0) 0 where
@@ -320,13 +337,20 @@ clausebody b0 = runReaderT (go b0) 0 where
 --   Types are extracted as @()@.
 --   @DontCare@ outside of irrelevant arguments is extracted as @error@.
 term :: Term -> ReaderT Nat TCM HS.Exp
-term tm0 = case ignoreSharing tm0 of
-  Var   i as -> do n <- ask; apps (hsVarUQ $ ihname "v" (n - i - 1)) as
+term tm0 = case unSpine $ ignoreSharing tm0 of
+  Var   i es -> do
+    let Just as = allApplyElims es
+    n <- ask
+    apps (hsVarUQ $ ihname "v" (n - i - 1)) as
   Lam   _ at -> do n <- ask; HS.Lambda dummy [HS.PVar $ ihname "v" n] <$>
                               local (1+) (term $ absBody at)
   Lit   l    -> lift $ literal l
-  Def   q as -> (`apps` as) . HS.Var =<< lift (xhqn "d" q)
-  Con   q as -> do
+  Def   q es -> do
+    let Just as = allApplyElims es
+    q <- lift $ xhqn "d" q
+    HS.Var q `apps` as
+  Con   c as -> do
+    let q = conName c
     kit <- lift coinductionKit
     if Just q == (nameOfSharp <$> kit)
       then (`apps` as) . HS.Var =<< lift (xhqn "d" q)
@@ -340,9 +364,9 @@ term tm0 = case ignoreSharing tm0 of
   where apps =  foldM (\h a -> HS.App h <$> term' a)
 
 -- | Irrelevant arguments are replaced by Haskells' ().
-term' :: Arg Term -> ReaderT Nat TCM HS.Exp
-term' (Arg _ Irrelevant _) = return HS.unit_con
-term' (Arg _ _          t) = term t
+term' :: I.Arg Term -> ReaderT Nat TCM HS.Exp
+term' a | isIrrelevant a = return HS.unit_con
+term' a = term $ unArg a
 
 literal :: Literal -> TCM HS.Exp
 literal l = case l of
@@ -509,8 +533,8 @@ outFile m = snd <$> outFile' m
 outFile_ :: TCM FilePath
 outFile_ = outFile =<< curHsMod
 
-callGHC :: Interface -> TCM ()
-callGHC i = do
+callGHC :: Bool -> Interface -> TCM ()
+callGHC modIsMain i = do
   setInterface i
   mdir          <- compileDir
   hsmod         <- prettyPrint <$> curHsMod
@@ -522,14 +546,13 @@ callGHC i = do
   opts       <- optGhcFlags <$> commandLineOptions
 
   let overridableArgs =
-        [ "-O"
-        , "-o", mdir </> show outputName
-        , "-Werror"
-        ]
+        [ "-O"] ++
+        (if modIsMain then ["-o", mdir </> show outputName] else []) ++
+        [ "-Werror"]
       otherArgs       =
-        [ "-i" ++ mdir
-        , "-main-is", hsmod
-        , fp
+        [ "-i" ++ mdir] ++
+        (if modIsMain then ["-main-is", hsmod] else []) ++
+        [ fp
         , "--make"
         , "-fwarn-incomplete-patterns"
         , "-fno-warn-overlapping-patterns"
