@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP #-}
+
 module Agda.Compiler.MAlonzo.Primitives where
 
 import Control.Monad.Reader
@@ -16,13 +17,12 @@ import Agda.Syntax.Internal
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Reduce
-import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Pretty
 import Agda.Utils.Monad
-import Agda.Utils.Impossible
 import qualified Agda.Utils.HashMap as HMap
 
 #include "../../undefined.h"
+import Agda.Utils.Impossible
 
 {- OLD
 -- | Check that the main function has type IO a, for some a.
@@ -117,12 +117,16 @@ declsForPrim = xForPrim $
       in do
         ds <- decls ["ZERO", "SUC"] to totxt from fromtxt
         let rule name = HS.Rule name HS.AlwaysActive (Just [HS.RuleVar $ HS.Ident "x"])
-            var = HS.Var . HS.UnQual . HS.Ident
+            qname = HS.UnQual . HS.Ident
+            var   = HS.Var . qname
             (%) = HS.App
         return $ [HS.RulePragmaDecl dummy
                   [ rule (to ++ "-" ++ from) (var to   % (var from % var "x")) (var "x")
                   , rule (from ++ "-" ++ to) (var from % (var to   % var "x")) (var "x")
-                  ]] ++ ds
+                  ],
+                  HS.InlineSig dummy True (HS.ActiveFrom 2) (qname to),
+                  HS.InlineSig dummy True (HS.ActiveFrom 2) (qname from)
+                  ] ++ ds
     decls cs n1 b1 n2 b2 =
       do cs' <- mapM pconName cs
          return $ zipWith (\ n -> fakeDS n . repl cs') [n1, n2] [b1, b2]
@@ -146,7 +150,7 @@ xForPrim table = do
   qs <- HMap.keys <$> curDefs
   bs <- toList <$> gets stBuiltinThings
   let getName (Builtin (Def q _))    = q
-      getName (Builtin (Con q _))    = q
+      getName (Builtin (Con q _))    = conName q
       getName (Builtin (Shared p))   = getName (Builtin $ derefPtr p)
       getName (Builtin _)            = __IMPOSSIBLE__
       getName (Prim (PrimFun q _ _)) = q
@@ -172,6 +176,11 @@ primBody s = maybe unimplemented (either (hsVarUQ . HS.Ident) id <$>) $
   , "primNatToInteger"   |-> bltQual' "NATURAL" mazNatToInteger
   , "primShowInteger"    |-> return "(show :: Integer -> String)"
 
+  -- Levels
+  , "primLevelZero"   |-> return "()"
+  , "primLevelSuc"    |-> return "(\\ _ -> ())"
+  , "primLevelMax"    |-> return "(\\ _ _ -> ())"
+
   -- Natural number functions
   , "primNatPlus"     |-> binNat "(+)"
   , "primNatMinus"    |-> binNat "(-)"
@@ -258,10 +267,8 @@ primBody s = maybe unimplemented (either (hsVarUQ . HS.Ident) id <$>) $
   opty t = t ++ "->" ++ t ++ "->" ++ t
   unimplemented = typeError $ NotImplemented s
 
-  lam x t = Lam Hidden (Abs x t)
-{- UNUSED
-  var x   = Arg Hidden Relevant (Var x [])
--}
+  lam x t = Lam (setHiding Hidden defaultArgInfo) (Abs x t)
+
 
 ----------------------
 
@@ -273,14 +280,14 @@ repl subs = go where
 
 pconName :: String -> TCM String
 pconName s = toS =<< getBuiltin s where
-  toS (Con q _) = prettyPrint <$> conhqn q
+  toS (Con q _) = prettyPrint <$> conhqn (conName q)
   toS (Lam _ t) = toS (unAbs t)
   toS _ = mazerror $ "pconName" ++ s
 
 hasCompiledData :: [String] -> TCM Bool
 hasCompiledData (s:_) = toB =<< getBuiltin s where
   toB (Con q _)         = do
-    def <- getConstInfo =<< ignoreAbstractMode (canonicalName q)
+    def <- getConstInfo =<< ignoreAbstractMode (canonicalName (conName q))
     return $ case compiledHaskell $ defCompiledRep def of
       Just{}  -> True
       Nothing -> False
