@@ -1,27 +1,28 @@
-{-# LANGUAGE CPP, FlexibleContexts, RelaxedPolyRec #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE RelaxedPolyRec #-}
 
 -- | Generates data used for precise syntax highlighting.
 
 module Agda.Interaction.Highlighting.Generate
   ( Level(..)
   , generateAndPrintSyntaxInfo
-  , generateTokenInfo
-  , printErrorInfo
+  , generateTokenInfo, generateTokenInfoFromString
+  , printErrorInfo, errorHighlighting
   , printUnsolvedInfo
   , printHighlightingInfo
   , highlightAsTypeChecked
+  , computeUnsolvedMetaWarnings
+  , computeUnsolvedConstraints
   , Agda.Interaction.Highlighting.Generate.tests
   ) where
 
 import Agda.Interaction.FindFile
 import Agda.Interaction.Response (Response(Resp_HighlightingInfo))
-import Agda.Interaction.Highlighting.Emacs   hiding (tests)
 import Agda.Interaction.Highlighting.Precise hiding (tests)
 import Agda.Interaction.Highlighting.Range   hiding (tests)
-import Agda.Interaction.EmacsCommand
 import qualified Agda.TypeChecking.Errors as E
 import Agda.TypeChecking.MetaVars (isBlockedTerm)
-import Agda.TypeChecking.Monad.Options (reportSLn, reportSDoc)
 import Agda.TypeChecking.Monad
   hiding (MetaInfo, Primitive, Constructor, Record, Function, Datatype)
 import qualified Agda.TypeChecking.Monad as M
@@ -37,8 +38,6 @@ import qualified Agda.Syntax.Literal as L
 import qualified Agda.Syntax.Parser as Pa
 import qualified Agda.Syntax.Parser.Tokens as T
 import qualified Agda.Syntax.Position as P
-import qualified Agda.Syntax.Scope.Base as S
-import qualified Agda.Syntax.Translation.ConcreteToAbstract as CA
 import Agda.Utils.List
 import Agda.Utils.TestHelpers
 import Agda.Utils.HashMap (HashMap)
@@ -50,23 +49,17 @@ import Control.Monad.Reader
 import Control.Applicative
 import Control.Arrow ((***))
 import Data.Monoid
-import Data.Function
 import Data.Generics.Geniplate
 import Agda.Utils.FileName
-import qualified Agda.Utils.IO.UTF8 as UTF8
-import Agda.Utils.Monad
 import Data.HashSet (HashSet)
 import qualified Data.HashSet as HSet
-import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Maybe
 import Data.List ((\\), isPrefixOf)
 import qualified Data.Foldable as Fold (toList, fold, foldMap)
-import System.Directory
-import System.IO
 
-import Agda.Utils.Impossible
 #include "../../undefined.h"
+import Agda.Utils.Impossible
 
 -- | @highlightAsTypeChecked rPre r m@ runs @m@ and returns its
 -- result. Some code may additionally be highlighted:
@@ -108,12 +101,13 @@ highlightAsTypeChecked rPre r m
 
 printHighlightingInfo :: MonadTCM tcm => HighlightingInfo -> tcm ()
 printHighlightingInfo x = do
+  modToSrc <- gets stModuleToSource
   liftTCM $ reportSLn "highlighting" 50 $
-    "Printing highlighting info:\n" ++ show x
+    "Printing highlighting info:\n" ++ show x ++ "\n" ++
+    "  modToSrc = " ++ show modToSrc
   unless (null $ ranges x) $ do
-    st <- get
-    liftTCM $ stInteractionOutputCallback st
-                (Resp_HighlightingInfo x (stModuleToSource st))
+    liftTCM $ appInteractionOutputCallback $
+        Resp_HighlightingInfo x modToSrc
 
 -- | Highlighting levels.
 
@@ -143,6 +137,7 @@ data Level
 -- 'stTokens'.
 
 generateAndPrintSyntaxInfo :: A.Declaration -> Level -> TCM ()
+generateAndPrintSyntaxInfo decl _ | P.noRange == P.getRange decl = return ()
 generateAndPrintSyntaxInfo decl hlLevel = do
   file <- envCurrentPath <$> ask
 
@@ -227,14 +222,18 @@ generateAndPrintSyntaxInfo decl hlLevel = do
     , Fold.foldMap getLam         $ universeBi decl
     , Fold.foldMap getTyped       $ universeBi decl
     , Fold.foldMap getPattern     $ universeBi decl
+    , Fold.foldMap getExpr        $ universeBi decl
+    , Fold.foldMap getPatSynArgs  $ universeBi decl
     , Fold.foldMap getModuleName  $ universeBi decl
     , Fold.foldMap getModuleInfo  $ universeBi decl
+    , Fold.foldMap getNamedArg    $ universeBi decl
     ]
     where
-    bound n = nameToFile modMap file []
-                         (A.nameConcrete n)
+    bound n = nameToFile modMap file [] (A.nameConcrete n)
                          (\isOp -> mempty { aspect = Just $ Name (Just Bound) isOp })
                          (Just $ A.nameBindingSite n)
+    patsyn n = nameToFileA modMap file n True $ \isOp ->
+                  mempty { aspect = Just $ Name (Just $ Constructor SC.Inductive) isOp }
     field m n = nameToFile modMap file m n
                            (\isOp -> mempty { aspect = Just $ Name (Just Field) isOp })
                            Nothing
@@ -254,6 +253,11 @@ generateAndPrintSyntaxInfo decl hlLevel = do
     getVarAndField (A.Rec _ fs) = mconcat $ map (field [] . fst) fs
     getVarAndField _            = mempty
 
+    -- Ulf, 2014-04-09: It would be nicer to have it on Named_ a, but
+    -- you can't have polymorphic functions in universeBi.
+    getNamedArg :: SC.RString -> File
+    getNamedArg x = singleton (rToR $ P.getRange x) mempty{ aspect = Just $ Name (Just Argument) False }
+
     getLet :: A.LetBinding -> File
     getLet (A.LetBind _ _ x _ _) = bound x
     getLet A.LetPatBind{}        = mempty
@@ -261,12 +265,16 @@ generateAndPrintSyntaxInfo decl hlLevel = do
     getLet A.LetOpen{}           = mempty
 
     getLam :: A.LamBinding -> File
-    getLam (A.DomainFree _ _ x) = bound x
+    getLam (A.DomainFree _ x) = bound x
     getLam (A.DomainFull {})  = mempty
 
     getTyped :: A.TypedBinding -> File
     getTyped (A.TBind _ xs _) = mconcat $ map bound xs
-    getTyped (A.TNoBind {})   = mempty
+    getTyped A.TLet{}         = mempty
+
+    getPatSynArgs :: A.Declaration -> File
+    getPatSynArgs (A.PatternSynDef _ xs _) = mconcat $ map (bound . SC.unArg) xs
+    getPatSynArgs _                        = mempty
 
     getPattern :: A.Pattern -> File
     getPattern (A.VarP x)    = bound x
@@ -274,8 +282,13 @@ generateAndPrintSyntaxInfo decl hlLevel = do
     getPattern (A.DotP pi _) =
       singleton (rToR $ P.getRange pi)
                 (mempty { otherAspects = [DottedPattern] })
+    getPattern (A.PatternSynP _ q _) = patsyn q
     getPattern _             = mempty
 
+    getExpr :: A.Expr -> File
+    getExpr (A.PatternSyn q) = patsyn q
+    getExpr _                = mempty
+
     getFieldDecl :: A.Declaration -> File
     getFieldDecl (A.RecDef _ _ _ _ _ _ fs) = Fold.foldMap extractField fs
       where
@@ -312,9 +325,20 @@ generateAndPrintSyntaxInfo decl hlLevel = do
 generateTokenInfo
   :: AbsolutePath          -- ^ The module to highlight.
   -> TCM CompressedFile
-generateTokenInfo file = do
-  tokens <- liftIO $ Pa.parseFile' Pa.tokensParser file
-  return $ merge $ map tokenToCFile tokens
+generateTokenInfo file =
+  liftIO $ tokenHighlighting <$> Pa.parseFile' Pa.tokensParser file
+
+-- | Same as 'generateTokenInfo' but takes a string instead of a filename.
+generateTokenInfoFromString :: P.Range -> String -> TCM CompressedFile
+generateTokenInfoFromString r _ | r == P.noRange = return mempty
+generateTokenInfoFromString r s = do
+  liftIO $ tokenHighlighting <$> Pa.parsePosString Pa.tokensParser p s
+  where
+    Just p = P.rStart r
+
+-- | Compute syntax highlighting for the given tokens.
+tokenHighlighting :: [T.Token] -> CompressedFile
+tokenHighlighting = merge . map tokenToCFile
   where
   -- Converts an aspect and a range to a file.
   aToF a r = singletonC (rToR r) (mempty { aspect = Just a })
@@ -358,11 +382,11 @@ nameKinds hlLevel decl = do
   imported <- fix . stImports <$> get
   local    <- case hlLevel of
     Full _ -> fix . stSignature <$> get
-    _      -> return $
+    _      -> return HMap.empty
       -- Traverses the syntax tree and constructs a map from qualified
       -- names to name kinds. TODO: Handle open public.
-      foldr ($) HMap.empty $ map declToKind $ universeBi decl
-  let merged = HMap.union local imported
+  let syntax = foldr ($) HMap.empty $ map declToKind $ universeBi decl
+  let merged = HMap.unions [local, imported, syntax]
   return (\n -> HMap.lookup n merged)
   where
   fix = HMap.map (defnToKind . theDef) . sigDefinitions
@@ -378,21 +402,18 @@ nameKinds hlLevel decl = do
     dropPostulates k         _ = k
 
   defnToKind :: Defn -> NameKind
-  defnToKind (M.Axiom {})                     = Postulate
-  defnToKind (M.Function {})                  = Function
-  defnToKind (M.Datatype {})                  = Datatype
-  defnToKind (M.Record {})                    = Record
-  defnToKind (M.Constructor { M.conInd = i }) = Constructor i
-  defnToKind (M.Primitive {})                 = Primitive
-
-  getAxiomName :: A.Declaration -> A.QName
-  getAxiomName (A.Axiom _ _ q _) = q
-  getAxiomName _                 = __IMPOSSIBLE__
+  defnToKind   M.Axiom{}                           = Postulate
+  defnToKind d@M.Function{} | isProperProjection d = Field
+                            | otherwise            = Function
+  defnToKind   M.Datatype{}                        = Datatype
+  defnToKind   M.Record{}                          = Record
+  defnToKind   M.Constructor{ M.conInd = i }       = Constructor i
+  defnToKind   M.Primitive{}                       = Primitive
 
   declToKind :: A.Declaration ->
                 HashMap A.QName NameKind -> HashMap A.QName NameKind
-  declToKind (A.Axiom _ _ q _)      = insert q Postulate
-  declToKind (A.Field _ q _)        = insert q Function
+  declToKind (A.Axiom _ _ _ q _)    = insert q Postulate
+  declToKind (A.Field _ q _)        = insert q Field -- Function
     -- Note that the name q can be used both as a field name and as a
     -- projection function. Highlighting of field names is taken care
     -- of by "theRest" above, which does not use NameKinds.
@@ -404,11 +425,12 @@ nameKinds hlLevel decl = do
   declToKind (A.Pragma {})          = id
   declToKind (A.ScopedDecl {})      = id
   declToKind (A.Open {})            = id
+  declToKind (A.PatternSynDef q _ _) = insert q (Constructor SC.Inductive)
   declToKind (A.FunDef  _ q _ _)    = insert q Function
   declToKind (A.DataSig _ q _ _)    = insert q Datatype
   declToKind (A.DataDef _ q _ cs)   = \m ->
                                       insert q Datatype $
-                                      foldr (\d -> insert (getAxiomName d)
+                                      foldr (\d -> insert (A.axiomName d)
                                                           (Constructor SC.Inductive))
                                             m cs
   declToKind (A.RecSig _ q _ _)     = insert q Record
@@ -447,7 +469,8 @@ generateConstructorInfo modMap file kinds decl = do
       terms    = universeBi (types, clauses)
 
       -- Find all constructors in the patterns and terms.
-      constrs = concatMap getConstructorP patterns ++
+      constrs = filter ((/= P.noRange) . P.getRange) $
+                concatMap getConstructorP patterns ++
                 concatMap getConstructor  terms
 
   -- Find all constructors in right-hand sides of delayed definitions.
@@ -513,33 +536,32 @@ generateConstructorInfo modMap file kinds decl = do
     getRHS (I.Bind b)   = getRHS (I.unAbs b)
 
   getConstructorP :: I.Pattern -> [A.QName]
-  getConstructorP (I.ConP q _ _) = [q]
+  getConstructorP (I.ConP c _ _) = [I.conName c]
   getConstructorP _              = []
 
   getConstructor :: I.Term -> [A.QName]
-  getConstructor (I.Con q _) = [q]
+  getConstructor (I.Con q _) = [I.conName q]
   getConstructor _           = []
 
 -- | Prints syntax highlighting info for an error.
 
 printErrorInfo :: TCErr -> TCM ()
 printErrorInfo e = do
-  file <- envCurrentPath <$> ask
-  let r = P.getRange e
-  case P.rStart r of
-    Just x | P.srcFile x == Just file -> do
-      s <- E.prettyError e
+  -- Erase previous highlighting.
+  printHighlightingInfo $ singletonC (rToR $ P.continuousPerLine $ P.getRange e) mempty
 
-      -- Erase previous highlighting.
-      p (P.continuousPerLine r) mempty
+  -- Print new highlighting.
+  printHighlightingInfo . compress =<< errorHighlighting e
 
-      -- Print new highlighting.
-      p r $ mempty { otherAspects = [Error]
-                   , note         = Just s
-                   }
-    _ -> internalError . ("invalid range when printing error: " ++) =<< E.prettyError e
+errorHighlighting :: TCErr -> TCM File
+errorHighlighting e = do
+  s <- E.prettyError e
+  return $ singleton (rToR r)
+         $ mempty { otherAspects = [Error]
+                  , note         = Just s
+                  }
   where
-  p r x = printHighlightingInfo $ singletonC (rToR r) x
+    r = P.getRange e
 
 -- | Generates and prints syntax highlighting information for unsolved
 -- meta-variables and certain unsolved constraints.
@@ -593,9 +615,15 @@ generate modMap file kinds (A.AmbQ qs) =
   mconcat $ map (\q -> nameToFileA modMap file q include m) qs
   where
     ks   = map kinds qs
-    kind = case (allEqual ks, ks) of
-             (True, Just k : _) -> Just k
-             _                  -> Nothing
+    -- Ulf, 2014-06-03: [issue1064] It's better to pick the first rather
+    -- than doing no highlighting if there's an ambiguity between an
+    -- inductive and coinductive constructor.
+    kind = case [ k | Just k <- ks ] of
+             k : _ -> Just k
+             []    -> Nothing
+    -- kind = case (allEqual ks, ks) of
+    --          (True, Just k : _) -> Just k
+    --          _                  -> Nothing
     -- Note that all names in an AmbiguousQName should have the same
     -- concrete name, so either they are all operators, or none of
     -- them are.
