@@ -1,4 +1,5 @@
-{-# LANGUAGE CPP, ViewPatterns #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE ViewPatterns #-}
 
 -- | Function for generating highlighted and aligned LaTeX from literate
 -- Agda source.
@@ -7,19 +8,17 @@ module Agda.Interaction.Highlighting.LaTeX
   ( generateLaTeX
   ) where
 
+import Prelude hiding (log)
 import Data.Char
 import Data.Maybe
 import Data.Function
-import Data.List
 import Control.Monad.RWS
 import Control.Monad.Error
-import Control.Monad.Trans
 import System.Directory
 import System.FilePath
-import System.Process
-import System.Exit
 import Data.Text (Text)
 import qualified Data.Text          as T
+import qualified Data.Text.IO       as T
 import qualified Data.Text.Encoding as E
 import qualified Data.ByteString    as BS
 
@@ -29,13 +28,16 @@ import qualified Data.Map  as Map
 import Paths_Agda
 
 import Agda.Syntax.Common
-import qualified Agda.Syntax.Abstract as A
+import Agda.Syntax.Concrete (TopLevelModuleName, moduleNameParts)
+import qualified Agda.Interaction.FindFile as Find
 import Agda.Interaction.Highlighting.Precise
 import Agda.TypeChecking.Monad (TCM)
 import qualified Agda.TypeChecking.Monad as TCM
 import Agda.Interaction.Options
 import Agda.Compiler.CallCompiler
 import qualified Agda.Utils.IO.UTF8 as UTF8
+import Agda.Utils.FileName (filePath)
+import Agda.Utils.Pretty (pretty, render)
 
 #include "../../undefined.h"
 import Agda.Utils.Impossible
@@ -43,11 +45,12 @@ import Agda.Utils.Impossible
 ------------------------------------------------------------------------
 -- * Datatypes.
 
--- | The @LaTeX@ monad is a combination of @ErrorT@ and @RWS@. The error
--- part is just used to keep track whether we finished or not, the
--- reader part isn't used, the writer is where the output goes and the
--- state is for keeping track of the tokens and some other useful info.
-type LaTeX = ErrorT String (RWS () Text State)
+-- | The @LaTeX@ monad is a combination of @ErrorT@, @RWST@ and
+-- @IO@. The error part is just used to keep track whether we finished
+-- or not, the reader part isn't used, the writer is where the output
+-- goes and the state is for keeping track of the tokens and some other
+-- useful info, and the I/O part is used for printing debugging info.
+type LaTeX = ErrorT String (RWST () Text State IO)
 
 data State = State
   { tokens     :: Tokens
@@ -56,21 +59,26 @@ data State = State
   , indentPrev :: Int
   , inCode     :: Bool    -- ^ Keeps track of whether we are in a code
                           -- block or not.
+  , debugs     :: [Debug] -- ^ Says what debug information should printed.
   }
 
 type Tokens = [Token]
 
 data Token = Token
-  { string   :: Text
+  { text     :: Text
   , info     :: MetaInfo
   , position :: Integer  -- ^ Is not used currently, but could
                          -- potentially be used for hyperlinks as in
                          -- the HTML output?
   }
+  deriving Show
+
+data Debug = MoveColumn | NonCode | Code | Spaces | Output
+  deriving (Eq, Show)
 
 -- | Run function for the @LaTeX@ monad.
-runLaTeX :: LaTeX a -> () -> State -> (Either String a, State, Text)
-runLaTeX = runRWS . runErrorT
+runLaTeX :: LaTeX a -> () -> State -> IO (Either String a, State, Text)
+runLaTeX = runRWST . runErrorT
 
 emptyState :: State
 emptyState = State
@@ -79,6 +87,7 @@ emptyState = State
   , indent     = 0
   , indentPrev = 0
   , inCode     = False
+  , debugs     = []
   }
 
 ------------------------------------------------------------------------
@@ -93,12 +102,19 @@ isInfixOf' needle haystack = go (T.tails haystack) 0
   go ((T.stripPrefix needle -> Just suf) : xss) n = Just (T.take n haystack, suf)
   go (_                                  : xss) n = go xss (n + 1)
 
-isSpaces :: Text -> Bool
-isSpaces (T.uncons -> Nothing)     = True
-isSpaces (T.uncons -> Just (c, s)) | isSpace c = isSpaces s
-                                   | otherwise = False
+-- Same as above, but starts searching from the back rather than the
+-- front.
+isInfixOfRev :: Text -> Text -> Maybe (Text, Text)
+isInfixOfRev needle haystack
+  = case T.reverse needle `isInfixOf'` T.reverse haystack of
+      Nothing         -> Nothing
+      Just (pre, suf) -> Just (T.reverse suf, T.reverse pre)
 
-isSpaces _                         = __IMPOSSIBLE__
+isSpaces :: Text -> Bool
+isSpaces = T.all isSpace
+
+isActualSpaces :: Text -> Bool
+isActualSpaces = T.all (== ' ')
 
 -- | Yields the next token, taking special care to begin/end code
 -- blocks. Junk occuring before and after the code blocks is separated
@@ -111,8 +127,8 @@ nextToken' = do
     []     -> throwError "Done"
 
     -- Clean begin/end code block or a LaTeX comment.
-    t : ts | string t == beginCode || string t == endCode ||
-             T.singleton '%' == T.take 1 (T.stripStart (string t)) -> do
+    t : ts | text t == beginCode || text t == endCode ||
+             T.singleton '%' == T.take 1 (T.stripStart (text t)) -> do
 
       modify $ \s -> s { tokens = ts }
       return t
@@ -123,27 +139,54 @@ nextToken' = do
       inCode <- gets inCode
       let code = if inCode then endCode else beginCode
 
-      case code `isInfixOf'` string t of
+      case code `isInfixOf'` text t of
         Nothing -> do
 
           -- Spaces take care of their own column tracking.
-          unless (isSpaces (string t)) $ do
-            moveColumn $ T.length $ string t
+          unless (isSpaces (text t)) $ do
+            log MoveColumn $ text t
+            moveColumn $ T.length $ text t
 
           return t
 
         Just (pre, suf) -> do
-          let t'  = t { string = code }
-          let t'' = t { string = suf }
-          modify $ \s -> s { tokens = t' : t'' : tokens s }
+
+          let (textToReturn, textsToPutBack) =
+
+               -- This bit fixes issue 954.
+
+               -- Drop spaces up until and including the first trailing
+               -- newline after begin code blocks.
+               if code == beginCode && isSpaces suf
+               then case T.singleton '\n' `isInfixOf'` suf of
+                     Nothing        -> (pre, [ beginCode ])
+                     Just (_, suf') -> (pre, [ beginCode, suf' ])
+
+               -- Do the converse thing for end code blocks.
+               else if code == endCode && isSpaces pre
+                    then case T.singleton '\n' `isInfixOfRev` pre of
+                           Nothing           -> (code, [ suf ])
+                           Just (pre', suf') ->
+                             (pre' <+> T.dropWhile (`elem` [' ', '\t']) suf',
+                              [ code, suf ])
+
+              -- This case happens for example when you have two code
+              -- blocks after each other, i.e. the begin code of the
+              -- second ends up in the suffix of the first's end code.
+                    else (pre, [ code, suf ])
+
+          let tokToReturn   = t { text = textToReturn }
+          let toksToPutBack = map (\txt -> t { text = txt }) textsToPutBack
 
           unless (isSpaces pre) $ do
+            log MoveColumn pre
             moveColumn $ T.length pre
 
-          return $ t { string = pre }
+          modify $ \s -> s { tokens = toksToPutBack ++ tokens s }
+          return tokToReturn
 
 nextToken :: LaTeX Text
-nextToken = string `fmap` nextToken'
+nextToken = text `fmap` nextToken'
 
 resetColumn :: LaTeX ()
 resetColumn = modify $ \s -> s { column = 0 }
@@ -151,8 +194,8 @@ resetColumn = modify $ \s -> s { column = 0 }
 moveColumn :: Int -> LaTeX ()
 moveColumn i = modify $ \s -> s { column = i + column s }
 
-moveIndent :: Int -> LaTeX ()
-moveIndent i = modify $ \s -> s { indent = i + indent s }
+setIndent :: Int -> LaTeX ()
+setIndent i = modify $ \s -> s { indent = i }
 
 resetIndent :: LaTeX ()
 resetIndent = modify $ \s -> s { indent = 0 }
@@ -169,6 +212,34 @@ setInCode = modify $ \s -> s { inCode = True }
 unsetInCode :: LaTeX ()
 unsetInCode = modify $ \s -> s { inCode = False }
 
+logHelper :: Debug -> Text -> [String] -> LaTeX ()
+logHelper debug text extra = do
+  debugs <- gets debugs
+  when (debug `elem` debugs) $ do
+    lift $ lift $ T.putStrLn $ T.pack (show debug ++ ": ") <+>
+      T.pack "'" <+> text <+> T.pack "' " <+>
+      if null extra
+         then T.empty
+         else T.pack "(" <+> T.pack (unwords extra) <+> T.pack ")"
+
+log :: Debug -> Text -> LaTeX ()
+log MoveColumn text = do
+  ind <- gets indent
+  logHelper MoveColumn text ["ind=", show ind]
+log Code text = do
+  ind <- gets indent
+  col <- gets column
+  logHelper Code text ["ind=", show ind, "col=", show col]
+log debug text = logHelper debug text []
+
+log' :: Debug -> String -> LaTeX ()
+log' d = log d . T.pack
+
+output :: Text -> LaTeX ()
+output text = do
+  log Output text
+  tell text
+
 ------------------------------------------------------------------------
 -- * LaTeX and polytable strings.
 
@@ -202,17 +273,18 @@ infixr'     = T.pack "infixr"
 nonCode :: LaTeX ()
 nonCode = do
   tok <- nextToken
+  log NonCode tok
 
   if tok == beginCode
 
      then do
-       tell tok
+       output $ beginCode <+> nl
        resetColumn
        setInCode
        code
 
      else do
-       tell tok
+       output tok
        nonCode
 
 -- | Deals with code blocks. Every token, except spaces, is pretty
@@ -220,20 +292,26 @@ nonCode = do
 code :: LaTeX ()
 code = do
 
-  col <- gets column
-  when (col == 0) $ do
-    tell ptOpen
+  -- Get the column information before grabbing the token, since
+  -- grabbing (possibly) moves the column.
+  col  <- gets column
 
   tok' <- nextToken'
-  let tok = string tok'
+  let tok = text tok'
+  log Code tok
+
+  when (tok == T.empty) code
+
+  when (col == 0 && not (isActualSpaces tok)) $ do
+    output ptOpen
 
   when (tok == endCode) $ do
-    tell $ ptClose <+> tok
+    output $ ptClose <+> nl <+> endCode
     unsetInCode
     nonCode
 
   when (tok `elem` [ infixl', infix', infixr' ]) $ do
-    tell $ cmdPrefix <+> T.pack "Keyword" <+> cmdArg tok
+    output $ cmdPrefix <+> T.pack "Keyword" <+> cmdArg tok
     fixity
     code
 
@@ -242,17 +320,28 @@ code = do
     code
 
   case aspect (info tok') of
-    Nothing -> tell $ escape tok
-    Just a  -> tell $ cmdPrefix <+> T.pack (cmd a) <+> cmdArg (escape tok)
+    Nothing -> output $ escape tok
+--    Just a  -> output $ cmdPrefix <+> T.pack (cmd a) <+> cmdArg (escape tok)
+--  Andreas, 2014-02-17 preliminary fix for issue 1062
+    Just a  -> case cmd a of
+      "" -> output $ escape tok
+      s  -> output $ cmdPrefix <+> T.pack s <+> cmdArg (escape tok)
+
   code
 
   where
   cmd :: Aspect -> String
-  cmd (Name mKind _) = maybe __IMPOSSIBLE__ showKind mKind
+--  cmd (Name mKind _) = maybe __IMPOSSIBLE__ showKind mKind
+--  Andreas, 2014-02-17 preliminary fix for issue 1062
+  cmd (Name mKind _) = maybe "" showKind mKind
     where
     showKind :: NameKind -> String
     showKind (Constructor Inductive)   = "InductiveConstructor"
     showKind (Constructor CoInductive) = "CoinductiveConstructor"
+    -- Andreas, 2014-02-17
+    -- It might be boring boilerplate, but please spell out the
+    -- remaining cases instead of using the brittle @show@ function.
+    -- What if a constructor in @NameKind@ gets renamed?
     showKind k                         = show k
   cmd a              = show a
 
@@ -270,9 +359,12 @@ escape (T.uncons -> Just (c, s)) = T.pack (replace c) <+> escape s
     '$'  -> "\\$"
     '&'  -> "\\&"
     '%'  -> "\\%"
-    '~'  -> "\textasciitilde"
-    '^'  -> "\textasciicircum"
-    '\\' -> "\textbackslash"
+    '~'  -> "\\textasciitilde"
+    '^'  -> "\\textasciicircum"
+    '\\' -> "\\textbackslash"
+    -- Escaping newlines seems to fix the problem caused by pattern
+    -- synonyms.
+    '\n' -> "\\<\\\\\n\\>"
     _    -> [ c ]
 escape _                         = __IMPOSSIBLE__
 
@@ -297,12 +389,12 @@ fixity = do
 
     -- Fixity level.
     (num, nls) | nls == T.empty -> do
-        tell $ cmdPrefix <+> T.pack "Number" <+> cmdArg num
+        output $ cmdPrefix <+> T.pack "Number" <+> cmdArg num
         fixity
 
     -- Operations followed by newlines.
     (ops, nls) | otherwise      -> do
-        tell $ escape ops
+        output $ escape ops
         spaces (T.group nls)
 
 
@@ -315,8 +407,12 @@ spaces ((T.uncons -> Nothing)       : ss) = __IMPOSSIBLE__
 
 -- Single spaces are ignored.
 spaces ((T.uncons -> Just (' ', s)) : []) | T.null s = do
+  col <- gets column
+  when (col == 0) $ do
+    output ptOpen
+
   moveColumn 1
-  tell $ T.singleton ' '
+  output $ T.singleton ' '
 
 -- Multiple spaces.
 spaces (s@(T.uncons -> Just (' ', _)) : ss) = do
@@ -325,43 +421,49 @@ spaces (s@(T.uncons -> Just (' ', _)) : ss) = do
   col <- gets column
   moveColumn len
 
-  if col == 0
+  if col /= 0
+
      then do
-
-       -- FIX: What's going on here?
-       ind     <- gets indent
-       indPrev <- gets indentPrev
-
-       if ind == len
-          then do
-            tell $ ptOpen' indPrev
-          else do
-            if len < ind
-               then do
-                 resetIndent
-                 resetIndentPrev
-                 tell $ ptOpen' $ if indPrev == 0 ||
-                                         len == indPrev
-                                     then 0
-                                     else ind - indPrev - len
-               else do
-                 moveIndent $ len - ind
-                 setIndentPrev ind
-                 tell $ ptOpen' ind
-
-       tell $ cmdIndent len
-       tell $ ptClose' len <+> nl <+> ptOpen' len
-     else do
-       tell $ T.singleton ' '
+       log' Spaces "col /= 0"
+       output $ T.singleton ' '
        col <- gets column
-       tell $ ptClose' col <+> nl <+> ptOpen' col
+       output $ ptClose' col <+> nl <+> ptOpen' col
+
+     else do
+       log' Spaces "col == 0"
+       indent <- gets indent
+       indentPrev <- gets indentPrev
+
+       case compare len indent of
+
+         GT -> do
+           log' Spaces "GT"
+           setIndent len
+           setIndentPrev indent
+           output $ ptOpen' indent
+           output $ cmdIndent len
+           output $ ptClose' len <+> nl <+> ptOpen' len
+
+         EQ -> do
+           log' Spaces "EQ"
+           output $ ptOpen' indentPrev
+           output $ cmdIndent len
+           output $ ptClose' len <+> nl <+> ptOpen' len
+
+         LT -> do
+           log' Spaces "LT"
+           setIndent len
+           resetIndentPrev
+           output $ ptOpen' 0
+           output $ cmdIndent len
+           output $ ptClose' len <+> nl <+> ptOpen' len
 
   spaces ss
 
 -- Newlines.
 spaces (s@(T.uncons -> Just ('\n', _)) : ss) = do
   resetColumn
-  tell $ ptClose <+> T.replicate (T.length s) ptNL
+  output $ ptClose <+> T.replicate (T.length s) ptNL
   spaces ss
 
 -- Treat tabs as if they were spaces.
@@ -375,7 +477,7 @@ spaces (_                              : ss) = __IMPOSSIBLE__
 defaultStyFile = "agda.sty"
 
 -- | The only exported function. It's (only) called in @Main.hs@.
-generateLaTeX :: A.ModuleName -> HighlightingInfo -> TCM ()
+generateLaTeX :: TopLevelModuleName -> HighlightingInfo -> TCM ()
 generateLaTeX mod hi = do
 
   options <- TCM.commandLineOptions
@@ -389,35 +491,35 @@ generateLaTeX mod hi = do
     , "Checking if " ++ defaultStyFile ++ " is found by the LaTeX environment."
     ]
 
-  merrors <- callCompiler' "kpsewhich" [ defaultStyFile ]
+  merrors <- callCompiler' "kpsewhich" [ "--path=" ++ dir,  defaultStyFile ]
 
   when (isJust merrors) $ do
     TCM.reportSLn "latex" 1 $ unlines
       [ ""
       , defaultStyFile ++ " was not found. Copying a default version of " ++
           defaultStyFile
-      , "into the working directory."
+      , "into the directory " ++ dir ++ "."
       ]
 
     liftIO $ do
       styFile <- getDataFileName defaultStyFile
-      liftIO $ copyFile styFile defaultStyFile
+      liftIO $ copyFile styFile (dir </> defaultStyFile)
+
+  let outPath = modToFile mod
+  inAbsPath <- liftM filePath (Find.findFile mod)
 
   liftIO $ do
-    source <- UTF8.readTextFile (modToFile mod <.> "lagda")
-    let latex = E.encodeUtf8 $ toLaTeX source hi
-    BS.writeFile (dir </> modToFile mod <.> "tex") latex
+    source <- UTF8.readTextFile inAbsPath
+    latex <- E.encodeUtf8 `fmap` toLaTeX source hi
+    createDirectoryIfMissing True $ dir </> takeDirectory outPath
+    BS.writeFile (dir </> outPath) latex
 
   where
-  modToFile :: A.ModuleName -> FilePath
-  modToFile mod = go $ show mod
-    where
-    go []         = []
-    go ('.' : cs) = pathSeparator : go cs
-    go (c   : cs) = c             : go cs
+    modToFile :: TopLevelModuleName -> FilePath
+    modToFile m = List.intercalate [pathSeparator] (moduleNameParts m) <.> "tex"
 
 -- | Transforms the source code into LaTeX.
-toLaTeX :: String -> HighlightingInfo -> Text
+toLaTeX :: String -> HighlightingInfo -> IO Text
 toLaTeX source hi
 
   = processTokens
@@ -426,7 +528,7 @@ toLaTeX source hi
   -- collect the characters into a string.
   . map (\xs -> case xs of
                     (mi, (pos, _)) : _ ->
-                        Token { string   = T.pack $ map (\(_, (_, c)) -> c) xs
+                        Token { text     = T.pack $ map (\(_, (_, c)) -> c) xs
                               , info     = maybe mempty id mi
                               , position = pos
                               }
@@ -446,10 +548,9 @@ toLaTeX source hi
   where
   infoMap = toMap (decompress hi)
 
-processTokens :: Tokens -> Text
-processTokens ts =
+processTokens :: Tokens -> IO Text
+processTokens ts = do
+  (x, _, s) <- runLaTeX nonCode () (emptyState { tokens = ts })
   case x of
-    Left "Done" -> s
+    Left "Done" -> return s
     _           -> __IMPOSSIBLE__
-  where
-  (x, _, s) = runLaTeX nonCode () (emptyState { tokens = ts })
