@@ -21,6 +21,7 @@ module Agda.Interaction.Highlighting.Precise
   , compressedFileInvariant
   , compress
   , decompress
+  , noHighlightingInRange
     -- ** Creation
   , singletonC
   , severalC
@@ -38,17 +39,14 @@ import Data.List
 import Data.Function
 import Data.Monoid
 import Control.Applicative ((<$>), (<*>))
-import Control.Arrow ((&&&))
 import Control.Monad
-import Agda.Utils.QuickCheck hiding (ranges)
+import Agda.Utils.QuickCheck
 import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Typeable (Typeable)
 
-import qualified Agda.Syntax.Abstract.Name as A
 import qualified Agda.Syntax.Common as C
 import qualified Agda.Syntax.Concrete as SC
-import qualified Agda.Syntax.Position as P
 
 import Agda.Interaction.Highlighting.Range
 
@@ -78,6 +76,7 @@ data NameKind
   | Postulate
   | Primitive               -- ^ Primitive.
   | Record                  -- ^ Record type.
+  | Argument                -- ^ Named argument, like x in {x = v}
     deriving (Eq, Show, Typeable)
 
 -- | Other aspects. (These can overlap with each other and with
@@ -245,6 +244,17 @@ prop_compress f =
   decompress c == f
   where c = compress f
 
+-- | Clear any highlighting info for the given ranges. Used to make sure
+--   unsolved meta highlighting overrides error highlighting.
+noHighlightingInRange :: Ranges -> CompressedFile -> CompressedFile
+noHighlightingInRange rs (CompressedFile hs) =
+    CompressedFile $ concatMap clear hs
+  where
+    clear (r, i) =
+      case minus (Ranges [r]) rs of
+        Ranges [] -> []
+        Ranges rs -> [ (r, i) | r <- rs ]
+
 ------------------------------------------------------------------------
 -- Operations that work directly with compressed files
 
@@ -385,6 +395,7 @@ instance CoArbitrary NameKind where
   coarbitrary Postulate         = variant 6
   coarbitrary Primitive         = variant 7
   coarbitrary Record            = variant 8
+  coarbitrary Argument          = variant 9
 
 instance Arbitrary OtherAspect where
   arbitrary = elements [minBound .. maxBound]
