@@ -1,19 +1,17 @@
-{-# LANGUAGE CPP #-}
+-- {-# LANGUAGE CPP #-}
 
 module Agda.Interaction.Highlighting.Vim where
 
 import Control.Monad.Trans
-import Data.Char
-import Data.Function ( on )
-import Data.Set ( Set )
-import Data.Map ( Map )
-import System.FilePath
 
+import Data.Function ( on )
 import qualified Data.List as List
 import qualified Data.Map as Map
-import qualified Data.Set as Set
+
+import System.FilePath
 
 import Agda.Syntax.Scope.Base
+import Agda.Syntax.Common
 import Agda.Syntax.Concrete.Name as CName
 
 import Agda.TypeChecking.Monad
@@ -21,12 +19,6 @@ import Agda.TypeChecking.Monad
 import qualified Agda.Utils.IO.UTF8 as UTF8
 import Agda.Utils.Tuple
 
-#include "../../undefined.h"
-import Agda.Utils.Impossible
-
--- defined in Data.Function
--- on f g x y = f (g x) (g y)
-
 vimFile :: FilePath -> FilePath
 vimFile file =
     case splitFileName file of
@@ -39,6 +31,9 @@ escape = concatMap esc
 	esc c	| c `elem` escchars = ['\\',c]
 		| otherwise	    = [c]
 
+wordBounded :: String -> String
+wordBounded s0 = concat ["\\<", s0, "\\>"]
+
 keyword :: String -> [String] -> String
 keyword _ [] = ""
 keyword cat ws	= "syn keyword " ++ unwords (cat : ws)
@@ -46,7 +41,7 @@ keyword cat ws	= "syn keyword " ++ unwords (cat : ws)
 match :: String -> [String] -> String
 match _ [] = ""
 match cat ws	= "syn match " ++ cat ++ " \"" ++
-		    concat (List.intersperse "\\|" $ map escape ws) ++ "\""
+		    concat (List.intersperse "\\|" $ map (wordBounded . escape) ws) ++ "\""
 
 matches :: [String] -> [String] -> [String] -> [String] -> [String] -> [String] -> [String]
 matches cons icons defs idefs flds iflds =
@@ -84,7 +79,7 @@ toVim ns = unlines $ matches mcons micons mdefs midefs mflds miflds
 
 	parts (NoName _ _) = []
 	parts (Name _ [_]) = []
-	parts (Name _ ps)  = [ x | Id x <- ps ]
+	parts (Name _ ps)  = [ rawNameToString x | Id x <- ps ]
 
 generateVimFile :: FilePath -> TCM ()
 generateVimFile file = do
