@@ -1,4 +1,6 @@
-{-# LANGUAGE StandaloneDeriving, DeriveFunctor #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE StandaloneDeriving #-}
 
 module Agda.Interaction.Options
     ( CommandLineOptions(..)
@@ -6,6 +8,7 @@ module Agda.Interaction.Options
     , OptionsPragma
     , Flag
     , Verbosity
+    , IncludeDirs
     , checkOpts
     , parseStandardOptions
     , parsePragmaOptions
@@ -13,6 +16,7 @@ module Agda.Interaction.Options
     , defaultOptions
     , defaultInteractionOptions
     , defaultVerbosity
+    , defaultCutOff
     , standardOptions_
     , unsafePragmaOptions
     , isLiterate
@@ -23,18 +27,21 @@ module Agda.Interaction.Options
 
 import Control.Monad            ( when )
 import Control.Monad.Error	( MonadError(..) )
-import Data.Maybe (isJust)
+import Data.Maybe               ( isJust )
 import Data.List		( isSuffixOf , intercalate )
-import System.Console.GetOpt	(getOpt, usageInfo, ArgOrder(ReturnInOrder)
+import System.Console.GetOpt	( getOpt, usageInfo, ArgOrder(ReturnInOrder)
 				, OptDescr(..), ArgDescr(..)
 				)
+
+import Agda.Termination.CutOff  ( CutOff(..) )
+
 import Agda.Utils.TestHelpers   ( runTests )
 import Agda.Utils.QuickCheck    ( quickCheck' )
 import Agda.Utils.FileName      ( AbsolutePath )
 import Agda.Utils.Monad		( readM )
-import Agda.Utils.List               ( wordsBy )
-import Agda.Utils.String             ( indent )
-import Agda.Utils.Trie               ( Trie )
+import Agda.Utils.List          ( wordsBy )
+import Agda.Utils.String        ( indent )
+import Agda.Utils.Trie          ( Trie )
 import qualified Agda.Utils.Trie as Trie
 
 -- | This should probably go somewhere else.
@@ -43,25 +50,31 @@ isLiterate file = ".lagda" `isSuffixOf` file
 
 -- OptDescr is a Functor --------------------------------------------------
 
+-- base-4.7 defines these
+#if !(MIN_VERSION_base(4,7,0))
 deriving instance Functor OptDescr
 deriving instance Functor ArgDescr
+#endif
 
 type Verbosity = Trie String Int
 
+type IncludeDirs = Either [FilePath] [AbsolutePath]
+     -- ^ 'Left' is used temporarily, before the paths have
+     -- been made absolute. An empty 'Left' list is
+     -- interpreted as @["."]@ (see
+     -- 'Agda.TypeChecking.Monad.Options.makeIncludeDirsAbsolute').
+
 data CommandLineOptions =
     Options { optProgramName          :: String
             , optInputFile            :: Maybe FilePath
-            , optIncludeDirs          :: Either [FilePath] [AbsolutePath]
-              -- ^ 'Left' is used temporarily, before the paths have
-              -- been made absolute. An empty 'Left' list is
-              -- interpreted as @["."]@ (see
-              -- 'Agda.TypeChecking.Monad.Options.makeIncludeDirsAbsolute').
+            , optIncludeDirs          :: IncludeDirs
             , optShowVersion          :: Bool
             , optShowHelp             :: Bool
             , optInteractive          :: Bool
             , optRunTests             :: Bool
             , optGHCiInteraction      :: Bool
             , optCompile              :: Bool
+            , optCompileNoMain        :: Bool
             , optEpicCompile          :: Bool
             , optJSCompile            :: Bool
             , optCompileDir           :: Maybe FilePath
@@ -92,7 +105,8 @@ data PragmaOptions = PragmaOptions
   , optAllowUnsolved             :: Bool
   , optDisablePositivity         :: Bool
   , optTerminationCheck          :: Bool
-  , optTerminationDepth          :: Int
+  , optTerminationDepth          :: CutOff
+    -- ^ Cut off structural order comparison at some depth in termination checker?
   , optCompletenessCheck         :: Bool
   , optUniverseCheck             :: Bool
   , optSizedTypes                :: Bool
@@ -102,7 +116,8 @@ data PragmaOptions = PragmaOptions
   , optIrrelevantProjections     :: Bool
   , optExperimentalIrrelevance   :: Bool  -- ^ irrelevant levels, irrelevant data matching
   , optWithoutK                  :: Bool
-  , optCopatterns                :: Bool  -- ^ definitions by copattern matching
+  , optCopatterns                :: Bool  -- ^ Allow definitions by copattern matching?
+  , optPatternMatching           :: Bool  -- ^ Is pattern matching allowed in the current file?
   }
   deriving Show
 
@@ -118,18 +133,11 @@ type OptionsPragma = [String]
 mapFlag :: (String -> String) -> OptDescr a -> OptDescr a
 mapFlag f (Option _ long arg descr) = Option [] (map f long) arg descr
 
--- | For batch usage.
 defaultVerbosity :: Verbosity
 defaultVerbosity = Trie.singleton [] 1
 
--- | For interactive usage, do not print any debug messages
---   by default.
-defaultInteractionVerbosity :: Verbosity
-defaultInteractionVerbosity = Trie.singleton [] 0
-
 defaultInteractionOptions :: PragmaOptions
 defaultInteractionOptions = defaultPragmaOptions
-  { optVerbose = defaultInteractionVerbosity }
 
 defaultOptions :: CommandLineOptions
 defaultOptions =
@@ -142,6 +150,7 @@ defaultOptions =
             , optRunTests             = False
             , optGHCiInteraction      = False
             , optCompile              = False
+            , optCompileNoMain        = False
             , optEpicCompile          = False
             , optJSCompile            = False
             , optCompileDir           = Nothing
@@ -171,17 +180,22 @@ defaultPragmaOptions = PragmaOptions
   , optAllowUnsolved             = False
   , optDisablePositivity         = False
   , optTerminationCheck          = True
-  , optTerminationDepth          = 0    -- this is the cutoff value
+  , optTerminationDepth          = defaultCutOff
   , optCompletenessCheck         = True
   , optUniverseCheck             = True
-  , optSizedTypes                = False
+  , optSizedTypes                = True
   , optInjectiveTypeConstructors = False
   , optGuardingTypeConstructors  = False
   , optUniversePolymorphism      = True
   , optWithoutK                  = False
   , optCopatterns                = False
+  , optPatternMatching           = True
   }
 
+-- | The default termination depth.
+
+defaultCutOff = CutOff 0 -- minimum value
+
 -- | The default output directory for LaTeX.
 
 defaultLaTeXDir = "latex"
@@ -204,11 +218,13 @@ type Flag opts = opts -> Either String opts
 
 checkOpts :: Flag CommandLineOptions
 checkOpts opts
-  | not (atMostOne [optAllowUnsolved . p, optCompile]) = Left
+  | not (atMostOne [optAllowUnsolved . p, \x -> optCompile x]) = Left
       "Unsolved meta variables are not allowed when compiling.\n"
+  | optCompileNoMain opts && not (optCompile opts) = Left
+      "--no-main only allowed in combination with --compile.\n"
   | not (atMostOne [optGHCiInteraction, isJust . optInputFile]) =
       Left "Choose at most one: input file or --interaction.\n"
-  | not (atMostOne $ interactive ++ [optCompile, optEpicCompile, optJSCompile]) =
+  | not (atMostOne $ interactive ++ [\x -> optCompile x, optEpicCompile, optJSCompile]) =
       Left "Choose at most one: compilers/--interactive/--interaction.\n"
   | not (atMostOne $ interactive ++ [optGenerateHTML]) =
       Left "Choose at most one: --html/--interactive/--interaction.\n"
@@ -240,7 +256,7 @@ unsafePragmaOptions opts =
   [ "--no-termination-check"                     | not (optTerminationCheck opts)    ] ++
   [ "--no-coverage-check"                        | not (optCompletenessCheck opts)   ] ++
   [ "--type-in-type"                             | not (optUniverseCheck opts)       ] ++
-  [ "--sized-types"                              | optSizedTypes opts                ] ++
+  -- [ "--sized-types"                              | optSizedTypes opts                ] ++
   [ "--injective-type-constructors"              | optInjectiveTypeConstructors opts ] ++
   [ "--guardedness-preserving-type-constructors" | optGuardingTypeConstructors opts  ] ++
   [ "--experimental-irrelevance"                 | optExperimentalIrrelevance opts   ] ++
@@ -283,6 +299,7 @@ dontCompletenessCheckFlag    o = return $ o { optCompletenessCheck         = Fal
 dontUniverseCheckFlag        o = return $ o { optUniverseCheck             = False
                                             , optUniversePolymorphism      = False }
 sizedTypes                   o = return $ o { optSizedTypes                = True  }
+noSizedTypes                 o = return $ o { optSizedTypes                = False  }
 injectiveTypeConstructorFlag o = return $ o { optInjectiveTypeConstructors = True  }
 guardingTypeConstructorFlag  o = return $ o { optGuardingTypeConstructors  = True  }
 universePolymorphismFlag     o = return $ o { optUniversePolymorphism      = True  }
@@ -290,17 +307,19 @@ noUniversePolymorphismFlag   o = return $ o { optUniversePolymorphism      = Fal
 noForcingFlag                o = return $ o { optForcing                   = False }
 withoutKFlag                 o = return $ o { optWithoutK                  = True  }
 copatternsFlag               o = return $ o { optCopatterns                = True  }
+noPatternMatchingFlag        o = return $ o { optPatternMatching           = False }
 
 interactiveFlag  o = return $ o { optInteractive    = True
                                 , optPragmaOptions  = (optPragmaOptions o)
                                                         { optAllowUnsolved = True }
                                 }
-compileFlag      o = return $ o { optCompile    = True }
-compileEpicFlag  o = return $ o { optEpicCompile = True}
-compileJSFlag    o = return $ o { optJSCompile = True}
-compileDirFlag f o = return $ o { optCompileDir = Just f }
-ghcFlag        f o = return $ o { optGhcFlags   = f : optGhcFlags o }
-epicFlagsFlag  s o = return $ o { optEpicFlags  = optEpicFlags o ++ [s]}
+compileFlag        o = return $ o { optCompile    = True }
+compileFlagNoMain  o = return $ o { optCompileNoMain = True }
+compileEpicFlag    o = return $ o { optEpicCompile = True}
+compileJSFlag      o = return $ o { optJSCompile = True}
+compileDirFlag f   o = return $ o { optCompileDir = Just f }
+ghcFlag        f   o = return $ o { optGhcFlags   = optGhcFlags o  ++ [f] }  -- NOTE: Quadratic in number of flags.
+epicFlagsFlag  s   o = return $ o { optEpicFlags  = optEpicFlags o ++ [s] }  -- NOTE: Quadratic in number of flags.
 
 htmlFlag      o = return $ o { optGenerateHTML = True }
 dependencyGraphFlag f o = return $ o { optDependencyGraph  = Just f }
@@ -324,7 +343,7 @@ verboseFlag s o =
 terminationDepthFlag s o =
     do k <- readM s `catchError` \_ -> usage
        when (k < 1) $ usage -- or: turn termination checking off for 0
-       return $ o { optTerminationDepth = k-1 }
+       return $ o { optTerminationDepth = CutOff $ k-1 }
     where usage = throwError "argument to termination-depth should be >= 1"
 
 integerArgument :: String -> String -> Either String Int
@@ -342,6 +361,8 @@ standardOptions =
                     "for use with the Emacs mode"
     , Option ['c']  ["compile"] (NoArg compileFlag)
                     "compile program using the MAlonzo backend (experimental)"
+    , Option []     ["no-main"] (NoArg compileFlagNoMain)
+                    "when compiling using the MAlonzo backend (experimental), do not treat the requested module as the main module of a program"
     , Option []     ["epic"] (NoArg compileEpicFlag) "compile program using the Epic backend"
     , Option []     ["js"] (NoArg compileJSFlag) "compile program using the JS backend"
     , Option []     ["compile-dir"] (ReqArg compileDirFlag "DIR")
@@ -406,7 +427,9 @@ pragmaOptions =
     , Option []	    ["type-in-type"] (NoArg dontUniverseCheckFlag)
 		    "ignore universe levels (this makes Agda inconsistent)"
     , Option []     ["sized-types"] (NoArg sizedTypes)
-                    "use sized types (inconsistent with coinduction)"
+                    "use sized types (default, inconsistent with `musical' coinduction)"
+    , Option []     ["no-sized-types"] (NoArg noSizedTypes)
+                    "disable sized types"
     , Option []     ["injective-type-constructors"] (NoArg injectiveTypeConstructorFlag)
                     "enable injective type constructors (makes Agda anti-classical and possibly inconsistent)"
     , Option []     ["guardedness-preserving-type-constructors"] (NoArg guardingTypeConstructorFlag)
@@ -420,9 +443,11 @@ pragmaOptions =
     , Option []     ["experimental-irrelevance"] (NoArg experimentalIrrelevanceFlag)
                     "enable potentially unsound irrelevance features (irrelevant levels, irrelevant data matching)"
     , Option []     ["without-K"] (NoArg withoutKFlag)
-                    "disable the K rule (maybe)"
+                    "disable the K rule in pattern matching"
     , Option []     ["copatterns"] (NoArg copatternsFlag)
                     "enable definitions by copattern matching"
+    , Option []     ["no-pattern-matching"] (NoArg noPatternMatchingFlag)
+                    "disable pattern matching completely"
     ]
 
 -- | Used for printing usage info.
