@@ -1,10 +1,14 @@
 ------------------------------------------------------------------------
 -- | Data type for all interactive responses
 ------------------------------------------------------------------------
-{-# LANGUAGE CPP, TypeSynonymInstances, FlexibleInstances #-}
+
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE TypeSynonymInstances #-}
 
 module Agda.Interaction.Response
   ( Response (..)
+  , MakeCaseVariant (..)
   , DisplayInfo (..)
   , Status (..)
   , GiveResult (..)
@@ -13,13 +17,14 @@ module Agda.Interaction.Response
   ) where
 
 import Agda.Interaction.Highlighting.Precise
-import Agda.Interaction.FindFile (ModuleToSource)
---import Agda.Interaction.MakeCase
 import Agda.TypeChecking.Monad.Base
+import Agda.Syntax.Common   (InteractionId(..))
 import Agda.Syntax.Concrete (Expr)
 import Agda.Utils.Pretty
 
+import Control.Monad.Trans
 import Data.Int
+import System.IO
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
@@ -35,14 +40,18 @@ data Response
     | Resp_JumpToError FilePath Int32
     | Resp_InteractionPoints [InteractionId]
     | Resp_GiveAction InteractionId GiveResult
-    | Resp_MakeCaseAction [String]
-    | Resp_MakeCase String [String]
+    | Resp_MakeCase MakeCaseVariant [String]
     | Resp_SolveAll [(InteractionId, Expr)]
     | Resp_DisplayInfo DisplayInfo
-    | Resp_RunningInfo String
+    | Resp_RunningInfo Int String
+      -- ^ The integer is the message's debug level.
     | Resp_ClearRunningInfo
     | Resp_ClearHighlighting
 
+-- | There are two kinds of \"make case\" commands.
+
+data MakeCaseVariant = Function | ExtendedLambda
+
 -- | Info to display at the end of an interactive command
 
 data DisplayInfo
@@ -59,11 +68,13 @@ data DisplayInfo
         -- ^ 'Info_Auto' denotes either an error or a success (when 'Resp_GiveAction' is present)
         --   TODO: split these into separate constructors
     | Info_ModuleContents Doc
+    | Info_WhyInScope Doc
     | Info_NormalForm Doc
     | Info_GoalType Doc
     | Info_CurrentGoal Doc
     | Info_InferredType Doc
     | Info_Context Doc
+    | Info_HelperFunction Doc
         deriving Show
 
 -- | Status information.
@@ -106,10 +117,21 @@ data GiveResult
 
 type InteractionOutputCallback = Response -> TCM ()
 
--- | The default 'InteractionOutputCallback' function
---   is set to @__@@IMPOSSIBLE__@ because in this way it is easier to
---   recognize that some response is lost due to an uninitialized
---   'InteractionOutputCallback' function.
+-- | The default 'InteractionOutputCallback' function prints certain
+-- things to stdout (other things generate internal errors).
 
 defaultInteractionOutputCallback :: InteractionOutputCallback
-defaultInteractionOutputCallback = __IMPOSSIBLE__
+defaultInteractionOutputCallback r = case r of
+  Resp_HighlightingInfo {}  -> __IMPOSSIBLE__
+  Resp_Status {}            -> __IMPOSSIBLE__
+  Resp_JumpToError {}       -> __IMPOSSIBLE__
+  Resp_InteractionPoints {} -> __IMPOSSIBLE__
+  Resp_GiveAction {}        -> __IMPOSSIBLE__
+  Resp_MakeCase {}          -> __IMPOSSIBLE__
+  Resp_SolveAll {}          -> __IMPOSSIBLE__
+  Resp_DisplayInfo {}       -> __IMPOSSIBLE__
+  Resp_RunningInfo _ s      -> liftIO $ do
+                                 putStr s
+                                 hFlush stdout
+  Resp_ClearRunningInfo {}  -> __IMPOSSIBLE__
+  Resp_ClearHighlighting {} -> __IMPOSSIBLE__
