@@ -9,59 +9,42 @@ import Control.Monad.State
 import Control.Monad.Error
 import Control.Applicative
 
-import Data.List
-import Data.Map (Map)
-import qualified Data.Map as Map
+import qualified Data.List as List
 import Data.Maybe
 
 import System.Environment
 import System.Exit
-import System.FilePath
 
-import Agda.Syntax.Position
-import Agda.Syntax.Parser
+import qualified Text.PrettyPrint.Boxes as Boxes
+
 import Agda.Syntax.Concrete.Pretty ()
-import qualified Agda.Syntax.Abstract as A
-import Agda.Syntax.Abstract.Pretty
-import Agda.Syntax.Translation.ConcreteToAbstract
-import Agda.Syntax.Translation.AbstractToConcrete
-import Agda.Syntax.Translation.InternalToAbstract
-import Agda.Syntax.Abstract.Name
-import Agda.Syntax.Scope.Base
+import Agda.Syntax.Abstract.Name (toTopLevelModuleName)
 
-import Agda.Interaction.Exceptions
 import Agda.Interaction.CommandLine.CommandLine
 import Agda.Interaction.Options
 import Agda.Interaction.Monad
-import Agda.Interaction.GhcTop (mimicGHCi)
+import Agda.Interaction.EmacsTop (mimicGHCi)
+import Agda.Interaction.Imports (MaybeWarnings(..))
 import qualified Agda.Interaction.Imports as Imp
 import qualified Agda.Interaction.Highlighting.Dot as Dot
 import qualified Agda.Interaction.Highlighting.LaTeX as LaTeX
 import Agda.Interaction.Highlighting.HTML
 
-import Agda.TypeChecker
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Reduce
+import Agda.TypeChecking.Monad.Benchmark
+import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 import Agda.TypeChecking.Errors
-import qualified Agda.TypeChecking.Serialise
-import Agda.TypeChecking.Serialise
-import Agda.TypeChecking.SizedTypes
 
 import Agda.Compiler.MAlonzo.Compiler as MAlonzo
 import Agda.Compiler.Epic.Compiler as Epic
 import Agda.Compiler.JS.Compiler as JS
 
-import Agda.Termination.TermCheck
-
 import Agda.Utils.Monad
-import Agda.Utils.FileName
-import Agda.Utils.Pretty
+import qualified Agda.Utils.Trie as Trie
 
 import Agda.Tests
 import Agda.Version
 
-import qualified System.IO as IO
-
 #include "undefined.h"
 import Agda.Utils.Impossible
 
@@ -69,7 +52,7 @@ import Agda.Utils.Impossible
 runAgda :: TCM ()
 runAgda = do
   progName <- liftIO getProgName
-  argv   <- liftIO getArgs
+  argv     <- liftIO getArgs
   let opts = parseStandardOptions argv
   case opts of
     Left err -> liftIO $ optionError err
@@ -85,15 +68,36 @@ runAgda = do
                             -> liftIO printUsage
       | otherwise           -> do
           setCommandLineOptions opts
-          checkFile
+          -- Main function.
+          -- Bill everything to root of Benchmark trie.
+          billTo [] $ checkFile
+
+          -- Print benchmarks.
+          whenM benchmarking $ do
+            (accounts, times) <- List.unzip . Trie.toList <$> getBenchmark
+            -- Generate a table.
+            let showAccount [] = "Total time"
+                showAccount ks = List.concat . List.intersperse "." . map show $ ks
+                -- First column is accounts.
+                col1 = Boxes.vcat Boxes.left $
+                       map (Boxes.text . showAccount) $
+                       accounts
+                -- Second column is times.
+                -- CPU times are in pico seconds, convert to milliseconds.
+                col2 = Boxes.vcat Boxes.right $
+                       map (Boxes.text . (++ " ms") . show . (`div` 1000000000)) $
+                       times
+                table = Boxes.hsep 1 Boxes.left [col1, col2]
+            reportBenchmarkingLn $ Boxes.render table
   where
     checkFile :: TCM ()
     checkFile = do
-      i       <- optInteractive     <$> liftTCM commandLineOptions
-      ghci    <- optGHCiInteraction <$> liftTCM commandLineOptions
-      compile <- optCompile         <$> liftTCM commandLineOptions
-      epic    <- optEpicCompile     <$> liftTCM commandLineOptions
-      js      <- optJSCompile       <$> liftTCM commandLineOptions
+      i             <- optInteractive     <$> liftTCM commandLineOptions
+      ghci          <- optGHCiInteraction <$> liftTCM commandLineOptions
+      compile       <- optCompile         <$> liftTCM commandLineOptions
+      compileNoMain <- optCompileNoMain   <$> liftTCM commandLineOptions
+      epic          <- optEpicCompile     <$> liftTCM commandLineOptions
+      js            <- optJSCompile       <$> liftTCM commandLineOptions
       when i $ liftIO $ putStr splashScreen
       let failIfNoInt (Just i) = return i
           -- The allowed combinations of command-line
@@ -104,40 +108,48 @@ runAgda = do
           failIfInt _ (Just _) = __IMPOSSIBLE__
 
           interaction :: TCM (Maybe Interface) -> TCM ()
-          interaction | i         = runIM . interactionLoop
-                      | ghci      = (failIfInt mimicGHCi =<<)
-                      | compile   = (MAlonzo.compilerMain =<<) . (failIfNoInt =<<)
-                      | epic      = (Epic.compilerMain    =<<) . (failIfNoInt =<<)
-                      | js        = (JS.compilerMain      =<<) . (failIfNoInt =<<)
-                      | otherwise = (() <$)
+          interaction | i             = runIM . interactionLoop
+                      | ghci          = (failIfInt mimicGHCi =<<)
+                      | compile && compileNoMain
+                                      = (MAlonzo.compilerMain False =<<) . (failIfNoInt =<<)
+                      | compile       = (MAlonzo.compilerMain True =<<) . (failIfNoInt =<<)
+                      | epic          = (Epic.compilerMain    =<<) . (failIfNoInt =<<)
+                      | js            = (JS.compilerMain      =<<) . (failIfNoInt =<<)
+                      | otherwise     = (() <$)
       interaction $ do
         hasFile <- hasInputFile
-        resetState
+        -- Andreas, 2013-10-30 The following 'resetState' kills the
+        -- verbosity options.  That does not make sense (see fail/Issue641).
+        -- 'resetState' here does not seem to serve any purpose,
+        -- thus, I am removing it.
+        -- resetState
         if not hasFile then return Nothing else do
           file    <- getInputFile
-          (i, mw) <- Imp.typeCheck file
+          (i, mw) <- Imp.typeCheckMain file
 
           unsolvedOK <- optAllowUnsolved <$> pragmaOptions
 
           result <- case mw of
-            Just (Warnings [] [] []) -> __IMPOSSIBLE__
-            Just (Warnings _ unsolved@(_:_) _)
+                          -- we get here if there are unfilled interaction
+                          -- points that have been solved by unification
+            SomeWarnings (Warnings [] [] []) -> return Nothing
+            SomeWarnings (Warnings _ unsolved@(_:_) _)
               | not unsolvedOK -> typeError $ UnsolvedMetas unsolved
-            Just (Warnings _ _ unsolved@(_:_))
+            SomeWarnings (Warnings _ _ unsolved@(_:_))
               | not unsolvedOK -> typeError $ UnsolvedConstraints unsolved
-            Just (Warnings termErrs@(_:_) _ _) ->
+            SomeWarnings (Warnings termErrs@(_:_) _ _) ->
               typeError $ TerminationCheckFailed termErrs
-            Just _  -> return Nothing
-            Nothing -> return $ Just i
+            SomeWarnings _  -> return Nothing
+            NoWarnings -> return $ Just i
 
           whenM (optGenerateHTML <$> commandLineOptions) $
-            generateHTML $ iModuleName i
+            generateHTML
 
           whenM (isJust . optDependencyGraph <$> commandLineOptions) $
             Dot.generateDot $ i
 
           whenM (optGenerateLaTeX <$> commandLineOptions) $
-            LaTeX.generateLaTeX (iModuleName i) (iHighlighting i)
+            LaTeX.generateLaTeX (toTopLevelModuleName $ iModuleName i) (iHighlighting i)
 
           return result
 
@@ -162,7 +174,7 @@ optionError err = do
 -- | Main
 main :: IO ()
 main = do
-    r <- runTCM $ runAgda `catchError` \err -> do
+    r <- runTCMTop $ runAgda `catchError` \err -> do
       s <- prettyError err
       liftIO $ putStrLn s
       throwError err
