@@ -1,4 +1,13 @@
-{-# LANGUAGE CPP, TupleSections, PatternGuards, DeriveFunctor, StandaloneDeriving, ScopedTypeVariables, TypeSynonymInstances, FlexibleContexts,  FlexibleInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE StandaloneDeriving #-}
+{-# LANGUAGE TupleSections #-}
+{-# LANGUAGE TypeSynonymInstances #-}
+
 module Agda.Syntax.Abstract.Copatterns (translateCopatternClauses) where
 
 import Prelude hiding (mapM)
@@ -14,16 +23,14 @@ import Data.Traversable as T
 
 import Agda.Syntax.Abstract
 import Agda.Syntax.Abstract.Name as A
-import Agda.Syntax.Common
+import Agda.Syntax.Common hiding (Arg, Dom, NamedArg)
+import qualified Agda.Syntax.Common as Common
 import qualified Agda.Syntax.Concrete.Name as C
 import Agda.Syntax.Info
 import Agda.Syntax.Position
-import Agda.Syntax.Scope.Base
 import Agda.Syntax.Scope.Monad
 
-import Agda.TypeChecking.Monad.Base (TypeError(..), Call(..), typeError,
-                                     TCErr(..))
-import Agda.Utils.List
+import Agda.TypeChecking.Monad.Base (TypeError(..), typeError)
 import Agda.Utils.Tuple
 
 #include "../../undefined.h"
@@ -218,9 +225,9 @@ pathToRecord pps =
             foldr abstract (return e) xs
 
           abstract :: NamedArg Name -> ScopeM Expr -> ScopeM Expr
-          abstract (Arg h r (Named Nothing x)) me =
-            Lam (ExprRange noRange) (DomainFree h r x) <$> me
-          abstract (Arg _ _ (Named Just{} _)) me = typeError $ NotImplemented $
+          abstract (Common.Arg info (Named Nothing x)) me =
+            Lam (ExprRange noRange) (DomainFree info x) <$> me
+          abstract (Common.Arg _ (Named Just{} _)) me = typeError $ NotImplemented $
             "named arguments in projection patterns"
 
 -- | Similar to 'groupClauses'.
@@ -263,7 +270,7 @@ instance Rename Expr where
       Def f                 -> e
       Con c                 -> e
       Lit l                 -> e
-      QuestionMark i        -> e
+      QuestionMark{}        -> e
       Underscore i          -> e
       App i e es            -> App i (rename rho e) (rename rho es)
       WithApp i e es        -> WithApp i (rename rho e) (rename rho es)
@@ -280,6 +287,7 @@ instance Rename Expr where
       RecUpdate i e fes     -> RecUpdate i (rename rho e) $ map (id -*- rename rho) fes
       ScopedExpr i e        -> ScopedExpr i (rename rho e)
       QuoteGoal i n e       -> QuoteGoal i n (rename rho e)
+      QuoteContext i n e    -> QuoteContext i n (rename rho e)
       Quote i               -> e
       QuoteTerm i           -> e
       Unquote i             -> e
@@ -305,7 +313,7 @@ instance Rename TypedBindings where
 
 instance Rename TypedBinding where
   rename rho (TBind r ns e) = TBind r ns (rename rho e)
-  rename rho (TNoBind    e) = TNoBind (rename rho e)
+  rename rho (TLet r lbs)   = TLet r (rename rho lbs)
 
 instance Rename Clause where
   rename rho (Clause lhs rhs wheredecls) =
