@@ -1,11 +1,18 @@
-{-# LANGUAGE CPP,
-  DeriveDataTypeable, GeneralizedNewtypeDeriving,
-  DeriveFunctor, DeriveFoldable, DeriveTraversable,
-  FlexibleContexts #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveTraversable #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
 
-{-| Abstract names should carry unique identifiers and stuff. Not right now though.
+{-| Abstract names carry unique identifiers and stuff.
 -}
-module Agda.Syntax.Abstract.Name where
+module Agda.Syntax.Abstract.Name
+  ( module Agda.Syntax.Abstract.Name
+  , IsNoName(..)
+  ) where
 
 import Control.Monad.State
 
@@ -19,14 +26,13 @@ import Data.Hashable
 import Agda.Syntax.Position
 import Agda.Syntax.Common
 import Agda.Syntax.Fixity
+import Agda.Syntax.Concrete.Name (IsNoName(..))
 import qualified Agda.Syntax.Concrete.Name as C
 
 import Agda.Utils.Fresh
 import Agda.Utils.Size
 import Agda.Utils.Suffix
 
-import Agda.Syntax.Notation
-
 #include "../../undefined.h"
 import Agda.Utils.Impossible
 
@@ -72,14 +78,10 @@ newtype ModuleName = MName { mnameToList :: [Name] }
 newtype AmbiguousQName = AmbQ { unAmbQ :: [QName] }
   deriving (Typeable, HasRange, Show)
 
-instance HasRange ModuleName where
-  getRange (MName []) = noRange
-  getRange (MName xs) = getRange xs
-
 -- | A module is anonymous if the qualification path ends in an underscore.
 isAnonymousModuleName :: ModuleName -> Bool
 isAnonymousModuleName (MName []) = False
-isAnonymousModuleName (MName ms) = C.isNoName $ nameConcrete $ last ms
+isAnonymousModuleName (MName ms) = isNoName $ last ms
 
 -- | Sets the ranges of the individual names in the module name to
 -- match those of the corresponding concrete names. If the concrete
@@ -118,19 +120,17 @@ mnameFromList = MName
 noModuleName :: ModuleName
 noModuleName = mnameFromList []
 
--- | The 'Range' sets the /definition site/ of the name, not the use
--- site.
+-- | Make a 'Name' from some kind of string.
+class MkName a where
+  -- | The 'Range' sets the /definition site/ of the name, not the use site.
+  mkName :: Range -> NameId -> a -> Name
 
-mkName :: Range -> NameId -> String -> Name
-mkName r i s = Name i (C.Name noRange (parseName s)) r defaultFixity'
-  where
-    parseName ""      = []
-    parseName ('_':s) = C.Hole : parseName s
-    parseName s = case break (== '_') s of
-      (s0, s1)	-> C.Id s0 : parseName s1
+  mkName_ :: NameId -> a -> Name
+  mkName_ = mkName noRange
+
+instance MkName String where
+  mkName r i s = Name i (C.Name noRange (C.stringNameParts s)) r defaultFixity'
 
-mkName_ :: NameId -> String -> Name
-mkName_ = mkName noRange
 
 qnameToList :: QName -> [Name]
 qnameToList (QName m x) = mnameToList m ++ [x]
@@ -201,9 +201,6 @@ freshName r s = do
   i <- fresh
   return $ mkName r i s
 
-freshName_ :: (MonadState s m, HasFresh NameId s) => String -> m Name
-freshName_ = freshName noRange
-
 freshNoName :: (MonadState s m, HasFresh NameId s) => Range -> m Name
 freshNoName r =
     do	i <- fresh
@@ -212,6 +209,22 @@ freshNoName r =
 freshNoName_ :: (MonadState s m, HasFresh NameId s) => m Name
 freshNoName_ = freshNoName noRange
 
+-- | Create a fresh name from @a@.
+class FreshName a where
+  freshName_ :: (MonadState s m, HasFresh NameId s) => a -> m Name
+
+instance FreshName (Range, String) where
+  freshName_ = uncurry freshName
+
+instance FreshName String where
+  freshName_ = freshName noRange
+
+instance FreshName Range where
+  freshName_ = freshNoName
+
+instance FreshName () where
+  freshName_ () = freshNoName_
+
 -- | Get the next version of the concrete name. For instance, @nextName "x" = "x₁"@.
 --   The name must not be a 'NoName'.
 nextName :: Name -> Name
@@ -226,27 +239,21 @@ nextName x = x { nameConcrete = C.Name noRange $ nextSuf ps }
 	nextStr s = case suffixView s of
 	    (s0, suf) -> addSuffix s0 (nextSuffix suf)
 
-instance Show NameId where
-  show (NameId x i) = show x ++ "@" ++ show i
+------------------------------------------------------------------------
+-- * Important instances: Eq, Ord, Hashable
+--
+--   For the identity and comparing of names, only the 'NameId' matters!
+------------------------------------------------------------------------
 
 instance Eq Name where
-  x == y  = nameId x == nameId y
+  (==) = (==) `on` nameId
 
 instance Ord Name where
-  compare x y = compare (nameId x) (nameId y)
-
-instance Show Name where
-  show x = show (nameConcrete x) -- ++ "|" ++ show (nameId x)
+  compare = compare `on` nameId
 
 instance Hashable Name where
   {-# INLINE hashWithSalt #-}
-  hashWithSalt salt = (hashWithSalt salt) . nameId
-
-instance Show QName where
-  show q = concat $ intersperse "." $ map show $ qnameToList q
-
-instance Show ModuleName where
-  show m = concat $ intersperse "." $ map show $ mnameToList m
+  hashWithSalt salt = hashWithSalt salt . nameId
 
 instance Eq QName where
   (==) = (==) `on` qnameName
@@ -256,14 +263,47 @@ instance Ord QName where
 
 instance Hashable QName where
   {-# INLINE hashWithSalt #-}
-  hashWithSalt salt = (hashWithSalt salt) . qnameName
+  hashWithSalt salt = hashWithSalt salt . qnameName
+
+------------------------------------------------------------------------
+-- * IsNoName instances (checking for "_")
+------------------------------------------------------------------------
+
+-- | An abstract name is empty if its concrete name is empty.
+instance IsNoName Name where
+  isNoName = isNoName . nameConcrete
+
+------------------------------------------------------------------------
+-- * Show instances
+------------------------------------------------------------------------
+
+instance Show Name where
+  show x = show (nameConcrete x) -- ++ "|" ++ show (nameId x)
+
+instance Show ModuleName where
+  show m = concat $ intersperse "." $ map show $ mnameToList m
+
+instance Show QName where
+  show q = concat $ intersperse "." $ map show $ qnameToList q
+
+------------------------------------------------------------------------
+-- * Range instances
+------------------------------------------------------------------------
+
+-- ** HasRange
 
 instance HasRange Name where
   getRange = getRange . nameConcrete
 
+instance HasRange ModuleName where
+  getRange (MName []) = noRange
+  getRange (MName xs) = getRange xs
+
 instance HasRange QName where
   getRange q = getRange (qnameModule q, qnameName q)
 
+-- ** SetRange
+
 instance SetRange Name where
   setRange r x = x { nameConcrete = setRange r $ nameConcrete x }
 
@@ -275,20 +315,32 @@ instance SetRange QName where
 instance SetRange ModuleName where
   setRange r (MName ns) = MName (map (setRange r) ns)
 
+-- ** KillRange
+
+instance KillRange Name where
+  killRange x = x { nameConcrete = killRange $ nameConcrete x
+                  -- Andreas, 2014-03-30
+                  -- An experiment: what happens if we preserve
+                  -- the range of the binding site, but kill all
+                  -- other ranges before serialization?
+                  -- , nameBindingSite = noRange
+                  }
+
+instance KillRange ModuleName where
+  killRange (MName xs) = MName $ killRange xs
+
 instance KillRange QName where
   killRange q = q { qnameModule = killRange $ qnameModule q
                   , qnameName   = killRange $ qnameName   q
                   }
 
-instance KillRange Name where
-  killRange x = x { nameConcrete = killRange $ nameConcrete x }
-
-instance KillRange ModuleName where
-  killRange (MName xs) = MName $ killRange xs
-
 instance KillRange AmbiguousQName where
   killRange (AmbQ xs) = AmbQ $ killRange xs
 
+------------------------------------------------------------------------
+-- * Sized instances
+------------------------------------------------------------------------
+
 instance Sized QName where
   size = size . qnameToList
 
