@@ -1,9 +1,8 @@
-
+{-# LANGUAGE FlexibleContexts #-}
 module Agda.Syntax.Abstract.Pretty where
 
 import Control.Applicative
 
-import Agda.Syntax.Abstract
 import Agda.Syntax.Concrete.Pretty ()
 import Agda.Syntax.Fixity
 import Agda.Syntax.Translation.AbstractToConcrete
@@ -16,6 +15,9 @@ showA x = show <$> abstractToConcrete_ x
 prettyA :: (Pretty c, ToConcrete a c) => a -> TCM Doc
 prettyA x = pretty <$> abstractToConcrete_ x
 
+prettyAs :: (Pretty c, ToConcrete a [c]) => a -> TCM Doc
+prettyAs x = fsep . map pretty <$> abstractToConcrete_ x
+
 -- | Variant of 'showA' which does not insert outermost parentheses.
 
 showATop :: (Show c, ToConcrete a c) => a -> TCM String
