@@ -1,4 +1,10 @@
-{-# LANGUAGE CPP, DeriveDataTypeable, DeriveFunctor, DeriveTraversable, DeriveFoldable #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveTraversable #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE TypeSynonymInstances #-}
 
 {-| The concrete syntax is a raw representation of the program text
     without any desugaring at all.  This is what the parser produces.
@@ -12,10 +18,14 @@ module Agda.Syntax.Concrete
     , module Agda.Syntax.Concrete.Name
     , appView, AppView(..)
       -- * Bindings
-    , LamBinding(..)
-    , TypedBindings(..)
-    , TypedBinding(..)
-    , BoundName(..), mkBoundName_
+    , LamBinding
+    , LamBinding'(..)
+    , TypedBindings
+    , TypedBindings'(..)
+    , TypedBinding
+    , TypedBinding'(..)
+    , ColoredTypedBinding(..)
+    , BoundName(..), mkBoundName_, mkBoundName
     , Telescope -- (..)
       -- * Declarations
     , Declaration(..)
@@ -27,34 +37,48 @@ module Agda.Syntax.Concrete
     , defaultImportDir
     , OpenShortHand(..), RewriteEqn, WithExpr
     , LHS(..), Pattern(..), LHSCore(..)
-    , RHS(..), WhereClause(..)
+    , RHS, RHS'(..), WhereClause, WhereClause'(..)
     , Pragma(..)
     , Module
     , ThingWithFixity(..)
     , topLevelModuleName
     -- * Pattern tools
     , patternHead, patternNames
+    -- * Concrete instances
+    , Color
+    , Arg
+--    , Dom
+    , NamedArg
+    , ArgInfo
     )
     where
 
+import Control.DeepSeq
 import Data.Typeable (Typeable)
-import Data.Foldable hiding (concatMap)
-import Data.Traversable
+import Data.Foldable (Foldable)
+import Data.Traversable (Traversable)
 import Agda.Syntax.Position
-import Agda.Syntax.Common
+import Agda.Syntax.Common hiding (Arg, Dom, NamedArg, ArgInfo)
+import qualified Agda.Syntax.Common as Common
 import Agda.Syntax.Fixity
 import Agda.Syntax.Notation
 import Agda.Syntax.Literal
 
 import Agda.Syntax.Concrete.Name
 
-import Agda.Utils.Impossible
 #include "../undefined.h"
+import Agda.Utils.Impossible
+
+type Color      = Expr
+type Arg a      = Common.Arg Color a
+-- type Dom a      = Common.Dom Color a
+type NamedArg a = Common.NamedArg Color a
+type ArgInfo    = Common.ArgInfo Color
 
 data OpApp e
         = SyntaxBindingLambda !Range [LamBinding] e -- ^ an abstraction inside a special syntax declaration (see Issue 358 why we introduce this).
         | Ordinary e
-    deriving (Typeable, Functor)
+    deriving (Typeable, Functor, Foldable, Traversable)
 
 fromOrdinary :: e -> OpApp e -> e
 fromOrdinary d (Ordinary e) = e
@@ -68,10 +92,10 @@ data Expr
 	| Underscore !Range (Maybe String)     -- ^ ex: @_@ or @_A_5@
 	| RawApp !Range [Expr]		       -- ^ before parsing operators
 	| App !Range Expr (NamedArg Expr)      -- ^ ex: @e e@, @e {e}@, or @e {x = e}@
-	| OpApp !Range QName [OpApp Expr]      -- ^ ex: @e + e@
+	| OpApp !Range QName [NamedArg (OpApp Expr)] -- ^ ex: @e + e@
         | WithApp !Range Expr [Expr]           -- ^ ex: @e | e1 | .. | en@
-	| HiddenArg !Range (Named String Expr) -- ^ ex: @{e}@ or @{x=e}@
-	| InstanceArg !Range (Named String Expr) -- ^ ex: @{{e}}@ or @{{x=e}}@
+	| HiddenArg !Range (Named_ Expr)       -- ^ ex: @{e}@ or @{x=e}@
+	| InstanceArg !Range (Named_ Expr)     -- ^ ex: @{{e}}@ or @{{x=e}}@
 	| Lam !Range [LamBinding] Expr	       -- ^ ex: @\\x {y} -> e@ or @\\(x:A){y:B} -> e@
         | AbsurdLam !Range Hiding              -- ^ ex: @\\ ()@
         | ExtendedLam !Range [(LHS,RHS,WhereClause)]       -- ^ ex: @\\ { p11 .. p1a -> e1 ; .. ; pn1 .. pnz -> en }@
@@ -89,21 +113,24 @@ data Expr
 	| Dot !Range Expr		       -- ^ ex: @.p@, only in patterns
         | ETel Telescope                       -- ^ only used for printing telescopes
         | QuoteGoal !Range Name Expr           -- ^ ex: @quoteGoal x in e@
+        | QuoteContext !Range Name Expr        -- ^ ex: @quoteContext ctx in e@
         | Quote !Range                         -- ^ ex: @quote@, should be applied to a name
         | QuoteTerm !Range                     -- ^ ex: @quoteTerm@, should be applied to a term
         | Unquote !Range                       -- ^ ex: @unquote@, should be applied to a term of type @Term@
         | DontCare Expr                        -- ^ to print irrelevant things
+        | Equal !Range Expr Expr               -- ^ ex: @a = b@, used internally in the parser
     deriving (Typeable)
 
+instance NFData Expr
 
 -- | Concrete patterns. No literals in patterns at the moment.
 data Pattern
 	= IdentP QName                            -- ^ @c@ or @x@
 	| AppP Pattern (NamedArg Pattern)         -- ^ @p p'@ or @p {x = p'}@
 	| RawAppP !Range [Pattern]                -- ^ @p1..pn@ before parsing operators
-	| OpAppP !Range QName [Pattern]           -- ^ eg: @p => p'@ for operator @_=>_@
-	| HiddenP !Range (Named String Pattern)   -- ^ @{p}@ or @{x = p}@
-	| InstanceP !Range (Named String Pattern) -- ^ @{{p}}@ or @{{x = p}}@
+	| OpAppP !Range QName [NamedArg Pattern]  -- ^ eg: @p => p'@ for operator @_=>_@
+	| HiddenP !Range (Named_ Pattern)         -- ^ @{p}@ or @{x = p}@
+	| InstanceP !Range (Named_ Pattern)       -- ^ @{{p}}@ or @{{x = p}}@
 	| ParenP !Range Pattern                   -- ^ @(p)@
 	| WildP !Range                            -- ^ @_@
 	| AbsurdP !Range                          -- ^ @()@
@@ -112,35 +139,45 @@ data Pattern
 	| LitP Literal                            -- ^ @0@, @1@, etc.
     deriving (Typeable)
 
+instance NFData Pattern
 
 -- | A lambda binding is either domain free or typed.
-data LamBinding
-	= DomainFree Hiding Relevance BoundName -- ^ . @x@ or @{x}@ or @.x@ or @.{x}@ or @{.x}@
-	| DomainFull TypedBindings              -- ^ . @(xs : e)@ or @{xs : e}@
-    deriving (Typeable)
+type LamBinding = LamBinding' TypedBindings
+data LamBinding' a
+	= DomainFree ArgInfo BoundName -- ^ . @x@ or @{x}@ or @.x@ or @.{x}@ or @{.x}@
+	| DomainFull a                 -- ^ . @(xs : e)@ or @{xs : e}@
+    deriving (Typeable, Functor, Foldable, Traversable)
 
 
 -- | A sequence of typed bindings with hiding information. Appears in dependent
 --   function spaces, typed lambdas, and telescopes.
-data TypedBindings = TypedBindings !Range (Arg TypedBinding)
+type TypedBindings = TypedBindings' TypedBinding
+data TypedBindings' a = TypedBindings !Range (Arg a)
 	-- ^ . @(xs : e)@ or @{xs : e}@
-    deriving (Typeable)
+    deriving (Typeable, Functor, Foldable, Traversable)
 
 
 data BoundName = BName { boundName   :: Name
+                       , boundLabel  :: Name    -- ^ for implicit function types the label matters and can't be alpha-renamed
                        , bnameFixity :: Fixity'
                        }
     deriving (Typeable)
 
 mkBoundName_ :: Name -> BoundName
-mkBoundName_ x = BName x defaultFixity'
+mkBoundName_ x = mkBoundName x defaultFixity'
+
+mkBoundName :: Name -> Fixity' -> BoundName
+mkBoundName x f = BName x x f
 
 -- | A typed binding.
-data TypedBinding
-	= TBind !Range [BoundName] Expr   -- Binding @x1,..,xn:A@
-	| TNoBind Expr		    -- No binding @A@, equivalent to @_ : A@.
-    deriving (Typeable)
+type TypedBinding = TypedBinding' Expr
+data TypedBinding' e
+    = TBind !Range [BoundName] e -- ^ Binding @(x1 ... xn : A)@.
+    | TLet  !Range [Declaration] -- ^ Let binding @(let Ds)@ or @(open M args)@.
+  deriving (Typeable, Functor, Foldable, Traversable)
 
+-- | Color a TypeBinding. Used by Pretty.
+data ColoredTypedBinding = WithColors [Color] TypedBinding
 
 -- | A telescope is a sequence of typed bindings. Bound variables are in scope
 --   in later types.
@@ -179,6 +216,8 @@ data LHSCore
              }
   deriving (Typeable)
 
+instance NFData LHSCore
+
 {- TRASH
 lhsCoreToPattern :: LHSCore -> Pattern
 lhsCoreToPattern (LHSHead f args) = OpAppP (fuseRange f args) (unqualify f) args
@@ -187,12 +226,19 @@ lhsCoreToPattern (LHSProj d ps1 lhscore ps2) = OpAppP (fuseRange d ps) (unqualif
         ps = ps1 ++ p : ps2
 -}
 
-data RHS = AbsurdRHS
-	 | RHS Expr
-    deriving (Typeable)
+type RHS = RHS' Expr
+data RHS' e
+    = AbsurdRHS -- ^ No right hand side because of absurd match.
+    | RHS e
+  deriving (Typeable, Functor, Foldable, Traversable)
 
-data WhereClause = NoWhere | AnyWhere [Declaration] | SomeWhere Name [Declaration]
-  deriving (Typeable)
+
+type WhereClause = WhereClause' [Declaration]
+data WhereClause' decls
+    = NoWhere               -- ^ No @where@ clauses.
+    | AnyWhere decls        -- ^ Ordinary @where@.
+    | SomeWhere Name decls  -- ^ Named where: @module M where@.
+  deriving (Typeable, Functor, Foldable, Traversable)
 
 
 -- | The things you are allowed to say when you shuffle names between name
@@ -257,17 +303,18 @@ type Constructor = TypeSignature
 -}
 
 data Declaration
-	= TypeSig Relevance Name Expr -- ^ Axioms and functions can be irrelevant.
+	= TypeSig ArgInfo Name Expr
+        -- ^ Axioms and functions can be irrelevant. (Hiding should be NotHidden)
         | Field Name (Arg Expr) -- ^ Record field, can be hidden and/or irrelevant.
 	| FunClause LHS RHS WhereClause
 	| DataSig     !Range Induction Name [LamBinding] Expr -- ^ lone data signature in mutual block
 	| Data        !Range Induction Name [LamBinding] (Maybe Expr) [Constructor]
 	| RecordSig   !Range Name [LamBinding] Expr -- ^ lone record signature in mutual block
-	| Record      !Range Name (Maybe Induction) (Maybe Name) [LamBinding] (Maybe Expr) [Declaration]
+	| Record      !Range Name (Maybe (Ranged Induction)) (Maybe Name) [LamBinding] (Maybe Expr) [Declaration]
           -- ^ The optional name is a name for the record constructor.
 	| Infix Fixity [Name]
         | Syntax      Name Notation -- ^ notation declaration for a name
-        | PatternSyn  !Range Name [Name] Pattern
+        | PatternSyn  !Range Name [Arg Name] Pattern
 	| Mutual      !Range [Declaration]
 	| Abstract    !Range [Declaration]
 	| Private     !Range [Declaration]
@@ -280,12 +327,15 @@ data Declaration
 	| Pragma      Pragma
     deriving (Typeable)
 
-data ModuleApplication = SectionApp Range [TypedBindings] Expr
-                       | RecordModuleIFS Range QName
+data ModuleApplication
+  = SectionApp Range [TypedBindings] Expr
+    -- ^ @tel. M args@
+  | RecordModuleIFS Range QName
+    -- ^ @M {{...}}@
     deriving (Typeable)
 
 data OpenShortHand = DoOpen | DontOpen
-    deriving (Typeable, Show)
+    deriving (Typeable, Eq, Show)
 
 -- Pragmas ----------------------------------------------------------------
 
@@ -294,6 +344,7 @@ data Pragma = OptionsPragma     !Range [String]
             | CompiledDataPragma !Range QName String [String]
             | CompiledTypePragma !Range QName String
             | CompiledPragma    !Range QName String
+            | CompiledExportPragma    !Range QName String
             | CompiledEpicPragma !Range QName String
             | CompiledJSPragma  !Range QName String
             | StaticPragma      !Range QName
@@ -334,9 +385,9 @@ appView (App r e1 e2) = vApp (appView e1) e2
 	vApp (AppView e es) arg = AppView e (es ++ [arg])
 appView (RawApp _ (e:es)) = AppView e $ map arg es
     where
-	arg (HiddenArg   _ e) = Arg Hidden    Relevant e
-	arg (InstanceArg _ e) = Arg Instance  Relevant e
-	arg e		      = Arg NotHidden Relevant (unnamed e)
+	arg (HiddenArg   _ e) = noColorArg Hidden    Relevant e
+	arg (InstanceArg _ e) = noColorArg Instance  Relevant e
+	arg e		      = noColorArg NotHidden Relevant (unnamed e)
 appView e = AppView e []
 
 {--------------------------------------------------------------------------
@@ -370,7 +421,7 @@ patternNames p =
     IdentP x             -> [unqualify x]
     AppP p p'            -> concatMap patternNames [p, namedArg p']
     RawAppP _ ps         -> concatMap patternNames  ps
-    OpAppP _ name ps     -> unqualify name : concatMap patternNames ps
+    OpAppP _ name ps     -> unqualify name : concatMap (patternNames . namedArg) ps
     HiddenP _ (namedPat) -> patternNames (namedThing namedPat)
     ParenP _ p           -> patternNames p
     WildP _              -> []
@@ -419,10 +470,12 @@ instance HasRange Expr where
 	    RecUpdate r _ _	-> r
             ETel tel            -> getRange tel
             QuoteGoal r _ _     -> r
+            QuoteContext r _ _  -> r
             Quote r             -> r
             QuoteTerm r         -> r
             Unquote r           -> r
             DontCare{}          -> noRange
+            Equal r _ _         -> r
 
 -- instance HasRange Telescope where
 --     getRange (TeleBind bs) = getRange bs
@@ -433,11 +486,11 @@ instance HasRange TypedBindings where
 
 instance HasRange TypedBinding where
     getRange (TBind r _ _) = r
-    getRange (TNoBind e)   = getRange e
+    getRange (TLet r _)    = r
 
 instance HasRange LamBinding where
-    getRange (DomainFree _ _ x)	= getRange x
-    getRange (DomainFull b)	= getRange b
+    getRange (DomainFree _ x) = getRange x
+    getRange (DomainFull b)   = getRange b
 
 instance HasRange BoundName where
   getRange = getRange . boundName
@@ -491,6 +544,7 @@ instance HasRange Pragma where
     getRange (CompiledDataPragma r _ _ _) = r
     getRange (CompiledTypePragma r _ _)   = r
     getRange (CompiledPragma r _ _)       = r
+    getRange (CompiledExportPragma r _ _) = r
     getRange (CompiledEpicPragma r _ _)   = r
     getRange (CompiledJSPragma r _ _)     = r
     getRange (StaticPragma r _)           = r
@@ -534,10 +588,10 @@ instance KillRange AsName where
   killRange (AsName n _) = killRange1 (flip AsName noRange) n
 
 instance KillRange BoundName where
-  killRange (BName n f) = killRange2 BName n f
+  killRange (BName n l f) = killRange3 BName n l f
 
 instance KillRange Declaration where
-  killRange (TypeSig r n e)         = killRange2 (TypeSig r) n e
+  killRange (TypeSig i n e)         = killRange2 (TypeSig i) n e
   killRange (Field n a)             = killRange2 Field n a
   killRange (FunClause l r w)       = killRange3 FunClause l r w
   killRange (DataSig _ i n l e)     = killRange4 (DataSig noRange) i n l e
@@ -559,37 +613,39 @@ instance KillRange Declaration where
   killRange (Pragma p)              = killRange1 Pragma p
 
 instance KillRange Expr where
-  killRange (Ident q)           = killRange1 Ident q
-  killRange (Lit l)             = killRange1 Lit l
-  killRange (QuestionMark _ n)  = QuestionMark noRange n
-  killRange (Underscore _ n)    = Underscore noRange n
-  killRange (RawApp _ e)        = killRange1 (RawApp noRange) e
-  killRange (App _ e a)         = killRange2 (App noRange) e a
-  killRange (OpApp _ n o)       = killRange2 (OpApp noRange) n o
-  killRange (WithApp _ e es)    = killRange2 (WithApp noRange) e es
-  killRange (HiddenArg _ n)     = killRange1 (HiddenArg noRange) n
-  killRange (InstanceArg _ n)   = killRange1 (InstanceArg noRange) n
-  killRange (Lam _ l e)         = killRange2 (Lam noRange) l e
-  killRange (AbsurdLam _ h)     = killRange1 (AbsurdLam noRange) h
-  killRange (ExtendedLam _ lrw) = killRange1 (ExtendedLam noRange) lrw
-  killRange (Fun _ e1 e2)       = killRange2 (Fun noRange) e1 e2
-  killRange (Pi t e)            = killRange2 Pi t e
-  killRange (Set _)             = Set noRange
-  killRange (Prop _)            = Prop noRange
-  killRange (SetN _ n)          = SetN noRange n
-  killRange (Rec _ ne)          = killRange1 (Rec noRange) ne
-  killRange (RecUpdate _ e ne)  = killRange2 (RecUpdate noRange) e ne
-  killRange (Let _ d e)         = killRange2 (Let noRange) d e
-  killRange (Paren _ e)         = killRange1 (Paren noRange) e
-  killRange (Absurd _)          = Absurd noRange
-  killRange (As _ n e)          = killRange2 (As noRange) n e
-  killRange (Dot _ e)           = killRange1 (Dot noRange) e
-  killRange (ETel t)            = killRange1 ETel t
-  killRange (QuoteGoal _ n e)   = killRange2 (QuoteGoal noRange) n e
-  killRange (Quote _)           = Quote noRange
-  killRange (QuoteTerm _)       = QuoteTerm noRange
-  killRange (Unquote _)         = Unquote noRange
-  killRange (DontCare e)        = killRange1 DontCare e
+  killRange (Ident q)            = killRange1 Ident q
+  killRange (Lit l)              = killRange1 Lit l
+  killRange (QuestionMark _ n)   = QuestionMark noRange n
+  killRange (Underscore _ n)     = Underscore noRange n
+  killRange (RawApp _ e)         = killRange1 (RawApp noRange) e
+  killRange (App _ e a)          = killRange2 (App noRange) e a
+  killRange (OpApp _ n o)        = killRange2 (OpApp noRange) n o
+  killRange (WithApp _ e es)     = killRange2 (WithApp noRange) e es
+  killRange (HiddenArg _ n)      = killRange1 (HiddenArg noRange) n
+  killRange (InstanceArg _ n)    = killRange1 (InstanceArg noRange) n
+  killRange (Lam _ l e)          = killRange2 (Lam noRange) l e
+  killRange (AbsurdLam _ h)      = killRange1 (AbsurdLam noRange) h
+  killRange (ExtendedLam _ lrw)  = killRange1 (ExtendedLam noRange) lrw
+  killRange (Fun _ e1 e2)        = killRange2 (Fun noRange) e1 e2
+  killRange (Pi t e)             = killRange2 Pi t e
+  killRange (Set _)              = Set noRange
+  killRange (Prop _)             = Prop noRange
+  killRange (SetN _ n)           = SetN noRange n
+  killRange (Rec _ ne)           = killRange1 (Rec noRange) ne
+  killRange (RecUpdate _ e ne)   = killRange2 (RecUpdate noRange) e ne
+  killRange (Let _ d e)          = killRange2 (Let noRange) d e
+  killRange (Paren _ e)          = killRange1 (Paren noRange) e
+  killRange (Absurd _)           = Absurd noRange
+  killRange (As _ n e)           = killRange2 (As noRange) n e
+  killRange (Dot _ e)            = killRange1 (Dot noRange) e
+  killRange (ETel t)             = killRange1 ETel t
+  killRange (QuoteGoal _ n e)    = killRange2 (QuoteGoal noRange) n e
+  killRange (QuoteContext _ n e) = killRange2 (QuoteContext noRange) n e
+  killRange (Quote _)            = Quote noRange
+  killRange (QuoteTerm _)        = QuoteTerm noRange
+  killRange (Unquote _)          = Unquote noRange
+  killRange (DontCare e)         = killRange1 DontCare e
+  killRange (Equal _ x y)        = Equal noRange x y
 
 instance KillRange ImportDirective where
   killRange (ImportDirective _ u r p) =
@@ -600,8 +656,8 @@ instance KillRange ImportedName where
   killRange (ImportedName   n) = killRange1 ImportedName   n
 
 instance KillRange LamBinding where
-  killRange (DomainFree h r b) = killRange2 (\h -> DomainFree h r) h b
-  killRange (DomainFull t)     = killRange1 DomainFull t
+  killRange (DomainFree i b) = killRange2 DomainFree i b
+  killRange (DomainFull t)   = killRange1 DomainFull t
 
 instance KillRange LHS where
   killRange (LHS p ps r w)     = killRange4 LHS p ps r w
@@ -635,6 +691,7 @@ instance KillRange Pragma where
   killRange (CompiledDataPragma _ q s ss) = killRange1 (\q -> CompiledDataPragma noRange q s ss) q
   killRange (CompiledTypePragma _ q s)    = killRange1 (\q -> CompiledTypePragma noRange q s) q
   killRange (CompiledPragma _ q s)        = killRange1 (\q -> CompiledPragma noRange q s) q
+  killRange (CompiledExportPragma _ q s)  = killRange1 (\q -> CompiledExportPragma noRange q s) q
   killRange (CompiledEpicPragma _ q s)    = killRange1 (\q -> CompiledEpicPragma noRange q s) q
   killRange (CompiledJSPragma _ q s)      = killRange1 (\q -> CompiledJSPragma noRange q s) q
   killRange (StaticPragma _ q)            = killRange1 (StaticPragma noRange) q
@@ -652,7 +709,7 @@ instance KillRange RHS where
 
 instance KillRange TypedBinding where
   killRange (TBind _ b e) = killRange2 (TBind noRange) b e
-  killRange (TNoBind e)   = killRange1 TNoBind e
+  killRange (TLet r ds)   = killRange2 TLet r ds
 
 instance KillRange TypedBindings where
   killRange (TypedBindings _ t) = killRange1 (TypedBindings noRange) t
