@@ -1,4 +1,6 @@
-{-# LANGUAGE CPP, PatternGuards, DeriveDataTypeable #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE PatternGuards #-}
 
 module Agda.Syntax.Concrete.Definitions
     ( NiceDeclaration(..)
@@ -10,38 +12,33 @@ module Agda.Syntax.Concrete.Definitions
     , notSoNiceDeclaration
     ) where
 
-import Control.Arrow ((***), (&&&))
+import Control.Arrow ((***))
 import Control.Applicative
+import Control.Monad.Error
+import Control.Monad.State
+
 import Data.Typeable (Typeable)
 import Data.Foldable hiding (concatMap, mapM_, notElem, elem, all)
 import qualified Data.Map as Map
 import Data.Map (Map)
-import Control.Monad.Error
-import Control.Monad.State
 import Data.List as List
-import Data.Maybe
 import Data.Traversable (traverse)
 
-import Debug.Trace (trace)
 import Agda.Syntax.Concrete
-import Agda.Syntax.Common
+import Agda.Syntax.Common hiding (Arg, Dom, NamedArg, ArgInfo)
+import qualified Agda.Syntax.Common as Common
 import Agda.Syntax.Position
 import Agda.Syntax.Fixity
 import Agda.Syntax.Notation
-import Agda.Syntax.Concrete.Pretty
+import Agda.Syntax.Concrete.Pretty ()
 import Agda.Utils.Pretty
 import Agda.Utils.List (mhead, isSublistOf)
 import Agda.Utils.Monad
+import Agda.Utils.Update
 
 #include "../../undefined.h"
 import Agda.Utils.Impossible
 
-import Control.Applicative
-import Control.Monad.Error
-import Data.Char
-
-import Debug.Trace
-
 {--------------------------------------------------------------------------
     Types
  --------------------------------------------------------------------------}
@@ -52,13 +49,13 @@ import Debug.Trace
     modifiers have been distributed to the individual declarations.
 -}
 data NiceDeclaration
-        = Axiom Range Fixity' Access Relevance Name Expr
-            -- ^ Axioms and functions can be declared irrelevant.
+        = Axiom Range Fixity' Access ArgInfo Name Expr
+            -- ^ Axioms and functions can be declared irrelevant. (Hiding should be NotHidden)
         | NiceField Range Fixity' Access IsAbstract Name (Arg Expr)
         | PrimitiveFunction Range Fixity' Access IsAbstract Name Expr
         | NiceMutual Range TerminationCheck [NiceDeclaration]
         | NiceModule Range Access IsAbstract QName Telescope [Declaration]
-        | NiceModuleMacro Range Access IsAbstract Name ModuleApplication OpenShortHand ImportDirective
+        | NiceModuleMacro Range Access Name ModuleApplication OpenShortHand ImportDirective
         | NiceOpen Range QName ImportDirective
         | NiceImport Range QName (Maybe AsName) OpenShortHand ImportDirective
         | NicePragma Range Pragma
@@ -67,11 +64,11 @@ data NiceDeclaration
         | NiceFunClause Range Access IsAbstract TerminationCheck Declaration
           -- ^ a uncategorized function clause, could be a function clause
           --   without type signature or a pattern lhs (e.g. for irrefutable let)x
-        | FunSig Range Fixity' Access Relevance TerminationCheck Name Expr
+        | FunSig Range Fixity' Access ArgInfo TerminationCheck Name Expr
         | FunDef  Range [Declaration] Fixity' IsAbstract TerminationCheck Name [Clause] -- ^ block of function clauses (we have seen the type signature before)
         | DataDef Range Fixity' IsAbstract Name [LamBinding] [NiceConstructor]
-        | RecDef Range Fixity' IsAbstract Name (Maybe Induction) (Maybe (ThingWithFixity Name)) [LamBinding] [NiceDeclaration]
-        | NicePatternSyn Range Fixity' Name [Name] Pattern
+        | RecDef Range Fixity' IsAbstract Name (Maybe (Ranged Induction)) (Maybe (ThingWithFixity Name)) [LamBinding] [NiceDeclaration]
+        | NicePatternSyn Range Fixity' Name [Arg Name] Pattern
     deriving (Typeable, Show)
 
 -- | Termination check? (Default = True).
@@ -129,7 +126,7 @@ instance HasRange NiceDeclaration where
   getRange (NiceField r _ _ _ _ _)         = r
   getRange (NiceMutual r _ _)              = r
   getRange (NiceModule r _ _ _ _ _)        = r
-  getRange (NiceModuleMacro r _ _ _ _ _ _) = r
+  getRange (NiceModuleMacro r _ _ _ _ _)   = r
   getRange (NiceOpen r _ _)                = r
   getRange (NiceImport r _ _ _ _)          = r
   getRange (NicePragma r _)                = r
@@ -173,9 +170,9 @@ instance Show DeclarationException where
   show (UnknownNamesInFixityDecl xs) = show $ fsep $
     pwords "Names out of scope in fixity declarations:" ++ map pretty xs
   show (UselessPrivate _)      = show $ fsep $
-    pwords "Using private here has no effect. Move it to the type signatures to make the definitions private."
+    pwords "Using private here has no effect. Private applies only to declarations that introduce new identifiers into the module, like type signatures and data, record, and module declarations."
   show (UselessAbstract _)      = show $ fsep $
-    pwords "Using abstract here has no effect. Move it to the definitions to make them abstract."
+    pwords "Using abstract here has no effect. Abstract applies only definitions like data definitions, record type definitions and function clauses."
   show (InvalidNoTerminationCheckPragma _) = show $ fsep $
     pwords "The NO_TERMINATION_CHECK pragma can only preceed a mutual block or a function definition."
   show (NotAllowedInMutual nd) = show $ fsep $
@@ -248,11 +245,6 @@ initNiceEnv = NiceEnv
 
 type Nice = StateT NiceEnv (Either DeclarationException)
 
-{- Andreas, 2012-04-30 MOVED to Utils.Monad
-localState :: Nice a -> Nice a
-localState m = bracket get put (const m)
--}
-
 addLoneSig :: DataRecOrFun -> Name -> Nice ()
 addLoneSig k x = modify $ \ niceEnv -> niceEnv { loneSigs = (k, x) : loneSigs niceEnv }
 
@@ -280,7 +272,11 @@ getFixity x = gets $ Map.findWithDefault defaultFixity' x . fixs
 runNice :: Nice a -> Either DeclarationException a
 runNice nice = nice `evalStateT` initNiceEnv
 
-data DeclKind = LoneSig DataRecOrFun Name | LoneDef DataRecOrFun Name | OtherDecl
+data DeclKind
+    = LoneSig DataRecOrFun Name
+    | LoneDef DataRecOrFun Name
+    | OtherDecl
+  deriving (Eq, Show)
 
 declKind (FunSig _ _ _ _ tc x _)      = LoneSig (FunName tc) x
 declKind (NiceRecSig _ _ _ x pars _)  = LoneSig (RecName $ parameters pars) x
@@ -293,9 +289,9 @@ declKind _                            = OtherDecl
 -- | Compute visible parameters of a data or record signature or definition.
 parameters :: [LamBinding] -> Params
 parameters = List.concat . List.map numP where
-  numP (DomainFree h _ _) = [h]
-  numP (DomainFull (TypedBindings _ (Arg h _ (TBind _ xs _)))) = List.replicate (length xs) h
-  numP (DomainFull (TypedBindings _ (Arg _ _ (TNoBind{})))) =  __IMPOSSIBLE__
+  numP (DomainFree i _) = [argInfoHiding i]
+  numP (DomainFull (TypedBindings _ (Common.Arg i (TBind _ xs _)))) = List.replicate (length xs) $ argInfoHiding i
+  numP (DomainFull (TypedBindings _ (Common.Arg _ TLet{})))         = []
 
 {- OLD:
 
@@ -304,7 +300,6 @@ numberOfPars :: [LamBinding] -> Params
 numberOfPars = List.sum . List.map numP where
   numP (DomainFree NotHidden _ _) = 1
   numP (DomainFull (TypedBindings _ (Arg NotHidden _ (TBind _ xs _)))) = length xs
-  numP (DomainFull (TypedBindings _ (Arg _ _ (TNoBind{})))) =  __IMPOSSIBLE__
   numP _ = 0  -- hidden / instance argument
 -- | Compute number of parameters of a data or record signature or definition.
 numberOfPars :: [LamBinding] -> Int
@@ -312,7 +307,6 @@ numberOfPars = List.sum . List.map numP where
   numP (DomainFree{}) = 1
   numP (DomainFull (TypedBindings _ arg)) = nP $ unArg arg where
     nP (TBind _ xs _) = length xs
-    nP (TNoBind _)    = __IMPOSSIBLE__
 -}
 
 niceDeclarations :: [Declaration] -> Nice [NiceDeclaration]
@@ -332,7 +326,7 @@ niceDeclarations ds = do
     declaredNames d = case d of
       TypeSig _ x _                                -> [x]
       Field x _                                    -> [x]
-      FunClause (LHS p [] _ _) _ _
+      FunClause (LHS p [] [] []) _ _
         | IdentP (QName x) <- removeSingletonRawAppP p -> [x]
       FunClause{}                                  -> []
       DataSig _ _ x _ _                            -> [x]
@@ -439,7 +433,7 @@ niceDeclarations ds = do
           (NiceModule r PublicAccess ConcreteDef x tel ds' :) <$> nice ds
 
         ModuleMacro r x modapp op is ->
-          (NiceModuleMacro r PublicAccess ConcreteDef x modapp op is :)
+          (NiceModuleMacro r PublicAccess x modapp op is :)
             <$> nice ds
 
         Infix _ _           -> nice ds
@@ -470,19 +464,22 @@ niceDeclarations ds = do
                       span (couldBeFunClauseOf (Map.lookup x fixs) x) (d : ds)
                , not (null fits)
                ] of
+
             -- case: clauses match none of the sigs
-            -- then the lhs can only be a single identifier
-            -- treat it as a function clause without a type signature
             [] -> case lhs of
-              LHS p [] _ _ | IdentP (QName x) <- removeSingletonRawAppP p -> do
+              -- Subcase: The lhs is single identifier.
+              -- Treat it as a function clause without a type signature.
+              LHS p [] [] [] | IdentP (QName x) <- removeSingletonRawAppP p -> do
                 ds <- nice ds
-                d  <- mkFunDef Relevant termCheck x Nothing [d] -- fun def without type signature is relevant
+                d  <- mkFunDef defaultArgInfo termCheck x Nothing [d] -- fun def without type signature is relevant
                 return $ d ++ ds
+              -- Subcase: The lhs is a proper pattern.
+              -- This could be a let-pattern binding. Pass it on.
+              -- A missing type signature error might be raise in ConcreteToAbstract
               _ -> do
                 ds <- nice ds
                 return $ NiceFunClause (getRange d) PublicAccess ConcreteDef termCheck d : ds
--- OLD:             _ -> throwError $ MissingTypeSignature lhs
--- error now raised in ConcreteToAbstract
+
             -- case: clauses match exactly one of the sigs
             [(x,(fits,rest))] -> do
                removeLoneSig x
@@ -491,17 +488,18 @@ niceDeclarations ds = do
                fx  <- getFixity x
                d   <- return $ FunDef (getRange fits) fits fx ConcreteDef termCheck x cs
                return $ d : ds1
+
             -- case: clauses match more than one sigs (ambiguity)
             l -> throwError $ AmbiguousFunClauses lhs (map fst l) -- "ambiguous function clause; cannot assign it uniquely to one type signature"
     niceFunClause _ _ _ = __IMPOSSIBLE__
 
     niceTypeSig :: TerminationCheck -> Declaration -> [Declaration] -> Nice [NiceDeclaration]
-    niceTypeSig termCheck d@(TypeSig rel x t) ds = do
+    niceTypeSig termCheck d@(TypeSig info x t) ds = do
       fx <- getFixity x
       -- register x as lone type signature, to recognize clauses later
       addLoneSig (FunName termCheck) x
       ds <- nice ds
-      return $ FunSig (getRange d) fx PublicAccess rel termCheck x t : ds
+      return $ FunSig (getRange d) fx PublicAccess info termCheck x t : ds
     niceTypeSig _ _ _ = __IMPOSSIBLE__
 
     -- We could add a default type signature here, but at the moment we can't
@@ -523,10 +521,9 @@ niceDeclarations ds = do
          [mkSig (fuseRange x t) f PublicAccess x tel t | Just t <- [mt] ] ++
          [mkDef (getRange x) f ConcreteDef x (concatMap dropType tel) ds | Just ds <- [mds] ]
       where
-        dropType (DomainFull (TypedBindings r (Arg h rel TNoBind{}))) =
-          [DomainFree h rel $ mkBoundName_ $ noName r]
-        dropType (DomainFull (TypedBindings r (Arg h rel (TBind _ xs _)))) =
-          map (DomainFree h rel) xs
+        dropType (DomainFull (TypedBindings r (Common.Arg i (TBind _ xs _)))) =
+          map (DomainFree i) xs
+        dropType (DomainFull (TypedBindings _ (Common.Arg _ TLet{}))) = []
         dropType b@DomainFree{} = [b]
 
     -- Translate axioms
@@ -547,10 +544,10 @@ niceDeclarations ds = do
     toPrim _                     = __IMPOSSIBLE__
 
     -- Create a function definition.
-    mkFunDef rel termCheck x mt ds0 = do
+    mkFunDef info termCheck x mt ds0 = do
       cs <- mkClauses x $ expandEllipsis ds0
       f  <- getFixity x
-      return [ FunSig (fuseRange x t) f PublicAccess rel termCheck x t
+      return [ FunSig (fuseRange x t) f PublicAccess info termCheck x t
              , FunDef (getRange ds0) ds0 f ConcreteDef termCheck x cs ]
         where
           t = case mt of
@@ -632,18 +629,19 @@ niceDeclarations ds = do
         _ -> False -- trace ("couldBe not (case default)") $ False
     couldBeFunClauseOf _ _ _ = False -- trace ("couldBe not (fun default)") $ False
 
+    -- ASR (27 May 2014). Commented out unused code.
     -- @isFunClauseOf@ is for non-mutual blocks where clauses must follow the
     -- type sig immediately
-    isFunClauseOf :: Name -> Declaration -> Bool
-    isFunClauseOf x (FunClause Ellipsis{} _ _) = True
-    isFunClauseOf x (FunClause (LHS p _ _ _) _ _) =
-     -- p is the whole left hand side, excluding "with" patterns and clauses
-      case removeSingletonRawAppP p of
-        IdentP (QName q)    -> x == q  -- lhs is just an identifier
-        _                   -> True
-            -- more complicated lhss must come with type signatures, so we just assume
-            -- it's part of the current definition
-    isFunClauseOf _ _ = False
+    -- isFunClauseOf :: Name -> Declaration -> Bool
+    -- isFunClauseOf x (FunClause Ellipsis{} _ _) = True
+    -- isFunClauseOf x (FunClause (LHS p _ _ _) _ _) =
+    --  -- p is the whole left hand side, excluding "with" patterns and clauses
+    --   case removeSingletonRawAppP p of
+    --     IdentP (QName q)    -> x == q  -- lhs is just an identifier
+    --     _                   -> True
+    --         -- more complicated lhss must come with type signatures, so we just assume
+    --         -- it's part of the current definition
+    -- isFunClauseOf _ _ = False
 
     removeSingletonRawAppP :: Pattern -> Pattern
     removeSingletonRawAppP (RawAppP _ [p]) = removeSingletonRawAppP p
@@ -655,22 +653,80 @@ niceDeclarations ds = do
         -- Check that there aren't any missing definitions
         checkLoneSigs loneNames
         -- Check that there are no declarations that aren't allowed in old style mutual blocks
-        case [ d | (d, OtherDecl) <- zip ds $ map declKind ds ] of
+        case filter notAllowedInMutual ds of
           []  -> return ()
           (NiceFunClause _ _ _ _ (FunClause lhs _ _)):_ -> throwError $ MissingTypeSignature lhs
           d:_ -> throwError $ NotAllowedInMutual d
-        return $ NiceMutual r True $ sigs ++ other  -- termCheck is adjusted later
+        return $ NiceMutual r (all termCheck ds) $ sigs ++ other
       where
+        -- Andreas, 2013-11-23 allow postulates in mutual blocks
+        notAllowedInMutual Axiom{} = False
+        notAllowedInMutual d       = declKind d == OtherDecl
         -- Pull type signatures to the top
         (sigs, other) = partition isTypeSig ds
+        isTypeSig Axiom{}                     = True
         isTypeSig d | LoneSig{} <- declKind d = True
-        isTypeSig _ = False
+        isTypeSig _                           = False
 
         sigNames  = [ (k, x) | LoneSig k x <- map declKind ds ]
         defNames  = [ (k, x) | LoneDef k x <- map declKind ds ]
         -- compute the set difference with equality just on names
-        loneNames = filter (\ (_, x) -> not (List.any (\ (_, x') -> x == x') defNames)) sigNames
+        loneNames = [ (k, x) | (k, x) <- sigNames, List.all ((x /=) . snd) defNames ]
 
+        -- Andreas, 2013-02-28 (issue 804):
+        -- do not termination check a mutual block if any of its
+        -- inner declarations comes with a {-# NO_TERMINATION_CHECK #-}
+        termCheck (FunSig _ _ _ _ tc _ _) = tc
+        termCheck (FunDef _ _ _ _ tc _ _) = tc
+        termCheck (NiceMutual _ tc _)     = tc
+        termCheck _                       = True
+
+    abstractBlock _ [] = return []
+    abstractBlock r ds = do
+      let (ds', anyChange) = runChange $ mapM mkAbstract ds
+          inherited        = r == noRange
+          -- hack to avoid failing on inherited abstract blocks in where clauses
+      if anyChange || inherited then return ds' else throwError $ UselessAbstract r
+
+    -- Make a declaration abstract
+    mkAbstract :: Updater NiceDeclaration
+    mkAbstract d =
+      case d of
+        NiceMutual r termCheck ds        -> NiceMutual r termCheck <$> mapM mkAbstract ds
+        FunDef r ds f a tc x cs          -> (\ a -> FunDef r ds f a tc x)  <$> setAbstract a <*> mapM mkAbstractClause cs
+        DataDef r f a x ps cs            -> (\ a -> DataDef r f a x ps)    <$> setAbstract a <*> mapM mkAbstract cs
+        RecDef r f a x i c ps cs         -> (\ a -> RecDef r f a x i c ps) <$> setAbstract a <*> mapM mkAbstract cs
+        NiceFunClause r p a termCheck d  -> (\ a -> NiceFunClause r p a termCheck d) <$> setAbstract a
+        -- no effect on fields or primitives, the InAbstract field there is unused
+        NiceField r f p _ x e            -> return $ NiceField r f p AbstractDef x e
+        PrimitiveFunction r f p _ x e    -> return $ PrimitiveFunction r f p AbstractDef x e
+        NiceModule{}                     -> return $ d
+        NiceModuleMacro{}                -> return $ d
+        Axiom{}                          -> return $ d
+        NicePragma{}                     -> return $ d
+        NiceOpen{}                       -> return $ d
+        NiceImport{}                     -> return $ d
+        FunSig{}                         -> return $ d
+        NiceRecSig{}                     -> return $ d
+        NiceDataSig{}                    -> return $ d
+        NicePatternSyn{}                 -> return $ d
+
+    setAbstract :: Updater IsAbstract
+    setAbstract a = case a of
+      AbstractDef -> return a
+      ConcreteDef -> dirty $ AbstractDef
+
+    mkAbstractClause :: Updater Clause
+    mkAbstractClause (Clause x lhs rhs wh with) = do
+        wh <- mkAbstractWhere wh
+        Clause x lhs rhs wh <$> mapM mkAbstractClause with
+
+    mkAbstractWhere :: Updater WhereClause
+    mkAbstractWhere  NoWhere         = return $ NoWhere
+    mkAbstractWhere (AnyWhere ds)    = dirty $ AnyWhere [Abstract noRange ds]
+    mkAbstractWhere (SomeWhere m ds) = dirty $SomeWhere m [Abstract noRange ds]
+
+{- OLD CODE
     abstractBlock _ [] = return []
     abstractBlock r ds
         -- hack to avoid failing on inherited abstract blocks in where clauses
@@ -684,29 +740,18 @@ niceDeclarations ds = do
       RecDef{}  -> False
       _         -> True
 
-    privateBlock _ [] = return []
-    privateBlock r ds
-      | all uselessPrivate ds = throwError $ UselessPrivate r
-      | otherwise             = return $ map mkPrivate ds
-
-    uselessPrivate d = case d of
-      FunDef{}  -> True
-      DataDef{} -> True
-      RecDef{}  -> True
-      _         -> False
-
     -- Make a declaration abstract
     mkAbstract d =
         case d of
             NiceField r f a _ x e            -> NiceField r f a AbstractDef x e
             PrimitiveFunction r f a _ x e    -> PrimitiveFunction r f a AbstractDef x e
             NiceMutual r termCheck ds        -> NiceMutual r termCheck (map mkAbstract ds)
-            NiceModuleMacro r a _ x ma op is -> NiceModuleMacro r a AbstractDef x ma op is
             FunDef r ds f _ tc x cs          -> FunDef r ds f AbstractDef tc x (map mkAbstractClause cs)
             DataDef r f _ x ps cs            -> DataDef r f AbstractDef x ps $ map mkAbstract cs
             RecDef r f _ x i c ps cs         -> RecDef r f AbstractDef x i c ps $ map mkAbstract cs
             NiceFunClause r a _ termCheck d  -> NiceFunClause r a AbstractDef termCheck d
             NiceModule{}                     -> d
+            NiceModuleMacro{}                -> d
             Axiom{}                          -> d
             NicePragma{}                     -> d
             NiceOpen{}                       -> d
@@ -722,6 +767,80 @@ niceDeclarations ds = do
     mkAbstractWhere  NoWhere         = NoWhere
     mkAbstractWhere (AnyWhere ds)    = AnyWhere [Abstract noRange ds]
     mkAbstractWhere (SomeWhere m ds) = SomeWhere m [Abstract noRange ds]
+-}
+    privateBlock _ [] = return []
+    privateBlock r ds = do
+      let (ds', anyChange) = runChange $ mapM mkPrivate ds
+      if anyChange then return ds' else throwError $ UselessPrivate r
+
+    -- Make a declaration private.
+    -- Andreas, 2012-11-17:
+    -- Mark computation 'dirty' if there was a declaration that could be privatized.
+    -- If no privatization is taking place, we want to complain about 'UselessPrivate'.
+    -- Alternatively, we could only dirty if a non-private thing was privatized.
+    -- Then, nested 'private's would sometimes also be complained about.
+    mkPrivate :: Updater NiceDeclaration
+    mkPrivate d =
+      case d of
+        Axiom r f p rel x e              -> (\ p -> Axiom r f p rel x e) <$> setPrivate p
+        NiceField r f p a x e            -> (\ p -> NiceField r f p a x e) <$> setPrivate p
+        PrimitiveFunction r f p a x e    -> (\ p -> PrimitiveFunction r f p a x e) <$> setPrivate p
+        NiceMutual r termCheck ds        -> NiceMutual r termCheck <$> mapM mkPrivate ds
+        NiceModule r p a x tel ds        -> (\ p -> NiceModule r p a x tel ds) <$> setPrivate p
+        NiceModuleMacro r p x ma op is   -> (\ p -> NiceModuleMacro r p x ma op is) <$> setPrivate p
+        FunSig r f p rel tc x e          -> (\ p -> FunSig r f p rel tc x e) <$> setPrivate p
+        NiceRecSig r f p x ls t          -> (\ p -> NiceRecSig r f p x ls t) <$> setPrivate p
+        NiceDataSig r f p x ls t         -> (\ p -> NiceDataSig r f p x ls t) <$> setPrivate p
+        NiceFunClause r p a termCheck d  -> (\ p -> NiceFunClause r p a termCheck d) <$> setPrivate p
+{-
+        Axiom r f _ rel x e              -> dirty $ Axiom r f PrivateAccess rel x e
+        NiceField r f _ a x e            -> dirty $ NiceField r f PrivateAccess a x e
+        PrimitiveFunction r f _ a x e    -> dirty $ PrimitiveFunction r f PrivateAccess a x e
+        NiceMutual r termCheck ds        -> NiceMutual r termCheck <$> mapM mkPrivate ds
+        NiceModule r _ a x tel ds        -> dirty $ NiceModule r PrivateAccess a x tel ds
+        NiceModuleMacro r _ x ma op is   -> dirty $ NiceModuleMacro r PrivateAccess x ma op is
+        FunSig r f _ rel tc x e          -> dirty $ FunSig r f PrivateAccess rel tc x e
+        NiceRecSig r f _ x ls t          -> dirty $ NiceRecSig r f PrivateAccess x ls t
+        NiceDataSig r f _ x ls t         -> dirty $ NiceDataSig r f PrivateAccess x ls t
+        NiceFunClause r _ a termCheck d  -> dirty $ NiceFunClause r PrivateAccess a termCheck d
+-}
+        NicePragma _ _                   -> return $ d
+        NiceOpen _ _ _                   -> return $ d
+        NiceImport _ _ _ _ _             -> return $ d
+        FunDef{}                         -> return $ d
+        DataDef{}                        -> return $ d
+        RecDef{}                         -> return $ d
+        NicePatternSyn _ _ _ _ _         -> return $ d
+
+    setPrivate :: Updater Access
+    setPrivate p = case p of
+      PrivateAccess -> return p
+      _             -> dirty $ PrivateAccess
+
+    -- Andreas, 2012-11-22: Q: is this necessary?
+    -- Are where clauses not always private?
+    mkPrivateClause :: Updater Clause
+    mkPrivateClause (Clause x lhs rhs wh with) = do
+        wh <- mkPrivateWhere wh
+        Clause x lhs rhs wh <$> mapM mkPrivateClause with
+
+    mkPrivateWhere :: Updater WhereClause
+    mkPrivateWhere  NoWhere         = return $ NoWhere
+    mkPrivateWhere (AnyWhere ds)    = dirty  $ AnyWhere [Private (getRange ds) ds]
+    mkPrivateWhere (SomeWhere m ds) = dirty  $ SomeWhere m [Private (getRange ds) ds]
+
+{- OLD CODE, with two functions (uselessPrivate, mkPrivate) to be maintained in sync.
+
+    privateBlock _ [] = return []
+    privateBlock r ds
+      | all uselessPrivate ds = throwError $ UselessPrivate r
+      | otherwise             = return $ map mkPrivate ds
+
+    uselessPrivate d = case d of
+      FunDef{}  -> True
+      DataDef{} -> True
+      RecDef{}  -> True
+      _         -> False
 
     -- Make a declaration private
     mkPrivate d =
@@ -731,7 +850,7 @@ niceDeclarations ds = do
             PrimitiveFunction r f _ a x e    -> PrimitiveFunction r f PrivateAccess a x e
             NiceMutual r termCheck ds        -> NiceMutual r termCheck (map mkPrivate ds)
             NiceModule r _ a x tel ds        -> NiceModule r PrivateAccess a x tel ds
-            NiceModuleMacro r _ a x ma op is -> NiceModuleMacro r PrivateAccess a x ma op is
+            NiceModuleMacro r _ x ma op is   -> NiceModuleMacro r PrivateAccess x ma op is
             FunSig r f _ rel tc x e          -> FunSig r f PrivateAccess rel tc x e
             NiceRecSig r f _ x ls t          -> NiceRecSig r f PrivateAccess x ls t
             NiceDataSig r f _ x ls t         -> NiceDataSig r f PrivateAccess x ls t
@@ -750,6 +869,7 @@ niceDeclarations ds = do
     mkPrivateWhere  NoWhere         = NoWhere
     mkPrivateWhere (AnyWhere ds)    = AnyWhere [Private (getRange ds) ds]
     mkPrivateWhere (SomeWhere m ds) = SomeWhere m [Private (getRange ds) ds]
+-}
 
 -- | Add more fixities. Throw an exception for multiple fixity declarations.
 plusFixities :: Map.Map Name Fixity' -> Map.Map Name Fixity' -> Nice (Map.Map Name Fixity')
@@ -793,10 +913,10 @@ notSoNiceDeclaration d =
     case d of
       Axiom _ _ _ rel x e              -> TypeSig rel x e
       NiceField _ _ _ _ x argt         -> Field x argt
-      PrimitiveFunction r _ _ _ x e    -> Primitive r [TypeSig Relevant x e]
+      PrimitiveFunction r _ _ _ x e    -> Primitive r [TypeSig defaultArgInfo x e]
       NiceMutual r _ ds                -> Mutual r $ map notSoNiceDeclaration ds
       NiceModule r _ _ x tel ds        -> Module r x tel ds
-      NiceModuleMacro r _ _ x ma o dir -> ModuleMacro r x ma o dir
+      NiceModuleMacro r _ x ma o dir   -> ModuleMacro r x ma o dir
       NiceOpen r x dir                 -> Open r x dir
       NiceImport r x as o dir          -> Import r x as o dir
       NicePragma _ p                   -> Pragma p
@@ -823,7 +943,7 @@ notSoNiceDeclarations = concatMap notNice
       PrimitiveFunction r _ _ _ x e    -> [Primitive r [TypeSig Relevant x e]]
       NiceMutual _ _ ds                -> concatMap notNice ds
       NiceModule r _ _ x tel ds        -> [Module r x tel ds]
-      NiceModuleMacro r _ _ x ma o dir -> [ModuleMacro r x ma o dir]
+      NiceModuleMacro r _ x ma o dir   -> [ModuleMacro r x ma o dir]
       NiceOpen r x dir                 -> [Open r x dir]
       NiceImport r x as o dir          -> [Import r x as o dir]
       NicePragma _ p                   -> [Pragma p]
