@@ -1,4 +1,5 @@
-{-# LANGUAGE CPP, ScopedTypeVariables #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 
 {-| The parser doesn't know about operators and parses everything as normal
     function application. This module contains the functions that parses the
@@ -27,20 +28,23 @@ module Agda.Syntax.Concrete.Operators
     , patternQNames
     ) where
 
+import Control.DeepSeq
 import Control.Applicative
 import Control.Monad
-import Control.Monad.Trans
-import Data.Typeable
+
+import Data.Either (partitionEithers)
+import Data.Function
+import Data.List
 import Data.Traversable (traverse)
 import qualified Data.Traversable as Trav
+import Data.Map (Map)
 import qualified Data.Map as Map
-import qualified Data.Set as Set
 import Data.Set (Set)
-import Data.List
-import Data.Function
+import qualified Data.Set as Set
 
 import Agda.Syntax.Concrete.Pretty ()
-import Agda.Syntax.Common
+import Agda.Syntax.Common hiding (Arg, Dom, NamedArg)
+import qualified Agda.Syntax.Common as Common
 import Agda.Syntax.Concrete hiding (appView)
 import Agda.Syntax.Concrete.Operators.Parser
 import qualified Agda.Syntax.Abstract.Name as A
@@ -51,18 +55,15 @@ import Agda.Syntax.Scope.Base
 import Agda.Syntax.Scope.Monad
 
 import Agda.TypeChecking.Monad.Base (typeError, TypeError(..), LHSOrPatSyn(..))
+import Agda.TypeChecking.Monad.Benchmark (billSub)
+import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 import Agda.TypeChecking.Monad.State (getScope)
 import Agda.TypeChecking.Monad.Options
-import Agda.TypeChecking.Monad.Statistics
 
 import Agda.Utils.Either
 import Agda.Utils.ReadP
-import Agda.Utils.Monad
-import Agda.Utils.Tuple
 import Agda.Utils.List
 
-import Debug.Trace
-
 #include "../../undefined.h"
 import Agda.Utils.Impossible
 
@@ -93,8 +94,12 @@ getDefinedNames kinds names =
   [ (x, A.nameFixity $ A.qnameName $ anameName d)
   | (x, ds) <- Map.toList names
   , d       <- take 1 ds
-  , anameKind d `elem` kinds
-  ]
+  , any (\ d -> anameKind d `elem` kinds) ds
+  -- Andreas, 2013-03-21 see Issue 822
+  -- Names can have different kinds, i.e., 'defined' and 'constructor'.
+  -- We need to consider all names that have *any* matching kind,
+  -- not only those whose first appearing kind is matching.
+ ]
 
 -- | Compute all names (first component) and operators (second component) in
 --   scope.
@@ -105,9 +110,7 @@ localNames flat = do
   return $ split $ uniqBy fst $ map localOp locals ++ defs
   where
     localOp (x, y) = (QName x, A.nameFixity y)
-    split ops = ([ x | Left x <- zs], [ y | Right y <- zs ])
-        where
-            zs = concatMap opOrNot ops
+    split ops = partitionEithers $ concatMap opOrNot ops
 
     opOrNot (q, Fixity' fx syn) = Left q
                                 :  case unqualify q of
@@ -270,7 +273,7 @@ instance IsExpr Pattern where
     exprView e = case e of
         IdentP x      -> LocalV x
         AppP e1 e2    -> AppV e1 e2
-        OpAppP r d es -> OpAppV d (map Ordinary es)
+        OpAppP r d es -> OpAppV d ((map . fmap . fmap) Ordinary es)
         HiddenP _ e   -> HiddenArgV e
         InstanceP _ e -> InstanceArgV e
         ParenP _ e    -> ParenV e
@@ -279,8 +282,8 @@ instance IsExpr Pattern where
     unExprView e = case e of
         LocalV x       -> IdentP x
         AppV e1 e2     -> AppP e1 e2
-        OpAppV d es    -> let ess :: [Pattern]
-                              ess = (map (fromOrdinary __IMPOSSIBLE__) es)
+        OpAppV d es    -> let ess :: [NamedArg Pattern]
+                              ess = (map . fmap . fmap) (fromOrdinary __IMPOSSIBLE__) es
                           in OpAppP (fuseRange d es) d ess
         HiddenArgV e   -> HiddenP (getRange e) e
         InstanceArgV e -> InstanceP (getRange e) e
@@ -315,8 +318,8 @@ lhsArgs p = case lhsArgs' p of
 --   into @(f, ps)@.
 lhsArgs' :: Pattern -> Maybe (Name, [NamedArg Pattern])
 lhsArgs' p = case patternAppView p of
-    Arg _ _ (Named _ (IdentP (QName x))) : ps -> Just (x, ps)
-    _                                         -> Nothing
+    Common.Arg _ (Named _ (IdentP (QName x))) : ps -> Just (x, ps)
+    _                                              -> Nothing
 
 -- | View a pattern @p@ as a list @p0 .. pn@ where @p0@ is the identifier
 --   (in most cases a constructor).
@@ -325,12 +328,10 @@ lhsArgs' p = case patternAppView p of
 patternAppView :: Pattern -> [NamedArg Pattern]
 patternAppView p = case p of
     AppP p arg    -> patternAppView p ++ [arg]
-    OpAppP _ x ps -> mkHead (IdentP x) : map notHidden ps
+    OpAppP _ x ps -> defaultNamedArg (IdentP x) : ps
     ParenP _ p    -> patternAppView p
     RawAppP _ _   -> __IMPOSSIBLE__
-    _             -> [ mkHead p ]
-  where mkHead    = Arg NotHidden Relevant . unnamed
-        notHidden = Arg NotHidden Relevant . unnamed
+    _             -> [ defaultNamedArg p ]
 
 
 ---------------------------------------------------------------------------
@@ -340,9 +341,10 @@ patternAppView p = case p of
 -- | Returns the list of possible parses.
 parsePat :: ReadP Pattern Pattern -> Pattern -> [Pattern]
 parsePat prs p = case p of
-    AppP p (Arg h r q) -> fullParen' <$> (AppP <$> parsePat prs p <*> (Arg h r <$> traverse (parsePat prs) q))
+    AppP p (Common.Arg info q) ->
+        fullParen' <$> (AppP <$> parsePat prs p <*> (Common.Arg info <$> traverse (parsePat prs) q))
     RawAppP _ ps     -> fullParen' <$> (parsePat prs =<< parse prs ps)
-    OpAppP r d ps    -> fullParen' . OpAppP r d <$> mapM (parsePat prs) ps
+    OpAppP r d ps    -> fullParen' . OpAppP r d <$> (mapM . traverse . traverse) (parsePat prs) ps
     HiddenP _ _      -> fail "bad hidden argument"
     InstanceP _ _    -> fail "bad instance argument"
     AsP r x p        -> AsP r x <$> parsePat prs p
@@ -399,7 +401,7 @@ parseLHS' lhsOrPatSyn top p = do
     patP <- buildParser (getRange p) flat DontUseBoundNames
     let cons = getNames [ConName, PatternSynName] flat
     let flds = getNames [FldName] flat
-    case [ res | p' <- parsePat patP p
+    case [ res | p' <- force $ parsePat patP p
                , res <- validPattern (PatternCheckConfig top cons flds) p' ] of
         [(p,lhs)] -> return lhs
         []        -> typeError $ NoParseForLHS lhsOrPatSyn p
@@ -428,12 +430,12 @@ classifyPattern conf p =
   case patternAppView p of
 
     -- case @f ps@
-    Arg _ _ (Named _ (IdentP x@(QName f))) : ps | Just f == topName conf -> do
+    Common.Arg _ (Named _ (IdentP x@(QName f))) : ps | Just f == topName conf -> do
       guard $ all validPat ps
       return $ Right (f, LHSHead f ps)
 
     -- case @d ps@
-    Arg _ _ (Named _ (IdentP x)) : ps | x `elem` fldNames conf -> do
+    Common.Arg _ (Named _ (IdentP x)) : ps | x `elem` fldNames conf -> do
       -- ps0 :: [NamedArg ParseLHS]
       ps0 <- mapM classPat ps
       let (ps1, rest) = span (isLeft . namedArg) ps0
@@ -453,8 +455,8 @@ classifyPattern conf p =
         classPat :: NamedArg Pattern -> Maybe (NamedArg ParseLHS)
         classPat = Trav.mapM (Trav.mapM (classifyPattern conf))
         fromR :: NamedArg (Either a (b, c)) -> (b, NamedArg c)
-        fromR (Arg h r (Named n (Right (b, c)))) = (b, Arg h r (Named n c))
-        fromR (Arg h r (Named n (Left  a     ))) = __IMPOSSIBLE__
+        fromR (Common.Arg info (Named n (Right (b, c)))) = (b, Common.Arg info (Named n c))
+        fromR (Common.Arg info (Named n (Left  a     ))) = __IMPOSSIBLE__
 
 
 
@@ -504,8 +506,8 @@ validConPattern cons p = case appView p of
 -- | Helper function for 'parseLHS' and 'parsePattern'.
 appView :: Pattern -> [Pattern]
 appView p = case p of
-    AppP p a         -> appView p ++ [namedThing (unArg a)]
-    OpAppP _ op ps   -> IdentP op : ps
+    AppP p a         -> appView p ++ [namedArg a]
+    OpAppP _ op ps   -> IdentP op : map namedArg ps
     ParenP _ p       -> appView p
     RawAppP _ _      -> __IMPOSSIBLE__
     HiddenP _ _      -> __IMPOSSIBLE__
@@ -542,10 +544,12 @@ parseApplication es = do
     -- Build the parser
     let ms = qualifierModules [ q | Ident q <- es ]
     flat <- flattenScope ms <$> getScope
-    p <- buildParser (getRange es) flat UseBoundNames
+    -- Andreas, 2014-04-27 Time for building the parser is negligible
+    p <- -- billSub [Bench.Parsing, Bench.Operators, Bench.BuildParser] $
+      buildParser (getRange es) flat UseBoundNames
 
     -- Parse
-    case parse p es of
+    case force $ parse p es of
         [e] -> return e
         []  -> do
           -- When the parser fails and a name is not in scope, it is more
@@ -574,49 +578,51 @@ fullParen' e = case exprView e of
     HiddenArgV _ -> e
     InstanceArgV _ -> e
     ParenV _     -> e
-    AppV e1 (Arg h r e2) -> par $ unExprView $ AppV (fullParen' e1) (Arg h r e2')
+    AppV e1 (Common.Arg info e2) -> par $ unExprView $ AppV (fullParen' e1) (Common.Arg info e2')
         where
-            e2' = case h of
+            e2' = case argInfoHiding info of
                 Hidden    -> e2
                 Instance  -> e2
                 NotHidden -> fullParen' <$> e2
-    OpAppV x es -> par $ unExprView $ OpAppV x $ map (fmap fullParen') es
+    OpAppV x es -> par $ unExprView $ OpAppV x $ (map . fmap . fmap . fmap) fullParen' es
     LamV bs e -> par $ unExprView $ LamV bs (fullParen e)
     where
         par = unExprView . ParenV
 
 paren :: Monad m => (QName -> m Fixity) -> Expr -> m (Precedence -> Expr)
-paren _   e@(App _ _ _)        = return $ \p -> mparen (appBrackets p) e
-paren f   e@(OpApp _ op _)     = do fx <- f op; return $ \p -> mparen (opBrackets fx p) e
-paren _   e@(Lam _ _ _)        = return $ \p -> mparen (lamBrackets p) e
-paren _   e@(AbsurdLam _ _)    = return $ \p -> mparen (lamBrackets p) e
-paren _   e@(ExtendedLam _ _)    = return $ \p -> mparen (lamBrackets p) e
-paren _   e@(Fun _ _ _)        = return $ \p -> mparen (lamBrackets p) e
-paren _   e@(Pi _ _)           = return $ \p -> mparen (lamBrackets p) e
-paren _   e@(Let _ _ _)        = return $ \p -> mparen (lamBrackets p) e
-paren _   e@(Rec _ _)          = return $ \p -> mparen (appBrackets p) e
-paren _   e@(RecUpdate _ _ _)  = return $ \p -> mparen (appBrackets p) e
-paren _   e@(WithApp _ _ _)    = return $ \p -> mparen (withAppBrackets p) e
-paren _   e@(Ident _)          = return $ \p -> e
-paren _   e@(Lit _)            = return $ \p -> e
-paren _   e@(QuestionMark _ _) = return $ \p -> e
-paren _   e@(Underscore _ _)   = return $ \p -> e
-paren _   e@(Set _)            = return $ \p -> e
-paren _   e@(SetN _ _)         = return $ \p -> e
-paren _   e@(Prop _)           = return $ \p -> e
-paren _   e@(Paren _ _)        = return $ \p -> e
-paren _   e@(As _ _ _)         = return $ \p -> e
-paren _   e@(Dot _ _)          = return $ \p -> e
-paren _   e@(Absurd _)         = return $ \p -> e
-paren _   e@(ETel _)           = return $ \p -> e
-paren _   e@(RawApp _ _)       = __IMPOSSIBLE__
-paren _   e@(HiddenArg _ _)    = __IMPOSSIBLE__
-paren _   e@(InstanceArg _ _)  = __IMPOSSIBLE__
-paren _   e@(QuoteGoal _ _ _)  = return $ \p -> mparen (lamBrackets p) e
-paren _   e@(Quote _)          = return $ \p -> e
-paren _   e@(QuoteTerm _)      = return $ \p -> e
-paren _   e@(Unquote _)        = return $ \p -> e
-paren _   e@(DontCare _)       = return $ \p -> e
+paren _   e@(App _ _ _)           = return $ \p -> mparen (appBrackets p) e
+paren f   e@(OpApp _ op _)        = do fx <- f op; return $ \p -> mparen (opBrackets fx p) e
+paren _   e@(Lam _ _ _)           = return $ \p -> mparen (lamBrackets p) e
+paren _   e@(AbsurdLam _ _)       = return $ \p -> mparen (lamBrackets p) e
+paren _   e@(ExtendedLam _ _)     = return $ \p -> mparen (lamBrackets p) e
+paren _   e@(Fun _ _ _)           = return $ \p -> mparen (lamBrackets p) e
+paren _   e@(Pi _ _)              = return $ \p -> mparen (lamBrackets p) e
+paren _   e@(Let _ _ _)           = return $ \p -> mparen (lamBrackets p) e
+paren _   e@(Rec _ _)             = return $ \p -> mparen (appBrackets p) e
+paren _   e@(RecUpdate _ _ _)     = return $ \p -> mparen (appBrackets p) e
+paren _   e@(WithApp _ _ _)       = return $ \p -> mparen (withAppBrackets p) e
+paren _   e@(Ident _)             = return $ \p -> e
+paren _   e@(Lit _)               = return $ \p -> e
+paren _   e@(QuestionMark _ _)    = return $ \p -> e
+paren _   e@(Underscore _ _)      = return $ \p -> e
+paren _   e@(Set _)               = return $ \p -> e
+paren _   e@(SetN _ _)            = return $ \p -> e
+paren _   e@(Prop _)              = return $ \p -> e
+paren _   e@(Paren _ _)           = return $ \p -> e
+paren _   e@(As _ _ _)            = return $ \p -> e
+paren _   e@(Dot _ _)             = return $ \p -> e
+paren _   e@(Absurd _)            = return $ \p -> e
+paren _   e@(ETel _)              = return $ \p -> e
+paren _   e@(RawApp _ _)          = __IMPOSSIBLE__
+paren _   e@(HiddenArg _ _)       = __IMPOSSIBLE__
+paren _   e@(InstanceArg _ _)     = __IMPOSSIBLE__
+paren _   e@(QuoteGoal _ _ _)     = return $ \p -> mparen (lamBrackets p) e
+paren _   e@(QuoteContext _ _ _)  = return $ \p -> mparen (lamBrackets p) e
+paren _   e@(Quote _)             = return $ \p -> e
+paren _   e@(QuoteTerm _)         = return $ \p -> e
+paren _   e@(Unquote _)           = return $ \p -> e
+paren _   e@(DontCare _)          = return $ \p -> e
+paren _   e@(Equal _ _ _)         = __IMPOSSIBLE__
 
 mparen :: Bool -> Expr -> Expr
 mparen True  e = Paren (getRange e) e
