@@ -1,13 +1,18 @@
-{-# LANGUAGE CPP, ScopedTypeVariables #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 
 module Agda.Syntax.Concrete.Operators.Parser where
 
+import Control.Exception (throw)
+
+import Data.Maybe
+
 import Agda.Syntax.Position
-import Agda.Syntax.Common
+import Agda.Syntax.Common hiding (Arg, Dom, NamedArg)
 import Agda.Syntax.Fixity
 import Agda.Syntax.Notation
 import Agda.Syntax.Concrete
-import Agda.Syntax.Concrete.Name
+import Agda.TypeChecking.Monad.Base (TCErr(Exception))
 import Agda.Utils.ReadP
 import Agda.Utils.Monad
 
@@ -19,9 +24,9 @@ data ExprView e
     | WildV e
     | OtherV e
     | AppV e (NamedArg e)
-    | OpAppV QName [OpApp e]
-    | HiddenArgV (Named String e)
-    | InstanceArgV (Named String e)
+    | OpAppV QName [NamedArg (OpApp e)]
+    | HiddenArgV (Named_ e)
+    | InstanceArgV (Named_ e)
     | LamV [LamBinding] e
     | ParenV e
 --    deriving (Show)
@@ -30,6 +35,9 @@ class HasRange e => IsExpr e where
     exprView   :: e -> ExprView e
     unExprView :: ExprView e -> e
 
+instance IsExpr e => HasRange (ExprView e) where
+  getRange = getRange . unExprView
+
 ---------------------------------------------------------------------------
 -- * Parser combinators
 ---------------------------------------------------------------------------
@@ -41,29 +49,11 @@ recursive f = p0
 	fs = f p0
 	p0 = foldr ( $ ) p0 fs
 
--- | Variant of chainr1
-chainr1' :: ReadP t a -> ReadP t (a -> a -> ReadP t a) -> ReadP t a
-chainr1' p op = scan
-  where scan   = p >>= rest
-        rest x = do f <- op
-                    y <- scan
-                    f x y
-                 +++ return x
-
--- | Variant of chainl1
-chainl1' :: ReadP t a -> ReadP t (a -> a -> ReadP t a) -> ReadP t a
-chainl1' p op = p >>= rest
-  where rest x = do f <- op
-                    y <- p
-                    fxy <- f x y
-                    rest fxy
-                 +++ return x
-
 ----------------------------
 -- Specific combinators
 
 -- | Parse a specific identifier as a NamePart
-partP :: IsExpr e => [Name] -> String -> ReadP e Range
+partP :: IsExpr e => [Name] -> RawName -> ReadP e Range
 partP ms s = do
     tok <- get
     case isLocal tok of
@@ -75,12 +65,12 @@ partP ms s = do
 	    LocalV y | str == show y -> Just (getRange y)
 	    _			     -> Nothing
 
-binop :: IsExpr e => ReadP e (NewNotation,Range,[e]) -> ReadP e (e -> e -> ReadP a e)
+binop :: IsExpr e => ReadP e (NewNotation,Range,[e]) -> ReadP e (e -> e -> e)
 binop middleP = do
   (nsyn,r,es) <- middleP
   return $ \x y -> rebuild nsyn r (x : es ++ [y])
 
-preop, postop :: IsExpr e => ReadP e (NewNotation,Range,[e]) -> ReadP e (e -> ReadP a e)
+preop, postop :: IsExpr e => ReadP e (NewNotation,Range,[e]) -> ReadP e (e -> e)
 preop middleP = do
   (nsyn,r,es) <- middleP
   return $ \x -> rebuild nsyn r (es ++ [x])
@@ -90,7 +80,6 @@ postop middleP = do
   return $ \x -> rebuild nsyn r (x : es)
 
 
-
 -- | Parse the "operator part" of the given syntax.
 -- holes at beginning and end are IGNORED.
 
@@ -114,33 +103,29 @@ opP p nsyn@(q,_,syn) = do
 
 -- | Given a name with a syntax spec, and a list of parsed expressions
 -- fitting it, rebuild the expression.
--- Note that this function must not parse any input (as guaranteed by the type)
-rebuild :: forall symbol e. IsExpr e => NewNotation -> Range -> [e] -> ReadP symbol e
-rebuild (name,_,syn) r es = do
-  exprs <- mapM findExprFor [0..lastHole]
-  return $ unExprView $ OpAppV (setRange r name) exprs
-  where filledHoles = zip es (filter isAHole syn)
-        lastHole = maximum [t | Just t <- map holeTarget syn]
-        findExprFor :: Int -> ReadP a (OpApp e)
-        findExprFor n = case [e | (e,NormalHole m) <- filledHoles, m == n] of
-                          [] -> fail $ "no expression for hole " ++ show n
-                          [x] -> case [e | (e,BindHole m) <- filledHoles, m == n] of
-                                   [] -> return (Ordinary x) -- no variable to bind
-                                   vars -> do bs <- mapM rebuildBinding $ map exprView vars
-                                              return $ SyntaxBindingLambda (fuseRange bs x) bs x
-                          _ -> fail $ "more than one expression for hole " ++ show n
+rebuild :: forall e. IsExpr e => NewNotation -> Range -> [e] -> e
+rebuild (name,_,syn) r es = unExprView $ OpAppV (setRange r name) exprs
+  where
+    exprs = map findExprFor [0..lastHole]
+    filledHoles = zip es (filter isAHole syn)
+    lastHole = maximum $ mapMaybe holeTarget syn
+    findExprFor :: Int -> NamedArg (OpApp e)
+    findExprFor n =
+      case [setArgColors [] $ fmap (e <$) m | (e, NormalHole m) <- filledHoles, namedArg m == n] of
+        []  -> __IMPOSSIBLE__
+        [x] -> case [e | (e, BindHole m) <- filledHoles, m == n] of
+                 [] -> (fmap . fmap) Ordinary x -- no variable to bind
+                 vars ->
+                  let bs = map (rebuildBinding . exprView) vars in
+                  (fmap . fmap) (SyntaxBindingLambda (fuseRange bs x) bs) x
+        _  -> __IMPOSSIBLE__
 
-rebuildBinding :: ExprView e -> ReadP a LamBinding
+rebuildBinding :: IsExpr e => ExprView e -> LamBinding
   -- Andreas, 2011-04-07 put just 'Relevant' here, is this correct?
-rebuildBinding (LocalV (QName name)) = return $ DomainFree NotHidden Relevant (mkBoundName_ name)
-rebuildBinding (WildV e)     =
-  return $ DomainFree NotHidden Relevant (mkBoundName_ $ Name noRange [Hole])
-rebuildBinding _ = fail "variable name expected"
-
-($$$) :: (e -> ReadP a e) -> ReadP a e -> ReadP a e
-f $$$ x = do
-   x' <- x
-   f x'
+rebuildBinding (LocalV (QName name)) = DomainFree defaultArgInfo $ mkBoundName_ name
+rebuildBinding (WildV e) =
+  DomainFree defaultArgInfo $ mkBoundName_ $ Name noRange [Hole]
+rebuildBinding e = throw $ Exception (getRange e) "Expected variable name in binding position"
 
 -- | Parse using the appropriate fixity, given a parser parsing the
 -- operator part, the name of the operator, and a parser of
@@ -149,27 +134,26 @@ infixP, infixrP, infixlP, postfixP, prefixP,nonfixP :: IsExpr e => ReadP e (NewN
 prefixP op p = do
     fs <- many (preop op)
     e  <- p
-    foldr (($$$)) (return e) fs
+    return $ foldr ( $ ) e fs
 
 postfixP op p = do
     e <- p
     fs <- many (postop op)
-    foldl (flip ( $$$ )) (return e) fs
+    return $ foldl (flip ( $ )) e fs
 
-infixlP op p = chainl1' p (binop op)
-infixrP op p = chainr1' p (binop op)
+infixlP op p = chainl1 p (binop op)
+infixrP op p = chainr1 p (binop op)
 infixP  op p = do
     e <- p
     restP e
     where
 	restP x = return x +++ do
 	    f <- binop op
-	    e <- p
-	    f x e
+	    f x <$> p
 
-nonfixP op p = (do
+nonfixP op p = do
   (nsyn,r,es) <- op
-  rebuild nsyn r es)
+  return $ rebuild nsyn r es
  +++ p
 
 appP :: IsExpr e => ReadP e e -> ReadP e e -> ReadP e e
@@ -187,20 +171,20 @@ appP top p = do
 	isInstance (InstanceArgV _) = True
 	isInstance _	            = False
 
-	nothidden = Arg NotHidden Relevant . unnamed <$> do
+	nothidden = defaultArg . unnamed <$> do
 	    e <- p
 	    case exprView e of
-		HiddenArgV _ -> pfail
+		HiddenArgV   _ -> pfail
 		InstanceArgV _ -> pfail
-		_	     -> return e
+		_	       -> return e
 
 	instanceH = do
 	    InstanceArgV e <- exprView <$> satisfy (isInstance . exprView)
-	    return $ Arg Instance Relevant e
+	    return $ makeInstance $ defaultArg e
 
 	hidden = do
 	    HiddenArgV e <- exprView <$> satisfy (isHidden . exprView)
-	    return $ Arg Hidden Relevant e
+	    return $ hide $ defaultArg e
 
 atomP :: IsExpr e => (QName -> Bool) -> ReadP e e
 atomP p = do
