@@ -1,4 +1,7 @@
-{-# LANGUAGE CPP, FlexibleInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE TypeSynonymInstances #-}
+
 {-# OPTIONS -fno-warn-orphans #-}
 
 {-| Pretty printer for the concrete syntax.
@@ -6,8 +9,11 @@
 module Agda.Syntax.Concrete.Pretty where
 
 import Data.Char
+import Data.Functor
+import Data.Maybe
 
-import Agda.Syntax.Common
+import qualified Agda.Syntax.Common as Common
+import Agda.Syntax.Common hiding (Arg, Dom, NamedArg, ArgInfo)
 import Agda.Syntax.Position
 import Agda.Syntax.Concrete
 import Agda.Syntax.Fixity
@@ -22,6 +28,7 @@ import Agda.Utils.Impossible
 instance Show Expr	      where show = show . pretty
 instance Show Declaration     where show = show . pretty
 instance Show Pattern	      where show = show . pretty
+instance Show TypedBinding    where show = show . pretty
 instance Show TypedBindings   where show = show . pretty
 instance Show LamBinding      where show = show . pretty
 instance Show ImportDirective where show = show . pretty
@@ -37,19 +44,34 @@ braces' d = case render d of
 dbraces :: Doc -> Doc
 dbraces = braces . braces'
 
+-- Lays out a list of documents [d₁, d₂, …] in the following way:
+-- @
+--   { d₁
+--   ; d₂
+--   ⋮
+--   }
+-- @
+-- If the list is empty, then the notation @{}@ is used.
+
+bracesAndSemicolons :: [Doc] -> Doc
+bracesAndSemicolons []       = text "{}"
+bracesAndSemicolons (d : ds) =
+  sep ([text "{" <+> d] ++ map (text ";" <+>) ds ++ [text "}"])
+
 arrow  = text "\x2192"
 lambda = text "\x03bb"
-underscore = text "_"
 
-pHidden :: Pretty a => Hiding -> a -> Doc
-pHidden Hidden	    = braces' . pretty
-pHidden Instance    = dbraces . pretty
-pHidden NotHidden   = pretty
+pHidden :: Pretty a => ArgInfo -> a -> Doc
+pHidden i = bracks h . pretty
+  where bracks Hidden   = braces'
+        bracks Instance = dbraces
+        bracks NotHidden= id
+        h = argInfoHiding i
 
-pRelevance :: Pretty a => Relevance -> a -> Doc
-pRelevance rel a =
+pRelevance :: Pretty a => ArgInfo -> a -> Doc
+pRelevance i a =
   let d = pretty a
-  in  if render d == "_" then d else pretty rel <> d
+  in  if render d == "_" then d else pretty (argInfoRelevance i) <> d
 {-
 pRelevance Forced     a = pretty a
 pRelevance UnusedArg  a = pretty a
@@ -62,6 +84,9 @@ pRelevance NonStrict a =
   in  if render d == "_" then d else text ".." <> d
 -}
 
+instance (Pretty a, Pretty b) => Pretty (a, b) where
+    pretty (a, b) = parens $ pretty a <> comma <+> pretty b
+
 instance Pretty (ThingWithFixity Name) where
     pretty (ThingWithFixity n _) = pretty n
 
@@ -136,7 +161,7 @@ instance Pretty Expr where
             AbsurdLam _ Instance -> lambda <+> text "{{}}"
             AbsurdLam _ Hidden -> lambda <+> text "{}"
 	    ExtendedLam _ pes ->
-		lambda <+> braces' (fsep $ punctuate (text ";") (map (\(x,y,z) -> prettyClause x y z) pes))
+              lambda <+> bracesAndSemicolons (map (\(x,y,z) -> prettyClause x y z) pes)
                    where prettyClause lhs rhs wh = sep [ pretty lhs
                                                        , nest 2 $ pretty' rhs
                                                        ] $$ nest 2 (pretty wh)
@@ -147,7 +172,7 @@ instance Pretty Expr where
 		    , pretty e2
 		    ]
 	    Pi tel e ->
-		sep [ fsep (map pretty (smashTel tel) ++ [arrow])
+		sep [ pretty (Tel $ smashTel tel) <+> arrow
 		    , pretty e
 		    ]
 	    Set _   -> text "Set"
@@ -161,54 +186,88 @@ instance Pretty Expr where
 	    As _ x e  -> pretty x <> text "@" <> pretty e
 	    Dot _ e   -> text "." <> pretty e
 	    Absurd _  -> text "()"
-	    Rec _ xs  -> sep (
-	        [ text "record {" ] ++
-	        punctuate (text ";") (map recPr xs)) <+> text "}"
+	    Rec _ xs  -> sep [text "record", bracesAndSemicolons (map recPr xs)]
 	    RecUpdate _ e xs ->
-	            sep [ text "record" <+> pretty e <+> text "{" ]
-	        <+> sep (punctuate (text ";") (map recPr xs))
-	        <+> text "}"
+              sep [text "record" <+> pretty e, bracesAndSemicolons (map recPr xs)]
             ETel []  -> text "()"
             ETel tel -> fsep $ map pretty tel
             QuoteGoal _ x e -> sep [text "quoteGoal" <+> pretty x <+> text "in",
                                     nest 2 $ pretty e]
+            QuoteContext _ x e -> sep [text "quoteContext" <+> pretty x <+> text "in",
+                                    nest 2 $ pretty e]
             Quote _ -> text "quote"
             QuoteTerm _ -> text "quoteTerm"
 	    Unquote _ -> text "unquote"
             -- Andreas, 2011-10-03 print irrelevant things as .(e)
             DontCare e -> text "." <> parens (pretty e)
+            Equal _ a b -> pretty a <+> text "=" <+> pretty b
 	where
 	  recPr (x, e) = sep [ pretty x <+> text "=" , nest 2 $ pretty e ]
 
 instance Pretty BoundName where
-  pretty = pretty . boundName
+  pretty BName{ boundName = x, boundLabel = l }
+    | x == l    = pretty x
+    | otherwise = pretty l <+> text "=" <+> pretty x
 
 instance Pretty LamBinding where
-    pretty (DomainFree h r x) = pRelevance r $ pHidden h $ pretty x
-    pretty (DomainFull b)     = pretty b
+    -- TODO guilhem: colors are unused (colored syntax disallowed)
+    pretty (DomainFree i x) = pRelevance i $ pHidden i $ pretty x
+    pretty (DomainFull b)   = pretty b
 
 instance Pretty TypedBindings where
-    pretty (TypedBindings _ (Arg h rel b)) =
-	pRelevance rel $ bracks $ pretty b
+    pretty (TypedBindings _ a) =
+	pRelevance (argInfo a) $ bracks $ pretty $ WithColors (argColors a) $ unArg a
 	where
-	    bracks = case h of
-			Hidden	    -> braces'
-			Instance    -> dbraces
-			NotHidden   -> parens
+	    bracks = case getHiding a of
+			Hidden                       -> braces'
+			Instance                     -> dbraces
+			NotHidden | isMeta (unArg a) -> id
+                                  | otherwise        -> parens
+            isMeta (TBind _ _ (Underscore _ Nothing)) = True
+            isMeta _ = False
 
+newtype Tel = Tel Telescope
+
+instance Pretty Tel where
+    pretty (Tel tel)
+      | any isMeta tel = text "∀" <+> fsep (map pretty tel)
+      | otherwise      = fsep (map pretty tel)
+      where
+        isMeta (TypedBindings _ (Common.Arg _ (TBind _ _ (Underscore _ Nothing)))) = True
+        isMeta _ = False
+
+
+instance Pretty ColoredTypedBinding where
+                -- (x y :{ i j } A) -> ...
+    pretty (WithColors [] (TBind _ xs (Underscore _ Nothing))) =
+        fsep (map pretty xs)
+    pretty (WithColors [] (TLet _ ds)) =
+        text "let" <+> vcat (map pretty ds)
+    pretty (WithColors _ (TLet _ _)) = __IMPOSSIBLE__
+    pretty (WithColors cs (TBind _ xs e)) =
+	sep [ fsep (map pretty xs)
+	    , pColors ":" cs <+> pretty e
+	    ]
+
+pColors :: String -> [Color] -> Doc
+pColors s [] = text s
+pColors s cs = text (s ++ "{") <+> fsep (map pretty cs) <+> text "}"
 
 instance Pretty TypedBinding where
-    pretty (TNoBind e) = pretty e
     pretty (TBind _ xs e) =
 	sep [ fsep (map pretty xs)
 	    , text ":" <+> pretty e
 	    ]
+    pretty (TLet _ ds) =
+        text "let" <+> vcat (map pretty ds)
 
 smashTel :: Telescope -> Telescope
-smashTel (TypedBindings r (Arg h  rel  (TBind r' xs e)) :
-          TypedBindings _ (Arg h' rel' (TBind _  ys e')) : tel)
-  | h == h' && rel == rel' && show e == show e' =
-    smashTel (TypedBindings r (Arg h rel (TBind r' (xs ++ ys) e)) : tel)
+smashTel (TypedBindings r (Common.Arg i  (TBind r' xs e)) :
+          TypedBindings _ (Common.Arg i' (TBind _  ys e')) : tel)
+  | show i == show i' && show e == show e' && all isUnnamed (xs ++ ys) =
+    smashTel (TypedBindings r (Common.Arg i (TBind r' (xs ++ ys) e)) : tel)
+  where
+    isUnnamed x = boundLabel x == boundName x
 smashTel (b : tel) = b : smashTel tel
 smashTel [] = []
 
@@ -262,13 +321,14 @@ instance Pretty ModuleApplication where
 instance Pretty Declaration where
     pretty d =
 	case d of
-	    TypeSig rel x e ->
-                sep [ pRelevance rel $ pretty x <+> text ":"
+	    TypeSig i x e ->
+                sep [ pRelevance i $ pretty x <+> pColors ":" (argInfoColors i)
 		    , nest 2 $ pretty e
 		    ]
-            Field x (Arg h rel e) ->
+            Field x (Common.Arg i e) ->
                 sep [ text "field"
-                    , nest 2 $ pRelevance rel $ pHidden h (TypeSig Relevant x e)
+                    , nest 2 $ pRelevance i $ pHidden i $
+                               TypeSig (i {argInfoRelevance = Relevant}) x e
                     ]
 	    FunClause lhs rhs wh ->
 		sep [ pretty lhs
@@ -328,19 +388,8 @@ instance Pretty Declaration where
 			    ]
                     pType Nothing  =
                               text "where"
-                    pInd = maybe [] (\i -> [text $ show i]) ind
-                    pCon = maybe [] (\c -> [text "constructor" <+> pretty c]) con
-{- ELIMINATED CUT-AND-PASTE CODE
-	    Record _ x con tel Nothing cs ->
-		sep [ hsep  [ text "record"
-			    , pretty x
-			    , fcat (map pretty tel)
-			    ]
-		    , nest 2 $ text "where"
-		    ] $$ nest 2 (vcat $ maybe [] (\c -> [text "constructor" <+> pretty c])
-                                              con ++
-                                        map pretty cs)
--}
+                    pInd = maybeToList $ text . show . rangedThing <$> ind
+                    pCon = maybeToList $ (text "constructor" <+>) . pretty <$> con
             Infix f xs	->
 		pretty f <+> (fsep $ punctuate comma $ map pretty xs)
             Syntax n xs -> text "syntax" <+> pretty n <+> text "..."
@@ -392,6 +441,8 @@ instance Pretty Pragma where
     pretty (BuiltinPragma _ b x)  = hsep [ text "BUILTIN", text b, pretty x ]
     pretty (CompiledPragma _ x hs) =
       hsep [ text "COMPILED", pretty x, text hs ]
+    pretty (CompiledExportPragma _ x hs) =
+      hsep [ text "COMPILED_EXPORT", pretty x, text hs ]
     pretty (CompiledTypePragma _ x hs) =
       hsep [ text "COMPILED_TYPE", pretty x, text hs ]
     pretty (CompiledDataPragma _ x hs hcs) =
@@ -418,12 +469,13 @@ instance Pretty Fixity where
 instance Pretty e => Pretty (Arg e) where
  -- Andreas 2010-09-21: do not print relevance in general, only in function types!
  -- Andreas 2010-09-24: and in record fields
-    pretty (Arg h r e) = -- pRelevance r $
-                         pHidden h e
+    pretty a = -- pRelevance r $
+               -- TODO guilhem: print colors
+               pHidden (argInfo a) $ unArg a
 
-instance Pretty e => Pretty (Named String e) where
+instance Pretty e => Pretty (Named_ e) where
     pretty (Named Nothing e) = pretty e
-    pretty (Named (Just s) e) = sep [ text s <+> text "=", pretty e ]
+    pretty (Named (Just s) e) = sep [ text (rawNameToString $ rangedThing s) <+> text "=", pretty e ]
 
 instance Pretty [Pattern] where
     pretty = fsep . map pretty
