@@ -1,14 +1,13 @@
 {-# LANGUAGE DeriveDataTypeable, StandaloneDeriving, FlexibleContexts, UndecidableInstances #-}
 {-| An info object contains additional information about a piece of abstract
     syntax that isn't part of the actual syntax. For instance, it might contain
-    the source code posisiton of an expression or the concrete syntax that
+    the source code position of an expression or the concrete syntax that
     an internal expression originates from.
 -}
 
 module Agda.Syntax.Info where
 
 import Data.Typeable (Typeable)
-import Text.Show.Functions
 
 import qualified Agda.Syntax.Concrete.Name as C
 import Agda.Syntax.Common
@@ -17,22 +16,16 @@ import Agda.Syntax.Concrete
 import Agda.Syntax.Fixity
 import Agda.Syntax.Scope.Base (ScopeInfo, emptyScopeInfo)
 
-{--------------------------------------------------------------------------
-    No information
- --------------------------------------------------------------------------}
-
-data Info = Nope
-
 {--------------------------------------------------------------------------
     Meta information
  --------------------------------------------------------------------------}
 
-data MetaInfo =
-	MetaInfo { metaRange	:: Range
-		 , metaScope	:: ScopeInfo
-		 , metaNumber	:: Maybe Nat
-                 , metaNameSuggestion :: String
-		 }
+data MetaInfo = MetaInfo
+  { metaRange          :: Range
+  , metaScope          :: ScopeInfo
+  , metaNumber         :: Maybe Nat  -- ^ The 'MetaId', not the 'InteractionId'.
+  , metaNameSuggestion :: String
+  }
   deriving (Typeable, Show)
 
 emptyMetaInfo :: MetaInfo
@@ -206,3 +199,26 @@ instance HasRange PatInfo where
 instance KillRange PatInfo where
   killRange (PatRange r)    = PatRange $ killRange r
   killRange (PatSource r f) = PatSource (killRange r) f
+
+-- | Empty range for patterns.
+patNoRange :: PatInfo
+patNoRange = PatRange noRange
+
+-- | Constructor pattern info.
+data ConPatInfo = ConPatInfo
+  { patImplicit :: Bool
+    -- ^ Does this pattern come form the eta-expansion of an implicit pattern.
+  , patInfo     :: PatInfo
+  }
+
+instance Show ConPatInfo where
+  show (ConPatInfo b i) = (if b then ("implicit " ++) else id) $ show i
+
+instance HasRange ConPatInfo where
+  getRange = getRange . patInfo
+
+instance KillRange ConPatInfo where
+  killRange (ConPatInfo b i) = ConPatInfo b $ killRange i
+
+instance SetRange ConPatInfo where
+  setRange r (ConPatInfo b i) = ConPatInfo b $ PatRange r
