@@ -1,4 +1,6 @@
-{-# LANGUAGE CPP #-}
+-- {-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE TypeSynonymInstances #-}
 
 module Agda.Syntax.Internal.Generic where
 
@@ -6,17 +8,47 @@ import Control.Applicative
 import Data.Traversable
 import Data.Monoid
 import Data.Foldable
-import Agda.Syntax.Common
 import Agda.Syntax.Internal
 
-#include "../../undefined.h"
-import Agda.Utils.Impossible
-
 class TermLike a where
   traverseTerm  :: (Term -> Term) -> a -> a
   traverseTermM :: (Monad m, Applicative m) => (Term -> m Term) -> a -> m a
   foldTerm      :: Monoid m => (Term -> m) -> a -> m
 
+-- * Constants
+
+instance TermLike Bool where
+  traverseTerm  _ = id
+  traverseTermM _ = pure
+  foldTerm _      = mempty
+
+instance TermLike Int where
+  traverseTerm  _ = id
+  traverseTermM _ = pure
+  foldTerm _      = mempty
+
+instance TermLike Integer where
+  traverseTerm  _ = id
+  traverseTermM _ = pure
+  foldTerm _      = mempty
+
+instance TermLike Char where
+  traverseTerm  _ = id
+  traverseTermM _ = pure
+  foldTerm _      = mempty
+
+instance TermLike QName where
+  traverseTerm  _ = id
+  traverseTermM _ = pure
+  foldTerm _      = mempty
+
+-- * Functors
+
+instance TermLike a => TermLike (Elim' a) where
+  traverseTerm  f = fmap (traverseTerm f)
+  traverseTermM f = traverse (traverseTermM f)
+  foldTerm f = foldMap (foldTerm f)
+
 instance TermLike a => TermLike (Arg a) where
   traverseTerm  f = fmap (traverseTerm f)
   traverseTermM f = traverse (traverseTermM f)
@@ -42,6 +74,16 @@ instance (TermLike a, TermLike b) => TermLike (a, b) where
   traverseTermM f (x, y) = (,) <$> traverseTermM f x <*> traverseTermM f y
   foldTerm f (x, y) = foldTerm f x `mappend` foldTerm f y
 
+instance (TermLike a, TermLike b, TermLike c) => TermLike (a, b, c) where
+  traverseTerm f (x, y, z) = (traverseTerm f x, traverseTerm f y, traverseTerm f z)
+  traverseTermM f (x, y, z) = (,,) <$> traverseTermM f x <*> traverseTermM f y <*> traverseTermM f z
+  foldTerm f (x, y, z) = mconcat [foldTerm f x, foldTerm f y, foldTerm f z]
+
+instance (TermLike a, TermLike b, TermLike c, TermLike d) => TermLike (a, b, c, d) where
+  traverseTerm f (x, y, z, u) = (traverseTerm f x, traverseTerm f y, traverseTerm f z, traverseTerm f u)
+  traverseTermM f (x, y, z, u) = (,,,) <$> traverseTermM f x <*> traverseTermM f y <*> traverseTermM f z <*> traverseTermM f u
+  foldTerm f (x, y, z, u) = mconcat [foldTerm f x, foldTerm f y, foldTerm f z, foldTerm f u]
+
 instance TermLike a => TermLike (Abs a) where
   traverseTerm f = fmap (traverseTerm f)
   traverseTermM f = traverse (traverseTermM f)
@@ -52,6 +94,8 @@ instance TermLike a => TermLike (Ptr a) where
   traverseTermM f = traverse (traverseTermM f)
   foldTerm f = foldMap (foldTerm f)
 
+-- * Real terms
+
 instance TermLike Term where
   traverseTerm f t = case t of
     Var i xs    -> f $ Var i $ traverseTerm f xs
@@ -132,4 +176,3 @@ instance TermLike Type where
 -- | Put it in a monad to make it possible to do strictly.
 copyTerm :: (TermLike a, Applicative m, Monad m) => a -> m a
 copyTerm = traverseTermM return
-
