@@ -1,4 +1,19 @@
-{-# LANGUAGE CPP, DeriveDataTypeable, DeriveFunctor, TypeSynonymInstances, FlexibleInstances, ScopedTypeVariables #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveTraversable #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE NoMonomorphismRestriction #-}
+
+#if __GLASGOW_HASKELL__ <= 700
+{-# LANGUAGE OverlappingInstances #-}
+#endif
+
+{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE TemplateHaskell #-}
+{-# LANGUAGE TypeSynonymInstances #-}
 
 {-| Position information for syntax. Crucial for giving good error messages.
 -}
@@ -24,6 +39,7 @@ module Agda.Syntax.Position
   , Range
   , Range'(..)
   , rangeInvariant
+  , rightMargin
   , noRange
   , posToRange
   , rStart
@@ -34,7 +50,10 @@ module Agda.Syntax.Position
   , HasRange(..)
   , SetRange(..)
   , KillRange(..)
+  , KillRangeT
   , killRange1, killRange2, killRange3, killRange4, killRange5, killRange6, killRange7
+  , killRange8, killRange9, killRange10, killRange11, killRange12, killRange13, killRange14
+  , killRange15, killRange16, killRange17, killRange18, killRange19
   , withRangeOf
   , fuseRange
   , fuseRanges
@@ -45,17 +64,24 @@ module Agda.Syntax.Position
   , tests
   ) where
 
-import Data.Typeable (Typeable)
-import Data.List
-import Data.Function
-import Data.Set (Set, (\\))
-import qualified Data.Set as Set
-import Data.Int
-import Agda.Utils.QuickCheck
 import Control.Applicative
 import Control.Monad
+
+import Data.Foldable (Foldable)
+import Data.Function
+import Data.Int
+import Data.List
+import Data.Set (Set)
+import qualified Data.Set as Set
+import Data.Traversable (Traversable)
+import Data.Typeable (Typeable)
+
+import Test.QuickCheck.All
+
 import Agda.Utils.FileName hiding (tests)
+import Agda.Utils.Maybe
 import Agda.Utils.TestHelpers
+import Agda.Utils.QuickCheck
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
@@ -77,18 +103,15 @@ data Position' a = Pn
   { srcFile :: a
     -- ^ File.
   , posPos  :: !Int32
-    -- ^ Position.
+    -- ^ Position, counting from 1.
   , posLine :: !Int32
     -- ^ Line number, counting from 1.
   , posCol  :: !Int32
     -- ^ Column number, counting from 1.
   }
-    deriving (Typeable, Functor)
+    deriving (Typeable, Functor, Foldable, Traversable)
 
-type SrcFile  = Maybe AbsolutePath
-type Position = Position' SrcFile
-
-positionInvariant :: Position -> Bool
+positionInvariant :: Position' a -> Bool
 positionInvariant p =
   posPos p > 0 && posLine p > 0 && posCol p > 0
 
@@ -100,22 +123,26 @@ instance Eq a => Eq (Position' a) where
 instance Ord a => Ord (Position' a) where
   compare = compare `on` importantPart
 
+type SrcFile     = Maybe AbsolutePath
+
+type Position = Position' SrcFile
+
 -- | An interval. The @iEnd@ position is not included in the interval.
 --
 -- Note the invariant which intervals have to satisfy: 'intervalInvariant'.
 data Interval' a = Interval { iStart, iEnd :: !(Position' a) }
-    deriving (Typeable, Eq, Ord, Functor)
+    deriving (Typeable, Eq, Ord, Functor, Foldable, Traversable)
 
-type Interval = Interval' SrcFile
+type Interval     = Interval' SrcFile
 
-intervalInvariant :: Interval -> Bool
+intervalInvariant :: Ord a => Interval' a -> Bool
 intervalInvariant i =
   all positionInvariant [iStart i, iEnd i] &&
   iStart i <= iEnd i
 
 -- | The length of an interval, assuming that the start and end
 -- positions are in the same file.
-iLength :: Interval -> Int32
+iLength :: Interval' a -> Int32
 iLength i = posPos (iEnd i) - posPos (iStart i)
 
 -- | A range is a list of intervals. The intervals should be
@@ -123,9 +150,9 @@ iLength i = posPos (iEnd i) - posPos (iStart i)
 --
 -- Note the invariant which ranges have to satisfy: 'rangeInvariant'.
 newtype Range' a = Range [Interval' a]
-  deriving (Typeable, Eq, Ord, Functor)
+  deriving (Typeable, Eq, Ord, Functor, Foldable, Traversable)
 
-type Range = Range' SrcFile
+type Range     = Range' SrcFile
 
 rangeInvariant :: Range -> Bool
 rangeInvariant (Range []) = True
@@ -133,6 +160,12 @@ rangeInvariant (Range is) =
   all intervalInvariant is &&
   and (zipWith (<) (map iEnd $ init is) (map iStart $ tail is))
 
+-- | Conflate a range to its right margin.
+rightMargin :: Range -> Range
+rightMargin r@(Range is) =
+  if null is then r else
+  let i = last is in Range [ i { iStart = iEnd i } ]
+
 -- | Things that have a range are instances of this class.
 class HasRange t where
     getRange :: t -> Range
@@ -179,7 +212,9 @@ instance SetRange Range where
 
 -- | Killing the range of an object sets all range information to 'noRange'.
 class KillRange a where
-  killRange :: a -> a
+  killRange :: KillRangeT a
+
+type KillRangeT a = a -> a
 
 killRange1 f a = f (killRange a)
 killRange2 f a = killRange1 (f $ killRange a)
@@ -188,10 +223,31 @@ killRange4 f a = killRange3 (f $ killRange a)
 killRange5 f a = killRange4 (f $ killRange a)
 killRange6 f a = killRange5 (f $ killRange a)
 killRange7 f a = killRange6 (f $ killRange a)
+killRange8 f a = killRange7 (f $ killRange a)
+killRange9 f a = killRange8 (f $ killRange a)
+killRange10 f a = killRange9 (f $ killRange a)
+killRange11 f a = killRange10 (f $ killRange a)
+killRange12 f a = killRange11 (f $ killRange a)
+killRange13 f a = killRange12 (f $ killRange a)
+killRange14 f a = killRange13 (f $ killRange a)
+killRange15 f a = killRange14 (f $ killRange a)
+killRange16 f a = killRange15 (f $ killRange a)
+killRange17 f a = killRange16 (f $ killRange a)
+killRange18 f a = killRange17 (f $ killRange a)
+killRange19 f a = killRange18 (f $ killRange a)
 
 instance KillRange Range where
   killRange _ = noRange
 
+instance KillRange () where
+  killRange = id
+
+instance KillRange Bool where
+  killRange = id
+
+instance KillRange Int where
+  killRange = id
+
 instance KillRange a => KillRange [a] where
   killRange = map killRange
 
@@ -247,34 +303,34 @@ startPos :: Maybe AbsolutePath -> Position
 startPos f = Pn { srcFile = f, posPos = 1, posLine = 1, posCol = 1 }
 
 -- | Ranges between two unknown positions
-noRange :: Range
+noRange :: Range' a
 noRange = Range []
 
 -- | Advance the position by one character.
 --   A newline character (@'\n'@) moves the position to the first
 --   character in the next line. Any other character moves the
 --   position to the next column.
-movePos :: Position -> Char -> Position
+movePos :: Position' a -> Char -> Position' a
 movePos (Pn f p l c) '\n' = Pn f (p + 1) (l + 1) 1
 movePos (Pn f p l c) _	  = Pn f (p + 1) l (c + 1)
 
 -- | Advance the position by a string.
 --
 --   > movePosByString = foldl' movePos
-movePosByString :: Position -> String -> Position
+movePosByString :: Position' a -> String -> Position' a
 movePosByString = foldl' movePos
 
 -- | Backup the position by one character.
 --
 -- Precondition: The character must not be @'\n'@.
-backupPos :: Position -> Position
+backupPos :: Position' a -> Position' a
 backupPos (Pn f p l c) = Pn f (p - 1) l (c - 1)
 
 -- | Extracts the interval corresponding to the given string, assuming
 -- that the string starts at the beginning of the given interval.
 --
 -- Precondition: The string must not be too long for the interval.
-takeI :: String -> Interval -> Interval
+takeI :: String -> Interval' a -> Interval' a
 takeI s i | genericLength s > iLength i = __IMPOSSIBLE__
           | otherwise = i { iEnd = movePosByString (iStart i) s }
 
@@ -283,12 +339,12 @@ takeI s i | genericLength s > iLength i = __IMPOSSIBLE__
 -- the interval.
 --
 -- Precondition: The string must not be too long for the interval.
-dropI :: String -> Interval -> Interval
+dropI :: String -> Interval' a -> Interval' a
 dropI s i | genericLength s > iLength i = __IMPOSSIBLE__
           | otherwise = i { iStart = movePosByString (iStart i) s }
 
 -- | Converts two positions to a range.
-posToRange :: Position -> Position -> Range
+posToRange :: Ord a => Position' a -> Position' a -> Range' a
 posToRange p1 p2 | p1 < p2   = Range [Interval p1 p2]
                  | otherwise = Range [Interval p2 p1]
 
@@ -300,13 +356,13 @@ rangeToInterval (Range is)   = Just $ Interval { iStart = iStart (head is)
                                                }
 
 -- | Returns the shortest continuous range containing the given one.
-continuous :: Range -> Range
+continuous :: Range' a -> Range' a
 continuous r = case rangeToInterval r of
   Nothing -> Range []
   Just i  -> Range [i]
 
 -- | Removes gaps between intervals on the same line.
-continuousPerLine :: Range -> Range
+continuousPerLine :: Ord a => Range' a -> Range' a
 continuousPerLine (Range [])     = Range []
 continuousPerLine (Range (i:is)) = Range $ fuse i is
   where
@@ -317,22 +373,22 @@ continuousPerLine (Range (i:is)) = Range $ fuse i is
     sameLine i j = posLine (iEnd i) == posLine (iStart j)
 
 -- | The initial position in the range, if any.
-rStart :: Range -> Maybe Position
+rStart :: Range' a -> Maybe (Position' a)
 rStart r = iStart <$> rangeToInterval r
 
 -- | The position after the final position in the range, if any.
-rEnd :: Range -> Maybe Position
+rEnd :: Range' a -> Maybe (Position' a)
 rEnd r = iEnd <$> rangeToInterval r
 
 -- | Finds the least interval which covers the arguments.
-fuseIntervals :: Interval -> Interval -> Interval
+fuseIntervals :: Ord a => Interval' a -> Interval' a -> Interval' a
 fuseIntervals x y = Interval { iStart = head ps, iEnd = last ps }
     where ps = sort [iStart x, iStart y, iEnd x, iEnd y]
 
 -- | @fuseRanges r r'@ unions the ranges @r@ and @r'@.
 --
 --   Meaning it finds the least range @r0@ that covers @r@ and @r'@.
-fuseRanges :: Range -> Range -> Range
+fuseRanges :: (Ord a) => Range' a -> Range' a -> Range' a
 fuseRanges (Range is) (Range js) = Range (helper is js)
   where
   helper []     js  = js
@@ -374,12 +430,12 @@ x `withRangeOf` y = setRange (getRange y) x
 -- end-point. This function assumes that the two end points belong to
 -- the same file. Note that the 'Arbitrary' instance for 'Position's
 -- uses a single, hard-wired file name.
-iPositions :: Interval -> Set Int32
+iPositions :: Interval' a -> Set Int32
 iPositions i = Set.fromList [posPos (iStart i) .. posPos (iEnd i)]
 
 -- | The positions corresponding to the range, including the
 -- end-points. All ranges are assumed to belong to a single file.
-rPositions :: Range -> Set Int32
+rPositions :: Range' a -> Set Int32
 rPositions (Range is) = Set.unions (map iPositions is)
 
 -- | Constructs the least interval containing all the elements in the
@@ -435,8 +491,8 @@ prop_beginningOfFile r = rangeInvariant (beginningOfFile r)
 
 instance Arbitrary a => Arbitrary (Position' a) where
   arbitrary = do
-    srcFile                    <- arbitrary
-    NonZero (NonNegative pos') <- arbitrary
+    srcFile       <- arbitrary
+    Positive pos' <- arbitrary
     let pos  = fromInteger pos'
         line = pred pos `div` 10 + 1
         col  = pred pos `mod` 10 + 1
@@ -445,7 +501,7 @@ instance Arbitrary a => Arbitrary (Position' a) where
 
 -- | Sets the 'srcFile' components of the interval.
 
-setFile :: SrcFile -> Interval -> Interval
+setFile :: a -> Interval' a -> Interval' a
 setFile f (Interval p1 p2) =
   Interval (p1 { srcFile = f }) (p2 { srcFile = f })
 
@@ -465,7 +521,7 @@ instance (Arbitrary a, Ord a) => Arbitrary (Interval' a) where
     let [p1', p2'] = sort [p1, p2 { srcFile = srcFile p1 }]
     return (Interval p1' p2')
 
-instance Arbitrary Range where
+instance (Ord a, Arbitrary a) => Arbitrary (Range' a) where
   arbitrary = Range . fuse . sort . fixFiles <$> arbitrary
     where
     fixFiles []       = []
@@ -476,21 +532,41 @@ instance Arbitrary Range where
       | otherwise            = i1 : fuse (i2 : is)
     fuse is = is
 
+prop_positionInvariant = positionInvariant
+prop_intervalInvariant = intervalInvariant
+prop_rangeInvariant    = rangeInvariant
+
+instance Show (Position' Integer) where show = show . fmap Just
+instance Show (Interval' Integer) where show = show . fmap Just
+instance Show (Range'    Integer) where show = show . fmap Just
+
+------------------------------------------------------------------------
+-- * All tests
+------------------------------------------------------------------------
+
+-- Template Haskell hack to make the following $quickCheckAll work
+-- under ghc-7.8.
+return [] -- KEEP!
+
 -- | Test suite.
 tests :: IO Bool
-tests = runTests "Agda.Syntax.Position"
-  [ quickCheck' positionInvariant
-  , quickCheck' intervalInvariant
-  , quickCheck' rangeInvariant
-  , quickCheck' prop_iLength
-  , quickCheck' prop_startPos
-  , quickCheck' prop_noRange
-  , quickCheck' prop_takeI_dropI
-  , quickCheck' prop_rangeToInterval
-  , quickCheck' prop_continuous
-  , quickCheck' prop_fuseIntervals
-  , quickCheck' prop_fuseRanges
-  , quickCheck' prop_beginningOf
-  , quickCheck' prop_beginningOfFile
-  , quickCheck' prop_intervalInSameFileAs
-  ]
+tests = do
+  putStrLn "Agda.Syntax.Position"
+  $quickCheckAll
+
+-- tests = runTests "Agda.Syntax.Position"
+--   [ quickCheck' (positionInvariant :: Position -> Bool)
+--   , quickCheck' intervalInvariant
+--   , quickCheck' rangeInvariant
+--   , quickCheck' prop_iLength
+--   , quickCheck' prop_startPos
+--   , quickCheck' prop_noRange
+--   , quickCheck' prop_takeI_dropI
+--   , quickCheck' prop_rangeToInterval
+--   , quickCheck' prop_continuous
+--   , quickCheck' prop_fuseIntervals
+--   , quickCheck' prop_fuseRanges
+--   , quickCheck' prop_beginningOf
+--   , quickCheck' prop_beginningOfFile
+--   , quickCheck' prop_intervalInSameFileAs
+--   ]
