@@ -1,16 +1,24 @@
-{-# LANGUAGE CPP, DeriveDataTypeable, GADTs, ScopedTypeVariables, TupleSections, RankNTypes #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE GADTs #-}
+{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE RankNTypes #-}
+{-# LANGUAGE TupleSections #-}
+
 
 {-| This module defines the notion of a scope and operations on scopes.
 -}
 module Agda.Syntax.Scope.Base where
 
-import Control.Arrow ((***), (&&&))
+import Control.Arrow ((***), first, second)
 import Control.Applicative
-import Data.Typeable (Typeable)
+
+import Data.Function
 import Data.List
 import Data.Map (Map)
 import qualified Data.Map as Map
-import Data.Function
+import Data.Maybe
+import Data.Typeable (Typeable)
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common
@@ -19,8 +27,8 @@ import Agda.Syntax.Abstract.Name as A
 import Agda.Syntax.Concrete.Name as C
 import Agda.Syntax.Concrete
   (ImportDirective(..), UsingOrHiding(..), ImportedName(..), Renaming(..))
+
 import qualified Agda.Utils.Map as Map
-import Agda.Utils.Tuple
 import Agda.Utils.List
 
 #include "../../undefined.h"
@@ -51,8 +59,9 @@ nameSpaceAccess :: NameSpaceId -> Access
 nameSpaceAccess PrivateNS = PrivateAccess
 nameSpaceAccess _         = PublicAccess
 
+-- | Get a 'NameSpace' from 'Scope'.
 scopeNameSpace :: NameSpaceId -> Scope -> NameSpace
-scopeNameSpace ns s = maybe __IMPOSSIBLE__ id $ lookup ns $ scopeNameSpaces s
+scopeNameSpace ns = fromMaybe __IMPOSSIBLE__ . lookup ns . scopeNameSpaces
 
 -- | The complete information about the scope at a particular program point
 --   includes the scope stack, the local variables, and the context precedence.
@@ -64,15 +73,30 @@ data ScopeInfo = ScopeInfo
       }
   deriving (Typeable)
 
--- | Local variables
+-- | Local variables.
 type LocalVars = [(C.Name, A.Name)]
 
+-- | Lens for 'scopeLocals'.
+modifyScopeLocals :: (LocalVars -> LocalVars) -> ScopeInfo -> ScopeInfo
+modifyScopeLocals f sc = sc { scopeLocals = f (scopeLocals sc) }
+
+setScopeLocals :: LocalVars -> ScopeInfo -> ScopeInfo
+setScopeLocals vars = modifyScopeLocals (const vars)
+
+------------------------------------------------------------------------
+-- * Name spaces
+--
+-- Map concrete names to lists of abstract names.
+------------------------------------------------------------------------
+
 -- | A @NameSpace@ contains the mappings from concrete names that the user can
 --   write to the abstract fully qualified names that the type checker wants to
 --   read.
 data NameSpace = NameSpace
       { nsNames	  :: NamesInScope
+        -- ^ Maps concrete names to a list of abstract names.
       , nsModules :: ModulesInScope
+        -- ^ Maps concrete module names to a list of abstract module names.
       }
   deriving (Typeable)
 
@@ -80,46 +104,82 @@ type ThingsInScope a = Map C.Name [a]
 type NamesInScope    = ThingsInScope AbstractName
 type ModulesInScope  = ThingsInScope AbstractModule
 
+-- | Set of types consisting of exactly 'AbstractName' and 'AbstractModule'.
+--
+--   A GADT just for some dependent-types trickery.
 data InScopeTag a where
   NameTag   :: InScopeTag AbstractName
   ModuleTag :: InScopeTag AbstractModule
 
+-- | Type class for some dependent-types trickery.
 class Eq a => InScope a where
   inScopeTag :: InScopeTag a
 
-inNameSpace :: forall a. InScope a => NameSpace -> ThingsInScope a
-inNameSpace = case inScopeTag :: InScopeTag a of
-  NameTag   -> nsNames
-  ModuleTag -> nsModules
-
-instance KillRange ScopeInfo where
-  killRange m = m
-
 instance InScope AbstractName where
   inScopeTag = NameTag
 
 instance InScope AbstractModule where
   inScopeTag = ModuleTag
 
--- | We distinguish constructor names from other names.
-data KindOfName = ConName | FldName | DefName | PatternSynName
+-- | @inNameSpace@ selects either the name map or the module name map from
+--   a 'NameSpace'.  What is selected is determined by result type
+--   (using the dependent-type trickery).
+inNameSpace :: forall a. InScope a => NameSpace -> ThingsInScope a
+inNameSpace = case inScopeTag :: InScopeTag a of
+  NameTag   -> nsNames
+  ModuleTag -> nsModules
+
+------------------------------------------------------------------------
+-- * Decorated names
+--
+-- - What kind of name? (defined, constructor...)
+-- - Where does the name come from? (to explain to user)
+------------------------------------------------------------------------
+
+-- | For the sake of parsing left-hand sides, we distinguish
+--   constructor and record field names from defined names.
+data KindOfName
+  = ConName        -- ^ Constructor name.
+  | FldName        -- ^ Record field name.
+  | DefName        -- ^ Ordinary defined name.
+  | PatternSynName -- ^ Name of a pattern synonym.
   deriving (Eq, Show, Typeable, Enum, Bounded)
 
+-- | A list containing all name kinds.
 allKindsOfNames :: [KindOfName]
 allKindsOfNames = [minBound..maxBound]
 
--- | Apart from the name, we also record whether it's a constructor or not and
---   what the fixity is.
-data AbstractName = AbsName
-      { anameName   :: A.QName
-      , anameKind   :: KindOfName
-      }
+-- | Where does a name come from?
+--
+--   This information is solely for reporting to the user,
+--   see 'Agda.Interaction.InteractionTop.whyInScope'.
+data WhyInScope
+  = Defined
+    -- ^ Defined in this module.
+  | Opened C.QName WhyInScope
+    -- ^ Imported from another module.
+  | Applied C.QName WhyInScope
+    -- ^ Imported by a module application.
   deriving (Typeable)
 
--- | For modules we record the arity. I'm not sure that it's every used anywhere.
+-- | A decoration of 'Agda.Syntax.Abstract.Name.QName'.
+data AbstractName = AbsName
+  { anameName    :: A.QName
+    -- ^ The resolved qualified name.
+  , anameKind    :: KindOfName
+    -- ^ The kind (definition, constructor, record field etc.).
+  , anameLineage :: WhyInScope
+    -- ^ Explanation where this name came from.
+  }
+  deriving (Typeable)
+
+-- | A decoration of abstract syntax module names.
 data AbstractModule = AbsModule
-      { amodName    :: A.ModuleName
-      }
+  { amodName    :: A.ModuleName
+    -- ^ The resolved module name.
+  , amodLineage :: WhyInScope
+    -- ^ Explanation where this name came from.
+  }
   deriving (Typeable)
 
 instance Eq AbstractName where
@@ -134,72 +194,14 @@ instance Eq AbstractModule where
 instance Ord AbstractModule where
   compare = compare `on` amodName
 
-instance Show ScopeInfo where
-  show (ScopeInfo this mods locals ctx) =
-    unlines $
-      [ "ScopeInfo"
-      , "  current = " ++ show this
-      ] ++
-      (if null locals then [] else [ "  locals  = " ++ show locals ]) ++
-      [ "  context = " ++ show ctx
-      , "  modules"
-      ] ++ map ("    "++) (relines . map show $ Map.elems mods)
-    where
-      relines = filter (not . null) . lines . unlines
-
-blockOfLines :: String -> [String] -> [String]
-blockOfLines _  [] = []
-blockOfLines hd ss = hd : map ("  "++) ss
-
-instance Show Scope where
-  show (scope @ Scope { scopeName = name, scopeParents = parents, scopeImports = imps }) =
-    unlines $
-      [ "* scope " ++ show name ] ++ ind (
-        concat [ blockOfLines (show nsid) (lines $ show $ scopeNameSpace nsid scope)
-               | nsid <- [minBound..maxBound] ]
-      ++ blockOfLines "imports"  (case Map.keys imps of
-                                    [] -> []
-                                    ks -> [ show ks ]
-                                 )
-      )
-    where ind = map ("  " ++)
-
-instance Show NameSpaceId where
-  show nsid = case nsid of
-    PublicNS        -> "public"
-    PrivateNS       -> "private"
-    ImportedNS      -> "imported"
-    OnlyQualifiedNS -> "only-qualified"
-
-instance Show NameSpace where
-  show (NameSpace names mods) =
-    unlines $
-      blockOfLines "names"   (map pr $ Map.toList names) ++
-      blockOfLines "modules" (map pr $ Map.toList mods)
-    where
-      pr :: (Show a, Show b) => (a,b) -> String
-      pr (x, y) = show x ++ " --> " ++ show y
-
-instance Show AbstractName where
-  show = show . anameName
-
-instance Show AbstractModule where
-  show = show . amodName
-
--- * Operations on names
-
-instance HasRange AbstractName where
-  getRange = getRange . anameName
-
-instance SetRange AbstractName where
-  setRange r x = x { anameName = setRange r $ anameName x }
-
 -- * Operations on name and module maps.
 
 mergeNames :: Eq a => ThingsInScope a -> ThingsInScope a -> ThingsInScope a
 mergeNames = Map.unionWith union
 
+------------------------------------------------------------------------
 -- * Operations on name spaces
+------------------------------------------------------------------------
 
 -- | The empty name space.
 emptyNameSpace :: NameSpace
@@ -234,25 +236,28 @@ mapNameSpaceM fd fm ns = do
   ms <- fm $ nsModules ns
   return $ ns { nsNames = ds, nsModules = ms }
 
+------------------------------------------------------------------------
 -- * General operations on scopes
+------------------------------------------------------------------------
 
 -- | The empty scope.
 emptyScope :: Scope
-emptyScope = Scope { scopeName	         = noModuleName
-                   , scopeParents        = []
-		   , scopeNameSpaces     = [ (nsid, emptyNameSpace) | nsid <- [minBound..maxBound] ]
-                   , scopeImports        = Map.empty
-		   , scopeDatatypeModule = False
-                   }
+emptyScope = Scope
+  { scopeName           = noModuleName
+  , scopeParents        = []
+  , scopeNameSpaces     = [ (nsid, emptyNameSpace) | nsid <- [minBound..maxBound] ]
+  , scopeImports        = Map.empty
+  , scopeDatatypeModule = False
+  }
 
 -- | The empty scope info.
 emptyScopeInfo :: ScopeInfo
 emptyScopeInfo = ScopeInfo
-		  { scopeCurrent    = noModuleName
-                  , scopeModules    = Map.singleton noModuleName emptyScope
-		  , scopeLocals	    = []
-		  , scopePrecedence = TopCtx
-		  }
+  { scopeCurrent    = noModuleName
+  , scopeModules    = Map.singleton noModuleName emptyScope
+  , scopeLocals	    = []
+  , scopePrecedence = TopCtx
+  }
 
 -- | Map functions over the names and modules in a scope.
 mapScope :: (NameSpaceId -> NamesInScope   -> NamesInScope  ) ->
@@ -453,6 +458,13 @@ restrictPrivate s = setNameSpace PrivateNS emptyNameSpace $ s { scopeImports = M
 removeOnlyQualified :: Scope -> Scope
 removeOnlyQualified s = setNameSpace OnlyQualifiedNS emptyNameSpace s
 
+-- | Add an explanation to why things are in scope.
+inScopeBecause :: (WhyInScope -> WhyInScope) -> Scope -> Scope
+inScopeBecause f = mapScope_ mapName mapMod
+  where
+    mapName = fmap . map $ \a -> a { anameLineage = f $ anameLineage a }
+    mapMod  = fmap . map $ \a -> a { amodLineage  = f $ amodLineage a  }
+
 -- | Get the public parts of the public modules of a scope
 publicModules :: ScopeInfo -> Map A.ModuleName Scope
 publicModules scope = Map.filterWithKey (\m _ -> reachable m) allMods
@@ -511,12 +523,10 @@ flattenScope ms scope =
           | (x, mods) <- Map.toList (getNames s)
           , let ms' = [ ms' | m':ms' <- ms, m' == x ]
           , not $ null ms'
-          , AbsModule m <- mods ]
+          , AbsModule m _ <- mods ]
 
     moduleScope :: A.ModuleName -> Scope
-    moduleScope name = case Map.lookup name (scopeModules scope) of
-      Nothing -> __IMPOSSIBLE__
-      Just s  -> s
+    moduleScope m = fromMaybe __IMPOSSIBLE__ $ Map.lookup m $ scopeModules scope
 
 -- | Look up a name in the scope
 scopeLookup :: InScope a => C.QName -> ScopeInfo -> [a]
@@ -525,17 +535,15 @@ scopeLookup q scope = map fst $ scopeLookup' q scope
 scopeLookup' :: forall a. InScope a => C.QName -> ScopeInfo -> [(a, Access)]
 scopeLookup' q scope = nubBy ((==) `on` fst) $ findName q root ++ imports
   where
-    this    :: A.ModuleName
-    this    = scopeCurrent scope
 
     current :: Scope
-    current = moduleScope this
+    current = moduleScope $ scopeCurrent scope
 
     root    :: Scope
     root    = mergeScopes $ current : map moduleScope (scopeParents current)
 
-    tag = inScopeTag :: InScopeTag a
-
+    -- return all possible splittings, e.g.
+    -- splitName X.Y.Z = [(X, Y.Z), (X.Y, Z)]
     splitName :: C.QName -> [(C.QName, C.QName)]
     splitName (C.QName x) = []
     splitName (C.Qual x q) = (C.QName x, q) : do
@@ -546,9 +554,9 @@ scopeLookup' q scope = nubBy ((==) `on` fst) $ findName q root ++ imports
     imported q = maybe [] ((:[]) . (, PublicAccess)) $ Map.lookup q $ scopeImports root
 
     topImports :: [(a, Access)]
-    topImports = case tag of
+    topImports = case (inScopeTag :: InScopeTag a) of
       NameTag   -> []
-      ModuleTag -> map (AbsModule *** id) (imported q)
+      ModuleTag -> map (first (`AbsModule` Defined)) (imported q)
 
     imports :: [(a, Access)]
     imports = topImports ++ do
@@ -557,9 +565,7 @@ scopeLookup' q scope = nubBy ((==) `on` fst) $ findName q root ++ imports
       findName x (restrictPrivate $ moduleScope m)
 
     moduleScope :: A.ModuleName -> Scope
-    moduleScope name = case Map.lookup name (scopeModules scope) of
-      Nothing -> __IMPOSSIBLE__
-      Just s  -> s
+    moduleScope m = fromMaybe __IMPOSSIBLE__ $ Map.lookup m $ scopeModules scope
 
     lookupName :: forall a. InScope a => C.Name -> Scope -> [(a, Access)]
     lookupName x s = maybe [] id $ Map.lookup x (allNamesInScope' s)
@@ -567,12 +573,19 @@ scopeLookup' q scope = nubBy ((==) `on` fst) $ findName q root ++ imports
     findName :: forall a. InScope a => C.QName -> Scope -> [(a, Access)]
     findName (C.QName x)  s = lookupName x s
     findName (C.Qual x q) s = do
-        m <- nub $ mods ++ defs -- record types will appear bot as a mod and a def
+        -- Andreas, 2013-05-01:  Issue 836 complains about the feature
+        -- that constructors can also be qualified by their datatype
+        -- and projections by their record type.  This feature is off
+        -- if we just consider the modules:
+        -- m <- mods
+        -- The feature is on if we consider also the data and record types:
+        m <- nub $ mods ++ defs -- record types will appear both as a mod and a def
         Just s' <- return $ Map.lookup m (scopeModules scope)
         findName q (restrictPrivate s')
       where
         mods, defs :: [ModuleName]
         mods = amodName . fst <$> lookupName x s
+        -- Andreas, 2013-05-01: Issue 836 debates this feature:
         -- Qualified constructors are qualified by their datatype rather than a module
         defs = mnameFromList . qnameToList . anameName . fst <$> lookupName x s
 
@@ -589,9 +602,8 @@ inverseScopeLookup name scope = case name of
     current = this : scopeParents (moduleScope this)
     scopes  = [ (m, restrict m s) | (m, s) <- Map.toList (scopeModules scope) ]
 
-    moduleScope name = case Map.lookup name (scopeModules scope) of
-      Nothing -> __IMPOSSIBLE__
-      Just s  -> s
+    moduleScope :: A.ModuleName -> Scope
+    moduleScope m = fromMaybe __IMPOSSIBLE__ $ Map.lookup m $ scopeModules scope
 
     restrict m s | m `elem` current = s
                  | otherwise = restrictPrivate s
@@ -650,3 +662,73 @@ inverseScopeLookupName x = inverseScopeLookup (Right x)
 -- | Takes the second component of 'inverseScopeLookup'.
 inverseScopeLookupModule :: A.ModuleName -> ScopeInfo -> Maybe C.QName
 inverseScopeLookupModule x = inverseScopeLookup (Left x)
+
+------------------------------------------------------------------------
+-- * (Debug) printing
+------------------------------------------------------------------------
+
+instance Show AbstractName where
+  show = show . anameName
+
+instance Show AbstractModule where
+  show = show . amodName
+
+instance Show NameSpaceId where
+  show nsid = case nsid of
+    PublicNS        -> "public"
+    PrivateNS       -> "private"
+    ImportedNS      -> "imported"
+    OnlyQualifiedNS -> "only-qualified"
+
+instance Show NameSpace where
+  show (NameSpace names mods) =
+    unlines $
+      blockOfLines "names"   (map pr $ Map.toList names) ++
+      blockOfLines "modules" (map pr $ Map.toList mods)
+    where
+      pr :: (Show a, Show b) => (a,b) -> String
+      pr (x, y) = show x ++ " --> " ++ show y
+
+instance Show Scope where
+  show (scope @ Scope { scopeName = name, scopeParents = parents, scopeImports = imps }) =
+    unlines $
+      [ "* scope " ++ show name ] ++ ind (
+        concat [ blockOfLines (show nsid) (lines $ show $ scopeNameSpace nsid scope)
+               | nsid <- [minBound..maxBound] ]
+      ++ blockOfLines "imports"  (case Map.keys imps of
+                                    [] -> []
+                                    ks -> [ show ks ]
+                                 )
+      )
+    where ind = map ("  " ++)
+
+-- | Add first string only if list is non-empty.
+blockOfLines :: String -> [String] -> [String]
+blockOfLines _  [] = []
+blockOfLines hd ss = hd : map ("  " ++) ss
+
+instance Show ScopeInfo where
+  show (ScopeInfo this mods locals ctx) =
+    unlines $
+      [ "ScopeInfo"
+      , "  current = " ++ show this
+      ] ++
+      (if null locals then [] else [ "  locals  = " ++ show locals ]) ++
+      [ "  context = " ++ show ctx
+      , "  modules"
+      ] ++ map ("    "++) (relines . map show $ Map.elems mods)
+    where
+      relines = filter (not . null) . lines . unlines
+
+------------------------------------------------------------------------
+-- * Boring instances
+------------------------------------------------------------------------
+
+instance KillRange ScopeInfo where
+  killRange m = m
+
+instance HasRange AbstractName where
+  getRange = getRange . anameName
+
+instance SetRange AbstractName where
+  setRange r x = x { anameName = setRange r $ anameName x }
