@@ -6,19 +6,21 @@
 module Agda.Syntax.Scope.Monad where
 
 import Prelude hiding (mapM)
+import Control.Arrow ((***), first, second)
 import Control.Applicative
 import Control.Monad hiding (mapM)
 import Control.Monad.Writer hiding (mapM)
 import Control.Monad.State hiding (mapM)
-import Data.Map (Map)
-import Data.Traversable
+
 import Data.List
+import Data.Map (Map)
 import qualified Data.Map as Map
+import Data.Maybe
+import Data.Traversable
 
 import Agda.Syntax.Common
 import Agda.Syntax.Position
 import Agda.Syntax.Fixity
-import Agda.Syntax.Notation
 import Agda.Syntax.Abstract.Name as A
 import Agda.Syntax.Concrete as C
 import Agda.Syntax.Scope.Base
@@ -43,9 +45,6 @@ type ScopeM = TCM
 
 -- * Errors
 
-notInScope :: C.QName -> ScopeM a
-notInScope x = typeError $ NotInScope [x]
-
 isDatatypeModule :: A.ModuleName -> ScopeM Bool
 isDatatypeModule m = do
    sc <- getScope
@@ -158,7 +157,7 @@ getLocalVars :: ScopeM LocalVars
 getLocalVars = scopeLocals <$> getScope
 
 setLocalVars :: LocalVars -> ScopeM ()
-setLocalVars vars = modifyScope $ \s -> s { scopeLocals = vars }
+setLocalVars vars = modifyScope $ setScopeLocals vars
 
 -- | Run a computation without changing the local variables.
 withLocalVars :: ScopeM a -> ScopeM a
@@ -171,10 +170,19 @@ withLocalVars m = do
 -- * Names
 
 -- | Create a fresh abstract name from a concrete name.
+--
+--   This function is used when we translate a concrete name
+--   in a binder.  The 'Range' of the concrete name is
+--   saved as the 'nameBindingSite' of the abstract name.
 freshAbstractName :: Fixity' -> C.Name -> ScopeM A.Name
 freshAbstractName fx x = do
   i <- fresh
-  return $ A.Name i x (getRange x) fx
+  return $ A.Name
+    { nameId          = i
+    , nameConcrete    = x
+    , nameBindingSite = getRange x
+    , nameFixity      = fx
+    }
 
 -- | @freshAbstractName_ = freshAbstractName defaultFixity@
 freshAbstractName_ :: C.Name -> ScopeM A.Name
@@ -199,12 +207,21 @@ data ResolvedName = VarName A.Name
 
 -- | Look up the abstract name referred to by a given concrete name.
 resolveName :: C.QName -> ScopeM ResolvedName
-resolveName x = do
+resolveName = resolveName' allKindsOfNames
+
+-- | Look up the abstract name corresponding to a concrete name of
+--   a certain kind.
+--   Sometimes we know already that we are dealing with a constructor
+--   or pattern synonym (e.g. when we have parsed a pattern).
+--   Then, we can ignore conflicting definitions of that name
+--   of a different kind. (See issue 822.)
+resolveName' :: [KindOfName] -> C.QName -> ScopeM ResolvedName
+resolveName' kinds x = do
   scope <- getScope
   let vars = map (C.QName -*- id) $ scopeLocals scope
   case lookup x vars of
     Just y  -> return $ VarName $ y { nameConcrete = unqualify x }
-    Nothing -> case scopeLookup' x scope of
+    Nothing -> case filter ((`elem` kinds) . anameKind . fst) $ scopeLookup' x scope of
       [] -> return UnknownName
       ds | all ((==ConName) . anameKind . fst) ds ->
         return $ ConstructorName
@@ -224,9 +241,9 @@ resolveModule :: C.QName -> ScopeM AbstractModule
 resolveModule x = do
   ms <- scopeLookup x <$> getScope
   case ms of
-    [AbsModule m] -> return $ AbsModule (m `withRangesOfQ` x)
-    []            -> typeError $ NoSuchModule x
-    ms            -> typeError $ AmbiguousModule x (map amodName ms)
+    [AbsModule m why] -> return $ AbsModule (m `withRangesOfQ` x) why
+    []                -> typeError $ NoSuchModule x
+    ms                -> typeError $ AmbiguousModule x (map amodName ms)
 
 -- | Get the fixity of a name. The name is assumed to be in scope.
 getFixity :: C.QName -> ScopeM Fixity'
@@ -263,10 +280,10 @@ bindName acc kind x y = do
     VarName z          -> typeError $ ClashingDefinition (C.QName x) $ A.qualify (mnameFromList []) z
     ConstructorName [] -> __IMPOSSIBLE__
     ConstructorName ds
-      | kind == ConName && all ((==ConName) . anameKind) ds -> return [ AbsName y kind ]
+      | kind == ConName && all ((==ConName) . anameKind) ds -> return [ AbsName y kind Defined ]
       | otherwise -> typeError $ ClashingDefinition (C.QName x) $ anameName (head' ds)
     PatternSynResName n -> typeError $ ClashingDefinition (C.QName x) $ anameName n
-    UnknownName         -> return [AbsName y kind]
+    UnknownName         -> return [AbsName y kind Defined]
   modifyCurrentScope $ addNamesToScope (localNameSpace acc) x ys
   where
     head' []    = {- ' -} __IMPOSSIBLE__
@@ -275,7 +292,7 @@ bindName acc kind x y = do
 -- | Bind a module name.
 bindModule :: Access -> C.Name -> A.ModuleName -> ScopeM ()
 bindModule acc x m = modifyCurrentScope $
-  addModuleToScope (localNameSpace acc) x (AbsModule m)
+  addModuleToScope (localNameSpace acc) x (AbsModule m Defined)
 
 -- | Bind a qualified module name. Adds it to the imports field of the scope.
 bindQModule :: Access -> C.QName -> A.ModuleName -> ScopeM ()
@@ -299,8 +316,8 @@ type WSM = StateT Out ScopeM
 -- | Create a new scope with the given name from an old scope. Renames
 --   public names in the old scope to match the new name and returns the
 --   renamings.
-copyScope :: A.ModuleName -> Scope -> ScopeM (Scope, (Ren A.ModuleName, Ren A.QName))
-copyScope new s = runStateT (copy new s) (Map.empty, Map.empty)
+copyScope :: C.QName -> A.ModuleName -> Scope -> ScopeM (Scope, (Ren A.ModuleName, Ren A.QName))
+copyScope cm new s = first (inScopeBecause $ Applied cm) <$> runStateT (copy new s) (Map.empty, Map.empty)
   where
     copy new s = do
       s0 <- lift $ getNamedScope new
@@ -313,13 +330,13 @@ copyScope new s = runStateT (copy new s) (Map.empty, Map.empty)
     old  = scopeName s
 
     copyM :: NameSpaceId -> ModulesInScope -> WSM ModulesInScope
-    copyM ImportedNS      ms = return ms
+    copyM ImportedNS      ms = traverse (mapM $ onMod renMod) ms
     copyM PrivateNS       _  = return Map.empty
     copyM PublicNS        ms = traverse (mapM $ onMod renMod) ms
     copyM OnlyQualifiedNS ms = traverse (mapM $ onMod renMod) ms
 
     copyD :: NameSpaceId -> NamesInScope -> WSM NamesInScope
-    copyD ImportedNS      ds = return ds
+    copyD ImportedNS      ds = traverse (mapM $ onName renName) ds
     copyD PrivateNS       _  = return Map.empty
     copyD PublicNS        ds = traverse (mapM $ onName renName) ds
     copyD OnlyQualifiedNS ds = traverse (mapM $ onName renName) ds
@@ -328,9 +345,12 @@ copyScope new s = runStateT (copy new s) (Map.empty, Map.empty)
       x <- f $ amodName m
       return m { amodName = x }
 
-    onName f d = do
-      x <- f $ anameName d
-      return d { anameName = x }
+    onName f d =
+      case anameKind d of
+        PatternSynName -> return d  -- Pattern synonyms are simply aliased, not renamed
+        _ -> do
+          x <- f $ anameName d
+          return d { anameName = x }
 
     addName x y = addNames (Map.singleton x y)
     addMod  x y = addMods (Map.singleton x y)
@@ -341,8 +361,13 @@ copyScope new s = runStateT (copy new s) (Map.empty, Map.empty)
     findName x = Map.lookup x <$> gets snd
     findMod  x = Map.lookup x <$> gets fst
 
+    isInOld qs = isPrefixOf (A.mnameToList old) qs
+
     -- Change a binding M.x -> old.M'.y to M.x -> new.M'.y
+    -- Ulf, 2013-11-06: We should run this also on the imported name space
+    -- (issue892), so make sure to only rename things with the prefix M.
     renName :: A.QName -> WSM A.QName
+    renName x | not (isInOld $ A.qnameToList x) = return x
     renName x = do
       -- Check if we've seen it already
       my <- findName x
@@ -360,6 +385,7 @@ copyScope new s = runStateT (copy new s) (Map.empty, Map.empty)
 
     -- Change a binding M.x -> old.M'.y to M.x -> new.M'.y
     renMod :: A.ModuleName -> WSM A.ModuleName
+    renMod x | not (isInOld $ A.mnameToList x) = return x
     renMod x = do
       -- Check if we've seen it already
       my <- findMod x
@@ -378,17 +404,18 @@ copyScope new s = runStateT (copy new s) (Map.empty, Map.empty)
       where
         dequalify = A.mnameFromList . drop (size old) . A.mnameToList
 
--- | Apply an importdirective and check that all the names mentioned actually
+-- | Apply an import directive and check that all the names mentioned actually
 --   exist.
 applyImportDirectiveM :: C.QName -> ImportDirective -> Scope -> ScopeM Scope
 applyImportDirectiveM m dir scope = do
-  xs <- filterM doesntExist names
-  reportSLn "scope.import.apply" 20 $ "non existing names: " ++ show xs
-  case xs of
-    []  -> case targetNames \\ nub targetNames of
-      []  -> return $ applyImportDirective dir scope
-      dup -> typeError $ DuplicateImports m dup
-    _   -> typeError $ ModuleDoesntExport m xs
+    let xs = filter doesntExist names
+    reportSLn "scope.import.apply" 20 $ "non existing names: " ++ show xs
+    unless (null xs)  $ typeError $ ModuleDoesntExport m xs
+
+    let dup = targetNames \\ nub targetNames
+    unless (null dup) $ typeError $ DuplicateImports m dup
+    return $ applyImportDirective dir scope
+
   where
     names :: [ImportedName]
     names = map renFrom (renaming dir) ++ case usingOrHiding dir of
@@ -402,14 +429,10 @@ applyImportDirectiveM m dir scope = do
       where
         renName r = (renFrom r) { importedName = renTo r }
 
-    doesntExist (ImportedName x) =
-      case Map.lookup x (allNamesInScope scope :: ThingsInScope AbstractName) of
-        Just _  -> return False
-        Nothing -> return True
-    doesntExist (ImportedModule x) =
-      case Map.lookup x (allNamesInScope scope :: ThingsInScope AbstractModule) of
-        Just _  -> return False
-        Nothing -> return True
+    doesntExist (ImportedName   x) = isNothing $
+      Map.lookup x (allNamesInScope scope :: ThingsInScope AbstractName)
+    doesntExist (ImportedModule x) = isNothing $
+      Map.lookup x (allNamesInScope scope :: ThingsInScope AbstractModule)
 
 -- | Open a module.
 openModule_ :: C.QName -> ImportDirective -> ScopeM ()
@@ -418,7 +441,7 @@ openModule_ cm dir = do
   m <- amodName <$> resolveModule cm
   let ns = namespace current m
   s <- setScopeAccess ns <$>
-        (applyImportDirectiveM cm dir . removeOnlyQualified . restrictPrivate =<< getNamedScope m)
+        (applyImportDirectiveM cm dir . inScopeBecause (Opened cm) . removeOnlyQualified . restrictPrivate =<< getNamedScope m)
   checkForClashes (scopeNameSpace ns s)
   modifyCurrentScope (`mergeScope` s)
   where
