@@ -1,6 +1,12 @@
-{-# LANGUAGE CPP, PatternGuards, MultiParamTypeClasses, FunctionalDependencies,
-             TypeSynonymInstances, FlexibleInstances, UndecidableInstances
-  #-}
+-- {-# OPTIONS -fwarn-unused-binds #-}
+
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE FunctionalDependencies #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE UndecidableInstances #-}
 
 {-| The translation of abstract syntax to concrete syntax has two purposes.
     First it allows us to pretty print abstract syntax values without having to
@@ -12,44 +18,45 @@ module Agda.Syntax.Translation.AbstractToConcrete
     ( ToConcrete(..)
     , toConcreteCtx
     , abstractToConcrete_
+    , abstractToConcreteEnv
     , runAbsToCon
     , RangeAndPragma(..)
     , abstractToConcreteCtx
     , withScope
     , makeEnv
-    , abstractToConcrete
     , AbsToCon, DontTouchMe, Env
     , noTakenNames
     ) where
 
-import Control.Applicative
+import Prelude hiding (null)
+
+import Control.Applicative hiding (empty)
 import Control.Monad.Reader
-import Data.Char
+
+import Data.List as List hiding (null)
 import qualified Data.Map as Map
-import Data.Map (Map)
 import qualified Data.Set as Set
 import Data.Set (Set)
-import Data.List as List
-import qualified Data.Traversable as Trav
+import Data.Traversable (traverse)
 
-import Agda.Syntax.Common
+import Agda.Syntax.Common hiding (Arg, Dom, NamedArg)
+import qualified Agda.Syntax.Common as Common
 import Agda.Syntax.Position
 import Agda.Syntax.Info
 import Agda.Syntax.Internal (MetaId(..))
 import Agda.Syntax.Fixity
 import Agda.Syntax.Concrete as C
-import Agda.Syntax.Concrete.Pretty
 import Agda.Syntax.Abstract as A
 import Agda.Syntax.Abstract.Views as AV
 import Agda.Syntax.Scope.Base
 
 import Agda.TypeChecking.Monad.State (getScope)
 import Agda.TypeChecking.Monad.Base  (TCM, NamedMeta(..))
+import Agda.TypeChecking.Monad.Options
 
-import Agda.Utils.Maybe
 import Agda.Utils.Monad hiding (bracket)
+import Agda.Utils.Null
 import Agda.Utils.Tuple
-import Agda.Utils.Suffix
 
 #include "../../undefined.h"
 import Agda.Utils.Impossible
@@ -90,6 +97,27 @@ noTakenNames = local $ \e -> e { takenNames = Set.empty }
 
 -- The Monad --------------------------------------------------------------
 
+-- | We put the translation into TCM in order to print debug messages.
+type AbsToCon = ReaderT Env TCM
+
+runAbsToCon :: AbsToCon c -> TCM c
+runAbsToCon m = do
+  scope <- getScope
+  runReaderT m (makeEnv scope)
+
+abstractToConcreteEnv :: ToConcrete a c => Env -> a -> TCM c
+abstractToConcreteEnv flags a = runReaderT (toConcrete a) flags
+
+abstractToConcreteCtx :: ToConcrete a c => Precedence -> a -> TCM c
+abstractToConcreteCtx ctx x = do
+  scope <- getScope
+  let scope' = scope { scopePrecedence = ctx }
+  abstractToConcreteEnv (makeEnv scope') x
+
+abstractToConcrete_ :: ToConcrete a c => a -> TCM c
+abstractToConcrete_ = runAbsToCon . toConcrete
+
+{-
 -- | We make the translation monadic for modularity purposes.
 type AbsToCon = Reader Env
 
@@ -98,8 +126,13 @@ runAbsToCon m = do
   scope <- getScope
   return $ runReader m (makeEnv scope)
 
+abstractToConcreteEnv :: ToConcrete a c => Env -> a -> TCM c
+abstractToConcreteEnv flags a = return $ runReader (toConcrete a) flags
+
+{- Andreas, 2013-02-26 discontinue non-monadic version in favor of debug msg.
 abstractToConcrete :: ToConcrete a c => Env -> a -> c
 abstractToConcrete flags a = runReader (toConcrete a) flags
+-}
 
 abstractToConcreteCtx :: ToConcrete a c => Precedence -> a -> TCM c
 abstractToConcreteCtx ctx x = do
@@ -113,6 +146,7 @@ abstractToConcrete_ :: ToConcrete a c => a -> TCM c
 abstractToConcrete_ x = do
   scope <- getScope
   return $ abstractToConcrete (makeEnv scope) x
+-}
 
 -- Dealing with names -----------------------------------------------------
 
@@ -132,18 +166,18 @@ lookupName x = do
   case lookup x $ map swap names of
       Just y  -> return y
       Nothing -> return $ nameConcrete x
-  where
-    swap (x, y) = (y, x)
 
 lookupQName :: A.QName -> AbsToCon C.QName
-lookupQName x =
-    do  scope <- asks currentScope
-        case inverseScopeLookupName x scope of
-            Just y  -> return y
-            Nothing
-              | show (qnameToConcrete x) == "_" -> return $ qnameToConcrete x
-              | otherwise -> return $ C.Qual (C.Name noRange [Id ""]) $ qnameToConcrete x
-                -- this is what happens for names that are not in scope (private names)
+lookupQName x = do
+  my <- inverseScopeLookupName x <$> asks currentScope
+  case my of
+    Just y  -> return y
+    Nothing -> do
+      let y = qnameToConcrete x
+      if isUnderscore y
+        then return y
+        else return $ C.Qual (C.Name noRange [Id empty]) y
+        -- this is what happens for names that are not in scope (private names)
 
 lookupModule :: A.ModuleName -> AbsToCon C.QName
 lookupModule x =
@@ -158,9 +192,9 @@ bindName x ret = do
   names <- asks takenNames
   let y = nameConcrete x
   case (Set.member y names) of
-    _ | C.isNoName y -> ret y
-    True             -> bindName (nextName x) ret
-    False            ->
+    _ | isNoName y -> ret y
+    True           -> bindName (nextName x) ret
+    False          ->
         local (\e -> e { takenNames   = Set.insert y $ takenNames e
                        , currentScope = (currentScope e)
                           { scopeLocals = (y, x) : scopeLocals (currentScope e)
@@ -192,6 +226,7 @@ bracketP_ par m =
     do  e <- m
         bracket' (ParenP (getRange e)) par e
 
+{- UNUSED
 -- | Pattern bracketing
 bracketP :: (Precedence -> Bool) -> (C.Pattern -> AbsToCon a)
                                  -> ((C.Pattern -> AbsToCon a) -> AbsToCon a)
@@ -199,6 +234,7 @@ bracketP :: (Precedence -> Bool) -> (C.Pattern -> AbsToCon a)
 bracketP par ret m = m $ \p -> do
     p <- bracket' (ParenP $ getRange p) par p
     ret p
+-}
 
 -- Dealing with infix declarations ----------------------------------------
 
@@ -211,8 +247,10 @@ withInfixDecl i x m = do
  where fixDecl = [C.Infix (theFixity $ defFixity i) [x] | theFixity (defFixity i) /= defaultFixity]
        synDecl = [C.Syntax x (theNotation (defFixity i))]
 
+{- UNUSED
 withInfixDecls :: [(DefInfo, C.Name)] -> AbsToCon [C.Declaration] -> AbsToCon [C.Declaration]
 withInfixDecls = foldr (.) id . map (uncurry withInfixDecl)
+-}
 
 -- Dealing with private definitions ---------------------------------------
 
@@ -269,12 +307,20 @@ instance (ToConcrete a1 c1, ToConcrete a2 c2, ToConcrete a3 c3) =>
         where
             reorder (x,(y,z)) = (x,y,z)
 
-instance ToConcrete a c => ToConcrete (Arg a) (Arg c) where
-    toConcrete (Arg h@Hidden    r x) = Arg h r <$> toConcreteCtx TopCtx x
-    toConcrete (Arg h@Instance  r x) = Arg h r <$> toConcreteCtx TopCtx x
-    toConcrete (Arg h@NotHidden r x) = Arg h r <$> toConcrete x
+instance ToConcrete (Common.ArgInfo ac) C.ArgInfo where
+  toConcrete info = -- do cs <- mapM toConcrete $ argInfoColors info
+                    return $ info { argInfoColors = [] } -- TODO: zapping ignoring colours
 
-    bindToConcrete (Arg h r x) ret = bindToConcreteCtx (hiddenArgumentCtx h) x $ ret . Arg h r
+instance ToConcrete a c => ToConcrete (Common.Arg ac a) (C.Arg c) where
+    toConcrete (Common.Arg info x) = liftM2 Common.Arg (toConcrete info) (f x)
+      where f = case getHiding info of
+                  Hidden    -> toConcreteCtx TopCtx
+                  Instance  -> toConcreteCtx TopCtx
+                  NotHidden -> toConcrete
+
+    bindToConcrete (Common.Arg info x) ret = do info <- toConcrete info
+                                                bindToConcreteCtx (hiddenArgumentCtx $ getHiding info) x $
+                                                                  ret . Common.Arg info
 
 instance ToConcrete a c => ToConcrete (Named name a) (Named name c) where
     toConcrete (Named n x) = Named n <$> toConcrete x
@@ -309,12 +355,15 @@ instance ToConcrete A.Expr C.Expr where
         -- variables)
     toConcrete (A.Lit l)            = return $ C.Lit l
 
-    toConcrete (A.QuestionMark i)   = return $ C.QuestionMark
-                                                (getRange i)
-                                                (metaNumber i)
+    -- Andreas, 2014-05-17  We print question marks with their
+    -- interaction id, in case @metaNumber /= Nothing@
+    toConcrete (A.QuestionMark i ii)= return $
+      C.QuestionMark (getRange i) $
+        interactionId ii <$ metaNumber i
+
     toConcrete (A.Underscore i)     = return $
       C.Underscore (getRange i) $
-       fmap (show . NamedMeta (metaNameSuggestion i) . MetaId) (metaNumber i)
+        show . NamedMeta (metaNameSuggestion i) . MetaId <$> metaNumber i
 
     toConcrete e@(A.App i e1 e2)    =
         tryToRecoverOpApp e
@@ -338,12 +387,12 @@ instance ToConcrete A.Expr C.Expr where
             (bs, e) ->
                 bindToConcrete (map makeDomainFree bs) $ \bs -> do
                     e  <- toConcreteCtx TopCtx e
-                    return $ C.Lam (getRange i) bs e
+                    return $ C.Lam (getRange i) (concat bs) e
         where
-            lamView (A.Lam _ b@(A.DomainFree _ _ _) e) =
+            lamView (A.Lam _ b@(A.DomainFree _ _) e) =
                 case lamView e of
                     ([], e)                        -> ([b], e)
-                    (bs@(A.DomainFree _ _ _ : _), e) -> (b:bs, e)
+                    (bs@(A.DomainFree _ _ : _), e) -> (b:bs, e)
                     _                              -> ([b], e)
             lamView (A.Lam _ b@(A.DomainFull _) e) =
                 case lamView e of
@@ -353,23 +402,36 @@ instance ToConcrete A.Expr C.Expr where
             lamView e = ([], e)
     toConcrete (A.ExtendedLam i di qname cs) =
         bracket lamBrackets $ do
-          decls <- toConcrete cs
+          decls <- concat <$> toConcrete cs
+          let namedPat np = case getHiding np of
+                 NotHidden -> namedArg np
+                 Hidden    -> C.HiddenP noRange (unArg np)
+                 Instance  -> C.InstanceP noRange (unArg np)
               -- we know all lhs are of the form `.extlam p1 p2 ... pn`,
               -- with the name .extlam leftmost. It is our mission to remove it.
-          let removeApp (C.RawAppP r (_:es)) = C.RawAppP r es
-              removeApp (C.AppP (C.IdentP _) np) = namedArg np
-              removeApp (C.AppP p np) = (C.AppP (removeApp p) np)
-              removeApp _ = __IMPOSSIBLE__
-          let decl2clause (C.FunClause lhs rhs wh) = (lhs {lhsOriginalPattern = removeApp $ lhsOriginalPattern lhs},rhs,wh)
+          let removeApp (C.RawAppP r (_:es)) = return $ C.RawAppP r es
+              removeApp (C.AppP (C.IdentP _) np) = return $ namedPat np
+              removeApp (C.AppP p np) = do
+                p <- removeApp p
+                return $ C.AppP p np
+              removeApp p = do
+                lift $ reportSLn "extendedlambda" 50 $ "abstractToConcrete removeApp p = " ++ show p
+                return p -- __IMPOSSIBLE__ -- Andreas, this is actually not impossible, my strictification exposed this sleeping bug
+          let decl2clause (C.FunClause lhs rhs wh) = do
+                let p = lhsOriginalPattern lhs
+                lift $ reportSLn "extendedlambda" 50 $ "abstractToConcrete extended lambda pattern p = " ++ show p
+                p' <- removeApp p
+                lift $ reportSLn "extendedlambda" 50 $ "abstractToConcrete extended lambda pattern p' = " ++ show p'
+                return (lhs{ lhsOriginalPattern = p' }, rhs, wh)
               decl2clause _ = __IMPOSSIBLE__
-          return $ C.ExtendedLam (getRange i) (map decl2clause $ concat decls)
+          C.ExtendedLam (getRange i) <$> mapM decl2clause decls
     toConcrete (A.Pi _ [] e) = toConcrete e
     toConcrete t@(A.Pi i _ _)  = case piTel t of
       (tel, e) ->
         bracket piBrackets
         $ bindToConcrete tel $ \b' -> do
              e' <- toConcreteCtx TopCtx e
-             return $ C.Pi b' e'
+             return $ C.Pi (concat b') e'
       where
         piTel (A.Pi _ tel e) = (tel ++) -*- id $ piTel e
         piTel e              = ([], e)
@@ -380,15 +442,16 @@ instance ToConcrete A.Expr C.Expr where
              b' <- toConcreteCtx TopCtx b
              return $ C.Fun (getRange i) (addRel a' $ mkArg a') b'
         where
-            irr        = argRelevance a `elem` [Irrelevant, NonStrict]
-            addRel a e = case argRelevance a of
+            irr        = getRelevance a `elem` [Irrelevant, NonStrict]
+            addRel a e = case getRelevance a of
                            Irrelevant -> addDot a e
                            NonStrict  -> addDot a (addDot a e)
                            _          -> e
             addDot a e = Dot (getRange a) e
-            mkArg (Arg Hidden    r e) = HiddenArg (getRange e) (unnamed e)
-            mkArg (Arg Instance  r e) = InstanceArg (getRange e) (unnamed e)
-            mkArg (Arg NotHidden r e) = e
+            mkArg (Common.Arg info e) = case getHiding info of
+                                          Hidden    -> HiddenArg   (getRange e) (unnamed e)
+                                          Instance  -> InstanceArg (getRange e) (unnamed e)
+                                          NotHidden -> e
 
     toConcrete (A.Set i 0)  = return $ C.Set (getRange i)
     toConcrete (A.Set i n)  = return $ C.SetN (getRange i) n
@@ -414,7 +477,7 @@ instance ToConcrete A.Expr C.Expr where
         return $ C.RecUpdate (getRange i) e $ zip xs es
 
     toConcrete (A.ETel tel) = do
-      tel <- toConcrete tel
+      tel <- concat <$> toConcrete tel
       return $ C.ETel tel
 
     toConcrete (A.ScopedExpr _ e) = toConcrete e
@@ -424,6 +487,11 @@ instance ToConcrete A.Expr C.Expr where
         bindToConcrete x $ \ x' -> do
             e' <- toConcrete e
             return $ C.QuoteGoal (getRange i) x' e'
+    toConcrete (A.QuoteContext i x e) =
+      bracket lamBrackets $
+        bindToConcrete x $ \ x' -> do
+            e' <- toConcrete e
+            return $ C.QuoteContext (getRange i) x' e'
     toConcrete (A.Quote i) = return $ C.Quote (getRange i)
     toConcrete (A.QuoteTerm i) = return $ C.QuoteTerm (getRange i)
     toConcrete (A.Unquote i) = return $ C.Unquote (getRange i)
@@ -440,43 +508,61 @@ instance ToConcrete A.Expr C.Expr where
     toConcrete (A.PatternSyn n) = C.Ident <$> toConcrete n
 
 makeDomainFree :: A.LamBinding -> A.LamBinding
-makeDomainFree b@(A.DomainFull (A.TypedBindings r (Arg h rel (A.TBind _ [x] t)))) =
+makeDomainFree b@(A.DomainFull (A.TypedBindings r (Common.Arg info (A.TBind _ [x] t)))) =
   case unScope t of
-    A.Underscore MetaInfo{metaNumber = Nothing} -> A.DomainFree h rel x
+    A.Underscore MetaInfo{metaNumber = Nothing} -> A.DomainFree info x
     _ -> b
-  where
-    unScope (A.ScopedExpr _ e) = unScope e
-    unScope e = e
 makeDomainFree b = b
 
 -- Binder instances -------------------------------------------------------
 
-instance ToConcrete A.LamBinding C.LamBinding where
-    bindToConcrete (A.DomainFree h rel x) ret = bindToConcrete x $ ret . C.DomainFree h rel . mkBoundName_
-    bindToConcrete (A.DomainFull b)   ret = bindToConcrete b $ ret . C.DomainFull
+instance ToConcrete A.LamBinding [C.LamBinding] where
+    bindToConcrete (A.DomainFree info x) ret = do info <- toConcrete info
+                                                  bindToConcrete x $ ret . (:[]) . C.DomainFree info . mkBoundName_
+    bindToConcrete (A.DomainFull b)      ret = bindToConcrete b $ ret . map C.DomainFull
 
-instance ToConcrete A.TypedBindings C.TypedBindings where
-    bindToConcrete (A.TypedBindings r bs) ret =
-        bindToConcrete bs $ \bs ->
-        ret (C.TypedBindings r bs)
+instance ToConcrete A.TypedBindings [C.TypedBindings] where
+  bindToConcrete (A.TypedBindings r bs) ret =
+    bindToConcrete bs $ \cbs ->
+    ret (map (C.TypedBindings r) $ recoverLabels bs cbs)
+    where
+      recoverLabels :: A.Arg A.TypedBinding -> C.Arg C.TypedBinding -> [C.Arg C.TypedBinding]
+      recoverLabels b cb
+        | getHiding b == NotHidden = [cb]   -- We don't care about labels for explicit args
+        | otherwise = traverse (recover (unArg b)) cb
+
+      recover (A.TBind _ xs _) (C.TBind r ys e) = tbind r e (zipWith label xs ys)
+      recover A.TLet{}         c@C.TLet{}       = [c]
+      recover _ _ = __IMPOSSIBLE__
+
+      tbinds r e [] = []
+      tbinds r e xs = [ C.TBind r xs e ]
+
+      tbind r e xs =
+        case span (\x -> boundLabel x == boundName x) xs of
+          (xs, x:ys) -> tbinds r e xs ++ [ C.TBind r [x] e ] ++ tbind r e ys
+          (xs, [])   -> tbinds r e xs
+
+      label x y = y { boundLabel = nameConcrete x }
 
 instance ToConcrete A.TypedBinding C.TypedBinding where
     bindToConcrete (A.TBind r xs e) ret =
         bindToConcrete xs $ \xs -> do
         e <- toConcreteCtx TopCtx e
         ret (C.TBind r (map mkBoundName_ xs) e)
-    bindToConcrete (A.TNoBind e) ret = do
-        e <- toConcreteCtx TopCtx e
-        ret (C.TNoBind e)
+    bindToConcrete (A.TLet r lbs) ret =
+        bindToConcrete lbs $ \ds -> do
+          ret (C.TLet r (concat ds))
 
 instance ToConcrete LetBinding [C.Declaration] where
-    bindToConcrete (LetBind i rel x t e) ret =
+    bindToConcrete (LetBind i info x t e) ret =
         bindToConcrete x $ \x ->
-        do  (t,(e, [], [], [])) <- toConcrete (t, A.RHS e)
-            ret [ C.TypeSig rel x t
-                , C.FunClause (C.LHS (C.IdentP $ C.QName x) [] [] [])
-                              e C.NoWhere
-                ]
+        do (t,(e, [], [], [])) <- toConcrete (t, A.RHS e)
+           info <- toConcrete info
+           ret [ C.TypeSig info x t
+               , C.FunClause (C.LHS (C.IdentP $ C.QName x) [] [] [])
+                             e C.NoWhere
+               ]
     -- TODO: bind variables
     bindToConcrete (LetPatBind i p e) ret = do
         p <- toConcrete p
@@ -557,6 +643,7 @@ instance ToConcrete (Maybe A.QName) (Maybe C.Name) where
     x' <- toConcrete (qnameName x)
     return $ Just x'
 
+{- UNUSED
 -- | Helper function used in instance @ToConcrete Definition@.
 telToTypedBindingss :: [C.LamBinding] -> [C.TypedBindings]
 telToTypedBindingss = map lamBindingToTypedBindings where
@@ -565,19 +652,21 @@ telToTypedBindingss = map lamBindingToTypedBindings where
   lamBindingToTypedBindings b =
     case b of
       C.DomainFull t     -> t
-      C.DomainFree h r n -> C.TypedBindings noRange $
-        Arg h r $ C.TBind noRange [n] $ C.Underscore noRange Nothing
+      C.DomainFree info n -> C.TypedBindings noRange $
+        Common.Arg info $ C.TBind noRange [n] $ C.Underscore noRange Nothing
+-}
 
 instance ToConcrete (Constr A.Constructor) C.Declaration where
   toConcrete (Constr (A.ScopedDecl scope [d])) =
     withScope scope $ toConcrete (Constr d)
-  toConcrete (Constr (A.Axiom i rel x t)) = do
+  toConcrete (Constr (A.Axiom _ i info x t)) = do
     x' <- unsafeQNameToName <$> toConcrete x
     t' <- toConcreteCtx TopCtx t
-    return $ C.TypeSig rel x' t'
+    info <- toConcrete info
+    return $ C.TypeSig info x' t'
   toConcrete (Constr d) = head <$> toConcrete d
 
-instance ToConcrete A.Clause [C.Declaration] where
+instance ToConcrete a C.LHS => ToConcrete (A.Clause' a) [C.Declaration] where
   toConcrete (A.Clause lhs rhs wh) =
       bindToConcrete lhs $ \lhs ->
         case lhs of
@@ -592,25 +681,26 @@ instance ToConcrete A.Clause [C.Declaration] where
 
 instance ToConcrete A.ModuleApplication C.ModuleApplication where
   toConcrete (A.SectionApp tel y es) = do
-    y  <- toConcrete y
+    y  <- toConcreteCtx FunctionCtx y
     bindToConcrete tel $ \tel -> do
-    es <- toConcrete es
+    es <- toConcreteCtx ArgumentCtx es
     let r = fuseRange y es
-    return $ C.SectionApp r tel (foldl (C.App r) (C.Ident y) es)
-  toConcrete (A.RecordModuleIFS rec) = do
-    rec <- toConcrete rec
-    return $ C.RecordModuleIFS (getRange rec) rec
+    return $ C.SectionApp r (concat tel) (foldl (C.App r) (C.Ident y) es)
+  toConcrete (A.RecordModuleIFS recm) = do
+    recm <- toConcrete recm
+    return $ C.RecordModuleIFS (getRange recm) recm
 
 instance ToConcrete A.Declaration [C.Declaration] where
   toConcrete (ScopedDecl scope ds) =
     withScope scope (declsToConcrete ds)
 
-  toConcrete (Axiom i rel x t) = do
+  toConcrete (Axiom _ i info x t) = do
     x' <- unsafeQNameToName <$> toConcrete x
     withAbstractPrivate i $
       withInfixDecl i x'  $ do
       t' <- toConcreteCtx TopCtx t
-      return [C.Postulate (getRange i) [C.TypeSig rel x' t']]
+      info <- toConcrete info
+      return [C.Postulate (getRange i) [C.TypeSig info x' t']]
 
   toConcrete (A.Field i x t) = do
     x' <- unsafeQNameToName <$> toConcrete x
@@ -624,7 +714,7 @@ instance ToConcrete A.Declaration [C.Declaration] where
     withAbstractPrivate i $
       withInfixDecl i x'  $ do
       t' <- toConcreteCtx TopCtx t
-      return [C.Primitive (getRange i) [C.TypeSig Relevant x' t']]
+      return [C.Primitive (getRange i) [C.TypeSig defaultArgInfo x' t']]
         -- Primitives are always relevant.
 
   toConcrete (A.FunDef i _ _ cs) =
@@ -635,26 +725,26 @@ instance ToConcrete A.Declaration [C.Declaration] where
     bindToConcrete bs $ \tel' -> do
       x' <- unsafeQNameToName <$> toConcrete x
       t' <- toConcreteCtx TopCtx t
-      return [ C.DataSig (getRange i) Inductive x' (map C.DomainFull tel') t' ]
+      return [ C.DataSig (getRange i) Inductive x' (map C.DomainFull $ concat tel') t' ]
 
   toConcrete (A.DataDef i x bs cs) =
     withAbstractPrivate i $
     bindToConcrete (map makeDomainFree bs) $ \tel' -> do
       (x',cs') <- (unsafeQNameToName -*- id) <$> toConcrete (x, map Constr cs)
-      return [ C.Data (getRange i) Inductive x' tel' Nothing cs' ]
+      return [ C.Data (getRange i) Inductive x' (concat tel') Nothing cs' ]
 
   toConcrete (A.RecSig i x bs t) =
     withAbstractPrivate i $
     bindToConcrete bs $ \tel' -> do
       x' <- unsafeQNameToName <$> toConcrete x
       t' <- toConcreteCtx TopCtx t
-      return [ C.RecordSig (getRange i) x' (map C.DomainFull tel') t' ]
+      return [ C.RecordSig (getRange i) x' (map C.DomainFull $ concat tel') t' ]
 
   toConcrete (A.RecDef  i x ind c bs t cs) =
     withAbstractPrivate i $
     bindToConcrete (map makeDomainFree bs) $ \tel' -> do
       (x',cs') <- (unsafeQNameToName -*- id) <$> toConcrete (x, map Constr cs)
-      return [ C.Record (getRange i) x' ind Nothing tel' Nothing cs' ]
+      return [ C.Record (getRange i) x' ind Nothing (concat tel') Nothing cs' ]
 
   toConcrete (A.Mutual i ds) = declsToConcrete ds
 
@@ -662,7 +752,7 @@ instance ToConcrete A.Declaration [C.Declaration] where
     x <- toConcrete x
     bindToConcrete tel $ \tel -> do
     ds <- declsToConcrete ds
-    return [ C.Module (getRange i) x tel ds ]
+    return [ C.Module (getRange i) x (concat tel) ds ]
 
   toConcrete (A.Apply i x modapp _ _) = do
     x  <- unsafeQNameToName <$> toConcrete x
@@ -686,6 +776,10 @@ instance ToConcrete A.Declaration [C.Declaration] where
     x <- toConcrete x
     return [C.Open (getRange i) x defaultImportDir]
 
+  toConcrete (A.PatternSynDef x xs p) = do
+    C.QName x <- toConcrete x
+    bindToConcrete xs $ \xs -> (:[]) . C.PatternSyn (getRange x) x xs <$> toConcrete p
+
 
 data RangeAndPragma = RangeAndPragma Range A.Pragma
 
@@ -704,6 +798,9 @@ instance ToConcrete RangeAndPragma C.Pragma where
         A.CompiledPragma x hs -> do
           x <- toConcrete x
           return $ C.CompiledPragma r x hs
+        A.CompiledExportPragma x hs -> do
+          x <- toConcrete x
+          return $ C.CompiledExportPragma r x hs
         A.CompiledEpicPragma x e -> do
           x <- toConcrete x
           return $ C.CompiledEpicPragma r x e
@@ -723,6 +820,9 @@ noImplicitPats = filter noImplicit
 noImplicit (A.ImplicitP _) = False
 noImplicit _               = True
 
+instance ToConcrete A.SpineLHS C.LHS where
+  bindToConcrete lhs = bindToConcrete (A.spineToLhs lhs :: A.LHS)
+
 instance ToConcrete A.LHS C.LHS where
     bindToConcrete (A.LHS i lhscore wps) ret = do
       bindToConcreteCtx TopCtx lhscore $ \lhs ->
@@ -782,7 +882,7 @@ instance ToConcrete A.Pattern C.Pattern where
         e <- toConcreteCtx DotPatternCtx e
         return $ C.DotP (getRange i) e
     -- just for debugging purposes (shouldn't show up in practise)
-    toConcrete (A.ImplicitP i) = return $ C.IdentP (C.QName $ C.Name noRange [C.Id "(implicit)"])
+    toConcrete (A.ImplicitP i) = return $ C.IdentP (C.QName $ C.Name noRange [C.Id $ stringToRawName "(implicit)"])
     toConcrete (A.PatternSynP i n _) = IdentP <$> toConcrete n
 
 
@@ -791,22 +891,24 @@ instance ToConcrete A.Pattern C.Pattern where
 data Hd = HdVar A.Name | HdCon A.QName | HdDef A.QName
 
 cOpApp :: Range -> C.QName -> [C.Expr] -> C.Expr
-cOpApp r n es = C.OpApp r n (map Ordinary es)
+cOpApp r n es = C.OpApp r n (map (defaultNamedArg . Ordinary) es)
 
 tryToRecoverOpApp :: A.Expr -> AbsToCon C.Expr -> AbsToCon C.Expr
 tryToRecoverOpApp e def = recoverOpApp bracket cOpApp view e def
   where
-    view e = case AV.appView e of
-      --NonApplication _   -> Nothing
-      Application (Var x) args -> Just (HdVar x, args)
-      Application (Def f) args -> Just (HdDef f, args)
-      Application (Con (AmbQ (c:_))) args -> Just (HdCon c, args)
-      Application (Con (AmbQ [])) args -> __IMPOSSIBLE__
-      _ -> Nothing
+    view e = do
+      let Application hd args = AV.appView e
+      case hd of
+        Var x            -> Just (HdVar x, args)
+        Def f            -> Just (HdDef f, args)
+        Con (AmbQ (c:_)) -> Just (HdCon c, args)
+        Con (AmbQ [])    -> __IMPOSSIBLE__
+        _                -> Nothing
 
 tryToRecoverOpAppP :: A.Pattern -> AbsToCon C.Pattern -> AbsToCon C.Pattern
-tryToRecoverOpAppP p def = recoverOpApp bracketP_ C.OpAppP view p def
+tryToRecoverOpAppP p def = recoverOpApp bracketP_ opApp view p def
   where
+    opApp r x ps = C.OpAppP r x (map defaultNamedArg ps)
     view p = case p of
       ConP _ (AmbQ (c:_)) ps -> Just (HdCon c, ps)
       DefP _ f            ps -> Just (HdDef f, ps)
@@ -815,8 +917,8 @@ tryToRecoverOpAppP p def = recoverOpApp bracketP_ C.OpAppP view p def
 recoverOpApp :: (ToConcrete a c, HasRange c)
   => ((Precedence -> Bool) -> AbsToCon c -> AbsToCon c)
   -> (Range -> C.QName -> [c] -> c)
-  -> (a -> Maybe (Hd, [NamedArg a]))
-  ->  a
+  -> (a -> Maybe (Hd, [A.NamedArg a]))
+  -> a
   -> AbsToCon c
   -> AbsToCon c
 recoverOpApp bracket opApp view e mDefault = case view e of
@@ -833,12 +935,6 @@ recoverOpApp bracket opApp view e mDefault = case view e of
     | otherwise           -> mDefault
   where
 
-  isNoName :: A.Name -> Bool
-  isNoName  x = C.isNoName $ A.nameConcrete x
-
-  notHidden :: NamedArg a -> Bool
-  notHidden a = argHiding a == NotHidden
-
   doQNameHelper fixityHelper conHelper n as = do
     x <- toConcrete n
     doQName (theFixity $ nameFixity $ fixityHelper n) (conHelper x) as
@@ -847,15 +943,17 @@ recoverOpApp bracket opApp view e mDefault = case view e of
   doQName _ n es
     | length xs == 1        = mDefault
     | length es /= numHoles = mDefault
-    | List.null es          = mDefault
+    | null es               = mDefault
     where
       xs       = C.nameParts $ C.unqualify n
       numHoles = length (filter (== Hole) xs)
+{- UNUSED
       msg      = concat [ "doQName "
                         , showList xs ""
                         , " on "
                         , show (length es)
                         , " args" ]
+-}
 
   -- binary case
   doQName fixity n as
