@@ -1,4 +1,10 @@
-{-# LANGUAGE TypeOperators, CPP, DeriveDataTypeable, DeriveFunctor, DeriveFoldable, DeriveTraversable #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveTraversable #-}
+{-# LANGUAGE TypeOperators #-}
+
 module Agda.TypeChecking.CompiledClause where
 
 import qualified Data.Map as Map
@@ -8,31 +14,40 @@ import Data.Typeable (Typeable)
 import Data.Foldable (Foldable)
 import Data.Traversable (Traversable)
 
-import Agda.Syntax.Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Literal
 
 import Agda.Utils.Pretty
-import Agda.Utils.Impossible
+
 #include "../undefined.h"
+import Agda.Utils.Impossible
 
 type key :-> value = Map key value
 
 data WithArity c = WithArity { arity :: Int, content :: c }
   deriving (Typeable, Functor, Foldable, Traversable)
 
+-- | Branches in a case tree.
 data Case c = Branches
-  { conBranches    :: QName :-> WithArity c -- ^ Map from constructor names to their arity and the case subtree
-  , litBranches    :: Literal :-> c         -- ^ Map from literal to case subtree
-  , catchAllBranch :: Maybe c               -- ^ (Possibly additional) catch-all clause
+  { conBranches    :: QName :-> WithArity c -- ^ Map from constructor (or projection) names to their arity and the case subtree.  (Projections have arity 0.)
+  , litBranches    :: Literal :-> c         -- ^ Map from literal to case subtree.
+  , catchAllBranch :: Maybe c               -- ^ (Possibly additional) catch-all clause.
   }
   deriving (Typeable, Functor, Foldable, Traversable)
 
+-- | Case tree with bodies.
 data CompiledClauses
   = Case Int (Case CompiledClauses)
     -- ^ @Case n bs@ stands for a match on the @n@-th argument
     -- (counting from zero) with @bs@ as the case branches.
-  | Done [Arg String] Term
+    -- If the @n@-th argument is a projection, we have only 'conBranches'.
+    -- with arity 0.
+{-
+  | CoCase Int (QName :-> CompiledClauses)
+    -- ^ @CoCase n bs@ matches on projections.
+    --   Catch-all is not meaningful here.
+-}
+  | Done [Arg ArgName] Term
     -- ^ @Done xs b@ stands for the body @b@ where the @xs@ contains hiding
     --   and name suggestions for the free variables. This is needed to build
     --   lambdas on the right hand side for partial applications which can
@@ -60,8 +75,11 @@ instance Monoid m => Monoid (Case m) where
              (Map.unionWith mappend ls ls')
              (mappend m m')
 
+-- * Pretty instances.
+
 instance Pretty a => Show (Case a) where
   show = show . pretty
+
 instance Show CompiledClauses where
   show = show . pretty
 
@@ -71,13 +89,15 @@ instance Pretty a => Pretty (WithArity a) where
 instance Pretty a => Pretty (Case a) where
   prettyPrec p (Branches cs ls m) =
     mparens (p > 0) $ vcat $
-      pr cs ++ pr ls ++ prC m
+      prettyMap cs ++ prettyMap ls ++ prC m
     where
       prC Nothing = []
       prC (Just x) = [text "_ ->" <+> pretty x]
-      pr m = [ sep [ text (show x ++ " ->")
-                   , nest 2 $ pretty v ]
-             | (x, v) <- Map.toList m ]
+
+prettyMap :: (Show k, Pretty v) => (k :-> v) -> [Doc]
+prettyMap m = [ sep [ text (show x ++ " ->")
+                    , nest 2 $ pretty v ]
+              | (x, v) <- Map.toList m ]
 
 instance Pretty CompiledClauses where
   pretty (Done hs t) = text ("done" ++ show hs) <+> text (show t)
@@ -86,3 +106,9 @@ instance Pretty CompiledClauses where
     sep [ text ("case " ++ show n ++ " of")
         , nest 2 $ pretty bs
         ]
+{-
+  pretty (CoCase n bs) =
+    sep [ text ("cocase " ++ show n ++ " of")
+        , nest 2 $ vcat $ prettyMap bs
+        ]
+-}
