@@ -1,17 +1,24 @@
-{-# LANGUAGE CPP, PatternGuards, FlexibleContexts, TupleSections #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE TupleSections #-}
 
 module Agda.TypeChecking.Coverage where
 
 import Control.Monad
 import Control.Monad.Error
-import Control.Applicative
+import Control.Applicative hiding (empty)
+
 import Data.List
 import qualified Data.Set as Set
 import Data.Set (Set)
+import qualified Data.Traversable as Trav
 
 import Agda.Syntax.Position
-import Agda.Syntax.Common
-import Agda.Syntax.Internal
+import qualified Agda.Syntax.Common as Common
+import Agda.Syntax.Common hiding (Arg,Dom)
+import qualified Agda.Syntax.Common as C
+import Agda.Syntax.Internal as I
 import Agda.Syntax.Internal.Pattern
 
 import Agda.TypeChecking.Monad.Base
@@ -22,6 +29,7 @@ import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Monad.Exception
 import Agda.TypeChecking.Monad.Context
 
+import Agda.TypeChecking.Rules.LHS.Problem (FlexibleVar(..),flexibleVarFromHiding)
 import Agda.TypeChecking.Rules.LHS.Unify
 import Agda.TypeChecking.Rules.LHS.Instantiate
 import Agda.TypeChecking.Rules.LHS
@@ -30,31 +38,33 @@ import qualified Agda.TypeChecking.Rules.LHS.Split as Split
 import Agda.TypeChecking.Coverage.Match
 import Agda.TypeChecking.Coverage.SplitTree
 
+import Agda.TypeChecking.Datatypes (getConForm)
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Reduce
-import Agda.TypeChecking.Primitive (constructorForm)
+import Agda.TypeChecking.Records (isRecordType)
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Irrelevance
 
 import Agda.Interaction.Options
 
-import Agda.Utils.Permutation
-import Agda.Utils.Size
-import Agda.Utils.Tuple
+import Agda.Utils.Functor (for, ($>))
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.Permutation
+import Agda.Utils.Size
+import Agda.Utils.Tuple
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
 
 data SplitClause = SClause
-      { scTel    :: Telescope      -- ^ Type of variables in @scPats@.
-      , scPerm   :: Permutation    -- ^ How to get from the variables in the patterns to the telescope.
-      , scPats   :: [Arg Pattern]  -- ^ The patterns leading to the currently considered branch of the split tree.
-      , scSubst  :: Substitution   -- ^ Substitution from @scTel@ to old context.
-      , scTarget :: Maybe Type     -- ^ The type of the rhs.
+      { scTel    :: Telescope            -- ^ Type of variables in @scPats@.
+      , scPerm   :: Permutation          -- ^ How to get from the variables in the patterns to the telescope.
+      , scPats   :: [I.NamedArg Pattern] -- ^ The patterns leading to the currently considered branch of the split tree.
+      , scSubst  :: Substitution         -- ^ Substitution from @scTel@ to old context.
+      , scTarget :: Maybe (I.Arg Type)   -- ^ The type of the rhs.
       }
 
 -- | A @Covering@ is the result of splitting a 'SplitClause'.
@@ -73,49 +83,11 @@ clauseToSplitClause :: Clause -> SplitClause
 clauseToSplitClause cl = SClause
   { scTel    = clauseTel  cl
   , scPerm   = clausePerm cl
-  , scPats   = clausePats cl
-  , scSubst  = __IMPOSSIBLE__
-  , scTarget = Nothing
+  , scPats   = namedClausePats cl
+  , scSubst  = idS  -- Andreas, 2014-07-15  TODO: Is this ok?
+  , scTarget = clauseType cl
   }
 
-data SplitError = NotADatatype (Closure Type) -- ^ neither data type nor record
-                | IrrelevantDatatype (Closure Type)   -- ^ data type, but in irrelevant position
-                | CoinductiveDatatype (Closure Type)  -- ^ coinductive data type
-{- UNUSED
-                | NoRecordConstructor Type  -- ^ record type, but no constructor
- -}
-                | CantSplit QName Telescope Args Args [Term]
-                | GenericSplitError String
-  deriving (Show)
-
-instance PrettyTCM SplitError where
-  prettyTCM err = case err of
-    NotADatatype t -> enterClosure t $ \ t -> fsep $
-      pwords "Cannot pattern match on non-datatype" ++ [prettyTCM t]
-    IrrelevantDatatype t -> enterClosure t $ \ t -> fsep $
-      pwords "Cannot pattern match on datatype" ++ [prettyTCM t] ++
-      pwords "since it is declared irrelevant"
-    CoinductiveDatatype t -> enterClosure t $ \ t -> fsep $
-      pwords "Cannot pattern match on the coinductive type" ++ [prettyTCM t]
-{- UNUSED
-    NoRecordConstructor t -> fsep $
-      pwords "Cannot pattern match on record" ++ [prettyTCM t] ++
-      pwords "because it has no constructor"
- -}
-    CantSplit c tel cIxs gIxs flex -> addCtxTel tel $ vcat
-      [ fsep $ pwords "Cannot decide whether there should be a case for the constructor" ++ [prettyTCM c <> text ","] ++
-               pwords "since the unification gets stuck on unifying the inferred indices"
-      , nest 2 $ prettyTCM cIxs
-      , fsep $ pwords "with the expected indices"
-      , nest 2 $ prettyTCM gIxs
-      ]
-    GenericSplitError s -> fsep $
-      pwords "Split failed:" ++ pwords s
-
-instance Error SplitError where
-  noMsg  = strMsg ""
-  strMsg = GenericSplitError
-
 type CoverM = ExceptionT SplitError TCM
 
 {- UNUSED
@@ -150,28 +122,16 @@ coverageCheck f t cs = do
       -- xs            = variable patterns fitting lgamma
       n            = size gamma
       lgamma       = telToList gamma
-      xs           = map (argFromDom . fmap (const $ VarP "_")) $ lgamma
+      xs           = map (argFromDom . fmap (namedVarP . fst)) $ lgamma
       -- construct the initial split clause
-      sc           = SClause gamma (idP n) xs idS $ Just a
-{- OLD
-  let -- n             = arity
-      -- lgamma/gamma' = telescope of non-dropped arguments
-      -- xs            = variable patterns fitting lgamma
-      n            = genericLength $ clausePats $ head cs
-      (lgamma, rgamma) = genericSplitAt n $ telToList gamma
-      gamma'       = telFromList lgamma
-      xs           = map (argFromDom . fmap (const $ VarP "_")) $ lgamma
-      -- construct the initial split clause
-      b            = telePi (telFromList rgamma) a
-      sc           = SClause gamma' (idP n) xs idS $ Just b
--}
+      sc           = SClause gamma (idP n) xs idS $ Just $ defaultArg a
   reportSDoc "tc.cover.top" 10 $ vcat
-    [ text "Coverage checking"
+    [ text $ "Coverage checking " ++ show f
     , nest 2 $ vcat $ map (text . show . clausePats) cs
     ]
   -- used = actually used clauses for cover
   -- pss  = uncovered cases
-  (splitTree, used, pss) <- cover cs sc
+  (splitTree, used, pss) <- cover f cs sc
   reportSDoc "tc.cover.splittree" 10 $ vcat
     [ text "generated split tree for" <+> prettyTCM f
     , text $ show splitTree
@@ -180,7 +140,7 @@ coverageCheck f t cs = do
     -- report an error if there are uncovered cases
     unless (null pss) $
         setCurrentRange (getRange cs) $
-          typeError $ CoverageFailure f pss
+          typeError $ CoverageFailure f (map (map $ fmap namedThing) pss)
   -- is = indices of unreachable clauses
   let is = Set.toList $ Set.difference (Set.fromList [0..genericLength cs - 1]) used
   -- report an error if there are unreachable clauses
@@ -191,26 +151,78 @@ coverageCheck f t cs = do
   return splitTree
 
 
--- | @cover cs (SClause _ _ ps _) = return (splitTree, used, pss)@.
+-- | @cover f cs (SClause _ _ ps _) = return (splitTree, used, pss)@.
 --   checks that the list of clauses @cs@ covers the given split clause.
 --   Returns the @splitTree@, the @used@ clauses, and missing cases @pss@.
-cover :: [Clause] -> SplitClause -> TCM (SplitTree, Set Nat, [[Arg Pattern]])
-cover cs sc@(SClause tel perm ps _ target) = do
+cover :: QName -> [Clause] -> SplitClause -> TCM (SplitTree, Set Nat, [[I.NamedArg Pattern]])
+cover f cs sc@(SClause tel perm ps _ target) = do
   reportSDoc "tc.cover.cover" 10 $ vcat
     [ text "checking coverage of pattern:"
     , nest 2 $ text "tel  =" <+> prettyTCM tel
     , nest 2 $ text "perm =" <+> text (show perm)
     , nest 2 $ text "ps   =" <+> text (show ps)
     ]
-  case match cs ps perm of
+  let ups = map (fmap namedThing) ps
+  case match cs ups perm of
     Yes i          -> do
       reportSLn "tc.cover.cover" 10 $ "pattern covered by clause " ++ show i
       -- Check if any earlier clauses could match with appropriate literals
-      let is = [ j | (j, c) <- zip [0..i-1] cs, matchLits c ps perm ]
+      let is = [ j | (j, c) <- zip [0..i-1] cs, matchLits c ups perm ]
       -- OLD: let is = [ j | (j, c) <- zip [0..] (genericTake i cs), matchLits c ps perm ]
       reportSLn "tc.cover.cover"  10 $ "literal matches: " ++ show is
       return (SplittingDone (size tel), Set.fromList (i : is), [])
-    No       -> return (SplittingDone (size tel), Set.empty, [ps])
+    No       ->  do
+      reportSLn "tc.cover" 20 $ "pattern is not covered"
+      return (SplittingDone (size tel), Set.empty, [ps])
+
+    -- case: split into projection patterns
+    BlockP   -> do
+      reportSLn "tc.cover" 20 $ "blocked by projection pattern"
+      -- if we want to split projections, but have no target type, we give up
+      let done = return (SplittingDone (size tel), Set.empty, [ps])
+      caseMaybeM (splitResult f sc) done $ \ (Covering n scs) -> do
+        (projs, (trees, useds, psss)) <- mapSnd unzip3 . unzip <$> do
+          forM scs $ \ (proj, sc') -> (proj,) <$> do cover f cs =<< fixTarget sc'
+          -- OR: mapM (traverseF $ cover f cs <=< fixTarget) scs
+        let tree = SplitAt n $ zip projs trees
+        return (tree, Set.unions useds, concat psss)
+
+      -- caseMaybe target done $ \ t -> do
+      --   isR <- addCtxTel tel $ isRecordType $ unArg t
+      --   case isR of
+      --     Just (_r, vs, Record{ recFields = fs }) -> do
+      --       reportSDoc "tc.cover" 20 $ sep
+      --         [ text $ "we are of record type _r = " ++ show _r
+      --         , text   "applied to parameters vs = " <+> (addCtxTel tel $ prettyTCM vs)
+      --         , text $ "and have fields       fs = " ++ show fs
+      --         ]
+      --       fvs <- freeVarsToApply f
+      --       let es = patternsToElims perm ps
+      --       let self  = defaultArg $ Def f (map Apply fvs) `applyE` es
+      --           pargs = vs ++ [self]
+      --       reportSDoc "tc.cover" 20 $ sep
+      --         [ text   "we are              self = " <+> (addCtxTel tel $ prettyTCM $ unArg self)
+      --         ]
+      --       (projs, (trees, useds, psss)) <- mapSnd unzip3 . unzip <$> do
+      --         forM fs $ \ proj -> do
+      --           -- compute the new target
+      --           dType <- defType <$> do getConstInfo $ unArg proj -- WRONG: typeOfConst $ unArg proj
+      --           let -- type of projection instantiated at self
+      --               target' = Just $ proj $> dType `apply` pargs
+      --               sc' = sc { scPats   = scPats sc ++ [fmap (Named Nothing . ProjP) proj]
+      --                        , scTarget = target'
+      --                        }
+      --           (unArg proj,) <$> do cover f cs =<< fixTarget sc'
+      --       let -- WRONG: -- n = length ps -- past the last argument, is pos. of proj pat.
+      --           -- n = size tel -- past the last variable, is pos. of proj pat. DURING SPLITTING
+      --           n = permRange perm -- Andreas & James, 2013-11-19 includes the dot patterns!
+      --           -- See test/succeed/CopatternsAndDotPatterns.agda for a case with dot patterns
+      --           -- and copatterns which fails for @n = size tel@ with a broken case tree.
+      --           tree = SplitAt n $ zip projs trees
+      --       return (tree, Set.unions useds, concat psss)
+      --     _ -> done
+
+    -- case: split on variable
     Block bs -> do
       reportSLn "tc.cover.strategy" 20 $ "blocking vars = " ++ show bs
       -- xs is a non-empty lists of blocking variables
@@ -232,12 +244,12 @@ cover cs sc@(SClause tel perm ps _ target) = do
         Right (Covering n []) ->
           return (SplittingDone (size tel), Set.empty, [])
         Right (Covering n scs) -> do
-          (trees, useds, psss) <- unzip3 <$> mapM (cover cs) (map snd scs)
+          (trees, useds, psss) <- unzip3 <$> mapM (cover f cs) (map snd scs)
           let tree = SplitAt n $ zipWith (\ (q,_) t -> (q,t)) scs trees
           return (tree, Set.unions useds, concat psss)
 
 splitStrategy :: BlockingVars -> Telescope -> TCM BlockingVars
-splitStrategy bs tel = return $ updateLast (mapSnd (const Nothing)) xs
+splitStrategy bs tel = return $ updateLast clearBlockingVarCons xs
   -- Make sure we do not insists on precomputed coverage when
   -- we make our last try to split.
   -- Otherwise, we will not get a nice error message.
@@ -264,7 +276,8 @@ isDatatype ind at = do
       throw f = throwException . f =<< do liftTCM $ buildClosure t
   t' <- liftTCM $ reduce t
   case ignoreSharing $ unEl t' of
-    Def d args -> do
+    Def d es -> do
+      let ~(Just args) = allApplyElims es
       def <- liftTCM $ theDef <$> getConstInfo d
       splitOnIrrelevantDataAllowed <- liftTCM $ optExperimentalIrrelevance <$> pragmaOptions
       case def of
@@ -272,16 +285,38 @@ isDatatype ind at = do
           | i == CoInductive && ind /= CoInductive ->
               throw CoinductiveDatatype
           -- Andreas, 2011-10-03 allow some splitting on irrelevant data (if only one constr. matches)
-          | domRelevance at == Irrelevant && not splitOnIrrelevantDataAllowed ->
+          | isIrrelevant at && not splitOnIrrelevantDataAllowed ->
               throw IrrelevantDatatype
           | otherwise -> do
               let (ps, is) = genericSplitAt np args
               return (d, ps, is, cs)
-        Record{recPars = np, recCon = c} ->
-          return (d, args, [], [c])
+        Record{recPars = np, recConHead = con} ->
+          return (d, args, [], [conName con])
         _ -> throw NotADatatype
     _ -> throw NotADatatype
 
+-- | update the target type, add more patterns to split clause
+-- if target becomes a function type.
+fixTarget :: SplitClause -> TCM SplitClause
+fixTarget sc@SClause{ scSubst = sigma, scTarget = target } =
+  caseMaybe target (return sc) $ \ a -> do
+    reportSDoc "tc.cover.target" 20 $
+      text "target type before substitution: " <+> prettyTCM a
+    TelV tel b <- telView $ applySubst sigma $ unArg a
+    reportSDoc "tc.cover.target" 10 $
+      text "telescope (after substitution): " <+> prettyTCM tel
+    let n      = size tel
+        lgamma = telToList tel
+        xs     = for lgamma $ \ (Common.Dom ai (x, _)) -> Common.Arg ai $ namedVarP "_"
+    if (n == 0) then return sc { scTarget = Just $ a $> b }
+     else return $ SClause
+      { scTel    = telFromList $ telToList (scTel sc) ++ lgamma
+      , scPerm   = liftP n $ scPerm sc
+      , scPats   = scPats sc ++ xs
+      , scSubst  = liftS n $ sigma
+      , scTarget = Just $ a $> b
+      }
+
 -- | @computeNeighbourhood delta1 delta2 perm d pars ixs hix hps con@
 --
 --   @
@@ -296,22 +331,40 @@ isDatatype ind at = do
 --      con      Constructor to fit into hole
 --   @
 --   @dtype == d pars ixs@
-computeNeighbourhood :: Telescope -> String -> Telescope -> Permutation -> QName -> Args -> Args -> Nat -> OneHolePatterns -> QName -> CoverM [SplitClause]
-computeNeighbourhood delta1 n delta2 perm d pars ixs hix hps con = do
+computeNeighbourhood
+  :: Telescope                  -- ^ Telescope before split point.
+  -> PatVarName                 -- ^ Name of pattern variable at split point.
+  -> Telescope                  -- ^ Telescope after split point.
+  -> Permutation                -- ^
+  -> QName                      -- ^ Name of datatype to split at.
+  -> Args                       -- ^ Data type parameters.
+  -> Args                       -- ^ Data type indices.
+  -> Nat                        -- ^ Index of split variable.
+  -> OneHolePatterns            -- ^ Patterns with hole at split point.
+  -> QName                      -- ^ Constructor to fit into hole.
+  -> CoverM (Maybe SplitClause) -- ^ New split clause if successful.
+computeNeighbourhood delta1 n delta2 perm d pars ixs hix hps c = do
 
   -- Get the type of the datatype
   dtype <- liftTCM $ (`piApply` pars) . defType <$> getConstInfo d
 
   -- Get the real constructor name
+  con <- liftTCM $ getConForm c
+  con <- return $ con { conName = c }  -- What if we restore the current name?
+                                       -- Andreas, 2013-11-29 changes nothing!
+{-
+  con <- conSrcCon . theDef <$> getConstInfo con
   Con con [] <- liftTCM $ ignoreSharing <$> (constructorForm =<< normalise (Con con []))
+-}
 
   -- Get the type of the constructor
-  ctype <- liftTCM $ defType <$> getConstInfo con
+  ctype <- liftTCM $ defType <$> getConInfo con
 
   -- Lookup the type of the constructor at the given parameters
   (gamma0, cixs) <- do
     TelV gamma0 (El _ d) <- liftTCM $ telView (ctype `piApply` pars)
-    let Def _ cixs = ignoreSharing d
+    let Def _ es = ignoreSharing d
+        Just cixs = allApplyElims es
     return (gamma0, cixs)
 
   -- Andreas, 2012-02-25 preserve name suggestion for recursive arguments
@@ -320,12 +373,16 @@ computeNeighbourhood delta1 n delta2 perm d pars ixs hix hps con = do
   let preserve (x, t@(El _ (Def d' _))) | d == d' = (n, t)
       preserve (x, (El s (Shared p))) = preserve (x, El s $ derefPtr p)
       preserve p = p
-      gamma = telFromList . map (fmap preserve) . telToList $ gamma0
+      gammal = map (fmap preserve) . telToList $ gamma0
+      gamma  = telFromList gammal
 
-  debugInit con ctype pars ixs cixs delta1 delta2 gamma hps hix
+  debugInit con ctype d pars ixs cixs delta1 delta2 gamma hps hix
 
   -- All variables are flexible
-  let flex = [0..size delta1 + size gamma - 1]
+  -- let flex = [0..size delta1 + size gamma - 1]
+  let gammaDelta1  = gammal ++ telToList delta1
+      makeFlex i d = flexibleVarFromHiding (getHiding d) i
+      flex = zipWith makeFlex [0 .. size gammaDelta1 - 1] gammaDelta1
 
   -- Unify constructor target and given type (in Δ₁Γ)
   let conIxs   = drop (size pars) cixs
@@ -337,11 +394,11 @@ computeNeighbourhood delta1 n delta2 perm d pars ixs hix hps con = do
   case r of
     NoUnify _ _ _ -> do
       debugNoUnify
-      return []
+      return Nothing
     DontKnow _    -> do
       debugCantSplit
-      throwException $ CantSplit con (delta1 `abstract` gamma) conIxs givenIxs
-                                 [ var i | i <- flex ]
+      throwException $ CantSplit (conName con) (delta1 `abstract` gamma) conIxs givenIxs
+                                 (map (var . flexVar) flex)
     Unifies sub   -> do
       debugSubst "sub" sub
 
@@ -358,7 +415,7 @@ computeNeighbourhood delta1 n delta2 perm d pars ixs hix hps con = do
 
       -- Plug the hole with the constructor and apply ρ
       -- TODO: Is it really correct to use Nothing here?
-      let conp = ConP con Nothing $ map (fmap VarP) $ teleArgNames gamma
+      let conp = ConP con Nothing $ map (fmap namedVarP) $ teleArgNames gamma
           ps   = plugHole conp hps
           ps'  = applySubst rho ps      -- Δ₁ΓΔ₂' ⊢ ps'
       debugPlugged ps ps'
@@ -396,19 +453,20 @@ computeNeighbourhood delta1 n delta2 perm d pars ixs hix hps con = do
 
       debugFinal theta' rperm rps
 
-      return [SClause theta' rperm rps rsub Nothing] -- target fixed later
+      return $ Just $ SClause theta' rperm rps rsub Nothing -- target fixed later
 
   where
-    debugInit con ctype pars ixs cixs delta1 delta2 gamma hps hix =
+    debugInit con ctype d pars ixs cixs delta1 delta2 gamma hps hix =
       liftTCM $ reportSDoc "tc.cover.split.con" 20 $ vcat
         [ text "computeNeighbourhood"
         , nest 2 $ vcat
           [ text "con    =" <+> prettyTCM con
           , text "ctype  =" <+> prettyTCM ctype
           , text "hps    =" <+> text (show hps)
+          , text "d      =" <+> prettyTCM d
           , text "pars   =" <+> prettyList (map prettyTCM pars)
-          , text "ixs    =" <+> addCtxTel (delta1 `abstract` gamma) (prettyList (map prettyTCM ixs))
-          , text "cixs   =" <+> prettyList (map prettyTCM cixs)
+          , text "ixs    =" <+> addCtxTel delta1 (prettyList (map prettyTCM ixs))
+          , text "cixs   =" <+> do addCtxTel gamma $ prettyList (map prettyTCM cixs)
           , text "delta1 =" <+> prettyTCM delta1
           , text "delta2 =" <+> prettyTCM delta2
           , text "gamma  =" <+> prettyTCM gamma
@@ -451,16 +509,26 @@ computeNeighbourhood delta1 n delta2 perm d pars ixs hix hps con = do
         ]
 
 -- | Entry point from @Interaction.MakeCase@.
---   @Abs@ is for absurd clause.
-splitClauseWithAbs :: Clause -> Nat -> TCM (Either SplitError (Either SplitClause Covering))
-splitClauseWithAbs c x = split' Inductive (clauseToSplitClause c) (x, Nothing)
+splitClauseWithAbsurd :: Clause -> Nat -> TCM (Either SplitError (Either SplitClause Covering))
+splitClauseWithAbsurd c x = split' Inductive (clauseToSplitClause c) (BlockingVar x Nothing)
 
 -- | Entry point from @TypeChecking.Empty@ and @Interaction.BasicOps@.
-splitLast :: Induction -> Telescope -> [Arg Pattern] -> TCM (Either SplitError Covering)
-splitLast ind tel ps = split ind sc (0, Nothing)
+splitLast :: Induction -> Telescope -> [I.NamedArg Pattern] -> TCM (Either SplitError Covering)
+splitLast ind tel ps = split ind sc (BlockingVar 0 Nothing)
   where sc = SClause tel (idP $ size tel) ps __IMPOSSIBLE__ Nothing
 
--- | @split _ Δ π ps x@. FIXME: Δ ⊢ ps, x ∈ Δ (deBruijn index)
+-- | @split ind splitClause x = return res@
+--   splits @splitClause@ at pattern var @x@ (de Bruijn index).
+--
+--   Possible results @res@ are:
+--
+--   1. @Left err@:
+--      Splitting failed.
+--
+--   2. @Right covering@:
+--      A covering set of split clauses, one for each valid constructor.
+--      This could be the empty set (denoting an absurd clause).
+
 split :: Induction
          -- ^ Coinductive constructors are allowed if this argument is
          -- 'CoInductive'.
@@ -468,28 +536,35 @@ split :: Induction
       -> BlockingVar
       -> TCM (Either SplitError Covering)
 split ind sc x = fmap (blendInAbsurdClause (splitDbIndexToLevel sc x)) <$>
-   split' ind sc x
-{- OLD
-split ind sc@SClause{ scTel = tel, scPerm = perm, scPats = ps } x =
-  r <- split' ind sc x
-  return $ case r of
-    Left err        -> Left err
-    Right (Left _)  -> Right $ Covering (dbIndexToLevel tel perm $ fst x) []
-    Right (Right c) -> Right c
--}
+    split' ind sc x
+  where
+    blendInAbsurdClause :: Nat -> Either SplitClause Covering -> Covering
+    blendInAbsurdClause n = either (const $ Covering n []) id
 
-blendInAbsurdClause :: Nat -> Either SplitClause Covering -> Covering
-blendInAbsurdClause n = either (const $ Covering n []) id
-
-splitDbIndexToLevel :: SplitClause -> BlockingVar -> Nat
-splitDbIndexToLevel sc@SClause{ scTel = tel, scPerm = perm } x =
-  dbIndexToLevel tel perm $ fst x
+    splitDbIndexToLevel :: SplitClause -> BlockingVar -> Nat
+    splitDbIndexToLevel sc@SClause{ scTel = tel, scPerm = perm } x =
+      dbIndexToLevel tel perm $ blockingVarNo x
 
 -- | Convert a de Bruijn index relative to a telescope to a de Buijn level.
 --   The result should be the argument (counted from left, starting with 0)
 --   to split at (dot patterns included!).
 dbIndexToLevel tel perm x = if n < 0 then __IMPOSSIBLE__ else n
-  where n = permute perm [0..] !! (size tel - x - 1)
+  where n = if k < 0 then __IMPOSSIBLE__ else permute perm [0..] !! k
+        k = size tel - x - 1
+
+-- | @split' ind splitClause x = return res@
+--   splits @splitClause@ at pattern var @x@ (de Bruijn index).
+--
+--   Possible results @res@ are:
+--
+--   1. @Left err@:
+--      Splitting failed.
+--
+--   2. @Right (Left splitClause')@:
+--      Absurd clause (type of @x@ has 0 valid constructors).
+--
+--   3. @Right (Right covering)@:
+--      A covering set of split clauses, one for each valid constructor.
 
 split' :: Induction
           -- ^ Coinductive constructors are allowed if this argument is
@@ -497,15 +572,15 @@ split' :: Induction
        -> SplitClause
        -> BlockingVar
        -> TCM (Either SplitError (Either SplitClause Covering))
-split' ind sc@(SClause tel perm ps _ target) (x, mcons) = liftTCM $ runExceptionT $ do
+split' ind sc@(SClause tel perm ps _ target) (BlockingVar x mcons) = liftTCM $ runExceptionT $ do
 
   debugInit tel perm x ps
 
   -- Split the telescope at the variable
   -- t = type of the variable,  Δ₁ ⊢ t
   (n, t, delta1, delta2) <- do
-    let (tel1, Dom h r (n, t) : tel2) = genericSplitAt (size tel - x - 1) $ telToList tel
-    return (n, Dom h r t, telFromList tel1, telFromList tel2)
+    let (tel1, C.Dom info (n, t) : tel2) = genericSplitAt (size tel - x - 1) $ telToList tel
+    return (n, C.Dom info t, telFromList tel1, telFromList tel2)
 
   -- Compute the one hole context of the patterns at the variable
   (hps, hix) <- do
@@ -524,14 +599,14 @@ split' ind sc@(SClause tel perm ps _ target) (x, mcons) = liftTCM $ runException
   -- cons = constructors of this datatype
   (d, pars, ixs, cons) <- inContextOfT $ isDatatype ind t
 
-  liftTCM $ whenM (optWithoutK <$> pragmaOptions) $
-    inContextOfT $ Split.wellFormedIndices (unDom t)
+  --liftTCM $ whenM (optWithoutK <$> pragmaOptions) $
+  --  inContextOfT $ Split.wellFormedIndices (unDom t)
 
   -- Compute the neighbourhoods for the constructors
-  ns <- concat <$> do
+  ns <- catMaybes <$> do
     forM cons $ \ con ->
-      map (con,) <$> do
-        mapM fixTarget =<<
+      fmap (con,) <$> do
+        Trav.mapM (\sc -> lift $ fixTarget $ sc { scTarget = target }) =<< do
           computeNeighbourhood delta1 n delta2 perm d pars ixs hix hps con
   case ns of
     []  -> do
@@ -549,13 +624,14 @@ split' ind sc@(SClause tel perm ps _ target) (x, mcons) = liftTCM $ runException
     -- Andreas, 2011-10-03
     -- if more than one constructor matches, we cannot be irrelevant
     -- (this piece of code is unreachable if --experimental-irrelevance is off)
-    (_ : _ : _) | unusableRelevance (domRelevance t) ->
+    (_ : _ : _) | unusableRelevance (getRelevance t) ->
       throwException . IrrelevantDatatype =<< do liftTCM $ buildClosure (unDom t)
 
   -- Andreas, 2012-10-10 fail if precomputed constructor set does not cover
   -- all the data type constructors
 
-    _ | Just pcons <- mcons,
+    _ | Just pcons' <- mcons,
+        let pcons = map conName pcons',
         let cons = (map fst ns),
         let diff = Set.fromList cons Set.\\ Set.fromList pcons,
         not (Set.null diff) -> do
@@ -570,24 +646,6 @@ split' ind sc@(SClause tel perm ps _ target) (x, mcons) = liftTCM $ runException
   where
     xDBLevel = dbIndexToLevel tel perm x
 
-    -- update the target type, add more patterns to split clause
-    -- if target becomes a function type.
-    fixTarget :: SplitClause -> CoverM SplitClause
-    fixTarget sc@SClause{ scSubst = sigma } =
-      flip (maybe $ return sc) target $ \ a -> do
-        TelV tel b <- lift $ telView $ applySubst sigma a
-        let n      = size tel
-            lgamma = telToList tel
-            xs     = map (argFromDom . fmap (const $ VarP "_")) $ lgamma
-        if (n == 0) then return sc { scTarget = Just b }
-         else return $ SClause
-          { scTel    = telFromList $ telToList (scTel sc) ++ lgamma
-          , scPerm   = liftP n $ scPerm sc
-          , scPats   = scPats sc ++ xs
-          , scSubst  = liftS n $ sigma
-          , scTarget = Just b
-          }
-
     inContextOfT :: MonadTCM tcm => tcm a -> tcm a
     inContextOfT = addCtxTel tel . escapeContext (x + 1)
 
@@ -608,9 +666,57 @@ split' ind sc@(SClause tel perm ps _ target) (x, mcons) = liftTCM $ runException
 
     debugHoleAndType delta1 delta2 s hps t =
       liftTCM $ reportSDoc "tc.cover.top" 10 $ nest 2 $ vcat $
-        [ text "p      =" <+> text s
+        [ text "p      =" <+> text (patVarNameToString s)
         , text "hps    =" <+> text (show hps)
         , text "delta1 =" <+> prettyTCM delta1
         , text "delta2 =" <+> inContextOfDelta2 (prettyTCM delta2)
         , text "t      =" <+> inContextOfT (prettyTCM t)
         ]
+
+-- | @splitResult f sc = return res@
+--
+--   If the target type of @sc@ is a record type, a covering set of
+--   split clauses is returned (@sc@ extended by all valid projection patterns),
+--   otherwise @res == Nothing@.
+--   Note that the empty set of split clauses is returned if the record has no fields.
+splitResult :: QName -> SplitClause -> TCM (Maybe Covering)
+splitResult f sc@(SClause tel perm ps _ target) = do
+  reportSDoc "tc.cover.split" 10 $ vcat
+    [ text "splitting result:"
+    , nest 2 $ text "f      =" <+> text (show f)
+    , nest 2 $ text "target =" <+> (addContext tel $ maybe empty prettyTCM target)
+    ]
+  -- if we want to split projections, but have no target type, we give up
+  let done = return Nothing
+  caseMaybe target done $ \ t -> do
+    isR <- addCtxTel tel $ isRecordType $ unArg t
+    case isR of
+      Just (_r, vs, Record{ recFields = fs }) -> do
+        reportSDoc "tc.cover" 20 $ sep
+          [ text $ "we are of record type _r = " ++ show _r
+          , text   "applied to parameters vs = " <+> (addCtxTel tel $ prettyTCM vs)
+          , text $ "and have fields       fs = " ++ show fs
+          ]
+        fvs <- freeVarsToApply f
+        let es = patternsToElims perm ps
+        let self  = defaultArg $ Def f (map Apply fvs) `applyE` es
+            pargs = vs ++ [self]
+        reportSDoc "tc.cover" 20 $ sep
+          [ text   "we are              self = " <+> (addCtxTel tel $ prettyTCM $ unArg self)
+          ]
+        let -- WRONG: -- n = length ps -- past the last argument, is pos. of proj pat.
+            -- n = size tel -- past the last variable, is pos. of proj pat. DURING SPLITTING
+            n = permRange perm -- Andreas & James, 2013-11-19 includes the dot patterns!
+            -- See test/succeed/CopatternsAndDotPatterns.agda for a case with dot patterns
+            -- and copatterns which fails for @n = size tel@ with a broken case tree.
+        Just . Covering n <$> do
+          forM fs $ \ proj -> do
+            -- compute the new target
+            dType <- defType <$> do getConstInfo $ unArg proj -- WRONG: typeOfConst $ unArg proj
+            let -- type of projection instantiated at self
+                target' = Just $ proj $> dType `apply` pargs
+                sc' = sc { scPats   = scPats sc ++ [fmap (Named Nothing . ProjP) proj]
+                         , scTarget = target'
+                         }
+            return (unArg proj, sc')
+      _ -> done
