@@ -3,35 +3,87 @@
 module Agda.TypeChecking.Datatypes where
 
 import Control.Applicative ((<$>))
-import Data.List
+
+import Data.Maybe (fromMaybe)
 
 import Agda.Syntax.Common
-import Agda.Syntax.Internal
+import Agda.Syntax.Internal as I
+
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Reduce
+import Agda.TypeChecking.Monad.Builtin (constructorForm)
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Substitute
 
 import Agda.Utils.Size
-import Agda.Utils.Impossible
+
 #include "../undefined.h"
+import Agda.Utils.Impossible
+
+---------------------------------------------------------------------------
+-- * Constructors
+---------------------------------------------------------------------------
+
+-- | Get true constructor with record fields.
+getConHead :: QName -> TCM ConHead
+getConHead c = conSrcCon . theDef <$> getConstInfo c
+
+-- | Get true constructor as term.
+getConTerm :: QName -> TCM Term
+getConTerm c = flip Con [] <$> getConHead c
+
+-- | Get true constructor with fields, expanding literals to constructors
+--   if possible.
+getConForm :: QName -> TCM ConHead
+getConForm c = do
+  Con con [] <- ignoreSharing <$> do constructorForm =<< getConTerm c
+  return con
+
+-- | Augment constructor with record fields (preserve constructor name).
+--   The true constructor might only surface via 'reduce'.
+getOrigConHead :: QName -> TCM ConHead
+getOrigConHead c = setConName c <$> getConHead c
+
+-- | Analogous to 'getConTerm'.
+getOrigConTerm :: QName -> TCM Term
+getOrigConTerm c = flip Con [] <$> getOrigConHead c
 
 -- | Get the name of the datatype constructed by a given constructor.
 --   Precondition: The argument must refer to a constructor
-getConstructorData :: QName -> TCM QName
+{-# SPECIALIZE getConstructorData :: QName -> TCM QName #-}
+getConstructorData :: HasConstInfo m => QName -> m QName
 getConstructorData c = do
   def <- getConstInfo c
   case theDef def of
     Constructor{conData = d} -> return d
     _                        -> __IMPOSSIBLE__
 
+-- | @getConType c t@ computes the constructor parameters from type @t@
+--   and returns the instantiated type of constructor @c@.
+--   @Nothing@ if @t@ is not a data/record type or does not have
+--   a constructor @c@.
+--   Precondition: @t@ is reduced.
+getConType :: ConHead -> Type -> TCM (Maybe Type)
+getConType c t = do
+  c <- getConHead $ conName c
+  case ignoreSharing $ unEl t of
+    Def d es -> do
+      def <- theDef <$> getConstInfo d
+      case def of
+        Datatype { dataPars = n, dataCons   = cs  } | conName c `elem` cs -> cont n
+        Record   { recPars  = n, recConHead = con } | c == con            -> cont n
+        _ ->  return Nothing
+      where
+        cont n = do
+          let pars = fromMaybe __IMPOSSIBLE__ $ allApplyElims $ take n es
+          Just . (`apply` pars) . defType <$> getConInfo c
+    _ -> return Nothing
 
 -- | Return the number of non-parameter arguments to a data constructor,
 --   or the field names of a record constructor.
 --
 --   For getting just the arity of constructor @c@,
 --   use @either id size <$> getConstructorArity c@.
-getConstructorArity :: QName -> TCM (Either Nat [Arg QName])
+getConstructorArity :: QName -> TCM (Either Nat [I.Arg QName])
 getConstructorArity c = do
   Defn{ defType = t, theDef = def } <- getConstInfo c
   case def of
@@ -48,6 +100,9 @@ getConstructorArity c = do
         _ -> __IMPOSSIBLE__
     _ -> __IMPOSSIBLE__
 
+---------------------------------------------------------------------------
+-- * Data types
+---------------------------------------------------------------------------
 
 -- | Check if a name refers to a datatype or a record with a named constructor.
 isDatatype :: QName -> TCM Bool
