@@ -1,3 +1,4 @@
+-- {-# LANGUAGE CPP #-}
 
 module Agda.TypeChecking.Empty where
 
@@ -14,9 +15,6 @@ import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 
-import Agda.Utils.Permutation
-import Agda.Utils.Size
-
 {- UNUSED
 -- | Make sure that a type is empty.
 isReallyEmptyType :: Range -> Type -> TCM ()
@@ -37,8 +35,8 @@ isEmptyType r t = do
     -- from the current context xs:ts, create a pattern list
     -- xs _ : ts t and try to split on _ (the last variable)
       tel0 <- getContextTelescope
-      let gamma = telToList tel0 ++ [domFromArg $ defaultArg ("_", t)]
-          ps    = [ Arg h r $ VarP x | Dom h r (x, _) <- gamma ]
+      let gamma = telToList tel0 ++ [domFromArg $ defaultArg (underscore, t)]
+          ps    = [ Arg info $ namedVarP x | Dom info (x, _) <- gamma ]
           tel   = telFromList gamma
 
       dontAssignMetas $ do
@@ -53,4 +51,4 @@ isEmptyType r t = do
         Right cov -> do
           let cs = splitClauses cov
           unless (null cs) $
-            typeError $ ShouldBeEmpty t $ map (unArg . last . scPats) $ cs
+            typeError $ ShouldBeEmpty t $ map (namedArg . last . scPats) $ cs
