@@ -1,4 +1,8 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE TupleSections #-}
+{-# LANGUAGE TypeSynonymInstances #-}
+
 module Agda.TypeChecking.Errors
     ( prettyError
     , PrettyTCM(..)
@@ -7,16 +11,16 @@ module Agda.TypeChecking.Errors
     , warningsToError
     ) where
 
-import Control.Applicative ( (<$>) )
 import Control.Monad.State
 import Control.Monad.Error
 
-import Data.List (nub)
+import Data.Function
+import Data.List (nub, sortBy)
+import Data.Maybe
 import qualified Data.Map as Map (empty)
 
-import System.FilePath
-
-import Agda.Syntax.Common
+import Agda.Syntax.Common hiding (Arg, Dom, NamedArg)
+import qualified Agda.Syntax.Common as Common
 import Agda.Syntax.Fixity
 import Agda.Syntax.Position
 import qualified Agda.Syntax.Info as A
@@ -25,19 +29,19 @@ import qualified Agda.Syntax.Concrete.Definitions as D
 import Agda.Syntax.Abstract as A
 import Agda.Syntax.Internal as I
 import qualified Agda.Syntax.Abstract.Pretty as P
-import qualified Agda.Syntax.Concrete.Pretty as P
 import Agda.Syntax.Translation.InternalToAbstract
 import Agda.Syntax.Translation.AbstractToConcrete
-import Agda.Syntax.Scope.Base (ScopeInfo(..))
 import Agda.Syntax.Scope.Monad (isDatatypeModule)
 import Agda.TypeChecking.Monad.Base
+import Agda.TypeChecking.Monad.Closure
 import Agda.TypeChecking.Monad.Context
 import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Reduce (instantiate)
 
 import Agda.Utils.FileName
 import Agda.Utils.Monad
-import Agda.Utils.Size
+import qualified Agda.Utils.Pretty as P
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
@@ -46,7 +50,8 @@ import Agda.Utils.Impossible
 -- * Top level function
 ---------------------------------------------------------------------------
 
-prettyError :: TCErr -> TCM String
+{-# SPECIALIZE prettyError :: TCErr -> TCM String #-}
+prettyError :: MonadTCM tcm => TCErr -> tcm String
 prettyError err = liftTCM $ liftM show $
     prettyTCM err
     `catchError` \err' -> text "panic: error when printing error!" $$ prettyTCM err'
@@ -89,7 +94,7 @@ warningsToError (Warnings w@(_:_) _ _) = TerminationCheckFailed w
 ---------------------------------------------------------------------------
 
 sayWhere :: HasRange a => a -> TCM Doc -> TCM Doc
-sayWhere x d = text (show $ getRange x) $$ d
+sayWhere x d = prettyTCM (getRange x) $$ d
 
 sayWhen :: Range -> Maybe (Closure Call) -> TCM Doc -> TCM Doc
 sayWhen r Nothing   m = sayWhere r m
@@ -100,7 +105,7 @@ panic s = fwords $ "Panic: " ++ s
 
 nameWithBinding :: QName -> TCM Doc
 nameWithBinding q =
-  sep [ prettyTCM q, text "bound at", text (show r) ]
+  sep [ prettyTCM q, text "bound at", prettyTCM r ]
   where
     r = nameBindingSite $ qnameName q
 
@@ -137,6 +142,7 @@ errorString err = case err of
     CyclicModuleDependency{}                 -> "CyclicModuleDependency"
     DataMustEndInSort{}                      -> "DataMustEndInSort"
 -- UNUSED:    DataTooManyParameters{}                  -> "DataTooManyParameters"
+    CantResolveOverloadedConstructorsTargetingSameDatatype{} -> "CantResolveOverloadedConstructorsTargetingSameDatatype"
     DifferentArities                         -> "DifferentArities"
     DoesNotConstructAnElementOf{}            -> "DoesNotConstructAnElementOf"
     DuplicateBuiltinBinding{}                -> "DuplicateBuiltinBinding"
@@ -149,6 +155,9 @@ errorString err = case err of
     GenericDocError{}                        -> "GenericDocError"
     IFSNoCandidateInScope{}                  -> "IFSNoCandidateInScope"
     IlltypedPattern{}                        -> "IlltypedPattern"
+    IllformedProjectionPattern{}             -> "IllformedProjectionPattern"
+    CannotEliminateWithPattern{}             -> "CannotEliminateWithPattern"
+    IllegalLetInTelescope{}                  -> "IllegalLetInTelescope"
     IncompletePatternMatching{}              -> "IncompletePatternMatching"
     IndexVariablesNotDistinct{}              -> "IndexVariablesNotDistinct"
     IndicesFreeInParameters{}                -> "IndicesFreeInParameters"
@@ -175,7 +184,8 @@ errorString err = case err of
     NoSuchPrimitiveFunction{}                -> "NoSuchPrimitiveFunction"
     NotAModuleExpr{}                         -> "NotAModuleExpr"
     NotAProperTerm                           -> "NotAProperTerm"
-    SetOmegaNotValidType                     -> "SetOmegaNotValidType"
+    SetOmegaNotValidType{}                   -> "SetOmegaNotValidType"
+    InvalidType{}                            -> "InvalidType"
     NotAValidLetBinding{}                    -> "NotAValidLetBinding"
     NotAnExpression{}                        -> "NotAnExpression"
     NotImplemented{}                         -> "NotImplemented"
@@ -202,7 +212,9 @@ errorString err = case err of
     ShouldBePi{}                             -> "ShouldBePi"
     ShouldBeRecordType{}                     -> "ShouldBeRecordType"
     ShouldBeRecordPattern{}                  -> "ShouldBeRecordPattern"
+    NotAProjectionPattern{}                  -> "NotAProjectionPattern"
     ShouldEndInApplicationOfTheDatatype{}    -> "ShouldEndInApplicationOfTheDatatype"
+    SplitError{}                             -> "SplitError"
     TerminationCheckFailed{}                 -> "TerminationCheckFailed"
     TooFewFields{}                           -> "TooFewFields"
     TooManyArgumentsInLHS{}                  -> "TooManyArgumentsInLHS"
@@ -213,11 +225,12 @@ errorString err = case err of
     UnequalBecauseOfUniverseConflict{}       -> "UnequalBecauseOfUniverseConflict"
     UnequalRelevance{}                       -> "UnequalRelevance"
     UnequalHiding{}                          -> "UnequalHiding"
-    UnequalLevel{}                           -> "UnequalLevel"
+--    UnequalLevel{}                           -> "UnequalLevel" -- UNUSED
     UnequalSorts{}                           -> "UnequalSorts"
     UnequalTerms{}                           -> "UnequalTerms"
     UnequalTypes{}                           -> "UnequalTypes"
-    UnequalTelescopes{}                      -> "UnequalTelescopes"
+--    UnequalTelescopes{}                      -> "UnequalTelescopes" -- UNUSED
+    UnequalColors{}                          -> "UnequalTelescopes"
     HeterogeneousEquality{}                  -> "HeterogeneousEquality"
     UnexpectedWithPatterns{}                 -> "UnexpectedWithPatterns"
     UninstantiatedDotPattern{}               -> "UninstantiatedDotPattern"
@@ -227,14 +240,22 @@ errorString err = case err of
     UnsolvedMetas{}                          -> "UnsolvedMetas"
     UnusedVariableInPatternSynonym           -> "UnusedVariableInPatternSynonym"
     WithClausePatternMismatch{}              -> "WithClausePatternMismatch"
+    WithoutKError{}                          -> "WithoutKError"
     WrongHidingInApplication{}               -> "WrongHidingInApplication"
     WrongHidingInLHS{}                       -> "WrongHidingInLHS"
     WrongHidingInLambda{}                    -> "WrongHidingInLambda"
     WrongIrrelevanceInLambda{}               -> "WrongIrrelevanceInLambda"
+    WrongNamedArgument{}                     -> "WrongNamedArgument"
     WrongNumberOfConstructorArguments{}      -> "WrongNumberOfConstructorArguments"
+    HidingMismatch{}                         -> "HidingMismatch"
+    RelevanceMismatch{}                      -> "RelevanceMismatch"
+    ColorMismatch{}                          -> "ColorMismatch"
 
 instance PrettyTCM TCErr where
     prettyTCM err = case err of
+        -- Andreas, 2014-03-23
+        -- This use of localState seems ok since we do not collect
+        -- Benchmark info during printing errors.
 	TypeError s e -> localState $ do
 	    put s
 	    sayWhen (envRange $ clEnv e) (envCall $ clEnv e) $ prettyTCM e
@@ -243,6 +264,18 @@ instance PrettyTCM TCErr where
 	PatternErr _    -> sayWhere err $ panic "uncaught pattern violation"
 	{- AbortAssign _   -> sayWhere err $ panic "uncaught aborted assignment" -- UNUSED -}
 
+instance PrettyTCM CallInfo where
+  prettyTCM c = do
+    let call = prettyTCM $ callInfoCall c
+        r    = callInfoRange c
+    case show r of
+      "" -> call
+      _  -> call $$ nest 2 (text "(at" <+> prettyTCM r <> text ")")
+
+-- | Drops the filename component of the qualified name
+dropTopLevelModule :: QName -> QName
+dropTopLevelModule (QName (MName ns) n) = QName (MName (drop 1 ns)) n
+
 instance PrettyTCM TypeError where
     prettyTCM err = do
 	case err of
@@ -255,15 +288,12 @@ instance PrettyTCM TypeError where
 	    TerminationCheckFailed because ->
               fwords "Termination checking failed for the following functions:"
               $$ (nest 2 $
-                    fsep (punctuate comma (map (text . show . qnameName)
+                    fsep (punctuate comma (map (text . show . dropTopLevelModule)
                                                (concatMap termErrFunctions because))))
               $$ fwords "Problematic calls:"
-              $$ (nest 2 $ vcat $
-                    map (\c -> let call = text (callInfoCall c) in
-                               case show (callInfoRange c) of
-                                 "" -> call
-                                 r  -> call $$ nest 2 (text "(at" <+> text r <> text ")"))
-                        (nub $ concatMap termErrCalls because))
+              $$ (nest 2 $ fmap (P.vcat . nub) $
+                    mapM prettyTCM $ sortBy (compare `on` callInfoRange) $
+                    concatMap termErrCalls because)
 	    PropMustBeSingleton -> fwords
 		"Datatypes in Prop must have at most one constructor when proof irrelevance is enabled"
 	    DataMustEndInSort t -> fsep $
@@ -282,35 +312,66 @@ instance PrettyTCM TypeError where
 		pwords "The pattern constructs an element of" ++ [prettyTCM q] ++
 		pwords "which is not the right datatype"
 	    ShouldBeRecordType t -> fsep $
-		pwords "Expected record type, found " ++ [prettyTCM t]
+		pwords "Expected non-abstract record type, found " ++ [prettyTCM t]
 	    ShouldBeRecordPattern p -> fsep $
 		pwords "Expected record pattern" -- ", found " ++ [prettyTCM p]
+	    NotAProjectionPattern p -> fsep $
+		pwords "Not a valid projection for a copattern: " ++ [ prettyA p ]
 	    DifferentArities ->
 		fwords "The number of arguments in the defining equations differ"
-	    WrongHidingInLHS t -> do
-		fwords "Found an implicit argument where an explicit argument was expected"
+	    WrongHidingInLHS -> do
+		fwords "Unexpected implicit argument"
 	    WrongHidingInLambda t -> do
 		fwords "Found an implicit lambda where an explicit lambda was expected"
 	    WrongIrrelevanceInLambda t -> do
 		fwords "Found an irrelevant lambda where a relevant lambda was expected"
+	    WrongNamedArgument a -> fsep $
+		pwords "Function does not accept argument " ++ [prettyTCM a] -- ++ pwords " (wrong argument name)"
 	    WrongHidingInApplication t -> do
 		fwords "Found an implicit application where an explicit application was expected"
+            HidingMismatch h h' -> fwords $
+              "Expected " ++ verbalize (Indefinite h') ++ " argument, but found " ++
+              verbalize (Indefinite h) ++ " argument"
+            RelevanceMismatch r r' -> fwords $
+              "Expected " ++ verbalize (Indefinite r') ++ " argument, but found " ++
+              verbalize (Indefinite r) ++ " argument"
+            ColorMismatch c c' -> fsep $   -- TODO guilhem
+               pwords "Expected argument color to be" ++ [prettyTCM c'] ++
+               pwords "but found color" ++ [prettyTCM c]
             NotInductive t -> fsep $
               [prettyTCM t] ++ pwords "is not an inductive data type"
             UninstantiatedDotPattern e -> fsep $
               pwords "Failed to infer the value of dotted pattern"
             IlltypedPattern p a -> fsep $
               pwords "Type mismatch"
+            IllformedProjectionPattern p -> fsep $
+              pwords "Ill-formed projection pattern " ++ [prettyA p]
+            CannotEliminateWithPattern p a -> do
+              let isProj = isJust (isProjP p)
+              fsep $
+                pwords "Cannot eliminate type" ++ prettyTCM a :
+                if isProj then
+                   pwords "with projection pattern" ++ [prettyA p]
+                 else
+                   pwords "with pattern" ++ prettyA p :
+                   pwords "(did you supply too many arguments?)"
             TooManyArgumentsInLHS a -> fsep $
               pwords "Left hand side gives too many arguments to a function of type" ++ [prettyTCM a]
             WrongNumberOfConstructorArguments c expect given -> fsep $
               pwords "The constructor" ++ [prettyTCM c] ++ pwords "expects" ++
-              [text (show expect)] ++ pwords "arguments, but has been given" ++ [text (show given)]
+              [text (show expect)] ++ pwords "arguments (including hidden ones), but has been given" ++ [text (show given)] ++ pwords "(including hidden ones)"
+            CantResolveOverloadedConstructorsTargetingSameDatatype d cs -> fsep $
+              pwords ("Can't resolve overloaded constructors targeting the same datatype (" ++ show d ++ "):") ++ map (text . show) cs
             DoesNotConstructAnElementOf c t -> fsep $
-              pwords "the constructor" ++ [prettyTCM c] ++
+              pwords "The constructor" ++ [prettyTCM c] ++
               pwords "does not construct an element of" ++ [prettyTCM t]
             ConstructorPatternInWrongDatatype c d -> fsep $
               [prettyTCM c] ++ pwords "is not a constructor of the datatype" ++ [prettyTCM d]
+            IndicesNotConstructorApplications [i] ->
+              fwords "The index"
+              $$ nest 2 (prettyTCM i)
+              $$ fsep (pwords "is not a constructor (or literal) applied to variables" ++
+                       pwords "(note that parameters count as constructor arguments)")
             IndicesNotConstructorApplications is ->
               fwords "The indices"
               $$ nest 2 (vcat $ map prettyTCM is)
@@ -333,7 +394,7 @@ instance PrettyTCM TypeError where
             ShadowedModule x ms@(m : _) -> fsep $
               pwords "Duplicate definition of module" ++ [prettyTCM x <> text "."] ++
               pwords "Previous definition of" ++ [help m] ++ pwords "module" ++ [prettyTCM x] ++
-              pwords "at" ++ [text $ show r]
+              pwords "at" ++ [prettyTCM r]
               where
                 help m = do
                   b <- isDatatypeModule m
@@ -367,6 +428,8 @@ instance PrettyTCM TypeError where
 		fwords "Found a malformed term"
 	    SetOmegaNotValidType ->
 		fwords "Setω is not a valid type"
+	    InvalidType v -> fsep $
+		[prettyTCM v] ++ pwords "is not a valid type"
             SplitOnIrrelevant p t -> fsep $
                 pwords "Cannot pattern match" ++ [prettyA p] ++
                 pwords "against irrelevant type" ++ [prettyTCM t]
@@ -376,14 +439,18 @@ instance PrettyTCM TypeError where
                 text "Variable" : prettyTCM x : pwords "is declared irrelevant, so it cannot be used here"
 	    UnequalBecauseOfUniverseConflict cmp s t -> fsep $
 		[prettyTCM s, notCmp cmp, prettyTCM t, text "because this would result in an invalid use of Setω" ]
- 	    UnequalTerms cmp s t a -> fsep $
-		[prettyTCM s, notCmp cmp, prettyTCM t] ++ pwords "of type" ++ [prettyTCM a]
-	    UnequalLevel cmp s t -> fsep $
-		[prettyTCM s, notCmp cmp, prettyTCM t]
-	    UnequalTelescopes cmp a b -> fsep $
-		[prettyTCM a, notCmp cmp, prettyTCM b]
-	    UnequalTypes cmp a b -> fsep $
-		[prettyTCM a, notCmp cmp, prettyTCM b]
+ 	    UnequalTerms cmp s t a -> do
+                (d1, d2, d) <- prettyInEqual s t
+		fsep $ [return d1, notCmp cmp, return d2] ++ pwords "of type" ++ [prettyTCM a] ++ [return d]
+-- UnequalLevel is UNUSED
+--	    UnequalLevel cmp s t -> fsep $
+--		[prettyTCM s, notCmp cmp, prettyTCM t]
+-- UnequalTelescopes is UNUSED
+--	    UnequalTelescopes cmp a b -> fsep $
+--		[prettyTCM a, notCmp cmp, prettyTCM b]
+	    UnequalTypes cmp a b -> prettyUnequal a (notCmp cmp) b
+--                fsep $ [prettyTCM a, notCmp cmp, prettyTCM b]
+            UnequalColors a b -> error "TODO guilhem 4"
 	    HeterogeneousEquality u a v b -> fsep $
                 pwords "Refuse to solve heterogeneous constraint" ++
                 [prettyTCM u] ++ pwords ":" ++ [prettyTCM a] ++ pwords "=?=" ++
@@ -416,13 +483,14 @@ instance PrettyTCM TypeError where
             UnexpectedWithPatterns ps -> fsep $
               pwords "Unexpected with patterns" ++ (punctuate (text " |") $ map prettyA ps)
             WithClausePatternMismatch p q -> fsep $
-              pwords "With clause pattern" ++ [prettyA p] ++
-              pwords "is not an instance of its parent pattern" -- TODO: pretty for internal patterns
+              pwords "With clause pattern " ++ [prettyA p] ++
+              pwords " is not an instance of its parent pattern " ++ [prettyTCM q]
+                 -- TODO: prettier printing for internal patterns
 	    MetaCannotDependOn m ps i -> fsep $
 		    pwords "The metavariable" ++ [prettyTCM $ MetaV m []] ++ pwords "cannot depend on" ++ [pvar i] ++
 		    pwords "because it" ++ deps
 		where
-		    pvar i = prettyTCM $ I.Var i []
+                    pvar = prettyTCM . I.var
 		    deps = case map pvar ps of
 			[]  -> pwords "does not depend on any variables"
 			[x] -> pwords "only depends on the variable" ++ [x]
@@ -445,6 +513,10 @@ instance PrettyTCM TypeError where
 	    BuiltinInParameterisedModule x -> fwords $
 		"The BUILTIN pragma cannot appear inside a bound context " ++
 		"(for instance, in a parameterised module or as a local declaration)"
+            IllegalLetInTelescope tb -> fsep $
+                -- pwords "The binding" ++
+                [pretty tb] ++
+                pwords " is not allowed in a telescope here."
 	    NoRHSRequiresAbsurdPattern ps -> fwords $
 		"The right-hand side can only be omitted if there " ++
 		"is an absurd pattern, () or {}, in the left-hand side."
@@ -456,10 +528,15 @@ instance PrettyTCM TypeError where
 		pwords "can refer to either a local module or an imported module"
 	    UnsolvedMetas rs ->
 		fsep ( pwords "Unsolved metas at the following locations:" )
-		$$ nest 2 (vcat $ map (text . show) rs)
+		$$ nest 2 (vcat $ map prettyTCM rs)
 	    UnsolvedConstraints cs ->
 		fsep ( pwords "Failed to solve the following constraints:" )
-		$$ nest 2 (vcat $ map prettyTCM cs)
+		$$ nest 2 (vcat $ map prettyConstraint cs)
+              where prettyConstraint :: ProblemConstraint -> TCM Doc
+                    prettyConstraint c = f (prettyTCM c)
+                      where
+                        r   = getRange c
+                        f d = if null (show r) then d else d $$ nest 4 (text "[ at" <+> prettyTCM r  <+> text "]")
 	    CyclicModuleDependency ms ->
 		fsep (pwords "cyclic module dependency:")
 		$$ nest 2 (vcat $ map pretty ms)
@@ -496,7 +573,7 @@ instance PrettyTCM TypeError where
 	    NotInScope xs ->
 		fsep (pwords "Not in scope:") $$ nest 2 (vcat $ map name xs)
 		where
-                  name x = fsep [ pretty x, text "at" <+> text (show $ getRange x), suggestion (show x) ]
+                  name x = fsep [ pretty x, text "at" <+> prettyTCM (getRange x), suggestion (show x) ]
                   suggestion s
                     | elem ':' s    = parens $ text "did you forget space around the ':'?"
                     | elem "->" two = parens $ text "did you forget space around the '->'?"
@@ -509,11 +586,13 @@ instance PrettyTCM TypeError where
 	      [ fsep $ pwords "Ambiguous name" ++ [pretty x <> text "."] ++
 		       pwords "It could refer to any one of"
 	      , nest 2 $ vcat $ map nameWithBinding ys
+              , fwords "(hint: Use C-c C-w (in Emacs) if you want to know why)"
 	      ]
 	    AmbiguousModule x ys -> vcat
 	      [ fsep $ pwords "Ambiguous module name" ++ [pretty x <> text "."] ++
 		       pwords "It could refer to any one of"
 	      , nest 2 $ vcat $ map help ys
+              , fwords "(hint: Use C-c C-w (in Emacs) if you want to know why)"
 	      ]
               where
                 help :: ModuleName -> TCM Doc
@@ -526,7 +605,7 @@ instance PrettyTCM TypeError where
 		) $$ nest 2 (hsep [ text "module", pretty x <> text "'", text "=", pretty x, text "e1 .. en" ])
 	    ClashingDefinition x y -> fsep $
 		pwords "Multiple definitions of" ++ [pretty x <> text "."] ++
-		pwords "Previous definition at" ++ [text $ show $ nameBindingSite $ qnameName y]
+		pwords "Previous definition at" ++ [prettyTCM $ nameBindingSite $ qnameName y]
 	    ClashingModule m1 m2 -> fsep $
 		pwords "The modules" ++ [prettyTCM m1, text "and", prettyTCM m2] ++ pwords "clash."
 	    ClashingImport x y -> fsep $
@@ -564,9 +643,31 @@ instance PrettyTCM TypeError where
 	    NoParseForApplication es -> fsep $
 		pwords "Could not parse the application" ++ [pretty $ C.RawApp noRange es]
 	    AmbiguousParseForApplication es es' -> fsep (
-		    pwords "Don't know how to parse" ++ [pretty (C.RawApp noRange es) <> text "."] ++
+		    pwords "Don't know how to parse" ++ [pretty_es <> (text ".")] ++
 		    pwords "Could mean any one of:"
-		) $$ nest 2 (vcat $ map pretty es')
+		) $$ nest 2 (vcat $ map pretty' es')
+              where
+                pretty_es :: TCM Doc
+                pretty_es = pretty $ C.RawApp noRange es
+
+                pretty' :: C.Expr -> TCM Doc
+                pretty' e = do
+                  p1 <- pretty_es
+                  p2 <- pretty e
+                  pretty $ if show p1 == show p2 then unambiguous e else e
+
+                unambiguous :: C.Expr -> C.Expr
+                unambiguous (C.OpApp r op xs) | all (isOrdinary . namedArg) xs
+                    = foldl (C.App r) (C.Ident op) $ (map . fmap . fmap) fromOrdinary xs
+                unambiguous e = e
+
+                isOrdinary :: C.OpApp e -> Bool
+                isOrdinary (C.Ordinary _) = True
+                isOrdinary _ = False
+
+                fromOrdinary :: C.OpApp e -> e
+                fromOrdinary (C.Ordinary e) = e
+                fromOrdinary _ = __IMPOSSIBLE__
             UnusedVariableInPatternSynonym -> fsep $
                 pwords "Unused variable in pattern synonym."
             PatternSynonymArityMismatch x -> fsep $
@@ -580,9 +681,29 @@ instance PrettyTCM TypeError where
 		pwords "Could not parse the pattern synonym" ++ [pretty p]
 -}
 	    AmbiguousParseForLHS lhsOrPatSyn p ps -> fsep (
-		    pwords "Don't know how to parse" ++ [pretty p <> text "."] ++
+		    pwords "Don't know how to parse" ++ [pretty_p <> text "."] ++
 		    pwords "Could mean any one of:"
-		) $$ nest 2 (vcat $ map pretty ps)
+		) $$ nest 2 (vcat $ map pretty' ps)
+              where
+                pretty_p :: TCM Doc
+                pretty_p = pretty p
+
+                pretty' :: C.Pattern -> TCM Doc
+                pretty' p' = do
+                  p1 <- pretty_p
+                  p2 <- pretty p'
+                  pretty $ if show p1 == show p2 then unambiguousP p' else p'
+
+                -- the entire pattern is shown, not just the ambiguous part,
+                -- so we need to dig in order to find the OpAppP's.
+                unambiguousP :: C.Pattern -> C.Pattern
+                unambiguousP (C.AppP x y)       = C.AppP (unambiguousP x) $ (fmap.fmap) unambiguousP y
+                unambiguousP (C.HiddenP r x)    = C.HiddenP r $ fmap unambiguousP x
+                unambiguousP (C.InstanceP r x)  = C.InstanceP r $ fmap unambiguousP x
+                unambiguousP (C.ParenP r x)     = C.ParenP r $ unambiguousP x
+                unambiguousP (C.AsP r n x)      = C.AsP r n $ unambiguousP x
+                unambiguousP (C.OpAppP r op xs) = foldl C.AppP (C.IdentP op) xs
+                unambiguousP e = e
 {- UNUSED
 	    AmbiguousParseForPatternSynonym p ps -> fsep (
 		    pwords "Don't know how to parse" ++ [pretty p <> text "."] ++
@@ -611,12 +732,15 @@ instance PrettyTCM TypeError where
                       then return ps
                       else return ps  -- TODO: remove implicit arguments which aren't constructors
 
-            CoverageCantSplitOn c tel cIxs gIxs -> inContext [] $ addCtxTel tel $ vcat
-              [ fsep $ pwords "Cannot decide whether there should be a case for the constructor" ++ [prettyTCM c <> text ","] ++
-                       pwords "since the unification gets stuck on unifying the inferred indices"
-              , nest 2 $ prettyTCM cIxs
-              , fsep $ pwords "with the expected indices"
-              , nest 2 $ prettyTCM gIxs ]
+            CoverageCantSplitOn c tel cIxs gIxs
+              | length cIxs /= length gIxs -> __IMPOSSIBLE__
+              | otherwise                  -> addCtxTel tel $ vcat (
+                  [ fsep $ pwords "I'm not sure if there should be a case for the constructor" ++
+                           [prettyTCM c <> text ","] ++
+                           pwords "because I get stuck when trying to solve the following" ++
+                           pwords "unification problems (inferred index ≟ expected index):"
+                  ] ++
+                  zipWith (\c g -> nest 2 $ prettyTCM c <+> text "≟" <+> prettyTCM g) cIxs gIxs)
 
             CoverageCantSplitIrrelevantType a -> fsep $
               pwords "Cannot split on argument of irrelevant datatype" ++ [prettyTCM a]
@@ -624,6 +748,11 @@ instance PrettyTCM TypeError where
             CoverageCantSplitType a -> fsep $
               pwords "Cannot split on argument of non-datatype" ++ [prettyTCM a]
 
+            SplitError e -> prettyTCM e
+
+            WithoutKError a u v -> fsep $
+              pwords "Cannot eliminate reflexive equation" ++ [prettyTCM u] ++ pwords "=" ++ [prettyTCM v] ++ pwords "of type" ++ [prettyTCM a] ++ pwords "because K has been disabled."
+
 	    NotStrictlyPositive d ocs -> fsep $
 		pwords "The datatype" ++ [prettyTCM d] ++ pwords "is not strictly positive, because"
 		++ prettyOcc "it" ocs
@@ -667,20 +796,78 @@ instance PrettyTCM TypeError where
               | n > 0 && not (null args) = parens
               | otherwise                = id
 
-            showArg :: Arg I.Pattern -> TCM Doc
-            showArg (Arg Hidden r x)    = braces $ showPat 0 x
-            showArg (Arg Instance r x)  = dbraces $ showPat 0 x
-            showArg (Arg NotHidden r x) = showPat 1 x
+            showArg :: I.Arg I.Pattern -> TCM Doc
+            showArg (Common.Arg info x) = case getHiding info of
+                    Hidden -> braces $ showPat 0 x
+                    Instance -> dbraces $ showPat 0 x
+                    NotHidden -> showPat 1 x
 
             showPat :: Integer -> I.Pattern -> TCM Doc
             showPat _ (I.VarP _)        = text "_"
             showPat _ (I.DotP _)        = text "._"
-            showPat n (I.ConP c _ args) = mpar n args $ prettyTCM c <+> fsep (map showArg args)
+            showPat n (I.ConP c _ args) = mpar n args $ prettyTCM c <+> fsep (map (showArg . fmap namedThing) args)
             showPat _ (I.LitP l)        = text (show l)
+            showPat _ (I.ProjP p)       = text (show p)
 
 notCmp :: Comparison -> TCM Doc
 notCmp cmp = text $ "!" ++ show cmp
 
+-- | Print two terms that are supposedly unequal.
+--   If they print to the same identifier, add some explanation
+--   why they are different nevertheless.
+prettyInEqual :: Term -> Term -> TCM (Doc, Doc, Doc)
+prettyInEqual t1 t2 = do
+  d1 <- prettyTCM t1
+  d2 <- prettyTCM t2
+  (d1, d2,) <$> do
+     -- if printed differently, no extra explanation needed
+    if P.render d1 /= P.render d2 then return P.empty else do
+      (v1, v2) <- instantiate (t1, t2)
+      case (ignoreSharing v1, ignoreSharing v2) of
+        (I.Var i1 _, I.Var i2 _)
+          | i1 == i2  -> __IMPOSSIBLE__   -- if they're actually the same we would get the error on the arguments instead
+          | otherwise -> varVar i1 i2
+        (I.Def{}, I.Con{}) -> __IMPOSSIBLE__  -- ambiguous identifiers
+        (I.Con{}, I.Def{}) -> __IMPOSSIBLE__
+        (I.Var{}, I.Def{}) -> varDef
+        (I.Def{}, I.Var{}) -> varDef
+        (I.Var{}, I.Con{}) -> varCon
+        (I.Con{}, I.Var{}) -> varCon
+        _                  -> return P.empty
+  where
+    varDef     = return $ P.parens $ P.fwords "because one is a variable and one a defined identifier"
+    varCon     = return $ P.parens $ P.fwords "because one is a variable and one a constructor"
+    varVar i j = return $ P.parens $ P.fwords $ "because one has deBruijn index " ++ show i ++ " and the other " ++ show j
+
+class PrettyUnequal a where
+  prettyUnequal :: a -> TCM Doc -> a -> TCM Doc
+
+instance PrettyUnequal Term where
+  prettyUnequal t1 ncmp t2 = do
+    (d1, d2, d) <- prettyInEqual t1 t2
+    fsep $ return d1 : ncmp : return d2 : return d : []
+
+instance PrettyUnequal Type where
+  prettyUnequal t1 ncmp t2 = prettyUnequal (unEl t1) ncmp (unEl t2)
+
+
+instance PrettyTCM SplitError where
+  prettyTCM err = case err of
+    NotADatatype t -> enterClosure t $ \ t -> fsep $
+      pwords "Cannot pattern match on non-datatype" ++ [prettyTCM t]
+    IrrelevantDatatype t -> enterClosure t $ \ t -> fsep $
+      pwords "Cannot pattern match on datatype" ++ [prettyTCM t] ++
+      pwords "since it is declared irrelevant"
+    CoinductiveDatatype t -> enterClosure t $ \ t -> fsep $
+      pwords "Cannot pattern match on the coinductive type" ++ [prettyTCM t]
+{- UNUSED
+    NoRecordConstructor t -> fsep $
+      pwords "Cannot pattern match on record" ++ [prettyTCM t] ++
+      pwords "because it has no constructor"
+ -}
+    CantSplit c tel cIxs gIxs flex -> prettyTCM (CoverageCantSplitOn c tel cIxs gIxs)
+    GenericSplitError s -> fsep $
+      pwords "Split failed:" ++ pwords s
 
 instance PrettyTCM Call where
     prettyTCM c = case c of
@@ -694,7 +881,7 @@ instance PrettyTCM Call where
 	    pwords "when checking the let binding" ++ [P.prettyA b]
 	InferExpr e _ -> fsep $
 	    pwords "when inferring the type of" ++ [prettyA e]
-	CheckExpr e t _ -> fsep $
+	CheckExprCall e t _ -> fsep $
 	    pwords "when checking that the expression"
 	    ++ [prettyA e] ++ pwords "has type" ++ [prettyTCM t]
 	IsTypeCall e s _ -> fsep $
@@ -710,7 +897,7 @@ instance PrettyTCM Call where
 	    fsep $ pwords "when checking the definition of" ++ [prettyTCM x]
 	CheckDataDef _ x ps cs _ ->
 	    fsep $ pwords "when checking the definition of" ++ [prettyTCM x]
-	CheckConstructor d _ _ (A.Axiom _ _ c _) _ -> fsep $
+	CheckConstructor d _ _ (A.Axiom _ _ _ c _) _ -> fsep $
 	    pwords "when checking the constructor" ++ [prettyTCM c] ++
 	    pwords "in the declaration of" ++ [prettyTCM d]
 	CheckConstructor _ _ _ _ _ -> __IMPOSSIBLE__
@@ -745,7 +932,7 @@ instance PrettyTCM Call where
 		   pwords "in the definition of" ++ [pretty x]
         NoHighlighting _ -> empty
 	SetRange r _ ->
-	    fsep $ pwords "when doing something at" ++ [text $ show r]
+	    fsep (pwords "when doing something at") <+> prettyTCM r
         CheckSectionApplication _ m1 modapp _ -> fsep $
           pwords "when checking the module application" ++
           [prettyA $ A.Apply info m1 modapp Map.empty Map.empty]
@@ -753,6 +940,44 @@ instance PrettyTCM Call where
             info = A.ModuleInfo noRange noRange Nothing Nothing Nothing
 
 	where
-	    hPretty a = pretty =<< abstractToConcreteCtx (hiddenArgumentCtx (argHiding a)) a
+            hPretty :: I.Arg (Named_ Expr) -> TCM Doc
+            hPretty a = do
+                info <- reify $ argInfo a
+                pretty =<< (abstractToConcreteCtx (hiddenArgumentCtx (getHiding a))
+                         $ Common.Arg info $ unArg a)
 
 	    simpleDecl = D.notSoNiceDeclaration
+
+---------------------------------------------------------------------------
+-- * Natural language
+---------------------------------------------------------------------------
+
+class Verbalize a where
+  verbalize :: a -> String
+
+instance Verbalize Hiding where
+  verbalize h =
+    case h of
+      Hidden    -> "hidden"
+      NotHidden -> "visible"
+      Instance  -> "instance"
+
+instance Verbalize Relevance where
+  verbalize r =
+    case r of
+      Relevant   -> "relevant"
+      Irrelevant -> "irrelevant"
+      NonStrict  -> "shape-irrelevant"
+      Forced     -> __IMPOSSIBLE__
+      UnusedArg  -> __IMPOSSIBLE__
+
+-- | Indefinite article.
+data Indefinite a = Indefinite a
+
+instance Verbalize a => Verbalize (Indefinite a) where
+  verbalize (Indefinite a) =
+    case verbalize a of
+      "" -> ""
+      w@(c:cs) | c `elem` ['a','e','i','o'] -> "an " ++ w
+               | otherwise                  -> "a " ++ w
+      -- Aarne Ranta would whip me if he saw this.
