@@ -1,9 +1,12 @@
-{-# LANGUAGE CPP #-}
+-- {-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE TypeSynonymInstances #-}
 
 -- | Computing the free variables of a term.
 module Agda.TypeChecking.Free
     ( FreeVars(..)
-    , Free
+    , Free(..)
+    , FreeConf(..), IgnoreSorts(..)
     , freeVars
     , allVars
     , relevantVars
@@ -18,12 +21,9 @@ module Agda.TypeChecking.Free
 import qualified Agda.Utils.VarSet as Set
 import Agda.Utils.VarSet (VarSet)
 
-import Agda.Syntax.Common
+import Agda.Syntax.Common hiding (Arg, Dom, NamedArg)
 import Agda.Syntax.Internal
 
-#include "../undefined.h"
-import Agda.Utils.Impossible
-
 -- | The distinction between rigid and strongly rigid occurrences comes from:
 --   Jason C. Reed, PhD thesis, 2009, page 96 (see also his LFMTP 2009 paper)
 --
@@ -200,13 +200,20 @@ instance Free a => Free (Maybe a) where
 instance (Free a, Free b) => Free (a,b) where
   freeVars' conf (x,y) = freeVars' conf x `union` freeVars' conf y
 
+instance Free a => Free (Elim' a) where
+  freeVars' conf (Apply a) = freeVars' conf a
+  freeVars' conf (Proj{} ) = empty
+
 instance Free a => Free (Arg a) where
-  freeVars' conf (Arg h Irrelevant a) = irrelevantly $ freeVars' conf a
-  freeVars' conf (Arg h UnusedArg  a) = unused $ freeVars' conf a
-  freeVars' conf (Arg h r          a) = freeVars' conf a
+  freeVars' conf a = f $ freeVars' conf $ unArg a
+    where f = case getRelevance a of
+               Irrelevant -> irrelevantly
+               UnusedArg  -> unused
+               _          -> id
+
 
 instance Free a => Free (Dom a) where
-  freeVars' conf (Dom h r a) = freeVars' conf a
+  freeVars' conf = freeVars' conf . unDom
 
 instance Free a => Free (Abs a) where
   freeVars' conf (Abs   _ b) = subtractFV 1 $ delete 0 $ freeVars' conf b
