@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE PatternGuards #-}
 
 {-| Functions for inserting implicit arguments at the right places.
 -}
@@ -8,37 +9,43 @@ import Control.Applicative
 import Control.Monad
 
 import Agda.Syntax.Common
-import Agda.Syntax.Internal
+import Agda.Syntax.Internal as I
+import qualified Agda.Syntax.Abstract as A (NamedArg)
 
 import Agda.TypeChecking.Irrelevance
 import {-# SOURCE #-} Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
-import Agda.TypeChecking.Telescope
 import {-# SOURCE #-} Agda.TypeChecking.InstanceArguments
 
-import Agda.Utils.Size
 import Agda.Utils.Tuple
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
 
-
 -- | @implicitArgs n expand t@ generates up to @n@ implicit arguments
 --   metas (unbounded if @n<0@), as long as @t@ is a function type
 --   and @expand@ holds on the hiding info of its domain.
 implicitArgs :: Int -> (Hiding -> Bool) -> Type -> TCM (Args, Type)
-implicitArgs 0 expand t0 = return ([], t0)
-implicitArgs n expand t0 = do
+implicitArgs n expand t = mapFst (map (fmap namedThing)) <$> do
+  implicitNamedArgs n (\ h x -> expand h) t
+
+-- | @implicitNamedArgs n expand t@ generates up to @n@ named implicit arguments
+--   metas (unbounded if @n<0@), as long as @t@ is a function type
+--   and @expand@ holds on the hiding and name info of its domain.
+implicitNamedArgs :: Int -> (Hiding -> ArgName -> Bool) -> Type -> TCM (NamedArgs, Type)
+implicitNamedArgs 0 expand t0 = return ([], t0)
+implicitNamedArgs n expand t0 = do
     t0' <- reduce t0
     case ignoreSharing $ unEl t0' of
-      Pi (Dom h rel a) b | expand h -> do
-          when (h == Instance) $ reportSLn "tc.term.args.ifs" 15 $
-            "inserting implicit meta for type " ++ show a
-          v  <- applyRelevanceToContext rel $ newMeta h (absName b) a
-          let arg = Arg h rel v
-          mapFst (arg:) <$> implicitArgs (n-1) expand (absApp b v)
+      Pi (Dom info a) b | let x = absName b, expand (getHiding info) x -> do
+          when (getHiding info == Instance) $ reportSLn "tc.term.args.ifs" 15 $
+            "inserting instance meta for type " ++ show a
+          v  <- applyRelevanceToContext (getRelevance info) $
+                newMeta (getHiding info) (argNameToString x) a
+          let narg = Arg info (Named (Just $ unranged x) v)
+          mapFst (narg :) <$> implicitNamedArgs (n-1) expand (absApp b v)
       _ -> return ([], t0')
   where
     newMeta Hidden   = newNamedValueMeta RunMetaOccursCheck
@@ -70,7 +77,7 @@ introImplicits expand t = do
 data ImplicitInsertion
       = ImpInsert [Hiding]	  -- ^ this many implicits have to be inserted
       | BadImplicits	  -- ^ hidden argument where there should have been a non-hidden arg
-      | NoSuchName String -- ^ bad named argument
+      | NoSuchName ArgName -- ^ bad named argument
       | NoInsertNeeded
   deriving (Show)
 
@@ -79,24 +86,25 @@ impInsert [] = NoInsertNeeded
 impInsert hs = ImpInsert hs
 
 -- | The list should be non-empty.
-insertImplicit :: NamedArg e -> [Arg String] -> ImplicitInsertion
+insertImplicit :: A.NamedArg e -> [I.Arg ArgName] -> ImplicitInsertion
 insertImplicit _ [] = __IMPOSSIBLE__
-insertImplicit a ts | argHiding a == NotHidden = impInsert $ nofHidden ts
+insertImplicit a ts | notHidden a = impInsert $ nofHidden ts
   where
-    nofHidden :: [Arg a] -> [Hiding]
-    nofHidden = takeWhile (NotHidden /=) . map argHiding
+    nofHidden :: [I.Arg a] -> [Hiding]
+    nofHidden = takeWhile (NotHidden /=) . map getHiding
 insertImplicit a ts =
   case nameOf (unArg a) of
-    Nothing -> maybe BadImplicits impInsert $ upto (argHiding a) $ map argHiding ts
-    Just x  -> find [] x (argHiding a) ts
+    Nothing -> maybe BadImplicits impInsert $ upto (getHiding a) $ map getHiding ts
+    Just x  -> find [] (rangedThing x) (getHiding a) ts
   where
     upto h [] = Nothing
     upto h (NotHidden:_) = Nothing
     upto h (h':_) | h == h' = Just []
     upto h (h':hs) = (h':) <$> upto h hs
-    find _ x _ (Arg NotHidden _ _ : _) = NoSuchName x
-    find hs x hidingx (Arg hidingy r y : ts)
-      | x == y && hidingx == hidingy = impInsert $ reverse hs
-      | x == y && hidingx /= hidingy = BadImplicits
-      | otherwise = find (hidingy:hs) x hidingx ts
+    find :: [Hiding] -> ArgName -> Hiding -> [I.Arg ArgName] -> ImplicitInsertion
+    find _ x _ (a@(Arg{}) : _) | notHidden a = NoSuchName x
+    find hs x hidingx (a@(Arg _ y) : ts)
+      | x == y && hidingx == getHiding a = impInsert $ reverse hs
+      | x == y && hidingx /= getHiding a = BadImplicits
+      | otherwise = find (getHiding a:hs) x hidingx ts
     find i x _ []			     = NoSuchName x
