@@ -4,16 +4,14 @@ module Agda.TypeChecking.Injectivity where
 
 import Prelude hiding (mapM)
 import Control.Applicative
-import Control.Monad hiding (mapM)
-import Control.Monad.Error hiding (mapM)
-import Control.Monad.State hiding (mapM)
-import Control.Monad.Reader hiding (mapM)
+import Control.Monad.Error hiding (mapM, forM)
+import Control.Monad.State hiding (mapM, forM)
+import Control.Monad.Reader hiding (mapM, forM)
 import Data.Map (Map)
 import qualified Data.Map as Map
 import qualified Data.Set as Set
 import Data.Maybe
-import Data.List
-import Data.Traversable
+import Data.Traversable hiding (for)
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
@@ -21,62 +19,45 @@ import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Primitive
-import Agda.TypeChecking.MetaVars
+import {-# SOURCE #-} Agda.TypeChecking.MetaVars
 import {-# SOURCE #-} Agda.TypeChecking.Conversion
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Polarity
 import Agda.Utils.List
-import Agda.Utils.Monad
+import Agda.Utils.Functor (for)
 import Agda.Utils.Permutation
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
 
--- | Reduce simple (single clause) definitions.
-reduceHead :: Term -> TCM (Blocked Term)
-reduceHead v = ignoreAbstractMode $ do
-  -- first, possibly rewrite literal v to constructor form
-  v <- constructorForm v
-  reportSDoc "tc.inj.reduce" 30 $ text "reduceHead" <+> prettyTCM v
-  case ignoreSharing v of
-    Def f args -> do
-      let v0 = Def f []
-      def <- theDef <$> getConstInfo f
-      case def of
-        -- Andreas, 2012-11-06 unfold aliases (single clause terminating functions)
-        -- see test/succeed/Issue747
-        -- We restrict this to terminating functions to not make the
-        -- type checker loop here on non-terminating functions.
-        -- see test/fail/TerminationInfiniteRecord
-        Function{ funClauses = [ _ ], funDelayed = NotDelayed, funTerminates = Just True }
-                                        -> unfoldDefinition False reduceHead v0 f args
-        Datatype{ dataClause = Just _ } -> unfoldDefinition False reduceHead v0 f args
-        Record{ recClause = Just _ }    -> unfoldDefinition False reduceHead v0 f args
-        _                               -> return $ notBlocked v
-    _ -> return $ notBlocked v
-
 headSymbol :: Term -> TCM (Maybe TermHead)
-headSymbol v = ignoreAbstractMode $ do
+headSymbol v = do -- ignoreAbstractMode $ do
+  -- Andreas, 2013-02-18 ignoreAbstractMode leads to information leakage
+
   v <- ignoreBlocking <$> reduceHead v
   case ignoreSharing v of
     Def f _ -> do
-      def <- theDef <$> getConstInfo f
+      def <- theDef <$> do ignoreAbstractMode $ getConstInfo f
+        -- Andreas, 2013-02-18
+        -- if we do not ignoreAbstractMode here, abstract Functions get turned
+        -- into Axioms, but we want to distinguish these.
       case def of
-        Datatype{}  -> return (Just $ ConHead f)
-        Record{}    -> return (Just $ ConHead f)
+        Datatype{}  -> return (Just $ ConsHead f)
+        Record{}    -> return (Just $ ConsHead f)
         Axiom{}     -> do
+          reportSLn "tc.inj.axiom" 50 $ "headSymbol: " ++ show f ++ " is an Axiom."
           -- Don't treat axioms in the current mutual block
           -- as constructors (they might have definitions we
           -- don't know about yet).
           fs <- lookupMutualBlock =<< currentOrFreshMutualBlock
           if Set.member f fs
             then return Nothing
-            else return (Just $ ConHead f)
+            else return (Just $ ConsHead f)
         Function{}    -> return Nothing
         Primitive{}   -> return Nothing
         Constructor{} -> __IMPOSSIBLE__
-    Con c _ -> return (Just $ ConHead c)
+    Con c _ -> return (Just $ ConsHead $ conName c)
     Sort _  -> return (Just SortHead)
     Pi _ _  -> return (Just PiHead)
     Lit _   -> return Nothing -- handle literal heads as well? can't think of
@@ -90,19 +71,25 @@ headSymbol v = ignoreAbstractMode $ do
 
 checkInjectivity :: QName -> [Clause] -> TCM FunctionInverse
 checkInjectivity f cs
-  | pointLess cs = return NotInjective
+  | pointLess cs = do
+      reportSLn "tc.inj.check" 20 $ "Injectivity of " ++ show f ++ " would be pointless."
+      return NotInjective
   where
     -- Is it pointless to use injectivity for this function?
     pointLess []      = True
     pointLess (_:_:_) = False
-    pointLess [Clause{clausePats = ps}] = all (noMatch . unArg) ps
-      where noMatch ConP{} = False
-            noMatch LitP{} = False
-            noMatch VarP{} = True
-            noMatch DotP{} = True
+    pointLess [cl] = not $ any (properlyMatching . unArg) $ clausePats cl
+        -- Andreas, 2014-06-12
+        -- If we only have record patterns, it is also pointless.
+        -- We need at least one proper match.
 checkInjectivity f cs = do
   reportSLn "tc.inj.check" 40 $ "Checking injectivity of " ++ show f
-  es <- concat <$> mapM entry cs
+  -- Extract the head symbol of the rhs of each clause (skip absurd clauses)
+  es <- catMaybes <$> do
+    forM cs $ \ c -> do             -- produces a list ...
+      forM (getBody c) $ \ v -> do  -- ... of maybes
+        h <- headSymbol v
+        return (h, c)
   let (hs, ps) = unzip es
   reportSLn "tc.inj.check" 40 $ "  right hand sides: " ++ show hs
   if all isJust hs && distinct hs
@@ -110,37 +97,28 @@ checkInjectivity f cs = do
       let inv = Map.fromList (map fromJust hs `zip` ps)
       reportSLn "tc.inj.check" 20 $ show f ++ " is injective."
       reportSDoc "tc.inj.check" 30 $ nest 2 $ vcat $
-        map (\ (h, c) -> text (show h) <+> text "-->" <+>
-                          fsep (punctuate comma $ map (text . show) $ clausePats c)
-            ) $ Map.toList inv
+        for (Map.toList inv) $ \ (h, c) ->
+          text (show h) <+> text "-->" <+>
+          fsep (punctuate comma $ map (prettyTCM . unArg) $ clausePats c)
       return $ Inverse inv
     else return NotInjective
-  where
-    entry c = do
-      mv <- rhs (clauseBody c)
-      case mv of
-        Nothing -> return []
-        Just v  -> do
-          h <- headSymbol v
-          return [(h, c)]
 
-    rhs (Bind b)   = underAbstraction_ b rhs
-    rhs (Body v)   = return $ Just v
-    rhs NoBody     = return Nothing
-
--- | Argument should be on weak head normal form.
+-- | Argument should be in weak head normal form.
 functionInverse :: Term -> TCM InvView
 functionInverse v = case ignoreSharing v of
-  Def f args -> do
+  Def f es -> do
     d <- theDef <$> getConstInfo f
     case d of
       Function{ funInv = inv } -> case inv of
         NotInjective  -> return NoInv
-        Inverse m     -> return $ Inv f args m
+        Inverse m     -> return $ Inv f es m
+          -- NB: Invertible functions are never classified as
+          --     projection-like, so this is fine, we are not
+          --     missing parameters.  (Andreas, 2013-11-01)
       _ -> return NoInv
   _ -> return NoInv
 
-data InvView = Inv QName Args (Map TermHead Clause)
+data InvView = Inv QName [Elim] (Map TermHead Clause)
              | NoInv
 
 useInjectivity :: Comparison -> Type -> Term -> Term -> TCM ()
@@ -152,6 +130,12 @@ useInjectivity cmp a u v = do
   uinv <- functionInverse u
   vinv <- functionInverse v
   case (uinv, vinv) of
+    -- Andreas, Francesco, 2014-06-12:
+    -- We know that one of u,v is neutral
+    -- (see calls to useInjectivity in Conversion.hs).
+    -- Otherwise, (e.g. if both were Blocked), the following case would be
+    -- unsound, since it assumes the arguments to be pointwise equal.
+    -- It would deliver non-unique solutions for metas.
     (Inv f fArgs _, Inv g gArgs _)
       | f == g    -> do
         a <- defType <$> getConstInfo f
@@ -163,7 +147,7 @@ useInjectivity cmp a u v = do
           , nest 2 $ text "and type" <+> prettyTCM a
           ]
         pol <- getPolarity' cmp f
-        compareArgs pol a (Def f []) fArgs gArgs
+        compareElims pol a (Def f []) fArgs gArgs
       | otherwise -> fallBack
     (Inv f args inv, NoInv) -> do
       a <- defType <$> getConstInfo f
@@ -185,13 +169,13 @@ useInjectivity cmp a u v = do
   where
     fallBack = addConstraint $ ValueCmp cmp a u v
 
-    invert :: Term -> QName -> Type -> Map TermHead Clause -> Args -> Maybe TermHead -> TCM ()
+    invert :: Term -> QName -> Type -> Map TermHead Clause -> [Elim] -> Maybe TermHead -> TCM ()
     invert _ _ a inv args Nothing  = fallBack
     invert org f ftype inv args (Just h) = case Map.lookup h inv of
       Nothing -> typeError $ UnequalTerms cmp u v a
-      Just (Clause{ clauseTel  = tel
-                  , clausePerm = perm
-                  , clausePats = ps }) -> do -- instArgs args ps
+      Just cl@Clause{ clauseTel  = tel
+                    , clausePerm = perm } -> do
+          let ps = clausePats cl
           -- These are what dot patterns should be instantiated at
           ms <- map unArg <$> newTelMeta tel
           reportSDoc "tc.inj.invert" 20 $ vcat
@@ -203,7 +187,7 @@ useInjectivity cmp a u v = do
           -- and this is the order the variables occur in the patterns
           let ms' = permute (invertP $ compactP perm) ms
           let sub = parallelS (reverse ms)
-          margs <- runReaderT (evalStateT (metaArgs ps) ms') sub
+          margs <- runReaderT (evalStateT (mapM metaElim ps) ms') sub
           reportSDoc "tc.inj.invert" 20 $ vcat
             [ text "inversion"
             , nest 2 $ vcat
@@ -219,7 +203,7 @@ useInjectivity cmp a u v = do
           -- The clause might not give as many patterns as there
           -- are arguments (point-free style definitions).
           let args' = take (length margs) args
-          compareArgs pol ftype org margs args'
+          compareElims pol ftype org margs args'
 {- Andreas, 2011-05-09 allow unsolved constraints as long as progress
           unless (null cs) $ do
             reportSDoc "tc.inj.invert" 30 $
@@ -257,12 +241,13 @@ useInjectivity cmp a u v = do
       sub <- ask
       return $ applySubst sub v
 
-    metaArgs args = mapM metaArg args
-    metaArg arg = traverse metaPat arg
+    metaElim (Arg _ (ProjP p)) = return $ Proj p
+    metaElim (Arg info p)         = Apply . Arg info <$> metaPat p
 
-    metaPat (DotP v) = dotP v
-    metaPat (VarP _) = nextMeta
-    metaPat (ConP c mt args) = do
-      args <- metaArgs args
-      return $ Con c args
-    metaPat (LitP l) = return $ Lit l
+    metaArgs args = mapM (traverse $ metaPat . namedThing) args
+
+    metaPat (DotP v)         = dotP v
+    metaPat (VarP _)         = nextMeta
+    metaPat (ConP c mt args) = Con c <$> metaArgs args
+    metaPat (LitP l)         = return $ Lit l
+    metaPat ProjP{}          = __IMPOSSIBLE__
