@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP #-}
+-- {-# LANGUAGE CPP #-}
 
 {-| Irrelevant function types.
 -}
@@ -11,12 +11,11 @@ import qualified Data.Map as Map
 
 import Agda.Interaction.Options hiding (tests)
 
-import Agda.Syntax.Common
+import Agda.Syntax.Common hiding (Arg, Dom, NamedArg, ArgInfo)
+import Agda.Syntax.Internal (Dom)
 
 import Agda.TypeChecking.Monad
 
-#include "../undefined.h"
-import Agda.Utils.Impossible
 import Agda.Utils.QuickCheck
 import Agda.Utils.TestHelpers
 
@@ -89,26 +88,17 @@ nonStrictToIrr rel       = rel
 -- | Prepare parts of a parameter telescope for abstraction in constructors
 --   and projections.
 hideAndRelParams :: Dom a -> Dom a
-hideAndRelParams a = a
-  { domRelevance = nonStrictToIrr (domRelevance a)
-  , domHiding    = Hidden
-  }
-
-{- UNUSED
--- | @modifyArgRelevance f arg@ applies @f@ to the 'argRelevance' component of @arg@.
-modifyArgRelevance :: (Relevance -> Relevance) -> Arg a -> Arg a
-modifyArgRelevance f a = a { argRelevance = f (argRelevance a) }
--}
+hideAndRelParams = setHiding Hidden . mapRelevance nonStrictToIrr
 
 -- | Used to modify context when going into a @rel@ argument.
 inverseApplyRelevance :: Relevance -> Dom a -> Dom a
-inverseApplyRelevance rel = mapDomRelevance (rel `inverseComposeRelevance`)
+inverseApplyRelevance rel = mapRelevance (rel `inverseComposeRelevance`)
 
 -- | Compose two relevance flags.
 --   This function is used to update the relevance information
 --   on pattern variables @a@ after a match against something @rel@.
 applyRelevance :: Relevance -> Dom a -> Dom a
-applyRelevance rel = mapDomRelevance (rel `composeRelevance`)
+applyRelevance rel = mapRelevance (rel `composeRelevance`)
 
 -- * Operations on 'Context'.
 
@@ -128,7 +118,7 @@ doWorkOnTypes = verboseBracket "tc.irr" 20 "workOnTypes" . workOnTypes' True
 workOnTypes' :: Bool -> TCM a -> TCM a
 workOnTypes' allowed cont =
   if allowed then
-    liftTCM $ modifyContext (modifyContextEntries $ mapDomRelevance $ irrToNonStrict) cont
+    liftTCM $ modifyContext (modifyContextEntries $ mapRelevance $ irrToNonStrict) cont
    else cont
 
 -- | (Conditionally) wake up irrelevant variables and make them relevant.
