@@ -1,21 +1,21 @@
 {-# LANGUAGE CPP #-}
+
 module Agda.TypeChecking.Level where
 
 import Control.Monad.Error
 import Control.Applicative
 import Data.List as List
 
-import Agda.Syntax.Position
 import Agda.Syntax.Common
-import Agda.Syntax.Literal
 import Agda.Syntax.Internal
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Reduce
+import Agda.TypeChecking.Reduce.Monad ()
 import Agda.TypeChecking.Monad.Builtin
 
-import Agda.Utils.Impossible
 #include "../undefined.h"
+import Agda.Utils.Impossible
 
 data LevelKit = LevelKit
   { lvlType  :: Term
@@ -28,6 +28,14 @@ data LevelKit = LevelKit
   , zeroName :: QName
   }
 
+-- | Get the 'primLevel as a 'Term', if present.
+mlevel :: TCM (Maybe Term)
+mlevel = liftTCM $ (Just <$> primLevel) `catchError` \_ -> return Nothing
+
+-- | Get the 'primLevel' as a 'Type'.
+levelType :: TCM Type
+levelType = El (mkType 0) <$> primLevel
+
 levelSucFunction :: TCM (Term -> Term)
 levelSucFunction = do
   suc <- primLevelSuc
@@ -66,8 +74,9 @@ unLevel (Level l)  = reallyUnLevelView l
 unLevel (Shared p) = unLevel (derefPtr p)
 unLevel v = return v
 
-reallyUnLevelView :: Level -> TCM Term
-reallyUnLevelView nv =
+{-# SPECIALIZE reallyUnLevelView :: Level -> TCM Term #-}
+reallyUnLevelView :: MonadTCM tcm => Level -> tcm Term
+reallyUnLevelView nv = liftTCM $ do
   case nv of
     Max []              -> primLevelZero
     Max [Plus 0 a]      -> return $ unLevelAtom a
@@ -85,7 +94,7 @@ reallyUnLevelView nv =
     unPlusV zer suc (ClosedLevel n) = foldr (.) id (genericReplicate n suc) zer
     unPlusV _   suc (Plus n a)      = foldr (.) id (genericReplicate n suc) (unLevelAtom a)
 
-maybePrimCon :: TCM Term -> TCM (Maybe QName)
+maybePrimCon :: TCM Term -> TCM (Maybe ConHead)
 maybePrimCon prim = liftTCM $ do
     Con c [] <- prim
     return (Just c)
@@ -100,26 +109,37 @@ maybePrimDef prim = liftTCM $ do
 levelView :: Term -> TCM Level
 levelView a = do
   reportSLn "tc.level.view" 50 $ "{ levelView " ++ show a
-  msuc <- maybePrimCon primLevelSuc
-  mzer <- maybePrimCon primLevelZero
-  mmax <- maybePrimDef primLevelMax
+  v <- runReduceM $ levelView' a
+  reportSLn "tc.level.view" 50 $ "  view: " ++ show v ++ "}"
+  return v
+
+levelView' :: Term -> ReduceM Level
+levelView' a = do
+  msuc <- (getCon =<<) <$> getBuiltin' builtinLevelSuc
+  mzer <- (getCon =<<) <$> getBuiltin' builtinLevelZero
+  mmax <- (getDef =<<) <$> getBuiltin' builtinLevelMax
   let view a = do
-        a <- reduce a
+        a <- reduce' a
         case ignoreSharing a of
           Level l -> return l
           Con s [arg]
             | Just s == msuc -> inc <$> view (unArg arg)
           Con z []
             | Just z == mzer -> return $ closed 0
-          Def m [arg1, arg2]
+          Def m [Apply arg1, Apply arg2]
             | Just m == mmax -> levelLub <$> view (unArg arg1) <*> view (unArg arg2)
           _                  -> mkAtom a
   v <- view a
-  reportSLn "tc.level.view" 50 $ "  view: " ++ show v ++ "}"
   return v
   where
+    getCon (Con c []) = Just c
+    getCon _          = Nothing
+
+    getDef (Def f []) = Just f
+    getDef _          = Nothing
+
     mkAtom a = do
-      b <- reduceB a
+      b <- reduceB' a
       return $ case ignoreSharing <$> b of
         NotBlocked (MetaV m as) -> atom $ MetaLevel m as
         NotBlocked _            -> atom $ NeutralLevel (ignoreBlocking b)
