@@ -1,8 +1,6 @@
 
 module Agda.TypeChecking.LevelConstraints ( simplifyLevelConstraint ) where
 
-import Data.List
-
 import Agda.Syntax.Common (Nat)
 import Agda.Syntax.Internal
 import Agda.TypeChecking.Monad.Base
@@ -28,7 +26,7 @@ simplifyLevelConstraint n new old =
     -- compute the common water level
     waterLevel :: Nat
     waterLevel = maximum (n:ns)
-    -- convert deBruijn indices to deBruijn levels to
+    -- raise deBruijn indices to largest context to
     -- enable comparing constraints under different contexts
     leqs = concatMap inequalities $ zipWith raise (map (waterLevel -) ns) ls
 
