@@ -1,11 +1,8 @@
-
+{-# LANGUAGE FlexibleInstances, TypeSynonymInstances #-}
 module Agda.TypeChecking.MetaVars.Mention where
 
-import Data.Traversable
-
 import Agda.Syntax.Common
-import Agda.Syntax.Internal
-import Agda.TypeChecking.Reduce
+import Agda.Syntax.Internal as I
 import Agda.TypeChecking.Monad
 
 
@@ -55,11 +52,12 @@ instance MentionsMeta Sort where
 instance MentionsMeta t => MentionsMeta (Abs t) where
   mentionsMeta x = mentionsMeta x . unAbs
 
-instance MentionsMeta t => MentionsMeta (Arg t) where
-  mentionsMeta x (Arg h Irrelevant t) = False  -- we don't have to look inside irrelevant arguments when deciding to wake constraints
+instance MentionsMeta t => MentionsMeta (I.Arg t) where
+  mentionsMeta x a | isIrrelevant a = False
+  -- ^ we don't have to look inside irrelevant arguments when deciding to wake constraints
   mentionsMeta x a = mentionsMeta x (unArg a)
 
-instance MentionsMeta t => MentionsMeta (Dom t) where
+instance MentionsMeta t => MentionsMeta (I.Dom t) where
   mentionsMeta x = mentionsMeta x . unDom
 
 instance MentionsMeta t => MentionsMeta [t] where
@@ -107,4 +105,3 @@ instance MentionsMeta Constraint where
 
 -- instance (Ord k, MentionsMeta e) => MentionsMeta (Map k e) where
 --   mentionsMeta = traverse mentionsMeta
-
