@@ -1,4 +1,6 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE TypeSynonymInstances #-}
 
 module Agda.TypeChecking.MetaVars.Occurs where
 
@@ -11,29 +13,30 @@ import Control.Monad.State
 import Data.List
 import Data.Maybe
 import Data.Set (Set)
-import qualified Data.Set
-import Data.Traversable (traverse)
+import qualified Data.Set as Set
 
 import Agda.Syntax.Common
-import Agda.Syntax.Internal
+import Agda.Syntax.Internal as I
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Free hiding (Occurrence(..))
 import Agda.TypeChecking.Substitute
-import Agda.TypeChecking.EtaContract
-import Agda.TypeChecking.Eliminators
 import Agda.TypeChecking.Records
-import Agda.TypeChecking.Datatypes (isDataOrRecordType, DataOrRecord(..))
+import Agda.TypeChecking.Datatypes (isDataOrRecordType)
 import {-# SOURCE #-} Agda.TypeChecking.MetaVars
+-- import Agda.TypeChecking.MetaVars
 
+import Agda.Utils.Either
+import Agda.Utils.List (takeWhileJust)
+import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Permutation
 import Agda.Utils.Size
-import qualified Agda.Utils.VarSet as Set
+import qualified Agda.Utils.VarSet as VarSet
 
-import Agda.Utils.Impossible
 #include "../../undefined.h"
+import Agda.Utils.Impossible
 
 {- To address issue 585 (meta var occurrences in mutual defs)
 
@@ -79,16 +82,16 @@ modifyOccursCheckDefs f = modify $ \ st ->
 initOccursCheck :: MetaVariable -> TCM ()
 initOccursCheck mv = modifyOccursCheckDefs . const =<<
   if (miMetaOccursCheck (mvInfo mv) == DontRunMetaOccursCheck)
-   then return Data.Set.empty
-   else maybe (return Data.Set.empty) lookupMutualBlock =<< asks envMutualBlock
+   then return Set.empty
+   else maybe (return Set.empty) lookupMutualBlock =<< asks envMutualBlock
 
 -- | Is a def in the list of stuff to be checked?
 defNeedsChecking :: QName -> TCM Bool
-defNeedsChecking d = Data.Set.member d <$> gets stOccursCheckDefs
+defNeedsChecking d = Set.member d <$> gets stOccursCheckDefs
 
 -- | Remove a def from the list of defs to be looked at.
 tallyDef :: QName -> TCM ()
-tallyDef d = modifyOccursCheckDefs $ \ s -> Data.Set.delete d s
+tallyDef d = modifyOccursCheckDefs $ \ s -> Set.delete d s
 
 data OccursCtx
   = Flex          -- ^ we are in arguments of a meta
@@ -124,12 +127,17 @@ strongly :: OccursCtx -> OccursCtx
 strongly Rigid = StronglyRigid
 strongly ctx = ctx
 
+patternViolation' :: Int -> String -> TCM a
+patternViolation' n err = do
+  reportSLn "tc.meta.occurs" n err
+  patternViolation
+
 abort :: OccursCtx -> TypeError -> TCM a
 abort Top           err = typeError err
 abort StronglyRigid err = typeError err -- here, throw an uncatchable error (unsolvable constraint)
-abort Flex  _ = patternViolation -- throws a PatternErr, which leads to delayed constraint
-abort Rigid _ = patternViolation
-abort Irrel _ = patternViolation
+abort Flex          err = patternViolation' 70 (show err) -- throws a PatternErr, which leads to delayed constraint
+abort Rigid         err = patternViolation' 70 (show err)
+abort Irrel         err = patternViolation' 70 (show err)
 
 -- | Distinguish relevant and irrelevant variables in occurs check.
 type Vars = ([Nat],[Nat])
@@ -171,13 +179,13 @@ occursCheck m xs v = liftTCM $ do
                , prettyTCM =<< instantiateFull v
                ]
       MetaCannotDependOn _ _ i ->
-        ifM ((&&) <$> isSortMeta m <*> (not <$> hasUniversePolymorphism))
+        ifM (isSortMeta m `and2M` (not <$> hasUniversePolymorphism))
         ( typeError . GenericError . show =<<
           fsep [ text ("Cannot instantiate the metavariable " ++ show m ++ " to")
                , prettyTCM v
                , text "since universe polymorphism is disabled"
                ]
-        )
+        ) {- else -}
         ( typeError . GenericError . show =<<
             fsep [ text ("Cannot instantiate the metavariable " ++ show m ++ " to solution")
                  , prettyTCM v
@@ -192,43 +200,61 @@ occursCheck m xs v = liftTCM $ do
 instance Occurs Term where
   occurs red ctx m xs v = do
     v <- unfold red v
+    -- occurs' ctx $ ignoreBlocking v  -- fails test/succeed/DontPruneBlocked
     case v of
       -- Don't fail on blocked terms or metas
-      Blocked _ v  -> occurs' Flex v
-      NotBlocked v -> occurs' ctx v
+      NotBlocked v        -> occurs' ctx v
+      -- Blocked _ v@MetaV{} -> occurs' ctx v  -- does not help with issue 856
+      Blocked _ v         -> occurs' Flex v
     where
-      occurs' ctx v = case v of
-        Var i vs   -> do
-          if (i `allowedVar` xs) then Var i <$> occ (weakly ctx) vs else do
+      occurs' ctx v = do
+      reportSDoc "tc.meta.occurs" 45 $
+        text ("occursCheck " ++ show m ++ " (" ++ show ctx ++ ") of ") <+> prettyTCM v
+      reportSDoc "tc.meta.occurs" 70 $
+        nest 2 $ text $ show v
+      case v of
+        Var i es   -> do
+          if (i `allowedVar` xs) then Var i <$> occ (weakly ctx) es else do
             -- if the offending variable is of singleton type,
             -- eta-expand it away
             isST <- isSingletonType =<< typeOfBV i
             case isST of
               -- cannot decide, blocked by meta-var
-              Left mid -> patternViolation
+              Left mid -> patternViolation' 70 $ "Disallowed var " ++ show i ++ " not obviously singleton"
               -- not a singleton type
               Right Nothing -> -- abort Rigid turns this error into PatternErr
                 abort (strongly ctx) $ MetaCannotDependOn m (takeRelevant xs) i
               -- is a singleton type with unique inhabitant sv
-              Right (Just sv) -> return $ sv `apply` vs
+              Right (Just sv) -> return $ sv `applyE` es
         Lam h f	    -> Lam h <$> occ (leaveTop ctx) f
         Level l     -> Level <$> occ ctx l  -- stay in Top
         Lit l	    -> return v
-        DontCare v  -> DontCare <$> occurs red Irrel m (goIrrelevant xs) v
-        Def d vs    -> Def d <$> occDef d (leaveTop ctx) vs
+        DontCare v  -> dontCare <$> occurs red Irrel m (goIrrelevant xs) v
+        Def d es    -> Def d <$> occDef d (leaveTop ctx) es
         Con c vs    -> Con c <$> occ (leaveTop ctx) vs  -- if strongly rigid, remain so
         Pi a b	    -> uncurry Pi <$> occ (leaveTop ctx) (a,b)
         Sort s	    -> Sort <$> occ (leaveTop ctx) s
         v@Shared{}  -> updateSharedTerm (occ ctx) v
-        MetaV m' vs -> do
+        MetaV m' es -> do
             -- Check for loop
             --   don't fail hard on this, since we might still be on the top-level
             --   after some killing (Issue 442)
+            --
+            -- Andreas, 2013-02-18  Issue 795 demonstrates that a recursive
+            -- occurrence of a meta could be solved by the identity.
+            --   ? (Q A) = Q (? A)
+            -- So, do not throw an error.
+            -- I guess the error was there from times when occurrence check
+            -- was done after the "lhs=linear variables" check, but now
+            -- occurrence check comes first.
+            {-
             when (m == m') $ if ctx == Top then patternViolation else
               abort ctx $ MetaOccursInItself m'
+            -}
+            when (m == m') $ patternViolation' 50 $ "occursCheck failed: Found " ++ show m
 
             -- The arguments of a meta are in a flexible position
-            (MetaV m' <$> occurs red Flex m xs vs) `catchError` \err -> do
+            (MetaV m' <$> occurs red Flex m xs es) `catchError` \err -> do
               reportSDoc "tc.meta.kill" 25 $ vcat
                 [ text $ "error during flexible occurs check, we are " ++ show ctx
                 , text $ show err
@@ -237,13 +263,17 @@ instance Occurs Term where
                 -- On pattern violations try to remove offending
                 -- flexible occurrences (if not already in a flexible context)
                 PatternErr{} | ctx /= Flex -> do
-                      reportSLn "tc.meta.kill" 20 $
-                        "oops, pattern violation for " ++ show m'
-                      killResult <- prune m' vs (takeRelevant xs)
-                      if (killResult == PrunedEverything)
-                        -- after successful pruning, restart occurs check
-                        then occurs red ctx m xs =<< instantiate (MetaV m' vs)
-                        else throwError err
+                  reportSLn "tc.meta.kill" 20 $
+                    "oops, pattern violation for " ++ show m'
+                  -- Andreas, 2014-03-02, see issue 1070:
+                  -- Do not prune when meta is projected!
+                  -- WAS: let vs = takeWhileJust isApplyElim es
+                  caseMaybe (allApplyElims es) (throwError err) $ \ vs -> do
+                  killResult <- prune m' vs (takeRelevant xs)
+                  if (killResult == PrunedEverything)
+                    -- after successful pruning, restart occurs check
+                    then occurs red ctx m xs =<< instantiate (MetaV m' es)
+                    else throwError err
                 _ -> throwError err
         where
           occ ctx v = occurs red ctx m xs v
@@ -253,6 +283,7 @@ instance Occurs Term where
             def <- theDef <$> getConstInfo d
             whenM (defNeedsChecking d) $ do
               tallyDef d
+              reportSLn "tc.meta.occurs" 30 $ "Checking for occurrences in " ++ show d
               metaOccurs m def
             if (defIsDataOrRecord def) then (occ ctx vs) else (occ (defArgs red ctx) vs)
 
@@ -269,7 +300,7 @@ instance Occurs Term where
       Pi a b     -> metaOccurs m (a,b)
       Sort s     -> metaOccurs m s
       Shared p   -> metaOccurs m $ derefPtr p
-      MetaV m' vs | m == m' -> patternViolation
+      MetaV m' vs | m == m' -> patternViolation' 50 $ "Found occurrence of " ++ show m
                   | otherwise -> metaOccurs m vs
 
 instance Occurs QName where
@@ -277,6 +308,7 @@ instance Occurs QName where
 
   metaOccurs m d = whenM (defNeedsChecking d) $ do
     tallyDef d
+    reportSLn "tc.meta.occurs" 30 $ "Checking for occurrences in " ++ show d
     metaOccurs m . theDef =<< getConstInfo d
 
 instance Occurs Defn where
@@ -322,9 +354,9 @@ instance Occurs LevelAtom where
       MetaLevel m' args -> do
         MetaV m' args <- ignoreSharing <$> occurs red ctx m xs (MetaV m' args)
         return $ MetaLevel m' args
-      NeutralLevel v   -> NeutralLevel   <$> occurs red ctx m xs v
-      BlockedLevel m v -> BlockedLevel m <$> occurs red Flex m xs v
-      UnreducedLevel v -> UnreducedLevel <$> occurs red ctx m xs v
+      NeutralLevel v    -> NeutralLevel    <$> occurs red ctx m xs v
+      BlockedLevel m' v -> BlockedLevel m' <$> occurs red Flex m xs v
+      UnreducedLevel v  -> UnreducedLevel  <$> occurs red ctx m xs v
 
   metaOccurs m l = do
     l <- instantiate l
@@ -359,7 +391,12 @@ instance Occurs Sort where
       Prop       -> return ()
       Inf        -> return ()
 
+instance Occurs a => Occurs (Elim' a) where
+  occurs red ctx m xs e@Proj{}  = return e
+  occurs red ctx m xs (Apply a) = Apply <$> occurs red ctx m xs a
 
+  metaOccurs m (Proj{} ) = return ()
+  metaOccurs m (Apply a) = metaOccurs m a
 
 instance (Occurs a, Subst a) => Occurs (Abs a) where
   occurs red ctx m xs b@(Abs   s x) = Abs   s <$> underAbstraction_ b (occurs red ctx m (liftUnderAbs xs))
@@ -368,16 +405,16 @@ instance (Occurs a, Subst a) => Occurs (Abs a) where
   metaOccurs m (Abs   s x) = metaOccurs m x
   metaOccurs m (NoAbs s x) = metaOccurs m x
 
-instance Occurs a => Occurs (Arg a) where
-  occurs red ctx m xs (Arg h r@Irrelevant{} x) = Arg h r <$>
+instance Occurs a => Occurs (I.Arg a) where
+  occurs red ctx m xs (Arg info x) | isIrrelevant info = Arg info <$>
     occurs red Irrel m (goIrrelevant xs) x
-  occurs red ctx m xs (Arg h r              x) = Arg h r <$>
+  occurs red ctx m xs (Arg info x) = Arg info <$>
     occurs red ctx m xs x
 
   metaOccurs m a = metaOccurs m (unArg a)
 
-instance Occurs a => Occurs (Dom a) where
-  occurs red ctx m xs (Dom h r x) = Dom h r <$> occurs red ctx m xs x
+instance Occurs a => Occurs (I.Dom a) where
+  occurs red ctx m xs (Dom info x) = Dom info <$> occurs red ctx m xs x
   metaOccurs m = metaOccurs m . unDom
 
 instance (Occurs a, Occurs b) => Occurs (a,b) where
@@ -396,9 +433,15 @@ instance Occurs a => Occurs [a] where
 --   free variables are not contained in @xs@.
 --   If successful, @m'@ is solved by the new, pruned meta variable and we
 --   return @True@ else @False@.
+--
+--   Issue 1147:
+--   If any of the meta args @vs@ is matchable, e.g., is a constructor term,
+--   we cannot prune, because the offending variables could be removed by
+--   reduction for a suitable instantiation of the meta variable.
 prune :: MetaId -> Args -> [Nat] -> TCM PruneResult
-prune m' vs xs = liftTCM $ do
-  kills <- mapM (hasBadRigid xs) $ map unArg vs
+prune m' vs xs = do
+  caseEitherM (runErrorT $ mapM (hasBadRigid xs) $ map unArg vs)
+    (const $ return PrunedNothing) $ \ kills -> do
   reportSDoc "tc.meta.kill" 10 $ vcat
     [ text "attempting kills"
     , nest 2 $ vcat
@@ -418,43 +461,68 @@ prune m' vs xs = liftTCM $ do
 -}
   killArgs kills m'
 
--- | @hasBadRigid xs v = True@ iff one of the rigid variables in @v@ is not in @xs@.
+-- | @hasBadRigid xs v = Just True@ iff one of the rigid variables in @v@ is not in @xs@.
 --   Actually we can only prune if a bad variable is in the head. See issue 458.
 --   Or in a non-eliminateable position (see succeed/PruningNonMillerPattern).
-hasBadRigid :: [Nat] -> Term -> TCM Bool
-hasBadRigid xs (Var x _)    = return $ notElem x xs
-hasBadRigid xs (Lam _ v)    = hasBadRigid (0 : map (+1) xs) (absBody v)
-hasBadRigid xs (DontCare v) = hasBadRigid xs v
--- The following types of arguments cannot be eliminated by a pattern
--- match: data, record, Pi, levels, sorts
--- Thus, their offending rigid variables are bad.
-hasBadRigid xs v@(Def f vs) =
-  ifM (isJust <$> isDataOrRecordType f) (return $ vs `rigidVarsNotContainedIn` xs) $ do
-    elV <- elimView v
-    case elV of
-      VarElim x els -> return $ notElem x xs
-      _             -> return $ False
-  -- Andreas, 2012-05-03: There is room for further improvement.
-  -- We could also consider a defined f which is not blocked by a meta.
-hasBadRigid xs (Pi a b)     = return $ (a,b) `rigidVarsNotContainedIn` xs
-hasBadRigid xs (Level v)    = return $ v `rigidVarsNotContainedIn` xs
-hasBadRigid xs (Sort s)     = return $ s `rigidVarsNotContainedIn` xs
--- Since constructors can be eliminated by pattern-matching,
--- offending variables under a constructor could be removed by
--- the right instantiation of the meta variable.
--- Thus, they are not rigid.
-hasBadRigid xs Con{}        = return $ False
-hasBadRigid xs Lit{}        = return $ False -- no variables in Lit
-hasBadRigid xs MetaV{}      = return $ False -- no rigid variables under a meta
-hasBadRigid xs (Shared p)   = hasBadRigid xs (derefPtr p)
+--
+--   @hasBadRigid xs v = Nothing@ means that
+--   we cannot prune at all as one of the meta args is matchable.
+--   (See issue 1147.)
+hasBadRigid :: [Nat] -> Term -> ErrorT () TCM Bool
+hasBadRigid xs t = do
+  -- We fail if we encounter a matchable argument.
+  let failure = throwError ()
+  t <- liftTCM $ reduce t
+  case ignoreSharing t of
+    (Var x _)    -> return $ notElem x xs
+    -- Issue 1153: A lambda has to be considered matchable.
+    -- (Lam _ v)    -> hasBadRigid (0 : map (+1) xs) (absBody v)
+    (Lam _ v)    -> failure
+    (DontCare v) -> hasBadRigid xs v
+    -- The following types of arguments cannot be eliminated by a pattern
+    -- match: data, record, Pi, levels, sorts
+    -- Thus, their offending rigid variables are bad.
+    v@(Def f es) -> ifNotM (isNeutral f es) failure $ {- else -} do
+      return $ es `rigidVarsNotContainedIn` xs
+    -- Andreas, 2012-05-03: There is room for further improvement.
+    -- We could also consider a defined f which is not blocked by a meta.
+    (Pi a b)     -> return $ (a,b) `rigidVarsNotContainedIn` xs
+    (Level v)    -> return $ v `rigidVarsNotContainedIn` xs
+    (Sort s)     -> return $ s `rigidVarsNotContainedIn` xs
+    -- Since constructors can be eliminated by pattern-matching,
+    -- offending variables under a constructor could be removed by
+    -- the right instantiation of the meta variable.
+    -- Thus, they are not rigid.
+    (Con c args) -> do
+      ifM (liftTCM $ isEtaCon (conName c))
+        -- in case of a record con, we can in principle prune
+        -- (but not this argument; the meta could become a projection!)
+        (and <$> mapM (hasBadRigid xs . unArg) args)  -- not andM, we need to force the exceptions!
+        failure
+    Lit{}        -> failure -- matchable
+    MetaV{}      -> failure -- potentially matchable
+    (Shared p)   -> __IMPOSSIBLE__
+
+-- | Check whether a term @Def f es@ is finally stuck.
+--   Currently, we give only a crude approximation.
+isNeutral :: MonadTCM tcm => QName -> Elims -> tcm Bool
+isNeutral f es = liftTCM $ do
+  let yes = return True
+  def <- getConstInfo f
+  case theDef def of
+    Axiom{}    -> yes
+    Datatype{} -> yes
+    Record{}   -> yes
+    _          -> return False
+      -- TODO: more precise analysis
+      -- We need to check whether a function is stuck on a variable
+      -- (not meta variable), but the API does not help us...
 
 -- This could be optimized, by not computing the whole variable set
 -- at once, but allow early failure
 rigidVarsNotContainedIn :: Free a => a -> [Nat] -> Bool
 rigidVarsNotContainedIn v xs =
-  not $ Set.isSubsetOf
-    (rigidVars $ freeVars v)
-    (Set.fromList xs)
+  not $ rigidVars (freeVars v) `VarSet.isSubsetOf` VarSet.fromList xs
 
 
 data PruneResult
@@ -472,7 +540,8 @@ killArgs kills _
   | not (or kills) = return NothingToPrune  -- nothing to kill
 killArgs kills m = do
   mv <- lookupMeta m
-  if mvFrozen mv == Frozen then return PrunedNothing else do
+  allowAssign <- asks envAssignMetas
+  if mvFrozen mv == Frozen || not allowAssign then return PrunedNothing else do
 {- Andreas 2011-04-26, allow pruning in MetaS
   case mvJudgement mv of
     IsSort _    -> return False
@@ -514,29 +583,29 @@ killArgs kills m = do
 --   Invariant: @k'i == True@ iff @ki == True@ and pruning the @i@th argument from
 --   type @b@ is possible without creating unbound variables.
 --   @t'@ is type @t@ after pruning all @k'i==True@.
-killedType :: [(Dom (String, Type), Bool)] -> Type -> ([Arg Bool], Type)
+killedType :: [(I.Dom (ArgName, Type), Bool)] -> Type -> ([I.Arg Bool], Type)
 killedType [] b = ([], b)
-killedType ((arg@(Dom h r _), kill) : kills) b
-  | dontKill  = (Arg h r False : args, mkPi arg b') -- OLD: telePi (telFromList [arg]) b')
-  | otherwise = (Arg h r True  : args, subst __IMPOSSIBLE__ b')
+killedType ((arg@(Dom info _), kill) : kills) b
+  | dontKill  = (Arg info False : args, mkPi arg b') -- OLD: telePi (telFromList [arg]) b')
+  | otherwise = (Arg info True  : args, subst __IMPOSSIBLE__ b')
   where
     (args, b') = killedType kills b
     dontKill = not kill || 0 `freeIn` b'
 
 -- The list starts with the last argument
-performKill :: [Arg Bool] -> MetaId -> Type -> TCM ()
+performKill :: [I.Arg Bool] -> MetaId -> Type -> TCM ()
 performKill kills m a = do
   mv <- lookupMeta m
   when (mvFrozen mv == Frozen) __IMPOSSIBLE__
   let perm = Perm (size kills)
-             [ i | (i, Arg _ _ False) <- zip [0..] (reverse kills) ]
+             [ i | (i, Arg _ False) <- zip [0..] (reverse kills) ]
   m' <- newMeta (mvInfo mv) (mvPriority mv) perm
                 (HasType __IMPOSSIBLE__ a)
   -- Andreas, 2010-10-15 eta expand new meta variable if necessary
   etaExpandMetaSafe m'
-  let vars = reverse [ Arg h r (var i) | (i, Arg h r False) <- zip [0..] kills ]
-      lam b a = Lam (argHiding a) (Abs "v" b)
-      u       = foldl' lam (MetaV m' vars) kills
+  let vars = reverse [ Arg info (var i) | (i, Arg info False) <- zip [0..] kills ]
+      lam b a = Lam (argInfo a) (Abs "v" b)
+      u       = foldl' lam (MetaV m' $ map Apply vars) kills
 {- OLD CODE
       hs   = reverse [ argHiding a | a <- kills ]
       lam h b = Lam h (Abs "v" b)
