@@ -1,23 +1,20 @@
 {-# LANGUAGE CPP #-}
+
 module Agda.TypeChecking.Monad.Constraints where
 
 import Control.Arrow ((&&&))
 import Control.Applicative
 import Control.Monad.State
 import Control.Monad.Reader
-import Data.Map as Map
 import Data.List as List
 
 import Agda.TypeChecking.Monad.Base
-import Agda.TypeChecking.Monad.Signature
-import Agda.TypeChecking.Monad.Env
-import Agda.TypeChecking.Monad.State
 import Agda.TypeChecking.Monad.Closure
 import Agda.TypeChecking.Monad.Options
 import Agda.Utils.Monad
-import Agda.Utils.Impossible
 
 #include "../../undefined.h"
+import Agda.Utils.Impossible
 
 -- | Get the current problem
 currentProblem :: TCM ProblemId
@@ -145,4 +142,3 @@ nowSolvingConstraints = local $ \e -> e { envSolvingConstraints = True }
 
 isSolvingConstraints :: TCM Bool
 isSolvingConstraints = asks envSolvingConstraints
-
