@@ -1,25 +1,27 @@
-{-# LANGUAGE CPP #-}
+{-# LANGUAGE TupleSections,
+      FlexibleInstances, TypeSynonymInstances, OverlappingInstances #-}
 
 module Agda.TypeChecking.Monad.Context where
 
 import Control.Monad.Reader
+
 import Data.List hiding (sort)
 import qualified Data.Map as Map
 
-import Agda.Syntax.Concrete.Name (isNoName)
 import Agda.Syntax.Abstract.Name
-import Agda.Syntax.Common
+import Agda.Syntax.Common hiding (Arg, Dom, NamedArg, ArgInfo)
+import qualified Agda.Syntax.Common as Common
 import Agda.Syntax.Internal
-import Agda.Syntax.Scope.Base
+import Agda.Syntax.Scope.Monad (getLocalVars, setLocalVars)
+
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Monad.Open
 
-import Agda.Utils.Monad
 import Agda.Utils.Fresh
-
-#include "../../undefined.h"
-import Agda.Utils.Impossible
+import Agda.Utils.Functor
+import Agda.Utils.List ((!!!), downFrom)
+import Agda.Utils.Monad
 
 -- * Modifying the context
 
@@ -34,7 +36,7 @@ modifyContextEntries f = map (modifyContextEntry f)
 -- | Modify a 'Context' in a computation.
 {-# SPECIALIZE modifyContext :: (Context -> Context) -> TCM a -> TCM a #-}
 modifyContext :: MonadTCM tcm => (Context -> Context) -> tcm a -> tcm a
-modifyContext f = local $ \ e -> e { envContext = f (envContext e) }
+modifyContext f = local $ \e -> e { envContext = f $ envContext e }
 
 {-# SPECIALIZE mkContextEntry :: Dom (Name, Type) -> TCM ContextEntry #-}
 mkContextEntry :: MonadTCM tcm => Dom (Name, Type) -> tcm ContextEntry
@@ -52,18 +54,18 @@ inContext xs ret = do
 -- | Change to top (=empty) context.
 {-# SPECIALIZE inTopContext :: TCM a -> TCM a #-}
 inTopContext :: MonadTCM tcm => tcm a -> tcm a
-inTopContext = modifyContext $ const []
+inTopContext cont = do
+  locals <- liftTCM $ getLocalVars
+  liftTCM $ setLocalVars []
+  a <- modifyContext (const []) cont
+  liftTCM $ setLocalVars locals
+  return a
 
 -- | Delete the last @n@ bindings from the context.
 {-# SPECIALIZE escapeContext :: Int -> TCM a -> TCM a #-}
 escapeContext :: MonadTCM tcm => Int -> tcm a -> tcm a
 escapeContext n = modifyContext $ drop n
 
--- | Deprecated.
-{-# SPECIALIZE escapeContextToTopLevel :: TCM a -> TCM a #-}
-escapeContextToTopLevel :: MonadTCM tcm => tcm a -> tcm a
-escapeContextToTopLevel = modifyContext $ const []
-
 -- * Adding to the context
 
 -- | @addCtx x arg cont@ add a variable to the context.
@@ -74,17 +76,57 @@ addCtx :: MonadTCM tcm => Name -> Dom Type -> tcm a -> tcm a
 addCtx x a ret = do
   ctx <- map (nameConcrete . fst . unDom) <$> getContext
   let x' = head $ filter (notTaken ctx) $ iterate nextName x
-  ce <- mkContextEntry $ fmap ((,) x') a
+  ce <- mkContextEntry $ (x',) <$> a
   modifyContext (ce :) ret
       -- let-bindings keep track of own their context
   where
-    notTaken xs x = isNoName (nameConcrete x) || nameConcrete x `notElem` xs
+    notTaken xs x = isNoName x || nameConcrete x `notElem` xs
 
+-- | Various specializations of @addCtx@.
+{-# SPECIALIZE addContext :: b -> TCM a -> TCM a #-}
+class AddContext b where
+  addContext :: MonadTCM tcm => b -> tcm a -> tcm a
+
+instance AddContext a => AddContext [a] where
+  addContext = flip (foldr addContext)
+
+instance AddContext (Name, Dom Type) where
+  addContext = uncurry addCtx
+
+instance AddContext (Dom (Name, Type)) where
+  addContext = addContext . distributeF
+  -- addContext dom = addCtx (fst $ unDom dom) (snd <$> dom)
+
+instance AddContext ([Name], Dom Type) where
+  addContext (xs, dom) = addContext (bindsToTel' id xs dom)
+
+instance AddContext (String, Dom Type) where
+  addContext (s, dom) ret = do
+    x <- freshName_ s
+    addCtx x dom ret
+
+instance AddContext (Dom (String, Type)) where
+  addContext = addContext . distributeF
+  -- addContext dom = addContext (fst $ unDom dom, snd <$> dom)
+
+instance AddContext Name where
+  addContext x = addContext (x, dummyDom)
+
+instance AddContext String where
+  addContext s = addContext (s, dummyDom)
+
+instance AddContext Telescope where
+  addContext tel ret = loop tel where
+    loop EmptyTel          = ret
+    loop (ExtendTel t tel) = underAbstraction t tel loop
+
+{-
 -- | N-ary variant of @addCtx@.
 {-# SPECIALIZE addContext :: [Dom (Name, Type)] -> TCM a -> TCM a #-}
 addContext :: MonadTCM tcm => [Dom (Name, Type)] -> tcm a -> tcm a
 addContext ctx m =
-  foldr (\arg -> addCtx (fst $ unDom arg) (fmap snd arg)) m ctx
+  foldr (\arg -> addCtx (fst $ unDom arg) (snd <$> arg)) m ctx
+-}
 
 -- | add a bunch of variables with the same type to the context
 {-# SPECIALIZE addCtxs :: [Name] -> Dom Type -> TCM a -> TCM a #-}
@@ -104,23 +146,23 @@ addCtxString s a m = do
 addCtxString_ :: MonadTCM tcm => String -> tcm a -> tcm a
 addCtxString_ s = addCtxString s dummyDom
 
+{-# SPECIALIZE addCtxStrings_ :: [String] -> TCM a -> TCM a #-}
+addCtxStrings_ :: MonadTCM tcm => [String] -> tcm a -> tcm a
+addCtxStrings_ = flip (foldr addCtxString_)
+
 -- | Context entries without a type have this dummy type.
 dummyDom :: Dom Type
-dummyDom = Dom NotHidden Relevant $ El Prop $ Sort Prop
+dummyDom = Common.Dom defaultArgInfo $ El Prop $ Sort Prop
 
 -- | Go under an abstraction.
 {-# SPECIALIZE underAbstraction :: Subst a => Dom Type -> Abs a -> (a -> TCM b) -> TCM b #-}
 underAbstraction :: (Subst a, MonadTCM tcm) => Dom Type -> Abs a -> (a -> tcm b) -> tcm b
 underAbstraction _ (NoAbs _ v) k = k v
-underAbstraction t a k = do
-    xs <- map (nameConcrete . fst . unDom) <$> getContext
+underAbstraction t a           k = do
     x <- freshName_ $ realName $ absName a
-    let y = head $ filter (notTaken xs) $ iterate nextName x
-    addCtx y t $ k $ absBody a
+    addCtx x t $ k $ absBody a
   where
-    notTaken xs x = isNoName (nameConcrete x) || notElem (nameConcrete x) xs
-    realName "_" = "x"
-    realName s   = s
+    realName s = if isNoName s then "x" else argNameToString s
 
 -- | Go under an abstract without worrying about the type to add to the context.
 {-# SPECIALIZE underAbstraction_ :: Subst a => Abs a -> (a -> TCM b) -> TCM b #-}
@@ -130,15 +172,15 @@ underAbstraction_ = underAbstraction dummyDom
 -- | Add a telescope to the context.
 {-# SPECIALIZE addCtxTel :: Telescope -> TCM a -> TCM a #-}
 addCtxTel :: MonadTCM tcm => Telescope -> tcm a -> tcm a
-addCtxTel EmptyTel	    ret = ret
-addCtxTel (ExtendTel t tel) ret = underAbstraction t tel $ \tel -> addCtxTel tel ret
-
+addCtxTel tel ret = loop tel where
+  loop EmptyTel          = ret
+  loop (ExtendTel t tel) = underAbstraction t tel loop
 
 -- | Add a let bound variable
-{-# SPECIALIZE addLetBinding :: Relevance -> Name -> Term -> Type -> TCM a -> TCM a #-}
-addLetBinding :: MonadTCM tcm => Relevance -> Name -> Term -> Type -> tcm a -> tcm a
-addLetBinding rel x v t0 ret = do
-    let t = Dom NotHidden rel t0
+{-# SPECIALIZE addLetBinding :: ArgInfo -> Name -> Term -> Type -> TCM a -> TCM a #-}
+addLetBinding :: MonadTCM tcm => ArgInfo -> Name -> Term -> Type -> tcm a -> tcm a
+addLetBinding info x v t0 ret = do
+    let t = Common.Dom (setHiding NotHidden info) t0
     vt <- liftTCM $ makeOpen (v, t)
     flip local ret $ \e -> e { envLetBindings = Map.insert x vt $ envLetBindings e }
 
@@ -155,23 +197,20 @@ getContext = asks $ map ctxEntry . envContext
 getContextSize :: MonadTCM tcm => tcm Nat
 getContextSize = genericLength <$> asks envContext
 
--- | Generate [Var n - 1, .., Var 0] for all declarations in the context.
+-- | Generate @[var (n - 1), ..., var 0]@ for all declarations in the context.
 {-# SPECIALIZE getContextArgs :: TCM Args #-}
 getContextArgs :: MonadTCM tcm => tcm Args
-getContextArgs = do
-  ctx <- getContext
-  return $ reverse $ [ Arg h r $ var i | (Dom h r _, i) <- zip ctx [0..] ]
+getContextArgs = reverse . zipWith mkArg [0..] <$> getContext
+  where mkArg i (Common.Dom info _) = Common.Arg info $ var i
 
 {-# SPECIALIZE getContextTerms :: TCM [Term] #-}
 getContextTerms :: MonadTCM tcm => tcm [Term]
-getContextTerms = map unArg <$> getContextArgs
+getContextTerms = map var . downFrom <$> getContextSize
 
--- | Get the current context as a 'Telescope' with the specified 'Hiding'.
+-- | Get the current context as a 'Telescope'.
 {-# SPECIALIZE getContextTelescope :: TCM Telescope #-}
 getContextTelescope :: MonadTCM tcm => tcm Telescope
-getContextTelescope = foldr extTel EmptyTel . reverse <$> getContext
-  where
-    extTel (Dom h r (x, t)) = ExtendTel (Dom h r t) . Abs (show x)
+getContextTelescope = telFromList' nameToArgName . reverse <$> getContext
 
 -- | Check if we are in a compatible context, i.e. an extension of the given context.
 {-# SPECIALIZE getContextId :: TCM [CtxId] #-}
@@ -180,12 +219,17 @@ getContextId = asks $ map ctxId . envContext
 
 -- | get type of bound variable (i.e. deBruijn index)
 --
+{-# SPECIALIZE lookupBV :: Nat -> TCM (Dom (Name, Type)) #-}
+lookupBV :: MonadTCM tcm => Nat -> tcm (Dom (Name, Type))
+lookupBV n = do
+  ctx <- getContext
+  let failure = fail $ "deBruijn index out of scope: " ++ show n ++
+                       " in context " ++ show (map (fst . unDom) ctx)
+  maybe failure (return . fmap (raise $ n + 1)) $ ctx !!! n
+
 {-# SPECIALIZE typeOfBV' :: Nat -> TCM (Dom Type) #-}
 typeOfBV' :: MonadTCM tcm => Nat -> tcm (Dom Type)
-typeOfBV' n =
-    do	ctx <- getContext
-	Dom h r (_,t) <- ctx !!! n
-	return $ Dom h r $ raise (n + 1) t
+typeOfBV' n = fmap snd <$> lookupBV n
 
 {-# SPECIALIZE typeOfBV :: Nat -> TCM Type #-}
 typeOfBV :: MonadTCM tcm => Nat -> tcm Type
@@ -193,19 +237,7 @@ typeOfBV i = unDom <$> typeOfBV' i
 
 {-# SPECIALIZE nameOfBV :: Nat -> TCM Name #-}
 nameOfBV :: MonadTCM tcm => Nat -> tcm Name
-nameOfBV n =
-    do	ctx <- getContext
-	Dom _ _ (x,_) <- ctx !!! n
-	return x
-
--- | TODO: move(?)
-xs !!! n = xs !!!! n
-    where
-	[]     !!!! _ = do
-            ctx <- getContext
-            fail $ "deBruijn index out of scope: " ++ show n ++ " in context " ++ show (map (fst . unDom) ctx)
-	(x:_)  !!!! 0 = return x
-	(_:xs) !!!! n = xs !!!! (n - 1)
+nameOfBV n = fst . unDom <$> lookupBV n
 
 -- | Get the term corresponding to a named variable. If it is a lambda bound
 --   variable the deBruijn index is returned and if it is a let bound variable
@@ -218,7 +250,7 @@ getVarInfo x =
 	case findIndex ((==x) . fst . unDom) ctx of
 	    Just n -> do
                 t <- typeOfBV' n
-                return (Var n [], t)
+                return (var n, t)
 	    _	    ->
 		case Map.lookup x def of
 		    Just vt -> liftTCM $ getOpen vt
