@@ -1,7 +1,9 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleContexts #-}
 
 module Agda.TypeChecking.Monad.Options where
 
+import Control.Applicative
 import Control.Monad.Error
 import Control.Monad.Reader
 import Control.Monad.State
@@ -9,19 +11,22 @@ import Data.Maybe
 import Text.PrettyPrint
 import System.Directory
 import System.FilePath
-import System.IO
+
+import Paths_Agda (getDataFileName)
+-- NB: find Paths_Agda.hs in dist/build/autogen/
 
 import Agda.Syntax.Concrete
 import {-# SOURCE #-} Agda.TypeChecking.Errors
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.State
-import Agda.Interaction.EmacsCommand as Emacs
 import Agda.Interaction.FindFile
 import Agda.Interaction.Options
+import qualified Agda.Interaction.Options.Lenses as Lens
+import Agda.Interaction.Response
+
 import Agda.Utils.FileName
 import Agda.Utils.Monad
 import Agda.Utils.List
-import Agda.Utils.String
 import Agda.Utils.Trie (Trie)
 import qualified Agda.Utils.Trie as Trie
 
@@ -45,8 +50,8 @@ setPragmaOptions opts = do
 --
 -- Relative include directories are made absolute with respect to the
 -- current working directory. If the include directories have changed
--- (and were previously @'Right' something@), then the state is reset
--- (completely) .
+-- (thus, they are 'Left' now, and were previously @'Right' something@),
+-- then the state is reset (completely, see setIncludeDirs) .
 --
 -- An empty list of relative include directories (@'Left' []@) is
 -- interpreted as @["."]@.
@@ -57,29 +62,28 @@ setCommandLineOptions opts =
     Left err   -> __IMPOSSIBLE__
     Right opts -> do
       incs <- case optIncludeDirs opts of
-        Right is -> return is
-        Left  is -> do
-          setIncludeDirs is CurrentDir
+        Right absolutePathes -> return absolutePathes
+        Left  relativePathes -> do
+          -- setIncludeDirs makes pathes (relative to CurrentDir) absolute
+          -- and possible adds the current directory (if no pathes given)
+          setIncludeDirs relativePathes CurrentDir
           getIncludeDirs
-      modify $ \s ->
-        s { stPersistent = (stPersistent s) {
-              stPersistentOptions = opts { optIncludeDirs = Right incs }
-            }
-          , stPragmaOptions = optPragmaOptions opts
-          }
+      modify $ Lens.setCommandLineOptions opts{ optIncludeDirs = Right incs }
+             . Lens.setPragmaOptions (optPragmaOptions opts)
 
--- | Returns the pragma options which are currently in effect.
+class (Functor m, Applicative m, Monad m) => HasOptions m where
+  -- | Returns the pragma options which are currently in effect.
+  pragmaOptions      :: m PragmaOptions
+  -- | Returns the command line options which are currently in effect.
+  commandLineOptions :: m CommandLineOptions
 
-pragmaOptions :: TCM PragmaOptions
-pragmaOptions = gets stPragmaOptions
+instance MonadIO m => HasOptions (TCMT m) where
+  pragmaOptions = gets stPragmaOptions
 
--- | Returns the command line options which are currently in effect.
-
-commandLineOptions :: TCM CommandLineOptions
-commandLineOptions = do
-  p  <- stPragmaOptions <$> get
-  cl <- stPersistentOptions . stPersistent <$> get
-  return $ cl { optPragmaOptions = p }
+  commandLineOptions = do
+    p  <- stPragmaOptions <$> get
+    cl <- stPersistentOptions . stPersistent <$> get
+    return $ cl { optPragmaOptions = p }
 
 setOptionsFromPragma :: OptionsPragma -> TCM ()
 setOptionsFromPragma ps = do
@@ -111,7 +115,8 @@ dontEtaContractImplicit = local $ \e -> e { envEtaContractImplicit = False }
 doEtaContractImplicit :: MonadTCM tcm => tcm a -> tcm a
 doEtaContractImplicit = local $ \e -> e { envEtaContractImplicit = True }
 
-shouldEtaContractImplicit :: TCM Bool
+{-# SPECIALIZE shouldEtaContractImplicit :: TCM Bool #-}
+shouldEtaContractImplicit :: MonadReader TCEnv m => m Bool
 shouldEtaContractImplicit = asks envEtaContractImplicit
 
 -- | Don't reify interaction points
@@ -159,8 +164,8 @@ setIncludeDirs
   -- ^ How should relative paths be interpreted?
   -> TCM ()
 setIncludeDirs incs relativeTo = do
-  opts <- commandLineOptions
-  let oldIncs = optIncludeDirs opts
+  -- save the previous include dirs
+  oldIncs <- gets Lens.getIncludeDirs
 
   (root, check) <- case relativeTo of
     CurrentDir -> do
@@ -170,28 +175,35 @@ setIncludeDirs incs relativeTo = do
       m <- moduleName' f
       return (projectRoot f m, checkModuleName m f)
 
-  let setIncs incs = modify $ \s ->
-        s { stPersistent =
-          (stPersistent s) { stPersistentOptions =
-            (stPersistentOptions $ stPersistent s)
-              { optIncludeDirs = Right incs
-            }
-          }
-        }
+  -- Add the current dir if no include path is given
+  incs <- return $ if null incs then ["."] else incs
+  -- Make pathes absolute
+  incs <- return $  map (mkAbsolute . (filePath root </>)) incs
 
-  setIncs (map (mkAbsolute . (filePath root </>)) $
-             case incs of
-               [] -> ["."]
-               _  -> incs)
+  -- Andreas, 2013-10-30  Add default include dir
+  libdir <- liftIO $ getDataFileName ("lib")
+      -- NB: This is an absolute file name, but
+      -- Agda.Utils.FilePath wants to check absoluteness anyway.
+  let primdir = mkAbsolute $ libdir </> "prim"
+      -- We add the default dir at the end, since it is then
+      -- printed last in error messages.
+      -- Might also be useful to overwrite default imports...
+  incs <- return $ incs ++ [primdir]
+  Lens.putIncludeDirs $ Right $ incs
 
-  incs <- getIncludeDirs
+  -- Check whether the include dirs have changed.  If yes, reset state.
+  -- Andreas, 2013-10-30 comments:
+  -- The logic, namely using the include-dirs variable as a driver
+  -- for the interaction, qualifies for a code-obfuscation contest.
+  -- I guess one Boolean more in the state cost 10.000 EUR at the time
+  -- of this implementation...
   case oldIncs of
     Right incs' | incs' /= incs -> do
-      ho <- stInteractionOutputCallback <$> get
+      ho <- getInteractionOutputCallback
       resetAllState
-      modify $ \st -> st { stInteractionOutputCallback = ho }
-      setIncs incs
-    _                           -> return ()
+      setInteractionOutputCallback ho
+      Lens.putIncludeDirs $ Right incs
+    _ -> return ()
 
   check
 
@@ -215,7 +227,8 @@ hasInputFile = isJust <$> optInputFile <$> commandLineOptions
 proofIrrelevance :: TCM Bool
 proofIrrelevance = optProofIrrelevance <$> pragmaOptions
 
-hasUniversePolymorphism :: TCM Bool
+{-# SPECIALIZE hasUniversePolymorphism :: TCM Bool #-}
+hasUniversePolymorphism :: HasOptions m => m Bool
 hasUniversePolymorphism = optUniversePolymorphism <$> pragmaOptions
 
 showImplicitArguments :: TCM Bool
@@ -268,12 +281,18 @@ typeInType = not . optUniverseCheck <$> pragmaOptions
 --   reportSLn
 --   reportSDoc
 
-getVerbosity :: TCM (Trie String Int)
+-- | Retrieve the current verbosity level.
+{-# SPECIALIZE getVerbosity :: TCM (Trie String Int) #-}
+getVerbosity :: HasOptions m => m (Trie String Int)
 getVerbosity = optVerbose <$> pragmaOptions
 
 type VerboseKey = String
 
-hasVerbosity :: VerboseKey -> Int -> TCM Bool
+-- | Check whether a certain verbosity level is activated.
+--
+--   Precondition: The level must be non-negative.
+{-# SPECIALIZE hasVerbosity :: VerboseKey -> Int -> TCM Bool #-}
+hasVerbosity :: HasOptions m => VerboseKey -> Int -> m Bool
 hasVerbosity k n | n < 0     = __IMPOSSIBLE__
                  | otherwise = do
     t <- getVerbosity
@@ -281,39 +300,38 @@ hasVerbosity k n | n < 0     = __IMPOSSIBLE__
 	m  = maximum $ 0 : Trie.lookupPath ks t
     return (n <= m)
 
--- | If this command is run under the Emacs mode, then it formats the
--- debug message in such a way that the Emacs mode can understand it.
-
-emacsifyDebugMessage :: String -- ^ The debug message.
-                     -> TCM String
-emacsifyDebugMessage s =
-  ifM (envEmacs <$> ask)
-      (return $ Emacs.response $
-         L [ A "agda2-verbose"
-           , A (quote s)
-           ])
-      (return s)
-
 -- | Displays a debug message in a suitable way.
-displayDebugMessage :: String -> TCM ()
-displayDebugMessage s =
-  liftIO . putStr =<< emacsifyDebugMessage s
+{-# SPECIALIZE displayDebugMessage :: Int -> String -> TCM () #-}
+displayDebugMessage :: MonadTCM tcm
+  => Int     -- ^ The message's debug level.
+  -> String  -- ^ Message.
+  -> tcm ()
+displayDebugMessage n s = liftTCM $
+  appInteractionOutputCallback (Resp_RunningInfo n s)
 
--- | Precondition: The level must be non-negative.
-verboseS :: VerboseKey -> Int -> TCM () -> TCM ()
-verboseS k n action = whenM (hasVerbosity k n) action
+-- | Run a computation if a certain verbosity level is activated.
+--
+--   Precondition: The level must be non-negative.
+{-# SPECIALIZE verboseS :: VerboseKey -> Int -> TCM () -> TCM () #-}
+verboseS :: MonadTCM tcm => VerboseKey -> Int -> tcm () -> tcm ()
+verboseS k n action = whenM (liftTCM $ hasVerbosity k n) action
 
-reportS :: VerboseKey -> Int -> String -> TCM ()
-reportS k n s = verboseS k n $ displayDebugMessage s
+-- | Conditionally print debug string.
+{-# SPECIALIZE reportS :: VerboseKey -> Int -> String -> TCM () #-}
+reportS :: MonadTCM tcm => VerboseKey -> Int -> String -> tcm ()
+reportS k n s = liftTCM $ verboseS k n $ displayDebugMessage n s
 
-reportSLn :: VerboseKey -> Int -> String -> TCM ()
-reportSLn k n s = verboseS k n $ do
-  displayDebugMessage (s ++ "\n")
-  liftIO $ hFlush stdout
+-- | Conditionally println debug string.
+{-# SPECIALIZE reportSLn :: VerboseKey -> Int -> String -> TCM () #-}
+reportSLn :: MonadTCM tcm => VerboseKey -> Int -> String -> tcm ()
+reportSLn k n s = verboseS k n $
+  displayDebugMessage n (s ++ "\n")
 
-reportSDoc :: VerboseKey -> Int -> TCM Doc -> TCM ()
-reportSDoc k n d = verboseS k n $ do
-  displayDebugMessage . (++ "\n") . show =<< do
+-- | Conditionally render debug 'Doc' and print it.
+{-# SPECIALIZE reportSDoc :: VerboseKey -> Int -> TCM Doc -> TCM () #-}
+reportSDoc :: MonadTCM tcm => VerboseKey -> Int -> TCM Doc -> tcm ()
+reportSDoc k n d = liftTCM $ verboseS k n $ do
+  displayDebugMessage n . (++ "\n") . show =<< do
     d `catchError` \ err ->
       (\ s -> (sep $ map text
                  [ "Printing debug message"
@@ -321,12 +339,9 @@ reportSDoc k n d = verboseS k n $ do
                  , "failed due to error:" ]) $$
               (nest 2 $ text s)) <$> prettyError err
 
-
-verboseBracket :: VerboseKey -> Int -> String -> TCM a -> TCM a
-verboseBracket k n s m = do
-  v <- hasVerbosity k n
-  if not v then m
-           else do
-    displayDebugMessage $ "{ " ++ s ++ "\n"
-    x <- m `finally` displayDebugMessage "}\n"
-    return x
+-- | Print brackets around debug messages issued by a computation.
+{-# SPECIALIZE verboseBracket :: VerboseKey -> Int -> String -> TCM a -> TCM a #-}
+verboseBracket :: MonadTCM tcm => VerboseKey -> Int -> String -> TCM a -> tcm a
+verboseBracket k n s m = liftTCM $ ifNotM (hasVerbosity k n) m $ {- else -} do
+  displayDebugMessage n $ "{ " ++ s ++ "\n"
+  m `finally` displayDebugMessage n "}\n"
