@@ -1,19 +1,22 @@
-{-# LANGUAGE CPP, PatternGuards #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE PatternGuards #-}
+
 module Agda.TypeChecking.Monad.Signature where
 
-import Control.Arrow ((***))
+import Control.Applicative
 import Control.Monad.State
 import Control.Monad.Reader
-import Data.Set (Set)
-import qualified Data.Set as Set
+
+import Data.List
 import Data.Map (Map)
 import qualified Data.Map as Map
-import Data.List
-import Data.Function
+import Data.Maybe
 
 import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Common
-import Agda.Syntax.Internal
+import Agda.Syntax.Internal as I
 import Agda.Syntax.Position
 
 import qualified Agda.Compiler.JS.Parser as JS
@@ -24,15 +27,15 @@ import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Monad.Env
 import Agda.TypeChecking.Monad.Mutual
 import Agda.TypeChecking.Monad.Open
-import Agda.TypeChecking.Free (isBinderUsed)
+import Agda.TypeChecking.Monad.State
 import Agda.TypeChecking.Substitute
--- import Agda.TypeChecking.Pretty -- leads to cyclicity
 import {-# SOURCE #-} Agda.TypeChecking.CompiledClause.Compile
 import {-# SOURCE #-} Agda.TypeChecking.Polarity
 import {-# SOURCE #-} Agda.TypeChecking.ProjectionLike
 
-import Agda.Utils.Monad
 import Agda.Utils.Map as Map
+import Agda.Utils.Maybe
+import Agda.Utils.Monad
 import Agda.Utils.Size
 import Agda.Utils.Permutation
 import Agda.Utils.Pretty
@@ -69,6 +72,9 @@ withSignature sig m =
 
 -- * modifiers for parts of the signature
 
+lookupDefinition :: QName -> Signature -> Maybe Definition
+lookupDefinition q sig = HMap.lookup q $ sigDefinitions sig
+
 updateDefinition :: QName -> (Definition -> Definition) -> Signature -> Signature
 updateDefinition q f sig = sig { sigDefinitions = HMap.adjust f q (sigDefinitions sig) }
 
@@ -87,13 +93,17 @@ updateDefPolarity f def = def { defPolarity = f (defPolarity def) }
 updateDefCompiledRep :: (CompiledRepresentation -> CompiledRepresentation) -> (Definition -> Definition)
 updateDefCompiledRep f def = def { defCompiledRep = f (defCompiledRep def) }
 
+updateFunClauses :: ([Clause] -> [Clause]) -> (Defn -> Defn)
+updateFunClauses f def@Function{ funClauses = cs} = def { funClauses = f cs }
+updateFunClauses f _                              = __IMPOSSIBLE__
+
 -- | Add a constant to the signature. Lifts the definition to top level.
 addConstant :: QName -> Definition -> TCM ()
 addConstant q d = do
   reportSLn "tc.signature" 20 $ "adding constant " ++ show q ++ " to signature"
   tel <- getContextTelescope
-  let tel' = killRange $ case theDef d of
-	      Constructor{} -> fmap (mapDomHiding (const Hidden)) tel
+  let tel' = replaceEmptyName "r" $ killRange $ case theDef d of
+	      Constructor{} -> fmap (setHiding Hidden) tel
 	      _		    -> tel
   let d' = abstract tel' $ d { defName = q }
   reportSLn "tc.signature" 30 $ "lambda-lifted definition = " ++ show d'
@@ -111,12 +121,29 @@ setTerminates q b = modifySignature $ updateDefinition q $ updateTheDef $ setT
     setT def@Function{} = def { funTerminates = Just b }
     setT def            = def
 
+-- | Modify the clauses of a function.
+modifyFunClauses :: QName -> ([Clause] -> [Clause]) -> TCM ()
+modifyFunClauses q f =
+  modifySignature $ updateDefinition q $ updateTheDef $ updateFunClauses f
+
+-- | Lifts clauses to the top-level and adds them to definition.
+addClauses :: QName -> [Clause] -> TCM ()
+addClauses q cls = do
+  tel <- getContextTelescope
+  modifyFunClauses q (++ abstract tel cls)
+
 addHaskellCode :: QName -> HaskellType -> HaskellCode -> TCM ()
 addHaskellCode q hsTy hsDef = modifySignature $ updateDefinition q $ updateDefCompiledRep $ addHs
   -- TODO: sanity checking
   where
     addHs crep = crep { compiledHaskell = Just $ HsDefn hsTy hsDef }
 
+addHaskellExport :: QName -> HaskellType -> String -> TCM ()
+addHaskellExport q hsTy hsName = modifySignature $ updateDefinition q $ updateDefCompiledRep $ addHs
+  -- TODO: sanity checking
+  where
+    addHs crep = crep { exportHaskell = Just (HsExport hsTy hsName)}
+
 addHaskellType :: QName -> HaskellType -> TCM ()
 addHaskellType q hsTy = modifySignature $ updateDefinition q $ updateDefCompiledRep $ addHs
   -- TODO: sanity checking
@@ -169,23 +196,27 @@ lookupSection m = do
 -- Add display forms to all names @xn@ such that @x = x1 es1@, ... @xn-1 = xn esn@.
 addDisplayForms :: QName -> TCM ()
 addDisplayForms x = do
+  def  <- getConstInfo x
   args <- getContextArgs
+{- OLD
   n    <- do
     proj <- isProjection x
     return $ case proj of
       Just (_, n) -> n
       Nothing     -> 0
-  add (drop (n - 1) args) x x []
+-}
+  add (drop (projectionArgs $ theDef def) args) x x []
   where
     add args top x vs0 = do
       def <- getConstInfo x
       let cs = defClauses def
       case cs of
-	[ Clause{ clausePats = pats, clauseBody = b } ]
-	  | all (isVar . unArg) pats
-          , Just (m, Def y vs) <- strip (b `apply` vs0) -> do
-	      let ps = raise 1 (map unArg vs)
-                  df = Display 0 ps $ DTerm $ Def top args
+	[ Clause{ namedClausePats = pats, clauseBody = b } ]
+	  | all (isVar . namedArg) pats
+          , Just (m, Def y es) <- strip (b `apply` vs0)
+          , Just vs <- mapM isApplyElim es -> do
+	      let ps = raise 1 $ map unArg vs
+                  df = Display 0 ps $ DTerm $ Def top $ map Apply args
 	      reportSLn "tc.display.section" 20 $ "adding display form " ++ show y ++ " --> " ++ show top
                                                 ++ "\n  " ++ show df
 	      addDisplayForm y df
@@ -196,14 +227,14 @@ addDisplayForms x = do
 		    _:_:_ -> "many clauses"
 		    [ Clause{ clauseBody = b } ] -> case strip b of
 		      Nothing -> "bad body"
-		      Just (m, Def y vs)
+		      Just (m, Def y es)
 			| m < length args -> "too few args"
 			| m > length args -> "too many args"
-			| otherwise	  -> "args=" ++ show args ++ " vs=" ++ show vs
+			| otherwise	  -> "args=" ++ show args ++ " es=" ++ show es
 		      Just (m, v) -> "not a def body"
 	      reportSLn "tc.display.section" 30 $ "no display form from " ++ show x ++ " because " ++ reason
 	      return ()
-    strip (Body v)   = return (0, v)
+    strip (Body v)   = return (0, unSpine v)
     strip  NoBody    = Nothing
     strip (Bind b)   = do
       (n, v) <- strip $ absBody b
@@ -212,9 +243,15 @@ addDisplayForms x = do
     isVar VarP{} = True
     isVar _      = False
 
-applySection ::
-  ModuleName -> Telescope -> ModuleName -> Args ->
-  Map QName QName -> Map ModuleName ModuleName -> TCM ()
+-- | Module application (followed by module parameter abstraction).
+applySection
+  :: ModuleName                -- ^ Name of new module defined by the module macro.
+  -> Telescope                 -- ^ Parameters of new module.
+  -> ModuleName                -- ^ Name of old module applied to arguments.
+  -> Args                      -- ^ Arguments of module application.
+  -> Map QName QName           -- ^ Imported names (given as renaming).
+  -> Map ModuleName ModuleName -- ^ Imported modules (given as renaming).
+  -> TCM ()
 applySection new ptel old ts rd rm = do
   sig  <- getSignature
   isig <- getImportedSignature
@@ -244,10 +281,19 @@ applySection new ptel old ts rd rm = do
     partOfOldM x = x `isSubModuleOf` old
     partOfOldD x = x `isInModule`    old
 
-    copyName x = maybe x id $ Map.lookup x rd
+    -- Andreas, 2013-10-29
+    -- Here, if the name x is not imported, it persists as
+    -- old, possibly out-of-scope name.
+    -- This old name may used by the case split tactic, leading to
+    -- names that cannot be printed properly.
+    -- I guess it would make sense to mark non-imported names
+    -- as such (out-of-scope) and let splitting fail if it would
+    -- produce out-of-scope constructors.
+    copyName x = Map.findWithDefault x x rd
 
     copyDef :: Args -> (QName, Definition) -> TCM ()
-    copyDef ts (x, d) = case Map.lookup x rd of
+    copyDef ts (x, d) =
+      case Map.lookup x rd of
 	Nothing -> return ()  -- if it's not in the renaming it was private and
 			      -- we won't need it
 	Just y	-> do
@@ -265,69 +311,85 @@ applySection new ptel old ts rd rm = do
         pol = defPolarity d `apply` ts
         occ = defArgOccurrences d `apply` ts
 	-- the name is set by the addConstant function
-	nd y = Defn (defRelevance d) y t pol occ [] (-1) noCompiledRep <$> def  -- TODO: mutual block?
+	nd y = Defn (defArgInfo d) y t pol occ [] (-1) noCompiledRep <$> def  -- TODO: mutual block?
         oldDef = theDef d
-	isCon = case oldDef of
-	  Constructor{} -> True
-	  _		-> False
-{- OLD
-        getOcc d = case d of
-          Function { funArgOccurrences  = os } -> os
-          Datatype { dataArgOccurrences = os } -> os
-          Record   { recArgOccurrences  = os } -> os
-          _ -> []
-        oldOcc = getOcc oldDef
+	isCon  = case oldDef of { Constructor{} -> True ; _ -> False }
+        mutual = case oldDef of { Function{funMutual = m} -> m              ; _ -> [] }
+        extlam = case oldDef of { Function{funExtLam = e} -> e              ; _ -> Nothing }
+        with   = case oldDef of { Function{funWith = w}   -> copyName <$> w ; _ -> Nothing }
+{- THIS BREAKS A LOT OF THINGS:
+        -- Andreas, 2013-10-21:
+        -- Even if we apply the record argument, we stay a projection.
+        -- This is because we may abstract the record argument later again.
+        -- See succeed/ProjectionNotNormalized.agda
+        proj   = case oldDef of
+          Function{funProjection = Just p@Projection{projIndex = n}}
+            -> Just $ p { projIndex    = n - size ts
+                        , projDropPars = projDropPars p `apply` ts
+                        }
+          _ -> Nothing
 -}
+        -- NB (Andreas, 2013-10-19):
+        -- If we apply the record argument, we are no longer a projection!
+        proj   = case oldDef of
+          Function{funProjection = Just p@Projection{projIndex = n}} | size ts < n
+            -> Just $ p { projIndex    = n - size ts
+                        , projDropPars = projDropPars p `apply` ts
+                        }
+          _ -> Nothing
+
 	def  = case oldDef of
                 Constructor{ conPars = np, conData = d } -> return $
-                  oldDef { conPars = np - size ts, conData = copyName d }
+                  oldDef { conPars = np - size ts
+                         , conData = copyName d
+                         }
                 Datatype{ dataPars = np, dataCons = cs } -> return $
-                  oldDef { dataPars = np - size ts, dataClause = Just cl, dataCons = map copyName cs
-                         {- , dataArgOccurrences = drop (length ts) oldOcc -} }
+                  oldDef { dataPars   = np - size ts
+                         , dataClause = Just cl
+                         , dataCons   = map copyName cs
+                         }
                 Record{ recPars = np, recConType = t, recTel = tel } -> return $
-                  oldDef { recPars = np - size ts, recClause = Just cl
-                         , recConType = apply t ts, recTel = apply tel ts
-                         {- , recArgOccurrences = drop (length ts) oldOcc -}
+                  oldDef { recPars    = np - size ts
+                         , recClause  = Just cl
+                         , recConType = apply t ts
+                         , recTel     = apply tel ts
                          }
 		_ -> do
                   cc <- compileClauses Nothing [cl] -- Andreas, 2012-10-07 non need for record pattern translation
                   let newDef = Function
                         { funClauses        = [cl]
-                        , funCompiled       = cc
+                        , funCompiled       = Just $ cc
                         , funDelayed        = NotDelayed
                         , funInv            = NotInjective
-{-
-                        , funPolarity       = []
-                        , funArgOccurrences = drop (length ts') oldOcc
--}
                         , funMutual         = mutual
                         , funAbstr          = ConcreteDef
                         , funProjection     = proj
                         , funStatic         = False
                         , funCopy           = True
                         , funTerminates     = Just True
+                        , funExtLam         = extlam
+                        , funWith           = with
                         }
                   reportSLn "tc.mod.apply" 80 $ "new def for " ++ show x ++ "\n  " ++ show newDef
                   return newDef
-                  where
-                    mutual = case oldDef of
-                      Function{funMutual = m} -> m
-                      _ -> []
-                    proj = case oldDef of
-                      Function{funProjection = Just (r, n)}
-                        | size ts < n -> Just (r, n - size ts)
-                      _ -> Nothing
+{-
         ts' | null ts   = []
             | otherwise = case oldDef of
-                Function{funProjection = Just (_, n)}
+                Function{funProjection = Just Projection{ projIndex = n}}
                   | n == 0       -> __IMPOSSIBLE__
                   | otherwise    -> drop (n - 1) ts
                 _ -> ts
-	cl = Clause { clauseRange = getRange $ defClauses d
-                    , clauseTel   = EmptyTel
-                    , clausePerm  = idP 0
-                    , clausePats  = []
-                    , clauseBody  = Body $ Def x ts'
+-}
+        head = case oldDef of
+                 Function{funProjection = Just Projection{ projDropPars = f}}
+                   -> f
+                 _ -> Def x []
+	cl = Clause { clauseRange     = getRange $ defClauses d
+                    , clauseTel       = EmptyTel
+                    , clausePerm      = idP 0
+                    , namedClausePats = []
+                    , clauseBody      = Body $ head `apply` ts
+                    , clauseType      = Just $ defaultArg t
                     }
 
     copySec :: Args -> (ModuleName, Section) -> TCM ()
@@ -354,7 +416,7 @@ canonicalName :: QName -> TCM QName
 canonicalName x = do
   def <- theDef <$> getConstInfo x
   case def of
-    Constructor{conSrcCon = c}                                -> return c
+    Constructor{conSrcCon = c}                                -> return $ conName c
     Record{recClause = Just (Clause{ clauseBody = body })}    -> canonicalName $ extract body
     Datatype{dataClause = Just (Clause{ clauseBody = body })} -> canonicalName $ extract body
     _                                                         -> return x
@@ -365,6 +427,12 @@ canonicalName x = do
     extract (Body _)         = __IMPOSSIBLE__
     extract (Bind b)         = extract (unAbs b)
 
+sameDef :: QName -> QName -> TCM (Maybe QName)
+sameDef d1 d2 = do
+  c1 <- canonicalName d1
+  c2 <- canonicalName d2
+  if (c1 == c2) then return $ Just c1 else return Nothing
+
 -- | Can be called on either a (co)datatype, a record type or a
 --   (co)constructor.
 whatInduction :: QName -> TCM Induction
@@ -393,56 +461,52 @@ singleConstructorType q = do
         _                          -> __IMPOSSIBLE__
     _ -> __IMPOSSIBLE__
 
--- | Lookup the definition of a name. The result is a closed thing, all free
---   variables have been abstracted over.
+class (Functor m, Applicative m, Monad m) => HasConstInfo m where
+  -- | Lookup the definition of a name. The result is a closed thing, all free
+  --   variables have been abstracted over.
+  getConstInfo :: QName -> m Definition
+
 {-# SPECIALIZE getConstInfo :: QName -> TCM Definition #-}
-getConstInfo :: MonadTCM tcm => QName -> tcm Definition
-getConstInfo q = liftTCM $ join $ pureTCM $ \st env ->
-  let defs  = sigDefinitions $ stSignature st
-      idefs = sigDefinitions $ stImports st
-      smash = (++) `on` maybe [] (:[])
-  in case smash (HMap.lookup q defs) (HMap.lookup q idefs) of
-      []  -> fail $ "Unbound name: " ++ show q ++ " " ++ showQNameId q
-      [d] -> mkAbs env d
-      ds  -> fail $ "Ambiguous name: " ++ show q
-  where
-    mkAbs env d
-      | treatAbstractly' q' env =
-        case makeAbstract d of
-          Just d	-> return d
-          Nothing	-> typeError $ NotInScope [qnameToConcrete q]
-            -- the above can happen since the scope checker is a bit sloppy with 'abstract'
-      | otherwise = return d
-      where
-        q' = case theDef d of
-          -- Hack to make abstract constructors work properly. The constructors
-          -- live in a module with the same name as the datatype, but for 'abstract'
-          -- purposes they're considered to be in the same module as the datatype.
-          Constructor{} -> dropLastModule q
-          _             -> q
 
-        dropLastModule q@QName{ qnameModule = m } =
-          q{ qnameModule = mnameFromList $ init' $ mnameToList m }
+instance HasConstInfo (TCMT IO) where
+  getConstInfo q = join $ pureTCM $ \st env ->
+    let defs  = sigDefinitions $ stSignature st
+        idefs = sigDefinitions $ stImports st
+    in case catMaybes [HMap.lookup q defs, HMap.lookup q idefs] of
+        []  -> fail $ "Unbound name: " ++ show q ++ " " ++ showQNameId q
+        [d] -> mkAbs env d
+        ds  -> fail $ "Ambiguous name: " ++ show q
+    where
+      mkAbs env d
+        | treatAbstractly' q' env =
+          case makeAbstract d of
+            Just d	-> return d
+            Nothing	-> notInScope $ qnameToConcrete q
+              -- the above can happen since the scope checker is a bit sloppy with 'abstract'
+        | otherwise = return d
+        where
+          q' = case theDef d of
+            -- Hack to make abstract constructors work properly. The constructors
+            -- live in a module with the same name as the datatype, but for 'abstract'
+            -- purposes they're considered to be in the same module as the datatype.
+            Constructor{} -> dropLastModule q
+            _             -> q
 
-        init' [] = {-'-} __IMPOSSIBLE__
-        init' xs = init xs
+          dropLastModule q@QName{ qnameModule = m } =
+            q{ qnameModule = mnameFromList $ init' $ mnameToList m }
+
+          init' [] = {-'-} __IMPOSSIBLE__
+          init' xs = init xs
+
+{-# INLINE getConInfo #-}
+{-# SPECIALIZE getConstInfo :: QName -> TCM Definition #-}
+getConInfo :: MonadTCM tcm => ConHead -> tcm Definition
+getConInfo = liftTCM . getConstInfo . conName
 
 -- | Look up the polarity of a definition.
 getPolarity :: QName -> TCM [Polarity]
 getPolarity q = defPolarity <$> getConstInfo q
 
-{- OLD
--- | Look up the polarity of a definition.
-getPolarity :: QName -> TCM [Polarity]
-getPolarity q = do
-  defn <- theDef <$> getConstInfo q
-  case defn of
-    Function{ funPolarity  = p } -> return p
-    Datatype{ dataPolarity = p } -> return p
-    Record{ recPolarity    = p } -> return p
-    _                            -> return []
--}
-
 -- | Look up polarity of a definition and compose with polarity
 --   represented by 'Comparison'.
 getPolarity' :: Comparison -> QName -> TCM [Polarity]
@@ -453,23 +517,6 @@ getPolarity' CmpLeq q = getPolarity q -- composition with Covariant is identity
 setPolarity :: QName -> [Polarity] -> TCM ()
 setPolarity q pol = modifySignature $ updateDefinition q $ updateDefPolarity $ const pol
 
-{- OLD
--- | Set the polarity of a definition.
-setPolarity :: QName -> [Polarity] -> TCM ()
-setPolarity q pol = do
-  modifySignature setP
-  where
-    setP sig = sig { sigDefinitions = HMap.adjust setPx q defs }
-      where
-	setPx def = def { theDef = setPd $ theDef def }
-        setPd d   = case d of
-          Function{} -> d { funPolarity  = pol }
-          Datatype{} -> d { dataPolarity = pol }
-          Record{}   -> d { recPolarity  = pol }
-          _          -> d
-	defs	  = sigDefinitions sig
--}
-
 -- | Return a finite list of argument occurrences.
 getArgOccurrences :: QName -> TCM [Occurrence]
 getArgOccurrences d = defArgOccurrences <$> getConstInfo d
@@ -513,20 +560,6 @@ getArgOccurrence d i = do
     _                                    -> Mixed
   where
     look i os = (os ++ repeat Mixed) !! fromIntegral i
-
-setArgOccurrences :: QName -> [Occurrence] -> TCM ()
-setArgOccurrences d os =
-  modifySignature setO
-  where
-    setO sig = sig { sigDefinitions = HMap.adjust setOx d defs }
-      where
-	setOx def = def { theDef = setOd $ theDef def }
-        setOd d   = case d of
-          Function{} -> d { funArgOccurrences  = os }
-          Datatype{} -> d { dataArgOccurrences = os }
-          Record{}   -> d { recArgOccurrences  = os }
-          _          -> d
-	defs	  = sigDefinitions sig
 -}
 
 -- | Get the mutually recursive identifiers.
@@ -560,7 +593,8 @@ getSecFreeVars m = do
   isig <- sigSections <$> getImportedSignature
   top <- currentModule
   case top `isSubModuleOf` m || top == m of
-    True  -> return $ maybe 0 secFreeVars $ Map.lookup m (Map.union sig isig)
+    True  -> return $ maybe 0 secFreeVars $
+               Map.lookup m sig <|> Map.lookup m isig
     False -> return 0
 
 -- | Compute the number of free variables of a module. This is the sum of
@@ -587,19 +621,30 @@ instantiateDef d = do
   verboseS "tc.sig.inst" 30 $ do
     ctx <- getContext
     m   <- currentModule
-    reportSLn "" 0 $ "instDef in " ++ show m ++ ": " ++ show (defName d) ++ " " ++
-			unwords (map show . take (size vs) . reverse . map (fst . unDom) $ ctx)
+    reportSLn "tc.sig.inst" 30 $
+      "instDef in " ++ show m ++ ": " ++ show (defName d) ++ " " ++
+      unwords (map show . take (size vs) . reverse . map (fst . unDom) $ ctx)
   return $ d `apply` vs
 
 -- | Give the abstract view of a definition.
 makeAbstract :: Definition -> Maybe Definition
-makeAbstract d = do def <- makeAbs $ theDef d
-		    return d { theDef = def }
-    where
-	makeAbs Datatype   {dataAbstr = AbstractDef} = Just Axiom
-	makeAbs Function   {funAbstr  = AbstractDef} = Just Axiom
-	makeAbs Constructor{conAbstr  = AbstractDef} = Nothing
-	makeAbs d                                    = Just d
+makeAbstract d =
+  case defAbstract d of
+    ConcreteDef -> return d
+    AbstractDef -> do
+      def <- makeAbs $ theDef d
+      return d { defArgOccurrences = [] -- no positivity info for abstract things!
+               , defPolarity       = [] -- no polarity info for abstract things!
+               , theDef = def
+               }
+  where
+    makeAbs Datatype   {} = Just Axiom
+    makeAbs Function   {} = Just Axiom
+    makeAbs Constructor{} = Nothing
+    -- Andreas, 2012-11-18:  Make record constructor and projections abstract.
+    makeAbs d@Record{}    = Just Axiom
+    -- Q: what about primitive?
+    makeAbs d             = Just d
 
 -- | Enter abstract mode. Abstract definition in the current module are transparent.
 inAbstractMode :: TCM a -> TCM a
@@ -613,7 +658,7 @@ inConcreteMode :: TCM a -> TCM a
 inConcreteMode = local $ \e -> e { envAbstractMode = ConcreteMode }
 
 -- | Ignore abstract mode. All abstract definitions are transparent.
-ignoreAbstractMode :: TCM a -> TCM a
+ignoreAbstractMode :: MonadReader TCEnv m => m a -> m a
 ignoreAbstractMode = local $ \e -> e { envAbstractMode = IgnoreAbstractMode,
                                        envAllowDestructiveUpdate = False }
                                        -- Allowing destructive updates when ignoring
@@ -622,8 +667,8 @@ ignoreAbstractMode = local $ \e -> e { envAbstractMode = IgnoreAbstractMode,
 -- | Check whether a name might have to be treated abstractly (either if we're
 --   'inAbstractMode' or it's not a local name). Returns true for things not
 --   declared abstract as well, but for those 'makeAbstract' will have no effect.
-treatAbstractly :: QName -> TCM Bool
-treatAbstractly q = treatAbstractly' q <$> ask
+treatAbstractly :: MonadReader TCEnv m => QName -> m Bool
+treatAbstractly q = asks $ treatAbstractly' q
 
 treatAbstractly' :: QName -> TCEnv -> Bool
 treatAbstractly' q env = case envAbstractMode env of
@@ -634,14 +679,18 @@ treatAbstractly' q env = case envAbstractMode env of
     current = envCurrentModule env
     m	    = qnameModule q
 
--- | get type of a constant
+-- | Get type of a constant, instantiated to the current context.
 typeOfConst :: QName -> TCM Type
 typeOfConst q = defType <$> (instantiateDef =<< getConstInfo q)
 
--- | get relevance of a constant
+-- | Get relevance of a constant.
 relOfConst :: QName -> TCM Relevance
 relOfConst q = defRelevance <$> getConstInfo q
 
+-- | Get colors of a constant.
+colOfConst :: QName -> TCM [Color]
+colOfConst q = defColors <$> getConstInfo q
+
 -- | The name must be a datatype.
 sortOfConst :: QName -> TCM Sort
 sortOfConst q =
@@ -651,9 +700,63 @@ sortOfConst q =
 	    _			   -> fail $ "Expected " ++ show q ++ " to be a datatype."
 
 -- | Is it the name of a record projection?
-isProjection :: QName -> TCM (Maybe (QName, Int))
-isProjection qn = do
-  def <- theDef <$> getConstInfo qn
+isProjection :: QName -> TCM (Maybe Projection)
+isProjection qn = isProjection_ . theDef <$> getConstInfo qn
+
+isProjection_ :: Defn -> Maybe Projection
+isProjection_ def =
   case def of
-    Function { funProjection = result } -> return $ result
-    _                                   -> return $ Nothing
+    Function { funProjection = result } -> result
+    _                                   -> Nothing
+
+isProperProjection :: Defn -> Bool
+isProperProjection = isJust . (projProper <=< isProjection_)
+-- isProperProjection = maybe False projProper . isProjection_
+
+-- | Number of dropped initial arguments.
+projectionArgs :: Defn -> Int
+projectionArgs = maybe 0 (pred . projIndex) . isProjection_
+
+-- | Apply a function @f@ to its first argument, producing the proper
+--   postfix projection if @f@ is a projection.
+applyDef :: QName -> I.Arg Term -> TCM Term
+applyDef f a = do
+  -- get the original projection, if existing
+  res <- (projProper =<<) <$> isProjection f
+  case res of
+    Nothing -> return $ Def f [Apply a]
+    Just f' -> return $ unArg a `applyE` [Proj f']
+
+-- | @getDefType f t@ computes the type of (possibly projection-(like))
+--   function @t@ whose first argument has type @t@.
+--   The `parameters' for @f@ are extracted from @t@.
+--   @Nothing@ if @f@ is projection(like) but
+--   @t@ is not a data/record/axiom type.
+--
+--   Precondition: @t@ is reduced.
+--
+--   See also: 'Agda.TypeChecking.Datatypes.getConType'
+getDefType :: QName -> Type -> TCM (Maybe Type)
+getDefType f t = do
+  def <- getConstInfo f
+  let a = defType def
+  -- if @f@ is not a projection (like) function, @a@ is the correct type
+  caseMaybe (isProjection_ $ theDef def) (return $ Just a) $
+    \ (Projection{ projIndex = n }) -> do
+      -- otherwise, we have to instantiate @a@ to the "parameters" of @f@
+      let npars | n == 0    = __IMPOSSIBLE__
+                | otherwise = n - 1
+      -- we get the parameters from type @t@
+      case ignoreSharing $ unEl t of
+        Def d es -> do
+          -- Andreas, 2013-10-22
+          -- we need to check this @Def@ is fully reduced.
+          -- If it is stuck due to disabled reductions
+          -- (because of failed termination check),
+          -- we will produce garbage parameters.
+          flip (ifM $ eligibleForProjectionLike d) (return Nothing) $ do
+            -- now we know it is reduced, we can safely take the parameters
+            let pars = fromMaybe __IMPOSSIBLE__ $ allApplyElims $ take npars es
+            -- pars <- maybe (return Nothing) return $ allApplyElims $ take npars es
+            return $ Just $ a `apply` pars
+        _ -> return Nothing
