@@ -2,7 +2,6 @@
 module Agda.TypeChecking.Monad.Trace where
 
 import Control.Monad.Reader
-import Control.Monad.State
 
 import Agda.Syntax.Position
 import Agda.TypeChecking.Monad.Base
@@ -34,21 +33,11 @@ traceCall mkCall m = do
   let call      = mkCall Nothing
       callRange = getRange call
   cl <- liftTCM $ buildClosure call
-  let trace = local $
-        (if interestingCall cl then
-           \e -> e { envCall = Just cl }
-         else
-           id) .
-        (if callRange /= noRange || isNoHighlighting call then
-           \e -> e { envHighlightingRange = callRange
-                   }
-         else
-           id) .
-        (if callRange /= noRange then
-           \e -> e { envRange = callRange
-                   }
-         else
-           id)
+  let trace = local $ foldr (.) id $
+        [ \e -> e { envCall = Just cl } | interestingCall cl ] ++
+        [ \e -> e { envHighlightingRange = callRange }
+          | callRange /= noRange && highlightCall call || isNoHighlighting call ] ++
+        [ \e -> e { envRange = callRange } | callRange /= noRange ]
   wrap <- ifM (do l <- envHighlightingLevel <$> ask
                   return (l == Interactive && highlightCall call))
               (do oldRange <- envHighlightingRange <$> ask
@@ -62,7 +51,7 @@ traceCall mkCall m = do
     CheckPattern _ _ _ _            -> True
     CheckLetBinding _ _             -> True
     InferExpr _ _                   -> True
-    CheckExpr _ _ _                 -> True
+    CheckExprCall _ _ _             -> True
     CheckDotPattern _ _ _           -> True
     CheckPatternShadowing _ _       -> True
     IsTypeCall _ _ _                -> True
