@@ -1,5 +1,8 @@
-{-# LANGUAGE CPP, FlexibleInstances, FlexibleContexts,
-             UndecidableInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE UndecidableInstances #-}
 
 -- | Check that a datatype is strictly positive.
 module Agda.TypeChecking.Positivity where
@@ -13,11 +16,11 @@ import qualified Data.Set as Set
 import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.List as List
-import Data.Maybe (mapMaybe)
+import Data.Maybe (mapMaybe, fromMaybe)
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common
-import Agda.Syntax.Internal
+import Agda.Syntax.Internal as I
 import Agda.TypeChecking.Datatypes (isDataOrRecordType, DataOrRecord(..))
 import Agda.TypeChecking.Records (unguardedRecord, recursiveRecord)
 import Agda.TypeChecking.Monad
@@ -25,20 +28,24 @@ import Agda.TypeChecking.Monad.Builtin (primInf, CoinductionKit(..), coinduction
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Substitute
+import Agda.TypeChecking.Telescope
 
-import Agda.Utils.Impossible
-import Agda.Utils.Permutation
 import Agda.Utils.Size
 import Agda.Utils.List
+import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.SemiRing
-import qualified Agda.Utils.Graph as Graph
-import Agda.Utils.Graph (Graph)
+import qualified Agda.Utils.Graph.AdjacencyMap as Graph
+import Agda.Utils.Graph.AdjacencyMap (Graph)
 
 #include "../undefined.h"
+import Agda.Utils.Impossible
 
 -- | Check that the datatypes in the mutual block containing the given
--- declarations are strictly positive.
+--   declarations are strictly positive.
+--
+--   Also add information about positivity and recursivity of records
+--   to the signature.
 checkStrictlyPositive :: Set QName -> TCM ()
 checkStrictlyPositive qs = disableDestructiveUpdate $ do
   -- compute the occurrence graph for qs
@@ -89,15 +96,15 @@ checkStrictlyPositive qs = disableDestructiveUpdate $ do
           setCurrentRange (getRange q) $ typeError $ GenericError (show err)
 
       -- if we find an unguarded record, mark it as such
-      (\ just noth -> maybe noth just (mhead [ how | Edge o how <- loops, o <= StrictPos ]))
-        (\ how -> do
+      case mhead [ how | Edge o how <- loops, o <= StrictPos ] of
+        Just how -> do
           reportSDoc "tc.pos.record" 5 $ sep
             [ prettyTCM q <+> text "is not guarded, because it occurs"
             , prettyTCM how
             ]
-          unguardedRecord q) $
+          unguardedRecord q
         -- otherwise, if the record is recursive, mark it as well
-        forM_ (take 1 [ how | Edge GuardPos how <- loops ]) $ \ how -> do
+        Nothing -> forM_ (take 1 [ how | Edge GuardPos how <- loops ]) $ \ how -> do
           reportSDoc "tc.pos.record" 5 $ sep
             [ prettyTCM q <+> text "is recursive, because it occurs"
             , prettyTCM how
@@ -120,12 +127,12 @@ checkStrictlyPositive qs = disableDestructiveUpdate $ do
 
     -- Set the polarity of the arguments to a definition
     setArgs g q = do
-      reportSDoc "tc.pos.args" 5 $ text "checking args of" <+> prettyTCM q
+      reportSDoc "tc.pos.args" 10 $ text "checking args of" <+> prettyTCM q
       n <- getDefArity =<< getConstInfo q
       let nArgs = maximum $ n :
                     [ i + 1 | (ArgNode q1 i) <- Set.toList $ Graph.nodes g
                     , q1 == q ]
-          findOcc i = maybe Unused id $ Graph.lookup (ArgNode q i) (DefNode q) g
+          findOcc i = fromMaybe Unused $ Graph.lookup (ArgNode q i) (DefNode q) g
           args = map findOcc [0..nArgs - 1]
       reportSDoc "tc.pos.args" 10 $ sep
         [ text "args of" <+> prettyTCM q <+> text "="
@@ -138,7 +145,7 @@ checkStrictlyPositive qs = disableDestructiveUpdate $ do
 
 getDefArity def = case theDef def of
   Function{ funClauses = cs, funProjection = proj } -> do
-    let dropped = maybe 0 (subtract 1 . snd) proj
+    let dropped = maybe 0 (subtract 1 . projIndex) proj
     subtract dropped . arity <$> instantiateFull (defType def)
   Datatype{ dataPars = n } -> return n
   Record{ recPars = n }    -> return n
@@ -343,10 +350,10 @@ class ComputeOccurrences a where
   occurrences :: a -> OccM Occurrences
 
 instance ComputeOccurrences Clause where
-  occurrences (Clause{ clausePats = ps0, clauseBody = body }) = do
-    let ps = map unArg ps0
+  occurrences cl = do
+    let ps = map unArg $ clausePats cl
     (concatOccurs (mapMaybe matching (zip [0..] ps)) >+<) <$>
-      walk (patItems ps) body
+      walk (patItems ps) (clauseBody cl)
     where
       matching (i, p)
         | properlyMatching p = Just $ occursAs Matched $ here $ AnArg i
@@ -377,8 +384,9 @@ instance ComputeOccurrences Clause where
       nVars p = case p of
         VarP{}      -> 1
         DotP{}      -> 1
-        ConP _ _ ps -> sum $ map (nVars . unArg) ps
+        ConP _ _ ps -> sum $ map (nVars . namedArg) ps
         LitP{}      -> 0
+        ProjP{}     -> 0
 
 instance ComputeOccurrences Term where
   occurrences v = case v of
@@ -439,10 +447,14 @@ instance ComputeOccurrences a => ComputeOccurrences (Abs a) where
   occurrences (Abs   _ b) = withExtendedOccEnv Nothing $ occurrences b
   occurrences (NoAbs _ b) = occurrences b
 
-instance ComputeOccurrences a => ComputeOccurrences (Arg a) where
+instance ComputeOccurrences a => ComputeOccurrences (Elim' a) where
+  occurrences Proj{}    = return $ Map.empty
+  occurrences (Apply a) = occurrences a
+
+instance ComputeOccurrences a => ComputeOccurrences (I.Arg a) where
   occurrences = occurrences . unArg
 
-instance ComputeOccurrences a => ComputeOccurrences (Dom a) where
+instance ComputeOccurrences a => ComputeOccurrences (I.Dom a) where
   occurrences = occurrences . unDom
 
 instance ComputeOccurrences a => ComputeOccurrences [a] where
@@ -466,9 +478,14 @@ computeOccurrences q = do
         mapM (getOccurrences []) cs
     Datatype{dataClause = Just c} -> getOccurrences [] =<< instantiateFull c
     Datatype{dataPars = np, dataCons = cs}       -> do
+      -- Andreas, 2013-02-27: first, each data index occurs as matched on.
+      TelV tel t <- telView $ defType def
+      let xs  = [np .. size tel - 1] -- argument positions corresponding to indices
+          ioccs = concatOccurs $ map (occursAs Matched . here . AnArg) xs
+      -- Then, we compute the occurrences in the constructor types.
       let conOcc c = do
             a <- defType <$> getConstInfo c
-            TelV tel t <- telView' <$> normalise a
+            TelV tel t <- telView' <$> normalise a -- normalization needed e.g. for test/succeed/Bush.agda
             let indices = case unEl t of
                             Def _ vs -> genericDrop np vs
                             _        -> __IMPOSSIBLE__
@@ -476,7 +493,7 @@ computeOccurrences q = do
                 vars np = map (Just . AnArg) $ downFrom np
             (>+<) <$> (occursAs (ConArgType c) <$> getOccurrences (vars np) tel')
                   <*> (occursAs (IndArgType c) . onlyVarsUpTo np <$> getOccurrences (vars $ size tel) indices)
-      concatOccurs <$> mapM conOcc cs
+      (>+<) ioccs <$> (concatOccurs <$> mapM conOcc cs)
     Record{recClause = Just c} -> getOccurrences [] =<< instantiateFull c
     Record{recPars = np, recTel = tel} -> do
       let tel' = telFromList $ genericDrop np $ telToList tel
@@ -493,16 +510,18 @@ computeOccurrences q = do
 --   This is used instead of special treatment of lambdas
 --   (which was unsound: issue 121)
 etaExpandClause :: Nat -> Clause -> Clause
-etaExpandClause n c@Clause{ clausePats = ps, clauseBody = b }
+etaExpandClause n c@Clause{ namedClausePats = ps, clauseBody = b }
   | m <= 0    = c
-  | otherwise = c { clausePats = ps ++ genericReplicate m (defaultArg $ VarP "_")
-                  , clauseBody = liftBody m b
+  | otherwise = c { namedClausePats = ps ++ genericReplicate m (defaultArg $ unnamed $ VarP underscore)
+                  , clauseBody      = liftBody m b
+                  , clauseTel       = __IMPOSSIBLE__
+                  , clausePerm      = __IMPOSSIBLE__
                   }
   where
     m = n - genericLength ps
 
     bind 0 = id
-    bind n = Bind . Abs "_" . bind (n - 1)
+    bind n = Bind . Abs underscore . bind (n - 1)
 
     vars = map (defaultArg . var) $ downFrom m
 --    vars = reverse [ defaultArg $ var i | i <- [0..m - 1] ]
@@ -533,21 +552,25 @@ instance PrettyTCM Occurrence where
   prettyTCM Mixed     = text "-[*]->"
   prettyTCM Unused    = text "-[ ]->"
 
-instance PrettyTCM n => PrettyTCM (n, Edge) where
-  prettyTCM (n, Edge o w) =
+-- | Pairing something with a node (for printing only).
+data WithNode n a = WithNode n a
+
+instance PrettyTCM n => PrettyTCM (WithNode n Edge) where
+  prettyTCM (WithNode n (Edge o w)) =
     prettyTCM o <+> prettyTCM n <+> fsep (pwords $ show w)
 
-instance PrettyTCM n => PrettyTCM (n, Occurrence) where
-  prettyTCM (n, o) = prettyTCM o <+> prettyTCM n
+instance PrettyTCM n => PrettyTCM (WithNode n Occurrence) where
+  prettyTCM (WithNode n o) = prettyTCM o <+> prettyTCM n
 
-instance (PrettyTCM n, PrettyTCM (n, e)) => PrettyTCM (Graph n e) where
+instance (PrettyTCM n, PrettyTCM (WithNode n e)) => PrettyTCM (Graph n e) where
   prettyTCM g = vcat $ map pr $ Map.assocs $ Graph.unGraph g
     where
       pr (n, es) = sep
         [ prettyTCM n
-        , nest 2 $ vcat $ map prettyTCM $ Map.assocs es
+        , nest 2 $ vcat $ map (prettyTCM . uncurry WithNode) $ Map.assocs es
         ]
 
+-- | Edge labels for the positivity graph.
 data Edge = Edge Occurrence OccursWhere
   deriving (Show)
 
