@@ -1,10 +1,16 @@
-{-# LANGUAGE CPP, TypeSynonymInstances, FlexibleInstances,
-             UndecidableInstances #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE UndecidableInstances #-}
+
 module Agda.TypeChecking.Pretty where
 
 import Control.Applicative hiding (empty)
 
-import Agda.Syntax.Common
+import Agda.Syntax.Position
+import Agda.Syntax.Common hiding (Arg, Dom, NamedArg, ArgInfo)
+import qualified Agda.Syntax.Common as Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Literal
 import Agda.Syntax.Translation.InternalToAbstract
@@ -15,12 +21,12 @@ import qualified Agda.Syntax.Abstract.Pretty as P
 import qualified Agda.Syntax.Concrete.Pretty as P
 
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Eliminators
 
 import qualified Agda.Utils.Pretty as P
-import Agda.Utils.Impossible
+import Agda.Utils.Permutation (Permutation)
 
 #include "../undefined.h"
+import Agda.Utils.Impossible
 
 ---------------------------------------------------------------------------
 -- * Wrappers for pretty printing combinators
@@ -36,6 +42,7 @@ colon      = return P.colon
 equals     = return P.equals
 pretty x   = return $ P.pretty x
 prettyA x  = P.prettyA x
+prettyAs x = P.prettyAs x
 text :: String -> TCM Doc
 text s	   = return $ P.text s
 pwords s   = map return $ P.pwords s
@@ -44,6 +51,7 @@ sep, fsep, hsep, vcat :: [TCM Doc] -> TCM Doc
 sep ds	   = P.sep <$> sequence ds
 fsep ds    = P.fsep <$> sequence ds
 hsep ds    = P.hsep <$> sequence ds
+hcat ds    = P.hcat <$> sequence ds
 vcat ds    = P.vcat <$> sequence ds
 ($$), ($+$), (<>), (<+>) :: TCM Doc -> TCM Doc -> TCM Doc
 d1 $$ d2   = (P.$$) <$> d1 <*> d2
@@ -76,13 +84,30 @@ instance PrettyTCM a => PrettyTCM (Closure a) where
 instance PrettyTCM a => PrettyTCM [a] where
   prettyTCM = prettyList . map prettyTCM
 
+instance (PrettyTCM a, PrettyTCM b) => PrettyTCM (a,b) where
+  prettyTCM (a, b) = parens $ prettyTCM a <> comma <> prettyTCM b
+
 instance PrettyTCM Nat where prettyTCM = text . show
+instance PrettyTCM Bool where prettyTCM = text . show
 instance PrettyTCM Term where prettyTCM x = prettyA =<< reify x
 instance PrettyTCM Type where prettyTCM x = prettyA =<< reify x
 instance PrettyTCM Sort where prettyTCM x = prettyA =<< reify x
 instance PrettyTCM DisplayTerm where prettyTCM x = prettyA =<< reify x
 instance PrettyTCM NamedClause where prettyTCM x = prettyA =<< reify x
 instance PrettyTCM Level where prettyTCM x = prettyA =<< reify (Level x)
+instance PrettyTCM Permutation where prettyTCM = text . show
+
+instance PrettyTCM Position where
+  prettyTCM p = do
+    text $ show p
+
+instance PrettyTCM Interval where
+  prettyTCM i = do
+    text $ show i
+
+instance PrettyTCM Range where
+  prettyTCM r = do
+    text $ show r
 
 instance PrettyTCM ClauseBody where
   prettyTCM b = do
@@ -93,7 +118,7 @@ instance PrettyTCM ClauseBody where
       walk (Body v) = (,) [] <$> prettyTCM v
       walk (Bind b) = do
         (bs, v) <- underAbstraction_ b walk
-        return (text (absName b) : bs, v)
+        return (text (argNameToString $ absName b) : bs, v)
 
 instance (PrettyTCM a, PrettyTCM b) => PrettyTCM (Judgement a b) where
   prettyTCM (HasType a t) = prettyTCM a <+> text ":" <+> prettyTCM t
@@ -108,19 +133,23 @@ instance PrettyTCM a => PrettyTCM (Blocked a) where
   prettyTCM (Blocked x a) = text "[" <+> prettyTCM a <+> text "]" <> text (show x)
   prettyTCM (NotBlocked x) = prettyTCM x
 
-instance (Reify a e, ToConcrete e c, P.Pretty c) => PrettyTCM (Named String a) where
+instance (Reify a e, ToConcrete e c, P.Pretty c) => PrettyTCM (Named_ a) where
     prettyTCM x = prettyA =<< reify x
 
-instance (ReifyWhen a e, ToConcrete e c, P.Pretty c) => PrettyTCM (Arg a) where
+instance (Reify a e, ToConcrete e c, P.Pretty c) => PrettyTCM (Arg a) where
     prettyTCM x = prettyA =<< reify x
 
 instance (Reify a e, ToConcrete e c, P.Pretty c) => PrettyTCM (Dom a) where
     prettyTCM x = prettyA =<< reify x
 
+-- instance (Reify a e, ToConcrete e c, P.Pretty c, PrettyTCM a) => PrettyTCM (Elim' a) where
 instance PrettyTCM Elim where
   prettyTCM (Apply v) = text "$" <+> prettyTCM v
   prettyTCM (Proj f)  = text "." <> prettyTCM f
 
+instance PrettyTCM a => PrettyTCM (MaybeReduced a) where
+  prettyTCM = prettyTCM . ignoreReduced
+
 instance PrettyTCM A.Expr where
     prettyTCM = prettyA
 
@@ -182,12 +211,9 @@ instance PrettyTCM Constraint where
                 sep [ text (show m) <+> text ":="
                     , nest 2 $ prettyTCM t
                     ]
-              PostponedTypeCheckingProblem cl -> enterClosure cl $ \(e, a, _) ->
+              PostponedTypeCheckingProblem cl _ -> enterClosure cl $ \p ->
                 sep [ text (show m) <+> text ":="
-                    , nest 2 $ sep [ prettyA e <+> text ":?"
-                                   , prettyTCM a
-                                   ]
-                    ]
+                    , nest 2 $ prettyTCM p ]
               Open{}  -> __IMPOSSIBLE__
               OpenIFS{}  -> __IMPOSSIBLE__
               InstS{} -> __IMPOSSIBLE__
@@ -202,6 +228,14 @@ instance PrettyTCM Constraint where
         IsEmpty r t ->
             sep [ text "Is empty:", nest 2 $ prettyTCM t ]
 
+instance PrettyTCM TypeCheckingProblem where
+  prettyTCM (CheckExpr e a) =
+    sep [ prettyA e <+> text ":?", prettyTCM a ]
+  prettyTCM (CheckArgs _ _ _ es t0 t1 _) =
+    sep [ parens $ text "_ :" <+> prettyTCM t0
+        , nest 2 $ prettyList $ map prettyA es
+        , nest 2 $ text ":?" <+> prettyTCM t1 ]
+
 instance PrettyTCM Literal where
   prettyTCM = text . show
 
@@ -214,10 +248,13 @@ instance PrettyTCM QName where
 instance PrettyTCM ModuleName where
   prettyTCM x = P.pretty <$> abstractToConcrete_ x
 
+instance PrettyTCM ConHead where
+  prettyTCM = prettyTCM . conName
+
 instance PrettyTCM Telescope where
   prettyTCM tel = P.fsep . map P.pretty <$> (do
       tel <- reify tel
-      runAbsToCon $ bindToConcrete tel return
+      runAbsToCon $ bindToConcrete tel (return . concat)
     )
 
 newtype PrettyContext = PrettyContext Context
@@ -225,16 +262,32 @@ newtype PrettyContext = PrettyContext Context
 instance PrettyTCM PrettyContext where
   prettyTCM (PrettyContext ctx) = P.fsep . reverse <$> pr (map ctxEntry ctx)
       where
+          pr :: [Dom (Name, Type)] -> TCM [P.Doc]
           pr []		   = return []
-          pr (Dom h r (x,t) : ctx) = escapeContext 1 $ do
+          pr (Common.Dom info (x,t) : ctx) = escapeContext 1 $ do
               d    <- prettyTCM t
               x    <- prettyTCM x
               dctx <- pr ctx
-              return $ P.pRelevance r (par h (P.hsep [ x, P.text ":", d])) : dctx
+              -- TODO guilhem: show colors
+              return $ P.pRelevance info' (par (P.hsep [ x, P.text ":", d])) : dctx
             where
-              par NotHidden = P.parens
-              par Hidden    = P.braces
-              par Instance  = P.dbraces
+              info' = mapArgInfoColors (const []) info
+              par = case argInfoHiding info of
+                     NotHidden -> P.parens
+                     Hidden    -> P.braces
+                     Instance  -> P.dbraces
 
 instance PrettyTCM Context where
   prettyTCM = prettyTCM . PrettyContext
+
+instance PrettyTCM Pattern where
+  prettyTCM = showPat
+    where
+      showPat (VarP x)                  = text $ patVarNameToString x
+      showPat (DotP t)                  = text $ ".(" ++ show t ++ ")"
+      showPat (ConP c Nothing ps)       = parens $
+        prettyTCM c <+> fsep (map (showPat . namedArg) ps)
+      showPat (ConP c (Just (b, t)) ps) = (if b then braces else parens) $
+        prettyTCM c <+> fsep (map (showPat . namedArg) ps) <+> text ":" <+> prettyTCM t
+      showPat (LitP l)                  = text (show l)
+      showPat (ProjP q)                 = text (show q)
