@@ -1,62 +1,47 @@
-{-# LANGUAGE CPP, FlexibleInstances, UndecidableInstances,
-             GeneralizedNewtypeDeriving, ScopedTypeVariables
-  #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE UndecidableInstances #-}
 
 {-| Primitive functions, such as addition on builtin integers.
 -}
 module Agda.TypeChecking.Primitive where
 
 import Control.Monad
-import Control.Monad.Error
+import Control.Applicative
+
+import Data.Char
 import Data.Map (Map)
 import qualified Data.Map as Map
-import Data.Char
 
 import Agda.Interaction.Options
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common hiding (Nat)
-import Agda.Syntax.Internal
+import Agda.Syntax.Internal as I
 import Agda.Syntax.Literal
 import Agda.Syntax.Concrete.Pretty ()
-import Agda.Syntax.Abstract.Name
-import qualified Agda.Syntax.Concrete.Name as C
 
-import Agda.TypeChecking.Monad
+import Agda.TypeChecking.Monad hiding (getConstInfo, typeOfConst)
+import qualified Agda.TypeChecking.Monad as TCM
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Reduce
+import Agda.TypeChecking.Reduce.Monad
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Errors
-import Agda.TypeChecking.Quote (quoteType, quotingKit)
-import Agda.TypeChecking.Pretty ()  -- instances only
-import {-# SOURCE #-} Agda.TypeChecking.Conversion
-import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Level
+import Agda.TypeChecking.Quote (quotingKit)
+import Agda.TypeChecking.Pretty ()  -- instances only
 
 import Agda.Utils.Monad
 import Agda.Utils.Pretty (pretty)
+import Agda.Utils.Maybe
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
-
--- | Rewrite a literal to constructor form if possible.
-constructorForm :: Term -> TCM Term
-constructorForm v = case ignoreSharing v of
-{- 2012-04-02 changed semantics of DontCare
--- Andreas, 2011-10-03, the following line restores IrrelevantLevel
-    DontCare v                  -> constructorForm v
--}
-    Lit (LitInt r n)            -> cons primZero primSuc (Lit . LitInt r) n
---     Level (Max [])              -> primLevelZero
---     Level (Max [ClosedLevel n]) -> cons primLevelZero primLevelSuc (Level . Max . (:[]) . ClosedLevel) n
-    _                           -> return v
-  where
-    cons pZero pSuc lit n
-      | n == 0  = pZero
-      | n > 0   = do
-        s <- pSuc
-        return $ s `apply` [defaultArg $ lit $ n - 1]
-      | otherwise	= return v
+import Debug.Trace
 
 ---------------------------------------------------------------------------
 -- * Primitive functions
@@ -70,7 +55,14 @@ newtype Str = Str { unStr :: String }
     deriving (Eq, Ord)
 
 newtype Nat = Nat { unNat :: Integer }
-    deriving (Eq, Ord, Num, Integral, Enum, Real)
+    deriving (Eq, Ord, Num, Enum, Real)
+
+-- TODO: ghc-7.7 bug: deriving Integral causes an unnecessary toInteger
+-- warning. Once 7.8 is out check if we can go back to deriving.
+instance Integral Nat where
+  toInteger = unNat
+  quotRem (Nat a) (Nat b) = (Nat q, Nat r)
+    where (q, r) = quotRem a b
 
 newtype Lvl = Lvl { unLvl :: Integer }
   deriving (Eq, Ord)
@@ -147,7 +139,7 @@ instance (PrimTerm a, ToTerm a) => ToTerm [a] where
 
 -- From Haskell value to Agda term
 
-type FromTermFunction a = Arg Term -> TCM (Reduced (MaybeReduced (Arg Term)) a)
+type FromTermFunction a = I.Arg Term -> ReduceM (Reduced (MaybeReduced (I.Arg Term)) a)
 
 class FromTerm a where
     fromTerm :: TCM (FromTermFunction a)
@@ -213,19 +205,19 @@ instance (ToTerm a, FromTerm a) => FromTerm [a] where
     where
       isCon (Lam _ b)  = isCon $ absBody b
       isCon (Con c _)  = return c
-      isCon (Shared p) = isCon (derefPtr p)
-      isCon v          = do
-        d <- prettyTCM v
-        typeError $ GenericError $ "expected constructor in built-in binding to " ++ show d
-                        -- TODO: check this when binding the things
+      isCon (Shared p) = __IMPOSSIBLE__ -- isCon (derefPtr p)
+      isCon v          = __IMPOSSIBLE__
 
       mkList nil cons toA fromA t = do
-        b <- reduceB t
+        b <- reduceB' t
         let t = ignoreBlocking b
-        let arg = Arg (argHiding t) (argRelevance t)
+        let arg = Arg (ArgInfo { argInfoHiding = getHiding t
+                               , argInfoRelevance = getRelevance t
+                               , argInfoColors = argColors t
+                               })
         case unArg t of
           Con c []
-            | c == nil  -> return $ YesReduction []
+            | c == nil  -> return $ YesReduction NoSimplification []
           Con c [x,xs]
             | c == cons ->
               redBind (toA x)
@@ -237,22 +229,22 @@ instance (ToTerm a, FromTerm a) => FromTerm [a] where
           _ -> return $ NoReduction (reduced b)
 
 -- | Conceptually: @redBind m f k = either (return . Left . f) k =<< m@
-redBind :: TCM (Reduced a a') -> (a -> b) ->
-	     (a' -> TCM (Reduced b b')) -> TCM (Reduced b b')
+redBind :: ReduceM (Reduced a a') -> (a -> b) ->
+	     (a' -> ReduceM (Reduced b b')) -> ReduceM (Reduced b b')
 redBind ma f k = do
     r <- ma
     case r of
-	NoReduction x	-> return $ NoReduction $ f x
-	YesReduction y	-> k y
+	NoReduction x    -> return $ NoReduction $ f x
+	YesReduction _ y -> k y
 
-redReturn :: a -> TCM (Reduced a' a)
-redReturn = return . YesReduction
+redReturn :: a -> ReduceM (Reduced a' a)
+redReturn = return . YesReduction YesSimplification
 
 fromReducedTerm :: (Term -> Maybe a) -> TCM (FromTermFunction a)
 fromReducedTerm f = return $ \t -> do
-    b <- reduceB t
+    b <- reduceB' t
     case f $ ignoreSharing $ unArg (ignoreBlocking b) of
-	Just x	-> return $ YesReduction x
+	Just x	-> return $ YesReduction NoSimplification x
 	Nothing	-> return $ NoReduction (reduced b)
 
 fromLiteral :: (Literal -> Maybe a) -> TCM (FromTermFunction a)
@@ -272,49 +264,70 @@ primTrustMe = do
           El (varSort 3) <$>
             primEquality <#> varM 3 <#> varM 2 <@> varM 1 <@> varM 0
   Con rf [] <- ignoreSharing <$> primRefl
-  n         <- conPars . theDef <$> getConstInfo rf
-  let refl x | n == 2    = Con rf [Arg Hidden Forced x]
+  n         <- conPars . theDef <$> getConInfo rf
+  let refl x | n == 2    = Con rf [setRelevance Forced $ hide $ defaultArg x]
              | n == 3    = Con rf []
              | otherwise = __IMPOSSIBLE__
   return $ PrimImpl t $ PrimFun __IMPOSSIBLE__ 4 $ \ts ->
       case ts of
-        [a, t, x, y] -> liftTCM $ do
+        [a, t, u, v] -> do
+            -- Andreas, 2013-07-22.
+            -- Note that we cannot call the conversion checker here,
+            -- because 'reduce' might be called in a context where
+            -- some bound variables do not have a type (just 'Prop),
+            -- and the conversion checker for eliminations does not
+            -- like this.
+            -- We can only do untyped equality, e.g., by normalisation.
+            (u', v') <- normalise' (u, v)
+            if (u' == v') then redReturn (refl $ unArg u) else
+              return (NoReduction $ map notReduced [a, t, u, v])
+{- OLD:
+
+              -- BAD:
               noConstraints $
-                equalTerm (El (Type $ lvlView $ unArg a) (unArg t))
-                          (unArg x) (unArg y)
-              redReturn (refl $ unArg x)
-            `catchError` \_ -> return (NoReduction $ map notReduced [a, t, x, y])
+                equalTerm (El (Type $ lvlView $ unArg a) (unArg t)) (unArg u) (unArg v)
+              redReturn (refl $ unArg u)
+            `catchError` \_ -> return (NoReduction $ map notReduced [a, t, u, v])
+-}
         _ -> __IMPOSSIBLE__
 
 primQNameType :: TCM PrimitiveImpl
-primQNameType = mkPrimFun1TCM (el primQName --> el primAgdaType) typeOfConst
+primQNameType = mkPrimFun1TCM (el primQName --> el primAgdaType)
+                              (\q -> defType <$> getConstInfo q)
+  -- Note: gets the top-level type! All bounds variables have been lifted.
 
 primQNameDefinition :: TCM PrimitiveImpl
 primQNameDefinition = do
-  let argQName qn = [defaultArg (Lit (LitQName noRange qn))]
-      app mt xs = do t <- mt
-                     return $ apply t xs
+  agdaFunDef                    <- primAgdaDefinitionFunDef
+  agdaDefinitionFunDef          <- primAgdaDefinitionFunDef
+  agdaDefinitionDataDef         <- primAgdaDefinitionDataDef
+  agdaDefinitionRecordDef       <- primAgdaDefinitionRecordDef
+  agdaDefinitionPostulate       <- primAgdaDefinitionPostulate
+  agdaDefinitionPrimitive       <- primAgdaDefinitionPrimitive
+  agdaDefinitionDataConstructor <- primAgdaDefinitionDataConstructor
 
-      con qn Function{}    = app primAgdaDefinitionFunDef    (argQName qn)
-      con qn Datatype{}    = app primAgdaDefinitionDataDef   (argQName qn)
-      con qn Record{}      = app primAgdaDefinitionRecordDef (argQName qn)
-      con _  Axiom{}       = app primAgdaDefinitionPostulate []
-      con _  Primitive{}   = app primAgdaDefinitionPrimitive []
-      con _  Constructor{} = app primAgdaDefinitionDataConstructor []
+  let argQName qn = [defaultArg (Lit (LitQName noRange qn))]
+      con qn Function{}    = apply agdaDefinitionFunDef    (argQName qn)
+      con qn Datatype{}    = apply agdaDefinitionDataDef   (argQName qn)
+      con qn Record{}      = apply agdaDefinitionRecordDef (argQName qn)
+      con _  Axiom{}       = apply agdaDefinitionPostulate []
+      con _  Primitive{}   = apply agdaDefinitionPrimitive []
+      con _  Constructor{} = apply agdaDefinitionDataConstructor []
 
   unquoteQName <- fromTerm
   t <- el primQName --> el primAgdaDefinition
   return $ PrimImpl t $ PrimFun __IMPOSSIBLE__ 1 $ \ts ->
     case ts of
-      [v] -> liftTCM $
+      [v] ->
         redBind (unquoteQName v)
             (\v' -> [v']) $ \x ->
-        redReturn =<< con x . theDef =<< getConstInfo x
+        redReturn =<< (con x . theDef <$> getConstInfo x)
       _ -> __IMPOSSIBLE__
 
 primDataConstructors :: TCM PrimitiveImpl
-primDataConstructors = mkPrimFun1TCM (el primAgdaDataDef --> el (list primQName))
-                                     (fmap (dataCons . theDef) . getConstInfo)
+primDataConstructors =
+  mkPrimFun1TCM (el primAgdaDataDef --> el (list primQName))
+                (fmap (dataCons . theDef) . getConstInfo)
 
 mkPrimLevelZero :: TCM PrimitiveImpl
 mkPrimLevelZero = do
@@ -324,26 +337,26 @@ mkPrimLevelZero = do
 mkPrimLevelSuc :: TCM PrimitiveImpl
 mkPrimLevelSuc = do
   t <- primType (id :: Lvl -> Lvl)
-  return $ PrimImpl t $ PrimFun __IMPOSSIBLE__ 1 $ \ ~[a] -> liftTCM $ do
-    l <- levelView $ unArg a
+  return $ PrimImpl t $ PrimFun __IMPOSSIBLE__ 1 $ \ ~[a] -> do
+    l <- levelView' $ unArg a
     redReturn $ Level $ levelSuc l
 
 mkPrimLevelMax :: TCM PrimitiveImpl
 mkPrimLevelMax = do
   t <- primType (max :: Op Lvl)
-  return $ PrimImpl t $ PrimFun __IMPOSSIBLE__ 2 $ \ ~[a, b] -> liftTCM $ do
-    Max as <- levelView $ unArg a
-    Max bs <- levelView $ unArg b
+  return $ PrimImpl t $ PrimFun __IMPOSSIBLE__ 2 $ \ ~[a, b] -> do
+    Max as <- levelView' $ unArg a
+    Max bs <- levelView' $ unArg b
     redReturn $ Level $ levelMax $ as ++ bs
 
-mkPrimFun1TCM :: (FromTerm a, ToTerm b) => TCM Type -> (a -> TCM b) -> TCM PrimitiveImpl
+mkPrimFun1TCM :: (FromTerm a, ToTerm b) => TCM Type -> (a -> ReduceM b) -> TCM PrimitiveImpl
 mkPrimFun1TCM mt f = do
     toA   <- fromTerm
     fromB <- toTerm
     t     <- mt
     return $ PrimImpl t $ PrimFun __IMPOSSIBLE__ 1 $ \ts ->
       case ts of
-        [v] -> liftTCM $
+        [v] ->
           redBind (toA v)
               (\v' -> [v']) $ \x ->
           redReturn . fromB =<< f x
@@ -358,7 +371,7 @@ mkPrimFun1 f = do
     t	  <- primType f
     return $ PrimImpl t $ PrimFun __IMPOSSIBLE__ 1 $ \ts ->
       case ts of
-        [v] -> liftTCM $
+        [v] ->
           redBind (toA v)
               (\v' -> [v']) $ \x ->
           redReturn $ fromB $ f x
@@ -374,11 +387,11 @@ mkPrimFun2 f = do
     t	  <- primType f
     return $ PrimImpl t $ PrimFun __IMPOSSIBLE__ 2 $ \ts ->
       case ts of
-        [v,w] -> liftTCM $
+        [v,w] ->
           redBind (toA v)
               (\v' -> [v', notReduced w]) $ \x ->
           redBind (toB w)
-              (\w' -> [ reduced $ notBlocked $ Arg (argHiding v) (argRelevance v) (fromA x)
+              (\w' -> [ reduced $ notBlocked $ Arg (argInfo v) (fromA x)
                       , w']) $ \y ->
           redReturn $ fromC $ f x y
         _ -> __IMPOSSIBLE__
@@ -398,9 +411,9 @@ mkPrimFun4 f = do
     t <- primType f
     return $ PrimImpl t $ PrimFun __IMPOSSIBLE__ 4 $ \ts ->
       let argFrom fromX a x =
-            reduced $ notBlocked $ Arg (argHiding a) (argRelevance a) (fromX x)
+            reduced $ notBlocked $ Arg (argInfo a) (fromX x)
       in case ts of
-        [a,b,c,d] -> liftTCM $
+        [a,b,c,d] ->
           redBind (toA a)
               (\a' -> a' : map notReduced [b,c,d]) $ \x ->
           redBind (toB b)
@@ -425,7 +438,7 @@ infixr 4 -->
 a --> b = do
     a' <- a
     b' <- b
-    return $ El (getSort a' `sLub` getSort b') $ Pi (Dom NotHidden Relevant a') (NoAbs "_" b')
+    return $ El (getSort a' `sLub` getSort b') $ Pi (Dom defaultArgInfo a') (NoAbs "_" b')
 
 infixr 4 .-->
 
@@ -433,19 +446,20 @@ infixr 4 .-->
 a .--> b = do
     a' <- a
     b' <- b
-    return $ El (getSort a' `sLub` getSort b') $ Pi (Dom NotHidden Irrelevant a') (NoAbs "_" b')
+    return $ El (getSort a' `sLub` getSort b') $
+             Pi (Dom (setRelevance Irrelevant defaultArgInfo) a') (NoAbs "_" b')
 
-gpi :: Hiding -> Relevance -> String -> TCM Type -> TCM Type -> TCM Type
-gpi h r name a b = do
+gpi :: I.ArgInfo -> String -> TCM Type -> TCM Type -> TCM Type
+gpi info name a b = do
   a <- a
-  x <- freshName_ name
-  b <- addCtx x (Dom h r a) b
-  return $ El (getSort a `dLub` Abs name (getSort b))
-              (Pi (Dom h r a) (Abs name b))
+  b <- addContext (name, Dom info a) b
+  let y = stringToArgName name
+  return $ El (getSort a `dLub` Abs y (getSort b))
+              (Pi (Dom info a) (Abs y b))
 
 hPi, nPi :: String -> TCM Type -> TCM Type -> TCM Type
-hPi = gpi Hidden Relevant
-nPi = gpi NotHidden Relevant
+hPi = gpi $ setHiding Hidden $ defaultArgInfo
+nPi = gpi defaultArgInfo
 
 varM :: Int -> TCM Term
 varM = return . var
@@ -456,7 +470,7 @@ gApply :: Hiding -> TCM Term -> TCM Term -> TCM Term
 gApply h a b = do
     x <- a
     y <- b
-    return $ x `apply` [Arg h Relevant y]
+    return $ x `apply` [Arg (setHiding h defaultArgInfo) y]
 
 (<@>),(<#>) :: TCM Term -> TCM Term -> TCM Term
 (<@>) = gApply NotHidden
@@ -474,15 +488,15 @@ el t = El (mkType 0) <$> t
 tset :: TCM Type
 tset = return $ sort (mkType 0)
 
--- | Abbreviation: @argN = 'Arg' 'NotHidden' 'Relevant'@.
+-- | Abbreviation: @argN = 'Arg' 'defaultArgInfo'@.
 
-argN = Arg NotHidden Relevant
-domN = Dom NotHidden Relevant
+argN = Arg defaultArgInfo
+domN = Dom defaultArgInfo
 
--- | Abbreviation: @argH = 'Arg' 'Hidden' 'Relevant'@.
+-- | Abbreviation: @argH = 'hide' 'Arg' 'defaultArgInfo'@.
 
-argH = Arg Hidden Relevant
-domH = Dom Hidden Relevant
+argH = Arg $ setHiding Hidden defaultArgInfo
+domH = Dom $ setHiding Hidden defaultArgInfo
 
 ---------------------------------------------------------------------------
 -- * The actual primitive functions
@@ -569,7 +583,8 @@ primitiveFunctions = Map.fromList
 
     -- Other stuff
     , "primTrustMe"         |-> primTrustMe
-    , "primQNameEquality"  |-> mkPrimFun2 ((==) :: Rel QName)
+    , "primQNameEquality"   |-> mkPrimFun2 ((==) :: Rel QName)
+    , "primShowQName"       |-> mkPrimFun1 (Str . show :: QName -> Str)
     ]
     where
 	(|->) = (,)
@@ -577,7 +592,7 @@ primitiveFunctions = Map.fromList
 lookupPrimitiveFunction :: String -> TCM PrimitiveImpl
 lookupPrimitiveFunction x =
     case Map.lookup x primitiveFunctions of
-	Just p	-> liftTCM p
+	Just p	-> p
 	Nothing	-> typeError $ NoSuchPrimitiveFunction x
 
 lookupPrimitiveFunctionQ :: QName -> TCM (String, PrimitiveImpl)
@@ -586,4 +601,3 @@ lookupPrimitiveFunctionQ q = do
             Name _ x _ _ -> show x
   PrimImpl t pf <- lookupPrimitiveFunction s
   return (s, PrimImpl t $ pf { primFunName = q })
-
