@@ -1,20 +1,27 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE TypeSynonymInstances #-}
 
 module Agda.TypeChecking.Quote where
 
 import Control.Applicative
 
+import Data.Maybe (fromMaybe)
+
 import Agda.Syntax.Position
 import Agda.Syntax.Literal
-import Agda.Syntax.Internal
+import Agda.Syntax.Internal as I
 import Agda.Syntax.Common
 
+import {-# SOURCE #-} Agda.TypeChecking.Datatypes
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Substitute
 
+import Agda.Utils.Monad
+
 #include "../undefined.h"
 import Agda.Utils.Impossible
 
@@ -28,6 +35,7 @@ quotingKit = do
   nil <- primNil
   cons <- primCons
   arg <- primArgArg
+  arginfo <- primArgArgInfo
   var <- primAgdaTermVar
   lam <- primAgdaTermLam
   def <- primAgdaTermDef
@@ -52,6 +60,10 @@ quotingKit = do
       quoteRelevance NonStrict  = relevant
       quoteRelevance Forced     = relevant
       quoteRelevance UnusedArg  = relevant
+      quoteColors _ = nil -- TODO guilhem
+      quoteArgInfo (ArgInfo h r cs) = arginfo @@ quoteHiding h
+                                              @@ quoteRelevance r
+                                --              @@ quoteColors cs
       quoteLit (LitInt   _ n)   = iterate suc zero !! fromIntegral n
       quoteLit _                = unsupported
       -- We keep no ranges in the quoted term, so the equality on terms
@@ -59,7 +71,7 @@ quotingKit = do
       quoteSortLevelTerm (Max [])              = setLit @@ Lit (LitInt noRange 0)
       quoteSortLevelTerm (Max [ClosedLevel n]) = setLit @@ Lit (LitInt noRange n)
       quoteSortLevelTerm (Max [Plus 0 (NeutralLevel v)]) = set @@ quote v
-      quoteSortLevelTerm _                     = unsupported
+      quoteSortLevelTerm _                     = unsupportedSort
       quoteSort (Type t)    = quoteSortLevelTerm t
       quoteSort Prop        = unsupportedSort
       quoteSort Inf         = unsupportedSort
@@ -67,28 +79,37 @@ quotingKit = do
       quoteType (El s t) = el @@ quoteSort s @@ quote t
       list [] = nil
       list (a : as) = cons @@ a @@ list as
-      zero = con @@ quoteName z @@ nil
-      suc n = con @@ quoteName s @@ list [arg @@ visible @@ relevant @@ n]
-      quoteDom q (Dom h r t) = arg @@ quoteHiding h @@ quoteRelevance r @@ q t
-      quoteArg q (Arg h r t) = arg @@ quoteHiding h @@ quoteRelevance r @@ q t
+      zero = con @@ quoteConName z @@ nil
+      suc n = con @@ quoteConName s @@ list [arg @@ quoteArgInfo defaultArgInfo @@ n]
+      quoteDom q (Dom info t) = arg @@ quoteArgInfo info @@ q t
+      quoteArg q (Arg info t) = arg @@ quoteArgInfo info @@ q t
       quoteArgs ts = list (map (quoteArg quote) ts)
-      quote (Var n ts) = var @@ Lit (LitInt noRange $ fromIntegral n) @@ quoteArgs ts
-      quote (Lam h t) = lam @@ quoteHiding h @@ quote (absBody t)
-      quote (Def x ts) = def @@ quoteName x @@ quoteArgs ts
-      quote (Con x ts) = con @@ quoteName x @@ quoteArgs ts
-      quote (Pi t u) = pi @@ quoteDom quoteType t
-                          @@ quoteType (absBody u)
-      quote (Level _) = unsupported
-      quote (Lit lit) = quoteLit lit
-      quote (Sort s)  = sort @@ quoteSort s
-      quote (Shared p) = quote $ derefPtr p
-      quote MetaV{}   = unsupported
-      quote DontCare{} = unsupported -- could be exposed at some point but we have to take care
+      quote v =
+        case unSpine v of
+          (Var n es)   ->
+             let ts = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
+             in  var @@ Lit (LitInt noRange $ fromIntegral n) @@ quoteArgs ts
+          (Lam info t) -> lam @@ quoteHiding (getHiding info) @@ quote (absBody t)
+          (Def x es)   ->
+             let ts = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
+             in  def @@ quoteName x @@ quoteArgs ts
+          (Con x ts)   -> con @@ quoteConName x @@ quoteArgs ts
+          (Pi t u)     -> pi @@ quoteDom quoteType t
+                        @@ quoteType (absBody u)
+          (Level _)    -> unsupported
+          (Lit lit)    -> quoteLit lit
+          (Sort s)     -> sort @@ quoteSort s
+          (Shared p)   -> quote $ derefPtr p
+          MetaV{}      -> unsupported
+          DontCare{}   -> unsupported -- could be exposed at some point but we have to take care
   return (quote, quoteType)
 
 quoteName :: QName -> Term
 quoteName x = Lit (LitQName noRange x)
 
+quoteConName :: ConHead -> Term
+quoteConName = quoteName . conName
+
 quoteTerm :: Term -> TCM Term
 quoteTerm v = ($v) . fst <$> quotingKit
 
@@ -101,7 +122,7 @@ agdaTermType = El (mkType 0) <$> primAgdaTerm
 qNameType :: TCM Type
 qNameType = El (mkType 0) <$> primQName
 
-isCon :: QName -> TCM Term -> TCM Bool
+isCon :: ConHead -> TCM Term -> TCM Bool
 isCon con tm = do t <- tm
                   case ignoreSharing t of
                     Con con' _ -> return (con == con')
@@ -117,28 +138,44 @@ unquoteFailed kind msg t = do doc <- prettyTCM t
 class Unquote a where
   unquote :: Term -> TCM a
 
-unquoteH :: Unquote a => Arg Term -> TCM a
-unquoteH (Arg Hidden Relevant x) = unquote x
-unquoteH _                       = unquoteFailedGeneric "argument. It should be `hidden'."
+unquoteH :: Unquote a => I.Arg Term -> TCM a
+unquoteH a | isHidden a && isRelevant a =
+    unquote $ unArg a
+unquoteH _ = unquoteFailedGeneric "argument. It should be `hidden'."
 
-unquoteN :: Unquote a => Arg Term -> TCM a
-unquoteN (Arg NotHidden Relevant x) = unquote x
-unquoteN _                          = unquoteFailedGeneric "argument. It should be `visible'"
+unquoteN :: Unquote a => I.Arg Term -> TCM a
+unquoteN a | notHidden a && isRelevant a =
+    unquote $ unArg a
+unquoteN _ = unquoteFailedGeneric "argument. It should be `visible'"
 
 choice :: Monad m => [(m Bool, m a)] -> m a -> m a
 choice [] dflt = dflt
-choice ((mb, mx) : mxs) dflt = do b <- mb
-                                  if b then mx else choice mxs dflt
+choice ((mb, mx) : mxs) dflt = ifM mb mx $ choice mxs dflt
 
-instance Unquote a => Unquote (Arg a) where
+instance Unquote I.ArgInfo where
   unquote t = do
     t <- reduce t
     case ignoreSharing t of
-      Con c [hid,rel,x] -> do
+      Con c [h,r] -> do
         choice
-          [(c `isCon` primArgArg, Arg <$> unquoteN hid <*> unquoteN rel <*> unquoteN x)]
-          (unquoteFailed "Arg" "arity 3 and not the `arg' constructor" t)
-      _ -> unquoteFailed "Arg" "not of arity 3" t
+          [(c `isCon` primArgArgInfo, ArgInfo <$> unquoteN h <*> unquoteN r <*> return [])]
+          (unquoteFailed "ArgInfo" "arity 2 and not the `arginfo' constructor" t)
+      _ -> unquoteFailed "ArgInfo" "not of arity 2" t
+
+instance Unquote a => Unquote (I.Arg a) where
+  unquote t = do
+    t <- reduce t
+    case ignoreSharing t of
+      Con c [info,x] -> do
+        choice
+          [(c `isCon` primArgArg, Arg <$> unquoteN info <*> unquoteN x)]
+          (unquoteFailed "Arg" "arity 2 and not the `arg' constructor" t)
+      _ -> unquoteFailed "Arg" "not of arity 2" t
+
+-- Andreas, 2013-10-20: currently, post-fix projections are not part of the
+-- quoted syntax.
+instance Unquote a => Unquote (Elim' a) where
+  unquote t = Apply <$> unquote t
 
 instance Unquote Integer where
   unquote t = do
@@ -195,9 +232,12 @@ instance Unquote QName where
       Lit (LitQName _ x) -> return x
       _                  -> unquoteFailed "QName" "not a literal qname value" t
 
+instance Unquote ConHead where
+  unquote t = getConHead =<< unquote t
+
 instance Unquote a => Unquote (Abs a) where
-  unquote t = do x <- freshNoName_
-                 Abs (show x) <$> unquote t
+  unquote t = do x <- freshNoName_ -- Andreas, 2014-07-11 This is pointless, as it does NOT generate a name suggestion.
+                 Abs (nameToArgName x) <$> unquote t
 
 instance Unquote Sort where
   unquote t = do
@@ -246,7 +286,7 @@ instance Unquote Term where
           [(c `isCon` primAgdaTermVar, Var <$> (fromInteger <$> unquoteN x) <*> unquoteN y)
           ,(c `isCon` primAgdaTermCon, Con <$> unquoteN x <*> unquoteN y)
           ,(c `isCon` primAgdaTermDef, Def <$> unquoteN x <*> unquoteN y)
-          ,(c `isCon` primAgdaTermLam, Lam <$> unquoteN x <*> unquoteN y)
+          ,(c `isCon` primAgdaTermLam, Lam <$> (flip setHiding defaultArgInfo <$> unquoteN x) <*> unquoteN y)
           ,(c `isCon` primAgdaTermPi,  Pi  <$> (domFromArg <$> unquoteN x) <*> unquoteN y)]
           (unquoteFailed "Term" "arity 2 and none of Var, Con, Def, Lam, Pi" t)
 
