@@ -1,4 +1,8 @@
-{-# LANGUAGE CPP, PatternGuards, TupleSections, FlexibleInstances, GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE TupleSections #-}
 
 -- | Code which replaces pattern matching on record constructors with
 -- uses of projection functions.
@@ -17,13 +21,11 @@ import Control.Monad.Reader
 import Control.Monad.State
 
 import Data.List
-import Data.Map (Map)
 import qualified Data.Map as Map
 import qualified Data.Traversable as Trav
 
 import Agda.Syntax.Common
-import Agda.Syntax.Internal
-
+import Agda.Syntax.Internal as I
 import Agda.TypeChecking.CompiledClause
 import Agda.TypeChecking.Coverage.SplitTree
 import Agda.TypeChecking.Datatypes
@@ -39,7 +41,7 @@ import Agda.Utils.Either
 import Agda.Utils.List
 import qualified Agda.Utils.Map as Map
 import Agda.Utils.Maybe
-import Agda.Utils.Permutation
+import Agda.Utils.Permutation hiding (dropFrom)
 import Agda.Utils.Size
 
 #include "../undefined.h"
@@ -62,14 +64,16 @@ recordPatternToProjections p =
     LitP{}             -> typeError $ ShouldBeRecordPattern p
     DotP{}             -> typeError $ ShouldBeRecordPattern p
     ConP c Nothing  ps -> typeError $ ShouldBeRecordPattern p
-    ConP c (Just t) ps -> do
+    ConP c (Just (_, t)) ps -> do
       t <- reduce t
       fields <- getRecordTypeFields (unArg t)
-      concat <$> zipWithM comb (map proj fields) (map unArg ps)
+      concat <$> zipWithM comb (map proj fields) (map namedArg ps)
+    ProjP{}            -> __IMPOSSIBLE__ -- copattern cannot appear here
   where
-    proj p = \ x -> Def (unArg p) [defaultArg x]
+    proj p = (`applyE` [Proj $ unArg p])
+--    proj p = \ x -> Def (unArg p) [defaultArg x]
     comb :: (Term -> Term) -> Pattern -> TCM [Term -> Term]
-    comb prj p = map (prj .) <$> recordPatternToProjections p
+    comb prj p = map (\ f -> f . prj) <$> recordPatternToProjections p
 
 
 ---------------------------------------------------------------------------
@@ -137,11 +141,16 @@ translateCompiledClauses cc = snd <$> loop cc
       (ccs, xssc, conMap)    <- Map.unzip3 <$> do
         Trav.forM (Map.mapWithKey (,) conMap) $ \ (c, WithArity ar cc) -> do
           (xs, cc)     <- loop cc
-          dataOrRecCon <- getConstructorArity c
+--          dataOrRecCon <- getConstructorArity c  -- TODO: c could be a projection
+          dataOrRecCon <- do
+            isProj <- isProjection c
+            case isProj of
+               Nothing -> getConstructorArity c
+               Just{}  -> return $ Left 0
           let (isRC, n)   = either (False,) ((True,) . size) dataOrRecCon
               (xs0, rest) = genericSplitAt i xs
               (xs1, xs2 ) = genericSplitAt n rest
-              -- if all dropped variables are virgins and we are record cons.
+              -- if all dropped variables (xs1) are virgins and we are record cons.
               -- then new variable x is also virgin
               -- and we can translate away the split
               x           = isRC && and xs1
@@ -165,11 +174,34 @@ translateCompiledClauses cc = snd <$> loop cc
                   , catchAllBranch = catchAll })
 
         -- case: translated away one record pattern
-        [cc] -> return (xs, cc)
+        [cc] -> do
+                -- Andreas, 2013-03-22
+                -- Due to catch-all-expansion this is actually possible:
+                -- -- we cannot have a catch-all if we had a record pattern
+                -- whenJust catchAll __IMPOSSIBLE__
+                -- We just drop the catch-all clause.  This is safe because
+                -- for record patterns we have expanded all the catch-alls.
+                return (xs, cc) -- mergeCatchAll cc catchAll)
 
         -- case: more than one record patterns (impossible)
         _    -> __IMPOSSIBLE__
 
+{- UNUSED
+instance Monoid CompiledClauses where
+  mempty = __IMPOSSIBLE__
+  mappend (Case n c) (Case n' c') | n == n' = Case n $ mappend c c'
+  mappend _ _ = __IMPOSSIBLE__
+
+mergeCatchAll :: CompiledClauses -> Maybe CompiledClauses -> CompiledClauses
+mergeCatchAll cc ca = maybe cc (mappend cc) ca
+{-
+  case (cc, ca) of
+    (_       , Nothing) -> cc
+    (Case n c, Just (Case n' c')) | n == n' -> Case n $ mappend c c'
+    _                   -> __IMPOSSIBLE__ -- this would mean non-determinism
+-}
+-}
+
 -- | @replaceByProjections i projs cc@ replaces variables @i..i+n-1@
 --   (counted from left) by projections @projs_1 i .. projs_n i@.
 --
@@ -199,9 +231,10 @@ replaceByProjections i projs cc =
           let (xs0,xs1,xs2)     = cutSublist i n xs
               names | null xs1  = ["r"]
                     | otherwise = map unArg xs1
-              x                 = defaultArg $ foldr1 (++) names
+              x                 = defaultArg $ foldr1 appendArgNames names
               xs'               = xs0 ++ x : xs2
-              us                = map (\ p -> Def p [defaultArg $ var 0]) (reverse projs)
+              us                = map (\ p -> Var 0 [Proj p]) (reverse projs)
+--              us                = map (\ p -> Def p [defaultArg $ var 0]) (reverse projs)
               -- go from level (i + n - 1) to index (subtract from |xs|-1)
               index             = length xs - (i + n)
           in  Done xs' $ applySubst (liftS (length xs2) $ us ++# raiseS 1) v
@@ -399,7 +432,7 @@ translateRecordPatterns clause = do
   -- cs: List of changes, with types in the context of the old
   -- telescope.
 
-  (ps, s, cs) <- runRecPatM $ translatePatterns $ clausePats clause
+  (ps, s, cs) <- runRecPatM $ translatePatterns $ namedClausePats clause
 
   let -- Number of variables + dot patterns in new clause.
       noNewPatternVars = size cs
@@ -470,10 +503,10 @@ translateRecordPatterns clause = do
 
       -- New clause.
       c = clause
-            { clauseTel  = newTel
-            , clausePerm = newPerm
-            , clausePats = applySubst lhsSubst ps
-            , clauseBody = translateBody cs rhsSubst $ clauseBody clause
+            { clauseTel       = newTel
+            , clausePerm      = newPerm
+            , namedClausePats = applySubst lhsSubst ps
+            , clauseBody      = translateBody cs rhsSubst $ clauseBody clause
             }
 
   reportSDoc "tc.lhs.recpat" 10 $
@@ -539,7 +572,7 @@ data Kind = VarPat | DotPat
 -- patterns, should be removed, and a new variable, with the name @x@,
 -- inserted instead. The type of the new variable is @t@.
 
-type Changes = [Either Pattern (Kind -> Nat, String, Dom Type)]
+type Changes = [Either Pattern (Kind -> Nat, ArgName, I.Dom Type)]
 
 -- | Record pattern trees.
 
@@ -547,7 +580,7 @@ data RecordTree
   = Leaf Pattern
     -- ^ Corresponds to variable and dot patterns; contains the
     -- original pattern.
-  | RecCon (Arg Type) [(Term -> Term, RecordTree)]
+  | RecCon (I.Arg Type) [(Term -> Term, RecordTree)]
     -- ^ @RecCon t args@ stands for a record constructor application:
     -- @t@ is the type of the application, and the list contains a
     -- projection function and a tree for every argument.
@@ -624,14 +657,12 @@ translatePattern p@(ConP _ (Just _) _) = do
 translatePattern p@VarP{} = removeTree (Leaf p)
 translatePattern p@DotP{} = removeTree (Leaf p)
 translatePattern p@LitP{} = return (p, [], [])
+translatePattern p@ProjP{}= __IMPOSSIBLE__
 
--- | 'translatePattern' lifted to lists of arguments.
-
-translatePatterns ::
-  [Arg Pattern] -> RecPatM ([Arg Pattern], [Term], Changes)
+translatePatterns :: [I.NamedArg Pattern] -> RecPatM ([I.NamedArg Pattern], [Term], Changes)
 translatePatterns ps = do
-  (ps', ss, cs) <- unzip3 <$> mapM (translatePattern . unArg) ps
-  return (ps' `withArgsFrom` ps, concat ss, concat cs)
+  (ps', ss, cs) <- unzip3 <$> mapM (translatePattern . namedArg) ps
+  return (zipWith (\p -> fmap (p <$)) ps' ps, concat ss, concat cs)
 
 -- | Traverses a pattern and returns one of two things:
 --
@@ -651,22 +682,24 @@ recordTree ::
   Pattern ->
   RecPatM (Either (RecPatM (Pattern, [Term], Changes)) RecordTree)
 recordTree p@(ConP _ Nothing _) = return $ Left $ translatePattern p
-recordTree (ConP c (Just t) ps) = do
-  rs <- mapM (recordTree . unArg) ps
+recordTree (ConP c ci@(Just (_, t)) ps) = do
+  rs <- mapM (recordTree . namedArg) ps
   case allRight rs of
-    Left rs ->
+    Nothing ->
       return $ Left $ do
         (ps', ss, cs) <- unzip3 <$> mapM (either id removeTree) rs
-        return (ConP c (Just t) (ps' `withArgsFrom` ps),
+        return (ConP c ci (ps' `withNamedArgsFrom` ps),
                 concat ss, concat cs)
-    Right ts -> liftTCM $ do
+    Just ts -> liftTCM $ do
       t <- reduce t
       fields <- getRecordTypeFields (unArg t)
-      let proj p = \x -> Def (unArg p) [defaultArg x]
+--      let proj p = \x -> Def (unArg p) [defaultArg x]
+      let proj p = (`applyE` [Proj $ unArg p])
       return $ Right $ RecCon t $ zip (map proj fields) ts
 recordTree p@VarP{} = return (Right (Leaf p))
 recordTree p@DotP{} = return (Right (Leaf p))
 recordTree p@LitP{} = return $ Left $ translatePattern p
+recordTree p@ProjP{}= __IMPOSSIBLE__
 
 ------------------------------------------------------------------------
 -- Translation of the clause telescope and body
@@ -677,10 +710,10 @@ translateTel
   :: Changes
      -- ^ Explanation of how the telescope should be changed. Types
      -- should be in the context of the old telescope.
-  -> [(String, Dom Type)]
+  -> [(ArgName, I.Dom Type)]
      -- ^ Old telescope, flattened, in textual left-to-right
      -- order.
-  -> [Maybe (String, Dom Type)]
+  -> [Maybe (ArgName, I.Dom Type)]
      -- ^ New telescope, flattened, in textual left-to-right order.
      -- 'Nothing' is used to indicate the locations of dot patterns.
 translateTel (Left (DotP{}) : rest)   tel = Nothing : translateTel rest tel
