@@ -1,15 +1,18 @@
-{-# LANGUAGE PatternGuards, CPP #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE TupleSections #-}
+
 module Agda.TypeChecking.Rules.Builtin (bindBuiltin, bindPostulatedName) where
 
 import Control.Applicative
 import Control.Monad
 import Control.Monad.Error
-import Data.Maybe
 import Data.List (find)
 
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Internal
 import Agda.Syntax.Common
+import Agda.Syntax.Position
 
 import Agda.TypeChecking.EtaContract
 import Agda.TypeChecking.Monad
@@ -19,16 +22,16 @@ import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Primitive
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Reduce
-import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Irrelevance
-import Agda.TypeChecking.SizedTypes ( builtinSizeHook )
+import Agda.TypeChecking.Monad.SizedTypes ( builtinSizeHook )
 import Agda.TypeChecking.Rules.Term ( checkExpr , inferExpr )
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Builtin.Coinduction
 
+import Agda.Utils.Maybe
 import Agda.Utils.Size
-import Agda.Utils.Impossible
 
-#include "../..//undefined.h"
+#include "../../undefined.h"
+import Agda.Utils.Impossible
 
 ---------------------------------------------------------------------------
 -- * Checking builtin pragmas
@@ -38,9 +41,10 @@ builtinPostulate :: TCM Type -> BuiltinDescriptor
 builtinPostulate = BuiltinPostulate Relevant
 
 coreBuiltins :: [BuiltinInfo]
-coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
+coreBuiltins = map (\ (x, z) -> BuiltinInfo x z)
   [ (builtinList               |-> BuiltinData (tset --> tset) [builtinNil, builtinCons])
   , (builtinArg                |-> BuiltinData (tset --> tset) [builtinArgArg])
+  , (builtinArgInfo            |-> BuiltinData tset [builtinArgArgInfo])
   , (builtinBool               |-> BuiltinData tset [builtinTrue, builtinFalse])
   , (builtinNat                |-> BuiltinData tset [builtinZero, builtinSuc])
   , (builtinLevel              |-> builtinPostulate tset)
@@ -65,7 +69,8 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
                                                [builtinRefl])
   , (builtinHiding             |-> BuiltinData tset [builtinHidden, builtinInstance, builtinVisible])
   , (builtinRelevance          |-> BuiltinData tset [builtinRelevant, builtinIrrelevant])
-  , (builtinRefl               |-> BuiltinDataCons (hPi "a" (el primLevel) $ hPi "A" (return $ sort $ varSort 0) $
+  , (builtinRefl               |-> BuiltinDataCons (hPi "a" (el primLevel) $
+                                                    hPi "A" (return $ sort $ varSort 0) $
                                                     hPi "x" (El (varSort 1) <$> varM 0) $
                                                     El (varSort 2) <$> primEquality <#> varM 2 <#> varM 1 <@> varM 0 <@> varM 0))
   , (builtinNil                |-> BuiltinDataCons (hPi "A" tset (el (list v0))))
@@ -74,7 +79,8 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
   , (builtinSuc                |-> BuiltinDataCons (tnat --> tnat))
   , (builtinTrue               |-> BuiltinDataCons tbool)
   , (builtinFalse              |-> BuiltinDataCons tbool)
-  , (builtinArgArg             |-> BuiltinDataCons (hPi "A" tset (thiding --> trelevance --> tv0 --> targ tv0)))
+  , (builtinArgArg             |-> BuiltinDataCons (hPi "A" tset (targinfo --> tv0 --> targ tv0)))
+  , (builtinArgArgInfo         |-> BuiltinDataCons (thiding --> trelevance --> targinfo))
   , (builtinAgdaTypeEl         |-> BuiltinDataCons (tsort --> tterm --> ttype))
   , (builtinAgdaTermVar        |-> BuiltinDataCons (tnat --> targs --> tterm))
   , (builtinAgdaTermLam        |-> BuiltinDataCons (thiding --> tterm --> tterm))
@@ -150,6 +156,8 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
         tbool      = el primBool
         thiding    = el primHiding
         trelevance = el primRelevance
+--        tcolors    = el (list primAgdaTerm) -- TODO guilhem
+        targinfo   = el primArgInfo
         ttype      = el primAgdaType
         tsort      = el primAgdaSort
         tdefn      = el primAgdaDefinition
@@ -158,7 +166,7 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
         trec       = el primAgdaRecordDef
 
         verifyPlus plus =
-            verify ["n","m"] $ \(@@) zero suc (==) choice -> do
+            verify ["n","m"] $ \(@@) zero suc (==) (===) choice -> do
                 let m = var 0
                     n = var 1
                     x + y = plus @@ x @@ y
@@ -172,7 +180,7 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
                     ]
 
         verifyMinus minus =
-            verify ["n","m"] $ \(@@) zero suc (==) choice -> do
+            verify ["n","m"] $ \(@@) zero suc (==) (===) choice -> do
                 let m = var 0
                     n = var 1
                     x - y = minus @@ x @@ y
@@ -185,7 +193,7 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
 
         verifyTimes times = do
             plus <- primNatPlus
-            verify ["n","m"] $ \(@@) zero suc (==) choice -> do
+            verify ["n","m"] $ \(@@) zero suc (==) (===) choice -> do
                 let m = var 0
                     n = var 1
                     x + y = plus  @@ x @@ y
@@ -203,7 +211,7 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
                     ]
 
         verifyDivSucAux dsAux =
-            verify ["k","m","n","j"] $ \(@@) zero suc (==) choice -> do
+            verify ["k","m","n","j"] $ \(@@) zero suc (==) (===) choice -> do
                 let aux k m n j = dsAux @@ k @@ m @@ n @@ j
                     k           = var 0
                     m           = var 1
@@ -215,7 +223,7 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
                 aux k m (suc n) (suc j) == aux k m n j
 
         verifyModSucAux dsAux =
-            verify ["k","m","n","j"] $ \(@@) zero suc (==) choice -> do
+            verify ["k","m","n","j"] $ \(@@) zero suc (==) (===) choice -> do
                 let aux k m n j = dsAux @@ k @@ m @@ n @@ j
                     k           = var 0
                     m           = var 1
@@ -227,7 +235,7 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
                 aux k m (suc n) (suc j) == aux (suc k) m n j
 
         verifyEquals eq =
-            verify ["n","m"] $ \(@@) zero suc (===) choice -> do
+            verify ["n","m"] $ \(@@) zero suc (==) (===) choice -> do
             true  <- primTrue
             false <- primFalse
             let x == y = eq @@ x @@ y
@@ -239,7 +247,7 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
             (zero  == suc n) === false
 
         verifyLess leq =
-            verify ["n","m"] $ \(@@) zero suc (===) choice -> do
+            verify ["n","m"] $ \(@@) zero suc (==) (===) choice -> do
             true  <- primTrue
             false <- primFalse
             let x < y = leq @@ x @@ y
@@ -255,18 +263,23 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
 
         verify' ::  TCM Term -> TCM Term -> TCM Term ->
                     [String] -> ( (Term -> Term -> Term) -> Term -> (Term -> Term) ->
+                                (Term -> Term -> TCM ()) ->
                                 (Term -> Term -> TCM ()) ->
                                 ([TCM ()] -> TCM ()) -> TCM a) -> TCM a
         verify' pNat pZero pSuc xs f = do
             nat  <- El (mkType 0) <$> pNat
             zero <- pZero
             s    <- pSuc
-            let x @@ y = x `apply` [defaultArg y]
-                x == y = noConstraints $ equalTerm nat x y
+            let x @@ y  = x `apply` [defaultArg y]
+                x == y  = noConstraints $ equalTerm nat x y
+                -- Andreas: 2013-10-21 I put primBool here on the inside
+                -- since some Nat-builtins do not require Bool-builtins
+                x === y = do bool <- El (mkType 0) <$> primBool
+                             noConstraints $ equalTerm bool x y
                 suc n  = s @@ n
                 choice = foldr1 (\x y -> x `catchError` \_ -> y)
             xs <- mapM freshName_ xs
-            addCtxs xs (domFromArg $ defaultArg nat) $ f (@@) zero suc (==) choice
+            addCtxs xs (domFromArg $ defaultArg nat) $ f (@@) zero suc (==) (===) choice
 
 
 inductiveCheck :: String -> Int -> Term -> TCM ()
@@ -310,28 +323,37 @@ bindPostulatedName builtin e m = do
   getName (A.ScopedExpr _ e) = getName e
   getName _                  = err
 
+bindBuiltinNat :: Term -> TCM ()
+bindBuiltinNat t = do
+  t' <- etaContract =<< normalise t
+  case ignoreSharing t' of
+    Def nat _ -> do
+      def <- theDef <$> getConstInfo nat
+      case def of
+        Datatype { dataCons = [c1, c2] } -> do
+          bindBuiltinName builtinNat t
+          let getArity c = arity <$> (normalise . defType =<< getConstInfo c)
+          [a1, a2] <- mapM getArity [c1, c2]
+          let (zero, suc) | a2 > a1   = (c1, c2)
+                          | otherwise = (c2, c1)
+              tnat = el primNat
+              rerange = setRange (getRange nat)
+          bindBuiltinInfo (BuiltinInfo builtinZero $ BuiltinDataCons tnat)
+                          (A.Con $ AmbQ [rerange zero])
+          bindBuiltinInfo (BuiltinInfo builtinSuc  $ BuiltinDataCons (tnat --> tnat))
+                          (A.Con $ AmbQ [rerange suc])
+        _ -> __IMPOSSIBLE__
+    _ -> __IMPOSSIBLE__
+
 bindBuiltinInfo :: BuiltinInfo -> A.Expr -> TCM ()
 bindBuiltinInfo i (A.ScopedExpr scope e) = setScope scope >> bindBuiltinInfo i e
 bindBuiltinInfo (BuiltinInfo s d) e = do
     case d of
       BuiltinData t cs -> do
-                           e' <- checkExpr e =<< t
-                           let n = length cs
-                           inductiveCheck s n e'
-                           bindBuiltinName s e'
-                           -- NAT and LEVEL must be different. (Why?)
-                           when (s `elem` [builtinNat, builtinLevel]) $ do
-                             nat   <- getBuiltin' builtinNat
-                             level <- getBuiltin' builtinLevel
-                             case (nat, level) of
-                               (Just nat, Just level) -> do
-                                  Def nat   _ <- ignoreSharing <$> normalise nat
-                                  Def level _ <- ignoreSharing <$> normalise level
-                                  when (nat == level) $ typeError $ GenericError $
-                                    builtinNat ++ " and " ++ builtinLevel ++
-                                    " have to be different types."
-                               _ -> return ()
-
+        e' <- checkExpr e =<< t
+        let n = length cs
+        inductiveCheck s n e'
+        if s == builtinNat then bindBuiltinNat e' else bindBuiltinName s e'
 
       BuiltinDataCons t -> do
 
@@ -362,7 +384,7 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
                 a   = defAbstract info
                 mcc = defCompiled info
             bindPrimitive pfname $ pf { primFunName = qx }
-            addConstant qx $ info { theDef = Primitive a pfname (Just cls) mcc }
+            addConstant qx $ info { theDef = Primitive a pfname cls mcc }
 
             -- needed? yes, for checking equations for mul
             bindBuiltinName s v
@@ -385,9 +407,11 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
           _ -> err
 
       BuiltinUnknown mt f -> do
-        e' <- maybe (fst <$> inferExpr e) (checkExpr e =<<) mt
-        f e'
-        bindBuiltinName s e'
+        (v, t) <- caseMaybe mt (inferExpr e) $ \ tcmt -> do
+          t <- tcmt
+          (,t) <$> checkExpr e t
+        f v t
+        bindBuiltinName s v
 
 -- | Bind a builtin thing to an expression.
 bindBuiltin :: String -> A.Expr -> TCM ()
@@ -397,6 +421,8 @@ bindBuiltin b e = do
     bind b e
     where
         bind b e
+            | b == builtinZero = typeError $ GenericError "Builtin ZERO does no longer exist. It is now bound by BUILTIN NATURAL"
+            | b == builtinSuc  = typeError $ GenericError "Builtin SUC does no longer exist. It is now bound by BUILTIN NATURAL"
             | b == builtinInf                                   = bindBuiltinInf e
             | b == builtinSharp                                 = bindBuiltinSharp e
             | b == builtinFlat                                  = bindBuiltinFlat e
