@@ -3,32 +3,39 @@
 module Agda.TypeChecking.Rules.LHS where
 
 import Data.Maybe
-import qualified Data.List as List
 
 import Control.Applicative
 import Control.Monad
+import Control.Monad.State
 
-import Agda.Syntax.Internal
+import Data.Traversable
+
+import Agda.Interaction.Options
+import Agda.Interaction.Options.Lenses
+
+import Agda.Syntax.Internal as I
 import Agda.Syntax.Internal.Pattern
+import Agda.Syntax.Abstract (IsProjP(..))
 import qualified Agda.Syntax.Abstract as A
+import Agda.Syntax.Abstract.Views (asView)
 import Agda.Syntax.Common
 import Agda.Syntax.Info
 import Agda.Syntax.Position
 
 import Agda.TypeChecking.Monad
+
+import Agda.TypeChecking.Conversion
+import Agda.TypeChecking.Constraints
+import Agda.TypeChecking.Datatypes
+import Agda.TypeChecking.Irrelevance
+import {-# SOURCE #-} Agda.TypeChecking.Empty
+import Agda.TypeChecking.Patterns.Abstract
 import Agda.TypeChecking.Pretty
-import Agda.TypeChecking.Records -- isRecord
+import Agda.TypeChecking.Records
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute hiding (Substitution)
 import qualified Agda.TypeChecking.Substitute as S (Substitution)
 import Agda.TypeChecking.Telescope
-import Agda.TypeChecking.Conversion
-import Agda.TypeChecking.Constraints
-import Agda.TypeChecking.Irrelevance
-import Agda.TypeChecking.Primitive (constructorForm)
-import {-# SOURCE #-} Agda.TypeChecking.Empty
--- Duplicate import??
--- import Agda.TypeChecking.Telescope (renamingR, teleArgs)
 
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Term (checkExpr)
 import Agda.TypeChecking.Rules.LHS.Problem
@@ -39,11 +46,10 @@ import Agda.TypeChecking.Rules.LHS.Implicit
 import Agda.TypeChecking.Rules.LHS.Instantiate
 import Agda.TypeChecking.Rules.Data
 
-import Agda.Interaction.Highlighting.Generate
-
+import Agda.Utils.Functor (($>))
+import Agda.Utils.Monad
 import Agda.Utils.Permutation
 import Agda.Utils.Size
-import Agda.Utils.Monad
 
 #include "../../undefined.h"
 import Agda.Utils.Impossible
@@ -51,10 +57,14 @@ import Agda.Utils.Impossible
 -- | Compute the set of flexible patterns in a list of patterns. The result is
 --   the deBruijn indices of the flexible patterns. A pattern is flexible if it
 --   is dotted or implicit.
-flexiblePatterns :: [NamedArg A.Pattern] -> TCM FlexibleVars
-flexiblePatterns nps = map fst <$> filterM (flexible . snd) (zip [0..] $ reverse ps)
+flexiblePatterns :: [A.NamedArg A.Pattern] -> TCM FlexibleVars
+flexiblePatterns nps = map setFlex <$> filterM (flexible . namedArg . snd) (zip [0..] $ reverse nps)
   where
-    ps = map namedArg nps
+    setFlex (i, Arg ai p)    = FlexibleVar (getHiding ai) (classify $ namedThing p) i
+    classify A.DotP{}        = DotFlex
+    classify A.ImplicitP{}   = ImplicitFlex
+    classify A.ConP{}        = RecordFlex
+    classify _               = __IMPOSSIBLE__
     flexible (A.DotP _ _)    = return True
     flexible (A.ImplicitP _) = return True
     flexible (A.ConP _ (A.AmbQ [c]) qs) =
@@ -64,7 +74,7 @@ flexiblePatterns nps = map fst <$> filterM (flexible . snd) (zip [0..] $ reverse
     flexible _               = return False
 
 -- | Compute the dot pattern instantiations.
-dotPatternInsts :: [NamedArg A.Pattern] -> Substitution -> [Dom Type] -> TCM [DotPatternInst]
+dotPatternInsts :: [A.NamedArg A.Pattern] -> Substitution -> [I.Dom Type] -> TCM [DotPatternInst]
 dotPatternInsts ps s as = dpi (map namedArg ps) (reverse s) as
   where
     dpi (_ : _)  []            _       = __IMPOSSIBLE__
@@ -79,7 +89,8 @@ dotPatternInsts ps s as = dpi (map namedArg ps) (reverse s) as
         A.ImplicitP _ -> dpi ps s as
         -- record pattern
         A.ConP _ (A.AmbQ [c]) qs -> do
-          Def r vs   <- ignoreSharing <$> reduce (unEl $ unDom a)
+          Def r es   <- ignoreSharing <$> reduce (unEl $ unDom a)
+          let Just vs = allApplyElims es
           (ftel, us) <- etaExpandRecord r vs u
           qs <- insertImplicitPatterns ExpandLast qs ftel
           let instTel EmptyTel _                   = []
@@ -87,12 +98,12 @@ dotPatternInsts ps s as = dpi (map namedArg ps) (reverse s) as
               instTel ExtendTel{} []               = __IMPOSSIBLE__
               bs0 = instTel ftel (map unArg us)
               -- Andreas, 2012-09-19 propagate relevance info to dot patterns
-              bs  = map (mapDomRelevance (composeRelevance (domRelevance a))) bs0
+              bs  = map (mapRelevance (composeRelevance (getRelevance a))) bs0
           dpi (map namedArg qs ++ ps) (map (Just . unArg) us ++ s) (bs ++ as)
 
         _           -> __IMPOSSIBLE__
 
-instantiatePattern :: Substitution -> Permutation -> [Arg Pattern] -> [Arg Pattern]
+instantiatePattern :: Substitution -> Permutation -> [I.NamedArg Pattern] -> [I.NamedArg Pattern]
 instantiatePattern sub perm ps
   | length sub /= length hps = error $ unlines [ "instantiatePattern:"
                                                , "  sub  = " ++ show sub
@@ -108,7 +119,7 @@ instantiatePattern sub perm ps
     merge Nothing   ps = ps
     merge (Just qs) ps = zipWith mergeA qs ps
       where
-        mergeA a1 a2 = a1 { unArg = mergeP (unArg a1) (unArg a2) }
+        mergeA a1 a2 = fmap (mergeP (namedArg a1) (namedArg a2) <$) a1
         mergeP (DotP s)  (DotP t)
           | s == t                    = DotP s
           | otherwise                 = __IMPOSSIBLE__
@@ -131,16 +142,30 @@ instantiatePattern sub perm ps
         mergeP (VarP _) (LitP _)      = __IMPOSSIBLE__
         mergeP (LitP _) (ConP _ _ _)  = __IMPOSSIBLE__
         mergeP (LitP _) (VarP _)      = __IMPOSSIBLE__
+        mergeP (ProjP x) (ProjP y)
+          | x == y                    = ProjP x
+          | otherwise                 = __IMPOSSIBLE__
+        mergeP ProjP{} _              = __IMPOSSIBLE__
+        mergeP _       ProjP{}        = __IMPOSSIBLE__
 
 -- | Check if a problem is solved. That is, if the patterns are all variables.
 isSolvedProblem :: Problem -> Bool
-isSolvedProblem = all (isVar . snd . asView . namedArg) . problemInPat
+isSolvedProblem problem = null (restPats $ problemRest problem) &&
+    all (isSolved . snd . asView . namedArg) (problemInPat problem)
   where
+    isSolved (A.DefP _ _ []) = False  -- projection pattern
+    isSolved (A.VarP _)      = True
+    isSolved (A.WildP _)     = True
+    isSolved (A.ImplicitP _) = True
+    isSolved (A.AbsurdP _)   = True
+    isSolved _               = False
+{-
     isVar (A.VarP _)      = True
     isVar (A.WildP _)     = True
     isVar (A.ImplicitP _) = True
     isVar (A.AbsurdP _)   = True
     isVar _               = False
+-}
 
 -- | For each user-defined pattern variable in the 'Problem', check
 -- that the corresponding data type (if any) does not contain a
@@ -149,14 +174,6 @@ isSolvedProblem = all (isVar . snd . asView . namedArg) . problemInPat
 --
 -- Precondition: The problem has to be solved.
 
-{-
-noShadowingOfConstructors
-  :: A.Clause
-     -- ^ The entire clause (used for error reporting).
-  -> Problem -> TCM ()
-noShadowingOfConstructors c problem =
-  traceCall (CheckPatternShadowing c) $ do
--}
 noShadowingOfConstructors
   :: (Maybe r -> Call) -- ^ Trace, e.g., @CheckPatternShadowing clause@
   -> Problem -> TCM ()
@@ -165,30 +182,31 @@ noShadowingOfConstructors mkCall problem =
     let pat = map (snd . asView . namedArg) $
                   problemInPat problem
         tel = map (unEl . snd . unDom) $ telToList $ problemTel problem
-    zipWithM' noShadowing pat tel
+    zipWithM noShadowing pat tel -- TODO: does not work for flexible arity and projection patterns
     return ()
   where
   noShadowing (A.WildP     {}) t = return ()
   noShadowing (A.AbsurdP   {}) t = return ()
   noShadowing (A.ImplicitP {}) t = return ()
   noShadowing (A.ConP      {}) t = return ()  -- only happens for eta expanded record patterns
-  noShadowing (A.DefP      {}) t = __IMPOSSIBLE__
+  noShadowing (A.DefP      {}) t = return ()  -- projection pattern
   noShadowing (A.AsP       {}) t = __IMPOSSIBLE__
   noShadowing (A.DotP      {}) t = __IMPOSSIBLE__
   noShadowing (A.LitP      {}) t = __IMPOSSIBLE__
   noShadowing (A.PatternSynP {}) t = __IMPOSSIBLE__
   noShadowing (A.VarP x)       t = do
-    t <- normalise t
+    reportSDoc "tc.lhs.shadow" 30 $ vcat
+      [ text $ "checking whether pattern variable " ++ show x ++ " shadows a constructor"
+      , nest 2 $ text "type of variable =" <+> prettyTCM t
+      ]
+    reportSLn "tc.lhs.shadow" 70 $ "  t = " ++ show t
+    t <- reduce t
     case t of
       Def t _ -> do
         d <- theDef <$> getConstInfo t
         case d of
           Datatype { dataCons = cs } -> do
-            let ns = map (\c -> (c, A.nameConcrete $ A.qnameName c)) cs
-                match x = catMaybes $
-                            map (\(c, n) -> if A.nameConcrete x == n
-                                            then Just c else Nothing) ns
-            case match x of
+            case filter ((A.nameConcrete x ==) . A.nameConcrete . A.qnameName) cs of
               []      -> return ()
               (c : _) -> setCurrentRange (getRange x) $
                 typeError $ PatternShadowsConstructor x c
@@ -215,7 +233,7 @@ noShadowingOfConstructors mkCall problem =
 
 -- | Check that a dot pattern matches it's instantiation.
 checkDotPattern :: DotPatternInst -> TCM ()
-checkDotPattern (DPI e v (Dom h r a)) =
+checkDotPattern (DPI e v (Dom info a)) =
   traceCall (CheckDotPattern e v) $ do
   reportSDoc "tc.lhs.dot" 15 $
     sep [ text "checking dot pattern"
@@ -223,7 +241,7 @@ checkDotPattern (DPI e v (Dom h r a)) =
         , nest 2 $ text "=" <+> prettyTCM v
         , nest 2 $ text ":" <+> prettyTCM a
         ]
-  applyRelevanceToContext r $ do
+  applyRelevanceToContext (argInfoRelevance info) $ do
     u <- checkExpr e a
     -- Should be ok to do noConstraints here
     noConstraints $ equalTerm a u v
@@ -233,13 +251,16 @@ checkDotPattern (DPI e v (Dom h r a)) =
 --   the same size as the pattern list.
 --   There could also be 'A.ConP's resulting from eta expanded implicit record
 --   patterns.
-bindLHSVars :: [NamedArg A.Pattern] -> Telescope -> TCM a -> TCM a
-bindLHSVars []       (ExtendTel _ _)   _   = __IMPOSSIBLE__
+bindLHSVars :: [A.NamedArg A.Pattern] -> Telescope -> TCM a -> TCM a
+bindLHSVars []        tel@ExtendTel{}  _   = do
+  reportSDoc "impossible" 10 $
+    text "bindLHSVars: no patterns left, but tel =" <+> prettyTCM tel
+  __IMPOSSIBLE__
 bindLHSVars (_ : _)   EmptyTel         _   = __IMPOSSIBLE__
 bindLHSVars []        EmptyTel         ret = ret
 bindLHSVars (p : ps) (ExtendTel a tel) ret =
   case namedArg p of
-    A.VarP x      -> addCtx x a $ bindLHSVars ps (absBody tel) ret
+    A.VarP x      -> addContext (x, a) $ bindLHSVars ps (absBody tel) ret
     A.WildP _     -> bindDummy (absName tel)
     A.ImplicitP _ -> bindDummy (absName tel)
     A.AbsurdP pi  -> do
@@ -248,10 +269,13 @@ bindLHSVars (p : ps) (ExtendTel a tel) ret =
       -- OLD CODE: isReallyEmptyType $ unArg a
       bindDummy (absName tel)
     A.ConP _ (A.AmbQ [c]) qs -> do -- eta expanded record pattern
-      Def r vs <- reduce (unEl $ unDom a)
+      Def r es <- reduce (unEl $ unDom a)
+      let Just vs = allApplyElims es
       ftel     <- (`apply` vs) <$> getRecordFieldTypes r
+      con      <- getConHead c
       let n   = size ftel
-          eta = Con c [ Var i [] <$ (namedThing <$> q) | (q, i) <- zip qs [n - 1, n - 2..0] ]
+          eta = Con con [ var i <$ (namedThing <$> setArgColors [] q) | (q, i) <- zip qs [n - 1, n - 2..0] ]
+          -- ^ TODO guilhem
       bindLHSVars (qs ++ ps) (ftel `abstract` absApp (raise (size ftel) tel) eta) ret
     A.ConP{}        -> __IMPOSSIBLE__
     A.DefP{}        -> __IMPOSSIBLE__
@@ -260,11 +284,9 @@ bindLHSVars (p : ps) (ExtendTel a tel) ret =
     A.LitP{}        -> __IMPOSSIBLE__
     A.PatternSynP{} -> __IMPOSSIBLE__
     where
-      name "_" = freshNoName_
-      name s   = freshName_ ("." ++ s)
       bindDummy s = do
-        x <- name s
-        addCtx x a $ bindLHSVars ps (absBody tel) ret
+        x <- if isUnderscore s then freshNoName_ else freshName_ ("." ++ argNameToString s)
+        addContext (x, a) $ bindLHSVars ps (absBody tel) ret
 
 -- | Bind as patterns
 bindAsPatterns :: [AsBinding] -> TCM a -> TCM a
@@ -274,40 +296,51 @@ bindAsPatterns (AsB x v a : asb) ret = do
     sep [ text ":" <+> prettyTCM a
         , text "=" <+> prettyTCM v
         ]
-  addLetBinding Relevant x v a $ bindAsPatterns asb ret
+  addLetBinding defaultArgInfo x v a $ bindAsPatterns asb ret
+
 
 -- | Check a LHS. Main function.
+--
+--   @checkLeftHandSide a ps a ret@ checks that user patterns @ps@ eliminate
+--   the type @a@ of the defined function, and calls continuation @ret@
+--   if successful.
+
 checkLeftHandSide
   :: (Maybe r -> Call)
      -- ^ Trace, e.g. @CheckPatternShadowing clause@
-  -> [NamedArg A.Pattern]
+  -> Maybe QName
+     -- ^ The name of the definition we are checking.
+  -> [A.NamedArg A.Pattern]
      -- ^ The patterns.
   -> Type
      -- ^ The expected type @a = Γ → b@.
-  -> (Maybe Telescope   -- Γ : The types of the patterns.
-                        -- 'Nothing' if more patterns than domain types in @a@.
-                        -- Used only to construct a @with@ function; see 'stripwithClausePatterns'.
-      -> Telescope      -- Δ : The types of the pattern variables.
-      -> S.Substitution -- σ : The patterns in form of a substitution Δ ⊢ σ : Γ
-      -> [String]       -- Names for the variables in Δ, for binding the body.
-      -> [Arg Pattern]  -- The patterns in internal syntax.
-      -> Type           -- The type of the body. Is @bσ@ if @Γ@ is defined.
-      -> Permutation    -- The permutation from pattern vars to @Δ@.
+  -> (Maybe Telescope         -- Γ : The types of the patterns.
+                              -- 'Nothing' if more patterns than domain types in @a@.
+                              -- Used only to construct a @with@ function; see 'stripwithClausePatterns'.
+      -> Telescope            -- Δ : The types of the pattern variables.
+      -> S.Substitution       -- σ : The patterns in form of a substitution Δ ⊢ σ : Γ
+      -> [ArgName]            -- Names for the variables in Δ, for binding the body.
+      -> [I.NamedArg Pattern] -- The patterns in internal syntax.
+      -> I.Arg Type           -- The type of the body. Is @bσ@ if @Γ@ is defined.
+                              -- 'Irrelevant' to indicate the rhs must be checked
+                              -- in irrelevant mode.
+      -> Permutation          -- The permutation from pattern vars to @Δ@.
       -> TCM a)
      -- ^ Continuation.
   -> TCM a
-checkLeftHandSide c ps a ret = do
+checkLeftHandSide c f ps a ret = do
   problem <- problemFromPats ps a
-  unless (noProblemRest problem) $ typeError $ TooManyArgumentsInLHS a
-  let (Problem _ _ gamma (ProblemRest _ b)) = problem
-      mgamma = if noProblemRest problem then Just gamma else Nothing
-      st     = LHSState problem idS [] []
+  -- Andreas, 2013-03-15 deactivating the following test allows
+  -- flexible arity
+  -- unless (noProblemRest problem) $ typeError $ TooManyArgumentsInLHS a
+  let mgamma = if noProblemRest problem then Just $ problemTel problem else Nothing
 
   -- doing the splits:
-  LHSState (Problem ps (perm, qs) delta rest) sigma dpi asb <- checkLHS st
+  LHSState (Problem ps (perm, qs) delta rest) sigma dpi asb
+    <- checkLHS $ LHSState problem idS [] []
+
   unless (null $ restPats rest) $ typeError $ TooManyArgumentsInLHS a
 
-  -- let b' = applySubst sigma b
   let b' = restType rest
 
   noPatternMatchingOnCodata qs
@@ -323,7 +356,8 @@ checkLeftHandSide c ps a ret = do
            , text "qs    = " <+> text (show qs)
 	   ]
          ]
-  bindLHSVars ps delta $ bindAsPatterns asb $ do
+  bindLHSVars (filter (isNothing . isProjP) ps) delta $ bindAsPatterns asb $ do
+    reportSDoc "tc.lhs.top" 10 $ text "bound pattern variables"
     reportSDoc "tc.lhs.top" 10 $ nest 2 $ text "type  = " <+> prettyTCM b'
 
     -- Check dot patterns
@@ -331,8 +365,9 @@ checkLeftHandSide c ps a ret = do
 
     let rho = renamingR perm -- I'm not certain about this...
         Perm n _ = perm
-        xs  = [ "h" ++ show n | n <- [0..n - 1] ]
-    ret mgamma delta rho xs qs b' perm
+        xs  = [ stringToArgName $ "h" ++ show n | n <- [0..n - 1] ]
+    applyRelevanceToContext (getRelevance b') $ do
+      ret mgamma delta rho xs qs b' perm
   where
     -- the loop: split at a variable in the problem until problem is solved
     checkLHS :: LHSState -> TCM LHSState
@@ -343,14 +378,38 @@ checkLeftHandSide c ps a ret = do
           noShadowingOfConstructors c problem
           return $ st { lhsProblem = problem }
         else do
-        sp <- splitProblem problem
+        unlessM (optPatternMatching <$> gets getPragmaOptions) $
+          typeError $ GenericError $ "Pattern matching is disabled"
+        sp <- splitProblem f problem
         reportSDoc "tc.lhs.split" 20 $ text "splitting completed"
         case sp of
-          Left NothingToSplit   -> nothingToSplitError problem
-          Left (SplitPanic err) -> __IMPOSSIBLE__
+          Left NothingToSplit   -> do
+            reportSLn "tc.lhs.split" 50 $ "checkLHS: nothing to split in problem " ++ show problem
+            nothingToSplitError problem
+          Left (SplitPanic err) -> do
+            reportSLn "impossible" 10 $ "checkLHS: panic: " ++ err
+            __IMPOSSIBLE__
+
+          -- Split problem rest (projection pattern)
+          Right (SplitRest projPat projType) -> do
+
+            -- Compute the new problem
+            let Problem ps1 (iperm, ip) delta (ProblemRest (p:ps2) b) = problem
+                -- ps'      = ps1 ++ [p]
+                ps'      = ps1 -- drop the projection pattern (already splitted)
+                rest     = ProblemRest ps2 (projPat $> projType)
+                ip'      = ip ++ [fmap (Named Nothing . ProjP) projPat]
+                problem' = Problem ps' (iperm, ip') delta rest
+            -- Jump the trampolin
+            st' <- updateProblemRest (LHSState problem' sigma dpi asb)
+            -- If the field is irrelevant, we need to continue in irr. cxt.
+            -- (see Issue 939).
+            applyRelevanceToContext (getRelevance projPat) $ do
+              checkLHS st'
+
 
           -- Split on literal pattern
-          Right (Split p0 xs (Arg _ _ (LitFocus lit iph hix a)) p1) -> do
+          Right (Split p0 xs (Arg _ (LitFocus lit iph hix a)) p1) -> do
 
             -- plug the hole with a lit pattern
             let ip    = plugHole (LitP lit) iph
@@ -378,8 +437,9 @@ checkLeftHandSide c ps a ret = do
             checkLHS st'
 
           -- Split on constructor pattern
-          Right (Split p0 xs (Arg h rel
+          Right (Split p0 xs (Arg info
                   ( Focus { focusCon      = c
+                          , focusImplicit = impl
                           , focusConArgs  = qs
                           , focusRange    = r
                           , focusOutPat   = iph
@@ -390,17 +450,17 @@ checkLeftHandSide c ps a ret = do
                           , focusType     = a
                           }
                   )) p1
-                ) -> traceCall (CheckPattern (A.ConP (PatRange r) (A.AmbQ [c]) qs)
+                ) -> traceCall (CheckPattern (A.ConP (ConPatInfo impl $ PatRange r) (A.AmbQ [c]) qs)
                                              (problemTel p0)
-                                             (El Prop $ Def d $ vs ++ ws)) $ do
+                                             (El Prop $ Def d $ map Apply $ vs ++ ws)) $ do
 
             let delta1 = problemTel p0
-            let typeOfSplitVar = Arg h rel a
+            let typeOfSplitVar = Arg info a
 
             reportSDoc "tc.lhs.split" 10 $ sep
               [ text "checking lhs"
               , nest 2 $ text "tel =" <+> prettyTCM (problemTel problem)
-              , nest 2 $ text "rel =" <+> (text $ show rel)
+              , nest 2 $ text "rel =" <+> (text $ show $ argInfoRelevance info)
               ]
 
             reportSDoc "tc.lhs.split" 15 $ sep
@@ -413,9 +473,13 @@ checkLeftHandSide c ps a ret = do
                 ]
               ]
 
+{-
+            c <- conSrcCon . theDef <$> getConstInfo c
             Con c' [] <- ignoreSharing <$> (constructorForm =<< normalise (Con c []))
             c  <- return $ c' `withRangeOf` c
-            ca <- defType <$> getConstInfo c
+-}
+            c <- (`withRangeOf` c) <$> getConForm c
+            ca <- defType <$> getConInfo c
 
             reportSDoc "tc.lhs.split" 20 $ nest 2 $ vcat
               [ text "ca =" <+> prettyTCM ca
@@ -428,7 +492,8 @@ checkLeftHandSide c ps a ret = do
             -- It will end in an application of the datatype
             (gamma', ca, d', us) <- do
               TelV gamma' ca@(El _ def) <- telView a
-              let Def d' us = ignoreSharing def
+              let Def d' es = ignoreSharing def
+                  Just us   = allApplyElims es
               return (gamma', ca, d', us)
 
             -- This should be the same datatype as we split on
@@ -441,7 +506,7 @@ checkLeftHandSide c ps a ret = do
 -}
 
             -- Andreas 2010-09-07  propagate relevance info to new vars
-            gamma' <- return $ fmap (applyRelevance rel) gamma'
+            gamma' <- return $ fmap (applyRelevance $ argInfoRelevance info) gamma'
 {-
             reportSDoc "tc.lhs.top" 20 $ nest 2 $ vcat
               [ text "gamma' =" <+> text (show gamma')
@@ -451,7 +516,7 @@ checkLeftHandSide c ps a ret = do
             qs' <- insertImplicitPatterns ExpandLast qs gamma'
 
             unless (size qs' == size gamma') $
-              typeError $ WrongNumberOfConstructorArguments c (size gamma') (size qs')
+              typeError $ WrongNumberOfConstructorArguments (conName c) (size gamma') (size qs')
 
             let gamma = useNamesFromPattern qs' gamma'
 
@@ -461,6 +526,9 @@ checkLeftHandSide c ps a ret = do
             -- Compute the flexible variables
             flex <- flexiblePatterns (problemInPat p0 ++ qs')
 
+            -- Compute the constructor indices by dropping the parameters
+            let us' = drop (size vs) us
+
 	    reportSDoc "tc.lhs.top" 15 $ addCtxTel delta1 $
 	      sep [ text "preparing to unify"
 		  , nest 2 $ vcat
@@ -469,13 +537,14 @@ checkLeftHandSide c ps a ret = do
 		    , text "gamma  =" <+> prettyTCM gamma
 		    , text "gamma' =" <+> prettyTCM gamma'
 		    , text "vs     =" <+> brackets (fsep $ punctuate comma $ map prettyTCM vs)
+		    , text "us'    =" <+> brackets (fsep $ punctuate comma $ map prettyTCM us')
 		    , text "ws     =" <+> brackets (fsep $ punctuate comma $ map prettyTCM ws)
 		    ]
 		  ]
 
             -- Unify constructor target and given type (in Δ₁Γ)
             sub0 <- addCtxTel (delta1 `abstract` gamma) $
-                    unifyIndices_ flex (raise (size gamma) da) (drop (size vs) us) (raise (size gamma) ws)
+                    unifyIndices_ flex (raise (size gamma) da) us' (raise (size gamma) ws)
 
             -- We should substitute c ys for x in Δ₂ and sigma
             let ys     = teleArgs gamma
@@ -505,11 +574,11 @@ checkLeftHandSide c ps a ret = do
             -- Thus, it has to be raised by 1 (the "hole" variable)
             -- plus the length of delta2 (the variables coming after the hole).
             storedPatternType <- ifM (isJust <$> isRecord d)
-              (return $ Just $ raise (1 + size delta2) $ typeOfSplitVar)
+              (return $ Just (impl, raise (1 + size delta2) typeOfSplitVar))
               (return $ Nothing)
 
             -- Plug the hole in the out pattern with c ys
-            let ysp = map (argFromDom . fmap (VarP . fst)) $ telToList gamma
+            let ysp = map (argFromDom . fmap (namedVarP . fst)) $ telToList gamma
                 ip  = plugHole (ConP c storedPatternType ysp) iph
                 ip0 = applySubst rho0 ip
 
@@ -628,17 +697,18 @@ checkLeftHandSide c ps a ret = do
 
 -- Ensures that we are not performing pattern matching on codata.
 
-noPatternMatchingOnCodata :: [Arg Pattern] -> TCM ()
-noPatternMatchingOnCodata = mapM_ (check . unArg)
+noPatternMatchingOnCodata :: [I.NamedArg Pattern] -> TCM ()
+noPatternMatchingOnCodata = mapM_ (check . namedArg)
   where
   check (VarP {})   = return ()
   check (DotP {})   = return ()
+  check (ProjP{})   = return ()
   check (LitP {})   = return ()  -- Literals are assumed not to be coinductive.
-  check (ConP q _ ps) = do
-    TelV _ t <- telView' . defType <$> getConstInfo q
+  check (ConP con _ ps) = do
+    TelV _ t <- telView' . defType <$> do getConstInfo $ conName con
     c <- isCoinductive t
     case c of
       Nothing    -> __IMPOSSIBLE__
-      Just False -> mapM_ (check . unArg) ps
+      Just False -> mapM_ (check . namedArg) ps
       Just True  -> typeError $
-        GenericError "Pattern matching on codata is not allowed"
+        GenericError "Pattern matching on coinductive types is not allowed"
