@@ -1,10 +1,18 @@
-{-# LANGUAGE CPP, MultiParamTypeClasses, FlexibleInstances, TupleSections,
-    GeneralizedNewtypeDeriving,
-    DeriveDataTypeable, DeriveFunctor, DeriveFoldable, DeriveTraversable #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveTraversable #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE TupleSections #-}
+{-# LANGUAGE TypeSynonymInstances #-}
 
 module Agda.TypeChecking.Rules.LHS.Unify where
 
-import Control.Arrow ((***), (&&&))
+import Control.Arrow ((***))
 import Control.Applicative hiding (empty)
 import Control.Monad.State
 import Control.Monad.Reader
@@ -19,15 +27,19 @@ import Data.Typeable (Typeable)
 import Data.Foldable (Foldable)
 import Data.Traversable (Traversable,traverse)
 
+import Agda.Interaction.Options (optInjectiveTypeConstructors)
+
 import Agda.Syntax.Common
-import Agda.Syntax.Internal
+import Agda.Syntax.Internal as I
 import Agda.Syntax.Literal
 import Agda.Syntax.Position
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Exception
+import Agda.TypeChecking.Monad.Builtin (constructorForm)
 import Agda.TypeChecking.Conversion -- equalTerm
 import Agda.TypeChecking.Constraints
+import Agda.TypeChecking.DropArgs
 import Agda.TypeChecking.Level (reallyUnLevelView)
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Pretty
@@ -36,24 +48,37 @@ import qualified Agda.TypeChecking.Substitute as S
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Free
 import Agda.TypeChecking.Records
-import Agda.TypeChecking.Primitive (constructorForm)
 import Agda.TypeChecking.MetaVars (assignV, newArgsMetaCtx)
 import Agda.TypeChecking.EtaContract
-import Agda.Interaction.Options (optInjectiveTypeConstructors)
+import Agda.Interaction.Options (optInjectiveTypeConstructors, optWithoutK)
 
 import Agda.TypeChecking.Rules.LHS.Problem
+-- import Agda.TypeChecking.SyntacticEquality
+
+import Agda.Utils.Maybe
+import Agda.Utils.Size
+import Agda.Utils.Monad
 
 #include "../../../undefined.h"
 import Agda.Utils.Impossible
-import Agda.Utils.Size
 
 newtype Unify a = U { unUnify :: ReaderT UnifyEnv (WriterT UnifyOutput (ExceptionT UnifyException (StateT UnifyState TCM))) a }
   deriving (Monad, MonadIO, Functor, Applicative, MonadException UnifyException, MonadWriter UnifyOutput)
 
+instance MonadTCM Unify where
+  liftTCM = U . lift . lift . lift . lift
+
+instance MonadState TCState Unify where
+  get = liftTCM $ get
+  put = liftTCM . put
+
 instance MonadReader TCEnv Unify where
   ask = U $ ReaderT $ \ _ -> ask
   local cont (U (ReaderT f)) = U $ ReaderT $ \ a -> local cont (f a)
 
+instance HasConstInfo Unify where
+  getConstInfo = U . lift . lift . lift . lift . getConstInfo
+
 data UnifyMayPostpone = MayPostpone | MayNotPostpone
 
 type UnifyEnv = UnifyMayPostpone
@@ -101,6 +126,8 @@ type Sub = Map Nat Term
 data UnifyException
   = ConstructorMismatch Type Term Term
   | StronglyRigidOccurrence Type Term Term
+  | UnclearOccurrence Type Term Term
+  | WithoutKException Type Term Term
   | GenericUnifyException String
 
 instance Error UnifyException where
@@ -113,6 +140,11 @@ data UnifyState = USt { uniSub	  :: Sub
 
 emptyUState = USt Map.empty []
 
+-- | Throw-away error message.
+projectionMismatch :: QName -> QName -> Unify a
+projectionMismatch f f' = throwException $ GenericUnifyException $
+  "projections " ++ show f ++ " and " ++ show f' ++ " do not match"
+
 constructorMismatch :: Type -> Term -> Term -> Unify a
 constructorMismatch a u v = throwException $ ConstructorMismatch a u v
 
@@ -120,13 +152,6 @@ constructorMismatchHH :: TypeHH -> Term -> Term -> Unify a
 constructorMismatchHH aHH = constructorMismatch (leftHH aHH)
   -- do not report heterogenity
 
-instance MonadState TCState Unify where
-  get = U . lift . lift . lift . lift $ get
-  put = U . lift . lift . lift . lift . put
-
-instance MonadTCM Unify where
-  liftTCM = U . lift . lift . lift . lift
-
 instance Subst Equality where
   applySubst rho (Equal a s t) =
     Equal (applySubst rho a) (applySubst rho s) (applySubst rho t)
@@ -142,6 +167,16 @@ checkEqualities eqs = noConstraints $ mapM_ checkEq eqs
   where
     checkEq (Equal (Hom a) s t) = equalTerm a s t
     checkEq (Equal (Het a1 a2) s t) = typeError $ HeterogeneousEquality s a1 t a2
+    -- Andreas, 2014-03-03:  Alternatively, one could try to get back
+    -- to a homogeneous situation.  Unless there is a case where this
+    -- actually helps, I leave it deactivated.
+    -- KEEP:
+    --
+    -- checkEq (Equal (Het a1 a2) s t) = do
+    --     noConstraints $ do
+    --       equalType a1 a2
+    --       equalTerm a1 s t
+    --   `catchError` \ _ -> typeError $ HeterogeneousEquality s a1 t a2
 
 -- | Force equality now instead of postponing it using 'addEquality'.
 checkEquality :: Type -> Term -> Term -> TCM ()
@@ -152,19 +187,25 @@ checkEquality a u v = noConstraints $ equalTerm a u v
 --   so we can throw them away and flag "dirty".
 checkEqualityHH :: TypeHH -> Term -> Term -> Unify ()
 checkEqualityHH (Hom a) u v = do
-    ok <- liftTCM $ noConstraints (True <$ equalTerm a u v)  -- no constraints left
-            `catchError` \ err -> return False
-    if ok then return () else addEquality a u v
+    ok <- liftTCM $ tryConversion $ equalTerm a u v  -- no constraints left
+    -- Jesper, 2013-11-21: Refuse to solve reflexive equations when --without-K is enabled
+    if ok
+      then (whenM (liftTCM $ optWithoutK <$> pragmaOptions)
+           (throwException $ WithoutKException a u v))
+      else (addEquality a u v)
 checkEqualityHH aHH@(Het a1 a2) u v = -- reportPostponing -- enter "dirty" mode
     addEqualityHH aHH u v -- postpone, enter "dirty" mode
 
 -- | Check whether heterogeneous situation is really homogeneous.
 --   If not, give up.
 forceHom :: TypeHH -> TCM Type
-forceHom (Hom a) = return a
-forceHom (Het a1 a2) = do
-  noConstraints $ equalType a1 a2
-  return a1
+forceHom (Hom a)     = return a
+forceHom (Het a1 a2) = a1 <$ do noConstraints $ equalType a1 a2
+
+-- | Check whether heterogeneous situation is really homogeneous.
+--   If not, return Nothing.
+makeHom :: TypeHH -> TCM (Maybe Type)
+makeHom aHH = (Just <$> forceHom aHH) `catchError` \ err -> return Nothing
 
 addEquality :: Type -> Term -> Term -> Unify ()
 addEquality a = addEqualityHH (Hom a)
@@ -198,13 +239,13 @@ occursCheck i u a = do
     -- any other recursive occurrence leads to unclear situation
     _             -> do
       liftTCM $ reportSDoc "tc.lhs.unify" 20 $ prettyTCM v <+> text "occurs in" <+> prettyTCM u
-      typeError $ UnequalTerms CmpEq v u a
+      throwException $ UnclearOccurrence a v u
 
 -- | Assignment with preceding occurs check.
 (|->) :: Nat -> (Term, Type) -> Unify ()
 i |-> (u, a) = do
   occursCheck i u a
-  liftTCM $ reportSDoc "tc.lhs.unify" 15 $ prettyTCM (var i) <+> text ":=" <+> prettyTCM u
+  liftTCM $ reportSDoc "tc.lhs.unify.assign" 15 $ prettyTCM (var i) <+> text ":=" <+> prettyTCM u
   modSub $ Map.insert i (killRange u)
   -- Apply substitution to itself (issue 552)
   rho  <- onSub id
@@ -226,14 +267,16 @@ class UReduce t where
 instance UReduce Term where
   ureduce u = doEtaContractImplicit $ do
     rho <- onSub makeSubstitution
-    liftTCM $ etaContract =<< normalise (applySubst rho u)
+-- Andreas, 2013-10-24 the following call to 'normalise' is problematic
+-- (see issue 924).  Instead, we only normalize if unifyAtomHH is undecided.
+--    liftTCM $ etaContract =<< normalise (applySubst rho u)
 -- Andreas, 2011-06-22, fix related to issue 423
 -- To make eta contraction work better, I switched reduce to normalise.
 -- I hope the performance penalty is not big (since we are dealing with
 -- l.h.s. terms only).
 -- A systematic solution would make unification type-directed and
 -- eta-insensitive...
---   liftTCM $ etaContract =<< reduce (applySubst rho u)
+    liftTCM $ etaContract =<< reduce (applySubst rho u)
 
 instance UReduce Type where
   ureduce (El s t) = El s <$> ureduce t
@@ -242,10 +285,38 @@ instance UReduce t => UReduce (HomHet t) where
   ureduce (Hom t)     = Hom <$> ureduce t
   ureduce (Het t1 t2) = Het <$> ureduce t1 <*> ureduce t2
 
+-- Andreas, 2014-03-03 A variant of ureduce that tries to get back
+-- to a homogeneous situation by checking syntactic equality.
+-- Did not solve issue 1071, so I am reverting to the old ureduce.
+-- However, KEEP THIS as an alternative to reconsider.
+-- Remember to import TypeChecking.SyntacticEquality!
+--
+-- instance (SynEq t, UReduce t) => UReduce (HomHet t) where
+--   ureduce (Hom t)     = Hom <$> ureduce t
+--   ureduce (Het t1 t2) = do
+--     t1 <- ureduce t1
+--     t2 <- ureduce t2
+--     ((t1,t2),equal) <- liftTCM $ checkSyntacticEquality t1 t2
+--     -- BRITTLE: syntactic equality only
+--     return $ if equal then Hom t1 else Het t1 t2
+
 instance UReduce t => UReduce (Maybe t) where
   ureduce Nothing = return Nothing
   ureduce (Just t) = Just <$> ureduce t
 
+instance (UReduce a, UReduce b) => UReduce (a, b) where
+  ureduce (a, b) = (,) <$> ureduce a <*> ureduce b
+
+instance (UReduce a, UReduce b, UReduce c) => UReduce (a, b, c) where
+  ureduce (a, b, c) = (\x y z -> (x, y, z)) <$> ureduce a <*> ureduce b <*> ureduce c
+
+instance (UReduce a) => UReduce (I.Arg a) where
+  ureduce (Arg c e) = Arg c <$> ureduce e
+
+instance (UReduce a) => UReduce [ a ] where
+  ureduce = sequence . (map ureduce)
+
+
 -- | Take a substitution σ and ensure that no variables from the domain appear
 --   in the targets. The context of the targets is not changed.
 --   TODO: can this be expressed using makeSubstitution and applySubst?
@@ -298,8 +369,9 @@ instance (PrettyTCM a) => PrettyTCM (HomHet a) where
 type TermHH    = HomHet Term
 type TypeHH    = HomHet Type
 --type FunViewHH = FunV TypeHH
-type TelHH     = Tele (Dom TypeHH)
+type TelHH     = Tele (I.Dom TypeHH)
 type TelViewHH = TelV TypeHH
+type ArgsHH    = HomHet Args
 
 absAppHH :: SubstHH t tHH => Abs t -> TermHH -> tHH
 absAppHH (Abs   _ t) u = substHH u t
@@ -341,11 +413,11 @@ instance SubstHH Type (HomHet Type) where
   -- terms we can match on
   trivialHH = Hom
 
-instance SubstHH a b => SubstHH (Arg a) (Arg b) where
+instance SubstHH a b => SubstHH (I.Arg a) (I.Arg b) where
   substUnderHH n u = fmap $ substUnderHH n u
   trivialHH = fmap trivialHH
 
-instance SubstHH a b => SubstHH (Dom a) (Dom b) where
+instance SubstHH a b => SubstHH (I.Dom a) (I.Dom b) where
   substUnderHH n u = fmap $ substUnderHH n u
   trivialHH = fmap trivialHH
 
@@ -364,7 +436,19 @@ instance SubstHH a b => SubstHH (Tele a) (Tele b) where
   trivialHH = fmap trivialHH
 
 -- | Unify indices.
-unifyIndices_ :: MonadTCM tcm => FlexibleVars -> Type -> [Arg Term] -> [Arg Term] -> tcm Substitution
+--
+--   In @unifyIndices_ flex a us vs@,
+--
+--   @a@ is the type eliminated by @us@ and @vs@
+--     (usally the type of a constructor),
+--     need not be reduced,
+--
+--   @us@ and @vs@ are the argument lists to unify,
+--
+--   @flex@ is the set of flexible (instantiable) variabes in @us@ and @vs@.
+--
+--   The result is the most general unifier of @us@ and @vs@.
+unifyIndices_ :: MonadTCM tcm => FlexibleVars -> Type -> Args -> Args -> tcm Substitution
 unifyIndices_ flex a us vs = liftTCM $ do
   r <- unifyIndices flex a us vs
   case r of
@@ -372,7 +456,7 @@ unifyIndices_ flex a us vs = liftTCM $ do
     DontKnow err  -> throwError err
     NoUnify a u v -> typeError $ UnequalTerms CmpEq u v a
 
-unifyIndices :: MonadTCM tcm => FlexibleVars -> Type -> [Arg Term] -> [Arg Term] -> tcm UnificationResult
+unifyIndices :: MonadTCM tcm => FlexibleVars -> Type -> Args -> Args -> tcm UnificationResult
 unifyIndices flex a us vs = liftTCM $ do
     a <- reduce a
     reportSDoc "tc.lhs.unify" 10 $
@@ -394,14 +478,21 @@ unifyIndices flex a us vs = liftTCM $ do
       Left (ConstructorMismatch     a u v)  -> return $ NoUnify a u v
       -- Andreas 2011-04-14:
       Left (StronglyRigidOccurrence a u v)  -> return $ NoUnify a u v
-      Left (GenericUnifyException     err)  -> fail err
+      Left (UnclearOccurrence a u v)        -> typeError $ UnequalTerms CmpEq u v a
+      Left (WithoutKException       a u v)  -> typeError $ WithoutKError a u v
+      Left (GenericUnifyException     err)  -> typeError $ GenericError err
       Right _                               -> do
         checkEqualities $ applySubst (makeSubstitution s) eqs
-        let n = maximum $ (-1) : flex
+        let n = maximum $ (-1) : flex'
         return $ Unifies $ flattenSubstitution [ Map.lookup i s | i <- [0..n] ]
-  `catchError` \err -> return $ DontKnow err
+  `catchError` \err -> case err of
+     TypeError _ (Closure {clValue = WithoutKError{}}) -> throwError err
+     _                                                 -> return $ DontKnow err
   where
-    flexible i = i `elem` flex
+    flex'      = map flexVar flex
+    flexible i = i `elem` flex'
+    findFlexible i = find ((i ==) . flexVar) flex
+    flexibleHid  i = fmap getHiding $ findFlexible i
 
     flexibleTerm (Var i []) = flexible i
     flexibleTerm (Shared p) = flexibleTerm (derefPtr p)
@@ -415,8 +506,8 @@ unifyIndices flex a us vs = liftTCM $ do
     unifyConstructorArgs ::
          TypeHH  -- ^ The ureduced type of the constructor, instantiated to the parameters.
                  --   Possibly heterogeneous, since pars of lhs and rhs might differ.
-      -> [Arg Term]  -- ^ the arguments of the constructor (lhs)
-      -> [Arg Term]  -- ^ the arguments of the constructor (rhs)
+      -> [I.Arg Term]  -- ^ the arguments of the constructor (lhs)
+      -> [I.Arg Term]  -- ^ the arguments of the constructor (rhs)
       -> Unify ()
     unifyConstructorArgs a12 [] [] = return ()
     unifyConstructorArgs a12 vs1 vs2 = do
@@ -438,15 +529,15 @@ unifyIndices flex a us vs = liftTCM $ do
       unifyConArgs tel12 vs1 vs2
 
     unifyConArgs ::
-         TelHH  -- ^ The telescope(s) of the constructor args    [length = n].
-      -> [Arg Term]  -- ^ the arguments of the constructor (lhs) [length = n].
-      -> [Arg Term]  -- ^ the arguments of the constructor (rhs) [length = n].
+         TelHH  -- ^ The telescope(s) of the constructor args [length = n].
+      -> Args   -- ^ the arguments of the constructor (lhs)   [length = n].
+      -> Args   -- ^ the arguments of the constructor (rhs)   [length = n].
       -> Unify ()
     unifyConArgs _ (_ : _) [] = __IMPOSSIBLE__
     unifyConArgs _ [] (_ : _) = __IMPOSSIBLE__
     unifyConArgs _ []      [] = return ()
     unifyConArgs EmptyTel _ _ = __IMPOSSIBLE__
-    unifyConArgs tel0@(ExtendTel a@(Dom _ rel bHH) tel) us0@(arg@(Arg _ _ u) : us) vs0@(Arg _ _ v : vs) = do
+    unifyConArgs tel0@(ExtendTel a@(Dom _ bHH) tel) us0@(arg@(Arg _ u) : us) vs0@(Arg _ v : vs) = do
       liftTCM $ reportSDoc "tc.lhs.unify" 15 $ sep
         [ text "unifyConArgs"
 	-- , nest 2 $ parens (prettyTCM tel0)
@@ -462,7 +553,7 @@ unifyIndices flex a us vs = liftTCM $ do
       -- in case of dependent function type, we cannot postpone
       -- unification of u and v, otherwise us or vs might be ill-typed
       -- skip irrelevant parts
-      uHH <- if (rel == Irrelevant) then return $ Hom u else
+      uHH <- if isIrrelevant a then return $ Hom u else
                ifClean (unifyHH bHH u v) (return $ Hom u) (return $ Het u v)
 
       liftTCM $ reportSDoc "tc.lhs.unify" 25 $
@@ -476,12 +567,50 @@ unifyIndices flex a us vs = liftTCM $ do
       unifyConArgs (tel `absAppHH` uHH) us vs
 
 
+    -- | Used for arguments of a 'Def'.
+    unifyElims :: Type -> Elims -> Elims -> Unify ()
+    unifyElims _ (_ : _) [] = __IMPOSSIBLE__
+    unifyElims _ [] (_ : _) = __IMPOSSIBLE__
+    unifyElims _ [] [] = return ()
+    unifyElims _ (Proj{} : _) (Apply{} : _) = __IMPOSSIBLE__ -- Andreas, 2013-10-19
+    unifyElims _ (Apply{} : _) (Proj{} : _) = __IMPOSSIBLE__ -- being optimistic
+    unifyElims t (Proj f : es1) (Proj f' : es2)
+      | f == f' = do
+          maybe __IMPOSSIBLE__ (\ t -> unifyElims t es1 es2) =<< do
+            liftTCM $ projectType t f
+      | otherwise = projectionMismatch f f'
+    unifyElims a us0@(Apply arg@(Arg _ u) : us) vs0@(Apply (Arg _ v) : vs) = do
+      liftTCM $ reportSDoc "tc.lhs.unify" 15 $ sep
+        [ text "unifyElims"
+	, nest 2 $ parens (prettyTCM a)
+	, nest 2 $ prettyList $ map prettyTCM us0
+	, nest 2 $ prettyList $ map prettyTCM vs0
+        ]
+      a <- ureduce a  -- Q: reduce sufficient?
+      case ignoreSharing $ unEl a of
+	Pi b _  -> do
+          -- Andreas, Ulf, 2011-09-08 (AIM XVI)
+          -- in case of dependent function type, we cannot postpone
+          -- unification of u and v, otherwise us or vs might be ill-typed
+          let dep = dependent $ unEl a
+          -- skip irrelevant parts
+	  unless (isIrrelevant b) $
+            (if dep then noPostponing else id) $
+              unify (unDom b) u v
+          arg <- traverse ureduce arg
+	  unifyElims (a `piApply` [arg]) us vs
+	_	  -> __IMPOSSIBLE__
+      where dependent (Pi _ NoAbs{}) = False
+            dependent (Pi b c)       = 0 `relevantIn` absBody c
+            dependent (Shared p)     = dependent (derefPtr p)
+            dependent _              = False
+{-
     -- | Used for arguments of a 'Def', not 'Con'.
-    unifyArgs :: Type -> [Arg Term] -> [Arg Term] -> Unify ()
+    unifyArgs :: Type -> [I.Arg Term] -> [I.Arg Term] -> Unify ()
     unifyArgs _ (_ : _) [] = __IMPOSSIBLE__
     unifyArgs _ [] (_ : _) = __IMPOSSIBLE__
     unifyArgs _ [] [] = return ()
-    unifyArgs a us0@(arg@(Arg _ _ u) : us) vs0@(Arg _ _ v : vs) = do
+    unifyArgs a us0@(arg@(Arg _ u) : us) vs0@(Arg _ v : vs) = do
       liftTCM $ reportSDoc "tc.lhs.unify" 15 $ sep
         [ text "unifyArgs"
 	, nest 2 $ parens (prettyTCM a)
@@ -496,7 +625,7 @@ unifyIndices flex a us vs = liftTCM $ do
           -- unification of u and v, otherwise us or vs might be ill-typed
           let dep = dependent $ unEl a
           -- skip irrelevant parts
-	  unless (domRelevance b == Irrelevant) $
+	  unless (isIrrelevant b) $
             (if dep then noPostponing else id) $
               unify (unDom b) u v
           arg <- traverse ureduce arg
@@ -506,7 +635,7 @@ unifyIndices flex a us vs = liftTCM $ do
             dependent (Pi b c)       = 0 `relevantIn` absBody c
             dependent (Shared p)     = dependent (derefPtr p)
             dependent _              = False
-
+-}
     -- | Check using conversion check.
     recheckEqualities :: Unify ()
     recheckEqualities = do
@@ -524,6 +653,11 @@ unifyIndices flex a us vs = liftTCM $ do
       i |-> x
       recheckConstraints
 
+    maybeAssign h i x = (i |->> x) `catchException` \e ->
+      case e of
+        UnclearOccurrence{} -> h
+        _                   -> throwException e
+
     unifySizes :: Term -> Term -> Unify ()
     unifySizes u v = do
       sz <- liftTCM sizeType
@@ -533,7 +667,7 @@ unifyIndices flex a us vs = liftTCM $ do
         (SizeSuc u, SizeSuc v) -> unify sz u v
         (SizeSuc u, SizeInf) -> unify sz u v
         (SizeInf, SizeSuc v) -> unify sz u v
-        _                    -> unifyAtom sz u v
+        _                    -> unifyAtomHH (Hom sz) u v checkEqualityHH
 
     -- | Possibly heterogeneous unification (but at same-shaped types).
     --   In het. situations, we only search for a mismatch!
@@ -543,11 +677,17 @@ unifyIndices flex a us vs = liftTCM $ do
         TypeHH  -- ^ one or two types, need not be in (u)reduced form
      -> Term -> Term -> Unify ()
     unifyHH aHH u v = do
+      liftTCM $ reportSDoc "tc.lhs.unify" 15 $
+	sep [ text "unifyHH"
+	    , nest 2 $ (parens $ prettyTCM u) <+> text "=?="
+	    , nest 2 $ parens $ prettyTCM v
+	    , nest 2 $ text ":" <+> prettyTCM aHH
+	    ]
       u <- liftTCM . constructorForm =<< ureduce u
       v <- liftTCM . constructorForm =<< ureduce v
       aHH <- ureduce aHH
-      liftTCM $ reportSDoc "tc.lhs.unify" 15 $
-	sep [ text "unifyHH"
+      liftTCM $ reportSDoc "tc.lhs.unify" 25 $
+	sep [ text "unifyHH (reduced)"
 	    , nest 2 $ (parens $ prettyTCM u) <+> text "=?="
 	    , nest 2 $ parens $ prettyTCM v
 	    , nest 2 $ text ":" <+> prettyTCM aHH
@@ -555,26 +695,45 @@ unifyIndices flex a us vs = liftTCM $ do
       -- obtain the (== Size) function
       isSizeName <- liftTCM isSizeNameTest
 
+      -- Andreas, 2013-10-24 (fixing issue 924)
+      -- Only if we cannot make progress, we try full normalization!
+      let tryAgain aHH u v = do
+            u <- liftTCM $ etaContract =<< normalise u
+            v <- liftTCM $ etaContract =<< normalise v
+            unifyAtomHH aHH u v $ \ aHH u v -> do
+              -- Andreas, 2014-03-03 (issue 1061)
+              -- As a last resort, normalize types to maybe get back
+              -- to the homogeneous case
+              caseMaybeM (liftTCM $ makeHom aHH) (checkEqualityHH aHH u v) $ \ a -> do
+                unifyAtomHH (Hom a) u v checkEqualityHH
+
       -- check whether types have the same shape
       (aHH, sh) <- shapeViewHH aHH
       case sh of
         ElseSh  -> checkEqualityHH aHH u v -- not a type or not same types
 
-        DefSh d -> if isSizeName d then unifySizes u v
-                                   else unifyAtomHH aHH u v
-        _ -> unifyAtomHH aHH u v
+        DefSh d | isSizeName d -> unifySizes u v
+
+        _ -> unifyAtomHH aHH u v tryAgain
 
     unifyAtomHH ::
          TypeHH -- ^ in ureduced form
-      -> Term -> Term -> Unify ()
-    unifyAtomHH aHH0 u v = do
+      -> Term
+      -> Term
+      -> (TypeHH -> Term -> Term -> Unify ())
+         -- ^ continuation in case unification was inconclusive
+      -> Unify ()
+    unifyAtomHH aHH0 u v tryAgain = do
       let (aHH, homogeneous, a) = case aHH0 of
             Hom a                -> (aHH0, True, a)
             Het a1 a2 | a1 == a2 -> (Hom a1, True, a1) -- BRITTLE: just checking syn.eq.
             _                    -> (aHH0, False, __IMPOSSIBLE__)
            -- use @a@ only if 'homogeneous' holds!
 
-          fallback = checkEqualityHH aHH u v
+          fallback = tryAgain aHH u v
+                   -- Try again if occurs check fails non-rigidly. It might be
+                   -- that normalising gets rid of the occurrence.
+          (|->?) = maybeAssign fallback
 
       liftTCM $ reportSDoc "tc.lhs.unify" 15 $
 	sep [ text "unifyAtom"
@@ -583,24 +742,53 @@ unifyIndices flex a us vs = liftTCM $ do
 	    , nest 2 $ prettyTCM v <> if flexibleTerm v then text " (flexible)" else empty
 	    , nest 2 $ text ":" <+> prettyTCM aHH
 	    ]
+      liftTCM $ reportSDoc "tc.lhs.unify" 60 $
+        text $ "aHH = " ++ show aHH
       case (ignoreSharing u, ignoreSharing v) of
         -- Ulf, 2011-06-19
         -- We don't want to worry about levels here.
         (Level l, _) -> do
             u <- liftTCM $ reallyUnLevelView l
-            unifyAtomHH aHH u v
+            unifyAtomHH aHH u v tryAgain
         (_, Level l) -> do
             v <- liftTCM $ reallyUnLevelView l
-            unifyAtomHH aHH u v
+            unifyAtomHH aHH u v tryAgain
 	(Var i us, Var j vs) | i == j  -> checkEqualityHH aHH u v
-	(Var i [], _) | homogeneous && flexible i -> i |->> (v, a)
-	(_, Var j []) | homogeneous && flexible j -> j |->> (u, a)
+-- Andreas, 2013-03-05: the following flex/flex case is an attempt at
+-- better dotting (see Issue811).  Does not work perfectly, maybe the best choice
+-- which variable to assign cannot made locally, but would need a look at the full
+-- picture!?  Or maybe the information on flexible variables in not yet good enough
+-- in the call to split in Coverage.
+        (Var i [], Var j []) | homogeneous, Just fi <- findFlexible i, Just fj <- findFlexible j -> do
+            liftTCM $ reportSDoc "tc.lhs.unify.flexflex" 20 $
+              sep [ text "unifying flexible/flexible"
+                  , nest 2 $ text "i =" <+> prettyTCM u <+> text ("; fi = " ++ show fi)
+                  , nest 2 $ text "j =" <+> prettyTCM v <+> text ("; fj = " ++ show fj)
+                  ]
+            -- We assign the "bigger" variable, where dotted, hidden, earlier is bigger
+            -- (in this order, see Problem.hs).
+            -- The comparison is total.
+            if fj >= fi then j |->? (u, a) else i |->? (v, a)
+	(Var i [], _) | homogeneous && flexible i -> i |->? (v, a)
+	(_, Var j []) | homogeneous && flexible j -> j |->? (u, a)
 	(Con c us, Con c' vs)
           | c == c' -> do
-              r <- ureduce =<< liftTCM (dataOrRecordTypeHH c aHH)
+              r <- liftTCM (dataOrRecordTypeHH' c aHH)
               case r of
-                Just a'HH -> unifyConstructorArgs a'HH us vs
-                Nothing   -> checkEqualityHH aHH u v
+                Just (d, bHH) -> do
+                  bHH <- ureduce bHH
+                  -- Jesper, 2014-05-03: When --without-K is enabled, we reconstruct
+                  -- datatype indices and unify them as well
+                  withoutKEnabled <- liftTCM $ optWithoutK <$> pragmaOptions
+                  when withoutKEnabled (do
+                      def   <- liftTCM $ getConstInfo d
+                      let parsHH  = fmap (\(b, pars, ixs) -> pars) bHH
+                          ixsHH   = fmap (\(b, pars, ixs) -> ixs) bHH
+                          dtypeHH = (defType def) `applyHH` parsHH
+                      unifyConstructorArgs dtypeHH (leftHH ixsHH) (rightHH ixsHH))
+                  let a'HH = fmap (\(b, pars, _) -> b `apply` pars) bHH
+                  unifyConstructorArgs a'HH us vs
+                Nothing -> checkEqualityHH aHH u v
           | otherwise -> constructorMismatchHH aHH u v
         -- Definitions are ok as long as they can't reduce (i.e. datatypes/axioms)
 	(Def d us, Def d' vs)
@@ -614,7 +802,8 @@ unifyIndices flex a us vs = liftTCM $ do
                     _          -> False
               inj <- liftTCM $ optInjectiveTypeConstructors <$> pragmaOptions
               if inj && ok
-                then unifyArgs (defType def) us vs
+                then unifyElims (defType def) us vs `catchException` \ _ ->
+                       constructorMismatchHH aHH u v
                 else checkEqualityHH aHH u v
           -- Andreas, 2011-05-30: if heads disagree, abort
           -- but do not raise "mismatch" because otherwise type constructors
@@ -627,7 +816,7 @@ unifyIndices flex a us vs = liftTCM $ do
         -- We can instantiate metas if the other term is inert (constructor application)
         -- Andreas, 2011-09-13: test/succeed/IndexInference needs this feature.
         (MetaV m us, _) | homogeneous -> do
-            ok <- liftTCM $ instMeta a m us v
+            ok <- liftTCM $ instMetaE a m us v
             liftTCM $ reportSDoc "tc.lhs.unify" 40 $
               vcat [ fsep [ text "inst meta", text $ if ok then "(ok)" else "(not ok)" ]
                    , nest 2 $ sep [ prettyTCM u, text ":=", prettyTCM =<< normalise u ]
@@ -635,7 +824,7 @@ unifyIndices flex a us vs = liftTCM $ do
             if ok then unify a u v
                   else addEquality a u v
         (_, MetaV m vs) | homogeneous -> do
-            ok <- liftTCM $ instMeta a m vs u
+            ok <- liftTCM $ instMetaE a m vs u
             liftTCM $ reportSDoc "tc.lhs.unify" 40 $
               vcat [ fsep [ text "inst meta", text $ if ok then "(ok)" else "(not ok)" ]
                    , nest 2 $ sep [ prettyTCM v, text ":=", prettyTCM =<< normalise v ]
@@ -671,12 +860,16 @@ unifyIndices flex a us vs = liftTCM $ do
     unify :: Type -> Term -> Term -> Unify ()
     unify a = unifyHH (Hom a)
 
-    unifyAtom :: Type -> Term -> Term -> Unify ()
-    unifyAtom a = unifyAtomHH (Hom a)
-
     -- The contexts are transient when unifying, so we should just instantiate to
     -- constructor heads and generate fresh metas for the arguments. Beware of
     -- constructors that aren't fully applied.
+    instMetaE :: Type -> MetaId -> Elims -> Term -> TCM Bool
+    instMetaE a m es v = do
+      case allApplyElims es of
+        Just us -> instMeta a m us v
+        Nothing -> return False
+
+    instMeta :: Type -> MetaId -> Args -> Term -> TCM Bool
     instMeta a m us v = do
       app <- inertApplication a v
       reportSDoc "tc.lhs.unify" 50 $
@@ -700,7 +893,7 @@ unifyIndices flex a us vs = liftTCM $ do
                 tel <- getContextTelescope
                 -- important: create the meta in the same environment as the original meta
                 newArgsMetaCtx b' tel us
-            noConstraints $ assignV m us (v' `apply` margs)
+            noConstraints $ assignV DirEq m us (v' `apply` margs)
             return True
           `catchError` \_ -> return False
 
@@ -708,7 +901,7 @@ unifyIndices flex a us vs = liftTCM $ do
     inertApplication a v =
       case ignoreSharing v of
         Con c vs -> fmap (\ b -> (Con c [], b, vs)) <$> dataOrRecordType c a
-        Def d vs -> do
+        Def d es | Just vs <- allApplyElims es -> do
           def <- getConstInfo d
           let ans = Just (Def d [], defType def, vs)
           return $ case theDef def of
@@ -725,35 +918,41 @@ unifyIndices flex a us vs = liftTCM $ do
 -- Precondition: The type has to correspond to an application of the
 -- given constructor.
 dataOrRecordType
-  :: QName -- ^ Constructor name.
+  :: ConHead -- ^ Constructor name.
   -> Type  -- ^ Type of constructor application (must end in data/record).
   -> TCM (Maybe Type) -- ^ Type of constructor, applied to pars.
-dataOrRecordType c a = fmap (\ (d, b, args) -> b `apply` args) <$> dataOrRecordType' c a
+dataOrRecordType c a = fmap (\ (d, b, pars, _) -> b `apply` pars) <$> dataOrRecordType' c a
 
 dataOrRecordType' ::
-     QName -- ^ Constructor name.
+     ConHead -- ^ Constructor name.
   -> Type  -- ^ Type of constructor application (must end in data/record).
-  -> TCM (Maybe (QName, Type, Args))
+  -> TCM (Maybe (QName, Type, Args, Args))
            -- ^ Name of data/record type,
-           --   type of constructor to be applied, and
-           --   data/record parameters
+           --   type of constructor to be applied,
+           --   data/record parameters, and
+           --   data indices
 dataOrRecordType' c a = do
   -- The telescope ends with a datatype or a record.
   (d, args) <- do
     TelV _ (El _ def) <- telView a
-    let Def d args = ignoreSharing def
+    let Def d es = ignoreSharing def
+        args = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
     return (d, args)
   def <- theDef <$> getConstInfo d
-  r <- case def of
-    Datatype{dataPars = n} -> Just . ((,) n) . defType <$> getConstInfo c
-    Record  {recPars  = n} -> Just . ((,) n) <$> getRecordConstructorType d
-    _		           -> return Nothing
-  return $ fmap (\ (n, a') -> (d, a', genericTake n args)) r
+  case def of
+    Datatype{dataPars = n} -> do
+      a' <- defType <$> getConInfo c
+      let (pars, ixs) = genericSplitAt n args
+      return $ Just (d, a', pars, ixs)
+    Record{} -> do
+      a' <- getRecordConstructorType d
+      return $ Just (d, a', args, [])
+    _ -> return Nothing
 
 -- | Heterogeneous situation.
 --   @a1@ and @a2@ need to end in same datatype/record.
 dataOrRecordTypeHH ::
-     QName      -- ^ Constructor name.
+     ConHead      -- ^ Constructor name.
   -> TypeHH     -- ^ Type(s) of constructor application (must end in same data/record).
   -> TCM (Maybe TypeHH) -- ^ Type of constructor, instantiated possibly heterogeneously to parameters.
 dataOrRecordTypeHH c (Hom a) = fmap Hom <$> dataOrRecordType c a
@@ -761,7 +960,7 @@ dataOrRecordTypeHH c (Het a1 a2) = do
   r1 <- dataOrRecordType' c a1
   r2 <- dataOrRecordType' c a2  -- b2 may have different parameters than b1!
   return $ case (r1, r2) of
-    (Just (d1, b1, pars1), Just (d2, b2, pars2)) | d1 == d2 -> Just $
+    (Just (d1, b1, pars1, _), Just (d2, b2, pars2, _)) | d1 == d2 -> Just $
         -- Andreas, 2011-09-15 if no parameters, we can stay homogeneous
         if null pars1 && null pars2 then Hom b1
          -- if parameters, go heterogeneous
@@ -769,6 +968,27 @@ dataOrRecordTypeHH c (Het a1 a2) = do
          else Het (b1 `apply` pars1) (b2 `apply` pars2)
     _ -> Nothing
 
+dataOrRecordTypeHH' ::
+     ConHead
+  -> TypeHH
+  -> TCM (Maybe (QName, HomHet (Type, Args, Args)))
+dataOrRecordTypeHH' c (Hom a) = do
+  r <- dataOrRecordType' c a
+  case r of
+    Just (d, a', pars, ixs) -> return $ Just (d, Hom (a', pars, ixs))
+    Nothing                 -> return $ Nothing
+dataOrRecordTypeHH' c (Het a1 a2) = do
+  r1 <- dataOrRecordType' c a1
+  r2 <- dataOrRecordType' c a2
+  return $ case (r1, r2) of
+    -- TODO: We should always have b1 == b2, check/force this in some way?
+    (Just (d1, b1, pars1, ixs1), Just (d2, b2, pars2, ixs2)) | d1 == d2 -> Just $
+        if null pars1 && null pars2 && null ixs1 && null ixs2
+          then (d1, Hom (b1, [], []))
+          else (d1, Het (b1, pars1, ixs1) (b2, pars2, ixs2))
+    _ -> Nothing
+
+
 -- | Return record type identifier if argument is a record type.
 isEtaRecordTypeHH :: MonadTCM tcm => TypeHH -> tcm (Maybe (QName, HomHet Args))
 isEtaRecordTypeHH (Hom a) = fmap (\ (d, ps) -> (d, Hom ps)) <$> liftTCM (isEtaRecordType a)
@@ -782,8 +1002,8 @@ isEtaRecordTypeHH (Het a1 a2) = do
 
 -- | Views an expression (pair) as type shape.  Fails if not same shape.
 data ShapeView a
-  = PiSh (Dom a) (Abs a)
-  | FunSh (Dom a) a
+  = PiSh (I.Dom a) (I.Abs a)
+  | FunSh (I.Dom a) a
   | DefSh QName   -- ^ data/record
   | VarSh Nat     -- ^ neutral type
   | LitSh Literal -- ^ built-in type
@@ -815,13 +1035,17 @@ shapeViewHH (Het a1 a2) = do
   (a2, sh2) <- shapeView a2
   return . (Het a1 a2,) $ case (sh1, sh2) of
 
-    (PiSh (Dom h1 r1 a1) b1, PiSh (Dom h2 r2 a2) b2)
-      | h1 == h2 ->
-      PiSh (Dom h1 (min r1 r2) (Het a1 a2)) (Abs (absName b1) (Het (absBody b1) (absBody b2)))
+    (PiSh d1@(Dom i1 a1) b1, PiSh (Dom i2 a2) b2)
+      | argInfoHiding i1 == argInfoHiding i2 ->
+      PiSh (Dom (setRelevance (min (getRelevance i1) (getRelevance i2)) i1)
+                (Het a1 a2))
+           (Abs (absName b1) (Het (absBody b1) (absBody b2)))
 
-    (FunSh (Dom h1 r1 a1) b1, FunSh (Dom h2 r2 a2) b2)
-      | h1 == h2 ->
-      FunSh (Dom h1 (min r1 r2) (Het a1 a2)) (Het b1 b2)
+    (FunSh d1@(Dom i1 a1) b1, FunSh (Dom i2 a2) b2)
+      | argInfoHiding i1 == argInfoHiding i2 ->
+      FunSh (Dom (setRelevance (min (getRelevance i1) (getRelevance i2)) i1)
+                 (Het a1 a2))
+            (Het b1 b2)
 
     (DefSh d1, DefSh d2) | d1 == d2 -> DefSh d1
     (VarSh x1, VarSh x2) | x1 == x2 -> VarSh x1
@@ -838,7 +1062,7 @@ telViewUpToHH n t = do
   (t, sh) <- shapeViewHH =<< liftTCM (traverse reduce t)
   case sh of
     PiSh a b  -> absV a (absName b) <$> telViewUpToHH (n-1) (absBody b)
-    FunSh a b -> absV a "_" <$> telViewUpToHH (n-1) (raise 1 b)
+    FunSh a b -> absV a underscore <$> telViewUpToHH (n-1) (raise 1 b)
     _         -> return $ TelV EmptyTel t
   where
     absV a x (TelV tel t) = TelV (ExtendTel a (Abs x tel)) t
