@@ -3,17 +3,14 @@
 module Agda.TypeChecking.Rules.Record where
 
 import Control.Applicative
-import Control.Monad.Trans
-import Control.Monad.Reader
 
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Common
-import Agda.Syntax.Internal
+import Agda.Syntax.Internal as I
 import Agda.Syntax.Position
 import qualified Agda.Syntax.Info as Info
 
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Monad.Builtin ( primIrrAxiom )
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Reduce
@@ -23,7 +20,7 @@ import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.CompiledClause.Compile
 
 import Agda.TypeChecking.Rules.Data ( bindParameters, fitsIn )
-import Agda.TypeChecking.Rules.Term ( isType_ )
+import Agda.TypeChecking.Rules.Term ( isType_, ConvColor(..) )
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Decl (checkDecl)
 
 import Agda.Utils.Size
@@ -49,16 +46,25 @@ import Agda.Utils.Impossible
 --     [@ps@]      Record parameters.
 --
 --     [@contel@]  Approximate type of constructor (@fields@ -> Set).
+--                 Does not include record parameters.
 --
 --     [@fields@]  List of field signatures.
 --
-checkRecDef :: Info.DefInfo -> QName -> Maybe Induction -> Maybe A.QName ->
-               [A.LamBinding] -> A.Expr -> [A.Field] -> TCM ()
+checkRecDef
+  :: Info.DefInfo              -- ^ Position and other info.
+  -> QName                     -- ^ Record type identifier.
+  -> Maybe (Ranged Induction)  -- ^ Optional: (co)inductive declaration.
+  -> Maybe A.QName             -- ^ Optional: constructor name.
+  -> [A.LamBinding]            -- ^ Record parameters.
+  -> A.Expr                    -- ^ Approximate type of constructor (@fields@ -> Set).
+                               --   Does not include record parameters.
+  -> [A.Field]                 -- ^ Field signatures.
+  -> TCM ()
 checkRecDef i name ind con ps contel fields =
   traceCall (CheckRecDef (getRange i) (qnameName name) ps fields) $ do
     reportSDoc "tc.rec" 10 $ vcat
       [ text "checking record def" <+> prettyTCM name
-      , nest 2 $ text "ps ="     <+> prettyList (map prettyA ps)
+      , nest 2 $ text "ps ="     <+> prettyList (map prettyAs ps)
       , nest 2 $ text "contel =" <+> prettyA contel
       , nest 2 $ text "fields =" <+> prettyA (map Constr fields)
       ]
@@ -74,13 +80,13 @@ checkRecDef i name ind con ps contel fields =
       -- WRONG: contype <- workOnTypes $ killRange <$> (instantiateFull =<< isType_ contel)
       contype <- killRange <$> (instantiateFull =<< isType_ contel)
 
-      -- compute the field telescope
+      -- compute the field telescope (does not include record parameters)
       let TelV ftel _ = telView' contype
 
           -- A record is irrelevant if all of its fields are.
           -- In this case, the associated module parameter will be irrelevant.
           -- See issue 392.
-          recordRelevance = minimum $ Irrelevant : (map domRelevance $ telToList ftel)
+          recordRelevance = minimum $ Irrelevant : (map getRelevance $ telToList ftel)
 
       -- Compute correct type of constructor
 
@@ -89,50 +95,56 @@ checkRecDef i name ind con ps contel fields =
       s <- case ignoreSharing $ unEl t0' of
 	Sort s	-> return s
 	_	-> typeError $ ShouldBeASort t0
-      gamma <- getContextTelescope
+      gamma <- getContextTelescope  -- the record params (incl. module params)
       -- record type (name applied to parameters)
-      let rect = El s $ Def name $ teleArgs gamma
+      let rect = El s $ Def name $ map Apply $ teleArgs gamma
 
       -- Put in @rect@ as correct target of constructor type.
       -- Andreas, 2011-05-10 use telePi_ instead of telePi to preserve
       -- even names of non-dependent fields in constructor type (Issue 322).
       let contype = telePi_ ftel (raise (size ftel) rect)
+        -- NB: contype does not contain the parameter telescope
 
       -- Obtain name of constructor (if present).
       (hasNamedCon, conName, conInfo) <- case con of
         Just c  -> return (True, c, i)
         Nothing -> do
           m <- killRange <$> currentModule
-          c <- qualify m <$> freshName_ "recCon-NOT-PRINTED"
+          c <- qualify m <$> freshName_ ("recCon-NOT-PRINTED" :: String)
           return (False, c, i)
 
       -- Add record type to signature.
       reportSDoc "tc.rec" 15 $ text "adding record type to signature"
 
-      let getName :: A.Declaration -> [Arg QName]
-          getName (A.Field _ x arg)    = [fmap (const x) arg]
+      let getName :: A.Declaration -> [A.Arg QName]
+          getName (A.Field _ x arg)    = [x <$ arg]
 	  getName (A.ScopedDecl _ [f]) = getName f
 	  getName _		       = []
 
-          indCo = maybe Inductive id ind -- default is 'Inductive' for backwards compatibility but should maybe be 'Coinductive'
+          fs = concatMap (convColor . getName) fields
+          con = ConHead conName $ map unArg fs
 
-      addConstant name $ Defn Relevant name t0 [] [] (defaultDisplayForm name) 0 noCompiledRep
+          -- Default is 'Inductive'.
+          -- This makes sense because all non-recursive records are "inductive",
+          -- meaning not coinductive.
+          -- Of course, one could make all *recursive* records coinductive
+          -- by default, but this would not be backwards-compatible.
+          indCo = maybe Inductive rangedThing ind
+
+      reportSDoc "tc.rec" 30 $ text "record constructor is " <+> text (show con)
+      addConstant name $ defaultDefn defaultArgInfo name t0
 		       $ Record { recPars           = 0
                                 , recClause         = Nothing
-                                , recCon            = conName
+                                , recConHead        = con
                                 , recNamedCon       = hasNamedCon
-                                , recConType        = contype
-				, recFields         = concatMap getName fields
-                                , recTel            = ftel
+                                , recConType        = contype  -- addConstant adds params!
+				, recFields         = fs
+                                , recTel            = ftel     -- addConstant adds params!
 				, recAbstr          = Info.defAbstract i
-                                , recEtaEquality    = True
+                                , recEtaEquality    = indCo == Inductive
                                 , recInduction      = indCo
                                 -- determined by positivity checker:
                                 , recRecursive      = False
-{-
-                                , recPolarity       = []
-                                , recArgOccurrences = []
--}
                                 , recMutual         = []
                                 }
 
@@ -140,17 +152,17 @@ checkRecDef i name ind con ps contel fields =
       -- Andreas, 2011-05-19 moved this here, it was below the record module
       --   creation
       addConstant conName $
-        Defn Relevant conName contype [] [] (defaultDisplayForm conName) 0 noCompiledRep $
+        defaultDefn defaultArgInfo conName contype $
              Constructor { conPars   = 0
-                         , conSrcCon = conName
+                         , conSrcCon = con
                          , conData   = name
                          , conAbstr  = Info.defAbstract conInfo
                          , conInd    = indCo
                          }
 
       -- Check that the fields fit inside the sort
-      let dummy = var 0  -- We're only interested in the sort here
-      telePi ftel (El s dummy) `fitsIn` s
+      let dummy = Sort Prop  -- We're only interested in the sort here
+      telePi ftel (El (raise (size ftel) s) dummy) `fitsIn` s
 
       {- Andreas, 2011-04-27 WRONG because field types are checked again
          and then non-stricts should not yet be irrelevant
@@ -159,23 +171,55 @@ checkRecDef i name ind con ps contel fields =
       -- ctx <- (reverse . map hideAndRelParams . take (size tel)) <$> getContext
       -}
 
-      -- make record parameters hidden
-      ctx <- (reverse . map (mapDomHiding $ const Hidden) . take (size tel)) <$> getContext
+{- Andreas, 2013-09-13 DEBUGGING the debug printout
+      reportSDoc "tc.rec" 80 $ sep
+        [ text "current module record telescope"
+        , nest 2 $ (prettyTCM =<< getContextTelescope)
+        ]
+      reportSDoc "tc.rec" 80 $ sep
+        [ text "current module record telescope"
+        , nest 2 $ (text . show =<< getContextTelescope)
+        ]
+      reportSDoc "tc.rec" 80 $ sep
+        [ text "current module record telescope"
+        , nest 2 $ (inTopContext . prettyTCM =<< getContextTelescope)
+        ]
+      reportSDoc "tc.rec" 80 $ sep
+        [ text "current module record telescope"
+        , nest 2 $ do
+           tel <- getContextTelescope
+           text (show tel) $+$ do
+           inContext [] $ do
+             prettyTCM tel $+$ do
+               telA <- reify tel
+               text (show telA) $+$ do
+               ctx <- getContextTelescope
+               text "should be empty:" <+> prettyTCM ctx
+        ]
+-}
 
       let -- name of record module
           m    = qnameToMName name
           -- make record parameters hidden and non-stricts irrelevant
 	  htel = map hideAndRelParams $ telToList tel
-	  tel' = telFromList $ htel ++ [Dom NotHidden recordRelevance ("r", rect)]
-          ext (Dom h r (x, t)) = addCtx x (Dom h r t)
+          info = setRelevance recordRelevance defaultArgInfo
+	  tel' = telFromList $ htel ++ [Dom info ("r", rect)]
+          ext (Dom info (x, t)) = addCtx x (Dom info t)
 
+      -- Add the record section
+      -- make record parameters hidden
+      ctx <- (reverse . map (setHiding Hidden) . take (size tel)) <$> getContext
+      reportSDoc "tc.rec" 80 $ sep
+        [ text "visibility-modified record telescope"
+        , nest 2 $ text "ctx =" <+> prettyTCM ctx
+        ]
       escapeContext (size tel) $ flip (foldr ext) ctx $
        -- the record variable has the empty name by intention, see issue 208
-       underAbstraction (Dom NotHidden recordRelevance rect) (Abs "" ()) $ \_ -> do
+       underAbstraction (Dom info rect) (Abs "" ()) $ \_ -> do
 	reportSDoc "tc.rec.def" 10 $ sep
 	  [ text "record section:"
 	  , nest 2 $ sep
-            [ prettyTCM m <+> (inContext [] . prettyTCM =<< getContextTelescope)
+            [ prettyTCM m <+> (inTopContext . prettyTCM =<< getContextTelescope)
             , fsep $ punctuate comma $ map (text . show . getName) fields
             ]
 	  ]
@@ -184,9 +228,14 @@ checkRecDef i name ind con ps contel fields =
           ]
 	addSection m (size tel')
 
-        -- Check the types of the fields
-        withCurrentModule m $
-          checkRecordProjections m name conName tel' (raise 1 ftel) fields
+      -- Check the types of the fields
+      -- Andreas, 2013-09-13 all module telescopes count as parameters to the record projections
+      -- thus, we set all context entries to @Hidden@
+      modifyContext (modifyContextEntries (setHiding Hidden)) $ do
+       underAbstraction (Dom info rect) (Abs "" ()) $ \_ -> do
+        withCurrentModule m $ do
+          tel' <- getContextTelescope
+          checkRecordProjections m name con tel' (raise 1 ftel) fields
 
       -- Andreas, 2011-05-19 here was the code "Add record constr..."
 
@@ -201,7 +250,7 @@ checkRecDef i name ind con ps contel fields =
 
     [@r@    ]  name of the record type
 
-    [@q@    ]  name of the record constructor
+    [@con@  ]  name of the record constructor
 
     [@tel@  ]  parameters and record variable r ("self")
 
@@ -210,9 +259,10 @@ checkRecDef i name ind con ps contel fields =
     [@fs@   ]  the fields to be checked
 -}
 checkRecordProjections ::
-  ModuleName -> QName -> QName -> Telescope -> Telescope ->
+  ModuleName -> QName -> ConHead -> Telescope -> Telescope ->
   [A.Declaration] -> TCM ()
-checkRecordProjections m r q tel ftel fs = checkProjs EmptyTel ftel fs
+checkRecordProjections m r con tel ftel fs = do
+    checkProjs EmptyTel ftel fs
   where
 
     checkProjs :: Telescope -> Telescope -> [A.Declaration] -> TCM ()
@@ -222,7 +272,7 @@ checkRecordProjections m r q tel ftel fs = checkProjs EmptyTel ftel fs
     checkProjs ftel1 ftel2 (A.ScopedDecl scope fs' : fs) =
       setScope scope >> checkProjs ftel1 ftel2 (fs' ++ fs)
 
-    checkProjs ftel1 (ExtendTel (Dom h rel t) ftel2) (A.Field info x _ : fs) = do
+    checkProjs ftel1 (ExtendTel (Dom ai t) ftel2) (A.Field info x _ : fs) = do
       -- Andreas, 2012-06-07:
       -- Issue 387: It is wrong to just type check field types again
       -- because then meta variables are created again.
@@ -230,8 +280,8 @@ checkRecordProjections m r q tel ftel fs = checkProjs EmptyTel ftel fs
       reportSDoc "tc.rec.proj" 5 $ sep
 	[ text "checking projection" <+> text (show x)
 	, nest 2 $ vcat
-	  [ text "top   =" <+> (inContext [] . prettyTCM =<< getContextTelescope)
-          , text "tel   =" <+> prettyTCM tel
+	  [ text "top   =" <+> (inTopContext . prettyTCM =<< getContextTelescope)
+          , text "tel   =" <+> (inTopContext . prettyTCM $ tel)
 	  , text "ftel1 =" <+> prettyTCM ftel1
 	  , text "t     =" <+> prettyTCM t
 	  , text "ftel2 =" <+> addCtxTel ftel1 (underAbstraction_ ftel2 prettyTCM)
@@ -257,15 +307,18 @@ checkRecordProjections m r q tel ftel fs = checkProjs EmptyTel ftel fs
       -- The type of the projection function should be
       --  {tel} -> (r : R Δ) -> t
       -- where Δ = Γ, tel is the current context
-      let finalt   = telePi tel t
+      let finalt   = telePi (replaceEmptyName "r" tel) t
 	  projname = qualify m $ qnameName x
-          projcall = Def projname [defaultArg $ var 0]
+          projcall = Var 0 [Proj projname]
+--          projcall = Def projname [defaultArg $ var 0]
+          rel      = getRelevance ai
           -- the recursive call
-          recurse  = checkProjs (abstract ftel1 $ ExtendTel (Dom h rel t)
-                                 $ Abs (show $ qnameName projname) EmptyTel)
+          recurse  = checkProjs (abstract ftel1 $ ExtendTel (Dom ai t)
+                                 $ Abs (nameToArgName $ qnameName projname) EmptyTel)
                                 (ftel2 `absApp` projcall) fs
 
-      reportSDoc "tc.rec.proj" 25 $ nest 2 $ text "finalt=" <+> prettyTCM finalt
+      reportSDoc "tc.rec.proj" 25 $ nest 2 $ text "finalt=" <+> do
+        inTopContext $ prettyTCM finalt
 
       -- Andreas, 2012-02-20 do not add irrelevant projections if
       -- disabled by --no-irrelevant-projections
@@ -273,7 +326,7 @@ checkRecordProjections m r q tel ftel fs = checkProjs EmptyTel ftel fs
 
       reportSDoc "tc.rec.proj" 10 $ sep
 	[ text "adding projection"
-	, nest 2 $ prettyTCM projname <+> text ":" <+> inContext [] (prettyTCM finalt)
+	, nest 2 $ prettyTCM projname <+> text ":" <+> inTopContext (prettyTCM finalt)
 	]
 
       -- The body should be
@@ -286,10 +339,10 @@ checkRecordProjections m r q tel ftel fs = checkProjs EmptyTel ftel fs
       -- PROBLEM: because of dropped parameters, cannot refer to t
 
       -- compute body modification for irrelevant projections
-      bodyMod <- do
-        case rel of
-          Relevant   -> return $ \ n x -> x -- no modification
-          Irrelevant -> return $ \ n x -> DontCare x
+      let bodyMod = case rel of
+            Relevant   -> id
+            Irrelevant -> DontCare
+            _          -> __IMPOSSIBLE__
 
 {- 2012-04-02: DontCare instead of irrAxiom
           Irrelevant -> do
@@ -302,37 +355,59 @@ checkRecordProjections m r q tel ftel fs = checkProjs EmptyTel ftel fs
                                   mkArg t = Arg Hidden Relevant $ Sort Prop
                               in  apply irrAxiom [mkArg levelOfT, mkArg (unEl t), Arg NotHidden Irrelevant x]
 -}
-          _          -> __IMPOSSIBLE__
 
       let -- Andreas, 2010-09-09: comment for existing code
           -- split the telescope into parameters (ptel) and the type or the record
           -- (rt) which should be  R ptel
           (ptel,[rt]) = splitAt (size tel - 1) $ telToList tel
+          projArgI    = domInfo rt
 	  conp	 = defaultArg
-		 $ ConP q (Just (argFromDom $ fmap snd rt))
-                   [ Arg h r (VarP "x") | Dom h r _ <- telToList ftel ]
+		 $ ConP con (Just (False, argFromDom $ fmap snd rt))
+                   [ Arg info $ unnamed $ VarP "x" | Dom info _ <- telToList ftel ]
 	  nobind 0 = id
 	  nobind n = Bind . Abs "_" . nobind (n - 1)
 	  body	 = nobind (size ftel1)
 		 $ Bind . Abs "x"
 		 $ nobind (size ftel2)
-		 $ Body $ bodyMod (size ftel) $ Var (size ftel2) []
+		 $ Body $ bodyMod $ var (size ftel2)
           cltel  = ftel
-	  clause = Clause { clauseRange = getRange info
-                          , clauseTel   = killRange cltel
-                          , clausePerm  = idP $ size ftel
-                          , clausePats  = [conp]
-                          , clauseBody  = body
+	  clause                            = Clause { clauseRange = getRange info
+                          , clauseTel       = killRange cltel
+                          , clausePerm      = idP $ size ftel
+                          , namedClausePats = [Named Nothing <$> conp]
+                          , clauseBody      = body
+                          , clauseType      = Just $ Arg ai t
                           }
 
-      reportSDoc "tc.rec.proj" 20 $ sep
+      -- Andreas, 2013-10-20
+      -- creating the projection construction function
+      let core = Lam projArgI $ Abs "r" $ bodyMod $ projcall
+          -- leading lambdas are to ignore parameter applications
+          proj = teleNoAbs ptel core
+          -- proj = foldr (\ (Dom ai (x, _)) -> Lam ai . NoAbs x) core ptel
+          projection = Projection
+            { projProper   = Just projname
+            -- name of the record type:
+            , projFromType = r
+            -- index of the record argument (in the type),
+            -- start counting with 1:
+            , projIndex    = size ptel + 1  -- which is @size tel@
+            , projDropPars = proj
+            , projArgInfo  = projArgI
+            }
+
+      reportSDoc "tc.rec.proj" 80 $ sep
 	[ text "adding projection"
 	, nest 2 $ prettyTCM projname <+> text (show clause)
 	]
-      reportSDoc "tc.rec.proj" 10 $ sep
+      reportSDoc "tc.rec.proj" 70 $ sep
 	[ text "adding projection"
 	, nest 2 $ prettyTCM projname <+> text (show (clausePats clause)) <+> text "=" <+>
-                     inContext [] (addCtxTel ftel (prettyTCM (clauseBody clause)))
+                     inTopContext (addCtxTel ftel (prettyTCM (clauseBody clause)))
+	]
+      reportSDoc "tc.rec.proj" 10 $ sep
+	[ text "adding projection"
+	, nest 2 $ prettyTCM (QNamed projname clause)
 	]
 
             -- Record patterns should /not/ be translated when the
@@ -347,24 +422,22 @@ checkRecordProjections m r q tel ftel fs = checkProjs EmptyTel ftel fs
             ]
 
       escapeContext (size tel) $ do
-	addConstant projname $ Defn rel projname (killRange finalt) [] [StrictPos] (defaultDisplayForm projname) 0 noCompiledRep
-          $ Function { funClauses        = [clause]
-                     , funCompiled       = cc
+	addConstant projname $
+          (defaultDefn ai projname (killRange finalt)
+            Function { funClauses        = [clause]
+                     , funCompiled       = Just cc
                      , funDelayed        = NotDelayed
                      , funInv            = NotInjective
                      , funAbstr          = ConcreteDef
-{-
-                     , funPolarity       = []
-                     , funArgOccurrences = [StrictPos]
--}
                      , funMutual         = []
-                     , funProjection     = Just (r, size ptel + 1)
-                       -- name of the record type and
-                       -- index of the record argument (in the type), start counting with 1
+                     , funProjection     = Just projection
                      , funStatic         = False
                      , funCopy           = False
                      , funTerminates     = Just True
-                     }
+                     , funExtLam         = Nothing
+                     , funWith           = Nothing
+                     })
+            { defArgOccurrences = [StrictPos] }
         computePolarity projname
 
       recurse
