@@ -1,11 +1,12 @@
-{-# LANGUAGE OverlappingInstances,
-             TypeSynonymInstances,
-             FlexibleInstances,
-             IncoherentInstances,
-             ExistentialQuantification,
-             ScopedTypeVariables,
-             CPP
-             #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE ExistentialQuantification #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE OverlappingInstances #-}
+{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE StandaloneDeriving #-}
+{-# LANGUAGE TypeSynonymInstances #-}
+
 {-# OPTIONS_GHC -O2 #-}
 
 -- | Structure-sharing serialisation of Agda interface files.
@@ -16,10 +17,8 @@
 -- -!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-!-
 
 module Agda.TypeChecking.Serialise
-  ( encode
-  , encodeFile
-  , decode
-  , decodeFile
+  ( encode, encodeFile, encodeInterface
+  , decode, decodeFile, decodeInterface, decodeHashes
   , EmbPrj
   )
   where
@@ -31,12 +30,11 @@ import Control.Monad.Reader
 import Control.Monad.State.Strict
 import Control.Monad.Error
 import Data.Array.IArray
-import Data.Bits (shiftR)
 import Data.Word
-import Data.ByteString.Lazy as L
+import qualified Data.ByteString.Lazy as L
 import Data.Hashable
 import qualified Data.HashTable.IO as H
-import Data.Int (Int32, Int64)
+import Data.Int (Int32)
 import Data.IORef
 import Data.Map (Map)
 import qualified Data.Map as M
@@ -44,6 +42,7 @@ import Data.Set (Set)
 import qualified Data.Set as S
 import qualified Data.Binary as B
 import qualified Data.Binary.Get as B
+import qualified Data.Binary.Put as B
 import qualified Data.List as List
 import Data.Function
 import Data.Typeable
@@ -54,13 +53,11 @@ import qualified Agda.Compiler.Epic.Interface as Epic
 import Agda.Syntax.Common
 import Agda.Syntax.Concrete.Name as C
 import qualified Agda.Syntax.Abstract as A
-import Agda.Syntax.Abstract.Name as A
 import Agda.Syntax.Info
 import Agda.Syntax.Internal as I
 import Agda.Syntax.Scope.Base
 import Agda.Syntax.Position (Position, Range, noRange)
 import qualified Agda.Syntax.Position as P
-import Agda.Syntax.Common
 import Agda.Syntax.Fixity
 import Agda.Syntax.Notation
 import Agda.Syntax.Literal
@@ -69,14 +66,19 @@ import qualified Agda.Interaction.Highlighting.Range   as HR
 import qualified Agda.Interaction.Highlighting.Precise as HP
 import Agda.Interaction.FindFile
 
+import qualified Agda.TypeChecking.Monad.Benchmark as Bench
+import Agda.TypeChecking.Monad.Benchmark (billTo)
+
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.CompiledClause
-import Agda.TypeChecking.Pretty
+-- import Agda.TypeChecking.Pretty
+
+import Agda.Utils.BiMap (BiMap)
+import qualified Agda.Utils.BiMap as BiMap
 import Agda.Utils.FileName
-import Agda.Utils.Monad
-import Agda.Utils.Tuple
 import Agda.Utils.Permutation
 import Agda.Utils.HashMap (HashMap)
+import Agda.Utils.Hash
 import qualified Agda.Utils.HashMap as HMap
 
 #include "../undefined.h"
@@ -87,7 +89,7 @@ import Agda.Utils.Impossible
 -- 32-bit machines). Word64 does not have these problems.
 
 currentInterfaceVersion :: Word64
-currentInterfaceVersion = 20121106 * 10 + 0
+currentInterfaceVersion = 20140611 * 10 + 0
 
 -- | Constructor tag (maybe omitted) and argument indices.
 
@@ -100,8 +102,13 @@ type Node = [Int32]
 -- 'H.LinearHashTable' and the hashtables from "Data.Hashtable" are
 -- much slower.
 
+#if defined(mingw32_HOST_OS) && defined(x86_64_HOST_ARCH)
+type HashTable k v = H.CuckooHashTable k v
+#else
 type HashTable k v = H.BasicHashTable k v
+#endif
 
+-- | State of the the encoder.
 data Dict = Dict{ nodeD     :: !(HashTable Node    Int32)
                 , stringD   :: !(HashTable String  Int32)
                 , integerD  :: !(HashTable Integer Int32)
@@ -115,9 +122,13 @@ data Dict = Dict{ nodeD     :: !(HashTable Node    Int32)
                 , fileMod   :: !SourceToModule
                 }
 
+-- | Universal type, wraps everything.
 data U    = forall a . Typeable a => U !a
+
+-- | Univeral memo structure, to introduce sharing during decoding
 type Memo = HashTable (Int32, TypeRep) U    -- (node index, type rep)
 
+-- | State of the decoder.
 data St = St
   { nodeE     :: !(Array Int32 Node)
   , stringE   :: !(Array Int32 String)
@@ -149,35 +160,67 @@ malformed :: R a
 malformed = throwError $ GenericError "Malformed input."
 
 class Typeable a => EmbPrj a where
-  icode :: a -> S Int32
-  value :: Int32 -> R a
+  icode :: a -> S Int32  -- ^ Serialization.
+  value :: Int32 -> R a  -- ^ Deserialization.
 
 -- | Encodes something. To ensure relocatability file paths in
 -- positions are replaced with module names.
 
-encode :: EmbPrj a => a -> TCM ByteString
+encode :: EmbPrj a => a -> TCM L.ByteString
 encode a = do
     fileMod <- sourceToModule
-    (x, shared, total) <- liftIO $ do
-      newD@(Dict nD sD iD dD _ _ _ _ _ stats _) <- emptyDict fileMod
-      root <- runReaderT (icode a) newD
-      nL <- l nD; sL <- l sD; iL <- l iD; dL <- l dD
-      (shared, total) <- readIORef stats
-      return (B.encode currentInterfaceVersion `L.append`
-              G.compress (B.encode (root, nL, sL, iL, dL)), shared, total)
+    newD@(Dict nD sD iD dD _ _ _ _ _ stats _) <- liftIO $ emptyDict fileMod
+    root <- liftIO $ runReaderT (icode a) newD
+    nL <- benchSort $ l nD
+    sL <- benchSort $ l sD
+    iL <- benchSort $ l iD
+    dL <- benchSort $ l dD
+    (shared, total) <- liftIO $ readIORef stats
+    let x = B.encode currentInterfaceVersion `L.append`
+            G.compress (B.encode (root, nL, sL, iL, dL))
     verboseS "profile.sharing" 10 $ do
       tickN "pointers (reused)" $ fromIntegral shared
       tickN "pointers" $ fromIntegral total
     return x
   where
-  l h = List.map fst . List.sortBy (compare `on` snd) <$> H.toList h
+    l h = List.map fst . List.sortBy (compare `on` snd) <$> H.toList h
+    benchSort = billTo [Bench.Serialization, Bench.Sort] . liftIO
+
+-- encode :: EmbPrj a => a -> TCM L.ByteString
+-- encode a = do
+--     fileMod <- sourceToModule
+--     (x, shared, total) <- liftIO $ do
+--       newD@(Dict nD sD iD dD _ _ _ _ _ stats _) <- emptyDict fileMod
+--       root <- runReaderT (icode a) newD
+--       nL <- l nD; sL <- l sD; iL <- l iD; dL <- l dD
+--       (shared, total) <- readIORef stats
+--       return (B.encode currentInterfaceVersion `L.append`
+--               G.compress (B.encode (root, nL, sL, iL, dL)), shared, total)
+--     verboseS "profile.sharing" 10 $ do
+--       tickN "pointers (reused)" $ fromIntegral shared
+--       tickN "pointers" $ fromIntegral total
+--     return x
+--   where
+--   l h = List.map fst . List.sortBy (compare `on` snd) <$> H.toList h
+
+-- | Data.Binary.runGetState is deprecated in favour of runGetIncremental.
+--   Reimplementing it in terms of the new function. The new Decoder type contains
+--   strict byte strings so we need to be careful not to feed the entire lazy byte
+--   string to the decoder at once.
+runGetState :: B.Get a -> L.ByteString -> B.ByteOffset -> (a, L.ByteString, B.ByteOffset)
+runGetState g s n = feed (B.runGetIncremental g) (L.toChunks s)
+  where
+    feed (B.Done s n' x) ss     = (x, L.fromChunks (s : ss), n + n')
+    feed (B.Fail _ _ err) _     = error err
+    feed (B.Partial f) (s : ss) = feed (f $ Just s) ss
+    feed (B.Partial f) []       = feed (f Nothing) []
 
 -- | Decodes something. The result depends on the include path.
 --
 -- Returns 'Nothing' if the input does not start with the right magic
 -- number or some other decoding error is encountered.
 
-decode :: EmbPrj a => ByteString -> TCM (Maybe a)
+decode :: EmbPrj a => L.ByteString -> TCM (Maybe a)
 decode s = do
   mf   <- stModuleToSource <$> get
   incs <- getIncludeDirs
@@ -188,13 +231,13 @@ decode s = do
 
   (mf, r) <- liftIO $ E.handle (\(E.ErrorCall s) -> noResult s) $ do
 
-    (ver, s, _) <- return $ B.runGetState B.get s 0
+    (ver, s, _) <- return $ runGetState B.get s 0
     if ver /= currentInterfaceVersion
      then noResult "Wrong interface version."
      else do
 
       ((r, nL, sL, iL, dL), s, _) <-
-        return $ B.runGetState B.get (G.decompress s) 0
+        return $ runGetState B.get (G.decompress s) 0
       if s /= L.empty
          -- G.decompress seems to throw away garbage at the end, so
          -- the then branch is possibly dead code.
@@ -214,9 +257,12 @@ decode s = do
   case r of
     Right x   -> return (Just x)
     Left  err -> do
-      reportSDoc "import.iface" 5 $
-        text "Error when decoding interface file:" $+$
-        nest 2 (prettyTCM err)
+      reportSLn "import.iface" 5 $ "Error when decoding interface file"
+      -- Andreas, 2014-06-11 deactivated debug printing
+      -- in order to get rid of dependency of Serialize on TCM.Pretty
+      -- reportSDoc "import.iface" 5 $
+      --   text "Error when decoding interface file:"
+      --   $+$ nest 2 (prettyTCM err)
       return Nothing
 
   where
@@ -224,25 +270,34 @@ decode s = do
 
   noResult s = return (Nothing, Left $ GenericError s)
 
+encodeInterface :: Interface -> TCM L.ByteString
+encodeInterface i = L.append hashes <$> encode i
+  where
+    hashes :: L.ByteString
+    hashes = B.runPut $ B.put (iSourceHash i) >> B.put (iFullHash i)
+
 -- | Encodes something. To ensure relocatability file paths in
 -- positions are replaced with module names.
 
-encodeFile :: EmbPrj a
-           => FilePath
-              -- ^ The encoded data is written to this file.
-           -> a
-              -- ^ Something.
-           -> TCM ()
-encodeFile f x = liftIO . L.writeFile f =<< encode x
+encodeFile :: FilePath -> Interface -> TCM ()
+encodeFile f i = liftIO . L.writeFile f =<< encodeInterface i
 
 -- | Decodes something. The result depends on the include path.
 --
 -- Returns 'Nothing' if the file does not start with the right magic
 -- number or some other decoding error is encountered.
 
-decodeFile :: EmbPrj a => FilePath -> TCM (Maybe a)
-decodeFile f = decode =<< liftIO (L.readFile f)
+decodeInterface :: L.ByteString -> TCM (Maybe Interface)
+decodeInterface s = decode $ L.drop 16 s
 
+decodeHashes :: L.ByteString -> Maybe (Hash, Hash)
+decodeHashes s
+  | L.length s < 16 = Nothing
+  | otherwise       = Just $ B.runGet getH $ L.take 16 s
+  where getH = (,) <$> B.get <*> B.get
+
+decodeFile :: FilePath -> TCM (Maybe Interface)
+decodeFile f = decodeInterface =<< liftIO (L.readFile f)
 
 instance EmbPrj String where
   icode   = icodeX stringD stringC
@@ -252,6 +307,15 @@ instance EmbPrj Integer where
   icode   = icodeX integerD integerC
   value i = (! i) `fmap` gets integerE
 
+instance EmbPrj Word64 where
+  icode i = icode2' (int32 q) (int32 r)
+    where (q, r) = quotRem i (2^32)
+          int32 :: Word64 -> Int32
+          int32 = fromIntegral
+  value = vcase valu where valu [a, b] = return $ n * mod (fromIntegral a) n + mod (fromIntegral b) n
+                           valu _      = malformed
+                           n = 2^32
+
 instance EmbPrj Int32 where
   icode i = return i
   value i = return i
@@ -304,7 +368,8 @@ instance EmbPrj AbsolutePath where
       Just m  -> icode m
       Nothing -> __IMPOSSIBLE__
   value m = do
-    m       <- value m
+    m :: TopLevelModuleName
+            <- value m
     mf      <- modFile  <$> get
     incs    <- includes <$> get
     (r, mf) <- liftIO $ findFile'' incs m mf
@@ -334,6 +399,10 @@ instance EmbPrj a => EmbPrj [a] where
 --                            valu [x, xs] = valu2 (:) x xs
 --                            valu _       = malformed
 
+instance (Ord a, Ord b, EmbPrj a, EmbPrj b) => EmbPrj (BiMap a b) where
+  icode m = icode (BiMap.toList m)
+  value m = BiMap.fromList <$> value m
+
 instance (Ord a, EmbPrj a, EmbPrj b) => EmbPrj (Map a b) where
   icode m = icode (M.toList m)
   value m = M.fromList `fmap` value m
@@ -408,23 +477,34 @@ instance EmbPrj NameSpace where
   value = vcase valu where valu [a, b] = valu2 NameSpace a b
                            valu _      = malformed
 
+instance EmbPrj WhyInScope where
+  icode Defined       = icode0'
+  icode (Opened a b)  = icode2 0 a b
+  icode (Applied a b) = icode2 1 a b
+  value = vcase valu where valu []        = valu0 Defined
+                           valu [0, a, b] = valu2 Opened a b
+                           valu [1, a, b] = valu2 Applied a b
+                           valu _         = malformed
+
 instance EmbPrj AbstractName where
-  icode (AbsName a b) = icode2' a b
-  value = vcase valu where valu [a, b] = valu2 AbsName a b
-                           valu _      = malformed
+  icode (AbsName a b c) = icode3' a b c
+  value = vcase valu where valu [a, b, c] = valu3 AbsName a b c
+                           valu _         = malformed
 
 instance EmbPrj AbstractModule where
-  icode (AbsModule a) = icode a
-  value n = AbsModule `fmap` value n
+  icode (AbsModule a b) = icode2' a b
+  value = vcase valu where valu [a, b] = valu2 AbsModule a b
+                           valu _      = malformed
 
 instance EmbPrj KindOfName where
   icode DefName = icode0'
   icode ConName = icode0 1
   icode FldName = icode0 2
-  icode PatternSynName = icode0 2
+  icode PatternSynName = icode0 3
   value = vcase valu where valu []  = valu0 DefName
                            valu [1] = valu0 ConName
                            valu [2] = valu0 FldName
+                           valu [3] = valu0 PatternSynName
                            valu _   = malformed
 
 instance EmbPrj Agda.Syntax.Fixity.Fixity where
@@ -473,33 +553,40 @@ instance (EmbPrj s, EmbPrj t) => EmbPrj (Named s t) where
   value = vcase valu where valu [a, b] = valu2 Named a b
                            valu _      = malformed
 
+instance EmbPrj a => EmbPrj (Ranged a) where
+  icode (Ranged r x) = icode2' r x
+  value = vcase valu where valu [r, x] = valu2 Ranged r x
+                           valu _      = malformed
+
+-- Only used for pattern synonyms
 instance EmbPrj A.Expr where
   icode (A.Var n)               = icode1 0 n
   icode (A.Def n)               = icode1 1 n
   icode (A.Con ns)              = icode1 2 ns
   icode (A.Lit l)               = icode1 3 l
-  icode (A.QuestionMark _)      = icode0 4
+  icode (A.QuestionMark{})      = icode0 4
   icode (A.Underscore _)        = icode0 5
   icode (A.App _ a b)           = icode2 6 a b
   icode (A.WithApp _ a b)       = icode2 7 a b
   icode (A.Lam  _ a b)          = icode2 8 a b
   icode (A.AbsurdLam _ a)       = icode1 9 a
-  icode (A.ExtendedLam _ _ _ _) = icode0 10
+  icode (A.ExtendedLam _ _ _ _) = __IMPOSSIBLE__
   icode (A.Pi   _ a b)          = icode2 11 a b
   icode (A.Fun  _ a b)          = icode2 12 a b
   icode (A.Set  _ a)            = icode1 13 a
   icode (A.Prop _)              = icode0 14
-  icode (A.Let  _ a b)          = icode2 15 a b
+  icode (A.Let  _ _ _)          = __IMPOSSIBLE__
   icode (A.ETel a)              = icode1 16 a
   icode (A.Rec  _ a)            = icode1 17 a
   icode (A.RecUpdate _ a b)     = icode2 18 a b
   icode (A.ScopedExpr a b)      = icode2 19 a b
   icode (A.QuoteGoal _ a b)     = icode2 20 a b
-  icode (A.Quote _)             = icode0 21
-  icode (A.QuoteTerm _)         = icode0 22
-  icode (A.Unquote _)           = icode0 23
-  icode (A.DontCare a)          = icode1 24 a
-  icode (A.PatternSyn a)        = icode1 25 a
+  icode (A.QuoteContext _ a b)  = icode2 21 a b
+  icode (A.Quote _)             = icode0 22
+  icode (A.QuoteTerm _)         = icode0 23
+  icode (A.Unquote _)           = icode0 24
+  icode (A.DontCare a)          = icode1 25 a
+  icode (A.PatternSyn a)        = icode1 26 a
 
   value = vcase valu
     where
@@ -507,29 +594,27 @@ instance EmbPrj A.Expr where
       valu [1, a]     = valu1 A.Def a
       valu [2, a]     = valu1 A.Con a
       valu [3, a]     = valu1 A.Lit a
-      valu [4]        = valu0 (A.QuestionMark emptyMetaInfo)
+      valu [4]        = valu0 (A.QuestionMark emptyMetaInfo 0)
       valu [5]        = valu0 (A.Underscore emptyMetaInfo)
       valu [6, a, b]  = valu2 (A.App i) a b
       valu [7, a, b]  = valu2 (A.WithApp i) a b
       valu [8, a, b]  = valu2 (A.Lam i) a b
       valu [9, a]     = valu1 (A.AbsurdLam i) a
-      valu [10]       = throwError $ NotSupported
-                            "importing pattern synonym containing extended lambda"
       valu [11, a, b] = valu2 (A.Pi i) a b
       valu [12, a, b] = valu2 (A.Fun i) a b
       valu [13, a]    = valu1 (A.Set i) a
       valu [14]       = valu0 (A.Prop i)
-      valu [15, a, b] = valu2 (A.Let i) a b
       valu [16, a]    = valu1 A.ETel a
       valu [17, a]    = valu1 (A.Rec i) a
       valu [18, a, b] = valu2 (A.RecUpdate i) a b
       valu [19, a, b] = valu2 A.ScopedExpr a b
       valu [20, a, b] = valu2 (A.QuoteGoal i) a b
-      valu [21]       = valu0 (A.Quote i)
-      valu [22]       = valu0 (A.QuoteTerm i)
-      valu [23]       = valu0 (A.Unquote i)
-      valu [24, a]    = valu1 A.DontCare a
-      valu [25, a]    = valu1 A.PatternSyn a
+      valu [21, a, b] = valu2 (A.QuoteContext i) a b
+      valu [22]       = valu0 (A.Quote i)
+      valu [23]       = valu0 (A.QuoteTerm i)
+      valu [24]       = valu0 (A.Unquote i)
+      valu [25, a]    = valu1 A.DontCare a
+      valu [26, a]    = valu1 A.PatternSyn a
       valu _          = malformed
 
       i = ExprRange noRange
@@ -549,7 +634,7 @@ instance EmbPrj A.Pattern where
   value = vcase valu
     where
      valu [0, a]    = valu1 A.VarP a
-     valu [1, a, b] = valu2 (A.ConP i) a b
+     valu [1, a, b] = valu2 (A.ConP (ConPatInfo False i)) a b
      valu [2, a, b] = valu2 (A.DefP i) a b
      valu [3]       = valu0 (A.WildP i)
      valu [4, a, b] = valu2 (A.AsP i) a b
@@ -560,29 +645,15 @@ instance EmbPrj A.Pattern where
      valu [9, a, b] = valu2 (A.PatternSynP i) a b
      valu _         = malformed
 
-     i = PatRange noRange
+     i = patNoRange
 
 instance EmbPrj A.LamBinding where
-  icode (A.DomainFree a b c) = icode3 0 a b c
-  icode (A.DomainFull a)     = icode1 1 a
+  icode (A.DomainFree i e) = icode2 0 i e
+  icode (A.DomainFull a)   = icode1 1 a
 
-  value = vcase valu where valu [0, a, b, c] = valu3 A.DomainFree a b c
-                           valu [1, a]       = valu1 A.DomainFull a
-                           valu _            = malformed
-
-instance EmbPrj A.TypedBindings where
-  icode (A.TypedBindings a b) = icode2' a b
-
-  value = vcase valu where valu [a, b] = valu2 A.TypedBindings a b
-                           valu _      = malformed
-
-instance EmbPrj A.TypedBinding where
-  icode (A.TBind a b c) = icode3 0 a b c
-  icode (A.TNoBind a)   = icode1 1 a
-
-  value = vcase valu where valu [0, a, b, c] = valu3 A.TBind a b c
-                           valu [1, a]       = valu1 A.TNoBind a
-                           valu _            = malformed
+  value = vcase valu where valu [0, i, e] = valu2 A.DomainFree i e
+                           valu [1, a]    = valu1 A.DomainFull a
+                           valu _         = malformed
 
 instance EmbPrj A.LetBinding where
   icode (A.LetBind _ a b c d)  = icode4 0 a b c d
@@ -598,6 +669,26 @@ instance EmbPrj A.LetBinding where
                                  "importing pattern synonym containing let module"
       valu _               = malformed
 
+instance EmbPrj A.TypedBindings where
+  icode (A.TypedBindings a b) = icode2' a b
+
+  value = vcase valu where valu [a, b] = valu2 A.TypedBindings a b
+                           valu _      = malformed
+
+instance EmbPrj A.TypedBinding where
+  icode (A.TBind a b c) = icode3 0 a b c
+  icode (A.TLet a b)    = icode2 1 a b
+
+  value = vcase valu where valu [0, a, b, c] = valu3 A.TBind a b c
+                           valu [1, a, b]    = valu2 A.TLet a b
+                           valu _            = malformed
+
+instance EmbPrj c => EmbPrj (Agda.Syntax.Common.ArgInfo c) where
+  icode (ArgInfo h r cs) = icode3' h r cs
+
+  value = vcase valu where valu [h, r, cs] = valu3 ArgInfo h r cs
+                           valu _          = malformed
+
 instance EmbPrj NameId where
   icode (NameId a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 NameId a b
@@ -629,15 +720,27 @@ instance EmbPrj Permutation where
   value = vcase valu where valu [a, b] = valu2 Perm a b
                            valu _      = malformed
 
-instance (EmbPrj a) => EmbPrj (Agda.Syntax.Common.Arg a) where
-  icode (Arg a b c) = icode3' a b c
-  value = vcase valu where valu [a, b, c] = valu3 Arg a b c
-                           valu _         = malformed
+instance EmbPrj a => EmbPrj (Drop a) where
+  icode (Drop a b) = icode2' a b
+  value = vcase valu where valu [a, b] = valu2 Drop a b
+                           valu _      = malformed
 
-instance (EmbPrj a) => EmbPrj (Agda.Syntax.Common.Dom a) where
-  icode (Dom a b c) = icode3' a b c
-  value = vcase valu where valu [a, b, c] = valu3 Dom a b c
-                           valu _         = malformed
+instance EmbPrj a => EmbPrj (Elim' a) where
+  icode (Apply a) = icode1' a
+  icode (Proj  a) = icode1 0 a
+  value = vcase valu where valu [a]    = valu1 Apply a
+                           valu [0, a] = valu1 Proj a
+                           valu _      = malformed
+
+instance (EmbPrj a, EmbPrj c) => EmbPrj (Agda.Syntax.Common.Arg c a) where
+  icode (Arg i e) = icode2' i e
+  value = vcase valu where valu [i, e] = valu2 Arg i e
+                           valu _      = malformed
+
+instance (EmbPrj a, EmbPrj c) => EmbPrj (Agda.Syntax.Common.Dom c a) where
+  icode (Dom i e) = icode2' i e
+  value = vcase valu where valu [i, e] = valu2 Dom i e
+                           valu _      = malformed
 
 instance EmbPrj Agda.Syntax.Common.Induction where
   icode Inductive   = icode0'
@@ -668,6 +771,11 @@ instance EmbPrj Agda.Syntax.Common.Relevance where
                            valu [4] = valu0 UnusedArg
                            valu _   = malformed
 
+instance EmbPrj I.ConHead where
+  icode (ConHead a b) = icode2' a b
+  value = vcase valu where valu [a, b] = valu2 ConHead a b
+                           valu _      = malformed
+
 instance EmbPrj I.Type where
   icode (El a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 El a b
@@ -780,12 +888,12 @@ instance EmbPrj DisplayTerm where
   icode (DDot     a  ) = icode1 1 a
   icode (DCon     a b) = icode2 2 a b
   icode (DDef     a b) = icode2 3 a b
-  icode (DWithApp a b) = icode2 4 a b
+  icode (DWithApp a b c) = icode3 4 a b c
   value = vcase valu where valu [a]       = valu1 DTerm a
                            valu [1, a]    = valu1 DDot a
                            valu [2, a, b] = valu2 DCon a b
                            valu [3, a, b] = valu2 DDef a b
-                           valu [4, a, b] = valu2 DWithApp a b
+                           valu [4, a, b, c] = valu3 DWithApp a b c
                            valu _         = malformed
 
 instance EmbPrj MutualId where
@@ -797,6 +905,17 @@ instance EmbPrj Definition where
   value = vcase valu where valu [rel, a, b, c, d, e, f, g, h] = valu9 Defn rel a b c d e f g h
                            valu _                             = malformed
 
+instance EmbPrj Projection where
+  icode (Projection a b c d e) = icode5' a b c d e
+  value = vcase valu where valu [a, b, c, d, e] = valu5 Projection a b c d e
+                           valu _               = malformed
+
+instance EmbPrj HaskellExport where
+  icode (HsExport a b) = icode2' a b
+  value = vcase valu where
+    valu [a,b] = valu2 HsExport a b
+    valu _ = malformed
+
 instance EmbPrj HaskellRepresentation where
   icode (HsType a)   = icode1' a
   icode (HsDefn a b) = icode2' a b
@@ -884,25 +1003,25 @@ instance EmbPrj Occurrence where
     valu _   = malformed
 
 instance EmbPrj CompiledRepresentation where
-  icode (CompiledRep a b c) = icode3' a b c
-  value = vcase valu where valu [a, b, c] = valu3 CompiledRep a b c
+  icode (CompiledRep a b c d) = icode4' a b c d
+  value = vcase valu where valu [a, b, c, d] = valu4 CompiledRep a b c d
                            valu _         = malformed
 
 instance EmbPrj Defn where
   icode Axiom                                   = icode0 0
-  icode (Function    a b c d e f g h i j)       = icode10 1 a b c d e f g h i j
-  icode (Datatype    a b c d e f g h)           = icode8 2 a b c d e f g h
+  icode (Function    a b c d e f g h i j k l)   = icode12 1 a b c d e f g h i j k l
+  icode (Datatype    a b c d e f g h i j)       = icode10 2 a b c d e f g h i j
   icode (Record      a b c d e f g h i j k l)   = icode12 3 a b c d e f g h i j k l
   icode (Constructor a b c d e)                 = icode5 4 a b c d e
   icode (Primitive   a b c d)                   = icode4 5 a b c d
   value = vcase valu where
-    valu [0]                                    = valu0 Axiom
-    valu [1, a, b, c, d, e, f, g, h, i, j]      = valu10 Function a b c d e f g h i j
-    valu [2, a, b, c, d, e, f, g, h]            = valu8 Datatype a b c d e f g h
-    valu [3, a, b, c, d, e, f, g, h, i, j, k, l]= valu12 Record  a b c d e f g h i j k l
-    valu [4, a, b, c, d, e]                     = valu5 Constructor a b c d e
-    valu [5, a, b, c, d]                        = valu4 Primitive   a b c d
-    valu _                                      = malformed
+    valu [0]                                     = valu0 Axiom
+    valu [1, a, b, c, d, e, f, g, h, i, j, k, l] = valu12 Function a b c d e f g h i j k l
+    valu [2, a, b, c, d, e, f, g, h, i, j]       = valu10 Datatype a b c d e f g h i j
+    valu [3, a, b, c, d, e, f, g, h, i, j, k, l] = valu12 Record  a b c d e f g h i j k l
+    valu [4, a, b, c, d, e]                      = valu5 Constructor a b c d e
+    valu [5, a, b, c, d]                         = valu4 Primitive   a b c d
+    valu _                                       = malformed
 
 instance EmbPrj a => EmbPrj (WithArity a) where
   icode (WithArity a b) = icode2' a b
@@ -937,12 +1056,12 @@ instance EmbPrj FunctionInverse where
                            valu _   = malformed
 
 instance EmbPrj TermHead where
-  icode SortHead    = icode0'
-  icode PiHead      = icode0 1
-  icode (ConHead a) = icode1 2 a
+  icode SortHead     = icode0'
+  icode PiHead       = icode0 1
+  icode (ConsHead a) = icode1 2 a
   value = vcase valu where valu []     = valu0 SortHead
                            valu [1]    = valu0 PiHead
-                           valu [2, a] = valu1 ConHead a
+                           valu [2, a] = valu1 ConsHead a
                            valu _      = malformed
 
 instance EmbPrj Agda.Syntax.Common.IsAbstract where
@@ -953,9 +1072,9 @@ instance EmbPrj Agda.Syntax.Common.IsAbstract where
                            valu _   = malformed
 
 instance EmbPrj I.Clause where
-  icode (Clause a b c d e) = icode5' a b c d e
-  value = vcase valu where valu [a, b, c, d, e] = valu5 Clause a b c d e
-                           valu _               = malformed
+  icode (Clause a b c d e f) = icode6' a b c d e f
+  value = vcase valu where valu [a, b, c, d, e, f] = valu6 Clause a b c d e f
+                           valu _                  = malformed
 
 instance EmbPrj I.ClauseBody where
   icode (Body   a) = icode1 0 a
@@ -978,10 +1097,12 @@ instance EmbPrj I.Pattern where
   icode (ConP a b c) = icode3' a b c
   icode (LitP a    ) = icode1 2 a
   icode (DotP a    ) = icode1 3 a
+  icode (ProjP a   ) = icode1 4 a
   value = vcase valu where valu [a]       = valu1 VarP a
                            valu [a, b, c] = valu3 ConP a b c
                            valu [2, a]    = valu1 LitP a
                            valu [3, a]    = valu1 DotP a
+                           valu [4, a]    = valu1 ProjP a
                            valu _         = malformed
 
 instance EmbPrj a => EmbPrj (Builtin a) where
@@ -1001,6 +1122,7 @@ instance EmbPrj HP.NameKind where
   icode HP.Postulate       = icode0 6
   icode HP.Primitive       = icode0 7
   icode HP.Record          = icode0 8
+  icode HP.Argument        = icode0 9
 
   value = vcase valu where
     valu []      = valu0 HP.Bound
@@ -1012,6 +1134,7 @@ instance EmbPrj HP.NameKind where
     valu [6]     = valu0 HP.Postulate
     valu [7]     = valu0 HP.Primitive
     valu [8]     = valu0 HP.Record
+    valu [9]     = valu0 HP.Argument
     valu _       = malformed
 
 instance EmbPrj HP.Aspect where
@@ -1097,11 +1220,11 @@ instance EmbPrj HP.CompressedFile where
     valu _   = malformed
 
 instance EmbPrj Interface where
-  icode (Interface a b c d e f g h i j) = icode10' a b c d e f g h i j
+  icode (Interface a b c d e f g h i j k) = icode11' a b c d e f g h i j k
   value = vcase valu
     where
-      valu [a, b, c, d, e, f, g, h, i, j] = valu10 Interface a b c d e f g h i j
-      valu _                              = malformed
+      valu [a, b, c, d, e, f, g, h, i, j, k] = valu11 Interface a b c d e f g h i j k
+      valu _                                 = malformed
 
 -- This is used for the Epic compiler backend
 instance EmbPrj Epic.EInterface where
@@ -1159,15 +1282,22 @@ icodeN :: [Int32] -> S Int32
 icodeN = icodeX nodeD nodeC
 
 {-# INLINE vcase #-}
+-- | @vcase value ix@ decodes thing represented by @ix :: Int32@
+--   via the @valu@ function and stores it in 'nodeMemo'.
+--   If @ix@ is present in 'nodeMemo', @valu@ is not used, but
+--   the thing is read from 'nodeMemo' instead.
 vcase :: forall a . EmbPrj a => (Node -> R a) -> Int32 -> R a
 vcase valu = \ix -> do
     memo <- gets nodeMemo
-    (aTyp, maybeU) <- liftIO $ do
-      let aTyp = typeOf (undefined :: a)
-      maybeU <- H.lookup memo (ix, aTyp)
-      return (aTyp, maybeU)
+    -- compute run-time representation of type a
+    let aTyp = typeOf (undefined :: a)
+    -- to introduce sharing, see if we have seen a thing
+    -- represented by ix before
+    maybeU <- liftIO $ H.lookup memo (ix, aTyp)
     case maybeU of
+      -- yes, we have seen it before, use the version from memo
       Just (U u) -> maybe malformed return (cast u)
+      -- no, it's new, so generate it via valu and insert it into memo
       Nothing    -> do
           v <- valu . (! ix) =<< gets nodeE
           liftIO $ H.insert memo (ix, aTyp) (U v)
