@@ -1,20 +1,28 @@
-{-# LANGUAGE CPP, TypeSynonymInstances, FlexibleInstances, OverlappingInstances,
-    DeriveDataTypeable, DeriveFunctor, StandaloneDeriving #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE OverlappingInstances #-}
+{-# LANGUAGE StandaloneDeriving #-}
+{-# LANGUAGE TupleSections #-}
+{-# LANGUAGE TypeSynonymInstances #-}
+
 module Agda.TypeChecking.Substitute where
 
-import Control.Monad.Identity
-import Control.Monad.Reader
 import Control.Arrow ((***))
 
-import Data.Typeable (Typeable)
-import Data.List hiding (sort)
-import qualified Data.List as List
 import Data.Function
+import Data.Functor
+import Data.List hiding (sort, drop)
+import qualified Data.List as List
 import Data.Map (Map)
-import qualified Data.Map as Map
-import qualified Data.Set as Set
+import Data.Typeable (Typeable)
 
-import Agda.Syntax.Common
+import Debug.Trace (trace)
+
+import Agda.Syntax.Common hiding (Arg, Dom, NamedArg)
+import qualified Agda.Syntax.Common as Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Position
 
@@ -22,44 +30,113 @@ import Agda.TypeChecking.Monad.Base as Base
 import Agda.TypeChecking.Free as Free
 import Agda.TypeChecking.CompiledClause
 
+import Agda.Utils.List
 import Agda.Utils.Monad
-import Agda.Utils.Size
 import Agda.Utils.Permutation
+import Agda.Utils.Size
+import Agda.Utils.Tuple
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
 
+---------------------------------------------------------------------------
+-- * Application
+---------------------------------------------------------------------------
+
 -- | Apply something to a bunch of arguments.
 --   Preserves blocking tags (application can never resolve blocking).
 class Apply t where
-    apply :: t -> Args -> t
+  apply  :: t -> Args -> t
+  applyE :: t -> Elims -> t
+
+  apply t args = applyE t $ map Apply args
+  applyE t es  = apply  t $ map argFromElim es
+    -- precondition: all @es@ are @Apply@s
 
 instance Apply Term where
-    apply m [] = m
-    apply m args@(a:args0) =
-        case m of
-            Var i args'   -> Var i (args' ++ args)
-            Def c args'   -> Def c (args' ++ args)
-            Con c args'   -> Con c (args' ++ args)
-            Lam _ u       -> absApp u (unArg a) `apply` args0
-            MetaV x args' -> MetaV x (args' ++ args)
-            Shared p      -> Shared $ apply p args
-            Lit{}         -> __IMPOSSIBLE__
-            Level{}       -> __IMPOSSIBLE__
-            Pi _ _        -> __IMPOSSIBLE__
-            Sort _        -> __IMPOSSIBLE__
-            DontCare mv   -> DontCare $ mv `apply` args  -- Andreas, 2011-10-02
-              -- need to go under DontCare, since "with" might resurrect irrelevant term
+  applyE m [] = m
+  applyE m es =
+    case m of
+      Var i es'   -> Var i (es' ++ es)
+      Def f es'   -> defApp f es' es  -- remove projection redexes
+      Con c args  -> conApp c args es
+      Lam _ b     ->
+        case es of
+          Apply a : es0 -> absApp b (unArg a) `applyE` es0
+          _             -> __IMPOSSIBLE__
+      MetaV x es' -> MetaV x (es' ++ es)
+      Shared p    -> Shared $ applyE p es
+      Lit{}       -> __IMPOSSIBLE__
+      Level{}     -> __IMPOSSIBLE__
+      Pi _ _      -> __IMPOSSIBLE__
+      Sort _      -> __IMPOSSIBLE__
+      DontCare mv -> dontCare $ mv `applyE` es  -- Andreas, 2011-10-02
+        -- need to go under DontCare, since "with" might resurrect irrelevant term
+
+-- | If $v$ is a record value, @canProject f v@
+--   returns its field @f@.
+canProject :: QName -> Term -> Maybe (Arg Term)
+canProject f v =
+  case ignoreSharing v of
+    (Con (ConHead _ fs) vs) -> do
+      i <- elemIndex f fs
+      mhead (drop i vs)
+    _ -> Nothing
+
+-- | Eliminate a constructed term.
+conApp :: ConHead -> Args -> Elims -> Term
+conApp ch                args []             = Con ch args
+conApp ch                args (Apply a : es) = conApp ch (args ++ [a]) es
+conApp ch@(ConHead c fs) args (Proj f  : es) =
+  let failure = flip trace __IMPOSSIBLE__ $
+        "conApp: constructor " ++ show c ++
+        " with fields " ++ show fs ++
+        " projected by " ++ show f
+      i = maybe failure id            $ elemIndex f fs
+      v = maybe failure argToDontCare $ mhead $ drop i args
+  in  applyE v es
+{-
+      i = maybe failure id    $ elemIndex f $ map unArg fs
+      v = maybe failure unArg $ mhead $ drop i args
+      -- Andreas, 2013-10-20 see Issue543a:
+      -- protect result of irrelevant projection.
+      r = maybe __IMPOSSIBLE__ getRelevance $ mhead $ drop i fs
+      u | Irrelevant <- r = DontCare v
+        | otherwise       = v
+  in  applyE v es
+-}
+
+-- | @defApp f us vs@ applies @Def f us@ to further arguments @vs@,
+--   eliminating top projection redexes.
+--   If @us@ is not empty, we cannot have a projection redex, since
+--   the record argument is the first one.
+defApp :: QName -> Elims -> Elims -> Term
+defApp f [] (Apply a : es) | Just v <- canProject f (unArg a)
+  = argToDontCare v `applyE` es
+defApp f es0 es = Def f $ es0 ++ es
+
+-- protect irrelevant fields (see issue 610)
+argToDontCare :: Common.Arg c Term -> Term
+argToDontCare (Common.Arg ai v)
+  | Irrelevant <- getRelevance ai     = dontCare v
+  | otherwise                         = v
 
 instance Apply Type where
   apply = piApply
+  -- Maybe an @applyE@ instance would be useful here as well.
+  -- A record type could be applied to a projection name
+  -- to yield the field type.
+  -- However, this works only in the monad where we can
+  -- look up the fields of a record type.
 
 instance Apply Sort where
-  apply s [] = s
-  apply s _  = __IMPOSSIBLE__
+  applyE s [] = s
+  applyE s _  = __IMPOSSIBLE__
 
 instance Apply a => Apply (Ptr a) where
-  apply p xs = fmap (`apply` xs) p
+  applyE p xs = fmap (`applyE` xs) p
+
+-- @applyE@ does not make sense for telecopes, definitions, clauses etc.
 
 instance Subst a => Apply (Tele a) where
   apply tel               []       = tel
@@ -67,47 +144,73 @@ instance Subst a => Apply (Tele a) where
   apply (ExtendTel _ tel) (t : ts) = absApp tel (unArg t) `apply` ts
 
 instance Apply Definition where
-  apply (Defn rel x t pol occ df m c d) args = Defn rel x (piApply t args) (apply pol args) (apply occ args) df m c (apply d args)
+  apply (Defn info x t pol occ df m c d) args = Defn info x (piApply t args) (apply pol args) (apply occ args) df m c (apply d args)
 
 instance Apply [Base.Occurrence] where
-  apply occ args = drop (length args) occ
+  apply occ args = List.drop (length args) occ
 
 instance Apply [Polarity] where
-  apply pol args = drop (length args) pol
+  apply pol args = List.drop (length args) pol
+
+instance Apply Projection where
+  apply p args = p
+    { projIndex    = projIndex p - size args
+    , projDropPars = projDropPars p `apply` args
+    }
 
 instance Apply Defn where
   apply d [] = d
   apply d args = case d of
     Axiom{} -> d
     Function{ funClauses = cs, funCompiled = cc, funInv = inv
-            , funProjection = Nothing {-, funArgOccurrences = occ -} } ->
+            , funProjection = Nothing } ->
       d { funClauses    = apply cs args
         , funCompiled   = apply cc args
         , funInv        = apply inv args
---        , funArgOccurrences = drop (length args) occ
         }
+
     Function{ funClauses = cs, funCompiled = cc, funInv = inv
-            , funProjection = Just (r, n) {-, funArgOccurrences = occ -} }
-      | m < n  -> d { funProjection = Just (r, n - m) }
-      | otherwise ->
+            , funProjection = Just p} ->
+      case p `apply` args of
+        p@Projection{ projIndex = n }
+          | n < 0     -> __IMPOSSIBLE__
+          -- case: applied only to parameters
+          | n > 0     -> d { funProjection = Just p }
+          -- case: applied also to record value
+          | otherwise ->
+              d { funClauses        = apply cs args'
+                , funCompiled       = apply cc args'
+                , funInv            = apply inv args'
+                , funProjection     = Nothing -- WAS: Just $ p { projIndex = 0 }
+                }
+              where args' = [last args]  -- the record value
+{-
+    Function{ funClauses = cs, funCompiled = cc, funInv = inv
+            , funProjection = Just p@Projection{ projIndex = n } }
+        -- case: only applying parameters
+      | size args < n -> d { funProjection = Just $ p `apply` args }
+        -- case: apply also to record value
+      | otherwise     ->
         d { funClauses        = apply cs args'
           , funCompiled       = apply cc args'
           , funInv            = apply inv args'
-          , funProjection     = Just (r, 0)
---          , funArgOccurrences = drop 1 occ
+          , funProjection     = Just $ p { projIndex = 0 } -- Nothing ?
           }
-      where args' = [last args]
-            m = size args
-    Datatype{ dataPars = np, dataClause = cl
+      where args' = [last args]  -- the record value
+-}
+    Datatype{ dataPars = np, dataSmallPars = sps, dataNonLinPars = nlps, dataClause = cl
             {-, dataArgOccurrences = occ-} } ->
-      d { dataPars = np - size args, dataClause = apply cl args
---        , dataArgOccurrences = drop (length args) occ
+      d { dataPars = np - size args
+        , dataSmallPars  = apply sps args
+        , dataNonLinPars = apply nlps args
+        , dataClause     = apply cl args
+--        , dataArgOccurrences = List.drop (length args) occ
         }
     Record{ recPars = np, recConType = t, recClause = cl, recTel = tel
           {-, recArgOccurrences = occ-} } ->
       d { recPars = np - size args, recConType = apply t args
         , recClause = apply cl args, recTel = apply tel args
---        , recArgOccurrences = drop (length args) occ
+--        , recArgOccurrences = List.drop (length args) occ
         }
     Constructor{ conPars = np } ->
       d { conPars = np - size args }
@@ -118,15 +221,19 @@ instance Apply PrimFun where
     apply (PrimFun x ar def) args   = PrimFun x (ar - size args) $ \vs -> def (args ++ vs)
 
 instance Apply Clause where
-    apply (Clause r tel perm ps b) args =
-      Clause r (apply tel args) (apply perm args)
-             (drop (size args) ps) (apply b args)
+    apply (Clause r tel perm ps b t) args =
+      Clause r
+             (apply tel args)
+             (apply perm args)
+             (List.drop (size args) ps)
+             (apply b args)
+             (applySubst (parallelS (map unArg args)) t)
 
 instance Apply CompiledClauses where
   apply cc args = case cc of
     Fail     -> Fail
     Done hs t
-      | length hs >= len -> Done (drop len hs)
+      | length hs >= len -> Done (List.drop len hs)
                                  (applySubst
                                     (parallelS $
                                        [ var i | i <- [0..length hs - len - 1]] ++
@@ -140,50 +247,71 @@ instance Apply CompiledClauses where
       len = length args
 
 instance Apply a => Apply (WithArity a) where
-  apply (WithArity n a) args = WithArity n $ apply a args
+  apply  (WithArity n a) args = WithArity n $ apply  a args
+  applyE (WithArity n a) es   = WithArity n $ applyE a es
 
 instance Apply a => Apply (Case a) where
   apply (Branches cs ls m) args =
     Branches (apply cs args) (apply ls args) (apply m args)
+  applyE (Branches cs ls m) es =
+    Branches (applyE cs es) (applyE ls es) (applyE m es)
 
 instance Apply FunctionInverse where
   apply NotInjective  args = NotInjective
   apply (Inverse inv) args = Inverse $ apply inv args
 
 instance Apply ClauseBody where
-    apply  b                 []       = b
-    apply (Bind (Abs   _ b)) (a:args) = subst (unArg a) b `apply` args
-    apply (Bind (NoAbs _ b)) (_:args) = b `apply` args
-    apply (Body v)           args     = Body $ v `apply` args
-    apply  NoBody             _       = NoBody
+  apply  b       []       = b
+  apply (Bind b) (a:args) = absApp b (unArg a) `apply` args
+  apply (Body v) args     = Body $ v `apply` args
+  apply  NoBody   _       = NoBody
+  applyE  b       []             = b
+
+  applyE (Bind b) (Apply a : es) = absApp b (unArg a) `applyE` es
+  applyE (Bind b) (Proj{}  : es) = __IMPOSSIBLE__
+  applyE (Body v) es             = Body $ v `applyE` es
+  applyE  NoBody   _             = NoBody
 
 instance Apply DisplayTerm where
   apply (DTerm v)          args = DTerm $ apply v args
   apply (DDot v)           args = DDot  $ apply v args
   apply (DCon c vs)        args = DCon c $ vs ++ map (fmap DTerm) args
   apply (DDef c vs)        args = DDef c $ vs ++ map (fmap DTerm) args
-  apply (DWithApp v args') args = DWithApp v $ args' ++ args
+  apply (DWithApp v ws args') args = DWithApp v ws $ args' ++ args
 
 instance Apply t => Apply [t] where
-    apply ts args = map (`apply` args) ts
+  apply  ts args = map (`apply` args) ts
+  applyE ts es   = map (`applyE` es) ts
 
 instance Apply t => Apply (Blocked t) where
-    apply b args = fmap (`apply` args) b
+  apply  b args = fmap (`apply` args) b
+  applyE b es   = fmap (`applyE` es) b
 
 instance Apply t => Apply (Maybe t) where
-  apply x args = fmap (`apply` args) x
+  apply  x args = fmap (`apply` args) x
+  applyE x es   = fmap (`applyE` es) x
 
 instance Apply v => Apply (Map k v) where
-  apply x args = fmap (`apply` args) x
+  apply  x args = fmap (`apply` args) x
+  applyE x es   = fmap (`applyE` es) x
 
 instance (Apply a, Apply b) => Apply (a,b) where
-    apply (x,y) args = (apply x args, apply y args)
+  apply  (x,y) args = (apply  x args, apply  y args)
+  applyE (x,y) es   = (applyE x es  , applyE y es  )
 
 instance (Apply a, Apply b, Apply c) => Apply (a,b,c) where
-    apply (x,y,z) args = (apply x args, apply y args, apply z args)
+  apply  (x,y,z) args = (apply  x args, apply  y args, apply  z args)
+  applyE (x,y,z) es   = (applyE x es  , applyE y es  , applyE z es  )
+
+instance DoDrop a => Apply (Drop a) where
+  apply x args = dropMore (size args) x
+
+instance DoDrop a => Abstract (Drop a) where
+  abstract tel x = unDrop (size tel) x
 
 instance Apply Permutation where
   -- The permutation must start with [0..m - 1]
+  -- NB: section (- m) not possible (unary minus), hence (flip (-) m)
   apply (Perm n xs) args = Perm (n - m) $ map (flip (-) m) $ genericDrop m xs
     where
       m = size args
@@ -199,9 +327,14 @@ piApply :: Type -> Args -> Type
 piApply t []                      = t
 piApply (El _ (Pi  _ b)) (a:args) = absApp b (unArg a) `piApply` args
 piApply (El s (Shared p)) args    = piApply (El s $ derefPtr p) args
-piApply _ _                       = __IMPOSSIBLE__
+piApply t args                    =
+  trace ("piApply t = " ++ show t ++ "\n  args = " ++ show args) __IMPOSSIBLE__
 
--- | @(abstract args v) args --> v[args]@.
+---------------------------------------------------------------------------
+-- * Abstraction
+---------------------------------------------------------------------------
+
+-- | @(abstract args v) `apply` args --> v[args]@.
 class Abstract t where
     abstract :: Telescope -> t -> t
 
@@ -220,8 +353,8 @@ instance Abstract Telescope where
   abstract (ExtendTel arg tel') tel = ExtendTel arg $ fmap (`abstract` tel) tel'
 
 instance Abstract Definition where
-  abstract tel (Defn rel x t pol occ df m c d) =
-    Defn rel x (abstract tel t) (abstract tel pol) (abstract tel occ) df m c (abstract tel d)
+  abstract tel (Defn info x t pol occ df m c d) =
+    Defn info x (abstract tel t) (abstract tel pol) (abstract tel occ) df m c (abstract tel d)
 
 instance Abstract [Base.Occurrence] where
   abstract tel []  = []
@@ -231,27 +364,34 @@ instance Abstract [Polarity] where
   abstract tel []  = []
   abstract tel pol = replicate (size tel) Invariant ++ pol -- TODO: check polarity
 
+instance Abstract Projection where
+  abstract tel p = p
+    { projIndex    = size tel + projIndex p
+    , projDropPars = abstract tel $ projDropPars p
+    }
+
 instance Abstract Defn where
   abstract tel d = case d of
     Axiom{} -> d
     Function{ funClauses = cs, funCompiled = cc, funInv = inv
-            , funProjection = Nothing {-, funArgOccurrences = occ-} } ->
-      d { funClauses = abstract tel cs, funCompiled = abstract tel cc
-        , funInv = abstract tel inv
---        , funArgOccurrences = replicate (size tel) Mixed ++ occ -- TODO: check occurrence
+            , funProjection = Nothing  } ->
+      d { funClauses  = abstract tel cs
+        , funCompiled = abstract tel cc
+        , funInv      = abstract tel inv
         }
-    Function{ funClauses = cs, funCompiled = cc, funInv = inv
-            , funProjection = Just (r, n) {-, funArgOccurrences = occ-} } ->
-      d { funProjection = Just (r, n + size tel) }
-    Datatype{ dataPars = np, dataClause = cl {-, dataArgOccurrences = occ-} } ->
-      d { dataPars = np + size tel, dataClause = abstract tel cl
---        , dataArgOccurrences = replicate (size tel) Mixed ++ occ -- TODO: check occurrence
+    Function{ funProjection = Just p } ->
+      d { funProjection = Just $ abstract tel p }
+    Datatype{ dataPars = np, dataSmallPars = sps, dataNonLinPars = nlps, dataClause = cl } ->
+      d { dataPars       = np + size tel
+        , dataSmallPars  = abstract tel sps
+        , dataNonLinPars = abstract tel nlps
+        , dataClause     = abstract tel cl
         }
-    Record{ recPars = np, recConType = t, recClause = cl, recTel = tel'
-          {-, recArgOccurrences = occ-} } ->
-      d { recPars = np + size tel, recConType = abstract tel t
-        , recClause = abstract tel cl, recTel = abstract tel tel'
---        , recArgOccurrences = replicate (size tel) Mixed ++ occ -- TODO: check occurrence
+    Record{ recPars = np, recConType = t, recClause = cl, recTel = tel' } ->
+      d { recPars    = np + size tel
+        , recConType = abstract tel t
+        , recClause  = abstract tel cl
+        , recTel     = abstract tel tel'
         }
     Constructor{ conPars = np } ->
       d { conPars = np + size tel }
@@ -263,9 +403,10 @@ instance Abstract PrimFun where
         where n = size tel
 
 instance Abstract Clause where
-  abstract tel (Clause r tel' perm ps b) =
+  abstract tel (Clause r tel' perm ps b t) =
     Clause r (abstract tel tel') (abstract tel perm)
-           (telVars tel ++ ps) (abstract tel b)
+           (namedTelVars tel ++ ps) (abstract tel b)
+           t -- nothing to do for t, since it lives under the telescope
 
 instance Abstract CompiledClauses where
   abstract tel Fail = Fail
@@ -281,8 +422,13 @@ instance Abstract a => Abstract (Case a) where
     Branches (abstract tel cs) (abstract tel ls) (abstract tel m)
 
 telVars :: Telescope -> [Arg Pattern]
-telVars EmptyTel                    = []
-telVars (ExtendTel (Dom h r a) tel) = Arg h r (VarP $ absName tel) : telVars (unAbs tel)
+telVars = map (fmap namedThing) . namedTelVars
+
+namedTelVars :: Telescope -> [NamedArg Pattern]
+namedTelVars EmptyTel                            = []
+namedTelVars (ExtendTel (Common.Dom info a) tel) =
+  Common.Arg info (namedVarP $ absName tel) :
+  namedTelVars (unAbs tel)
 
 instance Abstract FunctionInverse where
   abstract tel NotInjective  = NotInjective
@@ -304,9 +450,14 @@ instance Abstract v => Abstract (Map k v) where
 abstractArgs :: Abstract a => Args -> a -> a
 abstractArgs args x = abstract tel x
     where
-        tel   = foldr (\(Arg h r x) -> ExtendTel (Dom h r $ sort Prop) . Abs x) EmptyTel
-              $ zipWith (fmap . const) names args
-        names = cycle $ map (:[]) ['a'..'z']
+        tel   = foldr (\(Common.Arg info x) -> ExtendTel (Common.Dom info $ sort Prop) . Abs x)
+                      EmptyTel
+              $ zipWith (<$) names args
+        names = cycle $ map (stringToArgName . (:[])) ['a'..'z']
+
+---------------------------------------------------------------------------
+-- * Explicit substitutions
+---------------------------------------------------------------------------
 
 -- | Substitutions.
 
@@ -404,6 +555,10 @@ lookupS rho i = case rho of
              | otherwise -> raise n $ lookupS rho (i - n)
   EmptyS                 -> __IMPOSSIBLE__
 
+---------------------------------------------------------------------------
+-- * Substitution and raising/shifting/weakening
+---------------------------------------------------------------------------
+
 -- | Apply a substitution.
 
 -- For terms:
@@ -434,17 +589,17 @@ instance Subst Substitution where
 instance Subst Term where
   applySubst IdS t = t
   applySubst rho t    = case t of
-    Var i vs    -> lookupS rho i `apply` applySubst rho vs
+    Var i es    -> lookupS rho i `applyE` applySubst rho es
     Lam h m     -> Lam h $ applySubst rho m
-    Def c vs    -> Def c $ applySubst rho vs
+    Def f es    -> defApp f [] $ applySubst rho es
     Con c vs    -> Con c $ applySubst rho vs
-    MetaV x vs  -> MetaV x $ applySubst rho vs
+    MetaV x es  -> MetaV x $ applySubst rho es
     Lit l       -> Lit l
     Level l     -> levelTm $ applySubst rho l
     Pi a b      -> uncurry Pi $ applySubst rho (a,b)
     Sort s      -> sortTm $ applySubst rho s
     Shared p    -> Shared $ applySubst rho p
-    DontCare mv -> DontCare $ applySubst rho mv
+    DontCare mv -> dontCare $ applySubst rho mv
 
 instance Subst a => Subst (Ptr a) where
   applySubst rho = fmap (applySubst rho)
@@ -468,20 +623,27 @@ instance Subst PlusLevel where
   applySubst rho (Plus n l) = Plus n $ applySubst rho l
 
 instance Subst LevelAtom where
-  applySubst rho      (MetaLevel m vs)   = MetaLevel m    $ applySubst rho vs
-  applySubst rho      (BlockedLevel m v) = BlockedLevel m $ applySubst rho v
-  applySubst rho      (NeutralLevel v)   = UnreducedLevel $ applySubst rho v
-  applySubst rho      (UnreducedLevel v) = UnreducedLevel $ applySubst rho v
+  applySubst rho (MetaLevel m vs)   = MetaLevel m    $ applySubst rho vs
+  applySubst rho (BlockedLevel m v) = BlockedLevel m $ applySubst rho v
+  applySubst rho (NeutralLevel v)   = UnreducedLevel $ applySubst rho v
+  applySubst rho (UnreducedLevel v) = UnreducedLevel $ applySubst rho v
+
+instance Subst Bool where
+  applySubst rho = id
+
+instance Subst Name where
+  applySubst rho = id
 
 instance Subst Pattern where
   applySubst rho p = case p of
-    VarP s       -> VarP s
-    LitP l       -> LitP l
     ConP c mt ps -> ConP c (applySubst rho mt) $ applySubst rho ps
     DotP t       -> DotP $ applySubst rho t
+    VarP s       -> p
+    LitP l       -> p
+    ProjP _      -> p
 
 instance Subst t => Subst (Blocked t) where
-  applySubst rho b      = fmap (applySubst rho) b
+  applySubst rho b = fmap (applySubst rho) b
 
 instance Subst DisplayForm where
   applySubst rho (Display n ps v) =
@@ -489,15 +651,15 @@ instance Subst DisplayForm where
               (applySubst (liftS n rho) v)
 
 instance Subst DisplayTerm where
-  applySubst rho      (DTerm v)        = DTerm $ applySubst rho v
-  applySubst rho      (DDot v)         = DDot  $ applySubst rho v
-  applySubst rho      (DCon c vs)      = DCon c $ applySubst rho vs
-  applySubst rho      (DDef c vs)      = DDef c $ applySubst rho vs
-  applySubst rho      (DWithApp vs ws) = uncurry DWithApp $ applySubst rho (vs, ws)
+  applySubst rho (DTerm v)        = DTerm $ applySubst rho v
+  applySubst rho (DDot v)         = DDot  $ applySubst rho v
+  applySubst rho (DCon c vs)      = DCon c $ applySubst rho vs
+  applySubst rho (DDef c vs)      = DDef c $ applySubst rho vs
+  applySubst rho (DWithApp v vs ws) = uncurry3 DWithApp $ applySubst rho (v, vs, ws)
 
 instance Subst a => Subst (Tele a) where
-  applySubst rho  EmptyTel              = EmptyTel
-  applySubst rho (ExtendTel t tel)      = uncurry ExtendTel $ applySubst rho (t, tel)
+  applySubst rho  EmptyTel         = EmptyTel
+  applySubst rho (ExtendTel t tel) = uncurry ExtendTel $ applySubst rho (t, tel)
 
 instance Subst Constraint where
   applySubst rho c = case c of
@@ -526,6 +688,9 @@ instance Subst a => Subst (Abs a) where
 instance Subst a => Subst (Arg a) where
   applySubst rho = fmap (applySubst rho)
 
+instance Subst a => Subst (Named name a) where
+  applySubst rho = fmap (applySubst rho)
+
 instance Subst a => Subst (Dom a) where
   applySubst rho = fmap (applySubst rho)
 
@@ -541,22 +706,49 @@ instance Subst () where
 instance (Subst a, Subst b) => Subst (a,b) where
   applySubst rho (x,y) = (applySubst rho x, applySubst rho y)
 
+instance (Subst a, Subst b, Subst c) => Subst (a,b,c) where
+  applySubst rho (x,y,z) = (applySubst rho x, applySubst rho y, applySubst rho z)
+
+instance (Subst a, Subst b, Subst c, Subst d) => Subst (a,b,c,d) where
+  applySubst rho (x,y,z,u) = (applySubst rho x, applySubst rho y, applySubst rho z, applySubst rho u)
+
 instance Subst ClauseBody where
   applySubst rho (Body t) = Body $ applySubst rho t
   applySubst rho (Bind b) = Bind $ applySubst rho b
   applySubst _   NoBody   = NoBody
 
-data TelV a = TelV (Tele (Dom a)) a
-  deriving (Typeable, Show, Eq, Ord, Functor)
+---------------------------------------------------------------------------
+-- * Telescopes
+---------------------------------------------------------------------------
 
 type TelView = TelV Type
+data TelV a  = TelV { theTel :: Tele (Dom a), theCore :: a }
+  deriving (Typeable, Show, Eq, Ord, Functor)
 
-telFromList :: [Dom (String, Type)] -> Telescope
-telFromList = foldr (\(Dom h r (x, a)) -> ExtendTel (Dom h r a) . Abs x) EmptyTel
+type ListTel' a = [Dom (a, Type)]
+type ListTel = ListTel' ArgName
 
-telToList :: Telescope -> [Dom (String, Type)]
-telToList EmptyTel = []
-telToList (ExtendTel arg tel) = fmap ((,) $ absName tel) arg : telToList (absBody tel)
+telFromList' :: (a -> ArgName) -> ListTel' a -> Telescope
+telFromList' f = foldr extTel EmptyTel
+  where
+    extTel (Common.Dom info (x, a)) = ExtendTel (Common.Dom info a) . Abs (f x)
+
+telFromList :: ListTel -> Telescope
+telFromList = telFromList' id
+
+telToList :: Telescope -> ListTel
+telToList EmptyTel            = []
+telToList (ExtendTel arg tel) = fmap (absName tel,) arg : telToList (absBody tel)
+  -- Andreas, 2013-12-14: This would work also for 'NoAbs',
+  -- since 'absBody' raises.
+
+-- | Turn a typed binding @(x1 .. xn : A)@ into a telescope.
+bindsToTel' :: (Name -> a) -> [Name] -> Dom Type -> ListTel' a
+bindsToTel' f []     t = []
+bindsToTel' f (x:xs) t = fmap (f x,) t : bindsToTel' f xs (raise 1 t)
+
+bindsToTel :: [Name] -> Dom Type -> ListTel
+bindsToTel = bindsToTel' nameToArgName
 
 telView' :: Type -> TelView
 telView' t = case ignoreSharing $ unEl t of
@@ -566,22 +758,31 @@ telView' t = case ignoreSharing $ unEl t of
     absV a x (TelV tel t) = TelV (ExtendTel a (Abs x tel)) t
 
 -- | @mkPi dom t = telePi (telFromList [dom]) t@
-mkPi :: Dom (String, Type) -> Type -> Type
-mkPi (Dom h r (x, a)) b = el $ Pi (Dom h r a) (mkAbs x b)
+mkPi :: Dom (ArgName, Type) -> Type -> Type
+mkPi (Common.Dom info (x, a)) b = el $ Pi (Common.Dom info a) (mkAbs x b)
   where
     el = El $ dLub (getSort a) (Abs x (getSort b)) -- dLub checks x freeIn
 
-telePi :: Telescope -> Type -> Type
-telePi  EmptyTel         t = t
-telePi (ExtendTel u tel) t = el $ Pi u (reAbs b)
-  where
-    el = El (dLub s1 s2)
-    b = fmap (flip telePi t) tel
-    s1 = getSort $ unDom u
-    s2 = fmap getSort b
+telePi' :: (Abs Type -> Abs Type) -> Telescope -> Type -> Type
+telePi' reAbs = telePi where
+  telePi EmptyTel          t = t
+  telePi (ExtendTel u tel) t = el $ Pi u $ reAbs b
+    where
+      b  = (`telePi` t) <$> tel
+      s1 = getSort $ unDom u
+      s2 = getSort <$> b
+      el = El $ dLub s1 s2
 
--- | Everything will be a pi.
+-- | Uses free variable analysis to introduce 'noAbs' bindings.
+telePi :: Telescope -> Type -> Type
+telePi = telePi' reAbs
+
+-- | Everything will be a 'Abs'.
 telePi_ :: Telescope -> Type -> Type
+telePi_ = telePi' id
+
+{- OLD
+-- | Everything will be a pi.
 telePi_  EmptyTel        t = t
 telePi_ (ExtendTel u tel) t = el $ Pi u b
   where
@@ -589,10 +790,21 @@ telePi_ (ExtendTel u tel) t = el $ Pi u b
     b  = fmap (flip telePi_ t) tel
     s1 = getSort $ unDom u
     s2 = fmap getSort b
+-}
 
 teleLam :: Telescope -> Term -> Term
 teleLam  EmptyTel	  t = t
-teleLam (ExtendTel u tel) t = Lam (domHiding u) $ flip teleLam t <$> tel
+teleLam (ExtendTel u tel) t = Lam (domInfo u) $ flip teleLam t <$> tel
+
+-- | Performs void ('noAbs') abstraction over telescope.
+class TeleNoAbs a where
+  teleNoAbs :: a -> Term -> Term
+
+instance TeleNoAbs ListTel where
+  teleNoAbs tel t = foldr (\ (Common.Dom ai (x, _)) -> Lam ai . NoAbs x) t tel
+
+instance TeleNoAbs Telescope where
+  teleNoAbs tel = teleNoAbs $ telToList tel
 
 -- | Dependent least upper bound, to assign a level to expressions
 --   like @forall i -> Set i@.
@@ -604,12 +816,15 @@ dLub s1 b@(Abs _ s2) = case occurrence 0 $ freeVars s2 of
   Flexible      -> DLub s1 b
   Irrelevantly  -> DLub s1 b
   NoOccurrence  -> sLub s1 (absApp b __IMPOSSIBLE__)
-  Free.Unused   -> sLub s1 (absApp b __IMPOSSIBLE__)
+--  Free.Unused   -> sLub s1 (absApp b __IMPOSSIBLE__) -- triggers Issue784
+  Free.Unused   -> DLub s1 b
   StronglyRigid -> Inf
   WeaklyRigid   -> Inf
 
--- Functions on abstractions ----------------------------------------------
+---------------------------------------------------------------------------
+-- * Functions on abstractions
 --   and things we couldn't do before we could define 'absBody'
+---------------------------------------------------------------------------
 
 -- | Instantiate an abstraction
 absApp :: Subst t => Abs t -> Term -> t
@@ -620,7 +835,7 @@ absBody :: Subst t => Abs t -> t
 absBody (Abs   _ v) = v
 absBody (NoAbs _ v) = raise 1 v
 
-mkAbs :: (Subst a, Free a) => String -> a -> Abs a
+mkAbs :: (Subst a, Free a) => ArgName -> a -> Abs a
 mkAbs x v | 0 `freeIn` v = Abs x v
           | otherwise    = NoAbs x (raise (-1) v)
 
@@ -650,6 +865,37 @@ underLambdas n cont a v = loop n a v where
     Lam h b -> Lam h $ underAbs (loop $ n-1) a b
     _       -> __IMPOSSIBLE__
 
+-- | @getBody@ returns the properly raised clause 'Body',
+--   and 'Nothing' if 'NoBody'.
+--
+--   @getBodyUnraised@ just grabs the body, without raising the de Bruijn indices.
+--   This is useful if you want to consider the body in context 'clauseTel'.
+class GetBody a where
+  getBody         :: a -> Maybe Term
+  getBodyUnraised :: a -> Maybe Term
+
+instance GetBody ClauseBody where
+  getBody = body 0
+    where
+      -- collect all shiftings and do them in the end in one go
+      body :: Int -> ClauseBody -> Maybe Term
+      body _ NoBody             = Nothing
+      body n (Body v)           = Just $ raise n v
+      body n (Bind (NoAbs _ v)) = body n v
+      body n (Bind (Abs   _ v)) = body (n + 1) v
+
+  getBodyUnraised NoBody   = Nothing
+  getBodyUnraised (Body v) = Just v
+  getBodyUnraised (Bind b) = getBodyUnraised $ unAbs b  -- Does not raise!
+
+instance GetBody Clause where
+  getBody         = getBody         . clauseBody
+  getBodyUnraised = getBodyUnraised . clauseBody
+
+---------------------------------------------------------------------------
+-- * Syntactic equality and order
+---------------------------------------------------------------------------
+
 deriving instance (Subst a, Eq a) => Eq (Tele a)
 deriving instance (Subst a, Ord a) => Ord (Tele a)
 
@@ -736,7 +982,9 @@ instance (Subst a, Ord a) => Ord (Abs a) where
   Abs   _ a `compare` Abs   _ b = a `compare` b
   a         `compare` b         = absBody a `compare` absBody b
 
--- Level stuff ------------------------------------------------------------
+---------------------------------------------------------------------------
+-- * Level stuff
+---------------------------------------------------------------------------
 
 sLub :: Sort -> Sort -> Sort
 sLub s Prop = s
@@ -757,9 +1005,10 @@ levelMax :: [PlusLevel] -> Level
 levelMax as0 = Max $ ns ++ List.sort bs
   where
     as = Prelude.concatMap expand as0
+    -- ns is empty or a singleton
     ns = case [ n | ClosedLevel n <- as, n > 0 ] of
       []  -> []
-      ns  -> [ ClosedLevel n | n <- [Prelude.maximum ns], n > greatestB ]
+      ns  -> [ ClosedLevel n | let n = Prelude.maximum ns, n > greatestB ]
     bs = subsume [ b | b@Plus{} <- as ]
     greatestB | null bs   = 0
               | otherwise = Prelude.maximum [ n | Plus n _ <- bs ]
@@ -820,12 +1069,14 @@ levelTm l =
     Max [Plus 0 l] -> unLevelAtom l
     _              -> Level l
 
-unLevelAtom (MetaLevel x vs)   = MetaV x vs
+unLevelAtom (MetaLevel x es)   = MetaV x es
 unLevelAtom (NeutralLevel v)   = v
 unLevelAtom (UnreducedLevel v) = v
 unLevelAtom (BlockedLevel _ v) = v
 
--- Boring instances ----------------------------------------------------
+---------------------------------------------------------------------------
+-- * Boring instances
+---------------------------------------------------------------------------
 
 instance Sized Substitution where
   size IdS          = 1
