@@ -3,9 +3,11 @@
 module Agda.TypeChecking.Telescope where
 
 import Control.Applicative
+
 import Data.List
 
-import Agda.Syntax.Common
+import Agda.Syntax.Common hiding (Arg, Dom, NamedArg, ArgInfo)
+import qualified Agda.Syntax.Common as Common
 import Agda.Syntax.Internal
 
 import Agda.TypeChecking.Monad
@@ -46,6 +48,9 @@ flattenTel (ExtendTel a tel) = raise (size tel + 1) a : flattenTel (absBody tel)
 
 -- | Order a flattened telescope in the correct dependeny order: Γ ->
 --   Permutation (Γ -> Γ~)
+--
+--   Since @reorderTel tel@ uses free variable analysis of type in @tel@,
+--   the telescope should be 'normalise'd.
 reorderTel :: [Dom Type] -> Maybe Permutation
 reorderTel tel = topoSort comesBefore tel'
   where
@@ -59,7 +64,7 @@ reorderTel_ tel = case reorderTel tel of
 
 -- | Unflatten: turns a flattened telescope into a proper telescope. Must be
 --   properly ordered.
-unflattenTel :: [String] -> [Dom Type] -> Telescope
+unflattenTel :: [ArgName] -> [Dom Type] -> Telescope
 unflattenTel []	  []	        = EmptyTel
 unflattenTel (x : xs) (a : tel) = ExtendTel a' (Abs x tel')
   where
@@ -70,14 +75,14 @@ unflattenTel [] (_ : _) = __IMPOSSIBLE__
 unflattenTel (_ : _) [] = __IMPOSSIBLE__
 
 -- | Get the suggested names from a telescope
-teleNames :: Telescope -> [String]
+teleNames :: Telescope -> [ArgName]
 teleNames = map (fst . unDom) . telToList
 
-teleArgNames :: Telescope -> [Arg String]
+teleArgNames :: Telescope -> [Arg ArgName]
 teleArgNames = map (argFromDom . fmap fst) . telToList
 
 teleArgs :: Telescope -> Args
-teleArgs tel = [ Arg h r (var i) | (i, Dom h r _) <- zip (downFrom $ size l) l ]
+teleArgs tel = [ Common.Arg info (var i) | (i, Common.Dom info _) <- zip (downFrom $ size l) l ]
   where l = telToList tel
 
 -- | A telescope split in two.
