@@ -1,27 +1,28 @@
-{-# LANGUAGE CPP, MultiParamTypeClasses,
-             FunctionalDependencies, UndecidableInstances,
-             TypeSynonymInstances, FlexibleInstances
-  #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE FunctionalDependencies #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE TypeSynonymInstances #-}
+{-# LANGUAGE UndecidableInstances #-}
 
 module Agda.TypeChecking.Test.Generators where
 
 import Control.Applicative
-import Control.Monad
 import Control.Monad.State
 import qualified Data.List as List (sort, nub)
-import Agda.Utils.QuickCheck hiding (Args)
 
 import Agda.Syntax.Position
-import Agda.Syntax.Common
+import Agda.Syntax.Common as Common
 import Agda.Syntax.Literal
 import Agda.Syntax.Fixity
-import Agda.Syntax.Internal
+import Agda.Syntax.Internal as I
 import qualified Agda.Syntax.Concrete.Name as C
+
 import Agda.TypeChecking.Free
 import Agda.TypeChecking.Substitute
-import Agda.Utils.TestHelpers
-import Agda.Utils.Monad
+
 import Agda.Utils.QuickCheck hiding (Args)
+import Agda.Utils.TestHelpers
 import qualified Agda.Utils.VarSet as Set
 
 #include "../../undefined.h"
@@ -30,6 +31,7 @@ import Agda.Utils.Impossible
 data TermConfiguration = TermConf
       { tcDefinedNames	   :: [QName]
       , tcConstructorNames :: [QName]
+      , tcProjectionNames  :: [QName]
       , tcFreeVariables	   :: [Nat]
       , tcLiterals	   :: UseLiterals
       , tcFrequencies	   :: Frequencies
@@ -44,14 +46,16 @@ data TermConfiguration = TermConf
 
 data Frequencies = Freqs
       { hiddenFreqs :: HiddenFreqs
-      , nameFreqs   :: NameFreqs
+      , elimFreqs   :: ElimFreqs
       , sortFreqs   :: SortFreqs
       , termFreqs   :: TermFreqs
       }
   deriving Show
 
 data TermFreqs = TermFreqs
-      { nameFreq :: Int
+      { varFreq :: Int
+      , defFreq :: Int
+      , conFreq :: Int
       , litFreq	 :: Int
       , sortFreq :: Int
       , lamFreq	 :: Int
@@ -60,10 +64,9 @@ data TermFreqs = TermFreqs
       }
   deriving Show
 
-data NameFreqs = NameFreqs
-      { varFreq :: Int
-      , defFreq :: Int
-      , conFreq :: Int
+data ElimFreqs = ElimFreqs
+      { applyFreq :: Int
+      , projFreq  :: Int
       }
   deriving Show
 
@@ -81,8 +84,8 @@ data SortFreqs = SortFreqs
 
 defaultFrequencies :: Frequencies
 defaultFrequencies = Freqs
-      { termFreqs   = TermFreqs	  { nameFreq = 40, litFreq = 1, sortFreq = 2, lamFreq = 10, piFreq = 5, funFreq = 5 }
-      , nameFreqs   = NameFreqs	  { varFreq = 3, defFreq = 1, conFreq = 1 }
+      { termFreqs   = TermFreqs	  { varFreq = 24, defFreq = 8, conFreq = 8, litFreq = 1, sortFreq = 2, lamFreq = 10, piFreq = 5, funFreq = 5 }
+      , elimFreqs   = ElimFreqs   { applyFreq = 9, projFreq = 1 }
       , hiddenFreqs = HiddenFreqs { hiddenFreq = 1, notHiddenFreq = 5 }
       , sortFreqs   = SortFreqs	  { setFreqs = [3, 1], propFreq = 1 }
       }
@@ -128,10 +131,11 @@ extendConf conf = conf { tcFreeVariables = 0 : map (1+) (tcFreeVariables conf) }
 extendWithTelConf :: Telescope -> TermConfiguration -> TermConfiguration
 extendWithTelConf tel conf = foldr (const extendConf) conf (telToList tel)
 
-makeConfiguration :: [String] -> [String] -> [Nat] -> TermConfiguration
-makeConfiguration ds cs vs = TermConf
+makeConfiguration :: [RawName] -> [RawName] -> [RawName] -> [Nat] -> TermConfiguration
+makeConfiguration ds cs ps vs = TermConf
   { tcDefinedNames     = defs
   , tcConstructorNames = cons
+  , tcProjectionNames  = projs
   , tcFreeVariables    = List.sort $ List.nub vs
   , tcFrequencies      = defaultFrequencies
   , tcLiterals	       = noLiterals
@@ -139,8 +143,8 @@ makeConfiguration ds cs vs = TermConf
   , tcIsType	       = False
   }
   where
-    (defs, cons) = flip evalState 0 $
-		   (,) <$> mapM mkName ds <*> mapM mkName cs
+    (defs, cons, projs) = flip evalState 0 $ do
+      (,,) <$> mapM mkName ds <*> mapM mkName cs <*> mapM mkName ps
 
     tick     = do x <- get; put (x + 1); return x
     mkName s = do
@@ -161,7 +165,8 @@ newtype YesType a   = YesType	{ unYesType :: a     }
 newtype NoType  a   = NoType	{ unNoType  :: a     }
 newtype VarName	    = VarName	{ unVarName :: Nat   }
 newtype DefName	    = DefName	{ unDefName :: QName }
-newtype ConName	    = ConName	{ unConName :: QName }
+newtype ConName	    = ConName	{ unConName :: ConHead }
+newtype ProjName    = ProjName  { unProjName :: QName }
 newtype SizedList a = SizedList { unSizedList :: [a] }
 
 fixSize :: TermConfiguration -> Gen a -> Gen a
@@ -191,18 +196,35 @@ instance GenC Hiding where
       HiddenFreqs {hiddenFreq = hideF, notHiddenFreq = nohideF } =
 	hiddenFreqs $ tcFrequencies conf
 
-instance GenC a => GenC (Arg a) where
-  genC conf = (\ (h, a) -> Arg h Relevant a) <$> genC conf
+instance (GenC c, GenC a) => GenC (Common.Arg c a) where
+  genC conf = (\ (h, a) -> Arg (setHiding h defaultArgInfo) a) <$> genC conf
 
-instance GenC a => GenC (Dom a) where
-  genC conf = (\ (h, a) -> Dom h Relevant a) <$> genC conf
+instance (GenC c, GenC a) => GenC (Common.Dom c a) where
+  genC conf = (\ (h, a) -> Dom (setHiding h defaultArgInfo) a) <$> genC conf
 
 instance GenC a => GenC (Abs a) where
-  genC conf = Abs "x" <$> genC (extendConf conf)
+  genC conf = Abs x <$> genC (extendConf conf)
+    where x = stringToArgName "x"
+
+instance GenC a => GenC (Elim' a) where
+  genC conf = frequency [ (applyF, Apply <$> genC conf)
+                        , (projF, Proj . unProjName <$> genC conf) ]
+    where
+      ElimFreqs {applyFreq = applyF, projFreq = projF } =
+	elimFreqs $ tcFrequencies conf
+
+instance GenC DefName where
+  genC conf = DefName  <$> do elements $ tcDefinedNames conf
+
+instance GenC ProjName where
+  genC conf = ProjName <$> do elements $ tcProjectionNames conf
 
 genArgs :: TermConfiguration -> Gen Args
 genArgs conf = unSizedList <$> genC (isntTypeConf conf)
 
+genElims :: TermConfiguration -> Gen Elims
+genElims conf = unSizedList <$> genC (isntTypeConf conf)
+
 instance GenC Sort where
   genC conf = frequency $
     (propF, return Prop) :
@@ -249,7 +271,9 @@ instance GenC Term where
       Nothing -> sized $ \n -> genC $ fixSizeConf n conf
       Just n | n <= 0    -> genLeaf
 	     | otherwise -> frequency
-	[ (nameF, genName $ genArgs conf)
+	[ (varF, genVar $ genElims conf)
+	, (defF, genDef $ genElims conf)
+	, (conF, genCon $ genArgs conf)
 	, (litF,  Lit <$> genC conf)
 	, (sortF, Sort <$> genC conf)
 	, (lamF,  genLam)
@@ -264,41 +288,36 @@ instance GenC Term where
       useLits = map ($ tcLiterals conf) [ useLitInt, useLitFloat, useLitString, useLitChar ]
 
       varF  | null vars = 0
-	    | otherwise = freq (varFreq . nameFreqs)
+	    | otherwise = freq (varFreq . termFreqs)
       defF  | null defs = 0
-	    | otherwise = freq (defFreq . nameFreqs)
+	    | otherwise = freq (defFreq . termFreqs)
       conF  | null cons || isType = 0
-	    | otherwise	       = freq (conFreq . nameFreqs)
+	    | otherwise	          = freq (conFreq . termFreqs)
       litF  | or useLits && not isType = freq (litFreq . termFreqs)
 	    | otherwise		    = 0
-      nameF | 0 == varF + defF + conF = 0
-	    | otherwise		    = freq (nameFreq . termFreqs)
       lamF  | isType    = 0
 	    | otherwise = freq (lamFreq  . termFreqs)
       sortF = freq (sortFreq . termFreqs)
       piF   = freq (piFreq   . termFreqs)
 
       genLam :: Gen Term
-      genLam = Lam <$> genC conf <*> genC (isntTypeConf $ decrConf conf)
+      genLam = Lam <$> (flip setHiding defaultArgInfo <$> genC conf) <*> genC (isntTypeConf $ decrConf conf)
 
       genPi :: Gen Term
       genPi = uncurry Pi <$> genC conf
 
-      genVar, genDef, genCon :: Gen Args -> Gen Term
+      genVar, genDef :: Gen Elims -> Gen Term
       genVar args = Var <$> elements vars <*> args
       genDef args = Def <$> elements defs <*> args
-      genCon args = Con <$> elements cons <*> args
 
-      genName :: Gen Args -> Gen Term
-      genName args = frequency
-	[ (varF, genVar args)
-	, (defF, genDef args)
-	, (conF, genCon args)
-	]
+      genCon :: Gen Args -> Gen Term
+      genCon args = Con <$> (flip ConHead [] <$> elements cons) <*> args
 
       genLeaf :: Gen Term
       genLeaf = frequency
-	[ (nameF, genName $ return [])
+	[ (varF, genVar $ return [])
+	, (defF, genDef $ return [])
+	, (conF, genCon $ return [])
 	, (litF,  Lit  <$> genC conf)
 	, (sortF, Sort <$> genC conf)
 	]
@@ -308,11 +327,13 @@ genConf :: Gen TermConfiguration
 genConf = do
   ds <- listOf $ elements defs
   cs <- listOf $ elements cons
+  ps <- listOf1 $ elements projs
   vs <- listOf natural
-  return $ makeConfiguration ds cs vs
+  return $ makeConfiguration ds cs ps vs
   where
-    defs = [ [c] | c <- ['a'..'z'] ]
-    cons = [ [c] | c <- ['A'..'Z'] ]
+    defs = [ stringToRawName [c] | c <- ['a'..'n'] ++ ['r'..'z'] ]
+    cons = [ stringToRawName [c] | c <- ['A'..'Z'] ]
+    projs= [ stringToRawName [c] | c <- ['o'..'q'] ]
 
 instance Arbitrary TermConfiguration where
   arbitrary   = genConf
@@ -373,8 +394,8 @@ instance ShrinkC DefName QName where
   shrinkC conf (DefName c) = takeWhile (/= c) $ tcDefinedNames conf
   noShrink = unDefName
 
-instance ShrinkC ConName QName where
-  shrinkC conf (ConName c) = takeWhile (/= c) $ tcConstructorNames conf
+instance ShrinkC ConName ConHead where
+  shrinkC conf (ConName (ConHead{conName = c})) = map (flip ConHead []) $ takeWhile (/= c) $ tcConstructorNames conf
   noShrink = unConName
 
 instance ShrinkC Literal Literal where
@@ -403,12 +424,12 @@ instance ShrinkC a b => ShrinkC (Abs a) (Abs b) where
   shrinkC conf (Abs   s x) = Abs s <$> shrinkC (extendConf conf) x
   noShrink = fmap noShrink
 
-instance ShrinkC a b => ShrinkC (Arg a) (Arg b) where
-  shrinkC conf (Arg h r x) = (\ (h,x) -> Arg h r x) <$> shrinkC conf (h, x)
+instance ShrinkC a b => ShrinkC (I.Arg a) (I.Arg b) where
+  shrinkC conf (Arg info x) = (\ (h,x) -> Arg (setHiding h info) x) <$> shrinkC conf (argInfoHiding info, x)
   noShrink = fmap noShrink
 
-instance ShrinkC a b => ShrinkC (Dom a) (Dom b) where
-  shrinkC conf (Dom h r x) = (\ (h,x) -> Dom h r x) <$> shrinkC conf (h, x)
+instance ShrinkC a b => ShrinkC (I.Dom a) (I.Dom b) where
+  shrinkC conf (Dom info x) = (\ (h,x) -> Dom (setHiding h info) x) <$> shrinkC conf (argInfoHiding info, x)
   noShrink = fmap noShrink
 
 instance ShrinkC a b => ShrinkC (Blocked a) (Blocked b) where
@@ -416,6 +437,11 @@ instance ShrinkC a b => ShrinkC (Blocked a) (Blocked b) where
   shrinkC conf (NotBlocked x) = NotBlocked <$> shrinkC conf x
   noShrink = fmap noShrink
 
+instance ShrinkC a b => ShrinkC (Elim' a) (Elim' b) where
+  shrinkC conf (Apply a) = Apply <$> shrinkC conf a
+  shrinkC conf (Proj  p) = []
+  noShrink = fmap noShrink
+
 -- Andreas 2010-09-21: simplify? since Sort Prop is no longer abused as DontCare
 instance ShrinkC Sort Sort where
   shrinkC conf Prop = []
@@ -440,19 +466,21 @@ instance ShrinkC Term Term where
   shrinkC conf (DontCare _)  = []
   shrinkC conf (Sort Prop) = []
   shrinkC conf t	   = filter validType $ case ignoreSharing t of
-    Var i args   -> map unArg args ++
-		    (uncurry Var <$> shrinkC conf (VarName i, NoType args))
-    Def d args   -> map unArg args ++
-		    (uncurry Def <$> shrinkC conf (DefName d, NoType args))
-    Con d args   -> map unArg args ++
-		    (uncurry Con <$> shrinkC conf (ConName d, NoType args))
+    Var i es     -> map unArg (argsFromElims es) ++
+		    (uncurry Var <$> shrinkC conf (VarName i, NoType es))
+    Def d es     -> map unArg (argsFromElims es) ++
+		    (uncurry Def <$> shrinkC conf (DefName d, NoType es))
+    Con c args   -> map unArg args ++
+		    (uncurry Con <$> shrinkC conf (ConName c, NoType args))
     Lit l	 -> Lit <$> shrinkC conf l
     Level l      -> [] -- TODO
-    Lam h b      -> killAbs b : (uncurry Lam <$> shrinkC conf (h, b))
+    Lam info b   -> killAbs b : ((\(h,x) -> Lam (setHiding h defaultArgInfo) x)
+                                 <$> shrinkC conf (argInfoHiding info, b))
     Pi a b       -> unEl (unDom a) : unEl (killAbs b) :
 		    (uncurry Pi <$> shrinkC conf (a, b))
     Sort s       -> Sort <$> shrinkC conf s
-    MetaV m args -> map unArg args ++ (MetaV m <$> shrinkC conf (NoType args))
+    MetaV m es   -> map unArg (argsFromElims es) ++
+                    (MetaV m <$> shrinkC conf (NoType es))
     DontCare _   -> []
     Shared{}     -> __IMPOSSIBLE__
     where
@@ -495,10 +523,13 @@ instance KillVar Telescope where
   killVar i EmptyTel	      = EmptyTel
   killVar i (ExtendTel a tel) = uncurry ExtendTel $ killVar i (a, tel)
 
-instance KillVar a => KillVar (Arg a) where
+instance KillVar a => KillVar (Elim' a) where
   killVar i = fmap (killVar i)
 
-instance KillVar a => KillVar (Dom a) where
+instance KillVar a => KillVar (I.Arg a) where
+  killVar i = fmap (killVar i)
+
+instance KillVar a => KillVar (I.Dom a) where
   killVar i = fmap (killVar i)
 
 instance KillVar a => KillVar (Abs a) where
