@@ -5,19 +5,35 @@
 ------------------------------------------------------------------------
 
 module Agda.Utils.Either
-  ( mapEither, mapLeft, mapRight
+  ( whileLeft, caseEitherM
+  , mapEither, mapLeft, mapRight
   , isLeft, isRight
   , allRight
   , tests
   ) where
 
-import Control.Arrow
+import Control.Applicative
+
 import Agda.Utils.QuickCheck
 import Agda.Utils.TestHelpers
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
 
+-- | Loop while we have an exception.
+
+whileLeft :: Monad m => (a -> Either b c) -> (a -> b -> m a) -> (a -> c -> m d) -> a -> m d
+whileLeft test left right = loop where
+  loop a =
+    case test a of
+      Left  b -> loop =<< left a b
+      Right c -> right a c
+
+-- | Monadic version of 'either' with a different argument ordering.
+caseEitherM :: Monad m => m (Either a b) -> (a -> m c) -> (b -> m c) -> m c
+caseEitherM mm f g = either f g =<< mm
+
+
 -- | 'Either' is a bifunctor.
 
 mapEither :: (a -> c) -> (b -> d) -> Either a b -> Either c d
@@ -45,6 +61,36 @@ isLeft :: Either a b -> Bool
 isLeft (Right _) = False
 isLeft (Left _)  = True
 
+-- | Returns @'Just' <input with tags stripped>@ if all elements are
+-- to the right, and otherwise 'Nothing'.
+--
+-- @
+--  allRight xs ==
+--    if all isRight xs then
+--      Just (map (\(Right x) -> x) xs)
+--     else
+--      Nothing
+-- @
+
+allRight :: [Either a b] -> Maybe [b]
+allRight []             = Just []
+allRight (Left  _ : _ ) = Nothing
+allRight (Right b : xs) = (b:) <$> allRight xs
+
+prop_allRight xs =
+  allRight xs ==
+    if all isRight xs then
+      Just (map fromRight xs)
+     else
+      Nothing
+  where
+  fromRight (Right x) = x
+  fromRight (Left _)  = __IMPOSSIBLE__
+
+{- Andreas, 2012-12-01 I do not know why it makes sense to copy
+   the input (only extra work for the garbage collector...
+   So I disable the code below...
+
 -- | Returns @'Right' <input with tags stripped>@ if all elements are
 -- to the right, and otherwise @Left <input>@:
 --
@@ -72,6 +118,7 @@ prop_allRight xs =
   where
   fromRight (Right x) = x
   fromRight (Left _)  = __IMPOSSIBLE__
+-}
 
 ------------------------------------------------------------------------
 -- All tests
