@@ -1,9 +1,11 @@
-{-# LANGUAGE CPP, DeriveDataTypeable #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
 
 {-| Operations on file names. -}
 module Agda.Utils.FileName
   ( AbsolutePath
   , filePath
+  , rootName
   , mkAbsolute
   , absolute
   , (===)
@@ -11,14 +13,7 @@ module Agda.Utils.FileName
   , tests
   ) where
 
-import Agda.Utils.TestHelpers
-import Agda.Utils.QuickCheck
-import Data.Function
-import Data.Typeable (Typeable)
-import Data.List
-import Data.Maybe
 import Control.Applicative
-import Control.Monad
 import System.Directory
 import System.FilePath
 
@@ -27,6 +22,14 @@ import Control.Exception (bracket)
 import System.Win32 (findFirstFile, findClose, getFindDataFileName)
 #endif
 
+import Data.ByteString.Char8 (ByteString)
+import qualified Data.ByteString.Char8 as ByteString
+import Data.Function
+import Data.Typeable (Typeable)
+
+import Agda.Utils.TestHelpers
+import Agda.Utils.QuickCheck
+
 #include "../undefined.h"
 import Agda.Utils.Impossible
 
@@ -34,10 +37,20 @@ import Agda.Utils.Impossible
 --
 -- Note that the 'Eq' and 'Ord' instances do not check if different
 -- paths point to the same files or directories.
+--
+-- Andreas, 2014-03-30:
+-- For efficiency of serialization, 'AbsolutePath' is implemented
+-- as 'ByteString' which short-cuts equality testing using
+-- pointer equality.  This saves 20% of the serialization time
+-- of the standard library!
 
-newtype AbsolutePath = AbsolutePath { filePath :: FilePath }
+newtype AbsolutePath = AbsolutePath { byteStringPath :: ByteString }
   deriving (Eq, Ord, Typeable)
 
+-- | Extract the 'AbsolutePath' to be used as 'FilePath'.
+filePath :: AbsolutePath -> FilePath
+filePath = ByteString.unpack . byteStringPath
+
 instance Show AbsolutePath where
   show = filePath
 
@@ -45,11 +58,12 @@ instance Show AbsolutePath where
 -- trailing path separators.
 
 absolutePathInvariant :: AbsolutePath -> Bool
-absolutePathInvariant (AbsolutePath f) =
+absolutePathInvariant x =
   isAbsolute f &&
   isValid f &&
   f == normalise f &&
   f == dropTrailingPathSeparator f
+  where f = filePath x
 
 -- | Constructs 'AbsolutePath's.
 --
@@ -58,7 +72,7 @@ absolutePathInvariant (AbsolutePath f) =
 mkAbsolute :: FilePath -> AbsolutePath
 mkAbsolute f
   | isAbsolute f =
-      AbsolutePath $ dropTrailingPathSeparator $ normalise f
+      AbsolutePath $ ByteString.pack $ dropTrailingPathSeparator $ normalise f
   | otherwise    = __IMPOSSIBLE__
 
 prop_mkAbsolute f =
@@ -75,6 +89,10 @@ rootPath = joinDrive "C:" [pathSeparator]
 rootPath = [pathSeparator]
 #endif
 
+-- | maps @/bla/bla/bla/foo.bar.xxx@ to @foo.bar@.
+rootName :: AbsolutePath -> String
+rootName = dropExtension . snd . splitFileName . filePath
+
 -- | Makes the path absolute.
 --
 -- This function may raise an @\_\_IMPOSSIBLE\_\_@ error if
