@@ -1,19 +1,26 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE TemplateHaskell #-}
 
 {-| Utitlity functions on lists.
 -}
 module Agda.Utils.List where
 
-import Agda.Utils.TestHelpers
-import Agda.Utils.QuickCheck
-import Agda.Utils.Tuple
-
-import Text.Show.Functions
+import Data.Functor ((<$>))
 import Data.Function
 import Data.List
 import Data.Maybe
 import qualified Data.Set as Set
 
+import Text.Show.Functions ()
+import Test.QuickCheck
+import Test.QuickCheck.All
+
+import Agda.Utils.TestHelpers
+-- import Agda.Utils.QuickCheck -- Andreas, 2014-04-27 Inconvenient
+-- because cabal-only CPP directive
+import Agda.Utils.Tuple
+
 #include "../undefined.h"
 import Agda.Utils.Impossible
 
@@ -27,6 +34,25 @@ uncons :: [a] -> Maybe (a, [a])
 uncons []     = Nothing
 uncons (x:xs) = Just (x,xs)
 
+-- | Maybe cons.   @mcons ma as = maybeToList ma ++ as@
+mcons :: Maybe a -> [a] -> [a]
+mcons ma as = maybe as (:as) ma
+
+-- | 'init' and 'last' in one go, safe.
+initLast :: [a] -> Maybe ([a],a)
+initLast [] = Nothing
+initLast as = Just $ loop as where
+  loop []       = __IMPOSSIBLE__
+  loop [a]      = ([], a)
+  loop (a : as) = mapFst (a:) $ loop as
+
+-- | Lookup function (partially safe).
+(!!!) :: [a] -> Int -> Maybe a
+_        !!! n | n < 0 = __IMPOSSIBLE__
+[]       !!! _         = Nothing
+(x : _)  !!! 0         = Just x
+(_ : xs) !!! n         = xs !!! (n - 1)
+
 -- | downFrom n = [n-1,..1,0]
 downFrom :: Integral a => a -> [a]
 downFrom n | n <= 0     = []
@@ -49,6 +75,31 @@ deal f a ~(bs,cs) = case f a of
   Left  b -> (b:bs, cs)
   Right c -> (bs, c:cs)
 
+-- | A generalized version of @takeWhile@.
+--   (Cf. @mapMaybe@ vs. @filter@).
+takeWhileJust :: (a -> Maybe b) -> [a] -> [b]
+takeWhileJust p = loop
+  where
+    loop (a : as) | Just b <- p a = b : loop as
+    loop _ = []
+
+-- | A generalized version of @span@.
+spanJust :: (a -> Maybe b) -> [a] -> ([b], [a])
+spanJust p = loop
+  where
+    loop (a : as) | Just b <- p a = mapFst (b :) $ loop as
+    loop as                       = ([], as)
+
+-- | Partition a list into 'Nothing's and 'Just's.
+--   @'mapMaybe' f = snd . partitionMaybe f@.
+partitionMaybe :: (a -> Maybe b) -> [a] -> ([a], [b])
+partitionMaybe f = loop
+  where
+    loop []       = ([], [])
+    loop (a : as) = case f a of
+      Nothing -> mapFst (a :) $ loop as
+      Just b  -> mapSnd (b :) $ loop as
+
 -- | Sublist relation.
 isSublistOf :: Eq a => [a] -> [a] -> Bool
 isSublistOf []       ys = True
@@ -69,6 +120,22 @@ maybePrefixMatch (p:pat) (r:rest)
   | p == r    = maybePrefixMatch pat rest
   | otherwise = Nothing
 
+-- | Result of 'preOrSuffix'.
+data PreOrSuffix a
+  = IsPrefix a [a] -- ^ First list is prefix of second.
+  | IsSuffix a [a] -- ^ First list is suffix of second.
+  | IsBothfix      -- ^ The lists are equal.
+  | IsNofix        -- ^ The lists are incomparable.
+
+-- | Compare lists with respect to prefix partial order.
+preOrSuffix :: Eq a => [a] -> [a] -> PreOrSuffix a
+preOrSuffix []     []     = IsBothfix
+preOrSuffix []     (b:bs) = IsPrefix b bs
+preOrSuffix (a:as) []     = IsSuffix a as
+preOrSuffix (a:as) (b:bs)
+  | a == b    = preOrSuffix as bs
+  | otherwise = IsNofix
+
 -- | Split a list into sublists. Generalisation of the prelude function
 --   @words@.
 --
@@ -156,6 +223,13 @@ prop_groupBy' p xs =
 groupOn :: Ord b => (a -> b) -> [a] -> [[a]]
 groupOn f = groupBy ((==) `on` f) . sortBy (compare `on` f)
 
+-- | @splitExactlyAt n xs = Just (ys, zs)@ iff @xs = ys ++ zs@
+--   and @genericLength ys = n@.
+splitExactlyAt :: Integral n => n -> [a] -> Maybe ([a], [a])
+splitExactlyAt 0 xs       = return ([], xs)
+splitExactlyAt n []       = Nothing
+splitExactlyAt n (x : xs) = mapFst (x :) <$> splitExactlyAt (n-1) xs
+
 -- | @'extractNthElement' n xs@ gives the @n@-th element in @xs@
 -- (counting from 0), plus the remaining elements (preserving order).
 
@@ -175,7 +249,7 @@ prop_extractNthElement n xs =
     genericTake n rest ++ [elem] ++ genericDrop n rest == xs
   where (elem, rest) = extractNthElement n xs
 
--- A generalised variant of 'elemIndex'.
+-- | A generalised variant of 'elemIndex'.
 
 genericElemIndex :: (Eq a, Integral i) => a -> [a] -> Maybe i
 genericElemIndex x xs =
@@ -204,12 +278,16 @@ prop_zipWith' f =
     forAll (two $ vector n) $ \(xs, ys) ->
       zipWith' f xs ys == zipWith f xs ys
 
+{- UNUSED; a better type would be
+   zipWithTails :: (a -> b -> c) -> [a] -> [b] -> ([c], Either [a] [b])
+
 -- | Like zipWith, but returns the leftover elements of the input lists.
 zipWithTails :: (a -> b -> c) -> [a] -> [b] -> ([c], [a] , [b])
 zipWithTails f xs       []       = ([], xs, [])
 zipWithTails f []       ys       = ([], [] , ys)
 zipWithTails f (x : xs) (y : ys) = (f x y : zs , as , bs)
   where (zs , as , bs) = zipWithTails f xs ys
+-}
 
 -- | Efficient version of nub that sorts the list first. The tag function is
 --   assumed to be cheap. If it isn't pair up the elements with their tags and
@@ -223,15 +301,22 @@ uniqBy tag =
 prop_uniqBy :: [Integer] -> Bool
 prop_uniqBy xs = sort (nub xs) == uniqBy id xs
 
+-- Hack to make $quickCheckAll work under ghc-7.8
+return []
+
 ------------------------------------------------------------------------
 -- All tests
 
 tests :: IO Bool
-tests = runTests "Agda.Utils.List"
-  [ quickCheck' prop_distinct_fastDistinct
-  , quickCheck' prop_groupBy'
-  , quickCheck' prop_extractNthElement
-  , quickCheck' prop_genericElemIndex
-  , quickCheck' prop_zipWith'
-  , quickCheck' prop_uniqBy
-  ]
+tests = do
+  putStrLn "Agda.Utils.List"
+  $quickCheckAll
+
+-- tests = runTests "Agda.Utils.List"
+--   [ quickCheck' prop_distinct_fastDistinct
+--   , quickCheck' prop_groupBy'
+--   , quickCheck' prop_extractNthElement
+--   , quickCheck' prop_genericElemIndex
+--   , quickCheck' prop_zipWith'
+--   , quickCheck' prop_uniqBy
+--   ]
