@@ -1,28 +1,22 @@
-{-# LANGUAGE CPP, FlexibleContexts #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleContexts #-}
 
 module Agda.Utils.Monad
     ( module Agda.Utils.Monad
-{- Andreas 2012-04-21: I'd like to reexport Control.Monad except
-   patching when and unless, but the hiding syntax is not valid yet
-   (only a proposed language extension)
-
-    , module Control.Monad hiding (when, unless)
--}
+    , module Control.Monad
     , (<$>), (<*>)
+    , (<$)
     )
     where
 
 import Prelude		   hiding (concat)
-import Control.Monad
+import Control.Monad       hiding (mapM, forM)
 import Control.Monad.Error
-import Control.Monad.Reader
 import Control.Monad.State
-import qualified Control.Monad.State.Strict as SS
 import Control.Monad.Writer
 import Control.Applicative
-import Data.Traversable as Trav hiding (sequence)
+import Data.Traversable as Trav hiding (for, sequence)
 import Data.Foldable as Fold
-import Data.Monoid
 import Data.Maybe
 
 import Agda.Utils.List
@@ -40,23 +34,24 @@ when_ b m = when b $ do m >> return ()
 unless_ :: Monad m => Bool -> m a -> m ()
 unless_ b m = unless b $ do m >> return ()
 
-whenJust :: Monad m => Maybe a -> (a -> m ()) -> m ()
-whenJust m k = flip (maybe $ return ()) m k
-
 whenM :: Monad m => m Bool -> m a -> m ()
 whenM c m = c >>= (`when_` m)
 
 unlessM :: Monad m => m Bool -> m a -> m ()
 unlessM c m = c >>= (`unless_` m)
 
-whenJustM :: Monad m => m (Maybe a) -> (a -> m ()) -> m ()
-whenJustM c m = c >>= (`whenJust` m)
+-- whenJust, whenJustM moved to Utils.Maybe
 
+-- | Monadic if-then-else.
 ifM :: Monad m => m Bool -> m a -> m a -> m a
 ifM c m m' =
     do	b <- c
 	if b then m else m'
 
+-- | @ifNotM mc = ifM (not <$> mc)@
+ifNotM :: Monad m => m Bool -> m a -> m a -> m a
+ifNotM c = flip $ ifM c
+
 -- | Lazy monadic conjunction.
 and2M :: Monad m => m Bool -> m Bool -> m Bool
 and2M ma mb = ifM ma mb (return False)
@@ -87,7 +82,8 @@ altM1 f (a : as) = either (const $ altM1 f as) (return . Right) =<< f a
 --
 --     @mapM' f t = foldr mappend mempty <$> mapM f t@
 --
---   that collects results in right-to-left order (effects still right-to-left).
+--   that collects results in right-to-left order
+--   (effects still left-to-right).
 --   It might be preferable for right associative monoids.
 mapM' :: (Foldable t, Monad m, Monoid b) => (a -> m b) -> t a -> m b
 mapM' f = Fold.foldl (\ mb a -> liftM2 mappend mb (f a)) (return mempty)
@@ -112,12 +108,21 @@ thread f (x:xs) ret =
 zipWithM' :: Monad m => (a -> b -> m c) -> [a] -> [b] -> m [c]
 zipWithM' f xs ys = sequence (zipWith' f xs ys)
 
--- | A monadic version of @mapMaybe :: (a -> Maybe b) -> [a] -> [b]@.
+-- | A monadic version of @'mapMaybe' :: (a -> Maybe b) -> [a] -> [b]@.
 mapMaybeM :: (Monad m, Functor m) => (a -> m (Maybe b)) -> [a] -> m [b]
 mapMaybeM f xs = catMaybes <$> Trav.mapM f xs
 
+-- | A monadic version of @'dropWhile' :: (a -> Bool) -> [a] -> [a]@.
+dropWhileM :: (Monad m) => (a -> m Bool) -> [a] -> m [a]
+dropWhileM p []       = return []
+dropWhileM p (x : xs) = ifM (p x) (dropWhileM p xs) (return (x : xs))
+
 -- Error monad ------------------------------------------------------------
 
+-- | To simulate @MaybeT@ by @'ErrorT'@.
+instance Error () where
+  noMsg = ()
+
 -- | Finally for the 'Error' class. Errors in the finally part take
 -- precedence over prior errors.
 
@@ -171,20 +176,6 @@ readM s = case reads s of
 
 -- RETIRED STUFF ----------------------------------------------------------
 
-
-{- Andreas 2012-04-21: <.> is obsolete, it is called <=< in Control.Monad
-infixl 8 <.>
-
-(<.>) :: Monad m => (b -> m c) -> (a -> m b) -> a -> m c
-f <.> g = \x -> f =<< g x
--}
-
-{- RETIRED, Andreas, 2012-04-30.
-   For GHC >= 7, there is now Control.Monad.void.
-forgetM :: Applicative m => m a -> m ()
-forgetM m = const () <$> m
--}
-
 {- RETIRED, Andreas, 2012-04-30. Not used.
 concatMapM :: Applicative m => (a -> m [b]) -> [a] -> m [b]
 concatMapM f xs = concat <$> traverse f xs
@@ -197,21 +188,4 @@ forceM xs = do () <- length xs `seq` return ()
 
 commuteM :: (Traversable f, Applicative m) => f (m a) -> m (f a)
 commuteM = traverse id
-
--- these are just instances of traverse:
-
-fmapM :: (Traversable f, Applicative m) => (a -> m b) -> f a -> m (f b)
-fmapM f = commuteM . fmap f
-
-mapMaybeM :: Applicative m => (a -> m b) -> Maybe a -> m (Maybe b)
-mapMaybeM f = maybe (pure Nothing) (\x -> Just <$> f x)
-
--}
-
-{- UNUSED
-
--- Either -----------------------------------------------------------------
-
-liftEither :: MonadError e m => Either e a -> m a
-liftEither = either throwError return
 -}
