@@ -1,14 +1,34 @@
-{-# LANGUAGE DeriveDataTypeable, CPP #-}
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE DeriveTraversable #-}
+
 module Agda.Utils.Permutation where
 
+import Prelude hiding (drop)
+
+import Data.List hiding (drop)
+import qualified Data.List as List
+import Data.Maybe
+
+import Data.Foldable (Foldable)
+import Data.Traversable (Traversable)
 import Data.Typeable (Typeable)
-import Data.List
+
 import Agda.Utils.Size
-import Agda.Utils.Impossible
+import Agda.Utils.List ((!!!))
 
 #include "../undefined.h"
+import Agda.Utils.Impossible
 
--- | @permute [1,2,0] [x0,x1,x2] = [x1,x2,x0]@
+-- | Partial permutations. Examples:
+--
+--   @permute [1,2,0]   [x0,x1,x2] = [x1,x2,x0]@     (proper permutation).
+--
+--   @permute [1,0]     [x0,x1,x2] = [x1,x0]@        (partial permuation).
+--
+--   @permute [1,0,1,2] [x0,x1,x2] = [x1,x0,x1,x2]@  (not a permutation because not invertible).
 --
 --   Agda typing would be:
 --   @Perm : {m : Nat}(n : Nat) -> Vec (Fin n) m -> Permutation@
@@ -35,18 +55,22 @@ instance Sized Permutation where
 --   Agda typing:
 --   @permute (Perm {m} n is) : Vec A m -> Vec A n@
 permute :: Permutation -> [a] -> [a]
-permute (Perm _ is) xs = map (xs !!!) is
+permute (Perm _ is) xs = map (xs !!!!) is
   where
-    []     !!! _ = __IMPOSSIBLE__
-    (x:xs) !!! 0 = x
-    (x:xs) !!! n = xs !!! (n - 1)
+    xs !!!! n = fromMaybe __IMPOSSIBLE__ (xs !!! n)
 
+-- | Identity permutation.
 idP :: Int -> Permutation
 idP n = Perm n [0..n - 1]
 
+-- | Restrict a permutation to work on @n@ elements, discarding picks @>=n@.
 takeP :: Int -> Permutation -> Permutation
 takeP n (Perm m xs) = Perm n $ filter (< n) xs
 
+-- | Pick the elements that are not picked by the permutation.
+droppedP :: Permutation -> Permutation
+droppedP (Perm n xs) = Perm n $ [0..n-1] \\ xs
+
 -- | @liftP k@ takes a @Perm {m} n@ to a @Perm {m+k} (n+k)@.
 --   Analogous to 'Agda.TypeChecking.Substitution.liftS',
 --   but Permutations operate on de Bruijn LEVELS, not indices.
@@ -129,3 +153,34 @@ topoSort parent xs = fmap (Perm (size xs)) $ topo g
       where
 	xs = [ x | (x, []) <- g ]
 	remove x g = [ (y, filter (/= x) ys) | (y, ys) <- g, x /= y ]
+
+-- * Drop (apply) and undrop (abstract)
+
+-- | Delayed dropping which allows undropping.
+data Drop a = Drop
+  { dropN    :: Int  -- ^ Non-negative number of things to drop.
+  , dropFrom :: a    -- ^ Where to drop from.
+  }
+  deriving (Eq, Ord, Show, Typeable, Functor, Foldable, Traversable)
+
+-- | Things that support delayed dropping.
+class DoDrop a where
+
+  doDrop :: Drop a -> a              -- ^ Perform the dropping.
+
+  dropMore :: Int -> Drop a -> Drop a  -- ^ Drop more.
+  dropMore n (Drop m xs) = Drop (m + n) xs
+
+  unDrop :: Int -> Drop a -> Drop a  -- ^ Pick up dropped stuff.
+  unDrop n (Drop m xs)
+    | n <= m    = Drop (m - n) xs
+    | otherwise = __IMPOSSIBLE__
+
+instance DoDrop [a] where
+  doDrop   (Drop m xs) = List.drop m xs
+
+instance DoDrop Permutation where
+  doDrop (Drop k (Perm n xs)) =
+    Perm (n + m) $ [0..m-1] ++ map (+ m) (List.drop k xs)
+    where m = -k
+  unDrop m = dropMore (-m) -- allow picking up more than dropped
