@@ -5,7 +5,11 @@ module Agda.Utils.Pretty
     , module Text.PrettyPrint
     ) where
 
-import Text.PrettyPrint
+import Data.Function
+import Text.PrettyPrint hiding (TextDetails(Str))
+
+instance Eq Doc where
+  (==) = (==) `on` render
 
 class Pretty a where
     pretty	:: a -> Doc
