@@ -1,12 +1,10 @@
-
 module Agda.Utils.QuickCheck
   ( module Test.QuickCheck
   , module Agda.Utils.QuickCheck
   ) where
 
-import Control.Monad
-import Control.Applicative
-import Test.QuickCheck
+import Test.QuickCheck hiding ((===))
+import Test.QuickCheck.Property (Property(..))
 
 isSuccess :: Result -> Bool
 isSuccess Success{} = True
